
/******************************************************************************
* MODULE     : fft_fixed.hpp
* DESCRIPTION: inlined low-level fft "codelets"
* COPYRIGHT  : (C) 2007  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__FFT_FIXED__HPP
#define __MMX__FFT_FIXED__HPP
#include <algebramix/fft_naive.hpp>

namespace mmx {

/******************************************************************************
* Inlined fft codelet
******************************************************************************/

template<typename D, typename R, nat steps, nat todo> struct fft_large_helper;
template<typename D, typename R, nat steps, nat todo> struct fft_small_helper;

template<typename D, typename R, nat steps>
struct fft_helper {
  static const nat half1= steps >> 1;
  static const nat half2= steps - half1;
  static const nat n1   = 1 << half1;
  static const nat n2   = 1 << half2;
  typedef typename R::C C;
  typedef typename R::U U;

  static inline void
  tf (C* c, nat stride, U* u, nat shift) {
    fft_large_helper<D,R,half1,n2>::tf (c, stride<<half2, u, shift>>half2);
    fft_small_helper<D,R,half2,n1>::tf (c, stride, u, shift);
  }
};

template<typename R, nat steps>
struct fft_helper<fft_inverse,R,steps> {
  static const nat half1= steps >> 1;
  static const nat half2= steps - half1;
  static const nat n1   = 1 << half1;
  static const nat n2   = 1 << half2;
  typedef fft_inverse D;
  typedef typename R::C C;
  typedef typename R::U U;

  static inline void
  tf (C* c, nat stride, U* u, nat shift) {
    fft_small_helper<D,R,half2,n1>::tf (c, stride, u, shift);
    fft_large_helper<D,R,half1,n2>::tf (c, stride<<half2, u, shift>>half2);
  }
};

template<typename D, typename R>
struct fft_helper<D,R,1> {
  typedef typename R::C C;
  typedef typename R::U U;

  static inline void
  tf (C* c, nat stride, U* u, nat shift) {
    cross<R,D> (c, c + stride, u + shift);
  }
};

/******************************************************************************
* Unroll loop for large strides
******************************************************************************/

template<typename D, typename R, nat steps, nat todo>
struct fft_large_helper {
  typedef typename R::C C;
  typedef typename R::U U;

  static inline void
  tf (C* c, nat stride, U* u, nat shift) {
    fft_helper<D,R,steps>::tf (c, stride, u, shift);
    fft_large_helper<D,R,steps,todo-1>::tf (c, stride, u, shift);
  }
};

template<typename D, typename R, nat steps>
struct fft_large_helper<D,R,steps,1> {
  typedef typename R::C C;
  typedef typename R::U U;

  static inline void
  tf (C* c, nat stride, U* u, nat shift) {
    fft_helper<D,R,steps>::tf (c, stride, u, shift);
  }
};

/******************************************************************************
* Unroll loop for small strides
******************************************************************************/

template<typename D, typename R, nat steps, nat todo>
struct fft_small_helper {
  typedef typename R::C C;
  typedef typename R::U U;

  static inline void
  tf (C* c, nat stride, U* u, nat shift) {
    fft_helper<D,R,steps>::tf (c, stride, u, shift);
    fft_small_helper<D,R,steps,todo-1>::tf (c, stride, u, shift + (1<<steps));
  }
};

template<typename D, typename R, nat steps>
struct fft_small_helper<D,R,steps,1> {
  typedef typename R::C C;
  typedef typename R::U U;

  static inline void
  tf (C* c, nat stride, U* u, nat shift) {
    fft_helper<D,R,steps>::tf (c, stride, u, shift);
  }
};

/******************************************************************************
* Code for all sizes
******************************************************************************/

template<typename R, nat steps>
struct fft_fixed_helper {
  typedef typename R::C C;
  typedef typename R::U U;

  template<typename D> static inline void
  transform (nat s, C* c, nat stride, U* u, nat shift) {
    if (s < steps) fft_fixed_helper<R,steps>::
                     template transform<D> (s, c, stride, u, shift);
    else fft_helper<D,R,steps>::tf (c, stride, u, shift);
  }
};

template<typename R>
struct fft_fixed_helper<R,1> {
  typedef typename R::C C;
  typedef typename R::U U;

  template<typename D> static inline void
  transform (nat s, C* c, nat stride, U* u, nat shift) {
    (void) s;
    cross<R,D> (c, c + stride, u + shift);
  }
};

template<typename R>
struct fft_fixed_helper<R,2> {
  typedef typename R::C C;
  typedef typename R::U U;

  template<typename D> static inline void
  transform (nat s, C* c, nat stride, U* u, nat shift) {
    switch (s) {
    case 1: cross<R,D> (c, c + stride, u + shift); break;
    case 2: fft_helper<D,R,2>::tf (c, stride, u, shift); break;
    }
  }
};

template<typename R>
struct fft_fixed_helper<R,3> {
  typedef typename R::C C;
  typedef typename R::U U;

  template<typename D> static inline void
  transform (nat s, C* c, nat stride, U* u, nat shift) {
    switch (s) {
    case 1: cross<R,D> (c, c + stride, u + shift); break;
    case 2: fft_helper<D,R,2>::tf (c, stride, u, shift); break;
    case 3: fft_helper<D,R,3>::tf (c, stride, u, shift); break;
    }
  }
};

template<typename R>
struct fft_fixed_helper<R,4> {
  typedef typename R::C C;
  typedef typename R::U U;

  template<typename D> static inline void
  transform (nat s, C* c, nat stride, U* u, nat shift) {
    switch (s) {
    case 1: cross<R,D> (c, c + stride, u + shift); break;
    case 2: fft_helper<D,R,2>::tf (c, stride, u, shift); break;
    case 3: fft_helper<D,R,3>::tf (c, stride, u, shift); break;
    case 4: fft_helper<D,R,4>::tf (c, stride, u, shift); break;
    }
  }
};

template<typename R>
struct fft_fixed_helper<R,5> {
  typedef typename R::C C;
  typedef typename R::U U;

  template<typename D> static inline void
  transform (nat s, C* c, nat stride, U* u, nat shift) {
    switch (s) {
    case 1: cross<R,D> (c, c + stride, u + shift); break;
    case 2: fft_helper<D,R,2>::tf (c, stride, u, shift); break;
    case 3: fft_helper<D,R,3>::tf (c, stride, u, shift); break;
    case 4: fft_helper<D,R,4>::tf (c, stride, u, shift); break;
    case 5: fft_helper<D,R,5>::tf (c, stride, u, shift); break;
    }
  }
};

template<typename R>
struct fft_fixed_helper<R,6> {
  typedef typename R::C C;
  typedef typename R::U U;

  template<typename D> static inline void
  transform (nat s, C* c, nat stride, U* u, nat shift) {
    switch (s) {
    case 1: cross<R,D> (c, c + stride, u + shift); break;
    case 2: fft_helper<D,R,2>::tf (c, stride, u, shift); break;
    case 3: fft_helper<D,R,3>::tf (c, stride, u, shift); break;
    case 4: fft_helper<D,R,4>::tf (c, stride, u, shift); break;
    case 5: fft_helper<D,R,5>::tf (c, stride, u, shift); break;
    case 6: fft_helper<D,R,6>::tf (c, stride, u, shift); break;
    }
  }
};

} // namespace mmx
#endif //__MMX__FFT_FIXED__HPP
