
/******************************************************************************
* MODULE     : matrix.hpp
* DESCRIPTION: Dense matrices
* COPYRIGHT  : (C) 2004--2007  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_MATRIX_HPP
#define __MMX_MATRIX_HPP
#include <basix/iterator.hpp>
#include <algebramix/matrix_naive.hpp>
#include <basix/vector.hpp>
namespace mmx {
#define Matrix_variant(C) typename matrix_variant_helper<C>::MV
#define TMPL_DEF template<typename C, typename V= Matrix_variant(C)>
#define TMPL template<typename C, typename V>
#define Matrix matrix<C,V>
#define Matrix_rep matrix_rep<C,V>
#define Abs_matrix matrix<Abs_type(C),V>
TMPL class matrix_rep;
TMPL class matrix;
TMPL inline nat cols (const Matrix& m);
TMPL inline nat rows (const Matrix& m);
TMPL inline bool is_a_scalar   (const Matrix& m);
TMPL inline bool is_non_scalar (const Matrix& m);
TMPL inline C* tab (Matrix& m);
TMPL inline const C* tab (const Matrix& m);

/******************************************************************************
* Dense matrices of arbitrary sizes
******************************************************************************/

TMPL_DEF
class matrix_rep REP_STRUCT {
  C*   a;
  nat  nr;
  nat  nc;
  bool scalar_flag;
public:
  inline matrix_rep (C* a2, nat nr2, nat nc2, bool flag):
    a (a2), nr (nr2), nc (nc2), scalar_flag (flag) {}
  inline ~matrix_rep () { mmx_delete<C> (a, nr*nc); }
  friend class Matrix;
  friend nat cols LESSGTR (const Matrix& m);
  friend nat rows LESSGTR (const Matrix& m);
  friend C* tab LESSGTR (Matrix& m);
  friend const C* tab LESSGTR (const Matrix& m);
  friend bool is_a_scalar   LESSGTR (const Matrix& m);
  friend bool is_non_scalar LESSGTR (const Matrix& m);
};

TMPL_DEF
class matrix {
INDIRECT_PROTO_2 (matrix, matrix_rep, C, V)
private:
  static C zero;
public:
  matrix () {
    nat nc = V::def_rows;
    nat nr = V::def_cols;
    C* tab= mmx_new<C> (nr*nc);
    rep= new Matrix_rep (tab, nr, nc, false);
  }

  template<typename T> matrix (const T& c) {
    nat nr= V::init_rows;
    nat nc= V::init_cols;
    C* tab= mmx_new<C> (nr*nc);
    rep= new Matrix_rep (tab, nr, nc, true);
    V::set (rep->a, C(c), nr, nc);
  }

  template<typename T, typename W> matrix (const matrix<T,W>& m) {
    C* t= mmx_new<C> (rows(m)*cols(m));
    rep= new Matrix_rep (t, rows(m), cols(m), is_a_scalar (m));
    V::Vec::cast (rep->a, tab(m), rows(m) * cols(m));
  }

  matrix (const C& x) {
    nat nr= V::init_rows;
    nat nc= V::init_cols;
    C* tab= mmx_new<C> (nr*nc);
    rep= new Matrix_rep (tab, nr, nc, true);
    V::set (rep->a, x, nr, nc);
  }

  matrix (const C& x, nat nr, nat nc) {
    C* tab= mmx_new<C> (nr*nc);
    rep= new Matrix_rep (tab, nr, nc, false);
    V::set (rep->a, x, nr, nc);
  }

  matrix (C* tab, nat nr=V::def_rows, nat nc=V::def_cols) {
    rep= new Matrix_rep (tab, nr, nc, false);
  }

  matrix (C* tab, nat nr, nat nc, bool flag) {
    rep= new Matrix_rep (tab, nr, nc, flag);
  }

  matrix (const iterator<C>& it) {
    nat i, j, nr= V::def_rows, nc= V::def_cols;
    if (nr==0) nr=1;
    C* tab= mmx_new<C> (nr*nc);
    for (i=0; i<nr; i++, ++it)
      for (j=0; j<nc; j++, ++it)
	tab[V::index (i, j, nr, nc)]= *it;
    rep= new Matrix_rep (tab, nc, nr, false);
  }

  inline const C& operator () (nat i, nat j) const {
    VERIFY (is_non_scalar (*this), "non-scalar matrix expected");
    VERIFY (i<rep->nr, "row out of range");
    VERIFY (j<rep->nc, "column out of range");
    return rep->a[V::index (i, j, rows (*this), cols(*this))]; }
  inline C& operator () (nat i, nat j) {
    VERIFY (is_non_scalar (*this), "non-scalar matrix expected");
    VERIFY (i<rep->nr, "row out of range");
    VERIFY (j<rep->nc, "column out of range");
    return rep->a[V::index (i, j, rows (*this), cols(*this))]; }
  inline C scalar () const {
    VERIFY (is_a_scalar (*this), "scalar matrix expected");
    return rep->a[0]; }
  inline C& scalar () {
    VERIFY (is_a_scalar (*this), "scalar matrix expected");
    return rep->a[0]; }
};
INDIRECT_IMPL_2 (matrix, matrix_rep, typename C, C, typename V, V)

TMPL C Matrix::zero= C(0);
TMPL inline nat rows (const Matrix& m) { return m->nr; }
TMPL inline nat cols (const Matrix& m) { return m->nc; }
TMPL inline nat N (const Matrix& m) { return rows (m) * cols (m); }
TMPL inline bool is_a_scalar (const Matrix& m) { return m->scalar_flag; }
TMPL inline bool is_non_scalar (const Matrix& m) { return !m->scalar_flag; }
TMPL inline bool is_square_matrix (const Matrix& m) {
  return rows (m) == cols (m); }
TMPL inline const C& read (const Matrix& m, nat i, nat j) { return m(i,j); }
TMPL inline const C* tab (const Matrix& m) { return m->a; }
TMPL inline C* tab (Matrix& m) {
  VERIFY (is_non_scalar (m), "non-scalar matrix expected");
  m.secure(); return m->a; }
TMPL inline Matrix extend (const Matrix& m, const Matrix& n) {
  VERIFY (is_a_scalar (m), "scalar matrix expected");
  VERIFY (is_non_scalar (n), "non-scalar matrix expected");
  return Matrix (m.scalar(), rows (n), cols (n)); }

STYPE_TO_TYPE(TMPL,scalar_type,Matrix,C);
STYPE_TO_TYPE(TMPL,abs_type,Matrix,Abs_matrix);

TMPL
struct fast_helper<Matrix > {
  typedef Fast_type(C) FC;
  typedef matrix<FC> fast_type;
  typedef Matrix_variant(FC) FV;
  static inline fast_type dd (const Matrix& m) {
    FC* r= mmx_new<FC> (rows (m) * cols (m));
    FV::Vec::template vec_unary<fast_op> (r, tab (m), rows (m) * cols (m));
    return fast_type (r, rows (m), cols (m), is_a_scalar (m)); }
  static inline Matrix uu (const fast_type& m) {
    C* r= mmx_new<C> (rows (m) * cols (m));
    V::Vec::template vec_unary<slow_op> (r, tab (m), rows (m) * cols (m));
    return Matrix (r, rows (m), cols (m), is_a_scalar (m)); }
};

template<typename T, typename F, typename TV, typename FV>
struct as_helper<matrix<T,TV>,matrix<F,FV> > {
  static inline matrix<T,TV>
  cv (const matrix<F,FV>& m) {
    nat n= rows(m) * cols(m);
    const F* sm= tab (m);
    T* c= mmx_new<T> (n);
    for (nat i=0; i<n; i++) c[i]= as<T> (sm[i]);
    return matrix<T,TV> (c, rows (m), cols (m), is_a_scalar (m)); }
};

/******************************************************************************
* Fixed size dense matrices
******************************************************************************/

template<typename V, typename RS, typename CS>
struct matrix_fixed: public V {
  static const nat def_rows = RS::val;
  static const nat def_cols = CS::val;
  static const nat init_rows= RS::val;
  static const nat init_cols= CS::val;
};

#define FMatrix matrix<C,matrix_fixed<V,RS,CS> >
#define FMatrix_rep matrix_rep<C,matrix_fixed<V,RS,CS> >

template<typename C, typename V, typename RS, typename CS>
class FMatrix_rep REP_STRUCT {
  C* a;
public:
  inline matrix_rep (C* a2, nat dummy1, nat dummy2, bool dummy3):
    a (a2) { (void) dummy1; (void) dummy2; (void) dummy3; }
  inline virtual ~matrix_rep () { mmx_delete<C> (a, RS::val * CS::val); }
  friend class FMatrix;
  friend nat cols LESSGTR (const FMatrix& m);
  friend nat rows LESSGTR (const FMatrix& m);
  friend C*  tab  LESSGTR (const FMatrix& m);
  friend bool is_a_scalar   LESSGTR (const Matrix& m);
  friend bool is_non_scalar LESSGTR (const Matrix& m);
};

template<typename C, typename V, typename RS, typename CS> inline nat
cols (const FMatrix& m) { return CS::val; }
template<typename C, typename V, typename RS, typename CS> inline nat
rows (const FMatrix& m) { return RS::val; }
template<typename C, typename V, typename RS, typename CS> inline bool
is_a_scalar (const FMatrix& m) { (void) m; return false; }
template<typename C, typename V, typename RS, typename CS> inline bool
is_non_scalar (const FMatrix& m) { (void) m; return true; }

#undef FMatrix
#undef FMatrix_rep

/******************************************************************************
* Matrix iteration
******************************************************************************/

TMPL_DEF
class matrix_iterator_rep: public iterator_rep<C> {
  const Matrix m;  // the matrix to traverse
  nat i;           // current row
  nat j;           // current column
  nat nc;
protected:
  bool is_busy () { return i < rows(m); }
  void advance () { j++; if (j >= cols(m)) { j=0; i++; } }
  C current () {return m (i, j); }
public:
  matrix_iterator_rep (const Matrix& m2, nat c = V::def_cols):
    m(m2), i(0), j(0), nc(c) {}
};

TMPL iterator<C>
iterate (const Matrix& m) {
  ASSERT (is_non_scalar (m), "non-scalar matrix expected");
  return iterator<C> (new matrix_iterator_rep<C,V> (m));
}

/******************************************************************************
* Hashing and generic output routines for matrices
******************************************************************************/

TMPL syntactic
flatten (const Matrix& m) {
  if (is_a_scalar (m)) return flatten (m.scalar());
  int i, j, nr= rows(m), nc= cols(m);
  vector<syntactic> v;
  for (i=0; i<nr; i++) {
    vector<syntactic> h;
    for (j=0; j<nc; j++)
      h << flatten (m (i, j));
    v << apply (GEN_ROW, h);
  }
  return apply (GEN_SQTUPLE, v);
}

PRINT_SUGAR(TMPL,Matrix)

/******************************************************************************
* Abstract routines on dense matrices
******************************************************************************/

template<typename Op, typename C, typename V> nat
unary_hash (const matrix<C,V>& m) {
  register nat i, j, h= 214365, nr= rows (m), nc= cols (m);
  for (i=0; i<nr; i++)
    for (j=0; j<nc; j++)
      h= (h<<1) ^ (h<<5) ^ (h>>27) ^ Op::op (m (i, j));
  return h;
}

template<typename Op, typename C, typename V>
matrix<Unary_return_type(Op,C),V>
unary_map (const matrix<C,V>& m) {
  typedef Unary_return_type(Op,C) T;
  if (is_a_scalar (m)) return matrix<T,V> (Op::op (m.scalar()));
  nat nrows= rows (m);
  nat ncols= cols (m);
  T* r= mmx_new<T> (nrows*ncols);
  V::Vec::template vec_unary<Op> (r, tab (m), nrows*ncols);
  return matrix<T,V> (r, nrows, ncols);
}

template<typename Op, typename C1, typename C2, typename V>
matrix<Binary_return_type(Op,C1,C2),V>
binary_map (const matrix<C1,V>& m, const matrix<C2,V>& n) {
  typedef Binary_return_type(Op,C1,C2) T;
  if (is_a_scalar (m) || is_a_scalar (n)) {
    if (is_non_scalar (m)) return binary_map<Op> (m, extend (n, m));
    if (is_non_scalar (n)) return binary_map<Op> (extend (m, n), n);
    return matrix<T,V> (Op::op (m.scalar(), n.scalar()));
  }
  nat nrows= rows (m);
  nat ncols= cols (m);
  ASSERT (rows (n) == nrows, "unequal number of rows");
  ASSERT (cols (n) == ncols, "unequal number of columns");
  T* r= mmx_new<T> (nrows*ncols);
  V::Vec::template vec_binary<Op> (r, tab (m), tab (n), nrows*ncols);
  return matrix<T,V> (r, nrows, ncols);
}

template<typename Op, typename C, typename V, typename X>
matrix<Binary_return_type(Op,C,X),V>
binary_map_scalar (const matrix<C,V>& m, const X& x) {
  typedef Binary_return_type(Op,C,X) T;
  if (is_a_scalar (m)) return matrix<T,V> (Op::op (m.scalar(), x));
  nat nrows= rows (m);
  nat ncols= cols (m);
  T* r= mmx_new<T> (nrows*ncols);
  V::Vec::template vec_binary_scalar<Op> (r, tab (m), x, nrows*ncols);
  return matrix<T,V> (r, nrows, ncols);
}

template<typename Op, typename T, typename C, typename V> matrix<T,V>&
unary_update (matrix<T,V>& m, const matrix<C,V>& n) {
  if (is_a_scalar (m) || is_a_scalar (n)) {
    if (is_non_scalar (m))
      return unary_update<Op> (m, extend (n, m));
    else if (is_non_scalar (n))
      m= extend (m, n);
    else {
      Op::in_place (m.scalar(), n.scalar());
      return m;
    }
  }
  nat nrows= rows (m);
  nat ncols= cols (m);
  ASSERT (rows (n) == nrows, "unequal number of rows");
  ASSERT (cols (n) == ncols, "unequal number of columns");
  V::Vec::template vec_unary<Op> (tab (m), tab (n), nrows*ncols);
  return m;
}

template<typename Op, typename T, typename V, typename X> matrix<T,V>&
unary_update_scalar (matrix<T,V>& m, const X& x) {
  if (is_a_scalar (m)) {
    Op::in_place (m.scalar(), x);
    return m;
  }
  nat nrows= rows (m);
  nat ncols= cols (m);
  V::Vec::template vec_unary_scalar<Op> (tab (m), x, nrows*ncols);
  return m;
}

template<typename Op, typename C1, typename C2, typename V> bool
binary_test (const matrix<C1,V>& m, const matrix<C2,V>& n) {
  if (is_a_scalar (m) || is_a_scalar (n)) {
    if (is_non_scalar (m)) return binary_test<Op> (m, extend (n, m));
    if (is_non_scalar (n)) return binary_test<Op> (extend (m, n), n);
    return Op::op (m.scalar(), n.scalar());
  }
  nat nrows= rows (m);
  nat ncols= cols (m);
  if (rows (n) != nrows || cols (n) != ncols) return false;
  return V::Vec::template vec_binary_test<Op> (tab (m), tab (n), nrows*ncols);
}

/******************************************************************************
* Instantiations of abstract routines on matrices
******************************************************************************/

TMPL Matrix copy (const Matrix& m) {
  return unary_map<id_op> (m); }
TMPL Matrix operator - (const Matrix& m) {
  return unary_map<neg_op> (m); }
TMPL Matrix operator + (const Matrix& m, const Matrix& n) {
  return binary_map<add_op> (m, n); }
TMPL Matrix operator + (const Matrix& m, const C& n) {
  return binary_map<add_op> (m, Matrix (n)); }
TMPL Matrix operator + (const C& m, const Matrix& n) {
  return binary_map<add_op> (Matrix (m), n); }
TMPL Matrix operator - (const Matrix& m, const Matrix& n) {
  return binary_map<sub_op> (m, n); }
TMPL Matrix operator - (const Matrix& m, const C& n) {
  return binary_map<sub_op> (m, Matrix (n)); }
TMPL Matrix operator - (const C& m, const Matrix& n) {
  return binary_map<sub_op> (Matrix (m), n); }
TMPL Matrix operator * (const Matrix& m, const C& c) {
  return binary_map_scalar<rmul_op> (m, c); }
TMPL Matrix operator * (const C& c, const Matrix& m) {
  return binary_map_scalar<lmul_op> (m, c); }
TMPL Matrix operator / (const Matrix& m, const C& c) {
  return binary_map_scalar<rdiv_op> (m, c); }
TMPL Matrix quo (const Matrix& m, const C& c) {
  return binary_map_scalar<rquo_op> (m, c); }
TMPL Matrix rem (const Matrix& m, const C& c) {
  return binary_map_scalar<rrem_op> (m, c); }
TMPL Matrix& operator += (Matrix& m, const Matrix& n) {
  return unary_update<add_op> (m, n); }
TMPL Matrix& operator -= (Matrix& m, const Matrix& n) {
  return unary_update<sub_op> (m, n); }
TMPL Matrix& operator *= (Matrix& m, const C& x) {
  return unary_update_scalar<rmul_op> (m, x); }
TMPL Matrix& operator /= (Matrix& m, const C& x) {
  return unary_update_scalar<rdiv_op> (m, x); }
TMPL bool operator <= (const Matrix& m, const Matrix& n) {
  return binary_test<lesseq_op> (m, n); }
TMPL bool operator >= (const Matrix& m, const Matrix& n) {
  return binary_test<gtreq_op> (m, n); }
template<typename C, typename V, typename K> bool
operator == (const Matrix& m, const K& c) {
  return m == Matrix (c, rows (m), cols (m)); }
template<typename C, typename V, typename K> bool
operator != (const Matrix& m, const K& c) {
  return m != Matrix (c, rows (m), cols (m)); }
template<typename C, typename V, typename K> bool
operator <= (const Matrix& m, const C& c) {
  return m <= Matrix (c, rows (m), cols (m)); }
template<typename C, typename V, typename K> bool
operator >= (const Matrix& m, const C& c) {
  return m >= Matrix (c, rows (m), cols (m)); }

TRUE_IDENTITY_OP_SUGAR(TMPL,Matrix)
SOFT_IDENTITY_OP_SUGAR(TMPL,Matrix)
ARITH_INT_SUGAR(TMPL,Matrix,C);
STRICT_COMPARE_SUGAR(TMPL,Matrix)

TMPL Matrix derive (const Matrix& m) {
  return unary_map<derive_op> (m); }
TMPL Matrix integrate (const Matrix& m) {
  return unary_map<integrate_op> (m); }
TMPL Abs_matrix abs (const Matrix& m) {
  return unary_map<abs_op> (m); }

/******************************************************************************
* Explicit matrices
******************************************************************************/

template<typename C> matrix<C>
identity_matrix (const int& n) {
  return matrix<C> (C (1), n, n);
}

template<typename C> matrix<C>
fill_matrix (const C& x, const int& r, const int& c) {
  matrix<C> m (C(0), (nat) r, (nat) c);
  for (nat i=0; i<((nat) r); i++)
    for (nat j=0; j<((nat) c); j++)
      m (i, j)= x;
  return m;
}

template<typename C> matrix<C>
tensor_matrix (const vector<C>& v, const vector<C>& w) {
  ASSERT (is_non_scalar (v), "non-scalar vector expected");
  ASSERT (is_non_scalar (w), "non-scalar vector expected");
  matrix<C> m (C(0), N(v), N(w));
  for (nat i=0; i<N(v); i++)
    for (nat j=0; j<N(w); j++)
      m (i, j)= v[i] * w[j];
  return m;
}

template<typename C> matrix<C>
hilbert_matrix (const int& n) {
  matrix<C> m (C (0), n, n);
  for (nat i=0; i<(nat) n; i++)
    for (nat j=0; j<(nat) n; j++)
      m (i, j)= 1 / C (1 + i + j);
  return m;
}

template<typename C> matrix<C>
vandermonde (const vector<C>& v) {
  ASSERT (is_non_scalar (v), "non-scalar vector expected");
  matrix<C> m (C(0), N(v), N(v));
  for (nat i=0; i<N(v); i++) {
    C p= 1;
    for (nat j=0; j<N(v); j++, p *= v[i])
      m (i, j)= p;
  }
  return m;
}

/******************************************************************************
* Further routines on matrices
******************************************************************************/

TMPL Matrix
horizontal_join (const Matrix& m1, const Matrix& m2) {
  ASSERT (is_non_scalar (m1) || is_non_scalar (m2),
	  "non-scalar matrix expected");
  if (!is_non_scalar (m1))
    return horizontal_join (Matrix (m1.scalar(), rows (m2), rows (m2)), m2);
  if (!is_non_scalar (m2))
    return horizontal_join (m1, Matrix (m2.scalar(), rows (m1), rows (m1)));
  ASSERT (rows (m1) == rows (m2), "unequal number of rows");
  Matrix r (C(0), rows (m1), cols (m1) + cols (m2));
  for (nat i=0; i<rows(m1); i++) {
    for (nat j=0; j<cols(m1); j++)
      r(i,j)= m1(i,j);
    for (nat j=0; j<cols(m2); j++)
      r(i,j+cols(m1))= m2(i,j);
  }
  return r;
}

TMPL Matrix
vertical_join (const Matrix& m1, const Matrix& m2) {
  ASSERT (is_non_scalar (m1) || is_non_scalar (m2),
	  "non-scalar matrix expected");
  if (!is_non_scalar (m1))
    return vertical_join (Matrix (m1.scalar(), cols (m2), cols (m2)), m2);
  if (!is_non_scalar (m2))
    return vertical_join (m1, Matrix (m2.scalar(), cols (m1), cols (m1)));
  ASSERT (cols (m1) == cols (m2), "unequal number of columns");
  Matrix r (C(0), rows (m1) + rows (m2), cols (m1));
  for (nat j=0; j<cols(m1); j++) {
    for (nat i=0; i<rows(m1); i++)
      r(i,j)= m1(i,j);
    for (nat i=0; i<rows(m2); i++)
      r(i+rows(m1),j)= m2(i,j);
  }
  return r;
}

TMPL Matrix
transpose (const Matrix & m) {
  if (is_a_scalar (m)) return m;
  nat nrows= rows (m), ncols= cols (m);
  C* r = mmx_new<C> (nrows*ncols);
  V::transpose (r, tab (m), nrows, ncols);
  return Matrix (r, ncols, nrows);
}

TMPL Matrix
submatrix (const Matrix& m, nat rdeb, nat rfin, nat cdeb, nat cfin) {
  if (is_a_scalar (m)) return Matrix (m.scalar(), rfin-rdeb, cfin-cdeb);
  nat nrows= rows (m), ncols= cols (m);
  C* r = mmx_new<C> ((rfin-rdeb)*(cfin-cdeb));
  V::extract (r, tab (m), nrows, rdeb, rfin, ncols, cdeb, cfin);
  return Matrix (r, rfin-rdeb, cfin-cdeb);
}

TMPL Matrix
operator * (const Matrix& m, const Matrix& n) {
  if (is_a_scalar (m) || is_a_scalar (n)) {
    if (is_non_scalar (m)) return m * n.scalar();
    if (is_non_scalar (n)) return m.scalar() * n;
    return Matrix (m.scalar() * n.scalar());
  }
  nat mrows= rows (m), mcols= cols (m), nrows= rows(n), ncols= cols(n);
  ASSERT (nrows == mcols, "numbers of rows and columns don't match");
  C* r = mmx_new<C> (mrows*ncols);
  V::mul (r, tab (m), tab (n), mrows, mcols, ncols);
  return Matrix (r, mrows, ncols);
}

template<typename C, typename V, typename W> vector<C,W>
operator * (const matrix<C,V>& m, const vector<C,W>& v) {
  if (is_a_scalar (m)) return m.scalar() * v;
  ASSERT (is_non_scalar (v), "non-scalar vector expected");
  nat rr= rows (m), cc= cols (m);
  ASSERT (cc == N(v), "sizes don't match");
  C* a= mmx_new<C> (rr);
  for (nat i=0; i<rr; i++)
    a[i]= W::template vec_binary_big_stride<mul_add_op>
      (tab (m) + V::index (i, 0, rr, cc), V::index (0, 1, rr, cc),
       seg (v), 1, cc);
  return vector<C,W> (a, rr);
}

template<typename C, typename V, typename W> vector<C,W>
operator * (const vector<C,W>& v, const matrix<C,V>& m) {
  if (is_a_scalar (m)) return v * m.scalar();
  ASSERT (is_non_scalar (v), "non-scalar vector expected");
  nat rr= rows (m), cc= cols (m);
  ASSERT (rr == N(v), "sizes don't match");
  C* a= mmx_new<C> (cc);
  for (nat i=0; i<cc; i++)
    a[i]= W::template vec_binary_big_stride<mul_add_op>
      (seg (v), 1,
       tab (m) + V::index (0, i, rr, cc), V::index (1, 0, rr, cc), rr);
  return vector<C,W> (a, rr);
}

/******************************************************************************
* Operations on rows or columns
******************************************************************************/

TMPL vector<C>
row (const Matrix& m, nat i) {
  nat n= cols (m);
  nat l= default_aligned_size<C> (n);
  C* a= mmx_new<C> (l);
  for (nat j=0; j<n; j++) a[j]= m(i,j);
  return vector<C> (a, n, l);
}

TMPL vector<C>
column (const Matrix& m, nat j) {
  nat n= rows (m);
  nat l= default_aligned_size<C> (n);
  C* a= mmx_new<C> (l);
  for (nat i=0; i<n; i++) a[i]= m(i,j);
  return vector<C> (a, n, l);
}

TMPL void
swap_row (Matrix& m, nat i, nat j) {
  ASSERT (is_non_scalar (m), "non-scalar matrix expected");
  nat mrows= rows (m), mcols= cols (m);
  ASSERT (i < mrows && j < mrows, "out of range");
  V::row_swap (tab (m), i, j, mrows, mcols);
}

TMPL void
swap_col (Matrix& m, nat i, nat j) {
  ASSERT (is_non_scalar (m), "non-scalar matrix expected");
  nat mrows= rows (m), mcols= cols (m);
  ASSERT (i < mrows && j < mrows, "out of range");
  V::col_swap (tab (m), i, j, mrows, mcols);
}

#define INLINE_SCAL_LINE_OP(name1,name2)			\
TMPL void name1 (Matrix& m, C c, nat i) {			\
  ASSERT (is_non_scalar (m), "non-scalar matrix expected");	\
  nat mrows=rows(m), mcols=cols(m);				\
  V::name2 (tab(m), c, i, mrows, mcols);			\
}

#define INLINE_LINES_OP(name1, name2)				\
TMPL void name1 (Matrix& m, nat i, C ci, nat j, C cj) {		\
  ASSERT (is_non_scalar (m), "non-scalar matrix expected");	\
  nat mrows=rows(m), mcols=cols(m);				\
  V::name2 (tab(m), i, ci, j, cj, mrows, mcols);		\
}

INLINE_SCAL_LINE_OP(row_mul, row_mul);
INLINE_SCAL_LINE_OP(row_div, row_div);
INLINE_SCAL_LINE_OP(col_mul, col_mul);
INLINE_SCAL_LINE_OP(col_div, col_div);
INLINE_LINES_OP(rows_linsub,row_combine_sub);
INLINE_LINES_OP(cols_linsub,col_combine_sub);

/******************************************************************************
* Echelon forms
******************************************************************************/

TMPL Matrix
row_echelon (const Matrix& m) {
  ASSERT (is_non_scalar (m), "non-scalar matrix expected");
  Matrix c (copy (m));
  C* k= NULL;
  V::row_echelon (tab(c), k, rows(m), cols(m));
  return c;
}

TMPL Matrix
row_echelon (const Matrix& m, Matrix& k) {
  ASSERT (is_non_scalar (m), "non-scalar matrix expected");
  Matrix c (copy (m));
  k= Matrix (C(0), rows(m), rows(m));
  V::row_echelon (tab(c), tab(k), rows(m), cols(m));
  return c;
}

TMPL Matrix
col_echelon (const Matrix& m) {
  ASSERT (is_non_scalar (m), "non-scalar matrix expected");
  Matrix c (copy (m));
  C* k= NULL;
  V::col_echelon (tab(c), k, rows(m), cols(m));
  return c;
}

TMPL Matrix
col_echelon (const Matrix& m, Matrix& k) {
  ASSERT (is_non_scalar (m), "non-scalar matrix expected");
  Matrix c (copy (m));
  k= Matrix (C(0), cols(m), cols(m));
  V::col_echelon (tab(c), tab(k), rows(m), cols(m));
  return c;
}

/******************************************************************************
* System solving
******************************************************************************/

TMPL C
det (const Matrix& m) {
  ASSERT (is_non_scalar (m), "non-scalar matrix expected");
  ASSERT (cols(m) == rows(m), "square matrix expected");
  return V::det (tab(m), rows(m), cols(m));
}

TMPL C
det_no_div (const Matrix& m) {
  ASSERT (is_non_scalar (m), "non-scalar matrix expected");
  ASSERT (cols(m) == rows(m), "square matrix expected");
  return V::det_no_div (tab(m), rows(m), cols(m));
}

/******************************************************************************
* Applications
******************************************************************************/

TMPL Matrix
kernel (const Matrix& m) {
  ASSERT (is_non_scalar (m), "non-scalar matrix expected");
  Matrix k (C(0), cols(m), cols(m));
  nat dim= V::kernel (tab(k), tab(m), rows(m), cols(m));
  return submatrix (k, 0, cols(m), 0, dim);
}

TMPL Matrix
image (const Matrix& m) {
  ASSERT (is_non_scalar (m), "non-scalar matrix expected");
  Matrix k (C(0), rows(m), rows(m));
  nat dim= V::image (tab(k), tab(m), rows(m), cols(m));
  return submatrix (k, 0, rows(m), 0, dim);
}

TMPL nat
rank (const Matrix& m) {
  ASSERT (is_non_scalar (m), "non-scalar matrix expected");
  return V::rank (tab(m), rows(m), cols(m));
}

TMPL Matrix
invert (const Matrix& m) {
  ASSERT (is_non_scalar (m), "non-scalar matrix expected");
  ASSERT (cols(m) == rows(m), "square matrix expected");
  Matrix a (C(1), rows(m), cols(m));
  V::invert (tab(a), tab(m), rows(m), cols(m));
  return a;
}

/******************************************************************************
* Additionals
******************************************************************************/

TMPL Matrix
krylov (const Matrix& m, const Matrix& v) {
  ASSERT (is_non_scalar (m), "non-scalar matrix expected");
  ASSERT (is_square_matrix (m), "square matrix expected");
  Matrix r= image (v);
  Matrix p= m;
  while (true) {
    nat rk= rows (r);
    r= image (vertical_join (r, r*p));
    if (rows (r) <= rk) return r;
    p= p*p;
  }
}

#undef TMPL_DEF
#undef TMPL
#undef Matrix
#undef Matrix_rep
#undef Abs_matrix
} // namespace mmx
#endif // __MMX_MATRIX_HPP
