
/******************************************************************************
* MODULE     : matrix_block.hpp
* DESCRIPTION: unrolling and subdividing matrix multiplication
* COPYRIGHT  : (C) 2007  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__MATRIX_BLOCK__HPP
#define __MMX__MATRIX_BLOCK__HPP
#include <algebramix/matrix_fixed.hpp>

namespace mmx {

/******************************************************************************
* Complete a few missing rows, columns, inner products using another algorithm
******************************************************************************/

template<typename Op, typename V, typename D, typename S1, typename S2> void
mul_complete (D* dest, const S1* src1, const S2* src2,
	      nat r, nat rr, nat l, nat ll, nat c, nat cc,
	      nat hr, nat hl, nat hc)
{
  typedef typename Op::acc_op Acc;
  if (hr < r && hl != 0 && hc != 0)
    V::template mul<Op > (dest + V::index (hr, 0, rr, cc),
			  src1 + V::index (hr, 0, rr, ll),
			  src2,
			  r-hr, rr, hl, ll, hc, cc);
  if (hl != 0 && hc < c)
    V::template mul<Op > (dest + V::index (0, hc, rr, cc),
			  src1,
			  src2 + V::index (0, hc, ll, cc),
			  r , rr, hl, ll, c-hc, cc);
  if (hl < l)
    V::template mul<Acc> (dest,
			  src1 + V::index (0, hl, rr, ll),
			  src2 + V::index (hl, 0, ll, cc),
			  r , rr, l-hl, ll, c , cc);
}

/******************************************************************************
* Using a triple loop using big unrolled blocks
******************************************************************************/

template<typename Op, nat ur, nat ul, nat uc, typename V,
	 typename D, typename S1, typename S2> void
mul_unroll (D* dest, const S1* src1, const S2* src2,
	    nat r, nat rr, nat l, nat ll, nat c, nat cc)
{
  typedef typename Op::acc_op Acc;
  nat nr= r/ur, nl= l/ul, nc= c/uc;
  if (nl != 0)
    for (nat ir=0; ir<nr; ir++)
      for (nat ic=0; ic<nc; ic++) {
	nat il=0;
	for (; il<1; il++)
	  matrix_multiply_helper<Op,D,S1,S2,ur,ul,uc>::
	    mul_stride (dest + V::index (ir*ur, ic*uc, rr, cc),
			src1 + V::index (ir*ur, il*ul, rr, ll),
			src2 + V::index (il*ul, ic*uc, ll, cc),
			rr, ll);
	for (; il<nl; il++)
	  matrix_multiply_helper<Acc,D,S1,S2,ur,ul,uc>::
	    mul_stride (dest + V::index (ir*ur, ic*uc, rr, cc),
			src1 + V::index (ir*ur, il*ul, rr, ll),
			src2 + V::index (il*ul, ic*uc, ll, cc),
			rr, ll);
      }
  mul_complete<Op,V> (dest, src1, src2, r, rr, l, ll, c, cc,
		      ur*nr, ul*nl, uc*nc);
}

template<nat ur, nat ul, nat uc, typename V>
struct matrix_unroll_multiplication: public V {
  template<typename Op, typename D, typename S1, typename S2>
  static inline void
  mul (D* dest, const S1* src1, const S2* src2,
       nat r, nat rr, nat l, nat ll, nat c, nat cc)
  {
    mul_unroll<Op,ur,ul,uc,V> (dest, src1, src2, r, rr, l, ll, c, cc);
  }
};

template<typename V>
struct matrix_std_unroll_multiplication:
  public matrix_unroll_multiplication<4,4,4,
	   matrix_unroll_multiplication<2,2,2,
             V> > {};

/******************************************************************************
* Interface
******************************************************************************/

struct matrix_unrolled:
  public matrix_naive_application<
           matrix_enrich<
	     matrix_std_unroll_multiplication<
	       matrix_naive_multiplication<
		 matrix_naive_row_column<
		   matrix_vectorial<
		     matrix_naive_abstractions<
		       vector_naive> > > > > > >
{
  static const nat def_rows = 0;
  static const nat def_cols = 0;
  static const nat init_rows= 1;
  static const nat init_cols= 1;
};

} // namespace mmx
#endif //__MMX__MATRIX_BLOCK__HPP
