
/******************************************************************************
* MODULE     : matrix_complex.hpp
* DESCRIPTION: Complex multiplication of matrices
* COPYRIGHT  : (C) 2008 Raouf Dridi and Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__MATRIX_COMPLEX__HPP
#define __MMX__MATRIX_COMPLEX__HPP
#include <algebramix/matrix.hpp>
#include <numerix/complex.hpp>
#include <numerix/integer.hpp>

namespace mmx {
#define TMPL template<typename C>
#define Complex complex<C>

/******************************************************************************
* Complex multiplication
******************************************************************************/

template<typename V>
struct matrix_complex_multiplication: public V {

TMPL static inline void
mul (Complex* dest, const Complex* s1, const Complex* s2,
     nat r, nat l, nat c)
{
  nat n1 = r*l, spc1= ((n1+1)>>1) << 1; // force 128 bit alignment if necessary
  nat n2 = l*c, spc2= ((n2+1)>>1) << 1;
  nat nd = r*c, spcd= ((nd+1)>>1) << 1;
  nat spc= 2*spc1 + 2*spc2 + 3*spcd;
  C* real_x1= mmx_new<C> (spc);
  C* imag_x1= real_x1 + spc1;
  C* real_x2= imag_x1 + spc1;
  C* imag_x2= real_x2 + spc2;
  C* real_d = imag_x2 + spc2;
  C* imag_d = real_d  + spcd;
  C* temp_d = imag_d  + spcd;

  V::Vec::half_copy (real_x1, (C*) ((void*) s1), n1);
  V::Vec::half_copy (imag_x1, ((C*) ((void*) s1)) + 1, n1);
  V::Vec::half_copy (real_x2, (C*) ((void*) s2), n2);
  V::Vec::half_copy (imag_x2, ((C*) ((void*) s2)) + 1, n2);
  //cout << "real_x1= "; V::Vec::print (cout, real_x1, n1); cout << "\n";
  //cout << "real_x2= "; V::Vec::print (cout, real_x2, n2); cout << "\n";
  V::mul (real_d, real_x1, real_x2, r, l, c);
  //cout << "mul1_d = "; V::Vec::print (cout, real_d , nd); cout << "\n";
  //cout << "imag_x1= "; V::Vec::print (cout, imag_x1, n1); cout << "\n";
  //cout << "imag_x2= "; V::Vec::print (cout, imag_x2, n2); cout << "\n";
  V::mul (temp_d, imag_x1, imag_x2, r, l, c);
  //cout << "mul2_d = "; V::Vec::print (cout, temp_d , nd); cout << "\n";
  V::Vec::add (real_x1, imag_x1, n1);
  V::Vec::add (real_x2, imag_x2, n2);
  //cout << "sum_x1 = "; V::Vec::print (cout, real_x1, n1); cout << "\n";
  //cout << "sum_x2 = "; V::Vec::print (cout, real_x2, n2); cout << "\n";
  V::mul (imag_d, real_x1, real_x2, r, l, c);
  //cout << "mul3_d = "; V::Vec::print (cout, imag_d , nd); cout << "\n";
  V::Vec::sub (imag_d, real_d, nd);
  V::Vec::sub (imag_d, temp_d, nd);
  V::Vec::sub (real_d, temp_d, nd);
  V::Vec::double_copy ((C*) ((void*) dest), real_d, nd);
  V::Vec::double_copy (((C*) ((void*) dest)) + 1, imag_d, nd);

  mmx_delete<C> (real_x1, spc);
}

}; // matrix_complex_multiplication<V>

/******************************************************************************
* Use complexified multiplication by default
******************************************************************************/

template<typename V>
struct matrix_complex: 
  public matrix_naive_application<
           matrix_complex_multiplication<V> >
{
  static const nat def_rows = 0;
  static const nat def_cols = 0;
  static const nat init_rows= 1;
  static const nat init_cols= 1;
};

TMPL
struct matrix_variant_helper<Complex > {
  typedef typename matrix_variant_helper<C>::MV CMV;
  typedef matrix_complex<CMV> MV;
};

STMPL
struct matrix_variant_helper<complex<generic> > {
  typedef matrix_naive MV;
};

#undef TMPL
#undef Complex
} // namespace mmx
#endif //__MMX__MATRIX_COMPLEX__HPP
