
/******************************************************************************
* MODULE     : matrix_modular.hpp
* DESCRIPTION: modular matrices using SIMD instructions
* COPYRIGHT  : (C) 2007  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__MATRIX_MODULAR__HPP
#define __MMX__MATRIX_MODULAR__HPP
#include <numerix/modular.hpp>
#include <algebramix/modular_chinese.hpp>
#include <algebramix/matrix_simd.hpp>

namespace mmx {
#define TMPL template<typename V1, typename V2>
#define Modulus modulus<uint32_t,V1>
#define Modular modular<Modulus,V2>

template<typename V, typename V1, typename V2> void
encode_modular (double* dest, const Modular* src,
		nat r, nat rr, nat c, nat cc)
{
  nat dest_rs= V::index (1, 0, r, c);
  nat dest_cs= V::index (0, 1, r, c);
  nat src_rs = V::index (1, 0, rr, cc);
  nat src_cs = V::index (0, 1, rr, cc);
  double* dest_row= dest;
  const Modular* src_row= src;
  for (nat i=0; i<r; i++, dest_row += dest_rs, src_row += src_rs) {
    double* dest_col= dest_row;
    const Modular* src_col= src_row;
    for (nat j=0; j<c; j++, dest_col += dest_cs, src_col += src_cs)
      *dest_col= (double) (*(*src_col));
  }
}

template<typename Op, typename V, typename V1, typename V2> void
decode_modular (Modular* dest, const double* src,
		nat r, nat rr, nat c, nat cc)
{
  typedef typename Op::set_op Set;
  uint32_t p = *Modular::get_modulus ();
  nat dest_rs= V::index (1, 0, rr, cc);
  nat dest_cs= V::index (0, 1, rr, cc);
  nat src_rs = V::index (1, 0, r, c);
  nat src_cs = V::index (0, 1, r, c);
  Modular* dest_row= dest;
  const double* src_row= src;
  for (nat i=0; i<r; i++, dest_row += dest_rs, src_row += src_rs) {
    Modular* dest_col= dest_row;
    const double* src_col= src_row;
    for (nat j=0; j<c; j++, dest_col += dest_cs, src_col += src_cs)
      Set::in_place (*dest_col, Modular (((uint64_t) (*src_col)) % p));
  }
}

template<typename V>
struct matrix_modular_multiplication: public V {
  template<typename Op, typename V1, typename V2> static inline void
  mul (Modular* d, const Modular* s1, const Modular* s2,
       nat r, nat rr, nat l, nat ll, nat c, nat cc)
  {
    if (l <= (1 << FREE_BITS_IN_DOUBLE)) {
      nat sd = ((r * c + 1) >> 1) << 1; // force 128 bit alignment
      nat ss1= ((r * l + 1) >> 1) << 1; // force 128 bit alignment
      nat ss2= ((l * c + 1) >> 1) << 1; // force 128 bit alignment
      double*  xd = mmx_new<double> (sd);
      double*  xs1= mmx_new<double> (ss1);
      double*  xs2= mmx_new<double> (ss2);
      encode_modular<V> (xs1, s1, r, rr, l, ll);
      encode_modular<V> (xs2, s2, l, ll, c, cc);
      V::template mul<mul_op> (xd, xs1, xs2, r, r, l, l, c, c);
      decode_modular<Op,V> (d, xd, r, rr, c, cc);
      mmx_delete<double> (xd , sd );
      mmx_delete<double> (xs1, ss1);
      mmx_delete<double> (xs2, ss2);
    }
    else {
      typedef typename Op::acc_op Acc;
      const nat ur= 4 << FREE_BITS_IN_DOUBLE;
      const nat ul= 1 << FREE_BITS_IN_DOUBLE;
      const nat uc= 4 << FREE_BITS_IN_DOUBLE;      
      nat nr= (r + ur - 1) / ur;
      nat nl= (l + ul - 1) / ul;
      nat nc= (c + uc - 1) / uc;
      for (nat ir=0; ir<nr; ir++)
	for (nat ic=0; ic<nc; ic++) {
	  nat il=0;
	  for (; il<1; il++) {
	    mul<Op > (d  + V::index (ir*ur, ic*uc, rr, cc),
		      s1 + V::index (ir*ur, il*ul, rr, ll),
		      s2 + V::index (il*ul, ic*uc, ll, cc),
		      min (ur, r - ir*ur), rr,
		      min (ul, l - il*ul), ll,
		      min (uc, c - ic*uc), cc);
	  }
	  for (; il<nl; il++) {
	    mul<Acc> (d  + V::index (ir*ur, ic*uc, rr, cc),
		      s1 + V::index (ir*ur, il*ul, rr, ll),
		      s2 + V::index (il*ul, ic*uc, ll, cc),
		      min (ur, r - ir*ur), rr,
		      min (ul, l - il*ul), ll,
		      min (uc, c - ic*uc), cc);
	  }
	}
    }
  }
};

struct matrix_modular:
  public matrix_enrich<
           matrix_modular_multiplication<
             matrix_simd_multiplication<               
	       matrix_std_unroll_multiplication<
                 matrix_naive_multiplication<
	           matrix_naive_row_column<
	             matrix_vectorial<
                       matrix_naive_abstractions<
                         vector_naive> > > > > > > >
{
  static const nat def_rows = 0;
  static const nat def_cols = 0;
  static const nat init_rows= 1;
  static const nat init_cols= 1;
};

TMPL
struct matrix_variant_helper<Modular > {
  typedef matrix_modular MV;
};

#undef TMPL
#undef Modulus
#undef Modular
} // namespace mmx

#endif //__MMX__MATRIX_MODULAR__HPP
