
/******************************************************************************
* MODULE     : matrix_quotient.hpp
* DESCRIPTION: Multiply matrices by putting entries under a common denominator
* COPYRIGHT  : (C) 2008 Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_MATRIX_QUOTIENT_HPP
#define __MMX_MATRIX_QUOTIENT_HPP
#include <numerix/rational.hpp>
#include <algebramix/matrix_integer.hpp>
namespace mmx {
#define TMPL template<typename C>
#define NC Numerator_type(C)
#define DC Denominator_type(C)

/******************************************************************************
* Low level subroutines
******************************************************************************/

TMPL DC
least_common_denominator (const C* s, nat str, nat n) {
  switch (n) {
  case 0: return 1;
  case 1: return denominator (s[0]);
  case 2: return lcm (denominator (s[0]), denominator (s[str]));
  case 3: return lcm (denominator (s[0]), lcm (denominator (s[str]),
					       denominator (s[str<<1])));
  default:
    return lcm (least_common_denominator (s, str, n>>1),
		least_common_denominator (s + (n>>1) * str, str, n - (n>>1)));
  }
}

TMPL void
denominator_mul (NC* dest, const C* src, nat str, nat n, const DC& d) {
  for (; n != 0; n--, dest += str, src += str)
    *dest= numerator (*src) * (d / denominator (*src));
}

TMPL void
denominator_div (C* dest, const NC* src,
		 nat row_str, nat rows,
		 nat col_str, nat cols,
		 const DC* row_d, const DC* col_d)
{ 
  for (nat i=0; i < rows; i++, dest += row_str, src += row_str, row_d++) {
    C*        Dest = dest;
    const NC* Src  = src;
    const DC* Col_d= col_d;
    for (nat j=0; j < cols; j++, Dest += col_str, Src += col_str, Col_d++)
      *Dest= C (*Src) / ((*row_d) * (*Col_d));
  }
}

/******************************************************************************
* Multiply matrices by putting the entries under a common denominator
******************************************************************************/

template<nat thr, typename V>
struct matrix_quotient_multiplication: public V {

TMPL static void
mul (C* d, const C* s1, const C* s2,
     nat r, nat l, nat c) 
{
  if (r <= thr || l <= thr || c <= thr)
    V::Naive::template mul<mul_op> (d, s1, s2, r, r, l, l, c, c);
  else {
    DC* d1  = mmx_new<DC> (r);
    DC* d2  = mmx_new<DC> (c);
    NC* Src1= mmx_new<NC> (r*l);
    NC* Src2= mmx_new<NC> (c*l);
    NC* Dest= mmx_new<NC> (r*c);
    for (nat i=0; i<r; i++) 
      d1[i] = least_common_denominator (s1 + V::index (i, 0, r, l),
					V::index (0, 1, r, l), l);
    //cout << "d1= "; V::Vec::print (cout, d1, r); cout << "\n";
    for (nat j=0; j<c; j++) 
      d2[j] = least_common_denominator (s2 + V::index (0, j, l, c),
					V::index (1, 0, l, c), l);
    //cout << "d2= "; V::Vec::print (cout, d2, c); cout << "\n";

    for (nat i=0; i<r; i++) 
      denominator_mul 
	(Src1 + V::index (i, 0, r, l), s1 + V::index (i, 0, r, l),
	 V::index (0, 1, r, l), l, d1[i]);
    //cout << "Src1= "; V::Vec::print (cout, Src1, r*l); cout << "\n";
    for (nat j=0; j<c; j++) 
      denominator_mul 
	(Src2 + V::index (0, j, l, c), s2 + V::index (0, j, l, c),
	 V::index (1, 0, l, c), l, d2[j]);
    //cout << "Src2= "; V::Vec::print (cout, Src1, l*c); cout << "\n";

    V::mul (Dest, Src1, Src2, r, l, c);
    //cout << "Dest= "; V::Vec::print (cout, Dest, r*c); cout << "\n";

    denominator_div (d, Dest,
		     V::index (1, 0, r, c), r,
		     V::index (0, 1, r, c), c,
		     d1, d2);

    mmx_delete<DC> (d1, r);
    mmx_delete<DC> (d2, c);
    mmx_delete<NC> (Src1, r*l);
    mmx_delete<NC> (Src2, l*c);
    mmx_delete<NC> (Dest, r*c);
  }
}

}; // matrix_quotient_multiplication<thr, V>

/******************************************************************************
* Use multi-modular multiplication by default for rational matrices
******************************************************************************/

struct matrix_rational: 
  public matrix_naive_application<
           matrix_quotient_multiplication<7,
	     matrix_integer_multiplication<7,
               matrix_enrich<
                 matrix_naive_multiplication<
		   matrix_naive_row_column<
		     matrix_vectorial<
		       matrix_naive_abstractions<
			 vector_naive> > > > > > > >
{
  static const nat def_rows = 0;
  static const nat def_cols = 0;
  static const nat init_rows= 1;
  static const nat init_cols= 1;
};

STMPL
struct matrix_variant_helper<rational> {
  typedef matrix_rational MV;
};

#undef TMPL
#undef NC
#undef DC
}
#endif // __MMX_MATRIX_QUOTIENT_HPP

