
/******************************************************************************
* MODULE     : matrix_simd.hpp
* DESCRIPTION: matrices using SIMD instructions
* COPYRIGHT  : (C) 2007  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__MATRIX_SIMD__HPP
#define __MMX__MATRIX_SIMD__HPP
#include <algebramix/algebramix-config.hpp>
#include <numerix/simd.hpp>
#include <numerix/sse.hpp>
#include <algebramix/matrix_fixed.hpp>

namespace mmx {

#if defined (NUMERIX_ENABLE_SIMD) && defined (ALGEBRAMIX_ENABLE_SIMD) && defined (__SSE2__)

template<typename V, typename Op, typename D, typename S1, typename S2>
struct matrix_simd_helper {
  static inline void
  mul (D* d, const S1* s1, const S2* s2,
       nat r, nat rr, nat l, nat ll, nat c, nat cc)
  {
    V::template mul<Op> (d, s1, s2, r, rr, l, ll, c, cc);
  }
};

template<typename V, typename Op>
struct matrix_simd_helper<V, Op, double, double, double> {
  typedef simd_helper<double>::type simd_double;
  static inline void
  mul (double* d, const double* s1, const double* s2,
       nat r, nat rr, nat l, nat ll, nat c, nat cc)
  {
    VERIFY (((long nat) ((void*) d )) & 15 == 0, "alignment problem");
    VERIFY (((long nat) ((void*) s1)) & 15 == 0, "alignment problem");
    if (r == 0);
    else if ((r&1) != 0) {
      mul (d, s1, s2, r-1, rr, l, ll, c, cc);
      V::template mul<Op> (d  + V::index (r-1, 0, rr, cc),
			   s1 + V::index (r-1, 0, rr, ll), s2,
			   1, rr, l, ll, c, cc);
      // NOTE: row times matrix can be further optimized for SIMD instructions
    }
    else if ((rr&1) != 0) {
      double* xd = mmx_new<double> (r * c);
      double* xs1= mmx_new<double> (r * l);
      double* xs2= mmx_new<double> (l * c);
      V::template mat_unary_stride<id_op>
	(xs1, V::index (1, 0, r , l ), V::index (0, 1, r , l ),
	 s1 , V::index (1, 0, rr, ll), V::index (0, 1, rr, ll), r, l);
      V::template mat_unary_stride<id_op>
	(xs2, V::index (1, 0, l , c ), V::index (0, 1, l , c ),
	 s2 , V::index (1, 0, ll, cc), V::index (0, 1, ll, cc), l, c);
      mul (xd, xs1, xs2, r, r, l, l, c, c);
      V::template mat_unary_stride<typename Op::set_op>
	(d , V::index (1, 0, rr, cc), V::index (0, 1, rr, cc),
	 xd, V::index (1, 0, r , c ), V::index (0, 1, r , c ), r, c);
      mmx_delete<double> (xd , r * c);
      mmx_delete<double> (xs1, r * l);
      mmx_delete<double> (xs2, l * c);
    }
    else {
      simd_double* xd = (simd_double*) ((void*) d);
      simd_double* xs1= (simd_double*) ((void*) s1);
      V::template mul<Op> (xd, xs1, s2, r >> 1, rr >> 1, l, ll, c, cc);
    }
  }
};

template<typename V>
struct matrix_simd_multiplication: public V {
  typedef typename V::Naive Naive;
  typedef matrix_simd_multiplication<typename V::Positive> Positive;
  typedef typename V::No_simd No_simd;
  typedef matrix_simd_multiplication<typename V::No_thread> No_thread;
  typedef matrix_simd_multiplication<typename V::No_scaled> No_scaled;

  template<typename Op, typename D, typename S1, typename S2>
  static inline void
  mul (D* d, const S1* s1, const S2* s2,
       nat r, nat rr, nat l, nat ll, nat c, nat cc)
  {
    matrix_simd_helper<V,Op,D,S1,S2>::
      mul (d, s1, s2, r, rr, l, ll, c, cc);
  }
};

#else // not (*_ENABLE_SIMD and __SSE2__)

template<typename V>
struct matrix_simd_multiplication: public V {
  template<typename Op, typename D, typename S1, typename S2>
  static inline void
  mul (D* d, const S1* s1, const S2* s2,
       nat r, nat rr, nat l, nat ll, nat c, nat cc)
  {
    V::template mul<Op> (d, s1, s2, r, rr, l, ll, c, cc);
  }
};

#endif // *_ENABLE_SIMD and __SSE2__

} // namespace mmx

#endif // __MMX__MATRIX_SIMD__HPP
