
/******************************************************************************
* MODULE     : polynomial.hpp
* DESCRIPTION: Dense univariate polynomials
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_POLYNOMIAL_HPP
#define __MMX_POLYNOMIAL_HPP
#include <basix/string.hpp>
#include <basix/vector.hpp>
#include <algebramix/polynomial_naive.hpp>
namespace mmx {
#define Polynomial_multiply(C) polynomial_variant_helper<C>::MV
#define Polynomial_variant(C) polynomial_variant_helper<C>::PV
#define TMPL_DEF \
  template<typename C, typename V=typename Polynomial_variant(C) >
#define TMPL template<typename C, typename V>
#define Polynomial polynomial<C,V>
#define Polynomial_rep polynomial_rep<C,V>
#define Vector vector<C>
TMPL class polynomial_rep;
TMPL class polynomial;
TMPL inline int deg (const Polynomial& P);
TMPL inline int degree (const Polynomial& P);
TMPL inline nat N (const Polynomial& P);
TMPL inline const C* seg (const Polynomial& P);

/******************************************************************************
* Dense polynomials
******************************************************************************/

template<typename C>
struct polynomial_variant_helper {
  typedef typename Vector_variant(C) VV;
  typedef polynomial_multiply_naive<VV> MV;
  typedef polynomial_naive PV;
  // typedef polynomial_algos_naive<MV> PV;
  // Commented out so as to keep typenames simple when debugging
};

TMPL_DEF
class polynomial_rep REP_STRUCT {
  C*  a;
  nat n;
  nat l;  // allocated number of coefficients

public:
  inline void normalize () {
    while ((n != 0) && (a[n-1] == C(0))) n--; }
  inline polynomial_rep (C* a2, nat n2):
    a(a2), n(n2), l(n2) { normalize (); }
  inline polynomial_rep (C* a2, nat n2, nat l2):
    a(a2), n(n2), l(l2) { normalize (); }
  inline ~polynomial_rep () {
    mmx_delete<C> (a, l); }
  void extend (nat n2) {
    nat l2= aligned_size<C,V> (n2);
    C* b= mmx_new<C> (l2);
    V::copy (b, a, n);
    V::clear (b+n, l2-n);
    mmx_delete<C> (a, l);
    a= b;
    n= n2;
    l= l2;
  }

  friend class Polynomial;
  friend int deg LESSGTR (const Polynomial& P);
  friend int degree LESSGTR (const Polynomial& P);
  friend nat N LESSGTR (const Polynomial& P);
  friend const C* seg LESSGTR (const Polynomial& P);
};

// storage of the variable name
struct variable_name_local {
};

TMPL_DEF
class polynomial {
INDIRECT_PROTO_2 (polynomial, polynomial_rep, C, V)
public:
  typedef typename V::template variable_name_storage<Polynomial> S;
  static inline string get_variable_name () {
    return S::get_variable_name (); }
  static inline void set_variable_name (const string& x) {
    S::set_variable_name (x); }

  inline polynomial () {
    nat l= aligned_size<C,V> (0);
    rep= new Polynomial_rep (mmx_new<C> (l), 0, l);
  }
  template<typename T> inline polynomial (const T& x) {
    nat l= aligned_size<C,V> (1);
    C* a= mmx_new<C> (l);
    a[0]= x;
    rep= new Polynomial_rep (a, 1, l);
    rep->normalize ();
  }
  template<typename T> inline
  polynomial (const polynomial<T,V>& P) {
    nat l= aligned_size<C,V> (N(P));
    C* a= mmx_new<C> (l);
    V::cast (a, seg (P), N(P));
    rep= new Polynomial_rep (a, N(P), l);
    rep->normalize ();
  }
  template<typename T> polynomial (const T& x, nat n) {
    nat l= aligned_size<C,V> (n+1);
    C* a= mmx_new<C> (l);
    V::clear (a, n);
    a[n]= x;
    rep= new Polynomial_rep (a, n+1, l);
    rep->normalize ();
  }
  inline polynomial (C* a, nat n) {
    rep= new Polynomial_rep (a, n);
  }
  inline polynomial (C* a, nat n, nat l) {
    rep= new Polynomial_rep (a, n, l);
  }
  polynomial (const vector<C>& v) {
    nat l= aligned_size<C,V> (N(v));
    C* a= mmx_new<C> (l);
    V::copy (a, seg (v), N(v));
    rep= new Polynomial_rep (a, N(v), l);
    rep->normalize ();
  }
  polynomial (const iterator<C>& it) {
    nat i, l=0;
    nat s= aligned_size<C,V> (0);
    rep= new Polynomial_rep (mmx_new<C> (s), 0, s);
    for (i=0; busy (it); i++, ++it) {
      if (i >= l) {
	l= max ((nat) 1, l<<1);
	rep->extend (l);
      }
      rep->a[i]= *it;
    }
    rep->normalize ();
  }
  inline friend Polynomial copy (const Polynomial& P) {
    nat n= N(P), l= aligned_size<C,V> (n);
    C* r= mmx_new<C> (l);
    V::copy (r, seg (P), n);
    return Polynomial (r, n, l); }
  inline C operator[] (nat i) const { return i<rep->n? rep->a[i]: C(0); }
  Polynomial& operator += (const Polynomial& P);
  Polynomial& operator -= (const Polynomial& P);
  Polynomial& operator *= (const C& c);
  Polynomial& operator /= (const C& c);
};
INDIRECT_IMPL_2 (polynomial, polynomial_rep, typename C, C, typename V, V)

TMPL inline nat N (const Polynomial& P) { return P->n; }
TMPL inline int deg (const Polynomial& P) { return ((int) P->n) - 1; }
TMPL inline int degree (const Polynomial& P) { return ((int) P->n) - 1; }
TMPL inline const C* seg (const Polynomial& P) { return P->a; }

TMPL
struct scalar_type_helper<Polynomial > {
  typedef C val;
};

TMPL
struct fast_helper<Polynomial > {
  typedef Fast_type(C) FC;
  typedef polynomial<FC> fast_type;
  typedef typename Polynomial_variant(FC) FV;
  static inline fast_type dd (const Polynomial& p) {
    nat l= aligned_size<FC,FV> (N(p));
    FC* r= mmx_new<FC> (l);
    for (nat i=0; i<N(p); i++) r[i]= fast (p[i]);
    return fast_type (r, N(p), l); }
  static inline Polynomial uu (const fast_type& p) {
    nat l= aligned_size<C,V> (N(p));
    C* r= mmx_new<C> (l);
    for (nat i=0; i<N(p); i++) r[i]= slow<C> (p[i]);
    return Polynomial (r, N(p), l); }
};

template<typename T, typename F, typename TV, typename FV>
struct as_helper<polynomial<T,TV>,polynomial<F,FV> > {
  static inline polynomial<T,TV>
  cv (const polynomial<F,FV>& p) {
    nat l= aligned_size<T,TV> (N(p));
    T* c= mmx_new<T> (l);
    for (nat i=0; i<N(p); i++) c[i]= as<T> (p[i]);
    return polynomial<T,TV> (c, N(p), l); }
};

template<typename Op, typename C, typename V> nat
unary_hash (const Polynomial& p) {
  register nat i, h= 642531, n= N(p);
  for (i=0; i<n; i++)
    h= (h<<1) ^ (h<<5) ^ (h>>27) ^ Op::op (p[i]);
  return h;
}

/******************************************************************************
* Polynomial iteration
******************************************************************************/

TMPL_DEF
class polynomial_iterator_rep: public iterator_rep<C> {
  Polynomial P;  // the polynomial to traverse
  nat i;         // current position
protected:
  bool is_busy () { return ((int) i) <= deg (P); }
  void advance () { i++; }
  C current () { return P[i]; }
public:
  polynomial_iterator_rep (const Polynomial& P2): P(P2), i(0) {}
};

TMPL iterator<C>
iterate (const Polynomial& P) {
  return iterator<C> (new polynomial_iterator_rep<C,V> (P));
}

/******************************************************************************
* Output routines for dense polynomials
******************************************************************************/

TMPL inline generic var (const Polynomial& P) {
  (void) P; return Polynomial::get_variable_name (); }

TMPL syntactic
flatten (const Polynomial& P, const syntactic& v) {
  int i, len= deg(P);
  syntactic s= 0;
  for (i=len; i>=0; i--)
    s= s + flatten (P[i]) * pow (v, i);
  return s;
}

TMPL syntactic
flatten (const Polynomial& P) {
  return flatten (P, as_syntactic (var (P)));
}

PRINT_SUGAR(TMPL,Polynomial)

/******************************************************************************
* In place arithmetic
******************************************************************************/

TMPL Polynomial&
Polynomial::operator += (const Polynomial& P) {
  secure ();
  nat n1= rep->n, n2= N(P);
  if (n1 < n2) rep -> extend (n2);
  V::add (rep->a, seg (P), n2);
  if (n1 == n2) rep -> normalize ();
  return *this;
}

TMPL Polynomial&
Polynomial::operator -= (const Polynomial& P) {
  secure ();
  nat n1= rep->n, n2= N(P);
  if (n1 < n2) rep->extend (n2);
  V::sub (rep->a, seg (P), n2);
  if (n1 == n2) rep->normalize ();
  return *this;
}

TMPL Polynomial&
Polynomial::operator *= (const C& c) {
  secure ();
  V::mul_sc (rep->a, c, rep->n);
  rep->normalize ();
  return *this;
}

TMPL Polynomial&
Polynomial::operator /= (const C& c) {
  secure ();
  V::div_sc (rep->a, c, rep->n);
  rep->normalize ();
  return *this;
}

/******************************************************************************
* Basic arithmetic operations
******************************************************************************/

TMPL Polynomial
operator - (const Polynomial& P) {
  nat n= N(P);
  nat l= aligned_size<C,V> (n);
  C* r= mmx_new<C> (l);
  V::neg (r, seg (P), n);
  return Polynomial (r, n, l);
}

TMPL Polynomial
operator + (const Polynomial& P1, const Polynomial& P2) {
  nat m= min (N(P1), N(P2));
  nat n= max (N(P1), N(P2));
  nat l= aligned_size<C,V> (n);
  C* r= mmx_new<C> (l);
  V::add (r, seg (P1), seg (P2), m);
  if (N(P1) > m) V::copy (r+m, seg(P1)+m, n-m);
  if (N(P2) > m) V::copy (r+m, seg(P2)+m, n-m);
  return Polynomial (r, n, l);
}

TMPL Polynomial operator + (const Polynomial& P, const C& c) {
  return P + Polynomial (c); }
TMPL Polynomial operator + (const C& c, const Polynomial& P) {
  return Polynomial (c) + P; }

TMPL Polynomial
big_add (const vector<Polynomial >& a) {
  nat i, k= N(a), n=0;
  nat l= aligned_size<C,V> (n);
  for (i=0; i<k; i++) n= max (N(a[i]), n);
  C* r= mmx_new<C> (l);
  V::set (r, 0, n);
  for (i=0; i<k; i++) V::add (r, seg (a[i]), N(a[i]));
  return Polynomial (r, n, l);
}

TMPL Polynomial
operator - (const Polynomial& P1, const Polynomial& P2) {
  nat m= min (N(P1), N(P2));
  nat n= max (N(P1), N(P2));
  nat l= aligned_size<C,V> (n);
  C* r= mmx_new<C> (l);
  V::sub (r, seg (P1), seg (P2), m);
  if (N(P1) > m) V::copy (r+m, seg(P1)+m, n-m);
  if (N(P2) > m) V::neg  (r+m, seg(P2)+m, n-m);
  return Polynomial (r, n, l);
}

TMPL Polynomial operator - (const Polynomial& P, const C& c) {
  return P - Polynomial (c); }
TMPL Polynomial operator - (const C& c, const Polynomial& P) {
  return Polynomial (c) - P; }

TMPL Polynomial
operator * (const Polynomial& P1, const Polynomial& P2) {
  nat n1= N(P1), n2= N(P2);
  if (n1 == 0 || n2 == 0) return Polynomial ();
  nat l= aligned_size<C,V> (n1+n2-1);
  C* r= mmx_new<C> (l);
  V::mul (r, seg (P1), seg (P2), n1, n2);
  return Polynomial (r, n1+n2-1, l);
}

//! Transposed multiplication.
//! The returned polynomial has degree at most d2.
TMPL Polynomial
tmul (int d2, const Polynomial& P1, const Polynomial& P2) {
  nat n2 = max (0, d2 + 1), n1= N(P1), n= N(P2);
  if (n1 == 0 || n2 == 0) return Polynomial ();
  ASSERT (n < n1 + n2, "bad dimension in tmul");
  nat l2= aligned_size<C,V> (n2);
  C* r= mmx_new<C> (l2);
  if (n != n1 + n2 - 1) {
    nat l= aligned_size<C,V> (n1+n2-1);
    C* s2= mmx_new<C> (l);
    V::copy (s2, seg (P2), n);
    V::clear (s2 + n, n1+n2-n-1);
    V::tmul (r, seg (P1), s2, n1, n2);
    mmx_delete<C> (s2, l);
  }
  else
    V::tmul (r, seg (P1), seg (P2), n1, n2);
  return Polynomial (r, n2, l2);
}

TMPL Polynomial
square (const Polynomial& P) {
  nat n= N(P);
  if (n == 0) return Polynomial ();
  nat l= aligned_size<C,V> (2*n-1);
  C* r= mmx_new<C> (l);
  V::square (r, seg (P), n);
  return Polynomial (r, 2*n-1, l);
}

TMPL Polynomial
operator * (const Polynomial& P, const C& c) {
  nat n= N(P);
  nat l= aligned_size<C,V> (n);
  C* r= mmx_new<C> (l);
  V::mul_sc (r, seg (P), c, n);
  return Polynomial (r, n, l);
}

TMPL Polynomial
operator * (const C& c, const Polynomial& P) {
  nat n= N(P);
  nat l= aligned_size<C,V> (n);
  C* r= mmx_new<C> (l);
  V::mul_sc (r, seg (P), c, n);
  return Polynomial (r, n, l);
}

TMPL Polynomial
big_mul (const vector<Polynomial >& a) {
  return V::Vec::template vec_unary_big_dicho<mul_op> (seg (a), N(a));
}

TMPL Polynomial
operator / (const Polynomial& P, const C& c) {
  nat n= N(P);
  nat l= aligned_size<C,V> (n);
  C* r= mmx_new<C> (l);
  V::div_sc (r, seg (P), c, n);
  return Polynomial (r, n, l);
}

TMPL Polynomial
skew_div (const Polynomial& P, const C& c, bool left) {
  if (left) {
    nat n= N(P);
    nat l= aligned_size<C,V> (n);
    C* r= mmx_new<C> (n);
    V::div_sc (r, seg (P), c, n);
    return Polynomial (r, n, l);
  }
  else return P / c;
}

TMPL Polynomial
quo (const Polynomial& P, const C& c) {
  nat n= N(P);
  nat l= aligned_size<C,V> (n);
  C* r= mmx_new<C> (l);
  V::quo_sc (r, seg (P), c, n);
  return Polynomial (r, n, l);
}

TMPL Polynomial
rem (const Polynomial& P, const C& c) {
  nat n= N(P);
  nat l= aligned_size<C,V> (n);
  C* r= mmx_new<C> (l);
  V::rem_sc (r, seg (P), c, n);
  return Polynomial (r, n, l);
}

TMPL Polynomial
operator / (const Polynomial& P1, const Polynomial& P2) {
  nat n1= N(P1), n2= N(P2);
  if (n1 < n2 || n2 == 0) return Polynomial ();
  nat lq= aligned_size<C,V> (n1-n2+1);
  nat lr= aligned_size<C,V> (n1);
  C* q= mmx_new<C> (lq);
  C* r= mmx_new<C> (lr);
  V::copy (r, seg (P1), n1);
  V::quo_rem (q, r, seg (P2), n1, n2);
  ASSERT (V::equal (r, C(0), n1), "division must be exact");
  mmx_delete<C> (r, lr);
  return Polynomial (q, n1-n2+1, lq);
}

TMPL Polynomial
quo (const Polynomial& P1, const Polynomial& P2) {
  nat n1= N(P1), n2= N(P2);
  if (n1 < n2 || n2 == 0) return Polynomial ();
  nat lq= aligned_size<C,V> (n1-n2+1);
  nat lr= aligned_size<C,V> (n1);
  C* q= mmx_new<C> (lq);
  C* r= mmx_new<C> (lr);
  V::copy (r, seg (P1), n1);
  V::quo_rem (q, r, seg (P2), n1, n2);
  mmx_delete<C> (r, lr);
  return Polynomial (q, n1-n2+1, lq);
}

//! Transposed quotient in the division in degree d1 by P2.
TMPL Polynomial
tquo (int d1, const Polynomial& P1, const Polynomial& P2) {
  nat n1= max (0, d1+1), n= N(P1), n2= N(P2);
  ASSERT (n <= n1-n2+1, "bad dimension in tquo");
  if (n1 < n2 || n2 == 0) return Polynomial ();
  nat l= aligned_size<C,V> (n1);
  C* q= mmx_new<C> (l);
  C* r= mmx_new<C> (l);
  V::clear (r, n1);
  V::copy  (r+n2-1, seg (P1), n);
  V::tquo_rem (q, r, seg (P2), n1, n2);
  mmx_delete<C> (r, l);
  return Polynomial (q, n1, l);
}

TMPL Polynomial
rem (const Polynomial& P1, const Polynomial& P2) {
  nat n1= N(P1), n2= N(P2);
  if (n1 < n2 || n2 == 0) return P1;
  nat lq= aligned_size<C,V> (n1-n2+1);
  nat lr= aligned_size<C,V> (n1);
  C* q= mmx_new<C> (lq);
  C* r= mmx_new<C> (lr);
  V::copy (r, seg (P1), n1);
  V::quo_rem (q, r, seg (P2), n1, n2);
  mmx_delete<C> (q, lq);
  return Polynomial (r, n1, lr);
}

//! Transposed remainder in the division in degree d1 by P2.
TMPL Polynomial
trem (int d1, const Polynomial& P1, const Polynomial& P2) {
  nat n1= max (0, d1+1), n= N(P1), n2= N(P2);
  ASSERT (n <= n2-1, "bad dimension in trem");
  if (n1 < n2 || n2 == 0) return P1;
  nat l= aligned_size<C,V> (n1);
  C* q= mmx_new<C> (l);
  C* r= mmx_new<C> (l);
  V::copy  (r, seg (P1), n);
  V::clear (r+n, n1-n);
  V::tquo_rem (q, r, seg (P2), n1, n2);
  mmx_delete<C> (r, l);
  return Polynomial (q, n1, l);
}

//! Pseudo division
//! lc(P2)^pexponent (P1,P2) P1 = pquo(P1,P2) * P2 + prem(P1,P2)
TMPL inline int
pexponent (const Polynomial& P1, const Polynomial& P2) {
  return max (0, deg (P1) - deg (P2) + 1);
}

TMPL Polynomial
pquo (const Polynomial& P1, const Polynomial& P2) {
  nat n1= N(P1), n2= N(P2);
  if (n1 < n2 || n2 == 0) return Polynomial ();
  nat lq= aligned_size<C,V> (n1-n2+1);
  nat lr= aligned_size<C,V> (n1);
  C* q= mmx_new<C> (lq);
  C* r= mmx_new<C> (lr);
  V::copy (r, seg (P1), n1);
  V::pquo_rem (q, r, seg (P2), n1, n2);
  mmx_delete<C> (r, lr);
  return Polynomial (q, n1-n2+1, lq);
}

TMPL Polynomial
prem (const Polynomial& P1, const Polynomial& P2) {
  nat n1= N(P1), n2= N(P2);
  if (n1 < n2 || n2 == 0) return P1;
  nat lq= aligned_size<C,V> (n1-n2+1);
  nat lr= aligned_size<C,V> (n1);
  C* q= mmx_new<C> (lq);
  C* r= mmx_new<C> (lr);
  V::copy (r, seg (P1), n1);
  V::pquo_rem (q, r, seg (P2), n1, n2);
  mmx_delete<C> (q, lq);
  return Polynomial (r, n1, lr);
}

/******************************************************************************
* Gcd computations
******************************************************************************/

// gcd (P1, P2) = U1 P1 + U2 P2, with
// deg U1 < deg P2 and deg U2 < deg P1 (for non-degenerated cases)

TMPL Polynomial
gcd (const Polynomial& P1, const Polynomial& P2) {
  nat n1= N(P1), n2= N(P2);
  if (n2 == 0) return P1;
  if (n1 == 0) return P2;
  nat l= aligned_size<C,V> (min (n1, n2));
  C* g= mmx_new<C> (l); nat n;
  V::gcd (g, n, seg (P1), seg (P2), n1, n2);
  return Polynomial (g, n, l);
}

TMPL Polynomial
gcd (const Polynomial& P1, const Polynomial& P2,
     Polynomial& U1) {
  nat n1= N(P1), n2= N(P2);
  if (n2 == 0) { U1= 1; return P1; }
  if (n1 == 0) { U1= 0; return P2; }
  nat n= min (n1, n2), nu1;
  nat l = aligned_size<C,V> (n);
  nat l2= aligned_size<C,V> (n2);
  C* g= mmx_new<C> (l), * u1= mmx_new<C> (l2);
  V::gcd (g, n, seg (P1), seg (P2), n1, n2, u1, nu1);
  U1= Polynomial (u1, nu1, l2); return Polynomial (g, n, l);
}

TMPL Polynomial
gcd (const Polynomial& P1, const Polynomial& P2,
     Polynomial& U1, Polynomial& U2) {
  nat n1= N(P1), n2= N(P2);
  if (n2 == 0) { U1= 1; U2= 0; return P1; }
  if (n1 == 0) { U1= 0; U2= 1; return P2; }
  nat n = min (n1, n2), nu1, nu2;
  nat l = aligned_size<C,V> (n);
  nat l1= aligned_size<C,V> (n1), l2= aligned_size<C,V> (n2);
  C* g= mmx_new<C> (l), * u1= mmx_new<C> (l2), * u2= mmx_new<C> (l1);
  V::gcd(g, n, seg (P1), seg (P2), n1, n2, u1, u2, nu1, nu2);
  U1= Polynomial (u1, nu1, l2); U2= Polynomial (u2, nu2, l1);
  return Polynomial (g, n, l);
}

TMPL Polynomial
lcm (const Polynomial& P1, const Polynomial& P2) {
  return P1 * (P2 / gcd (P1, P2));
}

TMPL Polynomial
invert_modulo (const Polynomial& P, const Polynomial& Q) {
  // inverse of P modulo Q, return 0 if non-invertible
  Polynomial inv_P;
  Polynomial G= gcd (P, Q, inv_P);
  if (deg (G) != 0) inv_P= 0;
  return inv_P / G[0];
}

TMPL void
pade (const Polynomial& P, nat n, nat k,
      Polynomial& Num, Polynomial &Den) {
  // In return P = Num/Den + O(x^n), with
  // deg (Num) < k and deg (Den) <= n-k.
  // Note that Num and Den are not ensured to be coprime.
  VERIFY (k <= n, "invalid argument");
  nat m= N(P);
  if (n == 0 || m == 0) { Num= 0; Den= 1; return; }
  nat lnum = aligned_size<C,V> (k), lden= aligned_size<C,V> (n-k+1);
  C* num= mmx_new<C> (lnum), * den= mmx_new<C> (lden);
  V::pade (num, den, seg (P), m, n, k);
  Num= Polynomial (num, k, lnum); Den= Polynomial (den, n-k+1, lden);
}

/******************************************************************************
* Valuation and comparisons
******************************************************************************/

TMPL int
val (const Polynomial& P) {
  for (nat i=0; i<N(P); i++)
    if (P[i] != 0) return (int) i;
  return (int) (((nat) (-1)) >> 1);
}

TMPL int
sign (const Polynomial& P) {
  nat n= N(P);
  if (n == 0) return 0;
  return sign (P[n-1]);
}

TMPL int
compare (const Polynomial& P1, const Polynomial& P2) {
  nat n1= N(P1), n2= N(P2);
  if (n1 > n2) return sign (P1[n1-1]);
  if (n2 > n1) return -sign (P2[n2-1]);
  return V::cmp (seg (P1), seg (P2), n1);
}

template<typename Op,typename C,typename V> inline bool
binary_test (const Polynomial& P1, const Polynomial& P2) {
  nat n1= N(P1), n2= N(P2);
  if (n1 != n2) return false;
  for (nat i=0; i<n1; i++)
    if (Op::not_op (P1[i], P2[i])) return false;
  return true;
}

/******************************************************************************
* Differentiation, integration and composition
******************************************************************************/

TMPL Polynomial
derive (const Polynomial& P) {
  nat n= N(P);
  if (n <= 1) return 0;
  nat l= aligned_size<C,V> (n-1);
  C* r= mmx_new<C> (l);
  V::derive (r, seg (P), n);
  return Polynomial (r, n-1, l);
}

TMPL Polynomial
xderive (const Polynomial& P) {
  nat n= N(P);
  nat l= aligned_size<C,V> (n);
  C* r= mmx_new<C> (l);
  V::xderive (r, seg (P), n);
  return Polynomial (r, n, l);
}

TMPL Polynomial
integrate (const Polynomial& P) {
  nat n= N(P);
  nat l= aligned_size<C,V> (n+1);
  C* r= mmx_new<C> (l);
  V::integrate (r, seg (P), n);
  return Polynomial (r, n+1, l);
}

template<typename C,typename V,typename K> Polynomial
compose (const Polynomial& P1, const polynomial<K,V>& P2) {
  nat n1= N(P1), n2= N(P2);
  if (n1 <= 1) return P1;
  if (n2 == 0) return P1[0];
  nat n= (n1-1) * (n2-1) + 1;
  nat l= aligned_size<C,V> (n);
  C* r= mmx_new<C> (l);
  V::compose (r, seg (P1), seg (P2), n1, n2);
  return Polynomial (r, n, l);
}

TMPL Polynomial
shift (const Polynomial& P, const C& sh) {
  nat n= N(P);
  if (n <= 1 || sh == 0) return P;
  nat l= aligned_size<C,V> (n);
  C* r= mmx_new<C> (l);
  V::shift (r, seg (P), sh, n);
  return Polynomial (r, n, l);
}

TMPL Polynomial
shift (const Polynomial& P, int i) {
  return shift (P, C (i));
}

template<typename C,typename V,typename K> Polynomial
q_difference (const Polynomial& P, const K& q) {
  nat n= N(P);
  if (n <= 1) return P;
  nat l= aligned_size<C,V> (n);
  C* r= mmx_new<C> (l);
  V::q_difference (r, seg (P), q, n);
  return Polynomial (r, n, l);
}

TMPL Polynomial
dilate (const Polynomial& P, nat p) {
  if (p == 1) return P;
  nat n= N(P);
  if (n <= 1) return P;
  nat k= (n-1)*p + 1;
  nat l= aligned_size<C,V> (k);
  C* r= mmx_new<C> (l);
  V::dilate (r, seg (P), p, n);
  return Polynomial (r, k, l);
}

/******************************************************************************
* Evaluation
******************************************************************************/

TMPL Polynomial
annulator_bis (const Vector& v) {
  ASSERT (is_non_scalar (v), "non-scalar vector expected");
  nat l= aligned_size<C,V> (N(v)+1);
  C* r= mmx_new<C> (l);
  V::annulator (r, seg(v), N(v));
  return Polynomial (r, N(v)+1, l);
}

template<typename C> inline polynomial<C,typename Polynomial_variant(C) >
annulator (const Vector& v) {
  return annulator_bis<C,typename Polynomial_variant(C) > (v);
}

TMPL C
eval (const Polynomial& p, const C& c) {
  return V::eval (seg (p), c, N(p));
}

TMPL Vector
eval (const Polynomial& p, const Vector& v) {
  typedef typename Vector_variant(C) VV;
  ASSERT (is_non_scalar (v), "non-scalar vector expected");
  nat l= aligned_size<C,VV> (N(v));
  C* r= mmx_new<C> (l);
  V::eval (r, seg(p), seg(v), N(p), N(v));
  return Vector (r, N(v), l);
}

TMPL vector<Polynomial >
expand (const Polynomial& p, const Vector& v, const vector<nat>& mu) {
  nat k= N(v);
  ASSERT (N(mu) == k, "dimensions don't match");
  C** r= mmx_new<C*> (k);
  for (nat i=0; i<k; i++)
    r[i]= mmx_new<C> (aligned_size<C,V> (mu[i]));
  V::expand (r, seg (p), seg (v), seg (mu), N (p), k);
  nat l= aligned_size<Polynomial, typename Vector_variant (Polynomial)> (k);
  Polynomial* ret= mmx_new<Polynomial > (l);
  for (nat i=0; i<k; i++)
    ret[i]= Polynomial (r[i], mu[i], aligned_size<C,V> (mu[i]));
  mmx_delete<C*> (r, k);
  return vector<Polynomial > (ret, k, l);
}

/******************************************************************************
* Miscellaneous
******************************************************************************/

TMPL Polynomial
range (const Polynomial& P, nat start, nat end) {
  nat l= aligned_size<C,V> (end-start);
  C* r= mmx_new<C> (l);
  for (nat i=start; i<end; i++) r[i-start]= P[i];
  return Polynomial (r, end-start, l);
}

TMPL Polynomial
extract_mod (const Polynomial& P, nat k, nat p) {
  nat n= (N(P) - k + p - 1) / p;
  nat l= aligned_size<C,V> (n);
  C* r= mmx_new<C> (l);
  for (nat i=0; i<n; i++) r[i]= P[i*p+k];
  return Polynomial (r, n, l);
}

TMPL Polynomial
lshiftz (const Polynomial& P, const int& shift) {
  if (shift == 0) return P;
  else if (shift <= -((int) N(P))) return 0;
  else if (shift < 0) return range (P, (nat) (-shift), N(P));
  else {
    nat n= N(P) + shift;
    nat l= aligned_size<C,V> (n);
    C* r= mmx_new<C> (l, C(0));
    V::copy (r+shift, seg (P), N(P));
    return Polynomial (r, n, l);
  }
}

TMPL Polynomial
reverse (const Polynomial& P) {
  nat n= N(P);
  nat l= aligned_size<C,V> (n);
  C* r= mmx_new<C> (l);
  V::Vec::vec_reverse (r, seg (P), n);
  return Polynomial (r, n, l);
}

TMPL Polynomial
graeffe (const Polynomial& P) {
  nat n= N(P);
  nat l= aligned_size<C,V> (n);
  C* r= mmx_new<C> (l);
  V::graeffe (r, seg (P), n);
  return Polynomial (r, n, l);
}

TMPL Polynomial
invert_lo (const Polynomial& P) {
  nat n= N(P);
  nat l= aligned_size<C,V> (n);
  C* r= mmx_new<C> (l);
  V::invert_lo (r, seg (P), n);
  return Polynomial (r, n, l);
}

TMPL Polynomial
invert_hi (const Polynomial& P) {
  nat n= N(P);
  nat l= aligned_size<C,V> (n);
  C* r= mmx_new<C> (l);
  V::invert_hi (r, seg (P), n);
  return Polynomial (r, n, l);
}

/******************************************************************************
* Syntactic sugar
******************************************************************************/

TRUE_IDENTITY_OP_SUGAR(TMPL,Polynomial)
SOFT_IDENTITY_OP_SUGAR(TMPL,Polynomial)
ARITH_INT_SUGAR(TMPL,Polynomial,C)
SHIFT_INT_SUGAR(TMPL,Polynomial)
COMPARE_SUGAR(TMPL,Polynomial)
EQUAL_INT_SUGAR(TMPL,Polynomial,Polynomial)
COMPARE_INT_SUGAR(TMPL,Polynomial,Polynomial)

#undef TMPL_DEF
#undef TMPL
#undef Polynomial
#undef Polynomial_rep
#undef Vector
} // namespace mmx
#endif // __MMX_POLYNOMIAL_HPP
