
/******************************************************************************
* MODULE     : polynomial_complex.hpp
* DESCRIPTION: Complex multiplication of polynomials
* COPYRIGHT  : (C) 2007  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__POLYNOMIAL_COMPLEX__HPP
#define __MMX__POLYNOMIAL_COMPLEX__HPP
#include <algebramix/polynomial.hpp>
#include <algebramix/polynomial_dicho.hpp>
#include <numerix/complex.hpp>

namespace mmx {
#define TMPL template<typename C>

/******************************************************************************
* Complex multiplication
******************************************************************************/

template<typename V>
struct polynomial_complex_multiplication: public V {
  typedef typename V::Naive Naive;
  typedef typename V::Positive Positive;
  typedef polynomial_complex_multiplication<typename V::No_simd> No_simd;
  typedef polynomial_complex_multiplication<typename V::No_thread> No_thread;
  typedef polynomial_complex_multiplication<typename V::No_scaled> No_scaled;

TMPL static inline void
mul (complex<C>* dest, const complex<C>* s1, const complex<C>* s2,
     nat n1, nat n2)
{
  nat nd  = n1+n2-1;
  nat spc1= ((n1+1)>>1) << 1; // force 128 bit alignment if necessary
  nat spc2= ((n2+1)>>1) << 1;
  nat spcd= ((nd+1)>>1) << 1;
  nat spc = 5*(spc1+spc2);
  C* real_x1= mmx_new<C> (spc);
  C* imag_x1= real_x1 + spc1;
  C* real_x2= imag_x1 + spc1;
  C* imag_x2= real_x2 + spc2;
  C* real_d = imag_x2 + spc2;
  C* imag_d = real_d  + spcd;
  C* temp_d = imag_d  + spcd;
  
  V::Vec::half_copy (real_x1, (C*) ((void*) s1), n1);
  V::Vec::half_copy (imag_x1, ((C*) ((void*) s1)) + 1, n1);
  V::Vec::half_copy (real_x2, (C*) ((void*) s2), n2);
  V::Vec::half_copy (imag_x2, ((C*) ((void*) s2)) + 1, n2);
  V::mul (real_d, real_x1, real_x2, n1, n2);
  V::mul (temp_d , imag_x1, imag_x2, n1, n2);
  V::add (real_x1, imag_x1, n1);
  V::add (real_x2, imag_x2, n2);
  V::mul (imag_d, real_x1, real_x2, n1, n2);
  V::sub (imag_d, real_d, nd);
  V::sub (imag_d, temp_d, nd);
  V::sub (real_d, temp_d, nd);
  V::Vec::double_copy ((C*) ((void*) dest), real_d, nd);
  V::Vec::double_copy (((C*) ((void*) dest)) + 1, imag_d, nd);

  mmx_delete<C> (real_x1, spc);
}

TMPL static inline void
square (complex<C>* dest, const complex<C>* s1, nat n1) {
  nat nd  = n1+n1-1;
  nat spc1= ((n1+1)>>1) << 1; // force 128 bit alignment if necessary
  nat spcd= ((nd+1)>>1) << 1;
  nat spc = 8*spc1;
  C* real_x1= mmx_new<C> (spc);
  C* imag_x1= real_x1 + spc1;
  C* real_d = imag_x1 + spc1;
  C* imag_d = real_d  + spcd;
  C* temp_d = imag_d  + spcd;
  
  V::Vec::half_copy (real_x1, (C*) ((void*) s1), n1);
  V::Vec::half_copy (imag_x1, ((C*) ((void*) s1)) + 1, n1);
  V::square (real_d, real_x1, n1);
  V::square (temp_d, imag_x1, n1);
  V::add (real_x1, imag_x1, n1);
  V::square (imag_d, real_x1, n1);
  V::sub (imag_d, real_d, nd);
  V::sub (imag_d, temp_d, nd);
  V::sub (real_d, temp_d, nd);
  V::Vec::double_copy ((C*) ((void*) dest), real_d, nd);
  V::Vec::double_copy (((C*) ((void*) dest)) + 1, imag_d, nd);

  mmx_delete<C> (real_x1, spc);
}

TMPL static inline void
tmul (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
  (void) dest; (void) s1; (void) s2; (void) n1; (void) n2;
  ERROR ("transposed complexified multiplication not implemented");
}

}; // polynomial_complex_multiplication<V>

template<typename V>
struct polynomial_multiply_complex:
  public polynomial_complex_multiplication<V>
{};

/******************************************************************************
* Use complexified multiplication by default
******************************************************************************/

template<typename C>
struct polynomial_variant_helper<complex<C> > {
  typedef typename Vector_variant(C) CVV;
  typedef typename Polynomial_multiply(C) CMV;
  typedef typename Polynomial_variant(C) CPV;

  typedef typename Vector_variant(complex<C>) VV;
  typedef polynomial_multiply_complex<CMV> MV;
  typedef polynomial_variable_name_local<polynomial_algos_dicho<MV> > PV;
};

STMPL
struct polynomial_variant_helper<complex<generic> > {
  typedef vector_naive VV;
  typedef polynomial_multiply_naive<VV> MV;
  typedef polynomial_naive PV;
};

#undef TMPL
} // namespace mmx
#endif //__MMX__POLYNOMIAL_COMPLEX__HPP
