
/******************************************************************************
* MODULE     : polynomial_dicho.hpp
* DESCRIPTION: dichotomic algorithms including Karatsuba multiplication
* COPYRIGHT  : (C) 2003  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__POLYNOMIAL_DICHO__HPP
#define __MMX__POLYNOMIAL_DICHO__HPP
#include <algebramix/polynomial_naive.hpp>

namespace mmx {
#define TMPL template<typename C>

/******************************************************************************
* Helper class for thresholds
******************************************************************************/

struct dicho_threshold {};
template<nat n> struct uniform_threshold {};

template<typename C, typename Th>
struct threshold_helper {
  typedef fixed_value<nat,2> impl;
};

template<typename C, nat n>
struct threshold_helper<C,uniform_threshold<n> > {
  typedef fixed_value<nat,n> impl;
};

#define Threshold(C,Th) threshold_helper<C,Th >::impl::val

/******************************************************************************
* Multiplication
******************************************************************************/

template<typename V, typename Th= dicho_threshold>
struct polynomial_dicho_multiplication: public V {

TMPL static void
multiply (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
  if (n1 < Threshold(C,Th) || n2 < Threshold(C,Th))
    V::mul (dest, s1, s2, n1, n2);
  else {
    nat p1= n1 >> 1, p2= n2 >> 1, P= p1+p2-1;
    nat spc= aligned_size<C,V> (3*(p1+p2));
    C* buf= mmx_new<C> (spc);
    C* low1= dest;
    C* low2= low1 + p1;
    C* mid1= buf;
    C* mid2= buf + p1;
    C* hi1 = mid2 + p2;
    C* hi2 = hi1 + p1;
    C* Low = mid1;
    C* Mid = hi1;
    C* Hi  = Mid + p1 + p2;
    V::Vec::half_copy (low1 , s1  , p1);
    V::Vec::half_copy (hi1  , s1+1, p1);
    V::add            (mid1 , low1, hi1 , p1);
    V::Vec::half_copy (low2 , s2  , p2);
    V::Vec::half_copy (hi2  , s2+1, p2);
    V::add            (mid2 , low2, hi2 , p2);
    multiply          (Hi   , hi1 , hi2 , p1, p2);
    multiply          (Mid  , mid1, mid2, p1, p2);
    multiply          (Low  , low1, low2, p1, p2);
    V::sub            (Mid  , Low , P);
    V::sub            (Mid  , Hi  , P);
    V::add            (Low+1, Hi  , P-1);
    Low[P]= Hi[P-1];
    V::Vec::double_copy (dest  , Low, P+1);
    V::Vec::double_copy (dest+1, Mid, P);
    mmx_delete<C> (buf, spc);

    if ((n1 & 1) != 0) {
      dest[(P<<1)+1]= C(0);
      V::mul_add (dest + (p1<<1), s2, s1[p1<<1], p2<<1);
    }
    if ((n2 & 1) != 0) {
      dest[n1+n2-2]= C(0);
      V::mul_add (dest + (p2<<1), s1, s2[p2<<1], n1);
    }
  }
}

TMPL static inline void
mul (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
  if (n1 == 0 && n2 == 0) return;
  if (n1 == 0 || n2 == 0)
    V::clear (dest, n1 + n2 - 1);
  else
    multiply (dest, s1, s2, n1, n2);
}

TMPL static void
square (C* dest, const C* s, nat n) {
  if (n == 0) return;
  if (n < Threshold(C,Th))
    V::square (dest, s, n);
  else {
    nat p= n >> 1, P= 2*p-1, spc= aligned_size<C,V> (6*p);
    C* buf= mmx_new<C> (spc);
    C* low= dest;
    C* mid= buf;
    C* hi = buf + p;
    C* Low= buf;
    C* Mid= Low + 2*p;
    C* Hi = Mid + 2*p;
    V::Vec::half_copy (low  , s  , p);
    V::Vec::half_copy (hi   , s+1, p);
    V::add            (mid  , low, hi, p);
    square            (Hi   , hi , p);
    square            (Mid  , mid, p);
    square            (Low  , low, p);
    V::sub            (Mid  , Low, P);
    V::sub            (Mid  , Hi , P);
    V::add            (Low+1, Hi , P-1);
    Low[P]= Hi[P-1];
    V::Vec::double_copy (dest  , Low, P+1);
    V::Vec::double_copy (dest+1, Mid, P);
    mmx_delete<C> (buf, spc);

    if ((n & 1) != 0) {
      dest[(P<<1)+1]= C(0);
      V::Vec::template vec_binary_scalar<lmul_add_op,C,C,C>
	(dest + (p<<1), s, s[p<<1], p<<1);
      dest[2*n-2]= C(0);
      V::Vec::template vec_binary_scalar<rmul_add_op,C,C,C>
	(dest + (p<<1), s, s[p<<1], n);
    }
  }
}

TMPL static void
tmultiply (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
  if (n1 < Threshold(C,Th) || n2 < Threshold(C,Th))
    V::tmul (dest, s1, s2, n1, n2);
  else {
    nat p1= n1 >> 1, p2= n2 >> 1, P= p1+p2-1;
    nat spc= aligned_size<C,V> (3*(p1+P)+1);
    C* buf= mmx_new<C> (spc);
    C* low1= buf;
    C* low2= low1 + p1;
    C* mid1= low2 + P+1;
    C* mid2= mid1 + p1;
    C* hi1 = mid2 + P;
    C* hi2 = hi1  + p1;
    C* Low = mid1;
    C* Mid = hi1;
    C* Hi  = dest;
    V::Vec::half_copy (low1, s1  , p1);
    V::Vec::half_copy (hi1 , s1+1, p1);
    V::add            (mid1, low1, hi1, p1);
    V::Vec::half_copy (low2, s2  , P+1);
    V::Vec::half_copy (hi2 , s2+1, P);
    V::add            (mid2, low2+1, hi2, P);
    tmultiply (Hi , hi1 , hi2 , p1, p2);
    tmultiply (Mid, mid1, mid2, p1, p2);
    tmultiply (Low, low1, low2, p1, p2+1);
    V::sub (Mid, Hi   , p2);
    V::sub (Mid, Low+1, p2);
    V::add (Low, Hi   , p2);
    V::Vec::double_copy (dest  , Low, p2);
    V::Vec::double_copy (dest+1, Mid, p2);
    mmx_delete<C> (buf, spc);

    if ((n1 & 1) != 0)
      V::mul_add (dest, s2+n1-1, s1[n1-1], n2);

    if ((n2 & 1) != 0)
      dest[n2-1]= V::Vec::inn_prod (s1, s2+n2-1, n1);
  }
}

TMPL static inline void
tmul (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
  // transposed multiplication
  // s1 has length n1, dest has length n2 and s2 has length n1+n2-1
  V::clear (dest, n2);
  if (n1 == 0 || n2 == 0) return;
  tmultiply (dest, s1, s2, n1, n2);
}

}; // polynomial_dicho_multiplication<V>

/******************************************************************************
* Division
******************************************************************************/

template<typename V, typename Th= dicho_threshold>
struct polynomial_dicho_division: public V {

TMPL static void
quo_rem (C* dest, C* s1, const C* s2, nat n1, nat n2) {
  // NOTE: C can be an euclidean ring (not necessarily a field)
  if (n1 < n2);
  else if (n1 < Threshold(C,Th))
    V::quo_rem (dest, s1, s2, n1, n2);
  else {
    nat l= n1-n2+1;
    if (n2 > l) {
      nat d= n2-l;
      quo_rem (dest, s1+d, s2+d, n1-d, l);
      nat buf_size= aligned_size<C,V> (l+d-1);
      C* buf= mmx_new<C> (buf_size);
      V::mul (buf, dest, s2, l, d);
      V::sub (s1, buf, l+d-1);
      mmx_delete<C> (buf, buf_size);      
    }
    else {
      nat h= (n2+1) >> 1;
      nat d= l-h;
      quo_rem (dest+d, s1+d, s2, n1-d, n2);
      quo_rem (dest, s1, s2, n1-h, n2);
    }
  }
}

TMPL static void
tquo_rem (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
  // (s1, n2-1) contains the part corresponding to the remainder, while
  // (s2+n2-1, n1-n2+1) occupies the one of the quotient.
  // We assume n2>0 and s2[n2-1] != 0.
  // (dest, n1) contains the output. 
 // NOTE: C can be an euclidean ring (not necessarily a field)
  if (n1 < n2)
    V::copy (dest, s1, n1);
  else if (n1 < Threshold(C,Th))
    V::tquo_rem (dest, s1, s2, n1, n2);
  else {
    nat l= n1-n2+1;
    if (n2 > l) {
      nat d= n2-l;
      nat buf_size= aligned_size<C,V> (l+d-1);
      C* buf= mmx_new<C> (buf_size);
      nat tmp_size= aligned_size<C,V> (n1);
      C* tmp= mmx_new<C> (tmp_size);
      V::neg      (buf     , s1     , l+d-1);
      V::tmul     (tmp+n2-1, s2     , buf, d, l);
      V::add      (tmp+n2-1, s1+n2-1, l);
      V::copy     (tmp     , s1     , n2-1);
      V::tquo_rem (dest+d  , tmp+d  , s2+d, n1-d, l);
      V::copy     (dest    , s1     , d);
      mmx_delete<C> (buf, buf_size);
      mmx_delete<C> (tmp, tmp_size);
    }
    else {
      nat h= (n2+1) >> 1;
      nat d= l-h;
      nat tmp_size= aligned_size<C,V> (n1-d);
      C* tmp= mmx_new<C> (tmp_size);
      tquo_rem (dest    , s1      , s2, n1-h, n2);
      V::copy  (tmp     , dest + d, n2-1);
      V::copy  (tmp+n2-1, s1+n1-h , h);
      tquo_rem (dest+d  , tmp     , s2, n1-d, n2);
      mmx_delete<C> (tmp, tmp_size);
    }
  }
}

TMPL static void
pquo_rem (C* dest, C* s1, const C* s2, nat n1, nat n2) {
  if (n1 < n2);
  else if (n1 < Threshold(C,Th))
    V::pquo_rem (dest, s1, s2, n1, n2);
  else {
    nat l= n1-n2+1;
    if (n2 > l) {
      nat d= n2-l;
      pquo_rem (dest, s1+d, s2+d, n1-d, l);
      nat buf_size= aligned_size<C,V> (l+d-1);
      C* buf= mmx_new<C> (buf_size);
      V::mul (buf, dest, s2, l, d);
      V::mul_sc (s1, binpow (s2[n2-1], l), d);
      V::sub (s1, buf, l+d-1);
      mmx_delete<C> (buf, buf_size);      
    }
    else {
      nat h= (n2+1) >> 1;
      nat d= l-h;
      pquo_rem (dest+d, s1+d, s2, n1-d, n2);
      V::mul_sc (s1, binpow (s2[n2-1], h), d);
      pquo_rem (dest, s1, s2, n1-h, n2);
      V::mul_sc (dest+d, binpow (s2[n2-1], d), h);
    }
  }
}

}; // polynomial_dicho_division<V>

template<typename V, typename Th= dicho_threshold>
struct polynomial_newton_division: public V {

TMPL static void
invert_lo (C* dest, const C* src, nat n) {
  if (n == 1) *dest= C(1) / *src;
  else {
    nat h= (n+1) >> 1;
    nat l= n - h;
    invert_lo (dest, src, h);
    nat buf_size= aligned_size<C> (n << 1);
    C* buf= mmx_new<C> (buf_size);
    C* aux= buf + n;
    V::mul (buf, src, dest, n, h);
      // FIXME: use middle product
    V::mul (aux, dest, buf + h, l, l);
      // FIXME: use truncated product
    V::neg (dest + h, aux, l);
    mmx_delete<C> (buf, buf_size);
  }
}

TMPL static void
invert_hi (C* dest, const C* src, nat n) {
  if (n == 1) *dest= C(1) / *src;
  else {
    nat h= (n+1) >> 1;
    nat l= n - h;
    invert_hi (dest + l, src + l, h);
    nat buf_size= aligned_size<C,V> (n << 1);
    C* buf= mmx_new<C> (buf_size);
    C* aux= buf + l;
    V::mul (aux, src, dest + l, n, h);
      // FIXME: use middle product
    V::mul (buf, dest + h, aux + h - 1, l, l);
      // FIXME: use truncated product
    V::neg (dest, buf + l - 1, l);
    mmx_delete<C> (buf, buf_size);
  }
}

TMPL static void
quo_rem (C* dest, C* s1, const C* s2, nat n1, nat n2) {
  // NOTE: requires C to be a field
  if (n1 < n2);
  else if (n1 < Threshold(C,Th))
    V::quo_rem (dest, s1, s2, n1, n2);
  else {
    nat tot= aligned_size<C,V> ((n2 << 1) + n2);
    C* buf= mmx_new<C> (tot);
    C* inv= buf + (n2 << 1);
    nat nq = n1 + 1 - n2;
    nat l  = min (n2, nq);
    invert_hi (inv + n2 - l, s2 + n2 - l, l);
    while (n1 >= n2) {
      nat nq= n1 + 1 - n2;
      nat l = min (n2, nq);
      C* dest_hi= dest + nq - l;
      V::mul (buf, s1 + n1 - l, inv + n2 - l, l, l);
      V::copy (dest_hi, buf + l - 1, l);
      V::mul (buf, dest_hi, s2, l, n2);
      V::sub (s1 + n1 - (n2 + l - 1), buf, n2 + l - 1);
      n1 -= l;
    }
    mmx_delete<C> (buf, tot);
  }
}

TMPL static void
tquo_rem (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
  // (s1, n2-1) contains the part corresponding to the remainder, while
  // (s2+n2-1, n1-n2+1) occupies the one of the quotient.
  // We assume n2>0 and s2[n2-1] != 0.
  // (dest, n1) contains the output.
  // NOTE: requires C to be a field
  if (n1 < n2)
    V::copy (dest, s1, n1);
  else if (n1 < Threshold(C,Th))
    V::tquo_rem (dest, s1, s2, n1, n2);
  else {
    nat tot= aligned_size<C,V> (3 * n2);
    C* buf= mmx_new<C> (tot);
    C*  inv= buf + 2*n2;
    nat nq = n1 + 1 - n2;
    nat l  = min (n2, nq);
    invert_hi (inv + n2 - l, s2 + n2 - l, l);
    V::neg (inv + n2 - l, inv + n2 - l, l);
    V::copy (dest, s1, n2-1);
    V::clear (dest+n2-1, n1-n2+1);
    nat m = n2-1;
    while (m < n1) {
      nat nq= n1 - m;
      nat l = min (n2, nq);
      V::clear (buf     , l-1);
      V::tmul  (buf+l-1 , s2          , dest + m - (n2 - 1), n2, l);
      V::sub   (buf+l-1 , s1 + m      , l);      
      V::tmul  (dest + m, inv + n2 - l, buf, l , l);
      m += l;
    }
    mmx_delete<C> (buf, tot);
  }
}

// Pseudo division. C can be any ring.
TMPL static void
pinvert_hi (C* dest, const C* src, nat n) {
  if (n == 1) *dest= C(1);
  else {
    nat h= (n+1) >> 1;
    nat l= n - h;
    nat tmp_size= aligned_size<C,V> (l);
    C* tmp= mmx_new<C> (tmp_size);
    pinvert_hi (dest + l, src + l, h);
    pinvert_hi (tmp, src + h, l); // FIXME: increase from l to h
    nat buf_size= aligned_size<C,V> (n << 1);
    C* buf= mmx_new<C> (buf_size);
    C* aux= buf + l;
    V::mul (aux, src, dest + l, n, h);
      // FIXME: use middle product
    V::mul_sc (dest + l, binpow (src[n-1], l), h);
    V::mul (buf, tmp, aux + h - 1, l, l);
      // FIXME: use truncated product
    V::neg (dest, buf + l - 1, l);
    mmx_delete<C> (buf, buf_size);
    mmx_delete<C> (tmp, tmp_size);
  }
}

TMPL static void
pquo_rem (C* dest, C* s1, const C* s2, nat n1, nat n2) {
  if (n1 < n2);
  else if (n1 < Threshold(C,Th))
    V::pquo_rem (dest, s1, s2, n1, n2);
  else {
    nat tot= aligned_size<C,V> ((n2 << 1) + n2);
    C* buf= mmx_new<C> (tot);
    C* inv= buf + (n2 << 1);
    nat tmp_size= aligned_size<C,V> (n2);
    C* tmp= mmx_new<C> (tmp_size);
    nat nq= n1 + 1 - n2;
    nat l = min (n2, nq);
    nat l_end= nq % l;
    pinvert_hi (inv + n2 - l, s2 + n2 - l, l);
    if (l_end != 0)
      pinvert_hi (tmp + n2 - l_end, s2 + n2 - l_end, l_end);
    while (n1 >= n2) {
      nat nq= n1 + 1 - n2;
      nat l = min (n2, nq);
      C* dest_hi= dest + nq - l;
      if (l == l_end)
	V::mul (buf, s1 + n1 - l, tmp + n2 - l, l, l);
      else
	V::mul (buf, s1 + n1 - l, inv + n2 - l, l, l);
      V::copy (dest_hi, buf + l - 1, l);
      V::mul (buf, dest_hi, s2, l, n2);
      V::mul_sc (s1, binpow (s2[n2-1], l), n1);
      V::sub (s1 + n1 - (n2 + l - 1), buf, n2 + l - 1);
      V::mul_sc (dest_hi, binpow (s2[n2-1], nq - l), l);
      n1 -= l;
    }
    mmx_delete<C> (buf, tot);
    mmx_delete<C> (tmp, tmp_size);
  }
}

}; // polynomial_newton_division<V>

/******************************************************************************
* Dichotomic Gcd computations
******************************************************************************/

// "Modern Computer Algebra", Algorithm 11.4

struct dicho_gcd_threshold {};

template<typename V, typename Th= dicho_gcd_threshold>
struct polynomial_dicho_euclidean: public V {
private:

TMPL static void
dot_product (C* d, nat& nd,
	     const C* r0, const C* r1, nat nr0, nat nr1,
	     const C* s0, const C* s1, nat ns0, nat ns1,
	     C* t) {
  // d = r0 s0 + r1 s1
  if ((nr0 == 0 || ns0 == 0) && (nr1 == 0 || ns1 == 0)) {
    nd = 0; return;
  }
  if (nr0 == 0 || ns0 == 0) {
    nd = nr1 + ns1 - 1;
    V::mul (d, r1, s1, nr1, ns1);
    V::trim (d, nd);
    return;
  }
  if (nr1 == 0 || ns1 == 0) {
    nd = nr0 + ns0 - 1;
    V::mul (d, r0, s0, nr0, ns0);
    V::trim (d, nd);
  }
  if (nr0 + ns0 - 1 < nr1 + ns1 - 1) {
    nd = nr1 + ns1 - 1;
    V::mul (d, r1, s1, nr1, ns1);
    V::mul (t, r0, s0, nr0, ns0);
    V::add (d, t , nr0 + ns0 - 1);
  }
  else {
    nd = nr0 + ns0 - 1;
    V::mul (d, r0, s0, nr0, ns0);
    V::mul (t, r1, s1, nr1, ns1);
    V::add (d, t , nr1 + ns1 - 1);
  }
  V::trim (d, nd);
}
  
TMPL static void
matrix_vector_product (C* r0, C* r1, nat& nr0, nat& nr1,
		       const C* R00, const C* R01, const C* R10, const C* R11,
		       nat nR00, nat nR01, nat nR10, nat nR11,
		       const C* s0, const C* s1, nat n0, nat n1,
		       C* tp) {
  // (r0,r1) = R . (s0,s1)
  dot_product (r0, nr0, R00, R01, nR00, nR01, s0, s1, n0, n1, tp);
  dot_product (r1, nr1, R10, R11, nR10, nR11, s0, s1, n0, n1, tp);
}

TMPL static void
matrix_product (C* Q00, C* Q01, C* Q10, C* Q11,
		nat& nQ00, nat& nQ01, nat& nQ10, nat& nQ11,
		const C* S00, const C* S01, const C* S10, const C* S11,
		nat nS00, nat nS01, nat nS10, nat nS11,
		const C* R00, const C* R01, const C* R10, const C* R11,
		nat nR00, nat nR01, nat nR10, nat nR11, C* tp) {
  // Q = S . R
  matrix_vector_product (Q00, Q10, nQ00, nQ10,
			 S00, S01, S10, S11, nS00, nS01, nS10, nS11,
			 R00, R10, nR00, nR10, tp);
  matrix_vector_product (Q01, Q11, nQ01, nQ11, 
			 S00, S01, S10, S11, nS00, nS01, nS10, nS11,
			 R01, R11, nR01, nR11, tp);
}

TMPL static void
new_matrix (C*& M00, C*& M01, C*& M10, C*& M11, nat l) {
  M00= mmx_new<C> (l); M01= mmx_new<C> (l);
  M10= mmx_new<C> (l); M11= mmx_new<C> (l);
}

TMPL static void
delete_matrix (C* M00, C* M01, C* M10, C* M11, nat l) {
  mmx_delete<C> (M00, l); mmx_delete<C> (M01, l);
  mmx_delete<C> (M10, l); mmx_delete<C> (M11, l);
}

TMPL static void
half_gcd (C* Q00, C* Q01, C* Q10, C* Q11,
	  nat& nQ00, nat& nQ01, nat& nQ10, nat& nQ11,
	  const C* r0, const C* r1, nat n0, nat n1, nat k, C* tp) {
  // Performs the maximum numbers of reductions less than k.
  // tp must have size n0 + n1.
  //cout<<"--> half_gcd, "<<n0<<" "<<n1<<" "<<k<<indent<<lf;
  VERIFY (n0 >= n1, "bad input sizes");
  VERIFY (k  <= n0, "index k out of range");
  if (n1 == 0 || k < n0 - n1 + 1) {
    Q00[0] = 1; nQ00 = 1; nQ01 = 0;
    Q11[0] = 1; nQ10 = 0; nQ11 = 1;
    //cout<<"<-- half_gcd, "<<n0<<" "<<n1<<" "<<k<<unindent<<lf;
    return;
  }
  if (k == 1) {
    Q00[0] = 1; nQ00 = 1; nQ01 = 0;
    Q10[0] = 1; Q11[0] = - r0[n0-1] / r1[n1-1]; nQ10 = 1; nQ11 = 1;
    //cout<<"<-- half_gcd, "<<n0<<" "<<n1<<" "<<k<<unindent<<lf;
    return;
  }
  nat h = k >> 1, h2 = (h << 1) - 1;
  nat len_R= aligned_size<C,V> (h2);
  nat nR00, nR01, nR10, nR11;
  C* R00, * R01, * R10, * R11;
  new_matrix (R00, R01, R10, R11, len_R);
  if (h2 < n0 - n1)
    half_gcd (R00, R01, R10, R11, nR00, nR01, nR10, nR11,
	      r0, r1, n0, n1, h, tp);
  else
    half_gcd (R00, R01, R10, R11, nR00, nR01, nR10, nR11,
	      r0 + n0 - h2, r1 + n0 - h2,
	      h2, h2 - (n0 - n1), h, tp);    
  /* cout << "R00= "; vector_naive::print(cout, R00, nR00); cout << lf;
     cout << "R01= "; vector_naive::print(cout, R01, nR01); cout << lf;
     cout << "R10= "; vector_naive::print(cout, R10, nR10); cout << lf;
     cout << "R11= "; vector_naive::print(cout, R11, nR11); cout << lf; */
  nat len_r= aligned_size<C,V> (n0 + h);
  C* rjm1= mmx_new<C> (len_r), * rj  = mmx_new<C> (len_r);
  nat nj, njm1;
  // TODO << use truncated product
  matrix_vector_product (rjm1, rj, njm1, nj,
			 R00, R01, R10, R11, nR00, nR01, nR10, nR11,
			 r0, r1, n0, n1, tp);
  /* cout << "rjm1= "; vector_naive::print(cout, rjm1, njm1); cout << lf;
     cout << "rj  = "; vector_naive::print(cout, rj, nj); cout << lf; */
  if (nj == 0 || k < n0 - nj + 1) {
    V::copy (Q00, R00, nR00); nQ00= nR00;
    V::copy (Q01, R01, nR01); nQ01= nR01;
    V::copy (Q10, R10, nR10); nQ10= nR10;
    V::copy (Q11, R11, nR11); nQ11= nR11;
    //cout<<"<-- half_gcd, "<<n0<<" "<<n1<<" "<<k<<unindent<<lf;
    return;
  }
  nat len_S= aligned_size<C,V> (max (k - (n0 - nj), njm1 - nj + 1));
  nat nS00, nS01, nS10, nS11;
  C* S00, * S01, * S10, * S11;
  new_matrix (S00, S01, S10, S11, len_S);

  nat len_T= aligned_size<C,V> (njm1 - nj + h);
  nat nT00, nT01, nT10, nT11;
  C* T00, * T01, * T10, * T11;
  new_matrix (T00, T01, T10, T11, len_T);

  S01[0]= 1; S10[0]= 1;  
  nS00= 0; nS01= 1; nS10= 1; nS11= njm1 - nj + 1;
  quo_rem (S11, rjm1, rj, njm1, nj);
  V::Vec::neg (S11, nS11);
  V::trim (rj, nj); V::trim (rjm1, njm1);
  /* cout << "rjm1= "; vector_naive::print(cout, rjm1, njm1); cout << lf;
     cout << "rj  = "; vector_naive::print(cout, rj, nj); cout << lf; */
  matrix_product (T00, T01, T10, T11, nT00, nT01, nT10, nT11,
		  S00, S01, S10, S11, nS00, nS01, nS10, nS11,
		  R00, R01, R10, R11, nR00, nR01, nR10, nR11, tp);

  if (njm1 == 0 || k < n0 - njm1 + 1) {
    V::copy (Q00, T00, nT00); nQ00= nT00;
    V::copy (Q01, T01, nT01); nQ01= nT01;
    V::copy (Q10, T10, nT10); nQ10= nT10;
    V::copy (Q11, T11, nT11); nQ11= nT11;
    //cout<<"<-- half_gcd, "<<n0<<" "<<n1<<" "<<k<<unindent<<lf;
    return;
  }
  h = k - (n0 - nj); h2 = (h << 1) - 1;
  if (h2 < nj - njm1 || h2 > nj)
    half_gcd (S00, S01, S10, S11, nS00, nS01, nS10, nS11,
	      rj, rjm1, nj, njm1, h, tp);
  else
    half_gcd (S00, S01, S10, S11, nS00, nS01, nS10, nS11,
	      rj + nj - h2, rjm1 + nj - h2, h2, h2 - (nj - njm1), h, tp);
  /* cout << "S00= "; vector_naive::print(cout, S00, nS00); cout << lf;
     cout << "S01= "; vector_naive::print(cout, S01, nS01); cout << lf;
     cout << "S10= "; vector_naive::print(cout, S10, nS10); cout << lf;
     cout << "S11= "; vector_naive::print(cout, S11, nS11); cout << lf; */

  mmx_delete<C> (rjm1, len_r); mmx_delete<C> (rj, len_r);
  matrix_product (Q00, Q01, Q10, Q11, nQ00, nQ01, nQ10, nQ11,
		  S00, S01, S10, S11, nS00, nS01, nS10, nS11,
		  T00, T01, T10, T11, nT00, nT01, nT10, nT11, tp);
  /* cout << "Q00= "; vector_naive::print(cout, Q00, nQ00); cout << lf;
     cout << "Q01= "; vector_naive::print(cout, Q01, nQ01); cout << lf;
     cout << "Q10= "; vector_naive::print(cout, Q10, nQ10); cout << lf;
     cout << "Q11= "; vector_naive::print(cout, Q11, nQ11); cout << lf; */
  delete_matrix (R00, R01, R10, R11, len_R);
  delete_matrix (S00, S01, S10, S11, len_S);
  delete_matrix (T00, T01, T10, T11, len_T);
  //cout<<"<-- half_gcd, "<<n0<<" "<<n1<<" "<<k<<unindent<<lf;
}
  
public:

TMPL static void
gcd (C* g, nat& n, const C* s1, const C* s2, nat n1, nat n2) {
  // g must have allocated length min (n1, n2)
  if (n1 < Threshold(C,Th) || n2 < Threshold(C,Th)) {
    V::gcd (g, n, s1, s2, n1, n2); return; }
  VERIFY (n1>0 && n2>0 && s1[n1-1] != 0 && s2[n2-1] != 0,
	  "invalid hypothesis for gcd computation");
  if (n1 < n2) { swap (s1, s2); swap (n1, n2); }
  nat nu1, nu2, nv1, nv2;
  nat l1= aligned_size<C,V> (n1), l2= aligned_size<C,V> (n2);
  C* u1= mmx_new<C> (l2), * u2= mmx_new<C> (l1);
  C* v1= mmx_new<C> (l2), * v2= mmx_new<C> (l1);
  nat len_tp= aligned_size<C,V> (n1 + n2);
  C* tp1= mmx_new<C> (len_tp), * tp2= mmx_new<C> (len_tp);
  half_gcd (u1, u2, v1, v2, nu1, nu2, nv1, nv2,
	    s1, s2, n1, n2, n1, tp1);
  VERIFY (nu1 <= n2 && nv1 <= n2 && nu2 <= n1 && nv2 <= n1, "bug");
  // TODO << use truncated product
  dot_product (tp2, n, u1, u2, nu1, nu2, s1, s2, n1, n2, tp1);
  VERIFY (n <= n2, "bug");
  V::copy (g, tp2, n); V::clear (g + n, n2 - n);
  mmx_delete<C> (tp1, len_tp); mmx_delete<C> (tp2, len_tp);
  mmx_delete<C> (u1, l2); mmx_delete<C> (u2, l1);
  mmx_delete<C> (v1, l2); mmx_delete<C> (v2, l1);
}

TMPL static void
gcd (C* g, nat& n, const C* s1, const C* s2, nat n1, nat n2,
     C* uu1, nat& nuu1) {
  // g must have allocated length min (n1, n2)
  if (n1 < Threshold(C,Th) || n2 < Threshold(C,Th)) {
    V::gcd (g, n, s1, s2, n1, n2, uu1, nuu1); return; }
  VERIFY (n1>0 && n2>0 && s1[n1-1] != 0 && s2[n2-1] != 0,
	  "invalid hypothesis for gcd computation");
  nat nu1, nu2, nv1, nv2;
  nat l1= aligned_size<C,V> (n1), l2= aligned_size<C,V> (n2);
  C* u1= mmx_new<C> (l2), * u2= mmx_new<C> (l1);
  C* v1= mmx_new<C> (l2), * v2= mmx_new<C> (l1);
  nat len_tp= aligned_size<C,V> (n1 + n2);
  C* tp1= mmx_new<C> (len_tp), * tp2= mmx_new<C> (len_tp);
  if (n1 >= n2)
    half_gcd (u1, u2, v1, v2, nu1, nu2, nv1, nv2,
	      s1, s2, n1, n2, n1, tp1);
  else
    half_gcd (u2, u1, v2, v1, nu2, nu1, nv2, nv1,
	      s2, s1, n2, n1, n2, tp1);
  VERIFY (nu1 <= n2 && nv1 <= n2 && nu2 <= n1 && nv2 <= n1, "bug");
  dot_product (tp2, n, u1, u2, nu1, nu2, s1, s2, n1, n2, tp1);
  VERIFY (n <= min (n1, n2), "bug");
  V::copy (g, tp2, n); V::clear (g + n, min (n1, n2) - n);
  V::copy (uu1, u1, nu1); V::clear (uu1 + nu1, n2 - nu1); nuu1= nu1;
  mmx_delete<C> (tp1, len_tp); mmx_delete<C> (tp2, len_tp);
  mmx_delete<C> (u1, l2); mmx_delete<C> (u2, l1);
  mmx_delete<C> (v1, l2); mmx_delete<C> (v2, l1);
}  

TMPL static void
gcd (C* g, nat& n, const C* s1, const C* s2, nat n1, nat n2,
     C* uu1, C* uu2, nat& nuu1, nat& nuu2) {
  // g must have allocated length min (n1, n2)
  if (n1 < Threshold(C,Th) || n2 < Threshold(C,Th)) {
    V::gcd (g, n, s1, s2, n1, n2, uu1, uu2, nuu1, nuu2); return; }
  VERIFY (n1>0 && n2>0 && s1[n1-1] != 0 && s2[n2-1] != 0,
	  "invalid hypothesis for gcd computation");
  nat nu1, nu2, nv1, nv2;
  nat l1= aligned_size<C,V> (n1), l2= aligned_size<C,V> (n2);
  C* u1= mmx_new<C> (l2), * u2= mmx_new<C> (l1);
  C* v1= mmx_new<C> (l2), * v2= mmx_new<C> (l1);
  nat len_tp= aligned_size<C,V> (n1 + n2);
  C* tp1= mmx_new<C> (len_tp), * tp2= mmx_new<C> (len_tp);
  if (n1 >= n2)
    half_gcd (u1, u2, v1, v2, nu1, nu2, nv1, nv2,
	      s1, s2, n1, n2, n1, tp1);
  else
    half_gcd (u2, u1, v2, v1, nu2, nu1, nv2, nv1,
	      s2, s1, n2, n1, n2, tp1);
  VERIFY (nu1 <= n2 && nv1 <= n2 && nu2 <= n1 && nv2 <= n1, "bug");
  dot_product (tp2, n, u1, u2, nu1, nu2, s1, s2, n1, n2, tp1);
  VERIFY (n <= min (n1, n2), "bug");
  V::copy (g, tp2, n); V::clear (g + n, min (n1, n2) - n);
  V::copy (uu1, u1, nu1); V::clear (uu1 + nu1, n2 - nu1); nuu1= nu1;
  V::copy (uu2, u2, nu2); V::clear (uu2 + nu2, n1 - nu2); nuu2= nu2;
  mmx_delete<C> (tp1, len_tp); mmx_delete<C> (tp2, len_tp);
  mmx_delete<C> (u1, l2); mmx_delete<C> (u2, l1);
  mmx_delete<C> (v1, l2); mmx_delete<C> (v2, l1);
}  

TMPL static void
pade (C* r, C* t, const C* s, nat m, nat n, nat k) {
  // r (resp. t) must have allocated length at least k (resp. n-k+1)
  // s = r / t + O(x^n), deg r < k, deg t <= n-k
  // "Modern Computer Algebra", Corollary 5.21
  if (n < Threshold(C,Th) || k == n || k == 0) {
    V::pade (r, t, s, m, n, k); return; }
  VERIFY (n > k && n > 0 && m > 0 && m <= n && s[m-1] != 0,
	  "invalid hypothesis for gcd computation");
  nat nu1, nu2, nv1, nv2;
  nat n1= n+1, n2= m;
  nat l1= aligned_size<C,V> (n1), l2= aligned_size<C,V> (n2);
  C* u1= mmx_new<C> (l2), * u2= mmx_new<C> (l1);
  C* v1= mmx_new<C> (l2), * v2= mmx_new<C> (l1);
  C* s1= mmx_new<C> (l1); V::clear (s1, n); s1[n]= 1; const C* s2= s;
  nat len_tp= aligned_size<C,V> (n1 + n2);
  C* tp= mmx_new<C> (len_tp);
  half_gcd (u1, u2, v1, v2, nu1, nu2, nv1, nv2,
	    s1, s2, n1, n2, n-k, tp);

  nat len_r= aligned_size<C,V> (2*n1-k);
  C* rjm1= mmx_new<C> (len_r), * rj  = mmx_new<C> (len_r);
  nat nj, njm1;
  matrix_vector_product (rjm1, rj, njm1, nj,
			 u1, u2, v1, v2, nu1, nu2, nv1, nv2,
			 s1, s2, n1, n2, tp);
  if (nj <= k) {
    VERIFY (nv2 <= n-k+1, "bug");
    V::copy (r, rj, nj); V::clear (r + nj, k - nj);
    V::copy (t, v2, nv2); V::clear (t + nv2, n-k+1 - nv2);
  }
  else {
    VERIFY (nj > k, "bug");
    nat nq= njm1 - nj + 1, len_q= aligned_size<C,V> (nq);
    C* q= mmx_new<C> (len_q);
    quo_rem (q, rjm1, rj, njm1, nj);
    V::trim (rjm1, njm1);
    V::mul (tp, q, v2, nq, nv2);
    V::clear (u2 + nu2, n1 - nu2); nu2= n1;
    V::sub (u2, tp, nq + nv2 - 1);
    V::trim (u2, nu2);
    VERIFY (njm1 <= k, "bug");
    VERIFY (nu2 <= n-k+1, "bug");
    V::copy (r, rjm1, njm1); V::clear (r + njm1, k - njm1);
    V::copy (t, u2, nu2); V::clear (t + nu2, n-k+1 - nu2);
    mmx_delete<C> (q, len_q);
  }
  mmx_delete<C> (rjm1, len_r); mmx_delete<C> (rj, len_r);
  mmx_delete<C> (s1, l1); mmx_delete<C> (tp, len_tp);
  mmx_delete<C> (u1, l2); mmx_delete<C> (u2, l1);
  mmx_delete<C> (v1, l2); mmx_delete<C> (v2, l1);
}

}; // polynomial_dicho_euclidean<V>

/******************************************************************************
* Efficient evaluation of polynomials
******************************************************************************/

template<typename V, typename Th= uniform_threshold<64> >
struct polynomial_dicho_evaluation: public V {

TMPL static inline void
factorials (C* dest, nat n) {
  if (n > 0) dest[0]= C(1);
  for (nat i=1; i<n; i++)
    dest[i]= C(i) * dest[i-1];
}

TMPL static void
shift (C* dest, const C* s, const C& sh, nat n) {
  if (n <= 1 || sh == 0) V::copy (dest, s, n);
  else {
    nat l= aligned_size<C,V> (5 * n);
    C* u    = mmx_new<C> (l);
    C* v    = u + n;
    C* w    = v + n;
    C* facts= w + n + n;
    factorials (facts, n);
    V::Vec::mul (u, s, facts, n);
    V::Vec::vec_reverse (u, n);
    V::Vec::set (v, 1, n);
    if (sh != 0) V::q_difference (v, v, sh, n);
    V::Vec::div (v, facts, n);
    V::mul (w, u, v, n, n); // FIXME: rather use truncated multiplication
    V::Vec::vec_reverse (w, n);
    V::Vec::div (dest, w, facts, n);
    mmx_delete<C> (u, l);
  }
}

TMPL static void
multi_mod (C* r, const C* p, const C* q, const nat* d, nat n, nat k) {
  // The polynomial (p, n) is divided by k polynomials
  // (q, d[0]), (q + d[0], d[1]), ..., (q + d[0] + ... + d[k-2], d[k-1])
  // The remainders of these divisions are put into
  // (r, d[0]-1), (r + d[0], d[1]-1), ..., (r + d[0] + ... + d[k-2], d[k-1]-1)
  /*
  nat off= 0;
  cout << "p= "; V::Vec::print (cout, p, n); cout << "\n";
  for (nat i=0; i<k; i++, off += d[i]) {
    cout << "    q[" << i << "]= ";
    V::Vec::print (cout, q+off, d[i]);
    cout << "\n";
  }
  */
  if (k <= 1) {
    if (k == 0) return;
    else if (n < d[0]-1) {
      V::Vec::copy (r, p, n);
      V::Vec::set (r + n, 0, d[0]-n);
    }
    else {
      ASSERT (q[d[0]-1] == 1, "quotients not normalized");
      nat l= aligned_size<C,V> (n << 1);
      C* buf= mmx_new<C> (l);
      V::Vec::copy (buf, p, n);
      V::quo_rem (buf + n, buf, q, n, d[0]);
      V::copy (r, buf, d[0]);
      mmx_delete<C> (buf, l);
    }
  }
  else {
    nat tot= 0, start= 0, qoff= 0;
    for (nat i=0; i<=k; ) {
      ASSERT (i == k || q[d[i]-1] == 1, "quotients not normalized");
      nat len= tot + d[i] + d[i+1] - (i - start + 2);
      if (i+1 < k && len <= (n>>1)) {
	// Continue to find index of first packet of q[i]s
	// to be treated by dichotomy
	tot += d[i] + d[i+1];
	i   += 2;
      }
      else {
	if (start < i) {
	  // Treat divisions by q_start, ..., q_{i-1} by dichotomy
	  nat toff= 0, save_qoff= qoff;
	  nat* e= mmx_new<nat> ((i-start) >> 1);
	  nat l= aligned_size<C,V> (tot << 2);
	  C* t= mmx_new<C> (l);
	  C* u= t + tot;
	  C* v= u + tot;
	  C* w= v + tot;
	  for (nat j= start; j<i; j+=2) {
	    V::mul (t + toff, q + qoff, q + qoff + d[j], d[j], d[j+1]);
	    t[toff + d[j] + d[j+1] - 2]= 1; // NOTE: avoids precision loss
	    e[(j-start) >> 1]= d[j] + d[j+1] - 1;
	    qoff += d[j] + d[j+1];
	    toff += d[j] + d[j+1] - 1;
	  }
	  multi_mod (u, p, t, e, n, (i-start) >> 1);
	  V::Vec::copy (v, u, toff);
	  toff= 0; qoff= save_qoff;
	  for (nat j= start; j<i; j+=2) {
	    V::quo_rem (w, u + toff, q + qoff, d[j]+d[j+1]-2, d[j]);
	    V::quo_rem (w, v + toff, q + qoff + d[j], d[j]+d[j+1]-2, d[j+1]);
	    V::copy (r + qoff, u + toff, d[j]);
	    V::copy (r + qoff + d[j], v + toff, d[j+1]);
	    qoff += d[j] + d[j+1];
	    toff += d[j] + d[j+1] - 1;
	  }
	  mmx_delete<C> (t, l);
	  mmx_delete<nat> (e, (i-start) >> 1);
	  start= i;
	  tot  = 0;
	}
	else if (i<k) {
	  // i+1 = k or d[i] + d[i+1] is too large;
	  // perform the division by q_i apart
	  nat l= aligned_size<C,V> (n << 1);
	  C* buf= mmx_new<C> (l);
	  V::Vec::copy (buf, p, n);
	  V::quo_rem (buf + n, buf, q + qoff, n, d[i]);
	  V::copy (r + qoff, buf, d[i]);
	  mmx_delete<C> (buf, l);
	  qoff += d[i];
	  i++;
	  start= i;
	  tot  = 0;
	}
	else break;
      }
    }
  }
  /*
  off= 0;
  cout << "Results\n";
  for (nat i=0; i<k; i++, off += d[i]) {
    cout << "    r[" << i << "]= ";
    V::Vec::print (cout, r+off, d[i]-1);
    cout << "\n";
  }
  */
}

TMPL static inline void
annulator (C* d, const C* s, nat n) {
  V::annulator (d, s, n);
}

TMPL static inline C
eval (const C* p, const C& x, nat l) {
  return V::eval (p, x, l);
}

TMPL static void
eval (C* v, const C* p, const C* x, nat n, nat k) {
  if (n < Threshold(C,Th))
    V::eval (v, p, x, n, k);
  else {
    nat* d= mmx_new<nat> (k);
    nat l= aligned_size<C,V> (k << 2);
    C* q= mmx_new<C> (l);
    C* r= q + (k<<1);
    for (nat i=0; i<k; i++) {
      q[i<<1    ]= -x[i];
      q[(i<<1)+1]= 1;
      d[i]       = 2;
    }
    multi_mod (r, p, q, d, n, k);
    for (nat i=0; i<k; i++) v[i]= r[i<<1];
    mmx_delete<C> (q, l);
    mmx_delete<nat> (d, k);
  }
}

TMPL static void
q_binomial (C* dest, const C& q, nat mu) {
  dest[mu]= 1;
  for (nat i=1; i<=mu; i++)
    dest[mu-i]= (C (mu+1-i) * dest[mu+1-i] * q) / C(i);
}

TMPL static void
expand (C** v, const C* p, const C* x, const nat* nu, nat n, nat k) {
  nat tot= 0;
  for (nat i=0; i<k; i++)
    tot += nu[i] + 1;
  nat* d= mmx_new<nat> (k);
  nat l= aligned_size<C,V> (tot << 1);
  C* q= mmx_new<C> (l);
  C* r= q + tot;
  nat off= 0;
  for (nat i=0; i<k; i++) {
    q_binomial (q + off, -x[i], nu[i]);
    d[i]= nu[i] + 1;
    off += d[i];
  }
  multi_mod (r, p, q, d, n, k);
  off= 0;
  for (nat i=0; i<k; i++) {
    shift (v[i], r + off, x[i], nu[i]);
    off += d[i];
  }
  mmx_delete<C> (q, l);
  mmx_delete<nat> (d, k);
}

}; // polynomial_dicho_evaluation<V>

/******************************************************************************
* Dichotomic algorithms for polynomials
******************************************************************************/

template<typename V>
struct polynomial_multiply_dicho:
  public polynomial_dicho_multiplication<V>
{};

template<typename V>
struct polynomial_algos_euclidean:
  public polynomial_naive_evaluation<
           polynomial_dicho_euclidean<
	     polynomial_naive_euclidean<
	       polynomial_dicho_composition<
		 polynomial_graeffe<
		   polynomial_dicho_division<
		     polynomial_naive_division<V> > > > > > >
{};

struct polynomial_euclidean:
  public polynomial_variable_name_local<
           polynomial_algos_euclidean<
	     polynomial_multiply_dicho<
	       polynomial_multiply_naive<
		 vector_naive> > > >
{};

template<typename V>
struct polynomial_algos_dicho:
  public polynomial_variable_name_local<
           polynomial_dicho_evaluation<
	     polynomial_naive_evaluation<
	       polynomial_dicho_euclidean<
		 polynomial_naive_euclidean<
		   polynomial_dicho_composition<
		     polynomial_graeffe<
		       polynomial_newton_division<
			 polynomial_naive_division<V> > > > > > > > >
{};

struct polynomial_dicho:
  public polynomial_algos_dicho<
           polynomial_multiply_dicho<
	     polynomial_multiply_naive<
	       vector_naive> > >
{};

#undef TMPL
} // namespace mmx
#endif //__MMX__POLYNOMIAL_DICHO__HPP
