
/******************************************************************************
* MODULE     : polynomial_naive.hpp
* DESCRIPTION: naive low level polynomial arithmetic
* COPYRIGHT  : (C) 2003  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__POLYNOMIAL_NAIVE__HPP
#define __MMX__POLYNOMIAL_NAIVE__HPP
#include <basix/vector.hpp>

namespace mmx {
#define TMPL  template<typename C>
#define TMPLK template<typename C, typename K>
#define TMPLX template<typename C, typename X>

/******************************************************************************
* Variable name storage
******************************************************************************/

template<typename V>
struct polynomial_variable_name_local: public V {
  template<typename P>
  class variable_name_storage {
    static inline string& dyn_name () {
      static string name = "x";
      return name; }
  public:
    static inline void set_variable_name (const string& x) { dyn_name () = x; }
    static inline string get_variable_name () { return dyn_name (); }
  };
};

/******************************************************************************
* Vectorial operations
******************************************************************************/

template<typename V>
struct polynomial_vectorial {
typedef V Vec;

TMPL static inline nat
vec_aligned_size (nat n) {
  return Vec::template vec_aligned_size<C> (n); }

TMPLX static inline void set (C* dest, const X& c, nat n) {
  Vec::set (dest, c, n); }
TMPL static inline void clear (C* dest, nat n) {
  Vec::clear (dest, n); }
TMPL static inline void copy (C* dest, const C* s, nat n) {
  Vec::copy (dest, s, n); }
TMPLX static inline void cast (C* dest, const X* s, nat n) {
  Vec::cast (dest, s, n); }
TMPL static inline void reverse (C* dest, const C* s, nat n) {
  Vec::vec_reverse (dest, s, n); }
TMPL static inline void neg (C* dest, const C* s, nat n) {
  Vec::neg (dest, s, n); }
TMPL static inline void neg (C* dest, nat n) {
  Vec::neg (dest, n); }
TMPL static inline void add (C* dest, const C* s1, const C* s2, nat n) {
  Vec::add (dest, s1, s2, n); }
TMPL static inline void add (C* dest, const C* s, nat n) {
  Vec::add (dest, s, n); }
TMPL static inline void sub (C* dest, const C* s1, const C* s2, nat n) {
  Vec::sub (dest, s1, s2, n); }
TMPL static inline void sub (C* dest, const C* s, nat n) {
  Vec::sub (dest, s, n); }
TMPL static inline void mul_add (C* dest, const C* s1, const C* s2, nat n) {
  Vec::mul_add (dest, s1, s2, n); }
TMPLX static inline void mul_add (C* dest, const C* s, const X& c, nat n) {
  Vec::mul_add (dest, s, c, n); }
TMPLX static inline void mul_add (C* dest, const X& c, const C* s, nat n) {
  Vec::mul_add (dest, c, s, n); }
TMPL static inline void mul_sc (C* dest, const C* s, const C& c, nat n) {
  Vec::mul (dest, s, c, n); }
TMPL static inline void mul_sc (C* dest, const C& c, nat n) {
  Vec::mul (dest, c, n); }
TMPL static inline void div_sc (C* dest, const C* s, const C& c, nat n) {
  Vec::div (dest, s, c, n); }
TMPL static inline void div_sc (C* dest, const C& c, nat n) {
  Vec::div (dest, c, n); }
TMPL static inline void quo_sc (C* dest, const C* s, const C& c, nat n) {
  Vec::quo (dest, s, c, n); }
TMPL static inline void quo_sc (C* dest, const C& c, nat n) {
  Vec::quo (dest, c, n); }
TMPL static inline void rem_sc (C* dest, const C* s, const C& c, nat n) {
  Vec::rem (dest, s, c, n); }
TMPL static inline void rem_sc (C* dest, const C& c, nat n) {
  Vec::rem (dest, c, n); }
TMPL static inline bool equal (const C* s1, const C* s2, nat n) {
  return Vec::equal (s1, s2, n); }
TMPL static inline bool equal (const C* s, const C& c, nat n) {
  return Vec::equal (s, c, n); }
TMPL static inline void trim (const C* s, nat& n) {
  while (n > 0 && s[n-1] == 0) n--; }

}; // polynomial_vectorial<V>

/******************************************************************************
* Other linear time operators
******************************************************************************/

template<typename V>
struct polynomial_naive_linear: public V {

TMPL static inline int
sign (C* s, nat n) {
  while (n != 0) {
    int sgn= sign (*s);
    if (sgn != 0) return sgn;
    s++; n--;
  }
  return 0;
}

TMPL static inline int
cmp (const C* s1, const C* s2, nat n) {
  while (n != 0) {
    int sgn= compare (*s1, *s2);
    if (sgn != 0) return sgn;
    s1++; s2++; n--;
  }
  return 0;
}

TMPL static inline void
derive (C* dest, nat n) {
  for (nat i=1; i<n; i++)
    dest[i-1]= C(i) * dest[i];
}

TMPL static inline void
derive (C* dest, const C* s, nat n) {
  for (nat i=1; i<n; i++)
    dest[i-1]= C(i) * s[i];
}

TMPL static inline void
xderive (C* dest, nat n) {
  for (nat i=0; i<n; i++)
    dest[i]= C(i) * dest[i];
}

TMPL static inline void
xderive (C* dest, const C* s, nat n) {
  for (nat i=0; i<n; i++)
    dest[i]= C(i) * s[i];
}

TMPL static inline void
integrate (C* dest, const C* s, nat n) {
  dest[0]= 0;
  for (nat i=0; i<n; i++)
    dest[i+1]= s[i] / C(i+1);
}

TMPLK static inline void
q_difference (C* dest, const C* s, const K& q, nat n) {
  K p (1);
  while (n != 0) {
    *dest= p * (*s);
    dest++; s++; n--; p *= q;
  }
}

TMPL static inline void
dilate (C* dest, const C* s, nat p, nat n) {
  clear (dest, (n-1)*p + 1);
  while (n != 0) {
    *dest= *s;
    dest += p; s++; n--;
  }
}

}; // polynomial_naive_linear<V>

/******************************************************************************
* Multiplication
******************************************************************************/

template<typename V>
struct polynomial_naive_multiplication: public V {
  typedef polynomial_naive_multiplication<V> Naive;
  typedef polynomial_naive_multiplication<V> Positive;
  typedef polynomial_naive_multiplication<V> No_simd;
  typedef polynomial_naive_multiplication<V> No_thread;
  typedef polynomial_naive_multiplication<V> No_scaled;

TMPLK static void
mul (C* dest, const C* s1, const K* s2, nat n1, nat n2) {
  if (n1+n2>0)
    V::clear (dest, n1+n2-1);
  while (n2 != 0) {
    V::mul_add (dest, s1, *s2, n1);
    dest++; s2++; n2--;
  }
}

TMPL static void
square (C* dest, const C* s, nat n) {
  if (n>0)
    V::clear (dest, n+n-1);
  for (nat i=0; i<n-1; i++)
    V::mul_add (dest+i+i+1, s+i+1, s[i], n-i-1);
  if (n>1)
    V::add (dest+1, dest+1, n+n-3);
  for (nat i=0; i<n; i++)
    dest[i+i] += s[i] * s[i];
}

TMPL static void
tmul (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
  // transposed multiplication
  // s1 has length n1, dest has length n2 and s2 has length n1+n2-1
  V::clear (dest, n2);
  while (n1 != 0) {
    V::mul_add (dest, *s1, s2, n2);
    s1++; s2++; n1--;
  }
}

}; // polynomial_naive_multiplication<V>

/******************************************************************************
* Division
******************************************************************************/

template<typename V>
struct polynomial_naive_division: public V {

TMPL static void
quo_rem (C* dest, C* s1, const C* s2, nat n1, nat n2) {
  // (s1, n1) contains the numerator on input and the remainder on output
  // (s2, n2) contains the denominator. We assume n2>0 and s2[n2-1] != 0
  // (dest, n1-n2+1) contains the quotient
  while (n1 >= n2 && n1 != 0) {
    int d= n1-n2;
    C q= quo (s1[n1-1], s2[n2-1]);
    dest[d]= q;
    for (nat i=0; i<n2; i++)
      s1[i+d] -= q * s2[i];
    n1--;
  }
}

TMPL static void
tquo_rem (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
  // (s1, n2-1) contains the part corresponding to the remainder, while
  // (s2+n2-1, n1-n2+1) contains the one of the quotient.
  // We assume n2>0 and s2[n2-1] != 0.
  // (dest, n1) contains the output.
  C c = s2[n2-1];
  V::copy (dest, s1, n2-1);
  s1 += n2-1;
  for (nat i = n2-1; i < n1; i++, dest++, s1++)
    dest[n2-1] = quo (*s1 - V::Vec::inn_prod (dest, s2, n2-1), c);
}

TMPL static void
pquo_rem (C* dest, C* s1, const C* s2, nat n1, nat n2) {
  // (s1, n1) contains the numerator on input
  //          and the pseudo-remainder on output
  // (s2, n2) contains the denominator.
  // We assume n2>0 and s2[n2-1] != 0
  // (dest, n1-n2+1) contains the pseudo-quotient
  nat l= n1 - n2 + 1;
  while (n1 >= n2 && n1 != 0) {
    int d= n1-n2;
    dest[d]= s1[n1-1];
    for (nat i=0; i<n1-1; i++)
      s1[i]*= s2[n2-1];
    for (nat i=0; i<n2-1; i++)
      s1[i+d]-= s1[n1-1] * s2[i];
    s1[n1-1]=0;
    n1--;
  }
  C c= s2[n2-1];
  for (nat i = 1; i < l; i++) {
    dest[i] *= c;
    c *= s2[n2-1];
  }
}

}; // polynomial_naive_division<V>

/******************************************************************************
* Graeffe transformations
******************************************************************************/

template<typename V>
struct polynomial_graeffe: public V {

TMPL static void
graeffe (C* dest, const C* s, nat n) {
  nat l1= (n+1) >> 1, l2= n >> 1;
  nat l= aligned_size<C,V> (n << 1);
  C* c1 = mmx_new<C> (l);
  C* c2 = c1 + l1;
  C* aux= c2 + l2;
  if (n>0) dest[n-1]= C(0);
  V::Vec::half_copy (c1, s, l1);
  V::Vec::half_copy (c2, s+1, l2);
  V::square (dest, c1, l1);
  V::square (aux , c2, l2);
  V::sub (dest+1, aux, (l2<<1) - 1);
  mmx_delete<C> (c1, l);
}

}; // polynomial_graeffe<V>

/******************************************************************************
* Composition
******************************************************************************/

template<typename V>
struct polynomial_dicho_composition: public V {

TMPLK static void
compose (C* dest, K* p, const C* s1, const K* s2, nat n1, nat n2) {
  // computes compose ((s1, n1), (s2, n2)) and pow ((s2, n2), n1)
  if (n1 == 1) {
    *dest= *s1;
    V::copy (p, s2, n2);
  }
  else {
    nat d2= n2 - 1;
    nat h1= n1 >> 1, h2= n1 - h1;
    nat l1= (h1-1) * d2 + 1, l2= (h2-1) * d2 + 1;
    nat k1= l1 + d2, k2= l2 + d2;
    nat alloc_c1= aligned_size<C,V> (l1 + l2);
    C* c1= mmx_new<C> (alloc_c1);
    C* c2= c1 + l1;
    nat alloc_p1= aligned_size<K,V> (k1 + k2);
    C* p1= mmx_new<K> (alloc_p1);
    K* p2= p1 + k1;
    compose (c1, p1, s1, s2, h1, n2);
    compose (c2, p2, s1+h1, s2, h2, n2);
    V::mul (dest, c2, p1, l2, k1);
    V::add (dest, c1, l1);
    V::mul (p, p1, p2, k1, k2);
    mmx_delete<K> (p1, alloc_p1);
    mmx_delete<C> (c1, alloc_c1);
  }
}

TMPLK static void
compose (C* dest, const C* s1, const K* s2, nat n1, nat n2) {
  ASSERT (n1 > 0 && n2 > 0, "composition of empty segments");
  if (n1 == 1) *dest= *s1;
  else {
    nat d2= n2 - 1;
    nat h1= n1 >> 1, h2= n1 - h1;
    nat l1= (h1-1) * d2 + 1, l2= (h2-1) * d2 + 1;
    nat k1= l1 + d2;
    nat alloc_c1= aligned_size<C,V> (l1 + l2);
    C* c1= mmx_new<C> (alloc_c1);
    C* c2= c1 + l1;
    nat alloc_p1= aligned_size<K,V> (k1);
    K* p1= mmx_new<K> (alloc_p1);
    compose (c1, p1, s1, s2, h1, n2);
    compose (c2, s1+h1, s2, h2, n2);
    V::mul (dest, c2, p1, l2, k1);
    V::add (dest, c1, l1);
    mmx_delete<K> (p1, alloc_p1);
    mmx_delete<C> (c1, alloc_c1);
  }
}

TMPLK static void
shift (C* dest, const C* s, const K& sh, nat n) {
  if (n <= 1 || sh == 0) V::copy (dest, s, n);
  else {
    K q[2]; q[0]= sh; q[1]= 1;
    compose (dest, s, q, n, 2);
  }
}

}; // polynomial_dicho_composition<V>

/******************************************************************************
* Gcd computations
******************************************************************************/

template<typename V>
struct polynomial_naive_euclidean: public V {
private:
// d1 = u1 s1 + u2 s2,  d2 = v1 s1 + v2 s2
// s1 and s2 being the content of d1 and d2 at input
TMPL static void
gcd (C*& d1, C*& d2, nat& n1 , nat& n2,
     C*& u1, C*& u2, nat& nu1, nat& nu2,
     C*& v1, C*& v2, nat& nv1, nat& nv2, nat k=0) {
  // d1 and d2 must have allocated length max (n1, n2)
  // u1 and v1 (resp. u2 and v2) must have allocated length n2 (resp. n1)
  // Does the euclidian algorithm until a remainder has size at most k.
  if (u1 != NULL) {
    V::clear (u1, n2); u1[0]= 1; nu1= 1;
    V::clear (v1, n2);           nv1= 0; }
  if (u2 != NULL) {
    V::clear (u2, n1);           nu2= 0;
    V::clear (v2, n1); v2[0]= 1; nv2= 1; }
  while (n2 != 0) {
    if (n1 >= n2) {
      if (n2 <= k) return;
      C m= - d1[n1-1] / d2[n2-1]; nat s= n1-n2;
      V::Vec::mul_add (d1 + s, m, d2, n2); V::trim (d1, n1);
      if (u1 != NULL && nv1 != 0) {
	V::Vec::mul_add (u1 + s, m, v1, nv1);
	nu1 = max (nu1, nv1 + s); V::trim (u1, nu1);
      }
      if (u2 != NULL && nv2 != 0 ) {
	V::Vec::mul_add (u2 + s, m, v2, nv2);
	nu2 = max (nu2, nv2 + s); V::trim (u2, nu2);
      }
    }
    else {
      swap (d1, d2); swap (n1 , n2 );
      swap (u1, v1); swap (nu1, nv1);
      swap (u2, v2); swap (nu2, nv2);
    }
  }
  VERIFY (n2 == 0, "bug");
}
public:

TMPL static void
gcd (C* g, nat& n, const C* s1, const C* s2, nat n1, nat n2) {
  // g must have allocated length min (n1, n2)
  VERIFY (n1>0 && n2>0 && s1[n1-1] != 0 && s2[n2-1] != 0,
	  "invalid hypothesis for gcd computation");
  nat m= max (n1, n2); n = min (n1, n2);
  nat l= aligned_size<C,V> (m);
  C* d1= mmx_new<C> (l), * d2= mmx_new<C> (l),
   * u1= NULL, * u2= NULL, * v1= NULL, * v2= NULL;
  nat nu1=0, nu2=0, nv1=0, nv2=0;
  V::copy (d1, s1, n1); V::copy (d2, s2, n2);
  gcd (d1, d2, n1, n2, u1, u2, nu1, nu2, v1, v2, nv1, nv2);
  VERIFY (n1 <= n, "bug");
  V::copy (g, d1, n1); V::clear (g + n1, n - n1); n = n1; 
  mmx_delete<C> (d1, l); mmx_delete<C> (d2, l);
}

TMPL static void
gcd (C* g, nat& n, const C* s1, const C* s2, nat n1, nat n2,
     C* uu1, nat& nuu1) {
  // u1 must have allocated length at least n2
  VERIFY (n1>0 && n2>0 && s1[n1-1] != 0 && s2[n2-1] != 0,
	  "invalid hypothesis for gcd computation");
  nat m = max (n1, n2); n = min (n1, n2); nuu1 = n2;
  nat l = aligned_size<C,V> (m);
  nat l2= aligned_size<C,V> (n2);
  C* d1= mmx_new<C> (l) , * d2= mmx_new<C> (l);
  C* u1= mmx_new<C> (l2), * u2= NULL;
  C* v1= mmx_new<C> (l2), * v2= NULL;
  nat nu1=0, nu2=0, nv1=0, nv2=0;
  V::copy (d1, s1, n1); V::copy (d2, s2, n2);
  gcd (d1, d2, n1, n2, u1, u2, nu1, nu2, v1, v2, nv1, nv2);
  VERIFY (n1 <= n, "bug");
  VERIFY (nu1 <= nuu1, "bug");
  V::copy (g  , d1, n1  ); V::clear (g + n1   , n - n1    ); n = n1; 
  V::copy (uu1, u1, nu1 ); V::clear (uu1 + nu1, nuu1 - nu1); nuu1 = nu1; 
  mmx_delete<C> (d1, l ); mmx_delete<C> (d2, l);
  mmx_delete<C> (u1, l2);
  mmx_delete<C> (v1, l2);
}

TMPL static void
gcd (C* g, nat& n, const C* s1, const C* s2, nat n1, nat n2,
     C* uu1, C* uu2, nat& nuu1, nat& nuu2) {
  VERIFY (n1>0 && n2>0 && s1[n1-1] != 0 && s2[n2-1] != 0,
	  "invalid hypothesis for gcd computation");
  nat m = max (n1, n2); n = min (n1, n2); nuu1 = n2; nuu2 = n1;
  nat l = aligned_size<C,V> (m);
  nat l1= aligned_size<C,V> (n1), l2= aligned_size<C,V> (n2);
  C* d1= mmx_new<C> (l) , * d2= mmx_new<C> (l);
  C* u1= mmx_new<C> (l2), * u2= mmx_new<C> (l1);
  C* v1= mmx_new<C> (l2), * v2= mmx_new<C> (l1);
  nat nu1=0, nu2=0, nv1=0, nv2=0;
  V::copy (d1, s1, n1); V::copy (d2, s2, n2);
  gcd (d1, d2, n1, n2, u1, u2, nu1, nu2, v1, v2, nv1, nv2);
  VERIFY (n1 <= n, "bug");
  VERIFY (nu1 <= nuu1, "bug");
  VERIFY (nu2 <= nuu2, "bug");
  V::copy (g  , d1, n1  ); V::clear (g + n1   , n - n1    ); n = n1; 
  V::copy (uu1, u1, nu1 ); V::clear (uu1 + nu1, nuu1 - nu1); nuu1 = nu1; 
  V::copy (uu2, u2, nu2 ); V::clear (uu2 + nu2, nuu2 - nu2); nuu2 = nu2; 
  mmx_delete<C> (d1, l ); mmx_delete<C> (d2, l);
  mmx_delete<C> (v1, l2); mmx_delete<C> (v2, l1);
}

TMPL static void
pade (C* r, C* t, const C* s, nat m, nat n, nat k) {
  // r (resp. t) must have allocated length at least k (resp. n-k+1)
  // s = r / t + O(x^n), deg r < k, deg t <= n-k
  // "Modern Computer Algebra", Corollary 5.21
  VERIFY (n >= k && n > 0 && s[m-1]!= 0 && m <= n,
	  "invalid hypothesis for gcd computation");
  nat n1= n+1, n2= m, l1= aligned_size<C,V> (n1);
  C* d1= mmx_new<C> (l1) , * d2= mmx_new<C> (l1);
  C* u1= NULL, * u2= mmx_new<C> (l1);
  C* v1= NULL, * v2= mmx_new<C> (l1);
  nat nu1=0, nu2=0, nv1=0, nv2=0;
  V::clear (d1, n); d1[n]= 1; V::copy (d2, s, n2);
  gcd (d1, d2, n1, n2, u1, u2, nu1, nu2, v1, v2, nv1, nv2, k);
  VERIFY (n2  <= k, "bug");
  VERIFY (nv2 <= n-k+1, "bug");
  V::copy (r, d2, n2 ); V::clear (r + n2 , k - n2);
  V::copy (t, v2, nv2); V::clear (t + nv2, n-k+1 - nv2);
  mmx_delete<C> (d1, l1); mmx_delete<C> (d2, l1);
  mmx_delete<C> (u2, l1); mmx_delete<C> (v2, l1);
}

}; // polynomial_naive_euclidean<V>

/******************************************************************************
* Efficient evaluation of polynomials
******************************************************************************/

template<typename V>
struct polynomial_naive_evaluation: public V {

TMPL static void
annulator (C* d, const C* s, nat n) {
  // Given n constants s[0],...s[n-1],
  // construct the polynomial (z-s[0]) ... (z-s[n-1])
  switch (n) {
  case 0:
    d[0]= C(1);
    break;
  case 1:
    d[1]= C(1);
    d[0]= -s[0];
    break;
  case 2:
    d[2]= C(1);
    d[1]= -(s[0]+s[1]);
    d[0]= s[0]*s[1];
    break;
  case 3:
    d[3]= C(1);
    d[2]= -(s[0]+s[1]+s[2]);
    d[1]= s[0]*s[1] + s[1]*s[2]+ s[2]*s[0];
    d[0]= -s[0]*s[1]*s[2];
    break;
  default:
    {
      nat h= n>>1;
      nat l= aligned_size<C,V> (n+2);
      C* aux= mmx_new<C> (l);
      annulator (aux, s, h);
      annulator (aux+h+1, s+h, n-h);
      V::mul (d, aux, aux+h+1, h+1, n+1-h);
      mmx_delete<C> (aux, l);
    }
  }
}

TMPL static C
eval (const C* p, const C& x, nat l) {
  // Evaluate the polynomial (p, n) at x
  if (l == 0) return 0;
  C r= p[l-1];
  for (nat i=l-2; i!=((nat)(-1)); i--)
    r= (x*r) + p[i];
  return r;
}

TMPL static void
eval (C* v, const C* p, const C* x, nat l, nat n) {
  // Evaluate the polynomial (p, l) at points x[0], ..., x[n-1]
  for (nat j=0; j<n; j++)
    if (l == 0) v[j]= 0;
    else {
      C r= p[l-1];
      for (nat i=l-2; i!=((nat)(-1)); i--)
	r= (x[j]*r) + p[i];
      v[j]= r;
    }

}

TMPL static void
expand (C** v, const C* p, const C* x, const nat* nu, nat n, nat k) {
  (void) v; (void) p; (void) x; (void) nu; (void) n; (void) k;
  ERROR ("expand not yet implemented");
}

}; // polynomial_naive_evaluation<V>

/******************************************************************************
* Naive polynomials
******************************************************************************/

template<typename V>
struct polynomial_multiply_naive:
  public polynomial_naive_multiplication<
	   polynomial_naive_linear<
	     polynomial_vectorial<V> > >
{};

template<typename V>
struct polynomial_algos_naive:
  public polynomial_naive_evaluation<
	   polynomial_naive_euclidean<
	     polynomial_dicho_composition< // FIXME: implement naive algo
	       polynomial_graeffe<
	         polynomial_naive_division<V> > > > >
{};

struct polynomial_naive:
  public polynomial_variable_name_local<
           polynomial_algos_naive<
	     polynomial_multiply_naive<
	       vector_naive> > >
{};

#undef TMPL
#undef TMPLK
#undef TMPLX
} // namespace mmx
#endif //__MMX__POLYNOMIAL_NAIVE__HPP
