
/******************************************************************************
* MODULE     : polynomial_quotient.hpp
* DESCRIPTION: Multiply polynomials by reduction to a common denominator
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_POLYNOMIAL_QUOTIENT_HPP
#define __MMX_POLYNOMIAL_QUOTIENT_HPP
#include <numerix/rational.hpp>
#include <algebramix/polynomial_integer.hpp>
namespace mmx {
#define TMPL template<typename C>
#define NC Numerator_type(C)
#define DC Denominator_type(C)

/******************************************************************************
* Low level subroutines
******************************************************************************/

TMPL DC
least_common_denominator (const C* src, nat n) {
  switch (n) {
  case 0: return 1;
  case 1: return denominator (src[0]);
  case 2: return lcm (denominator (src[0]), denominator (src[1]));
  case 3: return lcm (denominator (src[0]), lcm (denominator (src[1]),
						 denominator (src[2])));
  default: return lcm (least_common_denominator (src, n>>1),
		       least_common_denominator (src + (n>>1), n - (n>>1)));
  }
}

TMPL void
denominator_mul (NC* dest, const C* src, nat n, const DC& d) {
  for (nat i=0; i<n; i++)
    dest[i]= numerator (src[i]) * (d / denominator (src[i]));
}

TMPL void
denominator_div (C* dest, const NC* src, nat n, const DC& d) {
  for (nat i=0; i<n; i++)
    dest[i]= C (src[i]) / d;
}

/******************************************************************************
* Multiply polynomials by putting the coefficients under a common denominator
******************************************************************************/

template<typename V, typename Th= dicho_threshold>
struct polynomial_quotient_multiplication: public V {

TMPL static void
mul (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
  if (min (n1, n2) < Threshold(C,Th))
    V::Naive::mul (dest, s1, s2, n1, n2);
  else {
    DC  d1  = least_common_denominator (s1, n1);
    DC  d2  = least_common_denominator (s2, n2);
    /*
      cout << "d1= " << d1 << ", " << "d2= " << d2 << "\n";
      for (nat i=0; i<n1; i++)
      cout << "  s" << i << "\t" << s1[i] << "\n";
      for (nat i=0; i<n2; i++)
      cout << "  t" << i << "\t" << s2[i] << "\n";
    */
    NC* Src1= mmx_new<NC> (n1);
    NC* Src2= mmx_new<NC> (n2);
    NC* Dest= mmx_new<NC> (n1+n2-1);
    denominator_mul (Src1, s1, n1, d1);
    denominator_mul (Src2, s2, n2, d2);
    V::mul (Dest, Src1, Src2, n1, n2);
    denominator_div (dest, Dest, n1+n2-1, d1*d2);
    mmx_delete<NC> (Src1, n1);
    mmx_delete<NC> (Src2, n2);
    mmx_delete<NC> (Dest, n1+n2-1);
  }
}

TMPL static void
square (C* dest, const C* s1, nat n1) {
  if (n1 < Threshold(C,Th))
    V::Naive::square (dest, s1, n1);
  else {
    DC  d1  = least_common_denominator (s1, n1);
    NC* Src1= mmx_new<NC> (n1);
    NC* Dest= mmx_new<NC> (n1+n1-1);
    denominator_mul (Src1, s1, n1, d1);
    V::square (Dest, Src1, n1);
    denominator_div (dest, Dest, n1+n1-1, mmx::square (d1));
    mmx_delete<NC> (Src1, n1);
    mmx_delete<NC> (Dest, n1+n1-1);
  }
}

TMPL static inline void
tmul (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
  (void) dest; (void) s1; (void) s2; (void) n1; (void) n2;
  ERROR ("transposed quotient multiplication not implemented");
}

}; // polynomial_quotient_multiplication<V>

template<typename V>
struct polynomial_multiply_quotient:
  public polynomial_quotient_multiplication<V>
{};

/******************************************************************************
* Use Kronecker multiplication by default for rational polynomials
******************************************************************************/

struct polynomial_rational_kronecker:
  public polynomial_algos_dicho<
         polynomial_multiply_quotient<
         polynomial_multiply_integer<
         polynomial_multiply_naive<
	   vector_naive> > > >
{};

STMPL
struct polynomial_variant_helper<rational> {
  typedef Vector_variant(rational) VV;
  typedef polynomial_multiply_naive<VV> MN;
  typedef polynomial_multiply_integer<MN> MI;
  typedef polynomial_multiply_quotient<MI> MV;
  typedef polynomial_rational_kronecker PV;
  // typedef polynomial_algos_dicho<MV> PV;
  // Commented out so as to keep typenames simple when debugging
};

#undef TMPL
#undef NC
#undef DC
}
#endif // __MMX_POLYNOMIAL_QUOTIENT_HPP
