
/******************************************************************************
* MODULE     : polynomial_schonhage.hpp
* DESCRIPTION: Fast product for any ring with unity
* COPYRIGHT  : (C) 2008  Gregoire Lecerf
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

// Ref. "Modern Computer Algebra", Chapter 8, Exercise 8.30.

#ifndef __MMX__POLYNOMIAL_SCHONHAGE__HPP
#define __MMX__POLYNOMIAL_SCHONHAGE__HPP
#include <numerix/modulus.hpp>
#include <numerix/modulus_naive_int.hpp>
#include <algebramix/polynomial_dicho.hpp>
#include <algebramix/polynomial_balanced.hpp>
#include <algebramix/polynomial_schonhage_strassen.hpp>
#include <algebramix/polynomial_schonhage_triadic.hpp>

namespace mmx {
#define TMPL template<typename C>

struct schonhage_threshold {};

/******************************************************************************
* Multiplication
******************************************************************************/

template<typename V, typename Th= schonhage_threshold, typename Th_rec=Th>
struct polynomial_schonhage_multiplication: public V {
  TMPL static inline void
  mul (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
    typedef polynomial_schonhage_strassen_multiplication<V, Th_rec> S2;
    typedef polynomial_schonhage_triadic_multiplication <V, Th_rec> S3;
    typedef modulus_naive_int<8*sizeof(nat)> MV;
    typedef modulus<nat, MV> Modulus;
    if (n1 < Threshold(C,Th) || n2 < Threshold(C,Th)) {
      V::mul (dest, s1, s2, n1, n2);
      return;
    }
    nat n = n1 + n2 - 1;
    nat l = aligned_size<C,V> (n);
    C* temp = mmx_new<C> (l);
    nat k2= S2::mul (dest, s1, s2, n1, n2, false);
    if (k2 == 0) return;
    nat k3= S3::mul (temp, s1, s2, n1, n2, false);
    if (k3 == 0) {
      V::Vec::copy (dest, temp, n);
      return;
    }
    nat u2, u3;
    nat c2= ((nat) 1) << k2, c3= binpow ((nat) 3, k3);
    MV::inv_mod (u2, binpow ((nat) 2, k2), Modulus (c3));
    MV::inv_mod (u3, binpow ((nat) 3, k3), Modulus (c2));
    u3= c2 - u3;
    VERIFY (u2*c2 - u3*c3 == 1, "bug");
    V::Vec::mul (dest, C(u2), n);
    V::Vec::mul (temp, C(u3), n);
    V::Vec::sub (dest, temp , n);
    mmx_delete<C> (temp, l); }

}; // polynomial_schonhage_multiplication<V>

/******************************************************************************
* Variant shortcut
******************************************************************************/

template<typename V>
struct polynomial_multiply_schonhage:
  public polynomial_balanced_multiplication <
           polynomial_schonhage_multiplication<V> >
{};

struct polynomial_schonhage:
  public polynomial_algos_dicho<
           polynomial_multiply_schonhage<
	     polynomial_multiply_dicho<
               polynomial_multiply_naive<
		 vector_naive> > > >
{};
#undef TMPL
} // namespace mmx
#endif //__MMX__POLYNOMIAL_SCHONHAGE__HPP
