/********************************************************************
 *   This file is part of the source code of the SYNAPS kernel.
 *   Author(s):
 *       Initial version:
 *          B. Mourrain, INRIA, GALAAD, <mourrain@sophia.inria.fr>
 *       Implementatation of Sturm Habicht by
 *         F. Carreras, U. of Santander, <fernando.carreras@unican.es>
 *         C. Tansescu, U. of Santander, <claudiu.tanasescu@unican.es>
 *         L. Gonzales-Vega, U. of Santander, <laureano.gonzalez@unican.es>
 *         O. Ruatta, INRIA, GALAAD, <oruatta@sophia.inria.fr>
 ********************************************************************/
#ifndef algebramix_sturm_seq_hpp
#define algebramix_sturm_seq_hpp
/********************************************************************/
#include <basix/vector.hpp>
#include <algebramix/polynomial.hpp>
/********************************************************************/
namespace mmx {

  inline int st_sign(int n)
  {
    return (n%2==0 ? 1:-1);
  }
    
template<class T>
T st_pow(const T& a, int i)
{
  assert(i>=0);
  if(i == 1) return a;
  if(i > 0)
    {
      if (a == T(1))  return a;
      if (a == T(-1)) return (i % 2  == 0) ? T(1) : T(-1);
      T y(a);
      T z(1);
      unsigned int n = i;
      unsigned int m;
      while (n > 0) {
        m = n / 2;
        if (n %2 )
          {
            z *= y;
          }
        y = y * y;
        n = m;
      }
      return z;
    }
  return T(1);
}

template <class Pol> inline
Pol prem(const Pol & a, const Pol & b)
{
  typedef typename scalar_type_helper<Pol>::val coeff_t;
  Pol A(0); A+=a;
  //  std::cout<<a<<" prem "<<b<<" = ?"<<std::endl;
  int c=0, db=deg(b), da= deg(A), dab =da-db+1;

  Pol t;
  coeff_t lb=b[db];

  while (da>=db) {
    Pol m(A[da],da-db);
    t=b*m;
    c++;
    A*= lb;
    //    t[da]=A[da]; //case of approx. coefficients.
    A-=t;
    da=deg(A);
    //    std::cout<<"\t"<<A<<std::endl;
  }
  A *= st_pow(lb, dab-c);
  //  std::cout<<a<<" prem "<<b<<" = "<<A<<std::endl;
  return A;
}

//--------------------------------------------------------------------
/** Construct the Sturm sequence of two polynomials @a@ and @b@,
 *  using Lalo algorithm.
 **/
template<class Pol>
vector<Pol> sturm_habicht_sequence(const Pol & a, const Pol & b)
{
  typedef typename scalar_type_helper<Pol>::val coeff_type;
  assert(deg(a) >=0 && deg(b) >= 0);
  Pol p=a, q=b;

  int n1 = deg(p), n2= deg(q);
  coeff_type lq = q[n2], lS1;

  vector<Pol> tab=vec<Pol>(p,q); //(Pol(0),n1+n2+3);

  int j=0,n,d,m,nS1,nS0,r;
  Pol S0, S1, NS, Sr;
  if (n1>n2)
    {
      n=n1-1; d=n-n2;
      if (d!=0)
	{
	  S1=q*st_pow(-1,d*(d+1)/2)*st_pow(lq,d);
	  S0=prem(p,q)*st_sign((d+2)*(d+3)/2);
	  //	  std::cout<<" prem "<<S0<<" "<<S1<<std::endl;

	  if (S0!=Pol(0))
	    {tab<<S1; tab<<S0;}
	  else
	    {tab<<S1;}
	}
      else
	{
	  S1=q;S0=-prem(p,q);
	  if (S0!=Pol(0)) {tab<<S0;}
	}
      j=n2-1;
    }
  else
    {
      n=n2; S1=q; S0=-prem(p,q); //rem(p*(lq*lq),q); 
      j=n-1;
      if (S0!=Pol(0)) {tab<<S0;}
    }

  nS1=deg(S1);
  lS1=S1[nS1];
  if (S0==Pol(0)) r=-1; else r=deg(S0);
  
  while (r>=0)
    {
      if (r>0)    {
	if (r==j)
	  {
	    NS=prem(S1,S0); 
	    NS /= st_pow(-lS1,j-r+2);
	    NS *= st_sign(((j-r+2)*(j-r+3))/2);
	    if (NS!=Pol(0)) {tab<<NS;}
	    S1=S0; S0=NS; nS1=deg(S1);lS1=S1[nS1]; j=r-1;
	  }
	else
	  {
	    for (m=r+1;m<j;m++)
	      {
		tab<<Pol(0);
	      }
	    Sr=prem(S1,S0); Sr /= st_pow(lS1,j-r+2);
	    nS0=deg(S0);
	    S1=S0*(st_pow(S0[nS0],j-r)); S1 /= st_pow(lS1,j-r);
	    S1*= st_sign(((j-r)*(j-r+1)/2));
	    S0 = Sr*st_sign(((j-r+2)*(j-r+3)/2));
	    if (S0!=Pol(0)) {
	      tab<<(S1);
	      tab<<(S0);}
	    else {
	      tab<<(S1);}
	    nS1=deg(S1);
	    lS1=S1[nS1]; j= r-1;
	  }
      }
      else
	{
	  if (j>0)
	    {
	      nS0=deg(S0); S1=S0*st_pow(S0[nS0],j); S1 /=st_pow(lS1,j);
	      S1*=st_sign((j*(j+1)/2)); tab<<(S1);
	    }
	  S0=Pol(0);
	}
      if (S0!=Pol(0)) r=deg(S0); else r=-1;
    }
  return tab;
}

template<typename C, typename V> vector<polynomial<C, V> >
sturm_sequence(const polynomial<C, V>& a, const polynomial<C, V>& b)
{
  return sturm_habicht_sequence(a,b);
}
//====================================================================
} //namespace mmx
#endif // algebramix_sturm_seq_hpp

