
/******************************************************************************
* MODULE     : polynomial_tangent.hpp
* DESCRIPTION: Multiplication of polynomials with order one jet coefficients
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__POLYNOMIAL_TANGENT__HPP
#define __MMX__POLYNOMIAL_TANGENT__HPP
#include <algebramix/polynomial.hpp>
#include <algebramix/polynomial_dicho.hpp>
#include <numerix/tangent.hpp>

namespace mmx {
#define TMPL template<typename C>

/******************************************************************************
* Tangent multiplication
******************************************************************************/

template<typename V>
struct polynomial_tangent_multiplication: public V {
  typedef typename V::Naive Naive;
  typedef polynomial_tangent_multiplication<typename V::Positive> Positive;
  typedef polynomial_tangent_multiplication<typename V::No_simd> No_simd;
  typedef polynomial_tangent_multiplication<typename V::No_thread> No_thread;
  typedef polynomial_tangent_multiplication<typename V::No_scaled> No_scaled;

TMPL static inline void
mul (tangent<C>* dest, const tangent<C>* s1, const tangent<C>* s2,
     nat n1, nat n2)
{
  nat nd  = n1+n2-1;
  nat spc1= ((n1+1)>>1) << 1; // force 128 bit alignment if necessary
  nat spc2= ((n2+1)>>1) << 1;
  nat spcd= ((nd+1)>>1) << 1;
  nat spc = 5*(spc1+spc2);
  C* base_x1 = mmx_new<C> (spc);
  C* slope_x1= base_x1  + spc1;
  C* base_x2 = slope_x1 + spc1;
  C* slope_x2= base_x2  + spc2;
  C* base_d  = slope_x2 + spc2;
  C* slope_d = base_d   + spcd;
  C* extra_d = slope_d  + spcd;

  V::Vec::half_copy (base_x1, (C*) ((void*) s1), n1);
  V::Vec::half_copy (slope_x1, ((C*) ((void*) s1)) + 1, n1);
  V::Vec::half_copy (base_x2, (C*) ((void*) s2), n2);
  V::Vec::half_copy (slope_x2, ((C*) ((void*) s2)) + 1, n2);
  V::mul (base_d, base_x1, base_x2, n1, n2);
  V::mul (slope_d, base_x1, slope_x2, n1, n2);
  V::mul (extra_d, slope_x1, base_x2, n1, n2);
  V::add (slope_d, extra_d, nd);
  V::Vec::double_copy ((C*) ((void*) dest), base_d, nd);
  V::Vec::double_copy (((C*) ((void*) dest)) + 1, slope_d, nd);

  mmx_delete<C> (base_x1, spc);
}

TMPL static inline void
square (tangent<C>* dest, const tangent<C>* s1, nat n1) {
  nat nd  = n1+n1-1;
  nat spc1= ((n1+1)>>1) << 1; // force 128 bit alignment if necessary
  nat spcd= ((nd+1)>>1) << 1;
  nat spc = 6*spc1;
  C* base_x1 = mmx_new<C> (spc);
  C* slope_x1= base_x1  + spc1;
  C* base_d  = slope_x1 + spc1;
  C* slope_d = base_d   + spcd;
  
  V::Vec::half_copy (base_x1, (C*) ((void*) s1), n1);
  V::Vec::half_copy (slope_x1, ((C*) ((void*) s1)) + 1, n1);
  V::square (base_d, base_x1, n1);
  V::mul (slope_d, base_x1, slope_x1, n1, n1);
  V::add (slope_d, slope_d, nd);
  V::Vec::double_copy ((C*) ((void*) dest), base_d, nd);
  V::Vec::double_copy (((C*) ((void*) dest)) + 1, slope_d, nd);

  mmx_delete<C> (base_x1, spc);
}

TMPL static inline void
tmul (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
  (void) dest; (void) s1; (void) s2; (void) n1; (void) n2;
  ERROR ("transposed tangent multiplication not implemented");
}

}; // polynomial_tangent_multiplication<V>

template<typename V>
struct polynomial_multiply_tangent:
  public polynomial_tangent_multiplication<V>
{};

/******************************************************************************
* Default variant for tangent coefficients
******************************************************************************/

template<typename C>
struct polynomial_variant_helper<tangent<C> > {
  typedef typename Vector_variant(C) CVV;
  typedef typename Polynomial_multiply(C) CMV;
  typedef typename Polynomial_variant(C) CPV;

  typedef typename Vector_variant(tangent<C>) VV;
  typedef polynomial_multiply_tangent<CMV> MV;
  typedef polynomial_algos_dicho<MV> PV;
};

STMPL
struct polynomial_variant_helper<tangent<generic> > {
  typedef vector_naive VV;
  typedef polynomial_multiply_naive<VV> MV;
  typedef polynomial_naive PV;
};

#undef TMPL
} // namespace mmx
#endif //__MMX__POLYNOMIAL_TANGENT__HPP
