
/******************************************************************************
* MODULE     : polynomial_unroll_naive.hpp
* DESCRIPTION: Loop unrolling for naive polynomial arithmetic
* COPYRIGHT  : (C) 2008  Gregoire Lecerf
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__POLYNOMIAL_UNROLL_NAIVE__HPP
#define __MMX__POLYNOMIAL_UNROLL_NAIVE__HPP
#include <algebramix/polynomial_naive.hpp>
#include <algebramix/polynomial.hpp>
#include <numerix/simd.hpp>

#define TMPL template<typename C>
#define TMPLK template<typename C, typename K>

namespace mmx {

/******************************************************************************
* Multiplication
******************************************************************************/

template<typename V, typename C, typename K>
struct polynomial_mul_helper {
  static void
  op (C* dest, const C* s1, const K* s2, nat n1, nat n2) {
    nat l = aligned_size<K,V> (n2);
    K* rev_s2 = mmx_new<K> (l);
    V::reverse (rev_s2, s2, n2);
    for (nat i = 1; i <= n2; i++, dest++)
      *dest = V::Vec::inn_prod (s1    , rev_s2 + n2 - i, min (i, n1));
    for (nat i = 1; i <  n1; i++, dest++)
      *dest = V::Vec::inn_prod (s1 + i, rev_s2         , min (n2, n1 - i));
    mmx_delete (rev_s2, l);
  } 
};

template<typename V, typename C>
struct polynomial_tmul_helper {
  static void
  op (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
    for (nat i = 0; i < n2; i++, dest++, s2++)
      *dest = V::Vec::inn_prod (s1, s2, n1); 
  }
};

template<typename V, nat m>
struct polynomial_unroll_naive_multiplication: public V {

TMPLK static void
mul (C* dest, const C* s1, const K* s2, nat n1, nat n2) {
  V::clear (dest, n1+n2-1);
  polynomial_mul_helper<V,C,K>::op (dest, s1, s2, n1, n2);
}

TMPL static inline void
square (C* dest, const C* s, nat n) {
  mul (dest, s, s, n, n);
}

TMPL static void
tmul (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
  V::clear (dest, n2);
  polynomial_tmul_helper<V,C>::op (dest, s1, s2, n1, n2);
}

}; // polynomial_unroll_naive_multiplication<V>

/******************************************************************************
* Division
******************************************************************************/

template<typename V, typename C>
struct polynomial_quo_rem_helper {
  static void
  op (C* dest, C* s1, const C* s2, nat n1, nat n2) {
    // (s1, n1) contains the numerator on input and the remainder on output
    // (s2, n2) contains the denominator. We assume n2>0 and s2[n2-1] != 0
    // (dest, n1-n2+1) contains the quotient
    while (n1 >= n2 && n1 != 0) {
      int d= n1-n2;
      C q= quo (s1[n1-1], s2[n2-1]);
      dest[d]= q;
      V::Vec::template vec_binary_scalar<mul_add_op,C,C,C> (s1 + d, s2, -q, n2);
      n1--;
    }
  }
};

template<typename V>
struct polynomial_unroll_naive_division: public V {

TMPL static void
quo_rem (C* dest, C* s1, const C* s2, nat n1, nat n2) {
  polynomial_quo_rem_helper<V,C>::op (dest, s1, s2, n1, n2);
}

};

/******************************************************************************
* Graeffe transformations
******************************************************************************/

template<typename V>
struct polynomial_unroll_graeffe: public V {

TMPL static void
graeffe (C* dest, const C* s, nat n) {
  nat l1= (n+1) >> 1, l2= n >> 1;
  // More allocation is done to ensure memory alignment
  nat l= aligned_size<C,V> (n + 1);
  C* c1= mmx_new<C> (l);
  C* c2= mmx_new<C> (l);
  if (n>0) dest[n-1]= C(0);
  V::Vec::half_copy (c1, s  , l1);
  V::square (dest, c1, l1);
  V::Vec::half_copy (c1, s+1, l2);
  V::square (c2 , c1, l2);
  V::copy (c1 + 1, c2, l2 << 1);
  c1[0] = C(0);
  V::sub (dest, c1, l2 << 1);
  mmx_delete<C> (c1, l);
  mmx_delete<C> (c2, l);
}

}; // polynomial_graeffe<V>

/******************************************************************************
* Efficient evaluation of polynomials
******************************************************************************/

template<typename V, typename C>
struct polynomial_eval_helper {
  static inline C
  op (const C* p, const C& x, nat l) {
    return polynomial_naive_evaluation<V>::eval (p, x, l); }
};

template<typename V>
struct polynomial_unroll_naive_evaluation: public V {

TMPL static void
annulator (C* d, const C* s, nat n) {
  (void) d; (void) s; (void) n;
  ERROR ("expand not yet implemented");
}

TMPL static C
eval (const C* p, const C& x, nat l) {
  return polynomial_eval_helper <V, C>::op (p, x, l);
}

TMPL static void
eval (C* v, const C* p, const C* x, nat l, nat n) {
  // Evaluate the polynomial (p, l) at points x[0], ..., x[n-1]
  for (nat j=0; j<n; j++)
    v[j]= eval (p, x[j], l);
}

TMPL static void
expand (C** v, const C* p, const C* x, const nat* nu, nat n, nat k) {
  (void) v; (void) p; (void) x; (void) nu; (void) n; (void) k;
  ERROR ("expand not yet implemented");
}

}; // polynomial_unroll_naive_evaluation<V>

/******************************************************************************
* Unroll naive polynomials
******************************************************************************/

template<typename V, nat m>
struct polynomial_multiply_unroll_naive:
  public polynomial_unroll_naive_multiplication<
	   polynomial_naive_linear<
	     polynomial_vectorial<V> >, m>
{};

template<typename V>
struct polynomial_algos_unroll_naive:
  public polynomial_unroll_naive_evaluation<
	   polynomial_naive_euclidean<
	     polynomial_dicho_composition< // FIXME: implement naive algo
	       polynomial_unroll_graeffe<
	         polynomial_unroll_naive_division<V> > > > >
{};

template<nat m>
struct polynomial_unroll_naive:
  public polynomial_algos_unroll_naive<
	   polynomial_multiply_unroll_naive<
	     vector_unroll<m>, m> >
{};

} // namespace mmx

#undef TMPL
#undef TMPLK
#endif //__MMX__POLYNOMIAL_UNROLL_NAIVE__HPP
