
/******************************************************************************
* MODULE     : subresultant_prem.hpp
* DESCRIPTION: Univariate subresultant of polynomials based on pseudo division
* COPYRIGHT  : (C) 2008  Gregoire Lecerf
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_SUBRESULTANT_PREM_HPP
#define __MMX_SUBRESULTANT_PREM_HPP
#include <algebramix/matrix.hpp>
#include <algebramix/polynomial.hpp>

namespace mmx {

#define TMPL     template<typename Polynomial> 
#define Scalar   typename scalar_type_helper<Polynomial>::val
#define PVector  vector    <Polynomial, typename V::VV>

template<typename V>
struct subresultant_prem_enrich: V {

TMPL static inline PVector
subresultants (const Polynomial& f, const Polynomial& g) {
  //! Subresultants from degree 0 to min (deg (f), deg(g)) - 1.
  int n= degree (f), m= degree (g);
  if (n < 0 || m < 0) return PVector (Polynomial (0), 0);
  if (n < m) {
    PVector s= V::subresultants (g, f);
    for (int k=0; k < n; k++) 
      s[k]= ((m+n-1) & (n-k) & 1) ? -s[k] : s[k];
    return s;
  }
  return V::subresultants (f, g); }

TMPL static Polynomial
subresultant (const Polynomial& f, const Polynomial& g, int k) {
  //! kth polynomial subresultant of f and g
  int m= min (degree (f), degree (g));
  if (m < 0 || k < 0 || k >= m) return Polynomial (Scalar(0));
  PVector s= subresultants (f, g); return s[k]; }

TMPL static inline Scalar
subresultant (const Polynomial& f, const Polynomial& g, int k, int l) {
  //! coefficient of degree l of the kth polynomial subresultant
  return subresultant (f, g, k)[l]; }

TMPL static Scalar
resultant (const Polynomial& f, const Polynomial& g) {
  int n= degree (f), m= degree (g);
  if (n < 0 || m < 0) return Scalar(0);
  if (m == 0) return binpow (g[0], n);
  if (n == 0) {
    Scalar r= binpow (f[0], m); 
    return (n & 1) ? -r : r;
  }
  return subresultant (f, g, 0, 0); }
};

template<typename V>
struct subresultant_prem_algo: V {
  //! Lionel Ducos, "Optimizations of the subresultant algorithm,
  //! JPAA, 1998 (p. 150).

TMPL static PVector
subresultants (const Polynomial& f, const Polynomial& g) {
  //! Subresultants from degree 0 to min (deg (f), deg(g)) - 1.
  int n= degree (f), m= degree (g);
  VERIFY (n >= m, "incorrect degrees");
  PVector s (Polynomial (0), m);
  if (m == 0) return s;
  Scalar d= binpow (g[m], n-m);
  Polynomial a= g, b= prem (f, g);
  s[m-1]= b;
  n= degree (a); m= degree (b);
  while (m >= 0) {
    s[m]= binpow (b[m], n-m-1) * b / binpow (d, n-m-1);
    if (m == 0) break;
    s[m-1]= prem (a, b) / (-a[n] * binpow (-d, n-m));
    a= s[m]; b= s[m-1]; d= a[m];
    n= degree (a); m= degree (b);
  }
  return s; }

};

template<typename V>
struct subresultant_optimized_prem_defected: V {

template<typename T> static inline T
defected_pow (const T& x, const T& y, int n) {
  // return x^n / y^(n-1)
  VERIFY (n >= 1, "bug");
  if (n == 1) return x;
  T z= square (defected_pow (x, y, n >> 1)) / y;
  return ((n & 1) == 0) ? z : ((x * z) / y); }

TMPL static inline Polynomial
defected_scaling (const Scalar& x, const Scalar& y, const Polynomial& b, int n) {
  // return (x^n * b) / y^n
  VERIFY (n >= 0, "bug");
  if (n == 0) return b;
  if (n == 1) return (x * b) / y;
  return (defected_pow (x, y, n) * b) / y; }
};
  
template<typename V>
struct subresultant_optimized_prem_twisted_op: V {

TMPL static inline Polynomial
defected_prem (const Polynomial& s_n, const Polynomial& s_nm1,
	       const Polynomial& s_m, const Scalar& x) {
  // Twisted pseudo remainder
  int n= degree (s_n), m= degree (s_nm1), k= n;
  Scalar c_m= s_m[degree (s_m)];
  Scalar y= s_nm1[m];
  Polynomial r= s_n;
  r = y * r - Polynomial (Scalar(1), n-m) * r[n] * s_nm1; k--;
  while (k >= m) {
    r = (y * r - Polynomial (Scalar(1), k-m) * r[k] * s_nm1) / x; k--;
  }
  return ((n-m+1) & 1) ? r / (-s_n[n]) : r / s_n[n]; }
};

template<typename V>
struct subresultant_optimized_prem_ducos_op: V {

TMPL static inline Polynomial
defected_prem (const Polynomial& s_n, const Polynomial& s_nm1,
	       const Polynomial& s_m, const Scalar& x) {
  // Ibid. p. 153.
  int n= degree (s_n), m= degree (s_nm1);
  Scalar c_n= s_n[n], c_m= s_m[m], c_nm1= s_nm1[m];
  // Polynomial r= rem (c_m * c_nm1 * s_n, s_nm1) / x;
  // return ((n-m+1) & 1) ? r / (-s_n[n]) : r / s_n[n];
  Polynomial h= - range (s_m, 0, m), t, r;
  Polynomial d= c_m * range (s_n, 0, m) + s_n[m] * h;
  for (int i= m+1; i < n; i++) {
    t= Polynomial (1,1) * h;
    h= t - (t[m] * s_nm1) / c_nm1;
    d += s_n[i] * h;
  }
  d /= c_n;
  t= Polynomial(1,1) * h;
  h= c_nm1 * t - t[m] * s_nm1;
  r= h + c_nm1 * d;
  return ((n-m+1) & 1) ? r / (-x) : r / x; }
};


template<typename V>
struct subresultant_optimized_prem_algo: V {

TMPL static PVector
subresultants (const Polynomial& f, const Polynomial& g) {
  int n= degree (f), m= degree (g);
  VERIFY (n >= m, "incorrect degrees");
  PVector s (Polynomial (0), m);
  if (m == 0) return s;
  Scalar d= binpow (g[m], n-m);
  Polynomial a= g, b= prem (f, g);
  s[m-1]= b;
  n= degree (a); m= degree (b);
  while (m >= 0) {
    s[m]= defected_scaling (b[m], d, b, n-m-1);
    if (m == 0) break;
    s[m-1]= defected_prem (a, b, s[m], d);
    a= s[m]; b= s[m-1]; d= a[m];
    n= degree (a); m= degree (b);
  }
  return s; }

};

#undef TMPL
#undef Polynomial
#undef Scalar
#undef PVector

/******************************************************************************
* Implementation of the subresultants via pseudo division
******************************************************************************/

struct subresultant_prem_base {
  typedef vector_naive     VV;
  typedef matrix_naive     MV;
  typedef polynomial_naive PV;
};

struct subresultant_prem :
    subresultant_prem_enrich <
      subresultant_prem_algo <subresultant_prem_base> > {};

struct subresultant_prem_twisted :
    subresultant_prem_enrich <
      subresultant_optimized_prem_algo <
	subresultant_optimized_prem_twisted_op <
	  subresultant_optimized_prem_defected <
	    subresultant_prem_base> > > > {};

struct subresultant_optimized_prem :
    subresultant_prem_enrich <
      subresultant_optimized_prem_algo <
	subresultant_optimized_prem_ducos_op <
	  subresultant_optimized_prem_defected <
	    subresultant_prem_base> > > > {};

} // namespace mmx
#endif // __MMX_SUBRESULTANT_PREM_HPP
