
/******************************************************************************
* MODULE     : vector_aligned.hpp
* DESCRIPTION: vectors aligned in memory
* COPYRIGHT  : (C) 2008  Joris van der Hoeven and Gregoire Lecerf
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__VECTOR_ALIGNED__HPP
#define __MMX__VECTOR_ALIGNED__HPP
#include <basix/basix.hpp>
#include <numerix/int.hpp>
#include <numerix/simd.hpp>

namespace mmx {

/******************************************************************************
* Helpers to be specialized on aligned hardware data
******************************************************************************/
// V is the variant to be used on base types
// W is the variant to be used on simd types (see vector_sse.hpp)

template<typename V, typename W,
	 typename Op, typename T>
struct vec_nullary_aligned_helper {
  static inline void op (T* dest, nat n) {
    V::template vec_nullary<Op,T> (dest, n); } };

template<typename V, typename W,
	 typename Op, typename T, typename C>
struct vec_unary_aligned_helper {
  static inline void op (T* dest, const C* s, nat n) {
    V::template vec_unary<Op,T,C> (dest, s, n); } };

template<typename V, typename W,
	 typename Op, typename T, typename C1, typename C2>
struct vec_binary_aligned_helper {
  static inline void op (T* dest, const C1* s1, const C2* s2, nat n) {
    V::template vec_binary<Op,T,C1,C2> (dest, s1, s2, n); } };

template<typename V, typename W,
	 typename Op, typename T, typename X>
struct vec_unary_scalar_aligned_helper {
  static inline void op (T* dest, const X& x, nat n) {
    V::template vec_unary_scalar<Op,T,X> (dest, x, n); } };

template<typename V, typename W,
	 typename Op, typename T, typename C, typename X>
struct vec_binary_scalar_aligned_helper {
  static inline void op (T* dest, const C* s, const X& x, nat n) {
    V::template vec_binary_scalar<Op,T,C,X> (dest, s, x, n); } };

template<typename V, typename W,
	 typename Op, typename C>
struct vec_unary_test_aligned_helper {
  static inline bool op (const C* s, nat n) {
    return V::template vec_unary_test (s, n); } };

template<typename V, typename W,
	 typename Op, typename C1, typename C2>
struct vec_binary_test_aligned_helper {
  static inline bool op (const C1* s1, const C2* s2, nat n) {
    return V::template vec_binary_test<Op,C1,C2> (s1, s2, n); } };

template<typename V, typename W,
	 typename Op, typename C, typename X>
struct vec_binary_test_scalar_aligned_helper {
  static inline bool op (const C* s, const X& x, nat n) {
    return V::template vec_binary_test_scalar<Op,C,X> (s, x, n); } };

template<typename V, typename W,
	 typename Op, typename C>
struct vec_unary_big_aligned_helper {
  static inline C op (const C* s, nat n) {
    return V::template vec_unary_big<Op,C> (s, n); } };

template<typename V, typename W,
	 typename Op, typename C1, typename C2>
struct vec_binary_big_aligned_helper {
  static inline C1 op (const C1* s1, const C2* s2, nat n) {
    return V::template vec_binary_big<Op,C1,C2> (s1, s2, n); } };

/******************************************************************************
* Abstract low level routines on vectors always aligned in memory
******************************************************************************/

template<typename V, typename W>
struct vector_always_aligned_abstractions : public V {

// Alignment 'len' of C is determined from its corresponding simd type.
// So far momory alignment is only suported for multiples of 16 bytes.
template<typename C>
struct mask_helper {
  static const nat m           = Simd_size (C);
  static const nat log_m       = int_bitsize_helper<nat, m>::value - 1;
  static const nat hi_mask_m   = ((nat) -1) << log_m;
  static const nat lo_mask_m   = m - 1;
  static const intptr_t len    = m * sizeof (C);
  static const intptr_t log_len= int_bitsize_helper<nat, len>::value - 1;
  static const intptr_t lo_mask_len= len - 1;
};

template<typename C> static inline nat
vec_aligned_size (nat n) {
  VERIFY (mask_helper<C>::m == (((nat) 1) << mask_helper<C>::log_m),
	  "simd size must be a power of two");
  return (n + mask_helper<C>::m - 1) & mask_helper<C>::hi_mask_m;
}

template<typename C> static inline nat
vec_floor_aligned_size (nat n) {
  return n & mask_helper<C>::hi_mask_m;
}

template<typename C> static inline bool
vec_is_aligned_size (nat n) {
  return (n & mask_helper<C>::lo_mask_m) == 0; }

template<typename C> static inline bool
vec_is_aligned (const C* s) {
  return (mask_helper<C>::len == 16) ?
    (((intptr_t) s) & mask_helper<C>::lo_mask_len) == 0 : false; }

template<typename C> static inline intptr_t
vec_ceil_shift (const C* s) {
  if (mask_helper<C>::len == 16) {
    intptr_t r= (intptr_t) s & mask_helper<C>::lo_mask_len;
    return (r != 0) ? mask_helper<C>::len - r : r;
  }
  return 0; }

template<typename Op, typename T> static inline void
vec_nullary (T* dest, nat n) {
  VERIFY (vec_is_aligned (dest), "address must be aligned");
  VERIFY (vec_is_aligned_size<T> (n), "size must be aligned");
  vec_nullary_aligned_helper<V,W,Op,T>::op (dest, n); }

template<typename Op, typename T, typename C> static inline void
vec_unary (T* dest, const C* s, nat n) {
  VERIFY (vec_is_aligned (dest), "address must be aligned");
  VERIFY (vec_is_aligned (s)   , "address must be aligned");
  VERIFY (vec_is_aligned_size<T> (n), "size must be aligned");
  VERIFY (vec_is_aligned_size<C> (n), "size must be aligned");
  vec_unary_aligned_helper<V,W,Op,T,C>::op (dest, s, n); }

template<typename Op, typename T, typename C1, typename C2> static inline void
vec_binary (T* dest, const C1* s1, const C2* s2, nat n) {
  VERIFY (vec_is_aligned (dest)  , "address must be aligned");
  VERIFY (vec_is_aligned (s1)    , "address must be aligned");
  VERIFY (vec_is_aligned (s2)    , "address must be aligned");
  VERIFY (vec_is_aligned_size<T> (n), "size must be aligned");
  VERIFY (vec_is_aligned_size<C1>(n), "size must be aligned");
  VERIFY (vec_is_aligned_size<C2>(n), "size must be aligned");
  vec_binary_aligned_helper<V,W,Op,T,C1,C2>::op (dest, s1, s2, n); }

template<typename Op, typename T, typename X> static inline void
vec_unary_scalar (T* dest, const X& x, nat n) {
  VERIFY (vec_is_aligned (dest), "address must be aligned");
  VERIFY (vec_is_aligned_size<T> (n), "size must be aligned");
  vec_unary_scalar_aligned_helper<V,W,Op,T,X>::op (dest, x, n); }

template<typename Op, typename T, typename C, typename X> static inline void
vec_binary_scalar (T* dest, const C* s, const X& x, nat n) {
  VERIFY (vec_is_aligned (dest), "address must be aligned");
  VERIFY (vec_is_aligned (s)   , "address must be aligned");
  VERIFY (vec_is_aligned_size<T> (n), "size must be aligned");
  VERIFY (vec_is_aligned_size<C> (n), "size must be aligned");
  vec_binary_scalar_aligned_helper<V,W,Op,T,C,X>::op (dest, s, x, n); }

template<typename Op, typename C> static inline bool
vec_unary_test (const C* s, nat n) {
  VERIFY (vec_is_aligned (s), "address must be aligned");
  VERIFY (vec_is_aligned_size<C> (n), "size must be aligned");
  return vec_unary_test_aligned_helper<V,W,Op,C>::op (s, n); }

template<typename Op, typename C1, typename C2> static inline bool
vec_binary_test (const C1* s1, const C2* s2, nat n) {
  VERIFY (vec_is_aligned (s1), "address must be aligned");
  VERIFY (vec_is_aligned (s2), "address must be aligned");
  VERIFY (vec_is_aligned_size<C1> (n), "size must be aligned");
  VERIFY (vec_is_aligned_size<C2> (n), "size must be aligned");
  return vec_binary_test_aligned_helper<V,W,Op,C1,C2>::op (s1, s2, n); }

template<typename Op, typename C, typename X> static inline bool
vec_binary_test_scalar (const C* s, const X& x, nat n) {
  VERIFY (vec_is_aligned (s), "address must be aligned");
  VERIFY (vec_is_aligned_size<C> (n), "size must be aligned");
  return vec_binary_test_scalar_aligned_helper<V,W,Op,C,X>::op (s, x, n); }

template<typename Op, typename C> static inline C
vec_unary_big (const C* s, nat n) {
  VERIFY (vec_is_aligned (s), "address must be aligned");
  VERIFY (vec_is_aligned_size<C> (n), "size must be aligned");
  return vec_unary_big_aligned_helper<V,W,Op,C>::op (s, n); }

template<typename Op, typename C> static inline C
vec_unary_big_dicho (const C* s, nat n) {
  return vec_unary_big<Op, C> (s, n); }
  
template<typename Op, typename C1, typename C2> static inline C1
vec_binary_big (const C1* s1, const C2* s2, nat n) {
  VERIFY (vec_is_aligned (s1), "address must be aligned");
  VERIFY (vec_is_aligned (s2), "address must be aligned");
  VERIFY (vec_is_aligned_size<C1> (n), "size must be aligned");
  VERIFY (vec_is_aligned_size<C2> (n), "size must be aligned");
  return vec_binary_big_aligned_helper<V,W,Op,C1,C2>::op (s1, s2, n); }

}; // vector_always_aligned_abstractions

/******************************************************************************
* Abstract low level routines on vectors not necessarily aligned in memory
******************************************************************************/

template<typename V, typename W>
struct vector_aligned_abstractions: public V {

typedef vector_always_aligned_abstractions<V, W> Aligned;

template<typename C> static inline nat
vec_aligned_size (nat n) {
  return Aligned::template vec_aligned_size<C> (n); }

template<typename C> static inline nat
vec_floor_aligned_size (nat n) {
  return Aligned::template vec_floor_aligned_size<C> (n); }

template<typename C> static inline bool
vec_is_aligned_size (nat n) {
  return Aligned::template vec_is_aligned_size<C> (n); }

template<typename C> static inline intptr_t
vec_ceil_shift (const C* s) {
  return Aligned::vec_ceil_shift (s); }

template<typename C> static inline bool
vec_is_aligned (const C* s) {
  return Aligned::vec_is_aligned (s); }

template<typename Op, typename T>
static inline void
vec_nullary (T* dest, nat n) {
  if (Aligned::template mask_helper<T>::len == 16) {
    nat r= min ((nat) vec_ceil_shift (dest), n);
    V::template vec_nullary<Op,T> (dest, r);
    dest += r; n -= r;
    r= vec_floor_aligned_size<T> (n);
    Aligned::template vec_nullary<Op,T> (dest, r);
    dest += r; n -= r;
  }
  V::template vec_nullary<Op,T> (dest, n);
}

template<typename Op, typename T, typename C> static inline void
vec_unary (T* dest, const C* s, nat n) {
  if (Aligned::template mask_helper<T>::len == 16 &&
      Aligned::template mask_helper<C>::len == 16) {
    nat rd= vec_ceil_shift (dest);
    nat rs= vec_ceil_shift (s);
    if (rd == rs) {
      nat r= min (rd, n);
      V::template vec_unary<Op,T,C> (dest, s, r);
      dest += r; s += r; n -= r;
      rd= vec_floor_aligned_size<T> (n);
      rs= vec_floor_aligned_size<C> (n);
      r= min (rd, rs);
      Aligned::template vec_unary<Op,T,C> (dest, s, r);
      dest += r; s += r; n -= r;
    }
  }
  V::template vec_unary<Op,T,C> (dest, s, n);
}

template<typename Op, typename T, typename C1, typename C2> static inline void
vec_binary (T* dest, const C1* s1, const C2* s2, nat n) {
  if (Aligned::template mask_helper<T >::len == 16 &&
      Aligned::template mask_helper<C1>::len == 16 &&
      Aligned::template mask_helper<C2>::len == 16) {
    nat rd = vec_ceil_shift (dest);
    nat rs1= vec_ceil_shift (s1);
    nat rs2= vec_ceil_shift (s2);
    if (rd == rs1 && rd == rs2) {
      nat r= min (rd, n);
      V::template vec_binary<Op,T,C1,C2> (dest, s1, s2, r);
      dest += r; s1 += r; s2 += r; n -= r;
      rd = vec_floor_aligned_size<T>  (n);
      rs1= vec_floor_aligned_size<C1> (n);
      rs2= vec_floor_aligned_size<C2> (n);
      r= min (min (rd, rs1), rs2);
      Aligned::template vec_binary<Op,T,C1,C2> (dest, s1, s2, r);
      dest += r; s1 += r; s2 += r; n -= r;
    }
  }
  V::template vec_binary<Op,T,C1,C2> (dest, s1, s2, n);
}

template<typename Op, typename T, typename X> static inline void
vec_unary_scalar (T* dest, const X& x, nat n) {
  if (Aligned::template mask_helper<T>::len == 16) {
    nat r= min ((nat) vec_ceil_shift (dest), n);
    V::template vec_unary_scalar<Op,T,X> (dest, x, r);
    dest += r; n -= r;
    r= vec_floor_aligned_size<T> (n);
    Aligned::template vec_unary_scalar<Op,T,X> (dest, x, r);
    dest += r; n -= r;
  }
  V::template vec_unary_scalar<Op,T,X> (dest, x, n);
}

template<typename Op, typename T, typename C, typename X> static inline void
vec_binary_scalar (T* dest, const C* s, const X& x, nat n) {
  if (Aligned::template mask_helper<T>::len == 16 &&
      Aligned::template mask_helper<C>::len == 16) {
    nat rd= vec_ceil_shift (dest);
    nat rs= vec_ceil_shift (s);
    if (rd == rs) {
      nat r= min (rd, n);
      V::template vec_binary_scalar<Op,T,C,X> (dest, s, x, r);
      dest += r; s += r; n -= r;
      rd= vec_floor_aligned_size<T> (n);
      rs= vec_floor_aligned_size<C> (n);
      r= min (rd, rs);
      Aligned::template vec_binary_scalar<Op,T,C,X> (dest, s, x, r);
      dest += r; s += r; n -= r;
    }
  }
  V::template vec_binary_scalar<Op,T,C,X> (dest, s, x, n);
}

template<typename Op, typename C> static inline bool
vec_unary_test (const C* s, nat n) {
  if (Aligned::template mask_helper<C>::len == 16) {
    nat r= min ((nat) vec_ceil_shift (s), n);
    if (! V::template vec_unary_test<Op,C> (s, r)) return false;
    s += r; n -= r;
    r= vec_floor_aligned_size<C> (n);
    if (! Aligned::template vec_unary_test<Op,C> (s, r)) return false;
    s += r; n -= r;
  }
  return V::template vec_unary_test<Op,C> (s, n);
}

template<typename Op, typename C1, typename C2> static inline bool
vec_binary_test (const C1* s1, const C2* s2, nat n) {
  if (Aligned::template mask_helper<C1>::len == 16 &&
      Aligned::template mask_helper<C2>::len == 16) {
    nat r1= vec_ceil_shift (s1);
    nat r2= vec_ceil_shift (s2);
    if (r1 == r2) {
      nat r= min (r1, r2);
      if (! V::template vec_binary_test<Op,C1,C2> (s1, s2, r))
	return false;
      s1 += r; s2 += r; n -= r;
      r1= vec_floor_aligned_size<C1> (n);
      r2= vec_floor_aligned_size<C2> (n);
      r= min (r1, r2);
      if (! Aligned::template vec_binary_test<Op,C1,C2> (s1, s2, r))
	return false;
      s1 += r; s2 += r; n -= r;
    }
  }
  return V::template vec_binary_test<Op,C1,C2> (s1, s2, n);
}
  
template<typename Op, typename C, typename X> static inline bool
vec_binary_test_scalar (const C* s, const X& x, nat n) {
  if (Aligned::template mask_helper<C>::len == 16) {
    nat r= min ((nat) vec_ceil_shift (s), n);
    if (! V::template vec_binary_test_scalar<Op,C,X> (s, x, r))
      return false;
    s += r; n -= r;
    r= vec_floor_aligned_size<C> (n);
    if (! Aligned::template vec_binary_test_scalar<Op,C,X> (s, x, r))
      return false;
    s += r; n -= r;
  }
  return V::template vec_binary_test_scalar<Op,C> (s, x, n);
}

template<typename Op, typename C> static inline C
vec_unary_big (const C* s, nat n) {
  if (Aligned::template mask_helper<C>::len == 16) {
    nat r= min ((nat) vec_ceil_shift (s), n);
    C c= V::template vec_unary_big<Op,C> (s, r);
    s += r; n -= r;
    r= vec_floor_aligned_size<C> (n);
    Op::in_place (c, Aligned::template vec_unary_big<Op,C> (s, r));
    s += r; n -= r;
    Op::in_place (c, V::template vec_unary_big<Op,C> (s, n));
    return c;
  }
  return V::template vec_unary_big<Op,C> (s, n);
}

template<typename Op, typename C> static inline C
vec_unary_big_dicho (const C* s, nat n) {
  return vec_unary_big<Op, C> (s, n);
}

template<typename Op, typename C1, typename C2> static inline C1
vec_binary_big (const C1* s1, const C2* s2, nat n) {
  if (vec_is_aligned (s1) && vec_is_aligned_size<C1> (n) &&
      vec_is_aligned (s2) && vec_is_aligned_size<C2> (n))
    return Aligned::template vec_binary_big<Op,C1,C2> (s1, s2, n);
  return V::template vec_binary_big<Op,C1,C2> (s1, s2, n);
}
  
}; // vector_aligned_abstractions

} // namespace mmx
#endif // __MMX__VECTOR_ALIGNED__HPP
