
/******************************************************************************
* MODULE     : analytic.hpp
* DESCRIPTION: Analytic functions
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_ANALYTIC_HPP
#define __MMX_ANALYTIC_HPP
#include <basix/vector.hpp>
#include <numerix/floating.hpp>
#include <numerix/complex_double.hpp>
#include <algebramix/series.hpp>
//#include <algebramix/series_relaxed.hpp>
#include <analyziz/newton_polygon.hpp>

namespace mmx {
class std_analytic {};
#define TMPL_DEF template<typename C,typename V= std_analytic>
#define TMPL template<typename C,typename V>
#define Analytic analytic<C,V>
#define Analytic_rep analytic_rep<C,V>
#define Implicit_analytic_rep implicit_analytic_rep<C,V>
#define Assume assume_bound<typename unvectorize<Error >::val>
#define Scalar typename unvectorize<C>::val
#define Series series<C>
#define Series_rep series_rep<C>
#define Implicit_series_rep implicit_series_rep<C>
#define Polynomial polynomial<C>
#define Error Abs_type(C)
TMPL class analytic;
TMPL Analytic lshiftz (const Analytic& f, const int& shift= 1);
TMPL Analytic rshiftz (const Analytic& f, const int& shift= 1);
typedef long long unsigned int llnat;

/******************************************************************************
* Important global parameters
******************************************************************************/

extern nat    fast_order;
extern double fast_convergence;
extern double fast_large_radius;
extern nat    radial_length;
extern double radial_resolution;
extern double radial_progress;
extern double radial_tolerance;

/******************************************************************************
* The analytic function class
******************************************************************************/

template<typename C>
struct unvectorize {
  typedef C val;
  static inline vector<val> encode (const C& x) {
    return vector<val> (x, 1); }
  static inline C decode (const vector<val>& v, const C& gauge) {
    (void) gauge;
    return v[0]; }
};

TMPL
class analytic_cache {
MMX_ALLOCATORS
public:
  Series              expansion;
  table<llnat,nat>    assumption;
  table<Error,nat>    head_bound;
  table<Error,nat>    tail_bound;
  table<Error,nat>    conv_bound;
  table<Analytic,Scalar > focus_at;
  table<Analytic,Scalar > move_to;
};

template<typename C>
class assume_bound {
MMX_ALLOCATORS
public:
  llnat serial;
  table<vector<C>,void*> assumption;
  table<vector<C>,void*> conclusion;
  inline assume_bound () {
    static llnat current_serial= 1;
    serial= current_serial++; }
};

TMPL_DEF
class analytic_rep REP_STRUCT {
public:
  analytic_cache<C,V>* cache;
  void init_cache ();

public:
  inline analytic_rep (): cache (NULL) {}
  inline virtual ~analytic_rep () { if (cache != NULL) delete cache; }
  virtual syntactic expression (const syntactic& z) const;

public:
  virtual Series Expand () const = 0;
  virtual Error Tail_bound (nat order, Assume& a) const = 0;
  virtual Analytic Focus (const Scalar& z) const = 0;
  virtual Analytic Move (const Scalar& z) const = 0;
  virtual Analytic Derive () const = 0;
  virtual Analytic Radial_move (const Scalar& z);
  virtual C Radial_eval (const Scalar& z);
  inline Analytic me () const;
  friend class Analytic;

  virtual C initial () const;
  virtual Analytic equation () const;
};

TMPL_DEF
class analytic {
INDIRECT_PROTO_2 (analytic, analytic_rep, C, V)
public:
  analytic (int c= 0);
  analytic (const C& c);
  analytic (const polynomial<C>& P);
  analytic (const vector<C>& coeffs);
  inline C operator [] (nat n) const;
};
INDIRECT_IMPL_2 (analytic, analytic_rep, typename C, C, typename V, V)

STYPE_TO_TYPE(TMPL,scalar_type,Analytic,C);

/******************************************************************************
* Forward definitions of operations on analytic functions
******************************************************************************/

TMPL Analytic operator + (const Analytic& f, const Analytic& g);
TMPL Analytic operator - (const Analytic& f);
TMPL Analytic operator - (const Analytic& f, const Analytic& g);
template<typename C,typename V,typename K> Analytic operator * (const K& c, const Analytic& f); 
TMPL Analytic operator * (const Analytic& f, const C& c);
TMPL Analytic operator / (const Analytic& f, const C& c);
TMPL Analytic operator * (const Analytic& f, const Analytic& g);
TMPL Analytic operator / (const Analytic& f, const Analytic& g);

TMPL Analytic integrate (const Analytic& f);
TMPL Analytic integrate_init (const Analytic& f, const C& c);

TMPL Analytic sqrt (const Analytic& f);
TMPL Analytic sqrt_init (const Analytic& f, const C& c);
TMPL Analytic exp (const Analytic& f);
TMPL Analytic log (const Analytic& f);
TMPL Analytic log_init (const Analytic& f, const C& c);
TMPL Analytic cos (const Analytic& f); // defined in analytic_linear.hpp
TMPL Analytic sin (const Analytic& f); // defined in analytic_linear.hpp
TMPL Analytic tan (const Analytic& f); // defined in analytic_linear.hpp
TMPL Analytic acos (const Analytic& f);
TMPL Analytic asin (const Analytic& f);
TMPL Analytic atan (const Analytic& f);

TMPL Analytic radial  (const Analytic& f); // defined in analytic_radial.hpp
TMPL Analytic riemann (const Analytic& f); // defined in analytic_riemann.hpp
TMPL Analytic heuristic (const polynomial<C>& f); // analytic_heuristic.hpp

/******************************************************************************
* Basic routines for analytic functions
******************************************************************************/

TMPL void
Analytic_rep::init_cache () {
  if (cache == NULL) {
    cache= new analytic_cache<C,V> ();
    cache->expansion = Expand ();
    cache->assumption= table<llnat,nat> (0);
  }
}

TMPL Series
expand (const Analytic& f) {
  inside (f) -> init_cache ();
  return f->cache->expansion;
}

TMPL Series
expand (const Analytic& f, nat order) {
  Series r= expand (f);
  if (order>0) (void) r[order-1];
  return r;
}

TMPL Polynomial
truncate (const Analytic& f, nat n) {
  return truncate (expand (f), n);
}

TMPL Error
tail_bound (const Analytic& f, nat order, Assume& a) {
  Analytic_rep* rep= inside (f);
  rep->init_cache ();
  if (read (rep->cache->assumption, order) != a.serial) {
    (void) expand (f, order);
    Error r= f->Tail_bound (order, a);
    rep->cache->assumption [order]= a.serial;
    rep->cache->tail_bound [order]= r;
  }
  return read (rep->cache->tail_bound, order);
}

template<typename C,typename T> C
minimum (const table<vector<C>,T>& t) {
  iterator<T> it= entries (t);
  C m= t[*it][0];
  for (++it; busy (it); ++it)
    for (nat i=0; i<N(t[*it]); i++)
      m= min (m, t[*it][i]);
  return m;
}

TMPL Error
tail_bound (const Analytic& f, nat order) {
  Assume a1;
  Error r= tail_bound (f, order, a1);
  if (N (a1.assumption) == 0) return r;
  nat iterations= (order >> 2) + 4;
  for (nat i=0; i<2*iterations; i++) {
    //cout << "Pass " << i << "\n";
    // One step with a simple iteration
    Assume a2;
    a2.assumption= a1.conclusion;
    r= tail_bound (f, order, a2);
    //cout << a2.assumption << " -> " << a2.conclusion << "\n";
    if (!(a2.assumption <= a2.conclusion)) return Infinity (Error);
    if (a2.assumption == a2.conclusion) return r;

    // One secant-method iteration, if possible
    Assume a3;
    table<vector<Error >,void*> d1= a1.conclusion - a1.assumption;
    table<vector<Error >,void*> d2= a2.conclusion - a2.assumption;
    if (d1 > d2) {
      Error l = minimum (d1 / (d1 - d2));
      a3.assumption= a1.assumption + l * (a1.conclusion - a1.assumption);
      r= tail_bound (f, order, a3);
      //cout << a3.assumption << " -> " << a3.conclusion << "\n";
      if (!(a3.assumption <= a3.conclusion)) return Infinity (Error);
      if (a3.assumption == a3.conclusion) return r;
    }
    else a3= a2;

    // Check whether we are done
    if (i >= iterations) {
      Assume a4;
      //a4.assumption= a3.conclusion + (a3.conclusion - a3.assumption);
      a4.assumption= Error (1.001) * a3.conclusion;
      r= tail_bound (f, order, a4);
      //cout << a4.assumption << " -> " << a4.conclusion << "\n";
      if (a4.conclusion <= a4.assumption) return r;
    }

    // Iterate
    a1.assumption= a3.assumption;
    a1.conclusion= a3.conclusion;
  }
  return Infinity (Error);
}

TMPL Error
head_bound (const Analytic& f, nat order) {
  Analytic_rep* rep= inside (f);
  rep->init_cache ();
  if (!rep->cache->head_bound->contains (order)) {
    Series s  = expand (f, order);
    Error  sum= 0;
    for (nat i=0; i<order; i++)
      sum += abs (s[i]);
    rep->cache->head_bound [order]= sum;
  }
  return read (rep->cache->head_bound, order);
}

TMPL Error
bound (const Analytic& f, nat order, Assume& a) {
  return head_bound (f, order) + tail_bound (f, order, a);
}

TMPL Error
conv_bound (const Analytic& me, const Analytic& f, const Analytic& g,
	    nat order)
{
  (void) me;
  Analytic_rep* rep= inside (f);
  rep->init_cache ();
  if (!rep->cache->conv_bound->contains (order)) {
    Error sum=0, cumul= 0;
    Series ff= expand (f, order);
    Series gg= expand (g, order);
    for (nat i=1; i<order; i++) {
      cumul += abs (gg[order-i]);
      sum   += abs (ff[i]) * cumul;
    }
    rep->cache->conv_bound [order]= sum;
  }
  return read (rep->cache->conv_bound, order);
}

TMPL Analytic
derive (const Analytic& f) {
  return f->Derive ();
}

TMPL Analytic
xderive (const Analytic& f) {
  return lshiftz (f->Derive ());
}

TMPL Analytic
focus (const Analytic& f, const Scalar& z) {
  Analytic_rep* rep= inside (f);
  rep->init_cache ();
  if (!rep->cache->focus_at->contains (z))
    rep->cache->focus_at [z]= rep->Focus (z);
  return read (rep->cache->focus_at, z);
}

TMPL Analytic
move (const Analytic& f, const Scalar& z) {
  Analytic_rep* rep= inside (f);
  rep->init_cache ();
  if (!rep->cache->move_to->contains (z))
    rep->cache->move_to [z]= rep->Move (z);
  return read (rep->cache->move_to, z);
}

TMPL C
eval (const Analytic& f, const Scalar& z) {
  nat order= mmx_bit_precision;
  // FIXME: determine 'order' in a better way
  C sum= f[order];
  for (nat i=1; i<=order; i++)
    sum= sum * z + f[order-i];
  return sum;
}

/******************************************************************************
* Radial analytic continuation
******************************************************************************/

template<typename C> Abs_type(C)
heuristic_radius (const polynomial<C>& p) {
  typedef Abs_type(C) R;
  nat order= fast_order;
  R* a= mmx_new<R> (order);
  for (nat i=0; i<order; i++) a[i]= abs (p[i]);
  nat n= order;
  while (n>0 && a[n-1] == 0) n--;
  while (n>0 && a[n-1] == 0) n--;
  if (n < (order>>2)) return Abs_type(C) (fast_large_radius);
  R r= 1 / newton_scale (a, n);
  mmx_delete<R> (a, order);
  return r;
}

template<typename C, typename Z> inline Z
heuristic_eval (const polynomial<C>& p, const Z& z) {
  ERROR ("not implemented");
  (void) p; (void) z; return z;
}

template<typename C> inline C
heuristic_eval (const polynomial<C>& p, const C& z) {
  typedef Abs_type(C) R;
  if (abs (z) < R(fast_convergence) * heuristic_radius (p)) return eval (p, z);
  return Nan (C);
}

TMPL Analytic
Analytic_rep::Radial_move (const Scalar& z) {
  return Move (z);
}

TMPL Analytic
radial_move (const Analytic& f, const Scalar& z) {
  return inside (f) -> Radial_move (z);
}

TMPL Analytic
broken_move (const Analytic& f, const vector<Scalar >& v) {
  Analytic r= f;
  for (nat i=0; i<N(v); i++)
    r= radial_move (r, v[i]);
  return r;
}

TMPL C
Analytic_rep::Radial_eval (const Scalar& z) {
  init_cache ();
  return heuristic_eval (truncate (cache->expansion, fast_order), z);
}

TMPL inline C
radial_eval (const Analytic& f, const Scalar& z) {
  return inside (f) -> Radial_eval (z);
}

TMPL C
broken_eval (const Analytic& f, const vector<Scalar >& v) {
  if (N(v) == 0) return radial_eval (f, 0);
  else {
    Analytic r= f;
    for (nat i=0; i<N(v)-1; i++)
      if (is_nan (radial_eval (r, v[i]))) return Nan (C);
      else r= radial_move (r, v[i]);
    return radial_eval (r, v[N(v)-1]);
  }
}

template<typename C> vector<C>
turn_around (const vector<C>& v, int eps= 1) {
  if (N(v) == 0) return v;
  vector<C> r= range (v, 0, N(v) - 1);
  C z= -v[N(v)-1];
  vector<C> aux= fill<C> (18);
  for (int i=0; i<=17; i++)
    aux[i]= z * exp ((2 * Pi (C) * Imaginary (C) * C (eps * i)) / 17);
  for (nat i=0; i<17; i++)
    r << (aux[i+1] - aux[i]);
  r << - reverse (range (v, 0, N(v) - 1));
  return r;
}

/******************************************************************************
* Standard functions
******************************************************************************/

TMPL generic
var (const Analytic& f) {
  return var (expand (f));
}

TMPL inline syntactic
Analytic_rep::expression (const syntactic& z) const {
  return flatten (expand (me ()), z);
}

TMPL inline Analytic
Analytic_rep::me () const {
  return Analytic (this, true);
}

TMPL C
Analytic_rep::initial () const {
  ERROR ("not implemented (initial)");
}

TMPL Analytic
Analytic_rep::equation () const {
  ERROR ("not implemented (equation)");
}

TMPL inline C
Analytic::operator [] (nat n) const {
  return expand (*this) [n];
}

HARD_TO_SOFT_IDENTITY_SUGAR(TMPL,Analytic)

TMPL inline nat hash (const Analytic& f) {
  return 7002007 + hash (expand (f)); }
TMPL inline bool operator == (const Analytic& f, const Analytic& g) {
  return expand (f) == expand (g); }
TMPL inline bool operator != (const Analytic& f, const Analytic& g) {
  return expand (f) != expand (g); }

TMPL syntactic
flatten (const Analytic& f, const syntactic& z) {
  if (mmx_series_formula_output) return f->expression (z);
  else return flatten (expand (f), z);
}

TMPL syntactic
flatten (const Analytic& f) {
  return flatten (f, as_syntactic (var (f)));
}

PRINT_SUGAR(TMPL,Analytic)

/******************************************************************************
* Implicit analytic functions
******************************************************************************/

template<typename C,typename V>
class analytic_implicit_series_rep:
  public Implicit_series_rep
{
protected:
  Analytic f;
public:
  analytic_implicit_series_rep (const Analytic& f2): f(f2) {}
  syntactic expression (const syntactic& z) const {
    return flatten (f, z); }
  Series initialize () {
    this->initial (0)= f->initial ();
    return expand (f->equation ()); }
};

TMPL_DEF
class implicit_analytic_rep: public Analytic_rep {
protected:
  C      init;
  Analytic equa;
public:
  inline implicit_analytic_rep () {}
  virtual syntactic expression (const syntactic& z) const {
    (void) z; return "implicit"; }
  inline void initialize () {
    init= this->Initial ();
    equa= this->Equation (); }
  C initial () const { return init; }
  Analytic equation () const { return equa; }
  virtual C Initial () const = 0;
  virtual Analytic Equation () const = 0;
  virtual Series Expand () const {
    Series_rep* rep= new analytic_implicit_series_rep<C,V> (this->me ());
    Series s (rep);
    this->cache->expansion= s;
    Series r= implicit (s);
    this->cache->expansion= r; // redudant, but for security
    return r; }
  virtual Error Tail_bound (nat order, Assume& a) const {
    void* code= (void*) this;
    Error zero= 0 * abs (init); // gets right dimensions in vector case
    if (!a.assumption->contains (code))
      a.assumption [code]= unvectorize<Error>::encode (zero);
    this->cache->assumption [order]= a.serial;
    this->cache->tail_bound [order]=
      unvectorize<Error>::decode (read (a.assumption, code), zero);
    Error bnd= tail_bound (this->equation (), order, a);
    if (order == 0) bnd += abs (this->initial ());
    this->cache->tail_bound [order]= bnd; // redundant, but for security
    a.conclusion [code]= unvectorize<Error>::encode (bnd);
    return bnd; }
};

TMPL Analytic
implicit (implicit_analytic_rep<C,V>* rep) {
  rep->initialize ();
  return (Analytic_rep*) rep;
}

/******************************************************************************
* Explicit analytic functions
******************************************************************************/

TMPL
class polynomial_analytic_rep: public Analytic_rep {
  polynomial<C> P;
public:
  polynomial_analytic_rep (const polynomial<C>& P2): P (P2) {}
  Series Expand () const { return Series (P); }
  Error Tail_bound (nat order, Assume& a) const {
    (void) a;
    Error  sum= 0;
    Series s  = expand (this->me (), order);
    for (nat i= order; i<N(P); i++)
      sum += abs (s[i]);
    return sum; }
  Analytic Focus (const Scalar& z) const {
    return Analytic (q_difference (P, z)); }
  Analytic Move (const Scalar& z) const {
    return Analytic (shift (P, C (z))); }
  //polynomial<Scalar> Q= seq (z, Scalar (1));
  //return Analytic (compose (P, Q)); }
  Analytic Derive () const {
    return Analytic (derive (P)); }
};

TMPL
Analytic::analytic (int c) {
  rep= new polynomial_analytic_rep<C,V> (polynomial<C> (C (c), 0));
}

TMPL
Analytic::analytic (const C& c) {
  rep= new polynomial_analytic_rep<C,V> (polynomial<C> (c, 0));
}

TMPL
Analytic::analytic (const polynomial<C>& P) {
  rep= new polynomial_analytic_rep<C,V> (P);
}

TMPL
Analytic::analytic (const vector<C>& coeffs) {
  rep= new polynomial_analytic_rep<C,V> (polynomial<C> (coeffs));
}

/******************************************************************************
* Abstract operations on analytic functions
******************************************************************************/

template<typename Op,typename C> inline analytic<C>
nullary_implicit_analytic (const C& c);

template<typename Op,typename C,typename V= std_analytic>
class nullary_implicit_analytic_rep: public Implicit_analytic_rep {
protected:
  C c;
public:
  inline nullary_implicit_analytic_rep (const C& c2): c (c2) {}
  syntactic expression (const syntactic& z) const { (void) z;
    return Op::op_init (flatten (c)); }
  C Initial () const {
    return c; }
  Analytic Equation () const {
    return Op::def (this->me ()); }
  Series Expand () const {
    return nullary_implicit_series<Op> (c); }
  Analytic Focus (const Scalar& z) const { (void) z;
    return nullary_implicit_analytic<Op> (c); }
  Analytic Move (const Scalar& z) const {
    return nullary_implicit_analytic<Op> (eval (this->me (), z)); }
  Analytic Derive () const {
    return Op::diff_op (this->me ()); }
};

template<typename Op,typename C> inline analytic<C>
nullary_implicit_analytic (const C& c= Op::template op<C> ()) {
  return implicit (new nullary_implicit_analytic_rep<Op,C> (c));
}

template<typename Op,typename C,typename V>
class unary_analytic_rep: public Analytic_rep {
protected:
  Analytic f;
public:
  inline unary_analytic_rep (const Analytic& f2): f(f2) {}
  Series Expand () const {
    return Op::op (expand (f)); }
  Analytic Focus (const Scalar& z) const {
    return Op::op (focus (f, z)); }
  Analytic Move (const Scalar& z) const {
    return Op::op (move (f, z)); }
  Analytic Derive () const {
    return Op::diff_op (this->me (), f); }
};

template<typename Op,typename C,typename V> inline Analytic
unary_analytic (const Analytic& f) {
  return (Analytic_rep*) new unary_analytic_rep<Op,C,V> (f);
}

template<typename Op,typename C,typename V>
class unary_implicit_analytic_rep: public Implicit_analytic_rep {
protected:
  Analytic f;
  bool with_init;
  C c;
public:
  inline unary_implicit_analytic_rep (const Analytic& f2):
    f (f2), with_init (false) {}
  inline unary_implicit_analytic_rep (const Analytic& f2, const C& c2):
    f (f2), with_init (true), c (c2) {}
  syntactic expression (const syntactic& z) const {
    if (with_init) return Op::op_init (flatten (f, z), flatten (c));
    else return Op::op (flatten (f, z)); }
  C Initial () const {
    if (with_init) return c;
    else return Op::op (expand (f) [0]); }
  Analytic Equation () const {
    return Op::def (this->me (), f); }
  Series Expand () const {
    if (with_init) return Op::op_init (expand (f), c);
    else return Op::op (expand (f)); }
  Analytic Focus (const Scalar& z) const {
    if (with_init) return Op::op_init (focus (f, z), c);
    else return Op::op (focus (f, z)); }
  Analytic Move (const Scalar& z) const {
    return Op::op_init (move (f, z), eval (this->me (), z)); }
    //if (with_init) return Op::op_init (move (f, z), c);
    //else return Op::op (move (f, z)); }
  Analytic Derive () const {
    return Op::diff_op (this->me (), f); }
};

template<typename Op,typename C,typename V> inline Analytic
unary_implicit_analytic (const Analytic& f) {
  return implicit (new unary_implicit_analytic_rep<Op,C,V> (f));
}

template<typename Op,typename C,typename V> inline Analytic
unary_implicit_analytic (const Analytic& f, const C& c) {
  return implicit (new unary_implicit_analytic_rep<Op,C,V> (f, c));
}

template<typename Op,typename C,typename V>
class binary_analytic_rep: public Analytic_rep {
protected:
  Analytic f, g;
public:
  inline binary_analytic_rep (const Analytic& f2, const Analytic& g2):
    f(f2), g (g2) {}
  Series Expand () const {
    return Op::op (expand (f), expand (g)); }
  Analytic Focus (const Scalar& z) const {
    return Op::op (focus (f, z), focus (g, z)); }
  Analytic Move (const Scalar& z) const {
    return Op::op (move (f, z), move (g, z)); }
  Analytic Derive () const {
    return Op::diff_op (this->me (), f, g); }
};

template<typename Op,typename C,typename V> inline Analytic
binary_analytic (const Analytic& f, const Analytic& g) {
  return (Analytic_rep*) new binary_analytic_rep<Op,C,V> (f, g);
}

template<typename Op,typename C,typename V>
class binary_implicit_analytic_rep: public Implicit_analytic_rep {
protected:
  Analytic f, g;
public:
  inline binary_implicit_analytic_rep (const Analytic& f2, const Analytic& g2):
    f (f2), g (g2) {}
  syntactic expression (const syntactic& z) const {
    return Op::op (flatten (f, z), flatten (g, z)); }
  C Initial () const {
    return Op::op (expand (f), expand (g)) [0]; }
  Analytic Equation () const {
    return Op::def (this->me (), f, g); }
  Series Expand () const {
    return Op::op (expand (f), expand (g)); }
  Analytic Focus (const Scalar& z) const {
    return Op::op (focus (f, z), focus (g, z)); }
  Analytic Move (const Scalar& z) const {
    return Op::op (move (f, z), move (g, z)); }
  Analytic Derive () const {
    return Op::diff_op (this->me (), f, g); }
};

template<typename Op,typename C,typename V> inline Analytic
binary_implicit_analytic (const Analytic& f, const Analytic& g) {
  return implicit (new binary_implicit_analytic_rep<Op,C,V> (f, g));
}

template<typename Op,typename C,typename V,typename X>
class binary_scalar_analytic_rep: public Analytic_rep {
protected:
  Analytic f;
  X x;
public:
  inline binary_scalar_analytic_rep (const Analytic& f2, const X& x2):
    f(f2), x (x2) {}
  Series Expand () const {
    return Op::op (expand (f), x); }
  Analytic Focus (const Scalar& z) const {
    return Op::op (focus (f, z), x); }
  Analytic Move (const Scalar& z) const {
    return Op::op (move (f, z), x); }
  Analytic Derive () const {
    return Op::diff_op (this->me (), f, x); }
};

template<typename Op,typename C,typename V,typename X> inline Analytic
binary_scalar_analytic (const Analytic& f, const X& x) {
  return (Analytic_rep*) new binary_scalar_analytic_rep<Op,C,V,X> (f, x);
}

/******************************************************************************
* Coefficient-wise operations
******************************************************************************/

TMPL
class add_analytic_rep: public binary_analytic_rep<add_op,C,V> {
public:
  add_analytic_rep (const Analytic& f, const Analytic& g):
    binary_analytic_rep<add_op,C,V> (f, g) {}
  Error Tail_bound (nat order, Assume& a) const {
    return tail_bound (this->f, order, a) + tail_bound (this->g, order, a); }
};

TMPL Analytic
operator + (const Analytic& f, const Analytic& g) {
  return (Analytic_rep*) new add_analytic_rep<C,V> (f, g);
}

TMPL Analytic
operator + (const C& f, const Analytic& g) {
  return (Analytic_rep*) new add_analytic_rep<C,V> (Analytic (f), g);
}

TMPL Analytic
operator + (const Analytic& f, const C& g) {
  return (Analytic_rep*) new add_analytic_rep<C,V> (f, Analytic (g));
}

TMPL
class neg_analytic_rep: public unary_analytic_rep<neg_op,C,V> {
public:
  neg_analytic_rep (const Analytic& f):
    unary_analytic_rep<neg_op,C,V> (f) {}
  Error Tail_bound (nat order, Assume& a) const {
    return tail_bound (this->f, order, a); }
};

TMPL Analytic
operator - (const Analytic& f) {
  return (Analytic_rep*) new neg_analytic_rep<C,V> (f);
}

TMPL
class sub_analytic_rep: public binary_analytic_rep<sub_op,C,V> {
public:
  sub_analytic_rep (const Analytic& f, const Analytic& g):
    binary_analytic_rep<sub_op,C,V> (f, g) {}
  Error Tail_bound (nat order, Assume& a) const {
    return tail_bound (this->f, order, a) + tail_bound (this->g, order, a); }
};

TMPL Analytic
operator - (const Analytic& f, const Analytic& g) {
  return (Analytic_rep*) new sub_analytic_rep<C,V> (f, g);
}

TMPL Analytic
operator - (const C& f, const Analytic& g) {
  return (Analytic_rep*) new sub_analytic_rep<C,V> (Analytic (f), g);
}

TMPL Analytic
operator - (const Analytic& f, const C& g) {
  return (Analytic_rep*) new sub_analytic_rep<C,V> (f, Analytic (g));
}

TMPL
class sc_mul_analytic_rep: public binary_scalar_analytic_rep<rmul_op,C,V,C> {
public:
  sc_mul_analytic_rep (const Analytic& f, const C& c):
    binary_scalar_analytic_rep<rmul_op,C,V,C> (f, c) {}
  Error Tail_bound (nat order, Assume& a) const {
    return tail_bound (this->f, order, a) * abs (this->x); }
};

TMPL Analytic
operator * (const Analytic& f, const C& c) {
  return (Analytic_rep*) new sc_mul_analytic_rep<C,V> (f, c);
}

template<typename C,typename V,typename K> Analytic
operator * (const K& c, const Analytic& f) {
  return (Analytic_rep*) new sc_mul_analytic_rep<C,V> (f, c);
}

TMPL Analytic
operator / (const Analytic& f, const C& c) {
  return (Analytic_rep*) new sc_mul_analytic_rep<C,V> (f, 1/c);
}

/******************************************************************************
* Multiplication and division
******************************************************************************/

TMPL
class mul_analytic_rep: public binary_analytic_rep<mul_op,C,V> {
public:
  mul_analytic_rep (const Analytic& f, const Analytic& g):
    binary_analytic_rep<mul_op,C,V> (f, g) {}
  Error Tail_bound (nat order, Assume& a) const {
    return
      conv_bound (this->me (), this->f, this->g, order) +
      head_bound (this->f, order) * tail_bound (this->g, order, a) +
      tail_bound (this->f, order, a) * bound (this->g, order, a); }
};

TMPL Analytic
operator * (const Analytic& f, const Analytic& g) {
  return (Analytic_rep*) new mul_analytic_rep<C,V> (f, g);
}

TMPL
class lshiftz_analytic_rep: public Analytic_rep {
protected:
  Analytic f;
  int shift;
public:
  lshiftz_analytic_rep (const Analytic& f2, int shift2):
    f (f2), shift (shift2) {}
  Series Expand () const {
    return lshiftz (expand (f), shift); }
  Error Tail_bound (nat order, Assume& a) const {
    Series s= expand (this->f, order);
    Error sum= tail_bound (this->f, order, a);
    if (shift >= 0)
      for (nat i=0; i < min (order, (nat) shift); i++)
	sum += abs (s[order-1-i]);
    else
      for (nat i=0; i < ((nat) -shift); i++)
	sum += abs (s[order+i]);
    return sum; }
  Analytic Focus (const Scalar& z) const {
    if (shift >= 0) return binpow (z, shift) * lshiftz (focus (f, z), shift);
    else return binpow (1/z, -shift) * lshiftz (focus (f, z), shift); }
  Analytic Move (const Scalar& z) const {
    (void) z;
    ERROR ("not implemented (Move)");
    return this->me (); }
  Analytic Derive () const {
    ERROR ("not implemented (Derive)");
    return this->me (); }
};

TMPL Analytic
lshiftz (const Analytic& f, const int& shift) {
  return (Analytic_rep*) new lshiftz_analytic_rep<C,V> (f, shift);
}

TMPL Analytic
rshiftz (const Analytic& f, const int& shift) {
  return (Analytic_rep*) new lshiftz_analytic_rep<C,V> (f, -shift);
}

TMPL Analytic
operator / (const Analytic& f, const Analytic& g) {
  return binary_implicit_analytic<div_op> (f, g);
}

TMPL Analytic
operator / (const C& f, const Analytic& g) {
  return binary_implicit_analytic<div_op> (Analytic (f), g);
}

ARITH_INT_SUGAR(TMPL,Analytic,C);
//VAR_ARITH_INT_SUGAR(TMPL,Analytic);
//TMPL inline Analytic operator * (const Analytic& x, int y) { return x * C(y); }
//TMPL inline Analytic operator / (const Analytic& x, int y) { return x / C(y); }
//TMPL inline Analytic operator / (int x, const Analytic& y) { return Analytic(x)/y; }

/******************************************************************************
* Integration
******************************************************************************/

TMPL
class integrate_analytic_rep: public Analytic_rep {
protected:
  Analytic f;
  C c;
public:
  integrate_analytic_rep (const Analytic& f2, const C& c2): f (f2), c (c2) {}
  Series Expand () const {
    return integrate_init (expand (f), c); }
  Error Tail_bound (nat order, Assume& a) const {
    Series s= expand (this->me (), order);
    return
      abs (s[order]) +
      (tail_bound (this->f, order, a) / (order+1)); }
  Analytic Focus (const Scalar& z) const {
    return integrate_init (C(z) * focus (f, z), c); }
  Analytic Move (const Scalar& z) const {
    return integrate_init (move (f, z), eval (this->me (), z)); }
  Analytic Derive () const {
    return f; }
};

TMPL Analytic
integrate (const Analytic& f) {
  return (Analytic_rep*) new integrate_analytic_rep<C,V> (f, C(0));
}

TMPL Analytic
integrate_init (const Analytic& f, const C& c) {
  return (Analytic_rep*) new integrate_analytic_rep<C,V> (f, c);
}

/******************************************************************************
* Elementary functions
******************************************************************************/

TMPL Analytic
sqrt (const Analytic& f) {
  return unary_implicit_analytic<sqrt_op> (f);
}

TMPL Analytic
sqrt_init (const Analytic& f, const C& c) {
  return unary_implicit_analytic<sqrt_op> (f, c);
}

TMPL Analytic
exp (const Analytic& f) {
  return unary_implicit_analytic<exp_op> (f);
}

TMPL Analytic
log (const Analytic& f) {
  return unary_implicit_analytic<log_op> (f);
}

TMPL Analytic
log_init (const Analytic& f, const C& c) {
  return unary_implicit_analytic<log_op> (f, c);
}

TMPL Analytic
pow (const Analytic& f, const C& c) {
  return exp (c * log (f));
}

TMPL Analytic
pow (const Analytic& f, const Analytic& g) {
  return exp (g * log (f));
}

// NOTE: cos, sin and tan defined in analytic_linear.hpp

TMPL Analytic
acos (const Analytic& f) {
  return unary_implicit_analytic<acos_op> (f);
}

TMPL Analytic
acos_init (const Analytic& f, const C& c) {
  return unary_implicit_analytic<acos_op> (f, c);
}

TMPL Analytic
asin (const Analytic& f) {
  return unary_implicit_analytic<asin_op> (f);
}

TMPL Analytic
asin_init (const Analytic& f, const C& c) {
  return unary_implicit_analytic<asin_op> (f, c);
}

TMPL Analytic
atan (const Analytic& f) {
  return unary_implicit_analytic<atan_op> (f);
}

TMPL Analytic
atan_init (const Analytic& f, const C& c) {
  return unary_implicit_analytic<atan_op> (f, c);
}

INV_TRIGO_SUGAR(TMPL,Analytic)
HYPER_SUGAR(TMPL,Analytic)
INV_HYPER_SUGAR(TMPL,Analytic)
ARG_HYPER_SUGAR(TMPL,Analytic)

/******************************************************************************
* Composition and functional inversion
******************************************************************************/

TMPL Analytic compose (const Analytic& f, const Analytic& g);
TMPL Analytic reverse (const Analytic& f);

TMPL
class compose_analytic_rep: public binary_analytic_rep<compose_op,C,V> {
public:
  compose_analytic_rep (const Analytic& f, const Analytic& g):
    binary_analytic_rep<compose_op,C,V> (f, g) {}
  Error Tail_bound (nat order, Assume& a) const {
    ERROR ("not yet implemented");
    return 0; }
  Analytic Focus (const Scalar& z) const {
    return compose (this->f, focus (this->g, z)); }
  Analytic Move (const Scalar& z) const {
    Analytic tg= move (this->g, z);
    Analytic tf= move (this->f, tg[0]);
    return compose (tf, tg - tg[0]); }
  Analytic Derive () const {
    return derive (this->g) * compose (derive (this->f), this->g); }
};

TMPL Analytic
compose (const Analytic& f, const Analytic& g) {
  return (Analytic_rep*) new compose_analytic_rep<C,V> (f, g);
}

TMPL
class reverse_analytic_rep: public unary_analytic_rep<reverse_op,C,V> {
public:
  reverse_analytic_rep (const Analytic& f):
    unary_analytic_rep<reverse_op,C,V> (f) {}
  Error Tail_bound (nat order, Assume& a) const {
    ERROR ("not yet implemented");
    return 0; }
  Analytic Focus (const Scalar& z) const {
    return reverse (this->f / z); }
  Analytic Move (const Scalar& a) const {
    C b= eval (this->me (), a);
    Analytic tf= move (this->f, b);
    return reverse (tf - tf[0]) + b; }
  Analytic Derive () const {
    return 1 / compose (derive (this->f), this->me ()); }
};

TMPL Analytic
reverse (const Analytic& f) {
  return (Analytic_rep*) new reverse_analytic_rep<C,V> (f);
}

#undef TMPL_DEF
#undef TMPL
#undef Analytic
#undef Analytic_rep
#undef Implicit_analytic_rep
#undef Assume
#undef Scalar
#undef Series
#undef Series_rep
#undef Implicit_series_rep
#undef Polynomial
#undef Error
} // namespace mmx
#endif // __MMX_ANALYTIC_HPP
