
/******************************************************************************
* MODULE     : analytic_radial.hpp
* DESCRIPTION: Radial continuation of analytic functions
* COPYRIGHT  : (C) 2007  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__ANALYTIC_RADIAL__HPP
#define __MMX__ANALYTIC_RADIAL__HPP
#include <analyziz/analytic.hpp>

namespace mmx {
#define TMPL template<typename CC,typename V>
// NOTE: typename unvectorize<CC>::val should coincide with CC
#define Analytic analytic<CC,V>
#define Analytic_rep analytic_rep<CC,V>
#define Series series<CC>
#define Assume assume_bound<typename unvectorize<Error >::val>
#define Error Abs_type(CC)

TMPL
class radial_analytic_rep: public Analytic_rep {
protected:
  typedef Fast_type(CC) C;
  typedef Real_type(C)  R;

  vector<analytic<CC> >  fs;
  vector<polynomial<C> > ps;
  vector<C>              cs;
  vector<R>              rs;

  void
  add_disk (const analytic<CC>& f, const C& c) {
    polynomial<C> p= fast (truncate (f, fast_order));
    R r= fast_convergence * heuristic_radius (p);
    fs << f;
    ps << p;
    cs << c;
    rs << r;
  }

  nat
  find_disk (const C& z) {
    for (nat i=0; i<N(fs); i++) {
      C c = cs[i];
      R r = rs[i];
      if (abs (z - c) < r) return i;
      if (i >= 1 && abs (Im (c * conj (z))) >= abs (z) *radial_tolerance* r) {
	// distance between line and center of disk is more than
	// radial_tolerance times its radius
	fs= range (fs, 0, i);
	ps= range (ps, 0, i);
	cs= range (cs, 0, i);
	rs= range (rs, 0, i);
	break;
      }
    }

    while (N(fs) < radial_length &&
	   abs (z) < radial_resolution * rs[N(rs)-1])
      {
	nat i= N(fs) - 1;
	C c = cs[i];
	R r = radial_progress * rs[i];
	R zz= norm (z);
	R cc= norm (c);
	R zc= Re (z * conj (c));
	R di= zc * zc - zz * (cc - r*r);
	R la= (zc + sqrt (di)) / zz;
	C u = la * z;
	add_disk (move (fs[i], u - c), u);
	i= N(fs) - 1;
	c= cs[i];
	r= rs[i];
	if (abs (z - c) < fast_convergence * r) return i;
      }

    return N(fs);
  }


public:
  inline radial_analytic_rep (const Analytic& f) {
    add_disk (f, C (0)); }
  syntactic expression (const syntactic& z) const {
    return apply ("radial", flatten (fs[0], z)); }
  Series Expand () const { return expand (fs[0]); }
  Error Tail_bound (nat order, Assume& a) const {
    return tail_bound (fs[0], order, a); }
  Analytic Focus (const CC& z) const {
    return radial (focus (fs[0], z)); }
  Analytic Move (const CC& z) const {
    return radial (move (fs[0], z)); }
  Analytic Derive () const {
    return radial (derive (fs[0])); }
  Analytic Radial_move (const CC& z) {
    nat i= find_disk (fast (z));
    ASSERT (i < N(fs), "convergence failed");
    return radial (move (fs[i], z - slow<CC> (cs[i]))); }
  CC Radial_eval (const CC& z) {
    nat i= find_disk (fast (z));
    if (i < N(fs)) {
      C ret= eval (ps[i], fast (z) - cs[i]);
      if (!is_nan (ret)) return slow<CC> (ret);
    }
    return Nan (CC); }
};

TMPL Analytic
radial (const Analytic& f) {
  // radial (f) returns the same analytic functions as f,
  // but with special optimizations for fast numeric continuation and
  // evaluation along radial straightline paths from the origin
  return (Analytic_rep*) new radial_analytic_rep<CC,V> (f);
}

#undef TMPL
#undef Analytic
#undef Analytic_rep
#undef Series
#undef Assume
#undef Error
} // namespace mmx

#endif //__MMX__ANALYTIC_RADIAL__HPP
