
/******************************************************************************
* MODULE     : analytic_riemann.hpp
* DESCRIPTION: Analytic contination with cache for underlying Riemann surface
* COPYRIGHT  : (C) 2007  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__ANALYTIC_RIEMANN__HPP
#define __MMX__ANALYTIC_RIEMANN__HPP
#include <analyziz/analytic_radial.hpp>
#include <analyziz/riemann_germ.hpp>
#include <analyziz/riemann_bundle.hpp>

namespace mmx {
#define TMPL template<typename CC,typename V>
// NOTE: typename unvectorize<CC>::val should coincide with CC
#define Analytic analytic<CC,V>
#define Analytic_rep analytic_rep<CC,V>
#define Series series<CC>
#define Assume assume_bound<typename unvectorize<Error >::val>
#define Error Abs_type(CC)
#define C Fast_type(CC)
#define R Real_type(C )
#define Surface riemann_surface<R >
#define Point riemann_point<R >
#define Germ riemann_germ<R >
#define Bundle riemann_bundle<Germ,R >

TMPL Analytic riemann (const Analytic& f, const Bundle& b, const Point& o);

TMPL
class riemann_analytic_rep: public Analytic_rep {
protected:
  Analytic f;
  Bundle   bundle;
  Point    origin;
  //nat      counter;

  Bundle
  make_bundle (const Analytic& phi, const C& z) {
    //counter= 0;
    polynomial<C > pol= fast (truncate (phi, fast_order));
    R r= fast_convergence * heuristic_radius (pol);
    ASSERT (r > 0, "convergence failed");
    Surface rs= riemann_disk<R,std_riemann> (z, 0.9999 * r);
    return Bundle (rs, Germ (pol, z, r));
  }

  Germ
  find_germ (const C& delta) {
    Point p= move (bundle, origin, delta);
    C     z= project (origin) + delta;
    while (project (p) != z) {
      //if (counter == 8) break;
      Analytic phi= radial_move (f, slow<CC> (project (p) - project (origin)));
      polynomial<C > pol= fast (truncate (phi, fast_order));
      R r= fast_convergence * heuristic_radius (pol);
      if (radial_resolution * r < abs (delta)) break;
      Surface rs= riemann_disk<R,std_riemann> (project (p), 0.9999 * r);
      Bundle  b = Bundle (rs, Germ (pol, project (p), r));
      //cout << "rs= " << b -> surface << "\n";
      Point   q = lift (b, project (p));
      glue (bundle, p, b, q);
      //counter++;
      p= move (bundle, p, z - project (p));
    }
    return germ (bundle, p);
  }

public:
  riemann_analytic_rep (const Analytic& f2):
    f (f2), bundle (make_bundle (f2, C(0))), origin (lift (bundle, C(0))) {}
  riemann_analytic_rep (const Analytic& f2, const Bundle& b, const Point& o):
    f (f2), bundle (b), origin (o) {}
  syntactic expression (const syntactic& z) const {
    return apply ("riemann", flatten (f, z)); }
  Series Expand () const { return expand (f); }
  Error Tail_bound (nat order, Assume& a) const {
    return tail_bound (f, order, a); }
  Analytic Focus (const CC& z) const {
    return riemann (focus (f, z)); }
  Analytic Move (const CC& z) const {
    riemann_analytic_rep<CC,V>* rep=
      const_cast<riemann_analytic_rep<CC,V>*> (this);
    (void) rep->find_germ (fast (z));
    return riemann (move (f, z), bundle, move (bundle, origin, fast (z))); }
  Analytic Derive () const {
    return riemann (derive (f)); }
  Analytic Radial_move (const CC& z) {
    (void) find_germ (fast (z));
    return riemann (radial_move (f, z), bundle,
		    move (bundle, origin, fast (z))); }
  CC Radial_eval (const CC& z) {
    Germ g= find_germ (fast (z));
    if (inside (project (origin) + fast (z), g)) {
      C ret= eval (g, project (origin) + fast (z));
      if (!is_nan (ret)) return slow<CC> (ret);
    }
    return Nan (CC); }
};

TMPL Analytic
riemann (const Analytic& f) {
  // riemann (f) returns the same analytic functions as f,
  // but with a cache for numeric evaluation of f
  return (Analytic_rep*) new riemann_analytic_rep<CC,V> (radial (f));
}

TMPL Analytic
riemann (const Analytic& f, const Bundle& b, const Point& o) {
  return (Analytic_rep*) new riemann_analytic_rep<CC,V> (f, b, o);
}

#undef TMPL
#undef Analytic
#undef Analytic_rep
#undef Series
#undef Assume
#undef Error
#undef C
#undef R
#undef Surface
#undef Point
#undef Bundle
} // namespace mmx

#endif //__MMX__ANALYTIC_RIEMANN__HPP
