
/******************************************************************************
* MODULE     : matrix_ball.hpp
* DESCRIPTION: Ball multiplication of matrices
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__MATRIX_BALL__HPP
#define __MMX__MATRIX_BALL__HPP
#include <numerix/ball_complex.hpp>
#include <algebramix/matrix_complex.hpp>

namespace mmx {

/******************************************************************************
* Accelerated bound computation
******************************************************************************/

#define TMPL template<typename C>

TMPL C
l2_norm_stride (const C* c, nat stride, nat n) {
  C sum= 0;
  for (; n != 0; c += stride, n--)
    sum += square (*c);
  return sqrt (sum);
}

template<typename V>
struct matrix_bound_multiplication: public V {

TMPL static void
mul (C* dest, const C* s1, const C* s2,
     nat r, nat l, nat c)
{
  nat n= 16;
  nat k= (l + n - 1) / n;
  nat q= l % n;

  nat n1= r*k, spc1= ((n1+1)>>1)<<1; // force 128 bit alignment if necessary
  nat n2= k*c, spc2= ((n2+1)>>1)<<1;
  nat spc= spc1 + spc2;
  C* t1= mmx_new<C> (spc);
  C* t2= t1 + spc1;

  for (nat i=0; i<r; i++)
    for (nat j=0; j<k; j++)
      t1[V::index (i, j, r, k)]=
	l2_norm_stride (s1 + V::index (i, j*n, r, l),
			V::index (0, 1, r, l),
			j == (k-1)? q: n);

  for (nat i=0; i<k; i++)
    for (nat j=0; j<c; j++)
      t2[V::index (i, j, k, c)]=
	l2_norm_stride (s2 + V::index (i*n, j, l, c),
			V::index (1, 0, l, c),
			i == (k-1)? q: n);

  //V::Vec::print (cout, s1, r*l); cout << "\n";
  //V::Vec::print (cout, t1, r*k); cout << "\n";
  //V::Vec::print (cout, s2, l*c); cout << "\n";
  //V::Vec::print (cout, t2, k*c); cout << "\n";

  V::mul (dest, t1, t2, r, k, c);

  mmx_delete<C> (t1, spc);
}

}; // matrix_bound_multiplication<V>

#undef TMPL

/******************************************************************************
* Ball multiplication
******************************************************************************/

#define TMPL template<typename Ball>
#define C Center_type(Ball)
#define R Radius_type(Ball)

template<typename V>
struct matrix_ball_multiplication: public V {

TMPL static void
mul (Ball* dest, const Ball* s1, const Ball* s2,
     nat r, nat l, nat c)
{
  nat n1= r*l, spc1= ((n1+1)>>1)<<1; // force 128 bit alignment if necessary
  nat n2= l*c, spc2= ((n2+1)>>1)<<1;
  nat nd= r*c, spcd= ((nd+1)>>1)<<1;
  nat spcc  = spc1 + spc2 + spcd;
  C* center1= mmx_new<C> (spcc);
  C* center2= center1 + spc1;
  C* centerd= center2 + spc2;
  nat spcr  = 3 * spc1 + 3 * spc2 + 3 * spcd;
  R* radius1= mmx_new<R> (spcr);
  R* radius2= radius1 + spc1;
  R* radiusd= radius2 + spc2;
  R* abs1   = radiusd + spcd;
  R* abs2   = abs1 + spc1;
  R* rad1   = abs2 + spc2;
  R* rad2   = rad1 + spc1;
  R* prod1  = rad2 + spc2;
  R* prod2  = prod1 + spcd;

  V::Vec::template vec_unary<center_op> (center1, s1, n1);
  V::Vec::template vec_unary<center_op> (center2, s2, n2);
  V::Vec::template vec_unary<radius_op> (radius1, s1, n1);
  V::Vec::template vec_unary<radius_op> (radius2, s2, n2);
  V::mul (centerd, center1, center2, r, l, c);
  {
    typedef matrix_bound_multiplication<typename V::Positive> W;
    with_rounding<R,MMX_ROUND_UP> local_rnd;
    W::Vec::template vec_unary<abs_as_op> (abs1, center1, n1);
    W::Vec::template vec_unary<abs_as_op> (abs2, center2, n2);
    W::Vec::template vec_unary<rnd_error_op> (rad1, center1, n1);
    W::Vec::template vec_unary<rnd_error_op> (rad2, center2, n2);
    W::Vec::mul (rad1, R(l), n1);
    W::Vec::mul (rad2, R(l), n2);
    W::Vec::add (rad1, radius1, n1);
    W::Vec::add (rad2, radius2, n2);
    W::Vec::add (abs1, radius1, n1);
    W::mul (prod1, abs1, rad2, r, l, c);
    W::mul (prod2, rad1, abs2, r, l, c);
    W::Vec::add (radiusd, prod1, prod2, nd);
  }
  V::Vec::template vec_binary<ball_op> (dest, centerd, radiusd, nd);

  mmx_delete<C> (center1, spcc);
  mmx_delete<R> (radius1, spcr);
}

}; // matrix_ball_multiplication<V>

#undef TMPL
#undef C
#undef R

/******************************************************************************
* Use ball multiplication by default
******************************************************************************/

template<typename V>
struct matrix_ball: 
  public matrix_naive_application<matrix_ball_multiplication<V> >
{
  static const nat def_rows = 0;
  static const nat def_cols = 0;
  static const nat init_rows= 1;
  static const nat init_cols= 1;
};

template<typename BC, typename BR, typename BV>
struct matrix_variant_helper<ball<BC,BR,BV> > {
  typedef typename matrix_variant_helper<BC>::MV CMV;
  typedef matrix_ball<CMV> MV;
};

STMPL
struct matrix_variant_helper<ball<generic,generic> > {
  typedef matrix_naive MV;
};

} // namespace mmx
#endif //__MMX__MATRIX_BALL__HPP
