
/******************************************************************************
* MODULE     : Matrix_floating.hpp
* DESCRIPTION: Multi-modular multiplication of (complex) floating matrices
* COPYRIGHT  : (C) 2008 Raouf Dridi
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_MATRIX_FLOATING_HPP
#define __MMX_MATRIX_FLOATING_HPP
#include <numerix/floating.hpp>
#include <algebramix/matrix_integer.hpp>
#include <algebramix/matrix_complex.hpp>
namespace mmx {
#define TMPL template<typename FV>
#define Floating floating<FV>

/******************************************************************************
* Low level subroutines
******************************************************************************/

TMPL void
decompose (integer* dest, mp_exp_t& e,
	   const Floating* src, nat str, nat n, nat bits)
{
  // Given a (scaled) segment [src, n], with a stride str, compute 
  // a segment [dest, n], with stride str, and a minimal exponent e,
  // such that each dest[i] fits into bits bits and
  // [dest, n] 2^e is an approximation of [src, n].
  if (n == 0) { e=0; return; }
  e= exponent (src[0]);
  for (nat i=1; i<n; i++)
    e= max (e, exponent (src[i*str]));
  e -= (bits-1);
  for (nat i=0; i<n; i++)
    dest[i*str]= as_integer (src[i*str] >> e);
}

TMPL void 
recompose (Floating* dest, const integer* src,
	   nat row_str, nat rows,
	   nat col_str, nat cols,
	   const mp_exp_t* row_e, const mp_exp_t* col_e)
{ 
  for (nat i=0; i < rows; i++, dest += row_str, src += row_str, row_e++) {
    Floating*      Dest = dest;
    const integer* Src  = src;
    const mp_exp_t* Col_e= col_e;
    for (nat j=0; j < cols; j++, Dest += col_str, Src += col_str, Col_e++)
      *Dest= Floating (*Src) << ((*row_e) + (*Col_e)); 
  }
}

/******************************************************************************
* Reduction of balanced floating multiplication to integer multiplication
******************************************************************************/

template<nat thr, typename V>
struct matrix_floating_multiplication: public V {
  typedef typename V::Naive Naive;
  typedef typename V::Positive Positive;
  typedef matrix_floating_multiplication<thr,typename V::No_simd> No_simd;
  typedef matrix_floating_multiplication<thr,typename V::No_thread> No_thread;
  typedef typename V::No_scaled No_scaled;

TMPL static void
mul (Floating* dest, const Floating* s1, const Floating* s2,
     nat r, nat l, nat c)
{  
  if (false && (r <= thr || l <= thr || c <= thr))
    V::Naive::template mul<mul_op> (dest, s1, s2, r, r, l, l, c, c);
  else {
    integer* mant1= mmx_new<integer> (r*l);
    integer* mant2= mmx_new<integer> (l*c);
    integer* dmant= mmx_new<integer> (r*c);
    mp_exp_t* e1  = mmx_new<mp_exp_t> (r);
    mp_exp_t* e2  = mmx_new<mp_exp_t> (c);
    nat bits= FV::prec() + FV::disc();
    for (nat i=0; i<r; i++) 
      decompose (mant1 + V::index (i, 0, r, l), e1[i], 
		 s1 + V::index (i, 0, r, l), V::index (0, 1, r, l), l, bits);
    //cout << "e1\t="; V::Vec::print (cout, e1, r); cout << "\n"; 
    for (nat j=0; j<c; j++)
      decompose (mant2 + V::index (0, j, l, c), e2[j],
		 s2 + V::index (0, j, l, c), V::index (1, 0, l, c), l, bits);
    V::mul (dmant, mant1, mant2, r, l, c);
    //for (nat i=0; i< r*c; i++)
    //  cout << "dmant[" << i << "]=\t " << dmant[i] << "\n";
    recompose (dest, dmant, 
	       V::index (1, 0, r, c), r,
    	       V::index (0, 1, r, c), c,
	       e1, e2);
    mmx_delete<integer> (mant1, r*l);
    mmx_delete<integer> (mant2, l*c);
    mmx_delete<integer> (dmant, r*c);
  }
}

}; // matrix_floating_multiplication<V>

/******************************************************************************
* Use Kronecker multiplication by default for floating matrices
******************************************************************************/

struct matrix_floating:
   public matrix_naive_application<
            matrix_floating_multiplication<7,
              matrix_integer_multiplication<7,
                matrix_enrich<                       
		  matrix_naive_multiplication<
		    matrix_naive_row_column<
		      matrix_vectorial<
			matrix_naive_abstractions<
			  vector_naive> > > > > > > >

{
  static const nat def_rows = 0;
  static const nat def_cols = 0;
  static const nat init_rows= 1;
  static const nat init_cols= 1;
};

TMPL
struct matrix_variant_helper<Floating > {
  typedef matrix_floating MV;
};

#undef TMPL
#undef Floating
} // namespace mmx
#endif // __MMX_MATRIX_FLOATING_HPP
