
/******************************************************************************
* MODULE     : matrix_numeric.hpp
* DESCRIPTION: Finding the roots of a complex polynomial
* COPYRIGHT  : (C) 2007  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_MATRIX_NUMERIC_HPP
#define __MMX_MATRIX_NUMERIC_HPP
#include <algebramix/matrix.hpp>
#include <analyziz/polynomial_roots.hpp>
#include <algebramix/fft_naive.hpp>
namespace mmx {
#define TMPL template<typename C>
#define Real Real_type(C)
#define Radius Radius_type(C)
#define Vector vector<C>
#define Matrix matrix<C>
#define Polynomial polynomial<C>

static nat
bit_reverse (nat i, nat p) {
  nat r=0;
  while (p != 1) {
    r= (r<<1) + (i&1);
    i= i>>1;
    p= p>>1;
  }
  return r;
}

TMPL Polynomial
characteristic_polynomial (const Matrix& m) {
  ASSERT (is_non_scalar (m), "non-scalar matrix expected");
  ASSERT (is_square_matrix (m), "square matrix expected");

  /*
  Real scale (0);
  for (nat i=0; i<rows(m); i++)
    for (nat j=0; j<cols(m); j++)
      scale= max (scale, abs (m(i,j)));
  if ((scale > 2 || 2*scale < 1) && scale > 0) {
    Polynomial chi= characteristic_polynomial (m / C(scale));
    return binpow (C(scale), rows(m)) * q_difference (chi, 1/C(scale));
  }
  */

  nat n= rows (m)+1;
  nat p= next_power_of_two (n);
  C* a= mmx_new<C> (p, C(0));
  C u= exp (2 * Pi(C) * Imaginary(C) / p);
  C v= 1;
  for (nat i=0; i<p; i++, v *= u) {
    //cout << "  " << i << ", " << bit_reverse (i, p) << "; " << v
    //<< " -> " << det<C> (v - m) << "\n";
    a[bit_reverse (i, p)]= det<C> (v - m);
  }
  inverse_fft (a, p);
  return Polynomial (a, n, p);
}

TMPL Vector
eigenvalues (const Matrix& m) {
  nat i, j;
  const Vector v= roots (characteristic_polynomial (m));
  Vector w;
  for (i=0; i<N(v); i++) {
    for (j=0; j<N(w); j++)
      if (v[i] == w[j]) break;
    if (j == i) w << v[i];
  }
  return w;
}

TMPL Matrix
eigenspace (const Matrix& m, const C& ev) {
  ASSERT (is_non_scalar (m), "non-scalar matrix expected");
  ASSERT (is_square_matrix (m), "square matrix expected");
  Matrix a= m - ev;
  Matrix k= kernel (a);
  while (true) {
    a= a*a;
    Matrix l= kernel (a);
    if (rows (l) == rows (k)) return k;
    k= l;
  }
}

TMPL Matrix
eigenvectors (const Matrix& m) {
  const Vector v= eigenvalues (m);
  if (N(v) == 0) return Matrix (C(0), 0, cols (m));
  Matrix ev= kernel (m - v[0]);
  for (nat i=1; i<N(v); i++)
    ev= vertical_join (ev, kernel (m - v[i]));
  return ev;
}

TMPL Matrix
eigenspaces (const Matrix& m) {
  const Vector v= eigenvalues (m);
  Matrix es (C(0), rows(m), cols(m));
  nat row= 0;
  for (nat i=0; i<N(v); i++) {
    const Matrix e= eigenspace (m, v[i]);
    for (nat r=0; r<rows(e); r++)
      for (nat c=0; c<cols(e); c++)
	es(r+row,c)= e(r,c);
    row += rows (e);
  }
  return es;
}

#undef TMPL
#undef Real
#undef Radius
#undef Vector
#undef Matrix
#undef Polynomial
} // namespace mmx
#endif // __MMX_MATRIX_NUMERIC_HPP
