
/******************************************************************************
* MODULE     : newton_polygon.hpp
* DESCRIPTION: Newton polygons
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_NEWTON_POLYGON_HPP
#define __MMX_NEWTON_POLYGON_HPP
#include <basix/vector.hpp>
#include <basix/double.hpp>
#include <numerix/complex.hpp>
#include <algebramix/polynomial.hpp>
namespace mmx {
#define TMPL template<typename C>
#define Vector vector<C>
#define Polynomial polynomial<C>

/******************************************************************************
* Convex hulls
******************************************************************************/

TMPL void
convex_hull_abscissae (nat* dest, nat& l, const C* src, nat n) {
  // find the x-coordinates of the Newton polygon of [src, n]
  l= 0;
  if (n == 0) return;
  dest[0]= 0; l= 1;
  if (n == 1) return;
  dest[1]= 1; l= 2;
  for (nat k=2; k<n; k++) {
    while (l>1) {
      nat i= dest[l-2], j= dest[l-1];
      if (C (k - j) * (src[j] - src[i]) < C (j - i) * (src[k] - src[j])) l--;
      else break;
    }
    dest[l++]= k;
  }
}

TMPL void
convex_hull (C* src, nat n) {
  // replace [src, n] by the Newton polygon of [src, n]
  nat l, j1= 0, j2;
  nat* a= mmx_new<nat> (n);
  convex_hull_abscissae (a, l, src, n);
  for (nat i=1; i<l; i++, j1= j2) {
    j2= a[i];
    if (j2 - j1 == 1) continue;
    else {
      C d= (src[j2] - src[j1]) / C (j2 - j1);
      for (nat j= j1+1; j<j2; j++)
	src[j]= src[j-1] + d;
    }
  }
  mmx_delete<nat> (a, n);
}

TMPL vector<nat>
convex_hull_abscissae (const Vector& v) {
  nat l;
  typedef Vector_variant(nat) V;
  nat s= aligned_size <nat,V> (N(v));
  nat* r= mmx_new<nat> (s);
  convex_hull_abscissae (r, l, seg (v), N(v));
  return vector<nat> (r, l, s);
}

TMPL Vector
convex_hull (const Vector& v) {
  typedef typename Vector_variant(C) V;
  nat l= aligned_size<C,V> (N(v));
  C* r= mmx_new<C> (l);
  V::copy (r, seg (v), N(v));
  convex_hull (r, N(v));
  return Vector (r, N(v), l);
}

TMPL Vector
increments (const Vector& v) {
  typedef typename Vector_variant(C) V;
  ASSERT (N(v)>1, "dimension must be at least two");
  nat l= aligned_size<C,V> (N(v) - 1);
  C* r= mmx_new<C> (l);
  vector_naive::sub (r, seg (v) + 1, seg (v), N(v) - 1);
  return Vector (r, N(v) - 1, l);
}

TMPL void
minkowski_sum (C* dest, const C* src1, const C* src2, nat n1, nat n2) {
  // Minkowski sum of two convex segments [src1, n1] and [src2, n2]
  nat i1= 0, i2= 0;
  while (i1+i2 < n1+n2-1) {
    dest[i1+i2]= src1[i1] + src2[i2];
    if (i1+1 == n1) i2++;
    else if (i2+1 == n2) i1++;
    else if (src1[i1+1] - src1[i1] > src2[i2+1] - src2[i2]) i1++;
    else i2++;
  }
}

TMPL Vector
minkowski_sum (const Vector& v1, const Vector& v2) {
  typedef typename Vector_variant(C) V;
  nat l= aligned_size<C,V> (N(v1) + N(v2) - 1);
  C* r= mmx_new<C> (l);
  minkowski_sum (r, seg (v1), seg (v2), N(v1), N(v2));
  return Vector (r, N(v1) + N(v2) - 1, l);
}

/******************************************************************************
* Finding a single edge of the Newton polygon
******************************************************************************/

TMPL void
newton_edge (nat& i, nat& j, const C* src, nat n, nat pos) {
  // Find the edge of the Newton polygon of [src, n] at position pos.
  if (n<3) {
    if (n == 2) { i=0; j=1; }
    else i=j=0;
  }
  else {
    i=0; j=1;
    for (nat k=2; k<n; k++) {
      if (C (k - j) * (src[j] - src[i]) < C (j - i) * (src[k] - src[j])) j=k;
      else if (j<pos) { i=j; j=k; }
    }
  }
}

TMPL C
newton_scale (const C* src, nat n, nat pos) {
  // Find the scaling factor for [src, n] at position pos.
  if (n<2) return C (1);
  else {
    nat i, j;
    double* expo_src= mmx_new<double> (n);
    for (nat k=0; k<n; k++)
      expo_src[k]= double_exponent (src[k]);
    newton_edge<double> (i, j, expo_src, n, pos);
    mmx_delete<double> (expo_src, n);
    if (src[i] != 0 && src[j] != 0)
      return exp (log (abs (src[j]) / abs (src[i])) / C (j-i));
    return C (1);
  }
}

TMPL inline C
newton_scale (const C* src, nat n) {
  return newton_scale (src, n, n>>1);
}

/******************************************************************************
* Exponents and Newton polygons
******************************************************************************/

TMPL double
magnitude (const complex<C>& z) {
  return magnitude (abs (z));
}

TMPL void
magnitudes (double* dest, const C* src, nat n) {
  for (nat i=0; i<n; i++)
    dest[i]= magnitude (src[i]);
}

TMPL vector<double>
magnitudes (const Vector& v) {
  typedef typename Vector_variant(double) V;
  nat l= aligned_size<double,V> (N(v));
  double* r= mmx_new<double> (l);
  magnitudes (r, seg (v), N(v));
  return vector<double> (r, N(v), l);
}

TMPL vector<double>
magnitudes (const Polynomial& p) {
  typedef typename Vector_variant(double) V;
  nat l= aligned_size<double,V> (N(p));
  double* r= mmx_new<double> (l);
  magnitudes (r, seg (p), N(p));
  return vector<double> (r, N(p), l);
}

TMPL double
double_exponent (const complex<C>& z) {
  return max (double_exponent (Re (z)), double_exponent (Im (z)));
}

TMPL void
exponents (double* dest, const C* src, nat n) {
  for (nat i=0; i<n; i++)
    dest[i]= double_exponent (src[i]);
}

TMPL vector<double>
exponents (const Vector& v) {
  typedef typename Vector_variant(double) V;
  nat l= aligned_size<double,V> (N(v));
  double* r= mmx_new<double> (l);
  exponents (r, seg (v), N(v));
  return vector<double> (r, N(v), l);
}

TMPL vector<double>
exponents (const Polynomial& p) {
  typedef typename Vector_variant(double) V;
  nat l= aligned_size<double,V> (N(p));
  double* r= mmx_new<double> (l);
  exponents (r, seg (p), N(p));
  return vector<double> (r, N(p), l);
}

TMPL vector<nat>
newton_abscissae (const Polynomial& p) {
  return convex_hull_abscissae (exponents (p));
}

TMPL vector<double>
newton_exponents (const Polynomial& p) {
  return convex_hull (exponents (p));
}

#undef TMPL
#undef Vector
#undef Polynomial
} // namespace mmx
#endif // __MMX_NEWTON_POLYGON_HPP
