
/******************************************************************************
* MODULE     : polynomial_ball.hpp
* DESCRIPTION: Polynomial multiplication by preconditioning based on scaling
* COPYRIGHT  : (C) 2007  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_POLYNOMIAL_BALL_HPP
#define __MMX_POLYNOMIAL_BALL_HPP
#include <numerix/ball_complex.hpp>
#include <algebramix/polynomial_complex.hpp>
#include <analyziz/newton_polygon.hpp>

namespace mmx {

/******************************************************************************
* Accelerated bound computation
******************************************************************************/

#define TMPL template<typename C>

template<typename V>
struct polynomial_newton_bound_multiplication: public V {

TMPL static void
mul (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
  //cout << "s1= "; V::Vec::print (cout, s1, n1); cout << "\n";
  //cout << "s2= "; V::Vec::print (cout, s2, n2); cout << "\n";
  nat nd = n1 + n2 - 1;
  nat spc= n1 + n2 + nd;
  double* m1= mmx_new<double> (spc);
  double* m2= m1 + n1;
  double* md= m2 + n2;
  magnitudes (m1, s1, n1);
  magnitudes (m2, s2, n2);
  convex_hull (m1, n1);
  convex_hull (m2, n2);
  minkowski_sum (md, m1, m2, n1, n2);
  C m= min (n1, n2);
  for (nat i=0; i<nd; i++)
    dest[i]= m * exp2 (C (md[i]));
  mmx_delete<double> (m1, spc);
  //cout << "dd= "; V::Vec::print (cout, dest, n1 + n2 - 1); cout << "\n";
}

}; // polynomial_newton_bound_multiplication<V>

#undef TMPL

/******************************************************************************
* Multiplication of balls
******************************************************************************/

#define TMPL template<typename Ball>
#define C Center_type(Ball)
#define R Radius_type(Ball)

template<typename V, typename W= polynomial_newton_bound_multiplication<V> >
struct polynomial_ball_multiplication: public V {
  typedef typename V::Naive Naive;
  typedef typename V::Positive Positive;
  typedef polynomial_ball_multiplication<typename V::No_simd> No_simd;
  typedef polynomial_ball_multiplication<typename V::No_thread> No_thread;
  typedef polynomial_ball_multiplication<typename V::No_scaled> No_scaled;

TMPL static void
mul (Ball* dest, const Ball* s1, const Ball* s2, nat n1, nat n2) {
  nat nd  = n1 + n2 - 1;
  nat spc1= ((n1+1)>>1)<<1; // force 128 bit alignment if necessary
  nat spc2= ((n2+1)>>1)<<1;
  nat spcd= ((nd+1)>>1)<<1;
  nat spcc  = spc1 + spc2 + spcd;
  C* center1= mmx_new<C> (spcc);
  C* center2= center1 + spc1;
  C* centerd= center2 + spc2;
  nat spcr  = 3 * spc1 + 3 * spc2 + 3 * spcd;
  R* radius1= mmx_new<R> (spcr);
  R* radius2= radius1 + spc1;
  R* radiusd= radius2 + spc2;
  R* abs1   = radiusd + spcd;
  R* abs2   = abs1 + spc1;
  R* rad1   = abs2 + spc2;
  R* rad2   = rad1 + spc1;
  R* prod1  = rad2 + spc2;
  R* prod2  = prod1 + spcd;

  V::Vec::template vec_unary<center_op> (center1, s1, n1);
  V::Vec::template vec_unary<center_op> (center2, s2, n2);
  V::Vec::template vec_unary<radius_op> (radius1, s1, n1);
  V::Vec::template vec_unary<radius_op> (radius2, s2, n2);
  V::mul (centerd, center1, center2, n1, n2);
  {
    with_rounding<R,MMX_ROUND_UP> local_rnd;
    W::Vec::template vec_unary<abs_as_op> (abs1, center1, n1);
    W::Vec::template vec_unary<abs_as_op> (abs2, center2, n2);
    W::Vec::template vec_unary<rnd_error_op> (rad1, center1, n1);
    W::Vec::template vec_unary<rnd_error_op> (rad2, center2, n2);
    W::Vec::mul (rad1, R(n1), n1);
    W::Vec::mul (rad2, R(n2), n2);
    W::Vec::add (rad1, radius1, n1);
    W::Vec::add (rad2, radius2, n2);
    W::Vec::add (abs1, radius1, n1);
    W::mul (prod1, abs1, rad2, n1, n2);
    W::mul (prod2, rad1, abs2, n1, n2);
    W::Vec::add (radiusd, prod1, prod2, nd);
  }
  V::Vec::template vec_binary<ball_op> (dest, centerd, radiusd, nd);

  mmx_delete<C> (center1, spcc);
  mmx_delete<R> (radius1, spcr);
}

TMPL static inline void
square (Ball* dest, const Ball* s1, nat n1) {
  mul (dest, s1, s1, n1, n1);
}

TMPL static inline void
tmul (Ball* dest, const Ball* s1, const Ball* s2, nat n1, nat n2) {
  (void) dest; (void) s1; (void) s2; (void) n1; (void) n2;
  ERROR ("transposed ball multiplication not implemented");
}

}; // polynomial_ball_multiplication<V>

template<typename V>
struct polynomial_multiply_ball:
  public polynomial_ball_multiplication<V>
{};

#undef TMPL
#undef C
#undef R

/******************************************************************************
* Take into account extra uniform multiplication error
******************************************************************************/

#define TMPL template<typename Ball>
#define C Center_type(Ball)
#define R Radius_type(Ball)

TMPL R
normsup (const Ball* src, nat n) {
  with_rounding<R,MMX_ROUND_UP> local_rnd;
  R r= 0;
  for (nat i=0; i<n; i++)
    r= max (r, as<R> (abs (center (src[i]))));
  return r;
}

template<typename W>
struct adjust_error_multiplication: public W::No_scaled {
  typedef typename W::No_scaled V;
  typedef typename W::Naive Naive;
  typedef typename W::Positive Positive;
  typedef adjust_error_multiplication<typename W::No_simd> No_simd;
  typedef adjust_error_multiplication<typename W::No_thread> No_thread;
  typedef V No_scaled;

TMPL static void
mul (Ball* dest, const Ball* s1, const Ball* s2, nat n1, nat n2) {
  V::mul (dest, s1, s2, n1, n2);
  int p= (int) (mmx_bit_precision + mmx_discrepancy - 1);
  R ns1= normsup (s1, n1);
  R ns2= normsup (s2, n2);
  R err= decexp2 (ns1 * ns2 * R (n1 + n2 - 1), p);
  for (nat i=0; i<n1+n2-1; i++)
    dest[i] += Ball (0, err);
}

TMPL static inline void
square (Ball* dest, const Ball* s1, nat n1) {
  mul (dest, s1, s1, n1, n1);
}

TMPL static inline void
tmul (Ball* dest, const Ball* s1, const Ball* s2, nat n1, nat n2) {
  (void) dest; (void) s1; (void) s2; (void) n1; (void) n2;
  ERROR ("transposed adjusted multiplication not implemented");
}

}; // polynomial_ball_multiplication<V>

template<typename V>
struct polynomial_adjusted_multiply_ball:
  public adjust_error_multiplication<polynomial_ball_multiplication<V> >
{};

#undef TMPL
#undef C
#undef R

/******************************************************************************
* Use ball multiplication by default
******************************************************************************/

template<typename C, typename R, typename V>
struct polynomial_variant_helper<ball<C,R,V> > {
  typedef ball<C,R,V> B;
  typedef typename Vector_variant(C) CVV;
  typedef typename Polynomial_multiply(C) CMV;
  typedef typename Polynomial_variant(C) CPV;

  typedef typename Vector_variant(B) VV;
  typedef polynomial_adjusted_multiply_ball<CMV> MV;
  typedef polynomial_algos_dicho<MV> PV;
};

STMPL
struct polynomial_variant_helper<ball<generic> > {
  typedef vector_naive VV;
  typedef polynomial_multiply_naive<VV> MV;
  typedef polynomial_naive PV;
};

}
#endif // __MMX_POLYNOMIAL_BALL_HPP
