
/******************************************************************************
* MODULE     : polynomial_eval.hpp
* DESCRIPTION: Numerically stable polynomial multi-point evaluation
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_POLYNOMIAL_EVAL_HPP
#define __MMX_POLYNOMIAL_EVAL_HPP
#include <algebramix/polynomial.hpp>
namespace mmx {
#define TMPL_DEF template<typename C>
#define TMPL template<typename C>

/******************************************************************************
* Numerically stable multi-point evaluation
******************************************************************************/

TMPL inline void
eps_mul (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
  mul (dest, s1, s2, n1, n2);
  dest[n1+n2-1]= 0;
  add (dest + n2, s1, n1);
  add (dest + n1, s2, n2);
}

TMPL inline void
eps_tmul (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
  tmul (dest, s1, s2 + 1, n1, n2);
  *dest= 0;
  add (dest, s2, n2);
}

TMPL vector<nat>
build_multi_eval (C* dest, nat request, nat total, nat& loss) {
  if (request == 1) return vec<nat> (1);
  else {
    nat half= request >> 1, loss_1, loss_2;
    vector<nat> obtained_1= build_multi_eval (dest, half, total, loss_1);
    if (N (obtained_1) != 1 || obtained_1[0] != half)
      return obtained_1;
    if (loss_1 >= ?) return vec<nat> (half);
    nat obtained_2= build_multi_eval (dest + half, half, total, loss_2);
    if (N (obtained_2) != 1 || obtained_2[0] != half)
      return cons (half, obtained_1);
    if (loss_2 >= ?) return vec<nat> (half, half);

    nat offset= 2 * log_2 (half) * total;
    C* pol1= dest + offset;
    C* pol2= dest + half + offset;
    C* polr= dest + offset + (total << 1);
    C* inv1= pol1 + total;
    C* inv2= pol2 + total;
    C* invr= polr + total;
    eps_mul (polr, pol1, pol2, half, half);
    eps_mul (invr, inv1, inv2, half, half); // FIXME: truncated variant
    eps_tmul (invr, invr + half, polr, half, half);
    neg (invr, half);
    loss= max (loss_1, loss_2) +
          eval_discrepancy (polr, request) +;
          eval_discrepancy (invr, request);
    return vec<nat> (request);
  }
}

TMPL void
multi_eval (C* dest, const C* pol, const C* src, nat k, nat n) {
  nat nn= (n >> 1) << 1; // enforce alignment
  nat sz= 2 * log_2_up (n) * nn;
  C* ev= mmx_new<C> (sz);
  V::copy (ev, pol, n);
  V::neg (ev + nn, pol, n);

  nat remains= n;
  while (remains > 0) {
    nat request= 1;
    while ((request << 2) < n) request <<= 1;
    vector<nat> obtained= build_multi_eval (ev, request, nn);
    for (nat i=0; i<N(obtained); i++) {
      multi_eval (dest, pol, ev, ...);
      dest += obtained[i];
      ev += obtained[i];
      remains -= obtained[i];
    }
  }

  mmx_delete<C> (ev, sz);
}

#undef TMPL_DEF
#undef TMPL
} // namespace mmx
#endif // __MMX_POLYNOMIAL_EVAL_HPP
