
/******************************************************************************
* MODULE     : polynomial_floating.hpp
* DESCRIPTION: Kronecker multiplication of (complex) floating polynomials
* COPYRIGHT  : (C) 2007  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_POLYNOMIAL_FLOATING_HPP
#define __MMX_POLYNOMIAL_FLOATING_HPP
#include <numerix/floating.hpp>
#include <algebramix/polynomial_integer.hpp>
#include <algebramix/polynomial_complex.hpp>
#include <analyziz/polynomial_scaled.hpp>
namespace mmx {
#define TMPL template<typename FV>
#define Floating floating<FV>

/******************************************************************************
* Reduction of balanced floating multiplication to integer multiplication
******************************************************************************/

TMPL void
decompose (integer* dest, mp_exp_t& e, const Floating* src, nat n, nat bits) {
  // Given a (scaled) segment [src, n], compute a segment [dest, i] and
  // a minimal exponent e, such that each dest[i] fits into bits bits and
  // [dest, n] 2^e is an approximation of [src, n].
  if (n == 0) { e=0; return; }
  e= exponent (src[0]);
  for (nat i=1; i<n; i++)
    e= max (e, exponent (src[i]));
  e -= (bits-1);
  for (nat i=0; i<n; i++)
    dest[i]= as_integer (src[i] >> e);
}

TMPL void
recompose (Floating* dest, const integer* src, nat n, mp_exp_t e) {
  // Set [dest, n] := [src, n] 2^e
  for (nat i=0; i<n; i++)
    dest[i]= Floating (src[i]) << e;
}

template<typename V, typename Th= dicho_threshold>
struct polynomial_floating_multiplication: public V {

TMPL static void
mul (Floating* dest, const Floating* s1, const Floating* s2, nat n1, nat n2) {
  // TODO: exceptions (NaN, etc.)
  if (min (n1, n2) < Threshold(Floating,Th))
    V::Naive::mul (dest, s1, s2, n1, n2);
  else {
    //cout << "--- multiply ---\n";
    //for (nat i=0; i<n1; i++)
    //cout << "  " << i << "\t -> " << s1[i] << "\n";
    //for (nat i=0; i<n2; i++)
    //cout << "  " << i << "\t -> " << s2[i] << "\n";
    integer* mant1= mmx_new<integer> (n1);
    integer* mant2= mmx_new<integer> (n2);
    integer* dmant= mmx_new<integer> (n1+n2-1);
    mp_exp_t e1, e2;
    nat bits= FV::prec() + FV::disc();
    decompose (mant1, e1, s1, n1, bits);
    decompose (mant2, e2, s2, n2, bits);
    //for (nat i=0; i<n1; i++)
    //cout << "  " << i << "\t -> " << mant1[i] << "\n";
    V::mul (dmant, mant1, mant2, n1, n2);
    //for (nat i=0; i<n1+n2-1; i++)
    //cout << "  " << i << "\t -> " << dmant[i] << "\n";
    recompose (dest, dmant, n1+n2-1, e1+e2);
    mmx_delete<integer> (mant1, n1);
    mmx_delete<integer> (mant2, n2);
    mmx_delete<integer> (dmant, n1+n2-1);
    //for (nat i=0; i<n1+n2-1; i++)
    //cout << "  " << i << "\t -> " << dest[i] << "\n";
  }
}

TMPL static void
square (Floating* dest, const Floating* s1, nat n1) {
  // TODO: exceptions (NaN, etc.)
  if (n1 < Threshold(Floating,Th))
    V::Naive::square (dest, s1, n1);
  else {
    integer* mant1= mmx_new<integer> (n1);
    integer* dmant= mmx_new<integer> (n1+n1-1);
    mp_exp_t e1;
    nat bits= FV::prec() + FV::disc();
    decompose (mant1, e1, s1, n1, bits);
    V::square (dmant, mant1, n1);
    recompose (dest, dmant, n1+n1-1, e1+e1);
    mmx_delete<integer> (mant1, n1);
    mmx_delete<integer> (dmant, n1+n1-1);
  }
}

TMPL static inline void
tmul (Floating* dest, const Floating* s1, const Floating* s2, nat n1, nat n2) {
  (void) dest; (void) s1; (void) s2; (void) n1; (void) n2;
  ERROR ("transposed floating multiplication not implemented");
}

}; // polynomial_floating_multiplication<V>

template<typename V>
struct polynomial_multiply_floating:
  public polynomial_floating_multiplication<V>
{};

/******************************************************************************
* Use Kronecker multiplication by default for floating polynomials
******************************************************************************/

struct polynomial_floating_kronecker:
  public polynomial_algos_dicho<
           polynomial_multiply_scaled<
           polynomial_multiply_floating<
           polynomial_multiply_integer<
           polynomial_multiply_naive<
             vector_naive> > > > >
{};

TMPL
struct polynomial_variant_helper<Floating > {
  typedef typename Vector_variant(Floating) VV;
  typedef polynomial_multiply_naive<VV> MN;
  typedef polynomial_multiply_integer<MN> MI;
  typedef polynomial_multiply_floating<MI> MF;
  typedef polynomial_multiply_scaled<MF> MV;
  typedef polynomial_floating_kronecker PV;
  // typedef polynomial_algos_dicho<MV> PV;
  // Commented out so as to keep typenames simple when debugging
};

#undef TMPL
#undef Floating
}
#endif // __MMX_POLYNOMIAL_FLOATING_HPP
