
/******************************************************************************
* MODULE     : polynomial_scaled.hpp
* DESCRIPTION: Polynomial multiplication by preconditioning based on scaling
* COPYRIGHT  : (C) 2007  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_POLYNOMIAL_SCALED_HPP
#define __MMX_POLYNOMIAL_SCALED_HPP
#include <algebramix/polynomial.hpp>
#include <analyziz/newton_polygon.hpp>
namespace mmx {
#define TMPL template<typename C>

/******************************************************************************
* Scale exponents using a single substitution x :-> lambda x
******************************************************************************/

TMPL void
scale (C* dest, const C* src, nat n, const C& c, const C& a) {
  // Set dest[i] = c src[i] a^i for all i
  if (n < 16) {
    if (n>0) dest[0]= c * src[0];
    C b= c;
    for (nat i=1; i<n; i++) {
      b *= a;
      dest[i]= b * src[i];
    }
  }
  else {
    nat i=1;
    C p= a;
    while ((i<<1) < n) {
      p *= p;
      i <<= 1;
    }
    scale (dest, src, i, c, a);
    scale (dest+i, src+i, n-i, p*c, a);
  }
}

/******************************************************************************
* Scale coefficients using a single substitution x :-> lambda x
******************************************************************************/

template<typename V, typename Th= dicho_threshold>
struct polynomial_scaled_multiplication: public V {
  typedef typename V::Naive Naive;
  typedef typename V::Positive Positive;
  typedef polynomial_scaled_multiplication<typename V::No_simd> No_simd;
  typedef polynomial_scaled_multiplication<typename V::No_thread> No_thread;
  typedef typename V::No_scaled No_scaled;

TMPL static void
mul (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
  // TODO: exceptions (NaN, etc.)
  if (min (n1, n2) < Threshold(C,Th))
    V::Naive::mul (dest, s1, s2, n1, n2);
  else {
    C* Src1= mmx_new<C> (n1);
    C* Src2= mmx_new<C> (n2);
    C* Dest= mmx_new<C> (n1+n2-1);
    C sc1= newton_scale (s1, n1);
    C sc2= newton_scale (s2, n2);
    C sc = max (sc1, sc2);
    //for (nat i=0; i<n1; i++)
    //cout << "  " << i << "\t -> " << s1[i] << "\n";
    scale (Src1, s1, n1, C (1), 1/sc);
    scale (Src2, s2, n2, C (1), 1/sc);
    //for (nat i=0; i<n1; i++)
    //cout << "  " << i << "\t -> " << Src1[i] << "\n";
    V::mul (Dest, Src1, Src2, n1, n2);
    //for (nat i=0; i<n1+n2-1; i++)
    //cout << "  " << i << "\t -> " << Dest[i] << "\n";
    scale (dest, Dest, n1+n2-1, C (1), sc);
    //for (nat i=0; i<n1+n2-1; i++)
    //cout << "  " << i << "\t -> " << dest[i] << "\n";
    mmx_delete<C> (Src1, n1);
    mmx_delete<C> (Src2, n2);
    mmx_delete<C> (Dest, n1+n2-1);
  }
}

TMPL static void
square (C* dest, const C* s1, nat n1) {
  // TODO: exceptions (NaN, etc.)
  if (n1 < Threshold(C,Th))
    V::Naive::square (dest, s1, n1);
  else {
    C* Src1= mmx_new<C> (n1);
    C* Dest= mmx_new<C> (n1+n1-1);
    C  sc  = newton_scale (s1, n1);
    scale (Src1, s1, n1, C (1), 1/sc);
    V::square (Dest, Src1, n1);
    scale (dest, Dest, n1+n1-1, C (1), sc);
    mmx_delete<C> (Src1, n1);
    mmx_delete<C> (Dest, n1+n1-1);
  }
}

TMPL static inline void
tmul (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
  (void) dest; (void) s1; (void) s2; (void) n1; (void) n2;
  ERROR ("transposed scaled multiplication not implemented");
}

}; // polynomial_scaled_multiplication<V>

template<typename V>
struct polynomial_multiply_scaled:
  public polynomial_scaled_multiplication<V>
{};

#undef TMPL
}
#endif // __MMX_POLYNOMIAL_SCALED_HPP
