
/******************************************************************************
* MODULE     : riemann_square.hpp
* DESCRIPTION: square building blocks for digital Riemann surfaces
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_RIEMANN_SQUARE_HPP
#define __MMX_RIEMANN_SQUARE_HPP
#include <basix/table.hpp>
#include <numerix/complex_double.hpp>

namespace mmx {
class std_riemann {};
#define TMPL_DEF template<typename R,typename V= std_riemann>
#define TMPL template<typename R,typename V>
#define C complex<R>
#define Square riemann_square<R,V>
#define Square_rep riemann_square_rep<R,V>
#define Squares table<bool,Square >

/******************************************************************************
* Square neighbourhoods on Riemann surfaces
******************************************************************************/

TMPL_DEF
class riemann_square_rep REP_STRUCT {
private:
  static double next_serial;
public:
  C c;
  R r;
  double serial;
public:
  inline riemann_square_rep (const C& c2, const R& r2):
    c (c2), r (r2), serial (next_serial++) {}
};

TMPL_DEF
class riemann_square {
INDIRECT_PROTO_2 (riemann_square, riemann_square_rep, R, V)
public:
  inline riemann_square ():
    rep (new Square_rep (C (0), R (0))) {}
  inline riemann_square (const C& c, const R& r):
    rep (new Square_rep (c, r)) {}
};
INDIRECT_IMPL_2 (riemann_square, riemann_square_rep,
		 typename R, R, typename V, V)

TMPL double Square_rep::next_serial= 0;
TMPL inline C center (const Square& s) { return s->c; }
TMPL inline R radius (const Square& s) { return s->r; }
TMPL inline R diameter (const Square& s) { return 2 * s->r; }

TMPL vector<C >
corners (const Square& s) {
  R r= radius (s);
  return center (s) + vec<C > (C (r, r), C (r, -r), C (-r, r), C (-r, -r));
}

/******************************************************************************
* Required functions
******************************************************************************/

template<typename Op, typename R, typename V> nat
unary_hash (const Square& s) {
  return hard_hash (s);
}

template<typename Op, typename R, typename V> bool
binary_test (const Square& s1, const Square& s2) {
  return hard_eq (s1, s2);
}

TMPL syntactic
flatten (const Square& z) {
  return apply ("square", flatten (center (z)), flatten (radius (z)));
}

TRUE_IDENTITY_OP_SUGAR(TMPL,Square)
SOFT_IDENTITY_OP_SUGAR(TMPL,Square)
PRINT_SUGAR(TMPL,Square)

/******************************************************************************
* Predicates
******************************************************************************/

TMPL bool
admissible (const Square& s) {
  R r= radius (s);
  C z= center (s) / r;
  R x= Re(z) - 2 * floor (Re (z) / 2);
  R y= Im(z) - 2 * floor (Im (z) / 2);
  return x == 1 && y == 1;
}

TMPL bool
intersect (const Square& s1, const Square& s2) {
  R d= radius (s1) + radius (s2);
  return abs (Re (center (s1)) - Re (center (s2))) <= d &&
         abs (Im (center (s1)) - Im (center (s2))) <= d;
}

TMPL bool
included (const Square& s1, const Square& s2) {
  R d= radius (s2) - radius (s1);
  return radius (s1) <= radius (s2) &&
         abs (Re (center (s1)) - Re (center (s2))) <= d &&
         abs (Im (center (s1)) - Im (center (s2))) <= d;
}

TMPL bool
adjacent (const Square& s1, const Square& s2) {
  R d= radius (s1) + radius (s2);
  return
    (abs (Re (center (s1)) - Re (center (s2))) == d &&
     abs (Im (center (s1)) - Im (center (s2))) < d) ||
    (abs (Re (center (s1)) - Re (center (s2))) < d &&
     abs (Im (center (s1)) - Im (center (s2))) == d);
}

TMPL bool
inside (const C& z, const Square& s) {
  return abs (Re (z) - Re (center (s))) <= radius (s) &&
         abs (Im (z) - Im (center (s))) <= radius (s);
}

/******************************************************************************
* Approximation of circular disks
******************************************************************************/

TMPL void
digital_disk_sub (Squares& ss,
		  const C& u, const R& d,
		  const C& z, const R& r, nat prec)
{
  if (R (prec) * d >= r) {
    R h= d / 2;
    Square s (u + C (h, h), h);
    ASSERT (admissible (s), "admissibility hypothesis violated");
    vector<C > v= corners (s);
    nat count= 0;
    for (nat i=0; i<N(v); i++)
      if (abs (v[i] - z) < r) count++;
    if (count == 0 && !intersect (s, Square (z, r)));
    else if (count == 4) ss[s]= true;
    else {
      digital_disk_sub<R,V> (ss, u, h, z, r, prec);
      digital_disk_sub<R,V> (ss, u + C (h, 0), h, z, r, prec);
      digital_disk_sub<R,V> (ss, u + C (0, h), h, z, r, prec);
      digital_disk_sub<R,V> (ss, u + C (h, h), h, z, r, prec);
    }
  }
}

TMPL Squares
digital_disk (const C& z, const R& r, nat prec= 8) {
  //cout << "digital disk " << z << ", " << r << "\n";
  Squares ss;
  R scale= incexp2 (R (1), exponent (r) + 2);
  R x= floor ((Re (z) - r) / scale) * scale;
  R y= floor ((Im (z) - r) / scale) * scale;
  digital_disk_sub<R,V> (ss, C (x, y), scale, z, r, prec);
  digital_disk_sub<R,V> (ss, C (x + scale, y), scale, z, r, prec);
  digital_disk_sub<R,V> (ss, C (x, y + scale), scale, z, r, prec);
  digital_disk_sub<R,V> (ss, C (x + scale, y + scale), scale, z, r, prec);
  return ss;
}

#undef TMPL_DEF
#undef TMPL
#undef C
#undef Square
#undef Square_rep
#undef Squares
} // namespace mmx
#endif // __MMX_RIEMANN_SQUARE_HPP
