
/******************************************************************************
* MODULE     : riemann_surface.hpp
* DESCRIPTION: digital Riemann surfaces
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_RIEMANN_SURFACE_HPP
#define __MMX_RIEMANN_SURFACE_HPP
#include <analyziz/riemann_point.hpp>

namespace mmx {
#define TMPL_DEF template<typename R,typename V= std_riemann>
#define TMPL template<typename R,typename V>
#define C complex<R>
#define Square riemann_square<R,V>
#define Point riemann_point<R,V>
#define Squares table<bool,Square >
#define Neighbours table<Squares,Square >
#define Upgrades table<Square,Square >
#define Iterator iterator<Square >
#define Surface riemann_surface<R,V>
#define Surface_rep riemann_surface_rep<R,V>

/******************************************************************************
* Digital Riemann surfaces
******************************************************************************/

TMPL_DEF
class riemann_surface_rep REP_STRUCT {
public:
  Squares    squares;
  Neighbours neighbours;
  Squares    invalid;
  Squares    replaced;
  Upgrades   upgrades;
public:
  inline riemann_surface_rep ():
    squares (false),
    neighbours (Squares (false)),
    invalid (false),
    replaced (false) {}
  bool add (const Square& s);
  bool add (const Squares& ss);
  bool connect (const Square& s1, const Square& s2);
  bool connect (const Neighbours& t);
  bool plane_connect (const Squares& ss);
  bool glue (const Square& s1, const Square& s2);
  Squares owners (const Square& s, const C& p) const;
  void normalize ();
  void clean ();
  Square upgrade (const Square& s);
  Point follow (const Square& orig, const C& from, const C& to) const;
  Point move (const Point& p, const C& delta) const;
};

TMPL_DEF
class riemann_surface {
INDIRECT_PROTO_2 (riemann_surface, riemann_surface_rep, R, V)
public:
  inline riemann_surface (): rep (new Surface_rep ()) {}
};
INDIRECT_IMPL_2 (riemann_surface, riemann_surface_rep,
		 typename R, R, typename V, V)

/******************************************************************************
* Required routines
******************************************************************************/

template<typename Op, typename R, typename V> nat
unary_hash (const Surface& rs) {
  return hard_hash (rs);
}

template<typename Op, typename R, typename V> bool
binary_test (const Surface& rs1, const Surface& rs2) {
  return hard_eq (rs1, rs2);
}

TMPL syntactic
flatten (const Surface& rs) {
  //return apply ("surface", flatten (rs->squares));
  return apply ("surface", flatten (rs->squares), flatten (rs->neighbours));
}

TRUE_IDENTITY_OP_SUGAR(TMPL,Surface)
SOFT_IDENTITY_OP_SUGAR(TMPL,Surface)
PRINT_SUGAR(TMPL,Surface)

/******************************************************************************
* Adding squares and connecting squares
******************************************************************************/

TMPL bool
Surface_rep::add (const Square& s) {
  // Add s to the Riemann surface
  if (squares->contains (s)) return false;
  squares[s]= true;
  neighbours[s]= Squares (false);
  return true;
}

TMPL bool
Surface_rep::add (const Squares& ss) {
  // Add all squares in ss to the Riemann surface
  bool r= false;
  for (Iterator it= entries (ss); busy (it); ++it)
    r= add (*it) || r;
  return r;
}

TMPL bool
Surface_rep::connect (const Square& s1, const Square& s2) {
  // Connect s1 and s2, if not already done
  // The projections of s1 and s2 are required to be neighbours
  if (!(squares->contains (s1) && squares->contains (s2))) return false;
  if (neighbours[s1]->contains (s2) && neighbours[s2]->contains (s1))
    return false;
  //cout << "    Connect " << s1 << ", " << s2 << "\n";
  invalid[s1]= true;
  invalid[s2]= true;
  neighbours[s1][s2]= true;
  neighbours[s2][s1]= true;
  return true;
}

TMPL bool
Surface_rep::connect (const Neighbours& t) {
  // Connect all squares in the neighbour table
  bool r= false;
  for (Iterator it= entries (t); busy (it); ++it)
    for (Iterator nb= entries (t[*it]); busy (nb); ++nb)
      r= connect (*it, *nb) || r;
  return r;
}

TMPL bool
Surface_rep::plane_connect (const Squares& ss) {
  // Connect each two squares in ss whose projections are neighbours
  // cout << "Plane connect " << ss << "\n";
  bool r= false;
  for (Iterator it1= entries (ss); busy (it1); ++it1)
    for (Iterator it2= entries (ss); busy (it2); ++it2)
      if (adjacent (*it1, *it2))
	r= connect (*it1, *it2) || r;
  return r;
}

TMPL bool
Surface_rep::glue (const Square& s1, const Square& s2) {
  // Glue s1 and s2 together
  if (!(squares->contains (s1) && squares->contains (s2))) return false;
  if (s1 == s2) return false;
  if (included (s1, s2) && (!included (s2, s1) || s2->serial < s1->serial)) {
    if (!squares->contains (s1)) return false;
    //cout << "  Glue " << s1 << ", " << s2 << "\n";
    Squares ns= neighbours[s1];
    upgrades[s1]= s2;
    replaced [s1]= true;
    reset (squares, s1);
    reset (neighbours, s1);
    for (Iterator nb= entries (ns); busy (nb); ++nb)
      reset (neighbours[*nb], s1);
    for (Iterator nb= entries (ns); busy (nb); ++nb)
      if (adjacent (*nb, s2)) connect (*nb, s2);
    for (Iterator nb= entries (ns); busy (nb); ++nb)
      if (included (*nb, s2)) glue (*nb, s2);
    return true;
  }
  else {
    ASSERT (included (s2, s1), "projections do not match");
    return glue (s2, s1);
  }
}

/******************************************************************************
* Normalization
******************************************************************************/

TMPL Squares
Surface_rep::owners (const Square& s, const C& p) const {
  // Return list of squares which contain the point p (on s)
  Squares r (false);
  Squares todo (false);
  todo[s]= true;
  while (N(todo) != 0) {
    Squares temp= todo;
    todo= Squares (false);
    for (Iterator it= entries (temp); busy (it); ++it) {
      r[*it]= true;
      for (Iterator nb= entries (neighbours[*it]); busy (nb); ++nb)
	if (inside (p, *nb) && !r[*nb]) todo[*nb]= true;
    }
  }
  return r;
}

TMPL inline Squares
owners (const Surface& rs, const Point& p) {
  return rs->owners (owner (p), project (p));
}

TMPL void
Surface_rep::normalize () {
  // After making non trivial connections,
  // the Riemann surface should be normalized using this routine
  while (N(invalid) != 0) {
    Squares ss= invalid;
    invalid= Squares (false);
    for (Iterator it= entries (ss); busy (it); ++it) {
      //cout << "Normalize " << *it << "\n";
      for (Iterator nb1= entries (neighbours[*it]); busy (nb1); ++nb1)
	for (Iterator nb2= entries (neighbours[*nb1]); busy (nb2); ++nb2) {
	  if (included (*it, *nb2)) glue (*it, *nb2);
	  if (included (*nb2, *it)) glue (*nb2, *it);
	}
      vector<C > cs= corners (*it);
      for (nat i=0; i<N(cs); i++) {
	Squares os= owners (*it, cs[i]);
	if (plane_connect (os)) continue;
	Square first= *entries (os);
	R r= diameter (first);
	bool same_diameter= true;
	for (Iterator it2= entries (os); busy (it2); ++it2)
	  same_diameter= same_diameter && (r == diameter (*it2));
	Square s (cs[i], r);
	if (same_diameter && admissible (s) && N(os) == 4) {
	  //cout << "  Join " << os << " -> " << s << "\n";
	  add (s);
	  glue (s, first);
	}
      }
    }
  }
}

TMPL void
Surface_rep::clean () {
  // At the end of normalize, the variable replaced contains
  // the list of all squares which have been upgraded during gluing
  // The present routine resets this variable and performs some
  // appropriate final cleaning
  for (Iterator it= entries (replaced); busy (it); ++it)
    reset (neighbours, *it);
  replaced= Squares (false);
}

TMPL Square
Surface_rep::upgrade (const Square& s) {
  if (squares->contains (s)) return s;
  ASSERT (upgrades->contains (s), "square not on Riemann surface");
  if (!squares->contains (upgrades[s]))
    upgrades[s]= upgrade (upgrades[s]);
  return upgrades[s];
}

TMPL inline Point
upgrade (const Surface& rs, const Point& p) {
  return Point (inside (rs)->upgrade (owner (p)), project (p));
}

/******************************************************************************
* Moving on the Riemann surface
******************************************************************************/

TMPL Point
Surface_rep::follow (const Square& s1, const C& z1, const C& z2) const {
  if (inside (z2, s1)) return Point (s1, z2);
  C c    = center (s1);
  R r    = radius (s1);
  C delta= z2 - z1;
  C z3   = center (s1);
  int dx = 0;
  int dy = 0;
  if (Re (z2) > Re (c) + r && Re (delta) != 0) {
    R x= Re (c) + r;
    R y= Im (z1) + (x - Re (z1)) * Im (delta) / Re (delta);
    if (abs (y - Im (c)) <= r) { z3= C (x, y); dx= 1; }
  }
  if (Re (z2) < Re (c) - r && Re (delta) != 0) {
    R x= Re (c) - r;
    R y= Im (z1) + (x - Re (z1)) * Im (delta) / Re (delta);
    if (abs (y - Im (c)) <= r) { z3= C (x, y); dx= -1; }
  }
  if (Im (z2) > Im (c) + r && Im (delta) != 0) {
    R y= Im (c) + r;
    R x= Re (z1) + (y - Im (z1)) * Re (delta) / Im (delta);
    if (abs (x - Re (c)) <= r) { z3= C (x, y); dy= 1; }
  }
  if (Im (z2) < Im (c) - r && Im (delta) != 0) {
    R y= Im (c) - r;
    R x= Re (z1) + (y - Im (z1)) * Re (delta) / Im (delta);
    if (abs (x - Re (c)) <= r) { z3= C (x, y); dy= -1; }
  }
  ASSERT (z3 != center (s1), "unexpected situation");
  for (Iterator it= entries (neighbours[s1]); busy (it); ++it) {
    Square s3= *it;
    if (inside (z3, s3)) {
      //cout << "Follow to " << s3 << "? " << dx << ", " << dy << "\n";
      R d= radius (s1) + radius (s3);
      if ((dx > 0 && Re (center (s3) - center (s1)) == d) ||
	  (dx < 0 && Re (center (s3) - center (s1)) == -d) ||
	  (dy > 0 && Im (center (s3) - center (s1)) == d) ||
	  (dy < 0 && Im (center (s3) - center (s1)) == -d))
	return follow (s3, z3, z2);
    }
  }
  return Point (s1, z3);
}

TMPL Point
Surface_rep::move (const Point& p, const C& delta) const {
  Square s= owner (p);
  C z1= project (p);
  C z2= z1 + delta;
  return follow (s, z1, z2);
}

/******************************************************************************
* Public interface
******************************************************************************/

TMPL Surface
riemann_disk (const C& z, const R& r, nat prec= 8) {
  Surface rs;
  Squares ds= digital_disk<R,V> (z, r, prec);
  inside (rs)->add (ds);
  inside (rs)->plane_connect (ds);
  inside (rs)->normalize ();
  inside (rs)->clean ();
  return rs;
}

TMPL Squares
glue (Surface& rs, const Point& p1, const Point& p2) {
  ASSERT (project (p1) == project (p2), "incompatible glue points");
  inside (rs)->glue (owner (upgrade (rs, p1)), owner (upgrade (rs, p2)));
  inside (rs)->normalize ();
  Squares replaced= rs->replaced;
  inside (rs)->clean ();
  return replaced;
}

TMPL Squares
glue (Surface& rs1, const Point& p1, const Surface& rs2, const Point& p2) {
  ASSERT (project (p1) == project (p2), "incompatible glue points");
  //cout << "--- Add ---\n";
  inside (rs1)->add (rs2->squares);
  inside (rs1)->connect (rs2->neighbours);
  Squares os1= owners (rs1, upgrade (rs1, p1));
  Squares os2= owners (rs2, upgrade (rs2, p2));
  for (Iterator it1= entries (os1); busy (it1); ++it1)
    for (Iterator it2= entries (os2); busy (it2); ++it2)
      if (included (*it1, *it2) || included (*it2, *it1)) {
	//cout << "--- Glue ---\n";
	inside (rs1)->glue (*it1, *it2);
	//cout << "--- Normalize ---\n";
	inside (rs1)->normalize ();
	Squares replaced= rs1->replaced;
	inside (rs1)->clean ();
	//cout << "--- Done ---\n";
	return replaced;
      }
  ERROR ("unexpected situation");
  return Squares (false);
}

TMPL void
glue (Surface& rs1, const Point& p1, const Surface& rs2) {
  Point p2= lift (rs2, project (p1));
  glue (rs1, p1, rs2, p2);
}

TMPL bool
is_plane (const Surface& rs) {
  for (Iterator it1= entries (rs->squares); busy (it1); ++it1)
    for (Iterator it2= entries (rs->squares); busy (it2); ++it2)
      if (adjacent (*it1, *it2))
	if (!rs->neighbours[*it1]->contains (*it2))
	  return false;
  return true;
}

TMPL Point
lift (const Surface& rs, const C& z) {
  for (Iterator it= entries (rs->squares); busy (it); ++it)
    if (inside (z, *it)) return Point (*it, z);
  cout << "rs= " << rs << "\n";
  cout << "z= " << z << "\n";
  ERROR ("unable to lift point");
  return Point ();
}

TMPL vector<Point >
lifts (const Surface& rs, const C& z) {
  vector<Point > v;
  for (Iterator it= entries (rs->squares); busy (it); ++it)
    if (inside (z, *it)) v << Point (*it, z);
  return v;
}

TMPL inline Point
move (const Surface& rs, const Point& p, const C& delta) {
  // Return p + t delta, where t <= 1 is maximal such that
  // the segment [p, p + t delta] lies on rs
  return rs->move (upgrade (rs, p), delta);
}

#undef TMPL_DEF
#undef TMPL
#undef C
#undef Square
#undef Point
#undef Squares
#undef Neighbours
#undef Upgrades
#undef Iterator
#undef Surface
#undef Surface_rep
} // namespace mmx
#endif // __MMX_RIEMANN_HPP
