
/******************************************************************************
* MODULE     : basix.hpp
* DESCRIPTION: Basic inclusions, definitions and routines
* COPYRIGHT  : (C) 2003  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

//! @file

#ifndef __MMX_BASIX_HPP
#define __MMX_BASIX_HPP
#include <basix/fast_new.hpp>
#include <cmath>

#ifdef BASIX_INTTYPES_H
#include <inttypes>
#else
#define uintptr_t unsigned long int
#endif

#if __GLIBC__ >= 2 && defined(__THROW)
extern "C" double hypot (double, double) __THROW __attribute__ ((const));
#else
extern "C" double hypot (double, double) __attribute__ ((const));
#endif

namespace mmx {

#define LESSGTR <>
#define STMPL template<>

//! @brief Special objects that affect output when printed.
enum print_format {
  blank,    //!< Set one indentation to two spaces.
  stroke,   //!< Set one indentation to "| ".
  indent,   //!< Increment indentation.
  unindent, //!< Decrement indentation.
  lf,       //!< Carriage return and indent the new line.
  hrule,    //!< Output 80 consecutive dashes.
};

ostream& operator << (ostream& out, print_format x);

extern int    mmx_argc;
extern char** mmx_argv;
void noop ();

/******************************************************************************
* Indirect data structures prototypes
******************************************************************************/

//! @brief Prototype of the target class @a T with representation @a R
#define INDIRECT_PROTO(T,R)			\
MMX_ALLOCATORS					\
protected:					\
  R* rep;					\
public:						\
  inline T (R* rep2);				\
  inline T (const R* rep2, bool with_inc);	\
  inline T (const T& x);			\
  inline ~T ();					\
  inline T& operator= (const T& x);		\
  inline const R* operator-> () const;		\
  inline void secure ();

//! @brief Prototype of the target class @a T<X>
//! with representation @a R<X>
#define INDIRECT_PROTO_1(T,R,X)			\
MMX_ALLOCATORS					\
protected:					\
  R<X >* rep;					\
public:						\
  inline T (R<X >* rep2);			\
  inline T (const R<X >* rep2, bool with_inc);	\
  inline T (const T& x);			\
  inline ~T ();					\
  inline T& operator= (const T& x);		\
  inline const R<X >* operator-> () const;	\
  inline R<X >* operator-> ();			\
  inline void secure ();                        \
  typedef X value_type;

//! @brief Prototype of the target class @a T<X,Y>
//!  with representation @a R<X,Y>
#define INDIRECT_PROTO_2(T,R,X,Y)		\
MMX_ALLOCATORS					\
protected:					\
  R<X,Y >* rep;					\
public:						\
  inline T (R<X,Y >* rep2);			\
  inline T (const R<X,Y >* rep2, bool inc);	\
  inline T (const T& x);			\
  inline ~T ();					\
  inline T& operator= (const T& x);		\
  inline const R<X,Y >* operator-> () const;	\
  inline void secure ();

//! @brief Prototype of the target class @a T<X,Y,Z>
//! with representation @a R<X,Y,Z>
#define INDIRECT_PROTO_3(T,R,X,Y,Z)		\
MMX_ALLOCATORS					\
protected:					\
  R<X,Y,Z >* rep;				\
public:						\
  inline T (R<X,Y,Z >* rep2);			\
  inline T (const R<X,Y,Z >* rep2, bool inc);	\
  inline T (const T& x);			\
  inline ~T ();					\
  inline T& operator= (const T& x);		\
  inline const R<X,Y,Z >* operator-> () const;	\
  inline void secure ();

/******************************************************************************
* Indirect data structures with reference counting
******************************************************************************/

struct rep_struct {
MMX_ALLOCATORS
  int ref_count;
  inline rep_struct (): ref_count (1) {}
  virtual inline ~rep_struct () {}
};

#define REP_STRUCT :public rep_struct
#define INC_COUNT(x) { (x)->ref_count++; }
#define DEC_COUNT(x) { if (0==--((x)->ref_count)) delete (x); }
#define INC_NULL_COUNT(x) { if ((x)!=NULL) (x)->ref_count++; }
#define DEC_NULL_COUNT(x) { \
  if ((x)!=NULL && 0==--((x)->ref_count)) delete (x); }

template<typename C> C copy (const C& x) {
  assert (false); // object should really be copied before access
  return x; }
template<typename C> inline nat hard_hash (const C& x) {
  return soft_hash (x); }
template<typename C> inline nat as_hash (const C* p) {
  return ((nat) ((uintptr_t) ((void*) p))) >> 3; }
template<typename C> inline bool hard_eq (const C& x, const C& y) {
  return eq (x, y); }
template<typename C> inline bool hard_neq (const C& x, const C& y) {
  return neq (x, y); }
template<typename C> inline bool hard_less (const C& x, const C& y) {
  return ((void*) inside (x)) < ((void*) inside (y)); }
template<typename C> inline bool hard_gtr (const C& x, const C& y) {
  return ((void*) inside (x)) > ((void*) inside (y)); }

#define INDIRECT_IMPL(T,R)						\
  inline T::T (R* rep2): rep(rep2) {}					\
  inline T::T (const R* rep2, bool with_inc):				\
    rep((R*) (void*) rep2) { (void) with_inc; INC_COUNT (rep); }	\
  inline T::T (const T& x): rep(x.rep) { INC_COUNT (rep); }		\
  inline T::~T () { DEC_COUNT (rep); }					\
  inline const R* T::operator -> () const { return rep; }		\
  inline T& T::operator = (const T& x) {				\
    INC_COUNT (x.rep); DEC_COUNT (rep);					\
    rep=x.rep; return *this; }						\
  inline void T::secure () {						\
    if (rep->ref_count>1) *this= copy (*this); }			\
  inline R* inside (const T& x) {					\
    return const_cast<R*> (x.operator -> ()); }				\
  STMPL inline nat hard_hash (const T& x) {				\
    return as_hash (x.operator -> ()); }				\
  STMPL inline bool hard_eq (const T& x, const T& y) {			\
    return (x.operator -> ()) == (y.operator -> ()); }			\
  STMPL inline bool hard_neq (const T& x, const T& y) {			\
    return (x.operator -> ()) != (y.operator -> ()); }

#define INDIRECT_NULL_IMPL(T,R)						\
  inline T::T (R* rep2): rep(rep2) {}					\
  inline T::T (const R* rep2, bool with_inc):				\
    rep((R*) (void*) rep2) { (void) with_inc; INC_NULL_COUNT (rep); }	\
  inline T::T (const T& x): rep(x.rep) { INC_NULL_COUNT (rep); }	\
  inline T::~T () { DEC_NULL_COUNT (rep); }				\
  inline const R* T::operator -> () const { return rep; }		\
  inline T& T::operator = (const T& x) {				\
    INC_NULL_COUNT (x.rep); DEC_NULL_COUNT (rep);			\
    rep=x.rep; return *this; }						\
  inline void T::secure () {						\
    if (rep->ref_count>1) *this= copy (*this); }			\
  inline R* inside (const T& x) {					\
    return const_cast<R*> (x.operator -> ()); }				\
  STMPL inline nat hard_hash (const T& x) {				\
    return as_hash (x.operator -> ()); }				\
  STMPL inline bool hard_eq (const T& x, const T& y) {			\
    return (x.operator -> ()) == (y.operator -> ()); }			\
  STMPL inline bool hard_neq (const T& x, const T& y) {			\
    return (x.operator -> ()) != (y.operator -> ()); }

#define INDIRECT_IMPL_1(T,R,XX,X)					\
  template<XX> inline T<X >::T (R<X >* rep2): rep(rep2) {}		\
  template<XX> inline T<X >::T (const R<X >* rep2, bool with_inc):	\
    rep((R<X >*) (void*) rep2) { (void) with_inc; INC_COUNT (rep); }	\
  template<XX> inline T<X >::T (const T<X >& x): rep(x.rep) {		\
    INC_COUNT (rep); }							\
  template<XX> inline T<X >::~T () { DEC_COUNT (rep); }			\
  template<XX> inline const R<X >* T<X >::operator -> () const {	\
    return rep; }							\
  template<XX> inline R<X >* T<X >::operator -> () {			\
    return rep; }							\
  template<XX> inline T<X >&						\
  T<X >::operator = (const T<X >& x) {					\
    INC_COUNT (x.rep); DEC_COUNT (rep);					\
    rep=x.rep; return *this; }						\
  template<XX> inline void T<X >::secure () {				\
    if (rep->ref_count>1) *this= copy (*this); }			\
  template<XX> R<X >* inside (const T<X >& x) {				\
    return const_cast<R<X >*> (x.operator -> ()); }			\
  template<XX> inline nat hard_hash (const T<X >& x) {			\
    return as_hash (x.operator -> ()); }				\
  template<XX> inline bool hard_eq (const T<X >& x, const T<X >& y) {	\
    return (x.operator -> ()) == (y.operator -> ()); }			\
  template<XX> inline bool hard_neq (const T<X >& x, const T<X >& y) {	\
    return (x.operator -> ()) != (y.operator -> ()); }

#define INDIRECT_NULL_IMPL_1(T,R,XX,X)					  \
  template<XX> inline T<X >::T (R<X >* rep2): rep(rep2) {}		  \
  template<XX> inline T<X >::T (const R<X >* rep2, bool with_inc):	  \
    rep((R<X >*) (void*) rep2) { (void) with_inc; INC_NULL_COUNT (rep); } \
  template<XX> inline T<X >::T (const T<X >& x): rep(x.rep) {		  \
    INC_NULL_COUNT (rep); }						  \
  template<XX> inline T<X >::~T () { DEC_NULL_COUNT (rep); }		  \
  template<XX> inline const R<X >* T<X >::operator -> () const {	  \
    return rep; }							  \
  template<XX> inline R<X >* T<X >::operator -> () {			  \
    return rep; }							  \
  template<XX> inline T<X >&						  \
  T<X >::operator = (const T<X >& x) {					  \
    INC_NULL_COUNT (x.rep); DEC_NULL_COUNT (rep);			  \
    rep=x.rep; return *this; }						  \
  template<XX> inline void T<X >::secure () {				  \
    if (rep->ref_count>1) *this= copy (*this); }			  \
  template<XX> R<X >* inside (const T<X >& x) {				  \
    return const_cast<R<X >*> (x.operator -> ()); }			  \
  template<XX> inline nat hard_hash (const T<X >& x) {			  \
    return as_hash (x.operator -> ()); }				  \
  template<XX> inline bool hard_eq (const T<X >& x, const T<X >& y) {	  \
    return (x.operator -> ()) == (y.operator -> ()); }			  \
  template<XX> inline bool hard_neq (const T<X >& x, const T<X >& y) {	  \
    return (x.operator -> ()) != (y.operator -> ()); }

#define INDIRECT_IMPL_2(T,R,XX,X,YY,Y)					      \
  template<XX,YY> inline T<X,Y >::T (R<X,Y >* rep2): rep(rep2) {}	      \
  template<XX,YY> inline T<X,Y >::T (const R<X,Y >* rep2, bool inc):	      \
    rep((R<X,Y >*) (void*) rep2) { (void) inc; INC_COUNT (rep); }	      \
  template<XX,YY> inline T<X,Y >::T (const T<X,Y >& x): rep(x.rep) {	      \
    INC_COUNT (rep); }							      \
  template<XX,YY> inline T<X,Y >::~T () { DEC_COUNT (rep); }		      \
  template<XX,YY> inline const R<X,Y >* T<X,Y >::operator -> () const {	      \
    return rep; }							      \
  template<XX,YY> inline T<X,Y >&					      \
  T<X,Y >::operator = (const T<X,Y >& x) {				      \
    INC_COUNT (x.rep); DEC_COUNT (rep);					      \
    rep=x.rep; return *this; }						      \
  template<XX,YY> inline void T<X,Y >::secure () {			      \
    if (rep->ref_count>1) *this= copy (*this); }			      \
  template<XX,YY> R<X,Y >* inside (const T<X,Y >& x) {			      \
    return const_cast<R<X,Y >*> (x.operator -> ()); }			      \
  template<XX,YY> inline nat hard_hash (const T<X,Y >& x) {		      \
    return as_hash (x.operator -> ()); }				      \
  template<XX,YY> inline bool hard_eq (const T<X,Y >& x, const T<X,Y >& y) {  \
    return (x.operator -> ()) == (y.operator -> ()); }			      \
  template<XX,YY> inline bool hard_neq (const T<X,Y >& x, const T<X,Y >& y) { \
    return (x.operator -> ()) != (y.operator -> ()); }

#define INDIRECT_IMPL_3(T,R,XX,X,YY,Y,ZZ,Z)				\
  template<XX,YY,ZZ> inline T<X,Y,Z >::T (R<X,Y,Z >* rep2):		\
    rep(rep2) {}							\
  template<XX,YY,ZZ> inline						\
  T<X,Y,Z >::T (const R<X,Y,Z >* rep2, bool inc):			\
    rep((R<X,Y,Z >*) (void*) rep2) { (void) inc; INC_COUNT (rep); }	\
  template<XX,YY,ZZ> inline T<X,Y,Z >::T (const T<X,Y,Z >& x):		\
    rep(x.rep) { INC_COUNT (rep); }					\
  template<XX,YY,ZZ> inline T<X,Y,Z >::~T () { DEC_COUNT (rep); }	\
  template<XX,YY,ZZ> inline const R<X,Y,Z >*				\
  T<X,Y,Z >::operator -> () const {					\
    return rep; }							\
  template<XX,YY,ZZ> inline T<X,Y,Z >&					\
  T<X,Y,Z >::operator = (const T<X,Y,Z >& x) {				\
    INC_COUNT (x.rep); DEC_COUNT (rep);					\
    rep=x.rep; return *this; }						\
  template<XX,YY,ZZ> inline void T<X,Y,Z >::secure () {			\
    if (rep->ref_count>1) *this= copy (*this); }			\
  template<XX,YY,ZZ> R<X,Y,Z >* inside (const T<X,Y,Z >& x) {		\
    return const_cast<R<X,Y,Z >*> (x.operator -> ()); }			\
  template<XX,YY,ZZ> inline nat hard_hash (const T<X,Y,Z>& x) {         \
    return as_hash (x.operator -> ()); }				\
  template<XX,YY,ZZ> inline bool					\
  hard_eq (const T<X,Y,Z >& x, const T<X,Y,Z >& y) {			\
    return (x.operator -> ()) == (y.operator -> ()); }			\
  template<XX,YY,ZZ> inline bool					\
  hard_neq (const T<X,Y,Z >& x, const T<X,Y,Z >& y) {			\
    return (x.operator -> ()) != (y.operator -> ()); }

/******************************************************************************
* Use the following classes instead of templating over values
******************************************************************************/

template<typename C, C x>
struct fixed_value {
  typedef C val_type;
  static const C val= x;
  static inline C dyn_val () { return val; }
};

template<typename C, typename V>
struct variable_value {
  typedef C val_type;
  static C val;
  static inline C dyn_val () { return val; }
};

template<typename C, typename V>
struct local_value {
  C old_value;
  inline local_value (const C& new_value) {
    old_value= variable_value<C,V>::val;
    variable_value<C,V>::val= new_value; }
  inline ~local_value () {
    variable_value<C,V>::val= old_value; }
};

/******************************************************************************
* Forward definitions of some classes
******************************************************************************/

class string;
struct vector_naive;
template<typename C> struct vector_variant_helper { typedef vector_naive VV; };
template<typename C, typename V=typename vector_variant_helper<C>::VV>
class vector;

} // namespace mmx
#endif // __MMX_BASIX_HPP
