
/******************************************************************************
* MODULE     : dag.hpp
* DESCRIPTION: Customizable dags over a domain of "scalars"
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__DAG__HPP
#define __MMX__DAG__HPP
#include <basix/exception.hpp>
#include <basix/dag_operators.hpp>
#include <basix/wrap.hpp>
#include <basix/vector.hpp>

namespace mmx {
struct basic_dag {};
#define TMPL_DEF template<typename C, typename DT= basic_dag>
#define TMPL template<typename C, typename DT>
#define TMPL1 template<typename A1, typename C, typename DT>
#define TMPL2 template<typename A1, typename A2, typename C, typename DT>
#define Dag dag<C,DT>
#define Dag_A1 dag<A1,DT>
#define Dag_A2 dag<A2,DT>
#define Dag_void dag<void,DT>
#define Dag_generic dag<generic,DT>
#define Dag_ref dag<C&,DT>
#define Dag_A1 dag<A1,DT>
#define Abstract abstract_dag<C,DT>

/******************************************************************************
* Abstract Dags
******************************************************************************/

TMPL_DEF
struct abstract_dag: public rep_struct {
  inline abstract_dag () {}
  inline virtual ~abstract_dag () {}
  virtual syntactic Flatten () const = 0;
};

TMPL_DEF
class dag {
INDIRECT_PROTO_2 (dag, abstract_dag, C, DT)
  inline void* operator * () const { return (void*) rep; }
  dag ();
};
INDIRECT_IMPL_2 (dag, abstract_dag, typename C, C, typename DT, DT)

TMPL inline syntactic flatten (const Dag& d) {
  return d->Flatten (); }

PRINT_SUGAR(TMPL,Dag)
WRAP_INDIRECT_IMPL(TMPL,Dag)

/******************************************************************************
* Concrete dags
******************************************************************************/

template<typename Op, typename C, typename DT, typename MT>
struct nullary_dag: public abstract_dag<C,MT> {
  inline nullary_dag () {}
  inline syntactic Flatten () const {
    return Op::template op<syntactic> (); }
};

template<typename Op, typename C, typename DT>
inline Dag make_dag () {
  return Dag ((Abstract*) new nullary_dag<Op,C,DT,DT> ()); }

template<typename C, typename DT>
dag<C,DT>::dag ():
  rep ((Abstract*) new nullary_dag<unimplemented_op,C,DT,DT> ()) {}

template<typename Op, typename C, typename A1, typename DT, typename MT>
struct unary_dag: public abstract_dag<C,MT> {
  A1 arg1;
  inline unary_dag (const A1& a1):
    arg1 (a1) {}
  inline syntactic Flatten () const {
    return Op::op (flatten (arg1)); }
};

template<typename Op, typename C, typename DT, typename A1>
inline Dag make_dag (const A1& a1) {
  return Dag ((Abstract*) new unary_dag<Op,C,A1,DT,DT> (a1)); }

template<typename Op, typename C, typename A1, typename A2,
	 typename DT, typename MT>
struct binary_dag: public abstract_dag<C,MT> {
  A1 arg1;
  A2 arg2;
  inline binary_dag (const A1& a1, const A2& a2):
    arg1 (a1), arg2 (a2) {}
  inline syntactic Flatten () const {
    return Op::op (flatten (arg1), flatten (arg2)); }
};

template<typename Op, typename C, typename DT, typename A1, typename A2>
inline Dag make_dag (const A1& a1, const A2& a2) {
  return Dag ((Abstract*) new binary_dag<Op,C,A1,A2,DT,DT> (a1, a2)); }

template<typename Op, typename C, typename A1, typename A2, typename A3,
	 typename DT, typename MT>
struct ternary_dag: public abstract_dag<C,MT> {
  A1 arg1;
  A2 arg2;
  A3 arg3;
  inline ternary_dag (const A1& a1, const A2& a2, const A3& a3):
    arg1 (a1), arg2 (a2), arg3 (a3) {}
  inline syntactic Flatten () const {
    return Op::op (flatten (arg1), flatten (arg2), flatten (arg3)); }
};

template<typename Op, typename C, typename DT,
	 typename A1, typename A2, typename A3>
inline Dag make_dag (const A1& a1, const A2& a2, const A3& a3) {
  return Dag ((Abstract*) new ternary_dag<Op,C,A1,A2,A3,DT,DT> (a1, a2, a3)); }

/******************************************************************************
* Standard specializations of basic operations
******************************************************************************/

TMPL inline Dag operator - (const Dag& d1) {
  return make_dag<neg_op,C,DT> (d1); }
TMPL inline Dag operator + (const Dag& d1, const Dag& d2) {
  return make_dag<add_op,C,DT> (d1, d2); }
TMPL inline Dag operator - (const Dag& d1, const Dag& d2) {
  return make_dag<sub_op,C,DT> (d1, d2); }
TMPL inline Dag operator * (const Dag& d1, const Dag& d2) {
  return make_dag<mul_op,C,DT> (d1, d2); }
TMPL inline Dag operator / (const Dag& d1, const Dag& d2) {
  return make_dag<div_op,C,DT> (d1, d2); }
TMPL inline Dag quo (const Dag& d1, const Dag& d2) {
  return make_dag<quo_op,C,DT> (d1, d2); }
TMPL inline Dag rem (const Dag& d1, const Dag& d2) {
  return make_dag<rem_op,C,DT> (d1, d2); }

TMPL1 inline Dag_A1 quote_equal (const Dag& d1, const Dag& d2) {
  return make_dag<equal_op,A1,DT> (d1, d2); }
TMPL1 inline Dag_A1 quote_unequal (const Dag& d1, const Dag& d2) {
  return make_dag<unequal_op,A1,DT> (d1, d2); }
TMPL1 inline Dag_A1 quote_less (const Dag& d1, const Dag& d2) {
  return make_dag<less_op,A1,DT> (d1, d2); }
TMPL1 inline Dag_A1 quote_lesseq (const Dag& d1, const Dag& d2) {
  return make_dag<lesseq_op,A1,DT> (d1, d2); }
TMPL1 inline Dag_A1 quote_gtr (const Dag& d1, const Dag& d2) {
  return make_dag<gtr_op,A1,DT> (d1, d2); }
TMPL1 inline Dag_A1 quote_gtreq (const Dag& d1, const Dag& d2) {
  return make_dag<gtreq_op,A1,DT> (d1, d2); }

TMPL1 inline Dag_A1 quote_assign (const Dag_ref& d1, const Dag& d2) {
  return make_dag<in_place_op<id_op>,A1,DT> (d1, d2); }
TMPL1 inline Dag_A1 quote_plus_assign (const Dag_ref& d1, const Dag& d2) {
  return make_dag<in_place_op<add_op>,A1,DT> (d1, d2); }
TMPL1 inline Dag_A1 quote_minus_assign (const Dag_ref& d1, const Dag& d2) {
  return make_dag<in_place_op<sub_op>,A1,DT> (d1, d2); }
TMPL1 inline Dag_A1 quote_times_assign (const Dag_ref& d1, const Dag& d2) {
  return make_dag<in_place_op<mul_op>,A1,DT> (d1, d2); }
TMPL1 inline Dag_A1 quote_over_assign (const Dag_ref& d1, const Dag& d2) {
  return make_dag<in_place_op<div_op>,A1,DT> (d1, d2); }

template<typename Op, typename R, typename DT>
inline Dag_generic make_routine () {
  return make_dag<nullary_routine_op<Op,R>,generic,DT> (); }
template<typename Op, typename R, typename A1, typename DT>
inline Dag_generic make_routine (const Dag_generic& d) {
  return make_dag<unary_routine_op<Op,R,A1>,generic,DT> (d); }
template<typename Op, typename R, typename A1, typename A2, typename DT>
inline Dag_generic make_routine (const Dag_generic& d, const Dag_generic& e) {
  return make_dag<binary_routine_op<Op,R,A1,A2>,generic,DT> (d, e); }

TMPL inline Dag_void define_variable (const generic& var, const Dag& val) {
  return make_dag<define_op,void,DT> (var, val); }
TMPL inline Dag constant_lookup (const generic& var) {
  return make_dag<constant_lookup_op,C,DT> (var); }
TMPL inline Dag_ref mutable_lookup (const generic& var) {
  return make_dag<mutable_lookup_op,C&,DT> (var); }

template<typename DT> inline Dag_void begin_do () {
  return make_dag<begin_op,void,DT> (); }
TMPL inline Dag begin_do (const Dag_void& l, const Dag& r) {
  return make_dag<begin_op,C,DT> (l, r); }
TMPL1 inline Dag_void if_then (const Dag_A1& c, const Dag& t) {
  return make_dag<if_op,void,DT> (c, t); }
TMPL1 inline Dag if_then_else (const Dag_A1& c, const Dag& t, const Dag& e) {
  return make_dag<if_op,C,DT> (c, t, e); }
TMPL1 inline Dag_void while_do (const Dag_A1& c, const Dag& b) {
  return make_dag<while_op,void,DT> (c, b); }
TMPL2 inline Dag_void
for_in_do (const generic& v, const Dag_A2& it, const Dag& b) {
  return make_dag<for_in_op<A1>,void,DT> (v, it, b); }

TMPL1 inline Dag
unary_closure (const generic& var, const Dag& body) {
  return make_dag<unary_closure_op<A1>,C,DT> (var, body); }
TMPL1 inline Dag_generic
unary_lambda (const generic& var, const Dag& body) {
  return make_dag<unary_lambda_op<A1>,generic,DT> (var, body); }
TMPL1 inline Dag
unary_apply (const Dag_generic& fun, const Dag_A1& arg) {
  return make_dag<unary_apply_op<A1>,C,DT> (fun, arg); }

TMPL1 inline Dag
n_ary_closure (const vector<generic>& vars, const Dag& body) {
  return make_dag<n_ary_closure_op<A1>,C,DT> (vars, body); }
TMPL1 inline Dag_generic
n_ary_lambda (const vector<generic>& vars, const Dag& body) {
  return make_dag<n_ary_lambda_op<A1>,generic,DT> (vars, body); }
TMPL1 inline Dag
n_ary_apply (const Dag_generic& fun, const vector<Dag_A1>& args) {
  return make_dag<n_ary_apply_op<A1>,C,DT> (fun, args); }

#undef TMPL_DEF
#undef TMPL
#undef TMPL1
#undef TMPL2
#undef Dag
#undef Dag_A1
#undef Dag_A2
#undef Dag_void
#undef Dag_generic
#undef Dag_ref
#undef Dag_A1
#undef Abstract
} // namespace mmx
#endif // __MMX__DAG__HPP
