
/******************************************************************************
* MODULE     : dag_operators.hpp
* DESCRIPTION: Extra operators for dags
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_DAG_OPERATORS_HPP
#define __MMX_DAG_OPERATORS_HPP
#include <basix/operators.hpp>

//! \file

namespace mmx {

struct unimplemented_op {
  template<typename C> static inline C
  op () {
    ERROR ("non implemented operation"); }
  template<typename A1> static inline void
  op (const A1& a1) { (void) a1;
    ERROR ("non implemented operation"); }
  template<typename A1, typename A2> static inline void
  op (const A1& a1, const A2& a2) { (void) a1; (void) a2;
    ERROR ("non implemented operation"); }
  template<typename A1, typename A2, typename A3> static inline void
  op (const A1& a1, const A2& a2, const A3& a3) {
    (void) a1; (void) a2; (void) a3;
    ERROR ("non implemented operation"); }
};

/******************************************************************************
* Wrapper operators for gluing functions
******************************************************************************/

template<typename Op, typename C>
struct nullary_routine_op: public Op {
  template<typename D> static inline C op () {
    return as<C> (Op::template op<D> ()); }
};

template<typename Op, typename C, typename A1>
struct unary_routine_op: public Op {
  static inline syntactic op (const syntactic& a1) {
    return Op::op (a1); }
  static inline generic op (const generic& a1) {
    return as<generic> (Op::op (as<A1> (a1))); }
};

template<typename Op, typename C, typename A1, typename A2>
struct binary_routine_op: public Op {
  static inline syntactic op (const syntactic& a1, const syntactic& a2) {
    return Op::op (a1, a2); }
  static inline generic op (const generic& a1, const generic& a2) {
    return as<generic> (Op::op (as<A1> (a1), as<A2> (a2))); }
};

template<typename Op, typename C, typename A1, typename A2, typename A3>
struct ternary_routine_op: public Op {
  static inline syntactic
  op (const syntactic& a1, const syntactic& a2, const syntactic& a3) {
    return Op::op (a1, a2, a3); }
  static inline generic
  op (const generic& a1, const generic& a2, const generic& a3) {
    return as<generic> (Op::op (as<A1> (a1), as<A2> (a2), as<A3> (a3))); }
};

/******************************************************************************
* Data lookup operators
******************************************************************************/

struct define_op {
  static generic name () { return GEN_DEFINE; }
  static inline syntactic
  op (const syntactic& var, const syntactic& val) {
    return apply (GEN_DEFINE, var, val); }
};

struct constant_lookup_op {
  static generic name () { return "constant_lookup"; }
  static inline syntactic op (const syntactic& var) { return var; }
};

struct mutable_lookup_op {
  static generic name () { return "mutable_lookup"; }
  static inline syntactic op (const syntactic& var) { return var; }
};

/******************************************************************************
* Program control operators
******************************************************************************/

struct begin_op {
  static generic name () { return GEN_BEGIN; }
  template<typename C> static inline C op () {
    return apply (GEN_BEGIN); }
  static inline syntactic op (const syntactic& l, const syntactic& r) {
    return apply (GEN_BEGIN, l, r); }
};

struct if_op {
  static generic name () { return GEN_IF; }
  static inline syntactic
  op (const syntactic& c, const syntactic& t) {
    return apply (GEN_IF, c, t); }
  static inline syntactic
  op (const syntactic& c, const syntactic& t, const syntactic& e) {
    return apply (GEN_IF, c, t, e); }
};

struct while_op {
  static generic name () { return GEN_WHILE; }
  static inline syntactic
  op (const syntactic& c, const syntactic& b) {
    return apply (GEN_WHILE, c, b); }
};

template<typename T>
struct for_in_op {
  static generic name () { return GEN_FOR; }
  static inline syntactic
  op (const syntactic& v, const syntactic& g, const syntactic& b) {
    return apply (GEN_FOR, apply (GEN_IN, v, g), b); }
};

template<typename A1>
struct unary_closure_op {
  static generic name () { return "closure"; }
  static inline syntactic
  op (const syntactic& vars, const syntactic& body) {
    return apply ("closure", vars, body); }
};

template<typename A1>
struct unary_lambda_op {
  static generic name () { return "closure"; }
  static inline syntactic
  op (const syntactic& vars, const syntactic& body) {
    return apply (GEN_LAMBDA, vars, body); }
};

template<typename A1>
struct unary_apply_op {
  static generic name () { return "apply"; }
  static inline syntactic
  op (const syntactic& f, const syntactic& a1) {
    return apply (f, a1); }
};

template<typename A1>
struct n_ary_closure_op {
  static generic name () { return "closure"; }
  static inline syntactic
  op (const syntactic& vars, const syntactic& body) {
    return apply ("closure", vars, body); }
};

template<typename A1>
struct n_ary_lambda_op {
  static generic name () { return "closure"; }
  static inline syntactic
  op (const syntactic& vars, const syntactic& body) {
    return apply (GEN_LAMBDA, vars, body); }
};

template<typename A1>
struct n_ary_apply_op {
  static generic name () { return "apply"; }
  static inline syntactic
  op (const syntactic& f, const syntactic& args) {
    return apply (f, arguments (args)); }
};

} // namespace mmx
#endif // __MMX_DAG_OPERATORS_HPP
