
/******************************************************************************
* MODULE     : double.hpp
* DESCRIPTION: Extra routines for doubles
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__DOUBLE_HPP__
#define __MMX__DOUBLE_HPP__
#include <fenv.h>
#include <basix/syntactic.hpp>

//! \file

namespace mmx {

/******************************************************************************
* Constants
******************************************************************************/

template<typename C> struct constant_helper;

STMPL
struct constant_helper<double > {
  static inline double nan () {
    return NAN; }
  static inline double plus_infinity () {
    return HUGE_VAL; }
  static inline double minus_infinity () {
    return -HUGE_VAL; }
  static inline double infinity () {
    return HUGE_VAL; }
  static inline double fuzz () {
    return 0.0; }
  static inline double log2 () {
    return 0.6931471805599453094172321214581766L; }
  static inline double pi () {
    return 3.1415926535897932384626433832795029L; }
  static inline double euler () {
    return 0.5772156649015328606065120900824024L; }
};

/******************************************************************************
* Shifting
******************************************************************************/

STMPL inline double
incexp2 (const double& x, const int& y) { return ldexp (x, y); }
STMPL inline void
incexp2_assign (double& x, const int& y) { x= ldexp (x, y); }
STMPL inline void
incexp2 (double& x, const double& y, const int& z) { x= ldexp (y, z); }
STMPL inline double
decexp2 (const double& x, const int& y) { return ldexp (x, -y); }
STMPL inline void
decexp2_assign (double& x, const int& y) { x= ldexp (x, -y); }
STMPL inline void
decexp2 (double& x, const double& y, const int& z) { x= ldexp (y, -z); }

/******************************************************************************
* Elementary functions
******************************************************************************/

inline double sqrt (const double& x) { return std::sqrt (x); }
inline double hypot (const double& x, const double& y) {
  return ::hypot (x, y); }
inline double pow (const double& x, const double& y) {
  return std::pow (x, y); }
inline double exp (const double& x) { return std::exp (x); }
inline double exp2 (const double& x) { return ::exp2 (x); }
inline double log (const double& x) { return std::log (x); }
inline double log2 (const double& x) { return ::log2 (x); }
inline double cos (const double& x) { return std::cos (x); }
inline double sin (const double& x) { return std::sin (x); }
inline double tan (const double& x) { return std::tan (x); }
inline double cosh (const double& x) { return std::cosh (x); }
inline double sinh (const double& x) { return std::sinh (x); }
inline double tanh (const double& x) { return std::tanh (x); }
inline double acos (const double& x) { return std::acos (x); }
inline double asin (const double& x) { return std::asin (x); }
inline double atan (const double& x) { return std::atan (x); }
inline double atan2 (const double& y, const double& x) {
  return std::atan2 (y, x); }

INV_TRIGO_SUGAR(,double)
INV_HYPER_SUGAR(,double)
ARG_HYPER_SUGAR(,double)

/******************************************************************************
* Other functions
******************************************************************************/

inline double as_double (const double& x) { return x; }
inline double rnd_error (const double& x) { return ldexp (x, -51); }
  inline bool is_nan (const double& x) { return std::isnan (x); }
  inline bool is_infinite (const double& x) { return std::isinf (x); }

inline double floor (const double& x) { return std::floor (x); }
inline double ceil (const double& x) { return std::ceil (x); }
inline double round (const double& x) { return std::floor (x + 0.5); }

STMPL inline double magnitude (const double& x) { return log2 (x); }
inline int exponent (const double& x) { return (int) logb (x); }
STMPL inline double double_exponent (const double& x) { return logb (x); }

}
#endif // __MMX__DOUBLE_HPP__
