
/******************************************************************************
* MODULE     : exception.hpp
* DESCRIPTION: The type for exceptional behaviour
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __EXCEPTION_HPP
#define __EXCEPTION_HPP
#include <basix/syntactic.hpp>

//! \file

namespace mmx {

/******************************************************************************
* The exception class for generic error messages
******************************************************************************/

class exception {
MMX_ALLOCATORS
  generic rep;
public:
  inline generic operator * () const { return rep; }
  inline exception (const generic& g): rep (g) {}
  inline exception (const exception& t): rep (t.rep) {}
  exception (const string& msg, const generic& where);
};

inline nat hash (const exception& c) { return hash (*c); }
inline nat soft_hash (const exception& c) { return soft_hash (*c); }
inline nat hard_hash (const exception& c) { return hard_hash (*c); }
inline bool operator == (const exception& c1, const exception& c2) {
  return (*c1) == (*c2); }
inline bool operator != (const exception& c1, const exception& c2) {
  return (*c1) != (*c2); }
inline bool eq (const exception& c1, const exception& c2) {
  return eq (*c1, *c2); }
inline bool neq (const exception& c1, const exception& c2) {
  return neq (*c1, *c2); }
inline bool hard_eq (const exception& c1, const exception& c2) {
  return hard_eq (*c1, *c2); }
inline bool hard_neq (const exception& c1, const exception& c2) {
  return hard_neq (*c1, *c2); }

inline syntactic flatten (const exception& c) { return flatten (*c); }
inline ostream& operator << (ostream& out, const exception& c) {
  return out << *c; }

generic std_exception (const string& msg, const generic& where);
generic std_exception (const string& msg, const generic& a, const generic& w);
generic user_exception (const vector<generic>& msg, const generic& where);
generic wrong_nr_args (const generic& where);
generic type_mismatch (const generic& expect, const generic& where);
generic trace_push (const generic& exc, const generic& where);
generic trace_pull (const generic& exc);
generic trace_top (const generic& exc);
generic trace_bottom (const generic& exc, nat n= 1);
bool    has_trace (const generic& exc);

/******************************************************************************
* C++ exceptions
******************************************************************************/

#ifdef BASIX_ENABLE_EXCEPTIONS
//! \brief Make an exception from @a x
exception as_exception (const generic& x);
//@{
//! \brief Make an exception from @a msg
exception error_message (const char* msg);
exception error_message (const string& msg);
//@}
#define STRINGIFY(x) #x
#define TOSTRING(x) STRINGIFY(x)
#define ERROR(msg) \
  throw mmx::error_message (__FILE__ ":" TOSTRING(__LINE__) ": " msg);
#define ASSERT(expr,msg)					 \
  if (!(expr)) ERROR (msg);
#else // disable exceptions
#define ERROR(msg) assert(false)
#define ASSERT(expr,msg) assert(expr)
#endif

#ifdef BASIX_ENABLE_VERIFY
#define VERIFY(expr,msg) ASSERT(expr,msg)
#else // disable verifications
#define VERIFY(expr,msg)
#endif

//! @brief Print @a message coming from @a routine in @a file. Exit with 1.
volatile void fatal_error (char* message, char* routine, char* file);

} // namespace mmx
#endif // __EXCEPTION_HPP
