
/******************************************************************************
* MODULE     : generic.hpp
* DESCRIPTION: Generic expressions
* COPYRIGHT  : (C) 2005  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __GENERIC_HPP
#define __GENERIC_HPP
#include <basix/defaults.hpp>

//! \file

namespace mmx {
class generic;
class syntactic;
class string;
template<typename T,typename F> struct as_helper;
template<typename T,typename F> inline T as (const F& x);
template<typename C> struct constant_helper;
template<typename C> struct inspector;
extern bool booting_done;

/******************************************************************************
* The generic class
******************************************************************************/

class generic_rep: public rep_struct {
protected:
  virtual nat get_type () const = 0;
  virtual nat get_symbolic_type () const = 0;
  virtual nat get_species_type () const = 0;
  virtual nat get_length () const = 0;
  virtual generic get_child (nat i) const = 0;
  virtual nat get_hard_hash_value () const = 0;
  virtual nat get_soft_hash_value () const = 0;
  virtual nat get_hash_value () const = 0;
  virtual bool is_hard_eq (const generic& rep) const = 0;
  virtual bool is_eq (const generic& rep) const = 0;
  virtual bool is_equal (const generic& rep) const = 0;
  virtual generic duplicate_me () const = 0;
  virtual syntactic expression () const = 0;

public:
  inline generic_rep () {}
  inline virtual ~generic_rep () {}

  friend class generic;
  inline friend nat type (const generic& g);
  inline friend nat symbolic_type (const generic& g);
  inline friend nat species_type (const generic& g);
  inline friend nat N (const generic& g);
  inline friend nat gen_hash (const generic& g);
  inline friend nat soft_hash (const generic& g);
  inline friend nat hash (const generic& g);
  inline friend bool gen_eq (const generic& g1, const generic& g2);
  inline friend bool gen_neq (const generic& g1, const generic& g2);
  inline friend bool eq (const generic& g1, const generic& g2);
  inline friend bool neq (const generic& g1, const generic& g2);
  friend bool operator == (const generic& g1, const generic& g2);
  friend bool operator != (const generic& g1, const generic& g2);
  friend generic duplicate (const generic& g);
  friend syntactic flatten (const generic& s);
};

class generic {
INDIRECT_PROTO (generic, generic_rep)
public:
  generic ();
  generic (const char* s);
  generic (const string& s);
  generic (int i);
  generic (nat i);
  generic operator () () const;
  generic operator () (const generic& g1) const;
  generic operator () (const generic& g1, const generic& g2) const;
  generic operator () (const generic& g1, const generic& g2,
		       const generic& g3) const;

  inline friend generic_rep* inspect (const generic& g) {
    return g.rep; }
  inline friend nat type (const generic& g) {
    return g->get_type (); }
  inline friend nat symbolic_type (const generic& g) {
    return g->get_symbolic_type (); }
  inline friend nat species_type (const generic& g) {
    return g->get_species_type (); }
  inline friend nat N (const generic& g) {
    return g->get_length (); }
  inline generic operator [] (nat i) const {
    return rep->get_child (i); }
  inline friend nat gen_hash (const generic& g) {
    return g->get_hard_hash_value (); }
  inline friend nat soft_hash (const generic& g) {
    return g->get_soft_hash_value (); }
  inline friend nat hash (const generic& g) {
    return g->get_hash_value (); }
  inline friend bool gen_eq (const generic& g1, const generic& g2) {
    return g1.rep->is_hard_eq (g2); }
  inline friend bool gen_neq (const generic& g1, const generic& g2) {
    return !g1.rep->is_hard_eq (g2); }
  inline friend bool eq (const generic& g1, const generic& g2) {
    return g1.rep->is_eq (g2); }
  inline friend bool neq (const generic& g1, const generic& g2) {
    return !g1.rep->is_eq (g2); }

  friend bool operator == (const generic& g1, const generic& g2);
  friend bool operator != (const generic& g1, const generic& g2);
  friend generic duplicate (const generic& g);
  friend syntactic flatten (const generic& g);

  friend class generic_rep;
};
INDIRECT_IMPL (generic, generic_rep)

ostream& operator << (ostream& out, const generic& g);

/******************************************************************************
* Type information
******************************************************************************/

nat new_type_id ();
inline nat binary_id (nat id1, nat id2) { return id1 ^ (0x10115 * id2); }
template<typename T> struct type_information { static nat id; };
template<typename T> nat type_information<T>::id= new_type_id ();
STMPL struct type_information<generic> { static nat id; };
template<typename C> inline bool is (const generic& x) {
  return type (x) == type_information<C>::id; }
generic type_name (const generic& g);

#define SYMBOLIC_UNKNOWN      0
#define SYMBOLIC_INTEGER      1
#define SYMBOLIC_RATIONAL     2
#define SYMBOLIC_FLOATING     3
#define SYMBOLIC_BALL         4
#define SYMBOLIC_REAL         5
#define SYMBOLIC_LITERAL      6
#define SYMBOLIC_COMPOUND     7
#define SYMBOLIC_SUM          8
#define SYMBOLIC_PRODUCT      9
#define SYMBOLIC_DERIVATIVE  10
#define SYMBOLIC_FUNCTION    11

template<typename T> struct symbolic_type_information {
  static const nat id= SYMBOLIC_UNKNOWN; };

#define SPECIES_DEFAULT   0
#define SPECIES_VECTOR    1
#define SPECIES_LIST      2
#define SPECIES_TABLE     3

template<typename T> struct species_type_information {
  static const nat id= SPECIES_DEFAULT; };
inline bool is_vector (const generic& x) {
  return species_type (x) == SPECIES_VECTOR; }
inline bool is_list (const generic& x) {
  return species_type (x) == SPECIES_LIST; }
inline bool is_table (const generic& x) {
  return species_type (x) == SPECIES_TABLE; }

/******************************************************************************
* Concrete instances
******************************************************************************/

template<typename C>
class generic_concrete_rep: public generic_rep {
public:
  C rep;

protected:
  nat get_type () const { return type_information<C>::id; }
  nat get_symbolic_type () const { return symbolic_type_information<C>::id; }
  nat get_species_type () const { return species_type_information<C>::id; }
  nat get_length () const { return inspector<C>::length (rep); }
  generic get_child (nat i) const { return inspector<C>::access (rep, i); }
  nat get_hard_hash_value () const { return hard_hash (rep); }
  nat get_soft_hash_value () const { return soft_hash (rep); }
  nat get_hash_value () const { return hash (rep); }
  bool is_hard_eq (const generic& g) const {
    if (type (g) != type_information<C>::id) return false;
    return hard_eq (rep, ((generic_concrete_rep<C>*) inspect (g)) -> rep); }
  bool is_eq (const generic& g) const {
    if (type (g) != type_information<C>::id) return false;
    return eq (rep, ((generic_concrete_rep<C>*) inspect (g)) -> rep); }
  bool is_equal (const generic& g) const {
    if (type (g) != type_information<C>::id) return false;
    return rep == ((generic_concrete_rep<C>*) inspect (g)) -> rep; }
  generic duplicate_me () const { return as<generic> (duplicate (rep)); }
  syntactic expression () const; // definition in syntactic.hpp

public:
  generic_concrete_rep (const C& rep2): rep (rep2) {}
};

template<typename C>
struct as_helper<generic,C> {
  static inline generic cv (const C& x) {
    return new generic_concrete_rep<C> (x); }
};

template<typename C>
struct as_helper<C,generic> {
  static inline C cv (const generic& x) {
    return ((generic_concrete_rep<C>*) inspect (x)) -> rep; }
};

STMPL
struct as_helper<generic,generic> {
  static inline generic cv (const generic& x) { return x; }
};

template<typename C> generic
as_generic (const C& c) {
  return construct (new generic_concrete_rep<C> (c));
}

/******************************************************************************
* Syntactic manipulation of generic objects
******************************************************************************/

generic lit (const char* s);
generic lit (const string& s);

generic gen (const generic& x1, const vector<generic>& v);
generic gen ();
generic gen (const generic& x1);
generic gen (const generic& x1, const string& x2);
generic gen (const generic& x1, const generic& x2);
generic gen (const generic& x1, const generic& x2, const generic& x3);
generic gen (const generic& x1, const generic& x2,
	     const generic& x3, const generic& x4);
generic gen (const generic& x1, const generic& x2,
	     const generic& x3, const generic& x4, const generic& x5);
generic gen (const generic& x1, const generic& x2,
	     const generic& x3, const generic& x4,
	     const generic& x5, const generic& x6);

generic void_value ();
generic comma ();
generic comma (const generic& x1, const generic& x2);
generic xgen (const generic& f, const generic& args);
generic xaccess (const generic& f, const generic& args);
generic xtuple (const generic& args);
generic xsqtuple (const generic& args);
generic xrow (const generic& args);
generic access (const generic& f, const generic& x);
generic access (const generic& f, const generic& x, const generic& y);
generic access (const generic& f, const vector<generic>& x);

generic as_object (const generic& v, nat tp_id);
generic as_object (const generic& v, const generic& tp);
generic as_generic (const generic& obj, nat tp_id);
generic as_generic (const generic& obj, const generic& tp);

bool is_scalar (const generic& g);    // for linear combinations
bool is_central (const generic& g);   // commutes with evrybody
bool is_operator (const generic& g);  // linear operators

bool scalar_expand ();          // a(b+c) = ab+ac for scalar a
bool scalar_power_expand ();    // (ab)^c = a^c b^c for scalar c
bool scalar_power_multiply ();  // (a^b)^c = a^(bc) for scalar c
bool central_expand ();         // a(b+c) = ab+ac for central a
bool central_power_expand ();   // (ab)^c = a^c b^c for central c
bool central_power_multiply (); // (a^b)^c = a^(bc) for central c

/******************************************************************************
* Further routines for syntactic manipulation (see generic_utils.cpp)
******************************************************************************/

generic range (const generic& g, nat start, nat end);
generic append (const generic& g1, const generic& g2);
generic cons (const generic& g1, const generic& g2);
generic cons (const generic& g1, const generic& g2, const generic& g3);
generic car (const generic& g);
generic cdr (const generic& g);

bool is_func (const generic& g, const char* f);
bool is_func (const generic& g, const char* f, nat n);
bool is_func (const generic& g, const generic& f);
bool is_func (const generic& g, const generic& f, nat n);
nat size (const generic& g);
int big_small_compare (const generic& g1, const generic& g2);
int small_big_compare (const generic& g1, const generic& g2);

/******************************************************************************
* Routines based on the current evaluator
******************************************************************************/

generic construct (const int& i);
generic construct (const nat& i);
generic construct (const double& x);
generic construct (const float& x);
generic construct (const generic& x);
generic eval (const generic& x);

STMPL
struct constant_helper<generic> {
  static generic pi ();
  static generic log2 ();
  static generic euler ();
  static generic catalan ();
  static generic imaginary ();
  static generic nan ();
  static generic fuzz ();
  static generic infinity ();
  static generic plus_infinity ();
  static generic minus_infinity ();
};

generic apply (const generic& f, const generic& x);
generic apply (const generic& f, const generic& x, const generic& y);
generic apply (const generic& f, const generic& x,
	       const generic& y, const generic& z);
generic apply (const generic& f, const vector<generic>& x);

generic operator - (const generic& x1);
generic operator + (const generic& x1, const generic& x2);
generic operator - (const generic& x1, const generic& x2);
generic operator * (const generic& x1, const generic& x2);
generic invert (const generic &x1);
generic operator / (const generic& x1, const generic& x2);
generic quo (const generic& x1, const generic& x2);
generic rem (const generic& x1, const generic& x2);
generic big_add (const vector<generic>& v);
generic big_mul (const vector<generic>& v);

generic numerator (const generic& x);
generic denominator (const generic& x);
generic gcd (const generic& x1, const generic& x2);
generic lcm (const generic& x1, const generic& x2);
template<typename C> struct xgcd_matrix;
xgcd_matrix<generic> xgcd (const generic& x1, const generic& x2);

generic pow (const generic& x1, const generic& x2);
generic pow (const generic& x1, const int& x2);
generic pow (const generic& x1, const nat& x2);
generic pow (const int& x1, const generic& x2);
generic pow (const nat& x1, const generic& x2);
generic sqrt (const generic& x1);
generic exp (const generic& x1);
generic log (const generic& x1);
generic cos (const generic& x1);
generic sin (const generic& x1);
generic tan (const generic& x1);
generic acos (const generic& x1);
generic asin (const generic& x1);
generic atan (const generic& x1);
generic trig (const generic& x1);
generic cosh (const generic& x1);
generic sinh (const generic& x1);
generic tanh (const generic& x1);
generic acosh (const generic& x1);
generic asinh (const generic& x1);
generic atanh (const generic& x1);
generic atan2 (const generic& x1, const generic& x2);
generic hypot (const generic& x1, const generic& x2);

bool operator == (const generic& x1, const generic& x2);
bool operator != (const generic& x1, const generic& x2);
bool operator <  (const generic& x1, const generic& x2);
bool operator <= (const generic& x1, const generic& x2);
bool operator >  (const generic& x1, const generic& x2);
bool operator >= (const generic& x1, const generic& x2);

generic min (const generic& x1, const generic& x2);
generic max (const generic& x1, const generic& x2);
generic abs (const generic& x1);
generic arg (const generic& x1);
generic Re (const generic& x1);
generic Im (const generic& x1);
generic conj (const generic& x1);
generic gaussian (const generic& x1, const generic& x2);
generic polar (const generic& x1, const generic& x2);

generic sharpen (const generic& x);

generic derive (const generic& x1);
generic derive (const generic& x1, const generic& v);
generic prime (const generic& x1);
generic integrate (const generic& x1);
generic integrate (const generic& x1, const generic& v);
generic substitute (const generic& x1, const generic& x2);
generic compose (const generic& x1, const generic& x2);
generic dilate (const generic& x1, const generic& x2);
generic solve (const generic& x1, const generic& x2);

generic sqrt_init (const generic& x, const generic& c);
generic log_init (const generic& x, const generic& c);
generic acos_init (const generic& x, const generic& c);
generic asin_init (const generic& x, const generic& c);
generic atan_init (const generic& x, const generic& c);
generic integrate_init (const generic& x, const generic& c);
generic solve_lde_init (const generic& x, const generic& c);

generic operator << (const generic& x1, const generic& x2);
generic operator >> (const generic& x1, const generic& x2);
generic lshiftz (const generic& x1, const generic& x2);
generic rshiftz (const generic& x1, const generic& x2);
generic operator | (const generic& x1, const generic& x2);
generic operator & (const generic& x1, const generic& x2);
generic lres (const generic& x1, const generic& x2);
generic rres (const generic& x1, const generic& x2);
generic uniform (const generic& x1, const generic& x2);
generic specialize (const generic& x1, const generic& x2);

} // namespace mmx
#endif // __GENERIC_HPP
