
/******************************************************************************
* MODULE     : operators.hpp
* DESCRIPTION: Templates and default implementations of several operations
* COPYRIGHT  : (C) 2003  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_OPERATORS_HPP
#define __MMX_OPERATORS_HPP
#include <basix/type_props.hpp>
#include <basix/identifiers.hpp>

//! \file

namespace mmx {

extern nat mmx_series_cancel_order;

/******************************************************************************
* Refer to standard C++ operators by another name
******************************************************************************/

#define TMPL template<typename T>

TMPL T neg_operator (const T& x) { return -x; }
TMPL T add_operator (const T& x, const T& y) { return x+y; }
TMPL T sub_operator (const T& x, const T& y) { return x-y; }
TMPL T mul_operator (const T& x, const T& y) { return x*y; }
TMPL T div_operator (const T& x, const T& y) { return x/y; }

TMPL bool equal_operator (const T& x, const T& y) { return x==y; }
TMPL bool unequal_operator (const T& x, const T& y) { return x!=y; }
TMPL bool less_operator (const T& x, const T& y) { return x<y; }
TMPL bool lesseq_operator (const T& x, const T& y) { return x<=y; }
TMPL bool gtr_operator (const T& x, const T& y) { return x>y; }
TMPL bool gtreq_operator (const T& x, const T& y) { return x>=y; }

#undef TMPL

/******************************************************************************
* Default versions of in place operators
******************************************************************************/

template<typename R> inline void
neg (R& x) { x= -x; }
template<typename R, typename C1> inline void
neg (R& x, const C1& y1) { x= -y1; }
template<typename R, typename C1, typename C2> inline void
add (R& x, const C1& y1, const C2& y2) { x= y1 + y2; }
template<typename R, typename C1, typename C2> inline void
sub (R& x, const C1& y1, const C2& y2) { x= y1 - y2; }
template<typename R, typename C1, typename C2> inline void
mul (R& x, const C1& y1, const C2& y2) { x= y1 * y2; }
template<typename R, typename C1, typename C2> inline void
div (R& x, const C1& y1, const C2& y2) { x= y1 / y2; }
template<typename R, typename C1, typename C2> inline void
mul_add (R& x, const C1& y1, const C2& y2) { x += y1 * y2; }
template<typename R, typename C1, typename C2> inline void
mul_sub (R& x, const C1& y1, const C2& y2) { x -= y1 * y2; }

/******************************************************************************
* Consider arbitrary routines as operators
******************************************************************************/

template<typename C, C (*fun) (const C&)>
struct unary_op {
  static inline C
  op (const C& x) { return fun (x); }
};

template<typename C, C (*fun) (const C&, const C&)>
struct binary_op {
  static inline C
  op (const C& x, const C& y) { return fun (x, y); }
};

/******************************************************************************
* Specification of default return types
******************************************************************************/

template<typename Op, typename C>
struct unary_return_type_helper {
  typedef C R;
};

template<typename Op, typename C1, typename C2>
struct binary_return_type_helper {
  typedef C1 R;
};

#define Unary_return_type(Op,C) \
  typename unary_return_type_helper<Op,C>::R
#define Binary_return_type(Op,C1,C2) \
  typename binary_return_type_helper<Op,C1,C2>::R

/******************************************************************************
* Specify the return type for operators which perform a cast
******************************************************************************/

template<typename Op, typename R>
struct replace_return_op {
  static inline R
  op () { return Op::template op<R> (); }
  template<typename X1> static inline R
  op (const X1& x1) { return Op::template op<R> (x1); }
  template<typename X1, typename X2> static inline R
  op (const X1& x1, const X2& x2) { return Op::template op<R> (x1, x2); }
  static inline void
  in_place (R& r) { Op::in_place (r); }
  template<typename X1> static inline void
  in_place (R& r, const X1& x1) { Op::in_place (r, x1); }
  template<typename X1, typename X2> static inline void
  in_place (R& r, const X1& x1, const X2& x2) { Op::in_place (r, x1, x2); }
};

template<typename Op, typename R>
struct return_op_helper {
  typedef Op Rop;
};

#define REPLACEABLE_RETURN(Op)         \
template<typename R>                   \
struct return_op_helper<Op,R> {        \
  typedef replace_return_op<Op,R> Rop; \
};

#define Return_op(Op,R) \
  typename return_op_helper<Op,R>::Rop

/******************************************************************************
* Get the in-place variant of an operator
******************************************************************************/

template<typename Op>
struct in_place_op {
  template<typename R> static inline void
  op (R& r) { (void) Op::in_place (r); }
  template<typename R, typename S1> static inline void
  op (R& r, const S1& s1) { (void) Op::in_place (r, s1); }
  template<typename R, typename S1, typename S2> static inline void
  op (R& r, const S1& s1, const S2& s2) { (void) Op::in_place (r, s1, s2); }
  static inline syntactic
  op (const syntactic& s1) {
    return apply (Op::in_place_name (), s1); }
  static inline syntactic
  op (const syntactic& s1, const syntactic& s2) {
    return apply (Op::in_place_name (), s1, s2); }
  static inline syntactic
  op (const syntactic& s1, const syntactic& s2, const syntactic& s3) {
    return apply (Op::in_place_name (), s1, s2, s3); }  
};

/******************************************************************************
* Constructors and copying
******************************************************************************/

struct id_op {
  static generic name () { return GEN_IDENTITY; }
  static generic in_place_name () { return GEN_ASSIGN; }
  template<typename C> static inline C
  op (const C& x) { return x; }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= y; }
};

struct as_op {
  template<typename R, typename C> static inline R
  op (const C& x) { return as<R> (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= as<R> (y); }
};

REPLACEABLE_RETURN(as_op);

struct copy_op {
  template<typename C> static inline C
  op (const C& x) { return copy (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= copy (y); }
  static inline generic
  op (const generic& x) { return x; }
  static inline syntactic
  op (const syntactic& x) { return x; }
  template<typename R> static inline void
  in_place (R& x, const generic& y) { x= y; }
  template<typename R> static inline void
  in_place (R& x, const syntactic& y) { x= y; }
};

struct duplicate_op {
  template<typename C> static inline C
  op (const C& x) { return duplicate (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= duplicate (y); }
};

struct fast_op {
  static generic name () { return GEN_FAST; }
  template<typename C> static inline Fast_type(C)
  op (const C& x) { return fast (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= fast (y); }
};

struct slow_op {
  static generic name () { return GEN_SLOW; }
  template<typename C> static inline C
  op (const Fast_type(C)& x) { return slow (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= slow (y); }
};

struct swap_op {
  template<typename C> static inline void
  in_place (C& x, C& y) { C aux; x= y; y= aux; }
};

struct flatten_op {
  template<typename C> static inline syntactic
  op (const C& x) { return flatten (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= flatten (y); }
};

template<typename C>
struct unary_return_type_helper<flatten_op,C> {
  typedef syntactic R;
};

/******************************************************************************
* Constants
******************************************************************************/

struct log2_as_op {
  template<typename R> static inline R
  op () { return Log2 (R); }
  template<typename R> static inline void
  in_place (R& x) { x= Log2 (R); }
};

struct pi_as_op {
  template<typename R> static inline R
  op () { return Pi (R); }
  template<typename R> static inline void
  in_place (R& x) { x= Pi (R); }
};

struct euler_as_op {
  template<typename R> static inline R
  op () { return Euler (R); }
  template<typename R> static inline void
  in_place (R& x) { x= Euler (R); }
};

struct catalan_as_op {
  template<typename R> static inline R
  op () { return Catalan (R); }
  template<typename R> static inline void
  in_place (R& x) { x= Catalan (R); }
};

struct imaginary_as_op {
  template<typename R> static inline R
  op () { return Imaginary (R); }
  template<typename R> static inline void
  in_place (R& x) { x= Imaginary (R); }
};

struct infinity_as_op {
  template<typename R> static inline R
  op () { return Infinity (R); }
  template<typename R> static inline void
  in_place (R& x) { x= Infinity (R); }
};

struct plus_infinity_as_op {
  template<typename R> static inline R
  op () { return Plus_infinity (R); }
  template<typename R> static inline void
  in_place (R& x) { x= Plus_infinity (R); }
};

struct minus_infinity_as_op {
  template<typename R> static inline R
  op () { return Minus_infinity (R); }
  template<typename R> static inline void
  in_place (R& x) { x= Minus_infinity (R); }
};

REPLACEABLE_RETURN(log2_as_op);
REPLACEABLE_RETURN(pi_as_op);
REPLACEABLE_RETURN(euler_as_op);
REPLACEABLE_RETURN(catalan_as_op);
REPLACEABLE_RETURN(imaginary_as_op);
REPLACEABLE_RETURN(infinity_as_op);
REPLACEABLE_RETURN(plus_infinity_as_op);
REPLACEABLE_RETURN(minus_infinity_as_op);

/******************************************************************************
* Hashing
******************************************************************************/

struct hash_op {
  static generic name () { return "hash"; }
  template<typename C> static inline nat
  op (const C& x) { return hash (x); }
};

struct soft_hash_op {
  static generic name () { return "soft_hash"; }
  template<typename C> static inline nat
  op (const C& x) { return soft_hash (x); }
};

struct hard_hash_op {
  static generic name () { return "hard_hash"; }
  template<typename C> static inline nat
  op (const C& x) { return hard_hash (x); }
};

/******************************************************************************
* Arithmetic operators
******************************************************************************/

struct neg_op {
  static generic name () { return GEN_MINUS; }
  template<typename C> static inline C
  op (const C& x) { return -x; }
  template<typename C> static inline void
  in_place (C& x) { neg (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { neg (x, y); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x) {
    (void) me; return -derive (x); }
};

struct add_op {
  static generic name () { return GEN_PLUS; }
  static generic in_place_name () { return GEN_PLUS_ASSIGN; }
  template<typename C> static inline C
  neutral () { return 0; }
  template<typename C> static inline C
  op (const C& x, const C& y) { return x + y; }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x += y; }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y1, const C2& y2) { add (x, y1, y2); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x, const C& y) {
    (void) me; return derive (x) + derive (y); }
};

struct radd_op {
  static generic name () { return GEN_PLUS; }
  template<typename C,typename S> static inline C
  op (const C& x, const S& y) { return x + y; }
  template<typename C,typename S> static inline void
  in_place (C& x, const S& y) { x += y; }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y1, const C2& y2) { add (x, y1, y2); }
};

struct ladd_op {
  static generic name () { return GEN_PLUS; }
  template<typename C,typename S> static inline C
  op (const C& y, const S& x) { return x + y; }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y2, const C2& y1) { add (x, y1, y2); }
};

struct sub_op {
  static generic name () { return GEN_MINUS; }
  static generic in_place_name () { return GEN_MINUS_ASSIGN; }
  template<typename C> static inline C
  op (const C& x, const C& y) { return x - y; }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x -= y; }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y1, const C2& y2) { sub (x, y1, y2); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x, const C& y) {
    (void) me; return derive (x) - derive (y); }
};

struct rsub_op {
  static generic name () { return GEN_MINUS; }
  template<typename C,typename S> static inline C
  op (const C& x, const S& y) { return x - y; }
  template<typename C,typename S> static inline void
  in_place (C& x, const S& y) { x -= y; }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y1, const C2& y2) { sub (x, y1, y2); }
};

struct lsub_op {
  static generic name () { return GEN_MINUS; }
  template<typename C,typename S> static inline C
  op (const C& y, const S& x) { return x - y; }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y2, const C2& y1) { sub (x, y1, y2); }
};

struct lmul_op;
struct rmul_op;
struct mul_add_op;

struct mul_op {
  typedef lmul_op sc_left;
  typedef rmul_op sc_right;
  typedef id_op set_op;
  typedef mul_add_op acc_op;
  static generic name () { return GEN_TIMES; }
  static generic in_place_name () { return GEN_TIMES_ASSIGN; }
  template<typename C> static inline C
  neutral () { return 1; }
  template<typename C> static inline C
  op (const C& x, const C& y) { return x * y; }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x *= y; }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y1, const C2& y2) { mul (x, y1, y2); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x, const C& y) {
    (void) me; return derive (x)*y + x*derive (y); }
};

struct rmul_op {
  static generic name () { return GEN_TIMES; }
  template<typename C,typename S> static inline C
  op (const C& x, const S& y) { return x * y; }
  template<typename C,typename S> static inline void
  in_place (C& x, const S& y) { x *= y; }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y1, const C2& y2) { mul (x, y1, y2); }
  template<typename C,typename S> static inline C
  diff_op (const C& me, const C& x, const S& y) {
    (void) me; return derive (x) * y; }
};

struct lmul_op {
  static generic name () { return GEN_TIMES; }
  template<typename C,typename S> static inline C
  op (const C& y, const S& x) { return x * y; }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y2, const C2& y1) { mul (x, y1, y2); }
  template<typename C,typename S> static inline C
  diff_op (const C& me, const C& y, const S& x) {
    (void) me; return x * derive (y); }
};

struct mul_add_op {
  typedef add_op set_op;
  typedef mul_op rhs_op;
  typedef mul_add_op acc_op;
  static generic name () { return GEN_MUL_ADD; }
  template<typename C> static inline C
  neutral () { return 0; }
  template<typename C> static inline C
  op (const C& x, const C& y, const C& z) { return x + y * z; }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y, const C2& z) { mul_add (x, y, z); }
};

struct rmul_add_op {
  static generic name () { return GEN_MUL_ADD; }
  template<typename R, typename C,typename S> static inline void
  in_place (R& x, const C& y, const S& z) { mul_add (x, y, z); }
};

struct lmul_add_op {
  static generic name () { return GEN_MUL_ADD; }
  template<typename R, typename C,typename S> static inline void
  in_place (R& x, const C& z, const S& y) { mul_add (x, y, z); }
};

struct mul_sub_op {
  static generic name () { return GEN_MUL_SUB; }
  template<typename C> static inline C
  op (const C& x, const C& y, const C& z) { return x - y * z; }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y, const C2& z) { mul_sub (x, y, z); }
};

struct square_op {
  static generic name () { return GEN_SQUARE; }
  template<typename C> static inline C
  op (const C& x) { return square (x); }
  template<typename R> static inline void
  in_place (R& x) { x= square (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= square (y); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x) {
    (void) me; return 2*derive (x)*x; }
};

struct graeffe_op {
  static generic name () { return GEN_GRAEFFE; }
  template<typename C> static inline C
  op (const C& x) { return graeffe (x); }
  template<typename R> static inline void
  in_place (R& x) { x= graeffe (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= graeffe (y); }
};

struct ldiv_op;
struct rdiv_op;

struct div_op {
  typedef ldiv_op sc_left;
  typedef rdiv_op sc_right;
  static generic name () { return GEN_OVER; }
  static generic in_place_name () { return GEN_OVER_ASSIGN; }
  template<typename C> static inline C
  op (const C& x, const C& y) { return x / y; }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x /= y; }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y1, const C2& y2) { div (x, y1, y2); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x, const C& y) {
    (void) me; return (derive (x)*y - x*derive (y)) / square (y); }
  static inline nat nr_init () { return 0; }
  template<typename C> static inline C
  def (const C& me, const C& f, const C& g) {
    if (g[0] == 0) {
      nat val;
      for (val=0; g[val]==0; val++)
	ASSERT (val != mmx_series_cancel_order,
		"valuation of denominator too high");
      return rshiftz (f, (int) val) / rshiftz (g, (int) val);
    }
    return (f - lshiftz (me * rshiftz (g))) / g[0]; }
};

struct rdiv_op {
  static generic name () { return GEN_OVER; }
  template<typename C,typename S> static inline C
  op (const C& x, const S& y) { return x / y; }
  template<typename C,typename S> static inline void
  in_place (C& x, const S& y) { x /= y; }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y1, const C2& y2) { div (x, y1, y2); }
  template<typename C,typename S> static inline C
  diff_op (const C& me, const C& x, const S& y) {
    (void) me; return derive (x) / y; }
};

struct ldiv_op {
  static generic name () { return GEN_OVER; }
  template<typename C,typename S> static inline C
  op (const C& y, const S& x) { return x / y; }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y2, const C2& y1) { div (x, y1, y2); }
  template<typename C,typename S> static inline C
  diff_op (const C& me, const C& y, const S& x) {
    (void) me; return -x * derive (y) / square (y); }
};

struct quo_op {
  static generic name () { return GEN_DIV; }
  template<typename C> static inline C
  op (const C& x, const C& y) { return quo (x, y); }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y1, const C2& y2) { x= quo (y1, y2); }
  static inline nat nr_init () { return 0; }
  template<typename C> static inline C
  def (const C& me, const C& f, const C& g) {
    if (g[0] == 0) {
      nat val;
      for (val=0; g[val]==0; val++)
	if (val == mmx_series_cancel_order) return 0;
      return quo (rshiftz (f, (int) val), rshiftz (g, (int) val));
    }
    return quo (f - lshiftz (me * rshiftz (g)), g[0]); }
};

struct rquo_op {
  static generic name () { return GEN_DIV; }
  template<typename C,typename S> static inline C
  op (const C& x, const S& y) { return quo (x, y); }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y1, const C2& y2) { x= quo (y1, y2); }
};

struct rem_op {
  static generic name () { return GEN_MOD; }
  template<typename C> static inline C
  op (const C& x, const C& y) { return rem (x, y); }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y1, const C2& y2) { x= rem (y1, y2); }
};

struct rrem_op {
  static generic name () { return GEN_MOD; }
  template<typename C,typename S> static inline C
  op (const C& x, const S& y) { return rem (x, y); }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y1, const C2& y2) { x= rem (y1, y2); }
};

struct lshift2_op {
  template<typename C,typename S> static inline C
  op (const C& x, const S& sh) { return lshift2 (x, sh); }
  template<typename C,typename S> static inline void
  in_place (C& x, const S& sh) { lshift2_assign (x, sh); }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y1, const C2& y2) { lshift2 (x, y1, y2); }
};

struct rshift2_op {
  template<typename C,typename S> static inline C
  op (const C& x, const S& sh) { return rshift2 (x, sh); }
  template<typename C,typename S> static inline void
  in_place (C& x, const S& sh) { rshift2_assign (x, sh); }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y1, const C2& y2) { rshift2 (x, y1, y2); }
};

struct lshiftz_op {
  template<typename C,typename S> static inline C
  op (const C& x, const S& sh) { return lshiftz (x, sh); }
  template<typename C,typename S> static inline void
  in_place (C& x, const S& sh) { lshiftz_assign (x, sh); }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y1, const C2& y2) { lshiftz (x, y1, y2); }
};

struct rshiftz_op {
  template<typename C,typename S> static inline C
  op (const C& x, const S& sh) { return rshiftz (x, sh); }
  template<typename C,typename S> static inline void
  in_place (C& x, const S& sh) { rshiftz_assign (x, sh); }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y1, const C2& y2) { rshiftz (x, y1, y2); }
};

struct incexp2_op {
  template<typename C,typename S> static inline C
  op (const C& x, const S& sh) { return incexp2 (x, sh); }
  template<typename C,typename S> static inline void
  in_place (C& x, const S& sh) { incexp2_assign (x, sh); }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y1, const C2& y2) { incexp2 (x, y1, y2); }
};

struct decexp2_op {
  template<typename C,typename S> static inline C
  op (const C& x, const S& sh) { return decexp2 (x, sh); }
  template<typename C,typename S> static inline void
  in_place (C& x, const S& sh) { decexp2_assign (x, sh); }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y1, const C2& y2) { decexp2 (x, y1, y2); }
};

struct or_op {
  template<typename C> static inline C
  neutral () { return false; }
  template<typename C> static inline C
  op (const C& x, const C& y) { return x | y; }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x |= y; }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y1, const C2& y2) { x= y1 | y2; }
};

struct and_op {
  template<typename C> static inline C
  neutral () { return true; }
  template<typename C> static inline C
  op (const C& x, const C& y) { return x & y; }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x &= y; }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y1, const C2& y2) { x= y1 & y2; }
};

/******************************************************************************
* Extra arithmetic operators
******************************************************************************/

struct gcd_op {
  static generic name () { return GEN_GCD; }
  template<typename C> static inline C
  neutral () { return C(0); }
  template<typename C> static inline C
  op (const C& x, const C& y) { return gcd (x, y); }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y1, const C2& y2) { x= gcd (y1, y2); }
};

struct lcm_op {
  static generic name () { return GEN_LCM; }
  template<typename C> static inline C
  neutral () { return C(1); }
  template<typename C> static inline C
  op (const C& x, const C& y) { return lcm (x, y); }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y1, const C2& y2) { x= lcm (y1, y2); }
};

template<typename C> struct xgcd_matrix;

struct xgcd_op {
  template<typename C> static inline xgcd_matrix<C>
  op (const C& x, const C& y) { return xgcd (x, y); }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y1, const C2& y2) { x= xgcd (y1, y2); }
};

/******************************************************************************
* Predicates
******************************************************************************/

struct hard_eq_op { // hard equality in memory
  static generic name () { return GEN_HARD_EQ; }
  template<typename C> static inline bool
  op (const C& x, const C& y) { return hard_eq (x, y); }
  static inline syntactic
  op (const syntactic& x, const syntactic& y) { return apply (name(), x, y); }
  template<typename C> static inline bool
  not_op (const C& x, const C& y) { return hard_neq (x, y); }
  template<typename C> static inline nat
  hash_op (const C& x) { return hard_hash (x); }
};

struct hard_neq_op { // hard inequality in memory
  static generic name () { return GEN_HARD_NEQ; }
  template<typename C> static inline bool
  op (const C& x, const C& y) { return hard_neq (x, y); }
  static inline syntactic
  op (const syntactic& x, const syntactic& y) { return apply (name(), x, y); }
  template<typename C> static inline bool
  not_op (const C& x, const C& y) { return hard_eq (x, y); }
  template<typename C> static inline nat
  hash_op (const C& x) { return hard_hash (x); }
};

struct hard_less_op { // hard address comparaison
  static generic name () { return "hard_less"; }
  template<typename C> static inline bool
  op (const C& x, const C& y) { return hard_less (x, y); }
};

struct hard_gtr_op { // hard address comparaison
  static generic name () { return "hard_less"; }
  template<typename C> static inline bool
  op (const C& x, const C& y) { return hard_gtr (x, y); }
};

struct eq_op { // syntactic equality
  static generic name () { return GEN_EQ; }
  template<typename C> static inline bool
  op (const C& x, const C& y) { return eq (x, y); }
  static inline syntactic
  op (const syntactic& x, const syntactic& y) { return apply (name(), x, y); }
  template<typename C> static inline bool
  not_op (const C& x, const C& y) { return neq (x, y); }
  template<typename C> static inline nat
  hash_op (const C& x) { return soft_hash (x); }
};

struct neq_op { // syntactic inequality
  static generic name () { return GEN_NEQ; }
  template<typename C> static inline bool
  op (const C& x, const C& y) { return neq (x, y); }
  static inline syntactic
  op (const syntactic& x, const syntactic& y) { return apply (name(), x, y); }
  template<typename C> static inline bool
  not_op (const C& x, const C& y) { return eq (x, y); }
  template<typename C> static inline nat
  hash_op (const C& x) { return soft_hash (x); }
};

struct equal_op {
  static generic name () { return GEN_EQUAL; }
  template<typename C> static inline bool
  op (const C& x, const C& y) { return x == y; }
  static inline syntactic
  op (const syntactic& x, const syntactic& y) { return apply (name(), x, y); }
  template<typename C> static inline bool
  not_op (const C& x, const C& y) { return x != y; }
  template<typename C> static inline nat
  hash_op (const C& x) { return hash (x); }
};

struct unequal_op {
  static generic name () { return GEN_UNEQUAL; }
  template<typename C> static inline bool
  op (const C& x, const C& y) { return x != y; }
  static inline syntactic
  op (const syntactic& x, const syntactic& y) { return apply (name(), x, y); }
  template<typename C> static inline bool
  not_op (const C& x, const C& y) { return x == y; }
  template<typename C> static inline nat
  hash_op (const C& x) { return hash (x); }
};

struct less_op {
  static generic name () { return GEN_LESS; }
  template<typename C> static inline bool
  op (const C& x, const C& y) { return x < y; }
  static inline syntactic
  op (const syntactic& x, const syntactic& y) { return apply (name(), x, y); }
  template<typename C> static inline bool
  not_op (const C& x, const C& y) { return !(x < y); }
};

struct lesseq_op {
  static generic name () { return GEN_LESSEQ; }
  template<typename C> static inline bool
  op (const C& x, const C& y) { return x <= y; }
  static inline syntactic
  op (const syntactic& x, const syntactic& y) { return apply (name(), x, y); }
  template<typename C> static inline bool
  not_op (const C& x, const C& y) { return !(x <= y); }
};

struct gtr_op {
  static generic name () { return GEN_GTR; }
  template<typename C> static inline bool
  op (const C& x, const C& y) { return x > y; }
  static inline syntactic
  op (const syntactic& x, const syntactic& y) { return apply (name(), x, y); }
  template<typename C> static inline bool
  not_op (const C& x, const C& y) { return !(x > y); }
};

struct gtreq_op {
  static generic name () { return GEN_GTREQ; }
  template<typename C> static inline bool
  op (const C& x, const C& y) { return x >= y; }
  static inline syntactic
  op (const syntactic& x, const syntactic& y) { return apply (name(), x, y); }
  template<typename C> static inline bool
  not_op (const C& x, const C& y) { return !(x >= y); }
};

/******************************************************************************
* Ordering related operators
******************************************************************************/

struct min_op {
  static generic name () { return GEN_MIN; }
  template<typename C> static inline C
  neutral () { return Plus_infinity (C); }
  template<typename C> static inline C
  op (const C& x, const C& y) { return min (x, y); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= min (x, y); }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y1, const C2& y2) { x= min (y1, y2); }
};

struct max_op {
  static generic name () { return GEN_MAX; }
  template<typename C> static inline C
  neutral () { return Minus_infinity (C); }
  template<typename C> static inline C
  op (const C& x, const C& y) { return max (x, y); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= max (x, y); }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y1, const C2& y2) { x= max (y1, y2); }
};

struct inf_op {
  static generic name () { return GEN_INF; }
  template<typename C> static inline C
  neutral () { return Plus_infinity (C); }
  template<typename C> static inline C
  op (const C& x, const C& y) { return inf (x, y); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= inf (x, y); }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y1, const C2& y2) { x= inf (y1, y2); }
};

struct sup_op {
  static generic name () { return GEN_SUP; }
  template<typename C> static inline C
  neutral () { return Minus_infinity (C); }
  template<typename C> static inline C
  op (const C& x, const C& y) { return sup (x, y); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= sup (x, y); }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y1, const C2& y2) { x= sup (y1, y2); }
};

struct abs_op {
  static generic name () { return GEN_ABS; }
  template<typename C> static inline Abs_type(C)
  op (const C& x) { return abs (x); }
  static inline generic
  op (const generic& x) { return abs (x); }
  static inline syntactic
  op (const syntactic& x) { return abs (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= abs (y); }
};

template<typename C>
struct unary_return_type_helper<abs_op,C> {
  typedef Abs_type(C) R;
};

struct abs_as_op {
  static generic name () { return GEN_ABS; }
  template<typename R, typename C> static inline R
  op (const C& x) { return as<R> (abs (x)); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= as<R> (abs (y)); }
};

REPLACEABLE_RETURN(abs_as_op);

struct sign_op {
  static generic name () { return GEN_SIGN; }
  template<typename C> static inline int
  op (const C& x) { return sign (x); }
  static inline generic
  op (const generic& x) { return gen (GEN_SIGN, x); }
  static inline syntactic
  op (const syntactic& x) { return apply (GEN_SIGN, x); }
  template<typename C> static inline void
  in_place (int& x, const C& y) { x= sign (y); }
};

/******************************************************************************
* Elementary functions
******************************************************************************/

struct sqrt_op {
  static generic name () { return GEN_SQRT; }
  template<typename C> static inline C
  op (const C& x) { return sqrt (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= sqrt (y); }
  template<typename C, typename I> static inline C
  op_init (const C& x, const I& i) { return sqrt_init (x, i); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x) {
    return derive (x) / (2 * me); }
  static inline nat nr_init () { return 1; }
  template<typename C> static inline C
  def (const C& me, const C& f) {
    if (f[0] == 0) {
      nat val= 0;
      for (val=0; val<mmx_series_cancel_order && f[val]==0; val++);
      ASSERT (val<mmx_series_cancel_order, "valuation too high");
      ASSERT ((val&1) == 0, "square root is not a series");
      return lshiftz (sqrt (rshiftz (f, (int) val)), (int) (val/2));
    }
    return integrate (derive (f) / (2*me)); }
};

struct cbrt_op {
  static generic name () { return "cbrt"; }
  template<typename C> static inline C
  op (const C& x) { return cbrt (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= cbrt (y); }
};

struct hypot_op {
  static generic name () { return "hypot"; }
  template<typename C> static inline C
  op (const C& x, const C& y) { return hypot (x, y); }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y1, const C2& y2) { x= hypot (y1, y2); }
};

struct exp_op {
  static generic name () { return GEN_EXP; }
  template<typename C> static inline C
  op (const C& x) { return exp (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= exp (y); }
  template<typename C, typename I> static inline C
  op_init (const C& x, const I& i) { (void) i; return exp (x); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x) {
    return derive (x) * me; }
  static inline nat nr_init () { return 1; }
  template<typename C> static inline C
  def (const C& me, const C& f) {
    return integrate (diff_op (me, f)); }
};

struct exp2_op {
  static generic name () { return "exp2"; }
  template<typename C> static inline C
  op (const C& x) { return exp2 (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= exp2 (y); }
};

struct exp10_op {
  static generic name () { return "exp10"; }
  template<typename C> static inline C
  op (const C& x) { return exp10 (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= exp10 (y); }
};

struct log_op {
  static generic name () { return GEN_LOG; }
  template<typename C> static inline C
  op (const C& x) { return log (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= log (y); }
  template<typename C,typename I> static inline C
  op_init (const C& x, const I& i) { return log_init (x, i); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x) {
    (void) me; return derive (x) / x; }
  static inline nat nr_init () { return 1; }
  template<typename C> static inline C
  def (const C& me, const C& f) {
    return integrate (diff_op (me, f)); }
};

struct log2_op {
  static generic name () { return "log2"; }
  template<typename C> static inline C
  op (const C& x) { return log2 (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= log2 (y); }
};

struct log10_op {
  static generic name () { return "log10"; }
  template<typename C> static inline C
  op (const C& x) { return log10 (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= log10 (y); }
};

struct pow_op {
  static generic name () { return GEN_POWER; }
  template<typename C> static inline C
  op (const C& x, const C& y) { return pow (x, y); }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y1, const C2& y2) { x= pow (y1, y2); }
};

struct trig_op {
  template<typename C> static inline C
  op (const C& x) { return trig (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= trig (y); }
  template<typename C, typename I> static inline C
  op_init (const C& x, const I& i) { (void) i; return trig (x); }
  static inline generic
  op (const generic& x) { return xsqtuple (comma (cos (x), sin (x))); }
  static inline syntactic
  op (const syntactic& x) { return as_syntactic (op (as_generic (x))); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x);
  static inline nat nr_init () { return 1; }
  template<typename V,typename C> static inline V
  def (const V& me, const C& f) {
    return integrate (diff_op (me, f)); }
};

struct cos_op {
  static generic name () { return GEN_COS; }
  template<typename C> static inline C
  op (const C& x) { return cos (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= cos (y); }
  template<typename C, typename I> static inline C
  op_init (const C& x, const I& i) { (void) i; return cos (x); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x) {
    (void) me; return -derive (x) * sin (x); }
  static inline nat nr_init () { return 1; }
  template<typename C> static inline C
  def (const C& me, const C& f) {
    (void) me; return cos_sin (f) [0]; }
};

struct sin_op {
  static generic name () { return GEN_SIN; }
  template<typename C> static inline C
  op (const C& x) { return sin (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= sin (y); }
  template<typename C, typename I> static inline C
  op_init (const C& x, const I& i) { (void) i; return sin (x); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x) {
    (void) me; return derive (x) * cos (x); }
  static inline nat nr_init () { return 1; }
  template<typename C> static inline C
  def (const C& me, const C& f) {
    (void) me; return cos_sin (f) [1]; }
};

struct tan_op {
  static generic name () { return GEN_TAN; }
  template<typename C> static inline C
  op (const C& x) { return tan (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= tan (y); }
  template<typename C, typename I> static inline C
  op_init (const C& x, const I& i) { (void) i; return tan (x); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x) {
    (void) me; return derive (x) / square (cos (x)); }
  static inline nat nr_init () { return 1; }
  template<typename C> static inline C
  def (const C& me, const C& f);
};

struct acos_op {
  static generic name () { return GEN_ARCCOS; }
  template<typename C> static inline C
  op (const C& x) { return acos (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= acos (y); }
  template<typename C, typename I> static inline C
  op_init (const C& x, const I& i) { return acos_init (x, i); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x) {
    (void) me; return -derive (x) / sqrt (1 - square (x)); }
  static inline nat nr_init () { return 1; }
  template<typename C> static inline C
  def (const C& me, const C& f) {
    return integrate (diff_op (me, f)); }
};

struct asin_op {
  static generic name () { return GEN_ARCSIN; }
  template<typename C> static inline C
  op (const C& x) { return asin (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= asin (y); }
  template<typename C, typename I> static inline C
  op_init (const C& x, const I& i) { return asin_init (x, i); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x) {
    (void) me; return derive (x) / sqrt (1 - square (x)); }
  static inline nat nr_init () { return 1; }
  template<typename C> static inline C
  def (const C& me, const C& f) {
    return integrate (diff_op (me, f)); }
};

struct atan_op {
  static generic name () { return GEN_ARCTAN; }
  template<typename C> static inline C
  op (const C& x) { return atan (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= atan (y); }
  template<typename C, typename I> static inline C
  op_init (const C& x, const I& i) { return atan_init (x, i); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x) {
    (void) me; return derive (x) / (square (x) + 1); }
  static inline nat nr_init () { return 1; }
  template<typename C> static inline C
  def (const C& me, const C& f) {
    return integrate (diff_op (me, f)); }
};

struct atan2_op {
  static generic name () { return "arctan2"; }
  template<typename C> static inline C
  op (const C& x, const C& y) { return atan2 (x, y); }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y1, const C2& y2) { x= atan2 (y1, y2); }
};

struct cosh_op {
  static generic name () { return GEN_CH; }
  template<typename C> static inline C
  op (const C& x) { return cosh (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= cosh (y); }
  template<typename C, typename I> static inline C
  op_init (const C& x, const I& i) { (void) i; return cosh (x); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x) {
    (void) me; return derive (x) * sinh (x); }
};

struct sinh_op {
  static generic name () { return GEN_SH; }
  template<typename C> static inline C
  op (const C& x) { return sinh (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= sinh (y); }
  template<typename C, typename I> static inline C
  op_init (const C& x, const I& i) { (void) i; return sinh (x); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x) {
    (void) me; return derive (x) * cosh (x); }
};

struct tanh_op {
  static generic name () { return GEN_TH; }
  template<typename C> static inline C
  op (const C& x) { return tanh (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= tanh (y); }
  template<typename C, typename I> static inline C
  op_init (const C& x, const I& i) { (void) i; return tanh (x); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x) {
    (void) me; return derive (x) / square (cosh (x)); }
};

struct acosh_op {
  static generic name () { return GEN_ARGCH; }
  template<typename C> static inline C
  op (const C& x) { return acosh (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= acosh (y); }
  template<typename C, typename I> static inline C
  op_init (const C& x, const I& i) { return acosh_init (x, i); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x) {
    (void) me; return derive (x) / sqrt (square (x) - 1); }
};

struct asinh_op {
  static generic name () { return GEN_ARGSH; }
  template<typename C> static inline C
  op (const C& x) { return asinh (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= asinh (y); }
  template<typename C, typename I> static inline C
  op_init (const C& x, const I& i) { return asinh_init (x, i); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x) {
    (void) me; return derive (x) / sqrt (square (x) + 1); }
};

struct atanh_op {
  static generic name () { return GEN_ARGTH; }
  template<typename C> static inline C
  op (const C& x) { return atanh (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= atanh (y); }
  template<typename C, typename I> static inline C
  op_init (const C& x, const I& i) { return atanh_init (x, i); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x) {
    (void) me; return derive (x) / (1 - square (x)); }
};

/******************************************************************************
* Special functions
******************************************************************************/

struct gamma_op {
  static generic name () { return GEN_GAMMA; }
  template<typename C> static inline C
  op (const C& x) { return gamma (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= gamma (y); }
};

struct zeta_op {
  static generic name () { return GEN_ZETA; }
  template<typename C> static inline C
  op (const C& x) { return zeta (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= zeta (y); }
};

struct erf_op {
  static generic name () { return GEN_ERF; }
  template<typename C> static inline C
  op (const C& x) { return erf (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= erf (y); }
};

/******************************************************************************
* Special operations for approximate arithmetic and reliable computations
******************************************************************************/

struct is_number_op {
  template<typename T> static inline bool
  op (const T& x) { return is_number (x); }
  template<typename T> static inline bool
  not_op (const T& x) { return !is_number (x); }
};

struct sharpen_op {
  // Make an imprecise number precise,
  // e.g. replace the ball B(c,r) by B(c,0)
  static generic name () { return GEN_SHARPEN; }
  template<typename C> static inline C
  op (const C& x) { return sharpen (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= sharpen (y); }
};

struct rnd_error_op {
  // When the number x is the result of a floating point computation,
  // rnd_error (x) is an upper bound for the distance between x and
  // the corresponding exact computation. For instance, for real x,
  // we may take rnd_error (x) = max (next_above x - x, x - next_below x).
  static generic name () { return "rnd_error"; }
  template<typename C> static inline Abs_type(C)
  op (const C& x) { return rnd_error (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= as<R> (rnd_error (y)); }
};

/******************************************************************************
* Operators on complex number like types
******************************************************************************/

struct gaussian_op {
  template<typename C> static inline Complex_type(C)
  op (const C& x, const C& y) { return gaussian (x, y); }
  static inline generic
  op (const generic& x, const generic& y) { return gaussian (x, y); }
  static inline syntactic
  op (const syntactic& x, const syntactic& y) { return gaussian (x, y); }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y1, const C2& y2) { x= gaussian (y1, y2); }
};

struct polar_op {
  template<typename C> static inline Complex_type(C)
  op (const C& x, const C& y) { return polar (x, y); }
  static inline generic
  op (const generic& x, const generic& y) { return polar (x, y); }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y1, const C2& y2) { x= polar (y1, y2); }
};

struct Re_op {
  static generic name () { return GEN_RE; }
  template<typename C> static inline Real_type(C)
  op (const C& z) { return Re (z); }
  static inline generic
  op (const generic& z) { return Re (z); }
  static inline syntactic
  op (const syntactic& z) { return Re (z); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= Re (y); }
};

struct Im_op {
  static generic name () { return GEN_IM; }
  template<typename C> static inline Real_type(C)
  op (const C& z) { return Im (z); }
  static inline generic
  op (const generic& z) { return Im (z); }
  static inline syntactic
  op (const syntactic& z) { return Im (z); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= Im (y); }
};

struct arg_op {
  static generic name () { return GEN_ARG; }
  template<typename C> static inline Real_type(C)
  op (const C& z) { return arg (z); }
  static inline generic
  op (const generic& z) { return arg (z); }
  static inline syntactic
  op (const syntactic& z) { return arg (z); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= arg (y); }
};

struct conj_op {
  static generic name () { return GEN_CONJ; }
  template<typename C> static inline C
  op (const C& z) { return conj (z); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= conj (y); }
};

/******************************************************************************
* Operators on ball like types
******************************************************************************/

struct ball_op {
  static generic name () { return GEN_BALL; }
  template<typename B> static inline B
  op (const Center_type(B)& c, const Radius_type(B)& r) {
    return make_ball<B> (c, r); }
  template<typename B, typename C, typename R> static inline void
  in_place (B& b, const C& c, const R& r) {
    b= make_ball<B> (c, r); }
};

struct center_op {
  static generic name () { return GEN_CENTER; }
  template<typename C> static inline Center_type(C)
  op (const C& x) { return center (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= center (y); }
};

struct radius_op {
  static generic name () { return GEN_RADIUS; }
  template<typename C> static inline Radius_type(C)
  op (const C& x) { return radius (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= radius (y); }
};

/******************************************************************************
* Calculus
******************************************************************************/

struct derive_op {
  static generic name () { return GEN_DERIVE; }
  template<typename C> static inline C
  op (const C& x) { return derive (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= derive (y); }
  template<typename C,typename V> static inline C
  op (const C& x, const V& v) { return derive (x, v); }
  template<typename R, typename C, typename V> static inline void
  in_place (R& x, const C& y, const V& v) { x= derive (y, v); }
};

struct xderive_op {
  static generic name () { return GEN_XDERIVE; }
  template<typename C> static inline C
  op (const C& x) { return xderive (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { return xderive (y); }
  template<typename C,typename V> static inline C
  op (const C& x, const V& v) { return xderive (x, v); }
  template<typename R, typename C, typename V> static inline void
  in_place (R& x, const C& y, const V& v) { x= xderive (y, v); }
};

struct integrate_op {
  static generic name () { return GEN_INTEGRATE; }
  template<typename C> static inline C
  op (const C& x) { return integrate (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= integrate (y); }
  template<typename C,typename V> static inline C
  op (const C& x, const V& v) { return integrate (x, v); }
  template<typename R, typename C, typename V> static inline void
  in_place (R& x, const C& y, const V& v) { x= integrate (y, v); }
  template<typename C,typename I> static inline C
  op_init (const C& x, const I& i) { return integrate_init (x, i); }
  template<typename C,typename V,typename I> static inline C
  op_init (const C& x, const V& v, const I& i) {
    return integrate_init (x, v, i); }
};

struct compose_op {
  static generic name () { return GEN_COMPOSE; }
  template<typename C1,typename C2> static inline C1
  op (const C1& x, const C2& y) { return compose (x, y); }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y1, const C2& y2) { x= compose (y1, y2); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x, const C& y) {
    (void) me; return derive (y) * compose (derive (x), y); }
};

struct reverse_op {
  static generic name () { return GEN_REVERSE; }
  template<typename C> static inline C
  op (const C& x) { return reverse (x); }
  template<typename R, typename C> static inline void
  in_place (R& x, const C& y) { x= reverse (y); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x) {
    return 1 / compose (derive (x), me); }
};

struct append_op {
  static generic name () { return GEN_APPEND; }
  template<typename C1,typename C2> static inline C1
  op (const C1& x, const C2& y) { return append (x, y); }
  template<typename R, typename C1, typename C2> static inline void
  in_place (R& x, const C1& y1, const C2& y2) { x= append (y1, y2); }
};

struct solve_vector_lde_op {
  template<typename C> static inline C
  op (const C& x) { ERROR ("must initialize (solve_vector_lde_op::op)"); }
  template<typename C, typename I> static inline C
  op_init (const C& x, const I& i) { return solve_vector_lde_init (x, i); }
  static inline generic op_init (const generic& x, const generic& i) { 
    return solve_lde_init (x, i); }
  static inline syntactic op_init (const syntactic& x, const syntactic& i) { 
    return solve_lde_init (x, i); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x);
  static inline nat nr_init () { return 1; }
  template<typename C> static inline C
  def (const C& me, const C& f) {
    return integrate (diff_op (me, f)); }
};

struct solve_matrix_lde_op {
  template<typename C> static inline C
  op (const C& x) { ERROR ("must initialize (solve_matrix_lde_op::op)"); }
  template<typename C, typename I> static inline C
  op_init (const C& x, const I& i) { return solve_matrix_lde_init (x, i); }
  static inline generic op_init (const generic& x, const generic& i) { 
    return solve_lde_init (x, i); }
  static inline syntactic op_init (const syntactic& x, const syntactic& i) { 
    return solve_lde_init (x, i); }
  template<typename C> static inline C diff_op (const C& me, const C& x) {
    return x * me; }
  static inline nat nr_init () { return 1; }
  template<typename C> static inline C def (const C& me, const C& f) {
    return integrate (diff_op (me, f)); }
};

/******************************************************************************
* Asymptotic relations
******************************************************************************/

struct dominant_part_op {
  template<typename T> static inline T
  op (const T& x) { return dominant_part (x); }
};

struct remainder_part_op {
  template<typename T> static inline T
  op (const T& x) { return remainder_part (x); }
};

struct infinitesimal_part_op {
  template<typename T> static inline T
  op (const T& x) { return infinitesimal_part (x); }
};

struct bounded_part_op {
  template<typename T> static inline T
  op (const T& x) { return bounded_part (x); }
};

struct infinite_part_op {
  template<typename T> static inline T
  op (const T& x) { return infinite_part (x); }
};

struct constant_part_op {
  template<typename T> static inline T
  op (const T& x) { return constant_part (x); }
};

struct non_constant_part_op {
  template<typename T> static inline T
  op (const T& x) { return non_constant_part (x); }
};

struct is_infinitesimal_op {
  template<typename T> static inline bool
  op (const T& x) { return x << 1; }
  template<typename T> static inline bool
  not_op (const T& x) { return !(x << 1); }
};

struct is_bounded_op {
  template<typename T> static inline bool
  op (const T& x) { return x <<= 1; }
  template<typename T> static inline bool
  not_op (const T& x) { return !(x <<= 1); }
};

struct is_infinite_op {
  template<typename T> static inline bool
  op (const T& x) { return x >> 1; }
  template<typename T> static inline bool
  not_op (const T& x) { return !(x >> 1); }
};

struct is_constant_op {
  template<typename T> static inline bool
  op (const T& x) { return x == 1; }
  template<typename T> static inline bool
  not_op (const T& x) { return x != 1; }
};

struct is_non_constant_op {
  template<typename T> static inline bool
  op (const T& x) { return x != 1; }
  template<typename T> static inline bool
  not_op (const T& x) { return x == 1; }
};

struct smaller_equal_op {
  template<typename T> static inline bool
  op (const T& x1, const T& x2) { return x1 << x2; }
  template<typename T> static inline bool
  not_op (const T& x1, const T& x2) { return !(x1 << x2); }
};

struct smaller_op {
  template<typename T> static inline bool
  op (const T& x1, const T& x2) { return x1 <<= x2; }
  template<typename T> static inline bool
  not_op (const T& x1, const T& x2) { return !(x1 <<= x2); }
};

struct larger_equal_op {
  template<typename T> static inline bool
  op (const T& x1, const T& x2) { return x1 >>= x2; }
  template<typename T> static inline bool
  not_op (const T& x1, const T& x2) { return !(x1 >>= x2); }
};

struct larger_op {
  template<typename T> static inline bool
  op (const T& x1, const T& x2) { return x1 >> x2; }
  template<typename T> static inline bool
  not_op (const T& x1, const T& x2) { return !(x1 >> x2); }
};

} // namespace mmx
#endif // __MMX_OPERATORS_HPP
