
/******************************************************************************
* MODULE     : program_basic.hpp
* DESCRIPTION: Programs which can be constructed and evaluated fast
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__PROGRAM_BASIC__HPP
#define __MMX__PROGRAM_BASIC__HPP
#include <basix/dag.hpp>

namespace mmx {
template<typename V> struct program_dag {};
#define Program dag<C,MT>
#define Program_A1 dag<A1,MT>
#define Program_A2 dag<A2,MT>
#define Program_A3 dag<A3,MT>

/******************************************************************************
* Abstract program dags
******************************************************************************/

template<typename C, typename V>
struct abstract_dag<C,program_dag<V> >:
  public abstract_dag<C,typename V::PT >
{
  virtual C Evaluate () const = 0;
  // NOTE: we would have liked to pass an environment parameter,
  // with a type E determined by V. However, it seems impossible
  // to take E = Void in C++. Since our main applications use
  // an implicit environment, directly implemented by V,
  // we therefore decided to go for the fast...
};

template<typename C, typename V> inline C
evaluate (const dag<C,program_dag<V> >& d) {
  return d->Evaluate ();
}

/******************************************************************************
* Concrete program dags
******************************************************************************/

template<typename Op, typename C, typename V, typename MT>
struct nullary_dag<Op,C,program_dag<V>,MT >:
  public V::template nullary_program<Op,C>
{
  inline nullary_dag ():
    V::template nullary_program<Op,C> () {}
};

template<typename Op, typename C, typename A1, typename V, typename MT>
struct unary_dag<Op,C,A1,program_dag<V>,MT >:
  public V::template unary_program<Op,C,A1>
{
  inline unary_dag (const A1& a1):
    V::template unary_program<Op,C,A1> (a1) {}
};

template<typename Op, typename C, typename A1, typename A2,
	 typename V, typename MT>
struct binary_dag<Op,C,A1,A2,program_dag<V>,MT >:
  public V::template binary_program<Op,C,A1,A2>
{
  inline binary_dag (const A1& a1, const A2& a2):
    V::template binary_program<Op,C,A1,A2> (a1, a2) {}
};

template<typename Op, typename C, typename A1, typename A2, typename A3,
	 typename V, typename MT>
struct ternary_dag<Op,C,A1,A2,A3,program_dag<V>,MT >:
  public V::template ternary_program<Op,C,A1,A2,A3>
{
  inline ternary_dag (const A1& a1, const A2& a2, const A3& a3):
    V::template ternary_program<Op,C,A1,A2,A3> (a1, a2, a3) {}
};

/******************************************************************************
* program_basic variant with default implementations of operations
******************************************************************************/

template<typename PT2, typename MT2>
struct program_basic {
  typedef PT2 PT;
  typedef MT2 MT;

template<typename Op, typename C>
struct nullary_program: public nullary_dag<Op,C,PT,MT> {
  inline nullary_program ():
    nullary_dag<Op,C,PT,MT> () {}
  inline C Evaluate () const {
    return Op::template op<C> (); }
};

template<typename Op, typename C, typename A1>
struct unary_program: unary_dag<Op,C,A1,PT,MT> {
  inline unary_program (const A1& a1):
    unary_dag<Op,C,A1,PT,MT> (a1) {}
  inline C Evaluate () const {
    return Op::op (this->arg1); }
};

template<typename Op, typename C, typename A1>
struct unary_program<Op,C,Program_A1 >:
  unary_dag<Op,C,Program_A1,PT,MT>
{
  inline unary_program (const Program_A1& a1):
    unary_dag<Op,C,Program_A1,PT,MT> (a1) {}
  inline C Evaluate () const {
    return Op::op (evaluate (this->arg1)); }
};

template<typename Op, typename C, typename A1, typename A2>
struct binary_program: binary_dag<Op,C,A1,A2,PT,MT> {
  inline binary_program (const A1& a1, const A2& a2):
    binary_dag<Op,C,A1,A2,PT,MT> (a1, a2) {}
  inline C Evaluate () const {
    return Op::op (this->arg1, this->arg2); }
};

/*
// FIXME: why does this not work?
// It seems that the compiler does not handle partial specialization well.

template<typename Op, typename C, typename A1, typename A2>
struct binary_program<Op,C,A1,Program_A2 >:
  binary_dag<Op,C,A1,Program_A2,PT,MT>
{
  inline binary_program (const A1& a1, const Program_A2& a2):
    binary_dag<Op,C,A1,Program_A2,PT,MT> (a1, a2) {}
  inline C Evaluate () const {
    return Op::op (this->arg1, evaluate (this->arg2)); }
};

template<typename Op, typename C, typename A1, typename A2>
struct binary_program<Op,C,Program_A1,Program_A2 >:
  binary_dag<Op,C,Program_A1,Program_A2,PT,MT>
{
  inline binary_program (const Program_A1& a1, const Program_A2& a2):
    binary_dag<Op,C,Program_A1,Program_A2,PT,MT> (a1, a2) {}
  inline C Evaluate () const {
    A1 a1= evaluate (this->arg1);
    A2 a2= evaluate (this->arg2); // force evaluation order
    return Op::op (a1, a2); }
};
*/

template<typename Op, typename C, typename A1, typename A2, typename DT>
struct binary_program<Op,C,dag<A1,DT>,dag<A2,DT> >:
  binary_dag<Op,C,dag<A1,DT>,dag<A2,DT>,PT,MT>
{
  inline binary_program (const dag<A1,DT>& a1, const dag<A2,DT>& a2):
    binary_dag<Op,C,dag<A1,DT>,dag<A2,DT>,PT,MT> (a1, a2) {}
  inline C Evaluate () const {
    A1 a1= evaluate (this->arg1);
    A2 a2= evaluate (this->arg2); // force evaluation order
    return Op::op (a1, a2); }
};

template<typename Op, typename C, typename A1, typename A2, typename A3>
struct ternary_program: ternary_dag<Op,C,A1,A2,A3,PT,MT> {
  inline ternary_program (const A1& a1, const A2& a2, const A3& a3):
    ternary_dag<Op,C,A1,A2,A3,PT,MT> (a1, a2, a3) {}
  inline C Evaluate () const {
    return Op::op (this->arg1, this->arg2, this->arg3); }
};

template<typename Op, typename C, typename A1, typename A2, typename A3>
struct ternary_program<Op,C,A1,A2,Program_A3 >:
  ternary_dag<Op,C,A1,A2,Program_A3,PT,MT>
{
  inline ternary_program (const A1& a1, const A2& a2, const Program_A3& a3):
    ternary_dag<Op,C,A1,A2,Program_A3,PT,MT> (a1, a2, a3) {}
  inline C Evaluate () const {
    return Op::op (this->arg1, this->arg2, evaluate (this->arg3)); }
};

template<typename Op, typename C, typename A1, typename A2, typename A3>
struct ternary_program<Op,C,A1,Program_A2,Program_A3 >:
  ternary_dag<Op,C,A1,Program_A2,Program_A3,PT,MT>
{
  inline ternary_program
    (const A1& a1, const Program_A2& a2, const Program_A3& a3):
      ternary_dag<Op,C,A1,Program_A2,Program_A3,PT,MT>
        (a1, a2, a3) {}
  inline C Evaluate () const {
    A2 a2= evaluate (this->arg2);
    A3 a3= evaluate (this->arg3); // force evaluation order
    return Op::op (this->arg1, a2, a3); }
};

template<typename Op, typename C, typename A1, typename A2, typename A3>
struct ternary_program<Op,C,Program_A1,Program_A2,Program_A3 >:
  ternary_dag<Op,C,Program_A1,Program_A2,Program_A3,PT,MT>
{
  inline ternary_program
    (const Program_A1& a1, const Program_A2& a2, const Program_A3& a3):
      ternary_dag<Op,C,Program_A1,Program_A2,Program_A3,PT,MT>
        (a1, a2, a3) {}
  inline C Evaluate () const {
    A1 a1= evaluate (this->arg1);
    A2 a2= evaluate (this->arg2);
    A3 a3= evaluate (this->arg3); // force evaluation order
    return Op::op (a1, a2, a3); }
};

};

#undef Program
#undef Program_A1
#undef Program_A2
#undef Program_A3
} // namespace mmx
#endif // __MMX__PROGRAM_BASIC__HPP
