
/******************************************************************************
* MODULE     : program_control.hpp
* DESCRIPTION: Construction of control structures
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__PROGRAM_CONTROL__HPP
#define __MMX__PROGRAM_CONTROL__HPP
#include <basix/program_basic.hpp>

namespace mmx {
#define Program dag<C,MT>
#define Program_A1 dag<A1,MT>
#define Program_A2 dag<A2,MT>
#define Program_A3 dag<A3,MT>
#define Program_void dag<void,MT>
#define Program_generic dag<generic,MT>
struct compound;

template<typename V>
struct program_control {
  typedef typename V::PT  PT;
  typedef typename V::MT MT;

/******************************************************************************
* Wrapper
******************************************************************************/

template<typename Op, typename C>
struct nullary_program: public V::template nullary_program<Op,C> {
  inline nullary_program ():
    V::template nullary_program<Op,C> () {}
};

template<typename Op, typename C, typename A1>
struct unary_program: public V::template unary_program<Op,C,A1> {
  inline unary_program (const A1& a1):
    V::template unary_program<Op,C,A1> (a1) {}
};

template<typename Op, typename C, typename A1, typename A2>
struct binary_program: public V::template binary_program<Op,C,A1,A2> {
  inline binary_program (const A1& a1, const A2& a2):
    V::template binary_program<Op,C,A1,A2> (a1, a2) {}
};

template<typename Op, typename C, typename A1, typename A2, typename A3>
struct ternary_program: public V::template ternary_program<Op,C,A1,A2,A3> {
  inline ternary_program (const A1& a1, const A2& a2, const A3& a3):
    V::template ternary_program<Op,C,A1,A2,A3> (a1, a2, a3) {}
};

/******************************************************************************
* Sequential execution
******************************************************************************/

// FIXME: compiler does not allow to do a template<>,
// so we template by a dummy parameter
template<typename Dummy>
struct nullary_program<begin_op,Dummy>:
  public nullary_dag<begin_op,Dummy,PT,MT>
{
  inline nullary_program (): nullary_dag<begin_op,Dummy,PT,MT> () {}
  inline Dummy Evaluate () const {}
};

template<typename A1, typename A2>
struct binary_program<begin_op,A2,Program_A1,Program_A2>:
  public binary_dag<begin_op,A2,Program_A1,Program_A2,PT,MT>
{
  inline binary_program (const Program_A1& a1, const Program_A2& a2):
    binary_dag<begin_op,A2,Program_A1,Program_A2,PT,MT> (a1, a2) {}
  inline A2 Evaluate () const {
    (void) evaluate (this->arg1);
    return evaluate (this->arg2); }
};

/******************************************************************************
* Conditional programs
******************************************************************************/

template<typename A1, typename A2>
struct binary_program<if_op,void,Program_A1,Program_A2 >:
  public binary_dag<if_op,void,Program_A1,Program_A2,PT,MT>
{
  inline binary_program (const Program_A1& a1, const Program_A2& a2):
    binary_dag<if_op,void,Program_A1,Program_A2,PT,MT> (a1, a2) {}
  inline void Evaluate () const {
    if (as<bool> (evaluate (this->arg1))) (void) evaluate (this->arg2); }
};

template<typename C, typename A1>
struct ternary_program<if_op,C,Program_A1,Program,Program >:
  public ternary_dag<if_op,C,Program_A1,Program,Program,PT,MT>
{
  inline ternary_program
    (const Program_A1& a1, const Program& a2, const Program& a3):
      ternary_dag<if_op,C,Program_A1,Program,Program,PT,MT>
        (a1, a2, a3) {}
  inline C Evaluate () const {
    if (as<bool> (evaluate (this->arg1))) return evaluate (this->arg2);
    else return evaluate (this->arg3); }
};

/******************************************************************************
* Simple loops
******************************************************************************/

template<typename A1, typename A2>
struct binary_program<while_op,void,Program_A1,Program_A2 >:
  public binary_dag<while_op,void,Program_A1,Program_A2,PT,MT>
{
  inline binary_program (const Program_A1& a1, const Program_A2& a2):
    binary_dag<while_op,void,Program_A1,Program_A2,PT,MT> (a1, a2) {}
  inline void Evaluate () const {
    while (as<bool> (evaluate (this->arg1))) (void) evaluate (this->arg2); }
};

template<typename A1, typename A2, typename A3>
struct ternary_program<for_in_op<A1>,void,generic,Program_A2,Program_A3 >:
    public ternary_dag<for_in_op<A1>,void,generic,Program_A2,Program_A3,PT,MT>
{
  A1& var;
  inline ternary_program
    (const generic& a1, const Program_A2& a2, const Program_A3& a3):
      ternary_dag<for_in_op<A1>,void,generic,Program_A2,Program_A3,PT,MT>
        (a1, a2, a3),
      var (V::template reference<A1,generic> (a1)) {}
  inline void Evaluate () const {
    A1 old= var;
    generic g= evaluate (this->arg2);
    if (is<compound> (g) && N(g) == 2 && g[0] == "protect") g= g[1];
    iterator<A1> it= as<iterator<A1> > (g);
    for (; busy (it); ++it) {
      var= *it;
      evaluate (this->arg3);
    }
    var= old; }
};

/******************************************************************************
* Unary function application
******************************************************************************/

template<typename C, typename A1>
struct binary_program<unary_closure_op<A1>,C,generic,Program >:
  public binary_dag<unary_closure_op<A1>,C,generic,Program,PT,MT>
{
  A1& local1;
  A1& param1;
  inline binary_program (const generic& a1, const Program& a2):
    binary_dag<unary_closure_op<A1>,C,generic,Program,PT,MT> (a1, a2),
    local1 (V::template reference<A1,generic> (a1)),
    param1 (V::template reference<A1,int> (0)) {}
  inline C Evaluate () const {
    A1 old1= local1;
    local1= param1;
    C r= evaluate (this->arg2);
    local1= old1;
    return r;
  }
};

template<typename C, typename A1>
struct binary_program<unary_lambda_op<A1>,generic,generic,Program >:
  binary_dag<unary_lambda_op<A1>,generic,generic,Program,PT,MT>
{
  inline binary_program (const generic& a1, const Program& a2):
    binary_dag<unary_lambda_op<A1>,generic,generic,Program,PT,MT> (a1, a2) {}
  inline generic Evaluate () const {
    return as<generic> (unary_closure<A1> (this->arg1, this->arg2)); }
};

template<typename C, typename A1>
struct binary_program<unary_apply_op<A1>,C,Program_generic,Program_A1 >:
  public binary_dag<unary_apply_op<A1>,C,Program_generic,Program_A1,PT,MT>
{
  A1& param1;
  inline binary_program (const Program_generic& a1, const Program_A1& a2):
    binary_dag<unary_apply_op<A1>,C,Program_generic,Program_A1,PT,MT> (a1, a2),
    param1 (V::template reference<A1,int> (0)) {}
  inline C Evaluate () const {
    Program body= as<Program > (evaluate (this->arg1));
    param1= evaluate (this->arg2);
    return evaluate (body);
  }
};

/******************************************************************************
* N_ary function application
******************************************************************************/

template<typename C, typename A1>
struct binary_program<n_ary_closure_op<A1>,C,vector<generic>,Program >:
  public binary_dag<n_ary_closure_op<A1>,C,vector<generic>,Program,PT,MT>
{
  typedef A1* pointer;
  nat n;
  pointer* local;
  pointer* param;
  inline binary_program (const vector<generic>& a1, const Program& a2):
    binary_dag<n_ary_closure_op<A1>,C,vector<generic>,Program,PT,MT>
      (a1, a2),
    n (N (a1)),
    local (mmx_new<pointer> (n)),
    param (mmx_new<pointer> (n))
  {
    for (nat i=0; i<n; i++) {
      local[i]= &(V::template reference<A1,generic> (a1[i]));
      param[i]= &(V::template reference<A1,int> ((int) i));
    }
  }
  inline ~binary_program () {
    mmx_delete<pointer> (local, n);
    mmx_delete<pointer> (param, n);
  }
  inline C Evaluate () const {
    A1 old[n];
    for (nat i=0; i<n; i++) {
      old[i]= *(local[i]);
      *(local[i])= *(param[i]);
    }
    C r= evaluate (this->arg2);
    for (nat i=0; i<n; i++)
      *(local[i])= old[i];
    return r;
  }
};

template<typename C, typename A1>
struct binary_program<n_ary_lambda_op<A1>,generic,vector<generic>,Program >:
  binary_dag<n_ary_lambda_op<A1>,generic,vector<generic>,Program,PT,MT>
{
  inline binary_program (const vector<generic>& a1, const Program& a2):
    binary_dag<n_ary_lambda_op<A1>,generic,vector<generic>,Program,PT,MT>
      (a1, a2) {}
  inline generic Evaluate () const {
    return as<generic> (n_ary_closure<A1> (this->arg1, this->arg2)); }
};

template<typename C, typename A1>
struct binary_program<n_ary_apply_op<A1>,C,
		      Program_generic,vector<Program_A1 > >:
  public binary_dag<n_ary_apply_op<A1>,C,
		    Program_generic,vector<Program_A1 >,PT,MT>
{
  typedef A1* pointer;
  nat n;
  pointer* param;
  inline binary_program
    (const Program_generic& a1, const vector<Program_A1>& a2):
      binary_dag<n_ary_apply_op<A1>,C,
		 Program_generic,vector<Program_A1 >,PT,MT> (a1, a2),
      n (N (a2)),
      param (mmx_new<pointer> (n))
  {
    for (nat i=0; i<n; i++)
      param[i]= &(V::template reference<A1,int> ((int) i));
  }
  inline ~binary_program () {
    mmx_delete<pointer> (param, n); }
  inline C Evaluate () const {
    Program body= as<Program> (evaluate (this->arg1));
    for (nat i=0; i<n; i++)
      *(param[i])= evaluate (this->arg2[i]);
    return evaluate (body);
  }
};

};

#undef Program
#undef Program_A1
#undef Program_A2
#undef Program_A3
#undef Program_void
#undef Program_generic
} // namespace mmx
#endif // __MMX__PROGRAM_CONTROL__HPP
