
/******************************************************************************
* MODULE     : program_data.hpp
* DESCRIPTION: Data storage for programs
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__PROGRAM_DATA__HPP
#define __MMX__PROGRAM_DATA__HPP
#include <basix/program_basic.hpp>
#include <basix/table.hpp>

namespace mmx {
#define Program dag<C,MT>
#define Program_A1 dag<A1,MT>
#define Program_A2 dag<A2,MT>
#define Program_A3 dag<A3,MT>

/******************************************************************************
* The following indirect class is used to store persistent pointer
* We might to extend the class and put it in a separate file later
******************************************************************************/

template<typename C>
struct indirect_rep REP_STRUCT {
  C val;
  inline indirect_rep () {}
  inline indirect_rep (const C& val2): val (val2) {}
};

template<typename C>
struct indirect {
INDIRECT_PROTO_1 (indirect, indirect_rep, C)
  inline indirect ():
    rep (new indirect_rep<C> ()) {}
  inline indirect (const C& c1):
    rep (new indirect_rep<C> (c1)) {}
};
INDIRECT_NULL_IMPL_1 (indirect, indirect_rep, typename C, C)

/******************************************************************************
* The reference template provides persistent storage of typed variables
* in auxiliary memory. Variables are addressed by generics (user variables)
* or ints (parameters to function calls)
******************************************************************************/

template<typename V>
struct program_data: public V {
  typedef typename V::PT PT;
  typedef typename V::MT MT;

  template<typename C, typename Var> static C&
  reference (const Var& var) {
    static table<indirect<C>,Var> t;
    if (!t->contains (var)) t[var]= indirect<C> ();
    return inside (read (t, var)) -> val;
  }

/******************************************************************************
* Wrapper
******************************************************************************/

template<typename Op, typename C>
struct nullary_program: public V::template nullary_program<Op,C> {
  inline nullary_program ():
    V::template nullary_program<Op,C> () {}
};

template<typename Op, typename C, typename A1>
struct unary_program: public V::template unary_program<Op,C,A1> {
  inline unary_program (const A1& a1):
    V::template unary_program<Op,C,A1> (a1) {}
};

template<typename Op, typename C, typename A1, typename A2>
struct binary_program: public V::template binary_program<Op,C,A1,A2> {
  inline binary_program (const A1& a1, const A2& a2):
    V::template binary_program<Op,C,A1,A2> (a1, a2) {}
};

template<typename Op, typename C, typename A1, typename A2, typename A3>
struct ternary_program: public V::template ternary_program<Op,C,A1,A2,A3> {
  inline ternary_program (const A1& a1, const A2& a2, const A3& a3):
    V::template ternary_program<Op,C,A1,A2,A3> (a1, a2, a3) {}
};

/******************************************************************************
* Variable definitions and lookup
******************************************************************************/

template<typename A2>
struct binary_program<define_op,void,generic,Program_A2 >:
  public binary_dag<define_op,void,generic,Program_A2,PT,MT>
{
  A2& var;
  inline binary_program (const generic& a1, const Program_A2& a2):
    binary_dag<define_op,void,generic,Program_A2,PT,MT> (a1, a2),
    var (reference<A2,generic> (a1)) {}
  inline void Evaluate () const {
    var= evaluate (this->arg2); }
};

template<typename C>
struct unary_program<constant_lookup_op,C,generic>:
  public unary_dag<constant_lookup_op,C,generic,PT,MT>
{
  C& var;
  inline unary_program (const generic& a1):
    unary_dag<constant_lookup_op,C,generic,PT,MT> (a1),
    var (reference<C,generic> (a1)) {}
  inline C Evaluate () const { return var; }
};

template<typename C>
struct unary_program<mutable_lookup_op,C&,generic>:
  unary_dag<mutable_lookup_op,C&,generic,PT,MT>
{
  C& var;
  inline unary_program (const generic& a1):
    unary_dag<mutable_lookup_op,C&,generic,PT,MT> (a1),
    var (reference<C,generic> (a1)) {}
  inline C& Evaluate () const { return var; }
};

};

#undef Program
#undef Program_A1
#undef Program_A2
#undef Program_A3
} // namespace mmx
#endif // __MMX__PROGRAM_DATA__HPP
