
/******************************************************************************
* MODULE     : sparse_vector.hpp
* DESCRIPTION: Hash sparse_vectors
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_SPARSE_VECTOR_HPP
#define __MMX_SPARSE_VECTOR_HPP
#include <basix/pair.hpp>
#include <basix/symbol.hpp>

//! \file

namespace mmx {
struct eq_sparse_vector;
#define TMPL_DEF template<typename C, typename T, typename V= eq_sparse_vector>
#define TMPL template<typename C, typename T, typename V>
#define Sparse_vector sparse_vector<C,T,V>
#define Sparse_vector_rep sparse_vector_rep<C,T,V>
#define Pair pair<T,C>
TMPL class sparse_vector_rep;
TMPL class sparse_vector;
TMPL inline nat N (const Sparse_vector& v);

/******************************************************************************
* Variants of sparse vectors
******************************************************************************/

struct hard_eq_sparse_vector {
  typedef hard_eq_op key_op;
  typedef equal_op val_op;
  typedef hard_less_op l_op;
};

struct eq_sparse_vector {
  typedef eq_op key_op;
  typedef equal_op val_op;
  typedef less_op l_op;
};

/******************************************************************************
* Sparse_vector class and its representation class
******************************************************************************/

TMPL_DEF
class sparse_vector_rep REP_STRUCT {
  Pair* a;   // the entries
  nat   n;   // number of entries
  nat   l;   // allocated number of entries

public:
  inline sparse_vector_rep (Pair* a2, nat n2):
    a (a2), n (n2), l (n2) {}
  inline sparse_vector_rep (Pair* a2, nat n2, nat l2):
    a (a2), n (n2), l (l2) {}
  inline ~sparse_vector_rep () {
    mmx_delete<Pair > (a, l); }

  friend class Sparse_vector;
  friend nat N LESSGTR (const Sparse_vector& v);
};

TMPL_DEF
class sparse_vector {
INDIRECT_PROTO_3 (sparse_vector, sparse_vector_rep, C, T, V)
  inline sparse_vector () {
    rep= new Sparse_vector_rep (mmx_new<Pair > (0), 0); }
  inline sparse_vector (const T& k, const C& v) {
    typedef typename V::val_op Eq;
    if (Eq::op (v, C (0))) rep= new Sparse_vector_rep (mmx_new<Pair > (0), 0);
    else rep= new Sparse_vector_rep (mmx_new<Pair > (1, Pair (k, v)), 1); }
  inline sparse_vector (const Pair& p) {
    rep= new Sparse_vector_rep (mmx_new<Pair > (1, p), 1); }
  inline sparse_vector (Pair* a, nat n) {
    rep= new Sparse_vector_rep (a, n); }
  inline sparse_vector (Pair* a, nat n, nat l) {
    rep= new Sparse_vector_rep (a, n, l); }
  inline Pair operator [] (nat i) const {
    VERIFY (i<rep->n, "index out of range");
    return rep->a[i]; }
  friend nat N LESSGTR (const Sparse_vector& v);
};
INDIRECT_IMPL_3 (sparse_vector, sparse_vector_rep,
                 typename C, C, typename T, T, typename V, V)

TMPL inline nat N (const Sparse_vector& v) { return v.rep->n; }

/******************************************************************************
* Printing sparse_vectors
******************************************************************************/

TMPL syntactic
flatten (const Sparse_vector& v) {
  vector<syntactic> r;
  for (nat i=0; i<N(v); i++)
    r << flatten (v[i]);
  return apply ("sparse_vector", r);
}

PRINT_SUGAR(TMPL,Sparse_vector)

/******************************************************************************
* Hashing and equality testing
******************************************************************************/

template<typename Op, typename C, typename T, typename V> nat
unary_hash (const Sparse_vector& v) {
  nat h=1236321;
  for (nat i=0; i<N(v); i++)
    h = (h << 5) ^ (h >> 27) ^ Op::op (v[i]);
  return h;
}

template<typename Op, typename C, typename T, typename V> bool
binary_test (const Sparse_vector& v1, const Sparse_vector& v2) {
  nat n1= N(v1), n2= N(v2);
  if (n1 != n2) return false;
  for (nat i=0; i<n1; i++)
    if (!Op::op (v1[i], v2[i]))
      return false;
  return true;
}

TRUE_IDENTITY_OP_SUGAR(TMPL,Sparse_vector)
SOFT_IDENTITY_OP_SUGAR(TMPL,Sparse_vector)

/******************************************************************************
* Unary operations
******************************************************************************/

template<typename Op, typename C, typename T, typename V> Sparse_vector
unary_map (const Sparse_vector& v) {
  typedef typename V::val_op Eq;
  nat i, j, n= N(v);
  Pair* r= mmx_new<Pair > (n);
  for (i= j= 0; i < n; i++) {
    Pair e (v[i].x1, Op::op (v[i].x2));
    if (Eq::not_op (e.x2, C (0))) r[j++]= e;
  }
  return Sparse_vector (r, j, n);
}

TMPL inline Sparse_vector
operator - (const Sparse_vector& t) {
  return unary_map<neg_op,C,T,V> (t); }

/******************************************************************************
* Binary operations
******************************************************************************/

template<typename Op>
struct neutral_helper {};

template<>
struct neutral_helper<add_op> {
  typedef id_op lop;
  typedef id_op rop;
};

template<>
struct neutral_helper<sub_op> {
  typedef id_op lop;
  typedef neg_op rop;
};

template<typename Op, typename C, typename T, typename V> Sparse_vector
binary_map (const Sparse_vector& v1, const Sparse_vector& v2) {
  // With optimization which is only correct for operations # with 0 # 0 = 0
  typedef typename V::val_op Eq;
  typedef typename V::l_op Less;
  nat i1, i2, n1= N(v1), n2= N(v2), j, n= n1 + n2;
  Pair* r= mmx_new<Pair > (n);
  for (i1= i2= j= 0; i1 < n1 && i2 < n2; ) {
    if (Less::op (v1[i1].x1, v2[i2].x1)) {
      typedef typename neutral_helper<Op>::lop Lop;
      r[j++]= Pair (v1[i1].x1, Lop::op (v1[i1].x2)); i1++;
    }
    else if (Less::op (v2[i2].x1, v1[i1].x1)) {
      typedef typename neutral_helper<Op>::rop Rop;
      r[j++]= Pair (v2[i2].x1, Rop::op (v2[i2].x2)); i2++;
    }
    else {
      Pair e (v2[i2].x1, Op::op (v1[i1].x2, v2[i2].x2));
      i1++; i2++; if (Eq::not_op (e.x2, C (0))) r[j++]= e;
    }
  }
  while (i1 < n1) {
    typedef typename neutral_helper<Op>::lop Lop;
    r[j++]= Pair (v1[i1].x1, Lop::op (v1[i1].x2)); i1++;
  }
  while (i2 < n2) {
    typedef typename neutral_helper<Op>::rop Rop;
    r[j++]= Pair (v2[i2].x1, Rop::op (v2[i2].x2)); i2++;
  }
  return Sparse_vector (r, j, n);
}

TMPL inline Sparse_vector
operator + (const Sparse_vector& t, const Sparse_vector& u) {
  return binary_map<add_op,C,T,V> (t, u); }
TMPL inline Sparse_vector
operator - (const Sparse_vector& t, const Sparse_vector& u) {
  return binary_map<sub_op,C,T,V> (t, u); }

/******************************************************************************
* Scalar operations
******************************************************************************/

template<typename Op, typename C, typename T, typename V>
Sparse_vector
binary_map_scalar (const Sparse_vector& v, const C& x) {
  typedef typename V::val_op Eq;
  nat i, j, n= N(v);
  Pair* r= mmx_new<Pair > (n);
  for (i=0, j=0; i<n; i++) {
    Pair e (v[i].x1, Op::op (v[i].x2, x));
    if (Eq::not_op (e.x2, C (0))) r[j++]= e;
  }
  return Sparse_vector (r, j, n);
}

TMPL inline Sparse_vector
operator * (const Sparse_vector& t, const C& sc) {
  return binary_map_scalar<rmul_op> (t, sc); }
TMPL inline Sparse_vector
operator * (const C& sc, const Sparse_vector& t) {
  return binary_map_scalar<lmul_op> (t, sc); }
TMPL inline Sparse_vector
operator / (const Sparse_vector& t, const C& sc) {
  return binary_map_scalar<rdiv_op> (t, sc); }
TMPL inline Sparse_vector
operator / (const C& sc, const Sparse_vector& t) {
  return binary_map_scalar<ldiv_op> (t, sc); }

#undef TMPL_DEF
#undef TMPL
#undef Sparse_vector
#undef Sparse_vector_rep
#undef Pair
} // namespace mmx
#endif // __MMX_SPARSE_VECTOR_HPP
