
/******************************************************************************
* MODULE     : string.hpp
* DESCRIPTION: Strings with possible zero-characters
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __STRING_HPP
#define __STRING_HPP
#include <basix/type_props.hpp>
//! @file

namespace mmx {

/******************************************************************************
* The string class
******************************************************************************/

class string;
class string_rep REP_STRUCT {
private:
  char* a;  // the string
  nat   n;  // length of string
  nat   l;  // allocated number of bytes

public:
  inline string_rep (nat n2): a ((char*) mmx_malloc (n2)), n(n2), l(n2) {}
  inline ~string_rep () { mmx_free ((void*) a, l); }
  void resize (nat n);
  void extend (nat n);

  friend class string;
  friend inline nat N (const string& s);
  friend inline bool busy (const string& s);
  friend inline char* S (string& s);
  friend inline const char* S (const string& s);
  friend inline char read (const string& s, nat i);
  friend string copy (const string& s);
  friend string operator * (const string& s, const string& t);
  friend int as_int (const string& s);
  friend double as_double (const string& s);
};

class string {
INDIRECT_PROTO (string, string_rep)
public:
  inline string (nat n=0): rep (new string_rep (n)) {}
  string (char c);
  string (const char *s);
  string (const char *s, nat n);

  //! Return the length of @a s
  friend inline nat N (const string& s) { return s.rep->n; }
  //! Check whether @a s is non empty
  friend inline bool busy (const string& s) { return s.rep->n != 0; }
  //! Return a pointer to a copy of @a s
  friend inline char* S (string& s) { s.secure (); return s.rep->a; }
  //! Return a const pointer to @a s
  friend inline const char* S (const string& s) { return s.rep->a; }
  //! Return the @a i th character of @a s
  friend inline char read (const string& s, nat i) { return s.rep->a[i]; }
  //! Return the @a i th character
  inline char operator [] (nat i) const { return rep->a[i]; }
  inline char& operator [] (nat i) { secure (); return rep->a[i]; }
  //! Return the substring from position @a start to @a end
  string operator () (nat start, nat end) const;
  //! Append @a c to the end of the string
  string& operator << (char c);
  //! Append @a t to the end of the string
  string& operator << (const string& t);
  //! Retrieve @a c from the end of the string
  string& operator >> (char& c);
  bool operator == (const char* s) const;
  bool operator != (const char* s) const;
  bool operator == (const string& s) const;
  bool operator != (const string& s) const;
  friend nat hash (const string& s);
  friend string copy (const string& s);
  friend ostream& operator << (ostream& out, const string& s);
  friend string operator * (const string& s, const string& t);
  friend string as_string (void* ptr);
  friend string as_string (short int i);
  friend string as_string (short unsigned int i);
  friend string as_string (int i);
  friend string as_string (unsigned int i);
  friend string as_string (long int i);
  friend string as_string (long unsigned int i);
  friend string as_string (long long int i);
  friend string as_string (long long unsigned int i);
  friend string as_string (float x);
  friend string as_string (double x);
  friend int as_int (const string& s);
  friend double as_double (const string& s);
  friend string unescape (const string& s);

  friend class string_rep;
};
INDIRECT_IMPL (string, string_rep)

TRUE_TO_SOFT_IDENTITY_SUGAR(,string)
template<> inline string duplicate (const string& s) { return copy (s); }

/******************************************************************************
* Conversion routines
******************************************************************************/

string as_string (int i);
string as_string (unsigned int i);
string as_string (void* ptr);
inline string as_string (short int i) {
  return as_string ((int) i); }
inline string as_string (short unsigned int i) {
  return as_string ((unsigned int) i); }
inline string as_string (signed char i) {
  return as_string ((int) i); }
inline string as_string (unsigned char i) {
  return as_string ((unsigned int) i); }
string as_string (long int i);
string as_string (long unsigned int i);
string as_string (long long int i);
string as_string (long long unsigned int i);
string as_string (float x);
string as_string (double x);
int as_int (const string& s);
double as_double (const string& s);

/******************************************************************************
* Further routines on strings
******************************************************************************/

bool operator <  (const string& s, const string& t);
bool operator <= (const string& s, const string& t);
bool operator >  (const string& s, const string& t);
bool operator >= (const string& s, const string& t);
//! Return a pointer to a copy of @a s.
//! The allocated memory must be freed with free_charp.
char* as_charp (const string& s);
//! Free memory allocated within as_charp.
void free_charp (char* s);
//! Replace "\n", "\t", and "\b" respectively by "\\n", "\\t", and "\\b".
string escape (const string& s);
//! Replace "\\n", "\\t", and "\\b" respectively by "\n", "\t", and "\b".
string unescape (const string& s);
//! Add " around @a s.
string quote (const string& s);
//! Remove all the "..." and /"..."/ around @a s.
string unquote (const string& s);
//! Replace upcases by locases
string locase (const string& s);
//! Replace locases by upcases
string upcase (const string& s);
//! Replace upcase by locase for first letter
string locase_first (const string& s);
//! Replace locase by upcase for first letter
string upcase_first (const string& s);
//! Tell if @a s starts with @a what
bool starts (const string& s, const string& what);
//! Tell if @a s ends with @a what
bool ends (const string& s, const string& what);
//! Replace all the occurences of @a what by @by in @a s.
string replace (const string& s, const string& what, const string& by);
//! Search first occurrence of @what@ in @s@ after @pos@ or @-1@
int search_forwards (const string& s, const string& what, const int& pos);
//! Search first occurrence of @what@ in @s@ before @pos@ or @-1@
int search_backwards (const string& s, const string& what, const int& pos);
//! Reverse the order of the characters in @a s.
string reverse (const string& s);
//! Cut @s into pieces, using @c as a separator
vector<string> tokenize (const string& s, const string& sep, bool keep=false);
//! Reconstruct string from pieces in @v using @sep as a separator
string recompose (const vector<string>& v, const string& sep, bool last=true);

/******************************************************************************
* Routines for binary serialization
******************************************************************************/

struct serialize_string_helper {
  template<typename M> static void write (M& out, const string& x) {
    raw_write_length (out, N(x));
    raw_write (out, S(x), N(x)); }
  template<typename M> static void read (M& in, string& x) {
    nat n;
    raw_read_length (in, n);
    x= string (n);
    raw_read (in, S(x), n); }
};

STYPE_TO_TYPE(STMPL,binary_variant,string,serialize_string_helper);

template<typename C> string
into_binary (const C& c) {
  string s;
  binary_write (c);
  return s;
}

template<typename C> C
from_binary (const string& s) {
  C c;
  string r= reverse (s);
  binary_read (c);
  return c;
}

} // namespace mmx
#endif // __STRING_HPP
