
/******************************************************************************
* MODULE     : type_props.hpp
* DESCRIPTION: Associate types, variables and functions to types
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_TYPE_PROPS_HPP
#define __MMX_TYPE_PROPS_HPP
#include <basix/exception.hpp>
//! \file
namespace mmx {

/******************************************************************************
* Types associated to types
******************************************************************************/

#define TYPE_TO_TYPE(name,expr)			\
template<typename C>				\
struct name##_helper {				\
  typedef expr val;				\
};

#define STYPE_TO_TYPE(TMPL,name,C,expr)		\
TMPL						\
struct name##_helper<C > {			\
  typedef expr val;				\
};

class unknown_scalar_type {};
class unknown_real_type {};
//class unknown_numerator_type {};
//class unknown_denominator_type {};
class unknown_center_type {};
class unknown_radius_type {};
class unknown_as_vector_type {};
class unknown_as_matrix_type {};
template<typename C> class complex;

TYPE_TO_TYPE(scalar_type,C);
TYPE_TO_TYPE(real_type,unknown_real_type);
TYPE_TO_TYPE(abs_type,C);
//TYPE_TO_TYPE(numerator_type,unknown_numerator_type);
//TYPE_TO_TYPE(denominator_type,unknown_denominator_type);
TYPE_TO_TYPE(numerator_type,C);
TYPE_TO_TYPE(denominator_type,C);
TYPE_TO_TYPE(center_type,unknown_center_type);
TYPE_TO_TYPE(radius_type,unknown_radius_type);
TYPE_TO_TYPE(default_radius_type,C);
TYPE_TO_TYPE(user_radius_type,C);
TYPE_TO_TYPE(complex_type,complex<C>);
TYPE_TO_TYPE(as_vector_type,unknown_radius_type);
TYPE_TO_TYPE(as_matrix_type,unknown_radius_type);

#define Scalar_type(C) typename scalar_type_helper<C >::val
#define Real_type(C) typename real_type_helper<C >::val
#define Abs_type(C) typename abs_type_helper<C >::val
#define Numerator_type(C) typename numerator_type_helper<C >::val
#define Denominator_type(C) typename denominator_type_helper<C >::val
#define Center_type(C) typename center_type_helper<C >::val
#define Radius_type(C) typename radius_type_helper<C >::val
#define Default_radius_type(C) typename default_radius_type_helper<C >::val
#define User_radius_type(C) typename user_radius_type_helper<C >::val
#define Complex_type(C) typename complex_type_helper<C >::val
#define As_vector_type(C) typename as_vector_type_helper<C >::val
#define As_matrix_type(C) typename as_matrix_type_helper<C >::val

/******************************************************************************
* Constants associated to types
******************************************************************************/

template<typename C>
struct constant_helper {
  static inline C pi () {
    ERROR ("not implemented (pi)"); return 0; }
  static inline C log2 () {
    ERROR ("not implemented (log_2)"); return 0; }
  static inline C euler () {
    ERROR ("not implemented (euler)"); return 0; }
  static inline C catalan () {
    ERROR ("not implemented (catalan)"); return 0; }
  static inline C imaginary () {
    ERROR ("not implemented (imaginary)"); return 0; }
  static inline C nan () {
    ERROR ("not implemented (nan)"); return 0; }
  static inline C fuzz () {
    ERROR ("not implemented (fuzz)"); return 0; }
  static inline C epsilon () {
    ERROR ("not implemented (epsilon)"); return 0; }
  static inline C infinity () {
    ERROR ("not implemented (infinity)"); return 0; }
  static inline C plus_infinity () {
    ERROR ("not implemented (plus_infinity)"); return 0; }
  static inline C minus_infinity () {
    ERROR ("not implemented (linus_infinity)"); return 0; }
};

#define Pi(C) constant_helper<C >::pi()
#define Log2(C) constant_helper<C >::log2()
#define Euler(C) constant_helper<C >::euler()
#define Catalan(C) constant_helper<C >::catalan()
#define Imaginary(C) constant_helper<C >::imaginary()
#define Nan(C) constant_helper<C >::nan()
#define Fuzz(C) constant_helper<C >::fuzz()
#define Epsilon(C) constant_helper<C >::epsilon()
#define Infinity(C) constant_helper<C >::infinity()
#define Plus_infinity(C) constant_helper<C >::plus_infinity()
#define Minus_infinity(C) constant_helper<C >::minus_infinity()

/******************************************************************************
* Helpers for conversions
******************************************************************************/

#ifndef WITH_AS
#define WITH_AS
template<typename T,typename F>
struct as_helper {
  static inline T cv (const F& x) { return x; }
};

template<typename T,typename F> inline T
as (const F& x) {
  return as_helper<T,F>::cv (x);
}
#endif

template<typename T>
struct fast_helper {
  typedef T fast_type;
  static inline fast_type dd (const T& x) { return x; }
  static inline T uu (const fast_type& x) { return x; }
};

#define Fast_type(T) typename fast_helper<T>::fast_type

template<typename T> Fast_type(T)
fast (const T& x) {
  return fast_helper<T>::dd (x);
}

template<typename T> T
slow (const Fast_type(T)& x) {
  return fast_helper<T>::uu (x);
}

/******************************************************************************
* Helpers for accessing children
******************************************************************************/

template<typename C>
struct inspector {
  static inline nat length (const C& c) { (void) c; return 0; }
  static inline generic access (const C& c, nat i) {
    (void) c; (void) i;
    VERIFY (false, "out of range");
    return generic (); }
};

/******************************************************************************
* Helpers for binary data serialization
******************************************************************************/

template<typename M> void
raw_write (M& m, const char* p, nat n) {
  for (nat i=0; i<n; i++) m << p[i];
}

template<typename M> void
raw_read (M& m, const char* p, nat n) {
  for (nat i=0; i<n; i++) {
    VERIFY (busy (m), "read failed");
    m >> p[i];
  }
}

template<typename M> void
raw_write_length (M& m, nat l) {
  if (l < (nat) 255) m << ((char) l);
  else {
    m << ((char) 255);
    raw_write (m, (const char*) ((const void*) (&l)), sizeof (nat));
  }
}

template<typename M> void
raw_read_length (M& m, nat& l) {
  char c;
  m >> c;
  if (c != ((char) 255)) l= (nat) ((uchar) c);
  else raw_read (m, (char*) ((void)* (&l)), sizeof (nat));
}

template<typename C>
struct unknown_binary_helper {
  template<typename M> static inline void write (M& out, const C& x) {
    (void) out; (void) x;
    ERROR ("binary write not implemented"); }
  template<typename M> static inline void read (M& in, C& x) {
    (void) in; (void) x;
    ERROR ("binary read not implemented"); }
};

template<typename C>
struct default_binary_helper {
  template<typename M> static inline void write (M& out, const C& x) {
    raw_write (out, (const char*) ((const void*) (&x)), sizeof (C)); }
  template<typename M> static inline void read (M& in, C& x) {
    raw_read (in, (char*) ((void*) (&x)), sizeof (C)); }
};

TYPE_TO_TYPE(binary_variant,unknown_binary_helper<C>);
STYPE_TO_TYPE(STMPL,binary_variant,char,default_binary_helper<char>);
STYPE_TO_TYPE(STMPL,binary_variant,uchar,default_binary_helper<uchar>);
STYPE_TO_TYPE(STMPL,binary_variant,short,default_binary_helper<short>);
STYPE_TO_TYPE(STMPL,binary_variant,ushort,default_binary_helper<ushort>);
STYPE_TO_TYPE(STMPL,binary_variant,int,default_binary_helper<int>);
STYPE_TO_TYPE(STMPL,binary_variant,nat,default_binary_helper<nat>);
STYPE_TO_TYPE(STMPL,binary_variant,long,default_binary_helper<long>);
STYPE_TO_TYPE(STMPL,binary_variant,ulong,default_binary_helper<ulong>);
STYPE_TO_TYPE(STMPL,binary_variant,float,default_binary_helper<float>);
STYPE_TO_TYPE(STMPL,binary_variant,double,default_binary_helper<double>);

#define Binary_variant(C) typename binary_variant_helper<C >::val

template<typename M, typename C> void
binary_write (M& out, C& x) {
  Binary_variant(C)::write (out, x);
}

template<typename M, typename C> void
binary_read (M& in, C& x) {
  Binary_variant(C)::read (in, x);
}

/******************************************************************************
* Helpers for rounding
******************************************************************************/

enum mmx_rounding {
  MMX_ROUND_NEAR,
  MMX_ROUND_ZERO,
  MMX_ROUND_UP,
  MMX_ROUND_DOWN
};

struct rounded_none;
template<typename V> struct rounded_operations;

template<typename C>
struct rounding_helper {
  typedef int rounding_mode;
  template<mmx_rounding Mode, typename K= nat> struct translate {
    static const rounding_mode val= -1; };
  static inline rounding_mode get_rounding () { return -1; }
  static inline void set_rounding (rounding_mode r) { (void) r; }
  static const bool exact_type = true;
  typedef rounded_operations<rounded_none> UV;
  typedef rounded_operations<rounded_none> DV;
};

template<typename C, mmx_rounding Mode>
struct with_rounding {
  typedef rounding_helper<C> H;
  typedef typename H::rounding_mode R;
  R old;
  inline with_rounding () {
    old= H::get_rounding ();
    H::set_rounding (H::template translate<Mode>::val); }
  inline ~with_rounding () {
    H::set_rounding (old); }
};

#define Round_up(C) typename rounding_helper<C>::UV
#define Round_down(C) typename rounding_helper<C>::DV

/******************************************************************************
* Helpers for interval/ball arithmetic
******************************************************************************/

template<typename C,typename A>
struct make_interval_helper {
  static inline C val (const A& l, const A& r) {
    typedef Radius_type(C) R;
    typedef Round_up(A) Up;
    A cc= (l+r) >> 1;
    R rr= as<R> (Up::sub (r, l)) >> 1;
    return C (cc, Up::template add_as<R> (rr, rnd_error (cc)));
  }
};

template<typename C,typename A> C
make_interval (const A& l, const A& r) {
  return make_interval_helper<C,A>::val (l, r);
}

template<typename B,typename C,typename R>
struct make_ball_helper {
  static inline B val (const C& c, const R& r) {
    typedef Round_up(C) Up;
    typedef Round_down(C) Down;
    return B (Down::template sub_as<C> (c, r), Up::template add_as<C> (c, r));
  }
};

template<typename B,typename C,typename R> B
make_ball (const C& c, const R& r) {
  return make_ball_helper<B,C,R>::val (c, r);
}

} // namespace mmx
#endif // __MMX_TYPE_PROPS_HPP
