
/******************************************************************************
* MODULE     : vector_naive.hpp
* DESCRIPTION: header for low level vectorial computations
* COPYRIGHT  : (C) 2003  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__VECTOR_NAIVE__HPP
#define __MMX__VECTOR_NAIVE__HPP
#include <basix/exception.hpp>
#include <basix/operators.hpp>

//! \file

namespace mmx {
#define TMPL  template<typename C>
#define TMPLX template<typename C, typename X>

/******************************************************************************
* Abstract low level vector routines on vectors in memory
******************************************************************************/

struct vector_naive_abstractions {

typedef vector_naive_abstractions Aligned;

TMPL static inline nat
vec_aligned_size (nat n) { return n; }

template<typename Op, typename T> static inline void
vec_nullary (T* dest, nat n) {
  for (; n != 0; dest++, n--)
    Op::in_place (*dest);
}

template<typename Op, typename T, typename C> static inline void
vec_unary (T* dest, const C* s, nat n) {
  for (; n != 0; dest++, s++, n--)
    Op::in_place (*dest, *s);
}

template<typename Op, typename T, typename C> static inline void
vec_unary_stride (T* dest, nat dest_stride, const C* s, nat s_stride, nat n) {
  for (; n != 0; dest += dest_stride, s += s_stride, n--)
    Op::in_place (*dest, *s);
}

template<typename Op, typename T, typename C1, typename C2> static inline void
vec_binary (T* dest, const C1* s1, const C2* s2, nat n) {
  for (; n != 0; dest++, s1++, s2++, n--)
    Op::in_place (*dest, *s1, *s2);
}

template<typename Op, typename T, typename C1, typename C2> static inline void
vec_binary_stride (T* dest, nat dest_stride, const C1* s1, nat s1_stride,
		   const C2* s2, nat s2_stride, nat n) {
  for (; n != 0; dest += dest_stride, s1 += s1_stride, s2 += s2_stride, n--)
    Op::in_place (*dest, *s1, *s2);
}

template<typename Op, typename T, typename X> static inline void
vec_unary_scalar (T* dest, const X& x, nat n) {
  for (; n != 0; dest++, n--)
    Op::in_place (*dest, x);
}

template<typename Op, typename T, typename X> static inline void
vec_unary_scalar_stride (T* dest, nat dest_stride, const X& x, nat n) {
  for (; n != 0; dest += dest_stride, n--)
    Op::in_place (*dest, x);
}

template<typename Op, typename T, typename C, typename X> static inline void
vec_binary_scalar (T* dest, const C* s, const X& x, nat n) {
  for (; n != 0; dest++, s++, n--)
    Op::in_place (*dest, *s, x);
}

template<typename Op, typename C> static inline bool
vec_unary_test (const C* s, nat n) {
  for (; n != 0; s++, n--)
    if (Op::not_op (*s)) return false;
  return true;
}

template<typename Op, typename C1, typename C2> static inline bool
vec_binary_test (const C1* s1, const C2* s2, nat n) {
  for (; n != 0; s1++, s2++, n--)
    if (Op::not_op (*s1, *s2)) return false;
  return true;
}

template<typename Op, typename C, typename X> static inline bool
vec_binary_test_scalar (const C* s, const X& x, nat n) {
  for (; n != 0; s++, n--)
    if (Op::not_op (*s, x)) return false;
  return true;
}

template<typename Op, typename C, typename X> static inline bool
vec_binary_test_scalar_stride (const C* s, nat s_stride, const X& x, nat n) {
  for (; n != 0; s += s_stride, n--)
    if (Op::not_op (*s, x)) return false;
  return true;
}

template<typename Op, typename C> static inline C
vec_unary_big (const C* s, nat n) {
  if (n == 0) return Op::template neutral<C> ();
  C r= *s;
  for (s++, n--; n != 0; s++, n--)
    Op::in_place (r, *s);
  return r;
}

template<typename Op, typename C1, typename C2> static inline C1
vec_binary_big (const C1* s1, const C2* s2, nat n) {
  C1 r= Op::template neutral<C1> ();
  for (; n != 0; s1++, s2++, n--)
    Op::in_place (r, *s1, *s2);
  return r;
}

template<typename Op, typename C1, typename C2> static inline C1
vec_binary_big_stride (const C1* s1, nat p1, const C2* s2, nat p2, nat n) {
  C1 r= Op::template neutral<C1> ();
  for (; n != 0; s1 += p1, s2 += p2, n--)
    Op::in_place (r, *s1, *s2);
  return r;
}

template<typename Op, typename C> static C
vec_unary_big_dicho (const C* s, nat n) {
  switch (n) {
  case 0: return Op::template neutral<C> ();
  case 1: return s[0];
  case 2: return Op::op (s[0], s[1]);
  case 3: return Op::op (s[0], Op::op (s[1], s[2]));
  case 4: return Op::op (Op::op (s[0], s[1]), Op::op (s[2], s[3]));
  default: return Op::op (vec_unary_big_dicho<Op> (s, n>>1),
			  vec_unary_big_dicho<Op> (s + (n>>1), n - (n>>1)));
  }
}

TMPL static inline void
vec_reverse (C* dest, const C* src, nat n) {
  dest += n-1;
  for (; n != 0; dest--, src++, n--) *dest= *src;
}

TMPL static inline void
vec_reverse (C* dest, nat n) {
  nat h= n >> 1;
  C* rev= dest + n - 1;
  C tmp;
  for (; h != 0; dest++, rev--, h--) {
    tmp  = *dest;
    *dest= *rev;
    *rev = tmp;
  }
}

}; // vector_naive_abstractions

/******************************************************************************
* Instantiations
******************************************************************************/

template<typename V>
struct vector_enrich: public V {

typedef vector_enrich<typename V::Aligned> Aligned;

TMPLX static inline void set (C* dest, const X& c, nat n) {
  V::template vec_unary_scalar<id_op> (dest, c, n); }
TMPL static inline void clear (C* dest, nat n) {
  set<C,int> (dest, 0, n); }
TMPL static inline void copy (C* dest, const C* s, nat n) {
  V::template vec_unary<id_op> (dest, s, n); }
TMPLX static inline void cast (C* dest, const X* s, nat n) {
  V::template vec_unary<id_op> (dest, s, n); }
TMPL static inline void neg (C* dest, const C* s, nat n) {
  V::template vec_unary<neg_op> (dest, s, n); }
TMPL static inline void neg (C* dest, nat n) {
  V::template vec_nullary<neg_op> (dest, n); }
TMPL static inline void add (C* dest, const C* s1, const C* s2, nat n) {
  V::template vec_binary<add_op> (dest, s1, s2, n); }
TMPL static inline void sub (C* dest, const C* s1, const C* s2, nat n) {
  V::template vec_binary<sub_op> (dest, s1, s2, n); }
TMPL static inline void mul (C* dest, const C* s1, const C* s2, nat n) {
  V::template vec_binary<mul_op> (dest, s1, s2, n); }
TMPL static inline void div (C* dest, const C* s1, const C* s2, nat n) {
  V::template vec_binary<div_op> (dest, s1, s2, n); }
TMPL static inline void mul (C* dest, const C* s, const C& c, nat n) {
  V::template vec_binary_scalar<mul_op> (dest, s, c, n); }
TMPL static inline void div (C* dest, const C* s, const C& c, nat n) {
  V::template vec_binary_scalar<div_op> (dest, s, c, n); }
TMPL static inline void quo (C* dest, const C* s, const C& c, nat n) {
  V::template vec_binary_scalar<quo_op> (dest, s, c, n); }
TMPL static inline void rem (C* dest, const C* s, const C& c, nat n) {
  V::template vec_binary_scalar<rem_op> (dest, s, c, n); }
TMPL static inline void add (C* dest, const C* s, nat n) {
  V::template vec_unary<add_op> (dest, s, n); }
TMPL static inline void sub (C* dest, const C* s, nat n) {
  V::template vec_unary<sub_op> (dest, s, n); }
TMPL static inline void mul (C* dest, const C* s, nat n) {
  V::template vec_unary<mul_op> (dest, s, n); }
TMPL static inline void div (C* dest, const C* s, nat n) {
  V::template vec_unary<div_op> (dest, s, n); }
TMPL static inline void mul (C* dest, const C& c, nat n) {
  V::template vec_unary_scalar<mul_op> (dest, c, n); }
TMPL static inline void div (C* dest, const C& c, nat n) {
  V::template vec_unary_scalar<div_op> (dest, c, n); }
TMPL static inline void quo (C* dest, const C& c, nat n) {
  V::template vec_unary_scalar<quo_op> (dest, c, n); }
TMPL static inline void rem (C* dest, const C& c, nat n) {
  V::template vec_unary_scalar<rem_op> (dest, c, n); }
TMPL static inline void mul_add (C* dest, const C* s1, const C* s2, nat n) {
  V::template vec_binary<mul_add_op> (dest, s1, s2, n); }
TMPLX static inline void mul_add (C* dest, const C* s, const X& c, nat n) {
  V::template vec_binary_scalar<rmul_add_op> (dest, s, c, n); }
TMPLX static inline void mul_add (C* dest, const X& c, const C* s, nat n) {
  V::template vec_binary_scalar<lmul_add_op> (dest, s, c, n); }
TMPL static inline C inn_prod (const C* s1, const C* s2, nat n) {
  return V::template vec_binary_big<mul_add_op> (s1, s2, n); }
TMPL static inline bool test_number (const C* s, nat n) {
  return V::template vec_unary_test<is_number_op> (s, n); }
TMPL static inline bool equal (const C* s1, const C* s2, nat n) {
  return V::template vec_binary_test<equal_op> (s1, s2, n); }
TMPL static inline bool lesseq (const C* s1, const C* s2, nat n) {
  return V::template vec_binary_test<lesseq_op> (s1, s2, n); }
TMPL static inline bool gtreq (const C* s1, const C* s2, nat n) {
  return V::template vec_binary_test<gtreq_op> (s1, s2, n); }
TMPL static inline bool equal (const C* s, const C& c, nat n) {
  return V::template vec_binary_test_scalar<equal_op> (s, c, n); }
TMPL static inline bool lesseq (const C* s, const C& c, nat n) {
  return V::template vec_binary_test_scalar<lesseq_op> (s, c, n); }
TMPL static inline bool gtreq (const C* s, const C& c, nat n) {
  return V::template vec_binary_test_scalar<gtreq_op> (s, c, n); }
TMPL static inline void half_copy (C* dest, const C* s, nat l) {
  V::template vec_unary_stride<id_op> (dest, 1, s, 2, l); }
TMPL static inline void double_copy (C* dest, const C* s, nat l) {
  V::template vec_unary_stride<id_op> (dest, 2, s, 1, l); }
TMPL static inline void double_add (C* dest, const C* s, nat l) {
  V::template vec_unary_stride<add_op> (dest, 2, s, 1, l); }
TMPL static inline void triple_copy (C* dest, const C* s, nat l) {
  V::template vec_unary_stride<id_op> (dest, 3, s, 1, l); }

TMPL static inline void
print (ostream& out, const C* s, nat n) {
  if (n == 0) out << "[]";
  else {
    out << "[ " << s[0];
    for (nat i=1; i<n; i++)
      out << ", " << s[i];
    out << " ]";
  }
}

}; // vector_enrich<V>

/******************************************************************************
* Naive vectors
******************************************************************************/

struct vector_naive:
  public vector_enrich<vector_naive_abstractions>
{
  static const nat def_len = 0;
  static const nat init_len= 1;
};

#undef TMPL
#undef TMPLX
} // namespace mmx
#endif //__MMX__VECTOR_NAIVE__HPP
