
/******************************************************************************
* MODULE     : extrapolate.hpp
* DESCRIPTION: Asymptotic extrapolation
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__EXTRAPOLATE__HPP
#define __MMX__EXTRAPOLATE__HPP
#include <numerix/floating.hpp>
#include <numerix/complex.hpp>
#include <basix/vector.hpp>
#include <basix/vector_sort.hpp>
#include <basix/routine.hpp>
#include <algebramix/matrix.hpp>
namespace mmx {
#define TMPL template<typename R>
#define Vector vector<R>
#define Matrix matrix<R,matrix_naive>

// TODO: more generous acceptance for log-polynomials
// TODO: special detection of c n / log n
// TODO: check relative error for increasing values of L (= lhi (f))
// TODO: can we detect series in 1 / log n ?

/******************************************************************************
* Utility functions
******************************************************************************/

TMPL R
relative_error (const Vector& v, const Vector& w, const int& start) {
  ASSERT (start >= 0 && start < ((int) N(v)), "out of range");
  ASSERT (N(v) == N(w), "lengths don't match");
  nat i, n= N(v);
  R max_err= 0;
  R last   = v[n-1];
  for (i= ((nat) start); i<n; i++) {
    if (is_zero (v[i]) && is_zero (w[i])) continue;
    R err= 2 * abs (v[i] - w[i]) / (abs (v[i]) + abs (w[i]));
    if (err > max_err) max_err= err;
  }
  return max_err;
}

TMPL bool
is_unprecise (const Vector& v, const int& start) {
  ASSERT (start >= 0 && start < ((int) N(v)), "out of range");
  nat i, n= N(v);
  R l= v[n-1];
  for (i= ((nat) start); i<n; i++)
    if (v[i] * l <= 0) return true;
  return false;
}

TMPL bool
is_constant (const Vector& v, const R& tol, const int& start) {
  ASSERT (start >= 0 && start < ((int) N(v)), "out of range");
  nat i, n= N(v);
  R l= v[n-1];
  for (i= ((nat) start); i<n; i++) {
    if (is_zero (v[i]) && is_zero (l)) continue;
    R err= 2 * abs (v[i] - l) / (abs (v[i]) + abs (l));
    if (!is_number (err) || err >= tol) return false;
  }
  return true;
}

TMPL bool
is_infinitesimal (const Vector& v, const R& tol, const int& start) {
  ASSERT (start >= 0 && start < ((int) N(v)), "out of range");
  nat i, n= N(v);
  for (i= ((nat) start); i<n; i++)
    if (abs (v[i]) >= tol) return false;
  return true;
}

TMPL inline int
lhi (const Vector& v) {
  return (int) (N(v) >> 1);
}

TMPL inline int
rhi (const Vector& v) {
  return (int) (N(v) - 1);
}

/******************************************************************************
* Linear interpolation
******************************************************************************/

TMPL R
eval_sub (const generic& fun, nat index) {
  if (is<Vector > (fun)) return as<Vector > (fun) [index];
  else if (is<routine> (fun)) {
    routine r= as<routine> (fun);
    return as<R> (r->apply (as<generic> (R ((int) index))));
  }
  else ERROR ("function or vector expected");
}

TMPL Vector
linear_interpolate (const Vector& v, const vector<generic>& g, const int& b) {
  nat n= N(g);
  nat k= N(v) - b;
  if (k < n) ERROR ("too little data for interpolation");
  Matrix m (R(0), n, n);
  Vector w= fill<R> (R(0), n);
  for (nat row= 0; row<n; row++) {
    nat index= (n == 1? N(v) - 1: b + (row * (k - 1)) / (n - 1));
    w[row]= v[index];
    for (nat col= 0; col<n; col++)
      m (row, col)= eval_sub<R> (g[col], index);
  }
  Vector r= invert (m) * w;
  Vector s= fill<R> (R(0), N(v));
  for (nat index= 0; index<N(v); index++)
    for (nat col= 0; col<n; col++)
      s[index] += r[col] * eval_sub<R> (g[col], index);
  return s;
}

TMPL R
linear_error (const Vector& v, const vector<generic>& g, const int& b) {
  return relative_error (v, linear_interpolate (v, g, b), b);
}

/******************************************************************************
* Transforms
******************************************************************************/

TMPL Vector
check (const Vector& orig, const Vector& v, const char* s) {
  nat i, n= N(v);
  for (i=0; i<n; i++)
    if (!is_number (v[i])) {
      cout << "Failed: " << s << "\n-------\n"
	   << orig << "\n-------\n"
	   << v << "\n-------\n";
      return v;
    }
  return v;
}

TMPL Vector
constant (const Vector& v, const R& c) {
  nat i, n= N(v);
  Vector r= fill<R> (n);
  for (i=0; i<n; i++)
    r[i]= c + R (0);
  return r;
}

TMPL Vector
scale (const Vector& v, const R& p) {
  nat i, n= N(v);
  Vector r= fill<R> (n);
  if (p == 0) r[0] == v[0];
  else r[0]= R (0);
  for (i=1; i<n; i++)
    r[i]= v[i] * pow (R (i), p);
  return check (v, r, "scale");
}

TMPL Vector
scale_log (const Vector& v, const R& p) {
  nat i, n= N(v);
  Vector r= fill<R> (n);
  r[0]= 0;
  r[1]= 0;
  for (i=2; i<n; i++)
    r[i]= v[i] * pow (log (R (i)), p);
  return check (v, r, "scale_log");
}

TMPL Vector
diff (const Vector& v) {
  nat i, n= N(v);
  Vector r= fill<R> (n-1);
  for (i=0; i<n-1; i++)
    r[i]= v[i+1] - v[i];
  return check (v, r, "diff");
}

TMPL Vector
summ (const Vector& v, const R& c, const int& at) {
  ASSERT (at >= 0 && at < ((int) (N(v)+1)), "out of range");
  nat i, n= N(v);
  Vector r= fill<R> (n+1);
  r[0]= R (0);
  for (i=0; i<n; i++)
    r[i+1]= r[i] + v[i];
  return check (v, r + (c - r[(nat) at]), "summ");
}

TMPL Vector
logg (const Vector& v) {
  nat i, n= N(v);
  Vector r= fill<R> (n);
  for (i=0; i<n; i++)
    if (is_zero (v[i])) r[i]= R (-100000000);
    else r[i]= log (abs (v[i]));
  return check (v, r, "llog");
}

TMPL Vector
expp (const Vector& v, const R& sgn) {
  nat i, n= N(v);
  Vector r= fill<R> (n);
  for (i=0; i<n; i++)
    r[i]= sgn * exp (min (R (100000000), v[i]));
  return check (v, r, "expp");
}

TMPL Vector
log_diff (const Vector& v) {
  nat i, n= N(v);
  Vector aux= fill<R> (n);
  for (i=0; i<n; i++)
    if (is_zero (v[i])) aux[i]= R (-100000000);
    else aux[i]= log (abs (v[i] / v[n-1]));
  return check (v, diff (aux), "log_diff");
}

TMPL Vector
exp_summ (const Vector& v, const R& c, const int& at) {
  Vector w= summ (v, R(0), at);
  nat i, n= N(w);
  Vector r= fill<R> (n);
  for (i=0; i<n; i++)
    r[i]= c * exp (min (R (100000000), w[i]));
  return check (w, r, "exp_summ");
}

/******************************************************************************
* Corresponding transforms on results of extrapolations
******************************************************************************/

inline generic
vector_access (const generic& g, nat i) {
  vector<generic> v= as<vector<generic> > (g);
  return v[i];
}

inline vector<generic>
recipe (const generic& g) {
  return as<vector<generic> > (vector_access (g, 0));
}

TMPL inline Vector
original (const generic& g) {
  return as<Vector > (vector_access (g, 1));
}

TMPL inline Vector
approximation (const generic& g) {
  return as<Vector > (vector_access (g, 2));
}

TMPL inline R
error_estimate (const generic& g) {
  return as<R> (vector_access (g, 3));
}

TMPL generic
extra (const vector<generic>& rec, const Vector& orig, const Vector& approx) {
  return as<generic>
    (vec<generic> (as<generic> (rec),
		   as<generic> (orig), as<generic> (approx),
		   as<generic> (relative_error (orig, approx, lhi (orig)))));
}

TMPL vector<generic>
extra_constant (const Vector& v, const R& c) {
  generic r= extra (vec<generic> (gen ("const", as<generic> (c))),
		    v, constant (v, c));
  return vec<generic> (r);
}

TMPL vector<generic>
extra_scale (const vector<generic>& v, const R& p) {
  nat i, n= N(v);
  vector<generic> r= fill<generic> (n);
  for (i=0; i<n; i++)
    r[i]= extra (cons (gen ("scale", as<generic> (p)), recipe (v[i])),
		 scale (original<R> (v[i]), p),
		 scale (approximation<R> (v[i]), p));
  return r;
}

TMPL vector<generic>
extra_summ (const vector<generic>& v, const R& c, const int& at) {
  nat i, n= N(v);
  vector<generic> r= fill<generic> (n);
  for (i=0; i<n; i++)
    r[i]= extra (cons (gen ("summ", as<generic> (c), as<generic> (at)),
		       recipe (v[i])),
		 summ (original<R> (v[i]), c, at),
		 summ (approximation<R> (v[i]), c, at));
  return r;
}

TMPL vector<generic>
extra_expp (const vector<generic>& v, const R& sgn) {
  nat i, n= N(v);
  vector<generic> r= fill<generic> (n);
  for (i=0; i<n; i++)
    r[i]= extra (cons (gen ("expp", as<generic> (sgn)), recipe (v[i])),
		 expp (original<R> (v[i]), sgn),
		 expp (approximation<R> (v[i]), sgn));
  return r;
}

TMPL vector<generic>
extra_exp_summ (const vector<generic>& v, const R& c, const int& at) {
  nat i, n= N(v);
  vector<generic> r= fill<generic> (n);
  for (i=0; i<n; i++)
    r[i]= extra (cons (gen ("exp_summ", as<generic> (c), as<generic> (at)),
		       recipe (v[i])),
		 exp_summ (original<R> (v[i]), c, at),
		 exp_summ (approximation<R> (v[i]), c, at));
  return r;
}

/******************************************************************************
* Automatic extrapolation
******************************************************************************/

TMPL vector<generic> extrapolate (const Vector& u, int order, int step);

TMPL vector<generic>
extrapolate_summ (const Vector& u, int order, int step) {
  if (order == 0 || !is_constant (u, R (0.25), lhi (u)))
    return vec<generic> ();
  cerr << "Diff" << lf;
  vector<generic> r= extrapolate (diff (u), order - 1, step + 1);
  cerr << "Summ " << lf;
  return extra_summ (r, u [rhi (u)], rhi (u));
}

TMPL vector<generic>
extrapolate_expp (const Vector& u, int order, int step) {
  if (order == 0 || is_constant (u, R (0.05), lhi (u)) ||
      u[N(u)-1] == 0 || is_unprecise (u, lhi (u)))
    return vec<generic> ();
  cerr << "Logg" << lf;
  vector<generic> r= extrapolate (logg (u), order, step + 1);
  cerr << "Expp" << lf;
  return extra_expp (r, u [rhi (u)] / abs (u [rhi (u)]));
}

TMPL vector<generic>
extrapolate_exp_summ (const Vector& u, int order, int step) {
  if (order == 0 || is_constant (u, R (0.05), lhi (u)) ||
      u[N(u)-1] == 0 || is_unprecise (u, lhi (u)))
    return vec<generic> ();
  cerr << "Log_diff" << lf;
  vector<generic> r= extrapolate (log_diff (u), order - 1, step + 1);
  cerr << "Exp_summ" << lf;
  return extra_exp_summ (r, u [rhi (u)], rhi (u));
}

TMPL vector<generic>
extrapolate_scale (const Vector& u, int order, int step, const R& p) {
  if (!is_constant (scale (u, -p), R (0.25), lhi (u)))
    return vec<generic> ();
  cerr << "Scale " << (-p) << lf;
  vector<generic> r= extrapolate (scale (u, -p), order, step + 1);
  cerr << "Scale " << p << lf;
  return extra_scale (r, p);
}

TMPL vector<generic>
extrapolate_poww (const Vector& u, int order, int step) {
  // Detect expansions c n^a + ...
  if (order < 2 || is_unprecise (u, lhi (u)))
    return vec<generic> ();
  Vector v1= logg (u);
  if (!is_constant (scale_log (v1, R (-1)), R (0.25), lhi (v1)))
    return vec<generic> ();
  Vector v2= diff (v1);
  if (!is_constant (scale (v2, R (1)), R (0.05), lhi (v2)))
    return vec<generic> ();
  Vector v3= scale (v2, R (1));
  Vector v4= diff (v3);
  Vector v5= scale_log (scale (v4, R (1)), R (2));
  if (!is_infinitesimal (v5, R (0.5), lhi (v5)))
    return vec<generic> ();

  R p= v3 [N(v3) - 1];
  cerr << "p= " << p << lf;
  if (abs (p) < R (1) / sqrt (R (N (u))))
    return extrapolate_summ (u, order, step);
  if (abs (p - round (p)) < R (1) / sqrt (R (N (u))))
    return extrapolate_scale (u, order, step, round (p));

  cerr << "Power transform " << p << lf;
  vector<generic> r4= extrapolate (v4, order - 2, step + 2);
  vector<generic> r3= extra_summ (r4, v3[rhi (v3)], rhi (v3));
  vector<generic> r2= extra_scale (r3, R (-1));
  vector<generic> r1= extra_summ (r2, v1[rhi (v1)], rhi (v1));
  cerr << "Inverse power transform " << p << lf;
  return extra_expp (r1, u [rhi (u)] / abs (u [rhi (u)]));
}

TMPL vector<generic>
extrapolate_logpol (const Vector& u, int order, int step, int d) {
  if (!is_constant (scale_log (u, R(-d)), R (0.25), lhi (u)))
    return vec<generic> ();
  Vector v1= diff (u);
  Vector v2= scale (v1, R (1));
  vector<generic> r2;
  if (d == 0) {
    if (!is_infinitesimal (v2, u[N(u)-1] / sqrt (R (N (u))), lhi (v2)))
      return vec<generic> ();
    cerr << "Log polynomial" << lf;
    r2= extrapolate (v2, order-1, step+1);
    cerr << "Log polynomial" << lf;
  }
  else r2= extrapolate_logpol (v2, order-1, step+1, d-1);
  vector<generic> r1= extra_scale (r2, R (-1));
  return extra_summ (r1, u[rhi (u)], rhi (u));
}

TMPL vector<generic>
extrapolate_logpol (const Vector& u, int order, int step, const R& p, int d) {
  // Detect expansions (a_d * (log n)^d + ... + a_0) n^p + ...
  if (order <= d) return vec<generic> ();
  Vector v= scale (u, -p);
  vector<generic> r= extrapolate_logpol (v, order, step, d);
  return extra_scale (r, p);
}

TMPL
struct better_extrapolation_op {
  static inline bool
  op (const generic& x, const generic& y) {
    return error_estimate<R> (x) <= error_estimate<R> (y); }
};

TMPL vector<generic>
extrapolate (const Vector& u, int order, int step) {
  cerr << "Extrapolating at order " << order << lf << indent;
  //cerr << "Extrapolating " << u << ", " << order << lf << indent;
  vector<generic> rr;
  if (N (rr) == 0)
    rr << extrapolate_poww (u, order, step);
  if (N (rr) == 0) {
    rr << extrapolate_summ (u, order, step);
    rr << extrapolate_scale (u, order, step, R (1));
    rr << extrapolate_scale (u, order, step, R (-1));
    rr << extrapolate_scale (u, order, step, R (-2));
    //rr << extrapolate_logpol (u, order, step, R (-1), 1);
    //rr << extrapolate_logpol (u, order, step, R (-2), 1);
    if (N (rr) == 0) {
      //rr << extrapolate_expp (u, order, step);
      rr << extrapolate_exp_summ (u, order, step);
    }
  }
  if (N (rr) == 0)
    rr << extra_constant (u, R (0));
  for (nat i=0; i<N(rr); i++)
    rr[i]= extra (recipe (rr[i]), u, approximation<R> (rr[i]));
  //cerr << unindent << "Results at order " << order << ": " << N (rr) << lf;
  cerr << unindent << "Errors at order " << order << ": ";
  for (nat i=0; i<N(rr); i++) {
    if (i != 0) cerr << ", ";
    cerr << error_estimate<R> (rr[i]);
  }
  cerr << lf;
  sort_leq<better_extrapolation_op<R> > (rr);
  if (N (rr) >= 10) rr= range (rr, 0, 10);
  return rr;
}

TMPL vector<generic>
jorix_extrapolate (const Vector& u, const int& order) {
  return extrapolate (u, order, 0);
}

/******************************************************************************
* Printing results
******************************************************************************/

TMPL void
diagnosis (const vector<generic>& v) {
  for (nat i=0; i<N(v); i++) {
    generic x= v[i];
    cout << "--------------------------------------------\n"; 
    cout << "Recipe: " << recipe (x) << "\n";
    cout << "Error : " << error_estimate<R> (x) << "\n";
    cout << "--------------------------------------------\n";
    Vector u= original<R> (x);
    Vector v= approximation<R> (x);
    for (nat k= lhi (u); k < N(u); k++)
      cout << k << "\t: " << u[k] << "\t->\t" << v[k] << "\n";
  }
}

#undef TMPL
#undef Vector
#undef Matrix
} // namespace mmx
#endif // __MMX__EXTRAPOLATE__HPP
