
/******************************************************************************
* MODULE     : guess.hpp
* DESCRIPTION: Guessing linear dependencies at singularities
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__GUESS__HPP
#define __MMX__GUESS__HPP
#include <numerix/floating.hpp>
#include <numerix/complex.hpp>
#include <analyziz/vector_floating.hpp>
#include <analyziz/matrix_floating.hpp>
#include <analyziz/polynomial_floating.hpp>
#include <algebramix/matrix_quotient.hpp>
#include <algebramix/polynomial_quotient.hpp>
#include <algebramix/series_relaxed.hpp>
namespace mmx {
#define TMPL template<typename R>
#define Vector vector<R>
#define Matrix matrix<R>
#define Polynomial polynomial<R>
#define Series series<R>

TMPL Matrix
orthogonalize (const Matrix& m) {
  Matrix r= copy (m);
  for (nat i=0; i<rows(r); i++)
    for (nat k=0; k<i; k++) {
      R sum1= 0, sum2= 0;
      for (nat j=0; j<cols(r); j++) {
	sum1 += r (i, j) * r (k, j);
	sum2 += r (k, j) * r (k, j); // * conj (r (k, j))
      }
      R lambda= sum1 / sum2;
      for (nat j=0; j<cols(r); j++)
	r (i, j) -= lambda * r (k, j);
    }
  return r;
}

TMPL Matrix
guess_encode (const vector<generic>& gv, nat n) {
  nat d= N(gv);
  vector<Polynomial > v= fill<Polynomial > (d);
  for (nat i=0; i<d; i++) {
    if (is<Polynomial > (gv[i]))
      v[i]= as<Polynomial > (gv[i]);
    if (is<polynomial<rational> > (gv[i]))
      v[i]= as<Polynomial > (as<polynomial<rational> > (gv[i]));
    else if (is<Series > (gv[i]))
      v[i]= truncate (as<Series > (gv[i]), n);
    else if (is<series<rational> > (gv[i]))
      v[i]= truncate (as<Series > (as<series<rational> > (gv[i])), n);
    else ERROR ("type mismatch");
  }
  Matrix m (R (0), d * n, n + d * n);
  for (nat i=0; i<d*n; i++) {
    for (nat j=0; j<n; j++) {
      nat ii= i % d;
      nat jj= (n - 1) - (i / d);
      if (j >= jj) m (i, j) = v[ii][j - jj];
    }
    m (i, n + i) = 1;
  }
  return m;
}

TMPL R
norm2 (const Vector& v) {
  R s= 0;
  for (nat i=0; i<N(v); i++) s += square (v[i]);
  return sqrt (s);
}

TMPL vector<generic>
guess_decode (const Matrix& m) {
  nat dn= rows (m), n= cols (m) - dn, d= dn / n;
  nat best_index= 0;
  R   best_norm = Plus_infinity (R);
  for (nat i=0; i<d; i++) {
    nat next_index= dn - 1 - i;
    R   next_norm = norm2 (row (m, dn - 1 - i));
    if (next_norm < best_norm) {
      best_index= next_index;
      best_norm = next_norm;
    }
  }
  cerr << "Best norm: " << best_norm << "\n";

  vector<generic> r= fill<generic> (generic (0), d);
  for (nat j=0; j<d; j++) {
    typedef typename Vector_variant(R) V;
    nat l= aligned_size<R,V> (n);
    R*  c= mmx_new<R> (l);
    for (nat k=0; k<n; k++)
      c[k]= m (best_index, n + (n - 1 - k) * d + j);
    r[j]= as<generic> (Series (Polynomial (c, n, l)));
  }
  return r;
}

TMPL matrix<generic>
guess_decode_basis (const Matrix& m) {
  nat dn= rows (m), n= cols (m) - dn, d= dn / n;
  matrix<generic> r (generic (0), d, d);
  for (nat i=0; i<d; i++)
    for (nat j=0; j<d; j++) {
      typedef typename Vector_variant(R) V;
      nat l= aligned_size<R,V> (n);
      R*  c= mmx_new<R> (l);
      for (nat k=0; k<n; k++)
	c[k]= m (dn - 1 - i, n + (n - 1 - k) * d + j);
      r (i, j)= as<generic> (Series (Polynomial (c, n, l)));
    }
  return r;
}

TMPL vector<generic>
guess (const vector<generic>& v, const int& n) {
  mmx_series_precision= (nat) n;
  return guess_decode<R> (orthogonalize (guess_encode<R> (v, (nat) n)));
}

TMPL matrix<generic>
guess_basis (const vector<generic>& v, const int& n) {
  mmx_series_precision= (nat) n;
  return guess_decode_basis<R> (orthogonalize (guess_encode<R> (v, (nat) n)));
}

#undef TMPL
#undef Vector
#undef Matrix
#undef Polynomial
#undef Series
} // namespace mmx
#endif // __MMX__GUESS__HPP
