
/******************************************************************************
* MODULE     : torn_evaluate.hpp.hpp
* DESCRIPTION: Numerically stable multi-point evaluation for torn polynomials
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_TORN_EVALUATE_HPP
#define __MMX_TORN_EVALUATE_HPP
#include <jorix/torn_polynomial.hpp>
namespace mmx {
template<typename C> struct standard_torn;
#define TMPL_DEF template<typename C, typename V= standard_torn<C> >
#define TMPL template<typename C, typename V>
#define Torn torn<C,V>
#define Torn_vector vector<Torn >
#define Torn_polynomial torn_polynomial<C,V>
#define Torn_polynomial_rep torn_polynomial_rep<C,V>
#define W V::PV
#define Polynomial typename V::PT
#define Exp typename V::ET
TMPL class torn_polynomial_rep;
TMPL class torn_polynomial;

/******************************************************************************
* Multi-point evaluation
******************************************************************************/

TMPL
struct torn_evaluator {
  Torn_polynomial tp;  // polynomial to be evaluated
  Torn_vector     tv;  // points where to evaluated (ordered by norm)
  Torn_vector     rv;  // values
  vector<Exp>     ex;  // raw exponents
  vector<nat>     ab;  // abcissae of Newton polygon
  vector<nat>     b1;  // lower border of non-neglectable area
  vector<nat>     bm;  // index which realises maximal norm
  vector<nat>     b2;  // upper border of non-neglectable area

  torn_evaluator (const Torn_polynomial& tp2, const Torn_vector& tv2):
    tp (tp2), tv (tv2)
  {
    ex= tp->exponents ();
    ab= convex_hull_abscissae (ex);
    linear_interpolate (ex, ab);
    rv= fill<Torn > (N(tv));
    b1= fill<nat> (0, N(tv));
    bm= fill<nat> (0, N(tv));
    b2= fill<nat> (0, N(tv));
    //cout << "tp= " << tp << "\n";
    //cout << "tv= " << tv << "\n";
    //cout << "ex= " << ex << "\n";
    nat k1= 0, km= 0, k2= 0;
    for (nat j=0; j<N(tv); j++) {
      Exp e= tv[j].e;
      while (km+1 < N(ex) && ex[km+1] + e > ex[km]) km++;
      bm[j]= km;
      rv[j].e= ex[km] + (km + tp->val) * e;
      Exp thr= ex[km] + km * e - V::prec() - V::disc();
      while (k1 < km && ex[k1] + k1 * e < thr) k1++;
      b1[j]= k1;
      k2= max (k2, km);
      while (k2+1 < N(ex) && ex[k2+1] + (k2+1) * e >= thr) k2++;
      b2[j]= k2;
      //cout << "  " << j << ": " << k1 << ", " << km << ", " << k2
      //<< "; " << as_double (rv[j].e) << "\n";
    }
  }

  Torn_vector
  naive_evaluate () {
    if (N (tp->p) == 1) {
      for (nat j=0; j<N(tv); j++)
	rv[j]= Torn (tp->p[0], tp->ex[0]) * pow (tv[j], tp->val);
    }
    else {
      nat j1=0, j2=0;
      for (nat k=0; k<N(tp->ab)-1; k++) {
	nat i1= tp->ab[k], i2= tp->ab[k+1];
	Exp sl= (tp->ex[i2] - tp->ex[i1]) / (i2 - i1);
	if (k == N(tp->ab) - 2) i2++;
	while (j1 < N(tv) && b2[j1] < i1) j1++;
	j2= max (j1, j2);
	while (j2 < N(tv) && b1[j2] < i2) j2++;
	for (nat j= j1; j<j2; j++) {
	  C x= V::shift (tv[j].x, tv[j].e + sl);
	  nat i3= max (i1, b1[j]);
	  nat i4= min (i2, b2[j]+1);
	  C y= W::eval (seg (tp->p) + i3, x, i4 - i3);
	  Torn p= pow (tv[j], tp->val + i3);
	  rv[j].x += V::shift (y * p.x, tp->ex[i3] + p.e - rv[j].e);
	}
      }
    }
    return rv;
  }
};

TMPL Torn_vector
eval (const Torn_polynomial& tp, const Torn_vector& tv) {
  if (N (tp->p) == 0) return fill<Torn > (Torn (), N (tv));
  torn_evaluator<C,V> te (tp, tv);
  return te.naive_evaluate ();
}

TMPL Torn_vector
eval (const Torn_polynomial& tp, const Torn_vector& tv, vector<nat>& ind) {
  if (N (tp->p) == 0) return fill<Torn > (Torn (), N (tv));
  torn_evaluator<C,V> te (tp, tv);
  Torn_vector rv= te.naive_evaluate ();
  ind= te.bm;
  return rv;
}

#undef TMPL_DEF
#undef TMPL
#undef Torn
#undef Torn_vector
#undef Torn_polynomial
#undef Torn_polynomial_rep
#undef W
#undef Polynomial
#undef Exp
} // namespace mmx
#endif // __MMX_TORN_EVALUATE_HPP
