
/******************************************************************************
* MODULE     : torn_number.hpp.hpp
* DESCRIPTION: Torn polynomials for fast numerically stable arithmetic
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_TORN_NUMBER_HPP
#define __MMX_TORN_NUMBER_HPP
#include <basix/double.hpp>
#include <numerix/floating.hpp>
#include <numerix/complex.hpp>
namespace mmx {
template<typename C> struct standard_torn;
#define TMPL_DEF template<typename C, typename V= standard_torn<C> >
#define TMPL template<typename C, typename V>
#define Torn torn<C,V>
#define Torn_vector vector<Torn >
#define Torn_polynomial torn_polynomial<C,V>
#define Torn_polynomial_rep torn_polynomial_rep<C,V>
#define W V::PV
#define Polynomial typename V::PT
#define Exp typename V::ET
TMPL class torn_polynomial_rep;
TMPL class torn_polynomial;

/******************************************************************************
* Missing helper routines
******************************************************************************/

inline integer floor (const rational& x) {
  return quo (numerator (x), denominator (x)); }

/******************************************************************************
* Variants for torn polynomials
******************************************************************************/

template<typename C>
struct standard_torn {
  typedef Real_type(C) R;
  typedef typename polynomial_variant_helper<C>::PV PV;
  typedef polynomial<C,PV> PT;
  typedef R ET;
  static inline ET get_exponent (const C& c) { return ET (exponent (c)); }
  static inline ET get_magnitude (const C& c) { return ET (magnitude (c)); }
  static inline R get_power (const ET& e) { return exp2 (R (e)); }
  static inline C shift (const C& c, const ET& e) { return c * get_power (e); }
  static inline nat prec () { return 52; }
  static inline nat disc () { return 52; }
};

STMPL
struct standard_torn<complex<floating<> > > {
  typedef complex<floating<> > C;
  typedef floating<> R;
  typedef polynomial_variant_helper<C>::PV PV;
  typedef polynomial<C,PV> PT;
  typedef rational ET;
  static inline ET get_exponent (const C& c) { return ET (exponent (c)); }
  static inline R get_power (const ET& e) { return exp2 (R (e)); }
  static inline C shift (const C& c, const ET& e) { return c * get_power (e); }
  static inline ET get_magnitude (const C& c) {
    double mag= magnitude (c);
    return ET ((int) floor (66536 * mag)) / 65536; }
  static inline nat prec () { return mmx_bit_precision; }
  static inline nat disc () { return mmx_discrepancy; }
};

/******************************************************************************
* Torn numbers
******************************************************************************/

TMPL_DEF
class torn {
MMX_ALLOCATORS;
public:
  C   x;
  Exp e;
  inline torn (): x (0), e (0) {}
  template<typename T>
  torn (const T& x2, const Exp& e2= 0): x (as<C> (x2)), e (e2) {
    if (x == 0) e= V::get_exponent (x);
    else {
      Exp inc= V::get_magnitude (x);
      x *= V::get_power (-inc);
      e += inc;
    }
  }
};

template<typename Op, typename C, typename V> nat
unary_hash (const Torn& t) {
  nat h= Op::op (t.x);
  return (h<<1) ^ (h<<3) ^ (h<<9) ^ (h>>29) ^ Op::op (t.e);
}

template<typename Op, typename C, typename V> bool
binary_test (const Torn& t1, const Torn& t2) {
  return Op::op (t1.x, t2.x) && Op::op (t1.e, t2.e);
}

TMPL inline syntactic
flatten (const Torn& t) {
  return flatten (t.x) * pow (flatten (2), flatten (t.e));
}

TRUE_IDENTITY_OP_SUGAR(TMPL,Torn)
SOFT_IDENTITY_OP_SUGAR(TMPL,Torn)
HARD_IDENTITY_OP_SUGAR(TMPL,Torn)
PRINT_SUGAR(TMPL,Torn)

TMPL inline C untwist (const Torn& t) { return V::shift (t.x, t.e); }
TMPL inline C mantissa (const Torn& t) { return t.x; }
TMPL inline Exp exponent (const Torn& t) { return t.e; }

TMPL vector<C>
untwist (const vector<Torn >& tv) {
  vector<C> r= fill<C> (N(tv));
  for (nat i=0; i<N(tv); i++) r[i]= untwist (tv[i]);
  return r;
}

TMPL vector<C>
mantissa (const vector<Torn >& v) {
  vector<C> r= fill<C> (N(v));
  for (nat i=0; i<N(v); i++) r[i]= mantissa (v[i]);
  return r;
}

TMPL vector<Exp >
exponent (const vector<Torn >& v) {
  vector<Exp > r= fill<Exp > (N(v));
  for (nat i=0; i<N(v); i++) r[i]= exponent (v[i]);
  return r;
}

/******************************************************************************
* Arithmetic on torn numbers
******************************************************************************/

TMPL Torn
operator - (const Torn& t) {
  return Torn (-t.x, t.e);
}

TMPL Torn
operator + (const Torn& t1, const Torn& t2) {
  if (t1.e >= t2.e) return Torn (t1.x + V::shift (t2.x, t2.e - t1.e), t1.e);
  else return Torn (V::shift (t1.x, t1.e - t2.e) + t2.x, t2.e);
}

TMPL Torn
operator - (const Torn& t1, const Torn& t2) {
  if (t1.e >= t2.e) return Torn (t1.x - V::shift (t2.x, t2.e - t1.e), t1.e);
  else return Torn (V::shift (t1.x, t1.e - t2.e) - t2.x, t2.e);
}

TMPL Torn
operator * (const Torn& t1, const Torn& t2) {
  return Torn (t1.x * t2.x, t1.e + t2.e);
}

TMPL Torn
square (const Torn& t) {
  return Torn (square (t.x), 2 * t.e);
}

TMPL Torn
operator / (const Torn& t1, const Torn& t2) {
  return Torn (t1.x / t2.x, t1.e - t2.e);
}

TMPL Torn
sqrt (const Torn& t) {
  return Torn (sqrt (t.x), t.e / 2);
}

template<typename C, typename V, typename K> Torn
pow (const Torn& t, const K& p) {
  typedef Real_type(C) R;
  return Torn (pow (t.x, R (p)), p * t.e);
}

#undef TMPL_DEF
#undef TMPL
#undef Torn
#undef Torn_vector
#undef Torn_polynomial
#undef Torn_polynomial_rep
#undef W
#undef Polynomial
#undef Exp
} // namespace mmx
#endif // __MMX_TORN_NUMBER_HPP
