
/******************************************************************************
* MODULE     : torn_polynomial.hpp
* DESCRIPTION: Torn polynomials for fast numerically stable arithmetic
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_TORN_POLYNOMIAL_HPP
#define __MMX_TORN_POLYNOMIAL_HPP
#include <jorix/torn_number.hpp>
#include <algebramix/polynomial.hpp>
#include <analyziz/newton_polygon.hpp>
namespace mmx {
template<typename C> struct standard_torn;
#define TMPL_DEF template<typename C, typename V= standard_torn<C> >
#define TMPL template<typename C, typename V>
#define Torn torn<C,V>
#define Torn_vector vector<Torn >
#define Torn_polynomial torn_polynomial<C,V>
#define Torn_polynomial_rep torn_polynomial_rep<C,V>
#define W V::PV
#define Polynomial typename V::PT
#define Exp typename V::ET
TMPL class torn_polynomial_rep;
TMPL class torn_polynomial;

template<typename C> polynomial<C>
as_polynomial (const vector<C>& v) {
  typedef typename Polynomial_variant(C) V;
  nat l= aligned_size<C,V> (N(v));
  C* p= mmx_new<C> (l);
  V::Vec::copy (p, seg (v), N(v));
  return polynomial<C> (p, N(v), l);
}

/******************************************************************************
* Torn univariate polynomials
******************************************************************************/

TMPL_DEF
class torn_polynomial_rep REP_STRUCT {
public:
  Polynomial   p;    // torn polynomial
  nat          val;  // valuation
  vector<Exp>  ex;   // exponents for the coefficients
  vector<nat>  ab;   // extremities of linear segments

public:
  vector<Exp> exponents () const;
  vector<nat> partition (const vector<nat>& ab, const vector<Exp>& v);
  void normalize ();
  inline torn_polynomial_rep (const Polynomial& p2, nat val2,
			      const vector<Exp>& ex2):
    p (p2), val (val2), ex (ex2) {
      normalize (); }
  friend class Torn_polynomial;
};

TMPL_DEF
class torn_polynomial {
INDIRECT_PROTO_2 (torn_polynomial, torn_polynomial_rep, C, V)
public:
  inline torn_polynomial (const Polynomial& p):
    rep (new Torn_polynomial_rep (p, 0, fill<Exp> (0, N(p)))) {}
  inline torn_polynomial (const Polynomial& p, nat val,
			  const vector<Exp>& ex):
    rep (new Torn_polynomial_rep (p, val, ex)) {}
  inline torn_polynomial (const vector<Torn >& v):
    rep (new Torn_polynomial_rep (as_polynomial (mantissa (v)), 0,
				  exponent (v))) {}
  inline Torn operator [] (nat i) const {
    if (i < rep->val || i >= rep->val + N(rep->p)) return Torn (0, 0);
    else return Torn (rep->p[i - rep->val], rep->ex[i - rep->val]); }
};
INDIRECT_IMPL_2 (torn_polynomial, torn_polynomial_rep,
		 typename C, C, typename V, V)

TMPL inline nat
N (const Torn_polynomial& tp) {
  return tp->val + N(tp->p);
}

/******************************************************************************
* Required operations
******************************************************************************/

template<typename Op, typename C, typename V> nat
unary_hash (const Torn_polynomial& pp) {
  return Op::op (pp->p) ^ Op::op (pp->val) ^ Op::op (pp->ex);
}

template<typename Op, typename C, typename V> bool
binary_test (const Torn_polynomial& pp1, const Torn_polynomial& pp2) {
  return Op::op (pp1->p  , pp2->p  ) &&
         Op::op (pp1->val, pp2->val) &&
         Op::op (pp1->ex , pp2->ex );
}

TMPL inline syntactic
flatten (const Torn_polynomial& pp) {
  return apply ("torn_polynomial", flatten (pp->p),
		flatten (pp->val), flatten (pp->ex), flatten (pp->ab));
}

TRUE_IDENTITY_OP_SUGAR(TMPL,Torn_polynomial)
SOFT_IDENTITY_OP_SUGAR(TMPL,Torn_polynomial)
PRINT_SUGAR(TMPL,Torn_polynomial)

/******************************************************************************
* Normalization
******************************************************************************/

TMPL vector<Exp>
Torn_polynomial_rep::exponents () const {
  vector<Exp> r= fill<Exp> (N(p));
  for (nat i=0; i<N(p); i++)
    r[i]= floor (ex[i] + Exp (exponent (p[i])));
  return r;
}

TMPL vector<nat>
Torn_polynomial_rep::partition (const vector<nat>& ab, const vector<Exp>& v) {
  vector<nat> r;
  nat i, j, k;
  nat max_disc= V::disc () >> 1;
  for (i=0; (i+1)<N(ab); i=j-1) {
    r << ab[i];
    k= i+1;
    for (j=i+2; j<N(ab); j++) {
      nat ii= ab[i], jj= ab[j], kk= ab[k];
      Exp slope= (v[jj] - v[ii]) / (jj - ii);
      Exp disc = v[kk] - (v[ii] + (kk - ii) * slope);
      while (k < j) {
	nat nn= ab[k+1];
	Exp new_disc= v[nn] - (v[ii] + (nn - ii) * slope);
	if (new_disc <= disc) break;
	k++; disc= new_disc;
      }
      if (disc > max_disc) break;
    }
  }
  r << ab[i];
  return r;
}

template<typename C> void
linear_interpolate (vector<C>& ex, const vector<nat>& ab) {
  for (nat i=0; i<N(ab)-1; i++) {
    C cur= ex[ab[i]];
    C inc= (ex[ab[i+1]] - cur) / (ab[i+1] - ab[i]);
    cur += inc;
    for (nat k=ab[i]+1; k<ab[i+1]; k++, cur += inc) ex[k]= cur;
  }
}

TMPL void
Torn_polynomial_rep::normalize () {
  nat xval= 0;
  while (xval < N(p) && p[xval] == 0) xval++;
  if (xval > 0) {
    p    = range (p, xval, N(p));
    val += xval;
    ex   = range (ex, xval, N(ex));
  }

  if (N(p) == 0) {
    ab = vector<nat> ();
    val= 0;
    ex = vector<Exp> ();
    return;
  }

  vector<Exp> new_ex= exponents ();
  //cout << "aux_ex= " << new_ex << "\n";
  vector<nat> aux_ab= convex_hull_abscissae (new_ex);
  //cout << "aux_ab= " << aux_ab << "\n";
  vector<nat> new_ab= partition (aux_ab, new_ex);
  //cout << "new_ab= " << new_ab << "\n";
  linear_interpolate (new_ex, new_ab);
  //cout << "old_ex= " << ex << "\n";
  //cout << "new_ex= " << new_ex << "\n";
  C* new_p = const_cast<C*> (seg (p));
  for (nat i=0; i<N(p); i++)
    new_p[i] *= V::get_power (ex[i] - new_ex[i]);
  ex= new_ex;
  ab= new_ab;
}

TMPL Polynomial
untwist (const Torn_polynomial& tp) {
  nat n= tp->val + N(tp->p);
  nat l= aligned_size<C, typename W> (n);
  C* p= mmx_new<C> (l);
  W::Vec::clear (p, tp->val);
  for (nat i=0; i<N(tp->p); i++)
    p[i + tp->val]= tp->p[i] * V::get_power (tp->ex[i]);
  return Polynomial (p, n, l);
}

/******************************************************************************
* Addition and subtraction
******************************************************************************/

TMPL inline Torn_polynomial
operator - (const Torn_polynomial& tp) {
  return Torn_polynomial (-tp->p, tp->val, tp->ex);
}

template<typename Op, typename C, typename V> Torn_polynomial
combine (const Torn_polynomial& tp1, const Torn_polynomial& tp2) {
  nat beg= min (tp1->val, tp2->val);
  nat end= max (N(tp1->p) + tp1->val, N(tp2->p) + tp2->val);
  if (end == beg) return tp1;
  nat l= aligned_size<C, typename W> (end - beg);
  C*  r= mmx_new<C> (l);
  nat alloc_e= default_aligned_size<Exp> (end-beg);
  Exp* e= mmx_new<Exp> (alloc_e);
  for (nat i= beg; i<end; i++) {
    bool b1= (i >= tp1->val && i < tp1->val + N(tp1->p));
    C    c1= (b1? tp1->p[i - tp1->val]: C(0));
    bool b2= (i >= tp2->val && i < tp2->val + N(tp2->p));
    C    c2= (b2? tp2->p[i - tp2->val]: C(0));
    if (b1 && b2) {
      Exp e1= tp1->ex[i-tp1->val];
      Exp e2= tp2->ex[i-tp2->val];
      if (e1 >= e2) {
	r[i-beg]= Op::op (c1, c2 * V::get_power (e2-e1));
	e[i-beg]= e1;
      }
      else {
	r[i-beg]= Op::op (c1 * V::get_power (e1-e2), c2);
	e[i-beg]= e2;
      }
    }
    else if (b1) {
      r[i-beg]= c1;
      e[i-beg]= tp1->ex[i-tp1->val];
    }
    else if (b2) {
      r[i-beg]= Op::op (c1, c2);
      e[i-beg]= tp2->ex[i-tp2->val];
    }
    else {
      if (i <= beg) {
	cout << "tp1= " << tp1 << "\n";
	cout << "tp2= " << tp2 << "\n";
      }
      ASSERT (i>beg, "invalid assumption");
      r[i-beg]= 0;
      e[i-beg]= e[i-beg-1];
    }
  }
  Polynomial  p  (r, end - beg, l);
  vector<Exp> ex (e, end - beg, alloc_e);
  return Torn_polynomial (p, beg, ex);
}

TMPL Torn_polynomial
operator + (const Torn_polynomial& tp1, const Torn_polynomial& tp2) {
  if (N (tp1->p) == 0) return tp2;
  if (N (tp2->p) == 0) return tp1;
  return combine<add_op> (tp1, tp2);
}

TMPL Torn_polynomial
operator - (const Torn_polynomial& tp1, const Torn_polynomial& tp2) {
  if (N (tp1->p) == 0) return -tp2;
  if (N (tp2->p) == 0) return tp1;
  return combine<sub_op> (tp1, tp2);
}

/******************************************************************************
* Other linear time operations
******************************************************************************/

TMPL Torn_polynomial
range (const Torn_polynomial& tp, nat i, nat j) {
  if (i >= tp->val + N(tp->p) || j <= tp->val || i >= j)
    return Torn_polynomial (Polynomial (0));
  else if (i < tp->val)
    return Torn_polynomial (range (tp->p , 0, j - tp->val), tp->val - i,
			    range (tp->ex, 0, j - tp->val));
  else
    return Torn_polynomial (range (tp->p , i - tp->val, j - tp->val), 0,
			    range (tp->ex, i - tp->val, j - tp->val));
}

TMPL Torn_polynomial
operator * (const C& c, const Torn_polynomial& tp) {
  if (c == 0 || N(tp->p) == 0) return Torn_polynomial (Polynomial (0));
  return Torn_polynomial (c * tp->p, tp->val, tp->ex);
}

TMPL Torn_polynomial
operator * (const Torn_polynomial& tp, const C& c) {
  if (c == 0 || N(tp->p) == 0) return Torn_polynomial (Polynomial (0));
  return Torn_polynomial (tp->p * c, tp->val, tp->ex);
}

TMPL Torn_polynomial
operator / (const Torn_polynomial& tp, const C& c) {
  if (N(tp->p) == 0) return Torn_polynomial (Polynomial (0));
  return Torn_polynomial (tp->p / c, tp->val, tp->ex);
}

TMPL Torn_polynomial
derive (const Torn_polynomial& tp) {
  if (tp->val == 0)
    return Torn_polynomial (derive (tp->p), 0,
			    cdr (tp->ex));
  else
    return Torn_polynomial (lshiftz (derive (tp->p), 1) + C(tp->val) * tp->p,
			    tp->val-1, tp->ex);
}

TMPL Torn_polynomial
q_difference (const Torn_polynomial& tp, const Torn& q) {
  Torn prod (1, 0);
  vector<Torn > v= fill<Torn > (N(tp));
  for (nat i=0; i<N(tp); i++, prod *= q)
    v[i]= prod * tp[i];
  return Torn_polynomial (v);
}

TMPL Torn_polynomial
laplace (const Torn_polynomial& tp) {
  Torn fact (1, 0);
  vector<Torn > v= fill<Torn > (N(tp));
  for (nat i=0; i<N(tp); i++, fact *= Torn (i))
    v[i]= fact * tp[i];
  return Torn_polynomial (v);
}

TMPL Torn_polynomial
borel (const Torn_polynomial& tp) {
  Torn fact (1, 0);
  vector<Torn > v= fill<Torn > (N(tp));
  for (nat i=0; i<N(tp); i++, fact *= Torn (i))
    v[i]= tp[i] / fact;
  return Torn_polynomial (v);
}

TMPL Torn_polynomial
reverse (const Torn_polynomial& tp, nat n) {
  vector<Torn > v= fill<Torn > (n);
  for (nat i=0; i<n; i++) v[i]= tp[n-1-i];
  return Torn_polynomial (v);
}

TMPL Torn_polynomial
shift (const Torn_polynomial& tp, const Torn& sh) {
  nat n= N(tp);
  Torn_polynomial o (fill<Torn > (Torn (1), n));
  Torn_polynomial u= reverse (laplace (tp), n);
  Torn_polynomial v= borel (q_difference (o, sh));
  return borel (reverse (u * v, n));
}

/******************************************************************************
* Multiplication
******************************************************************************/

template<typename E> vector<nat>
border (const vector<E>& e, const E& prec,
	const vector<E>& e1, const vector<E>& e2)
{
  nat i1=0, i2=0;
  vector<nat> r= fill<nat> (0, N(e1));
  while (true) {
    if (i1 > 0 && i2+1 < N(e2) && e1[i1-1] + e2[i2+1] >= e[i1+i2] - prec) {
      i1--; i2++; }
    else if (i1+1 < N(e1)) i1++;
    else if (i2+1 < N(e2)) i2++;
    else break;
    r[i1]= i2;
  }
  return r;
}

TMPL
struct torn_multiplier {
  typedef typename V::R R;

  Torn_polynomial tp1;
  Torn_polynomial tp2;
  nat             nn;
  C*              p1;
  C*              p2;
  C*              pm;
  vector<Exp>     ex;
  vector<nat>     b1;
  vector<nat>     b2;

  torn_multiplier (const Torn_polynomial& tp1b, const Torn_polynomial& tp2b):
    tp1 (tp1b), tp2 (tp2b)
  {
    nn= N(tp1->p) + N(tp2->p) - 1;
    nat ll= aligned_size<C, typename W> (nn);
    p1= mmx_new<C> (ll);
    p2= mmx_new<C> (ll);
    pm= mmx_new<C> (ll);
    W::clear (pm, nn);
    ex= minkowski_sum (tp1->ex, tp2->ex);
    //cout << "ex= " << ex << "\n";
    b1= border (ex, Exp (V::prec () + V::disc ()), tp1->ex, tp2->ex);
    //cout << "b1= " << b1 << "\n";
    b2= border (ex, Exp (V::prec () + V::disc ()), tp2->ex, tp1->ex);
    //cout << "b2= " << b2 << "\n";
  }

  inline ~torn_multiplier () {
    nat ll= aligned_size<C, typename W> (nn);
    mmx_delete<C> (p1, ll);
    mmx_delete<C> (p2, ll);
    // pm is consumed by multiply and square
  }

  bool
  top_negligible (nat i, nat j) {
    nat lo1= tp1->ab[i];
    nat hi2= tp2->ab[j+1];
    if (j == N(tp2->ab) - 2) hi2++;
    return b2[hi2-1] < lo1;
  }

  bool
  bot_negligible (nat i, nat j) {
    nat hi1= tp1->ab[i+1];
    nat lo2= tp2->ab[j];
    if (i == N(tp1->ab) - 2) hi1++;
    return b1[hi1-1] < lo2;
  }

  void
  scale (C* dest, const C* src,
	 nat lo1, nat hi1, nat lo2, nat hi2,
	 const Exp& inc)
  {
    Exp cur= tp1->ex[lo1] + tp2->ex[lo2];
    for (nat i= lo1 + lo2; i < hi1 + hi2 - 1; i++, cur += inc)
      mul (dest[i], src[i], V::get_power (cur - ex[i]));
  }
  
  void
  accumulate_ver (nat Lo1, nat hi1, nat lo2, nat Hi2,
		  const Exp& inc, const R& sc)
  {
    //cout << "Accumulate vertical (" << Lo1 << ", " << lo2 << ") - ("
    //<< hi1 << ", " << Hi2 << ")\n";
    //nat lo1= Lo1, hi2= Hi2;
    nat lo1= hi1, hi2= lo2;
    while (lo1 > Lo1 && b1[lo1-1] >= lo2) lo1--;
    while (hi2 < Hi2 && b1[hi1-1] >= hi2) hi2++;
    //cout << "Accumulate vertical (" << lo1 << ", " << lo2 << ") - ("
    //<< hi1 << ", " << hi2 << ")\n";
    //cout << "  sc = " << sc << "\n";
    //cout << "  inc= " << inc << "\n";
    if (hi1 == lo1 || hi2 == lo2) return;
    if (lo1 == lo2 && hi1 == hi2 && hard_eq (tp1, tp2)) {
      if ((hi1 - lo1) * V::prec () < 16384) {
	typedef polynomial_naive X;
	X::square (p1 + lo1 + lo2, seg (tp1->p) + lo1, hi1 - lo1);
      }
      else {
	typedef typename W::No_scaled X;
	X::square (p1 + lo1 + lo2, seg (tp1->p) + lo1, hi1 - lo1);
      }
    }
    else {
      W::q_difference (p2 + lo2, seg (tp2->p) + lo2, sc, hi2 - lo2);
      if (min (hi1 - lo1, hi2 - lo2) * V::prec () < 16384) {
	typedef polynomial_naive X;
	X::mul (p1 + lo1 + lo2, seg (tp1->p) + lo1, p2 + lo2,
		hi1 - lo1, hi2 - lo2);
      }
      else {
	typedef typename W::No_scaled X;
	X::mul (p1 + lo1 + lo2, seg (tp1->p) + lo1, p2 + lo2,
		hi1 - lo1, hi2 - lo2);
      }
    }
    scale (p2, p1, lo1, hi1, lo2, hi2, inc);
    W::add (pm + lo1 + lo2, p2 + lo1 + lo2, hi1 - lo1 + hi2 - lo2 - 1);
  }

  void
  accumulate_hor (nat lo1, nat Hi1, nat Lo2, nat hi2,
		  const Exp& inc, const R& sc)
  {
    //cout << "Accumulate horizontal (" << lo1 << ", " << Lo2 << ") - ("
    //<< Hi1 << ", " << hi2 << ")\n";
    //nat hi1= Hi1, lo2= Lo2;
    nat hi1= lo1, lo2= hi2;
    while (hi1 < Hi1 && b2[hi2-1] >= hi1) hi1++;
    while (lo2 > Lo2 && b2[lo2-1] >= lo1) lo2--;
    //cout << "Accumulate horizontal (" << lo1 << ", " << lo2 << ") - ("
    //<< hi1 << ", " << hi2 << ")\n";
    if (hi1 == lo1 || hi2 == lo2) return;
    W::q_difference (p1 + lo1, seg (tp1->p) + lo1, sc, hi1 - lo1);
    if (min (hi1 - lo1, hi2 - lo2) * V::prec () < 16384) {
      typedef polynomial_naive X;
      X::mul (p2 + lo1 + lo2, p1 + lo1, seg (tp2->p) + lo2,
	      hi1 - lo1, hi2 - lo2);
    }
    else {
      typedef typename W::No_scaled X;
      X::mul (p2 + lo1 + lo2, p1 + lo1, seg (tp2->p) + lo2,
	      hi1 - lo1, hi2 - lo2);
    }
    scale (p1, p2, lo1, hi1, lo2, hi2, inc);
    W::add (pm + lo1 + lo2, p1 + lo1 + lo2, hi1 - lo1 + hi2 - lo2 - 1);
  }

  void
  accumulate (nat i, nat j) {
    //cout << "Accumulate " << i << ", " << j << "\n";
    int lo1= tp1->ab[i], hi1= tp1->ab[i+1];
    int lo2= tp2->ab[j], hi2= tp2->ab[j+1];
    Exp q1= (tp1->ex[hi1] - tp1->ex[lo1]) / (hi1 - lo1);
    Exp q2= (tp2->ex[hi2] - tp2->ex[lo2]) / (hi2 - lo2);
    Exp dq= q1 - q2;
    R   sc= V::get_power (-abs (dq));
    if (i == N(tp1->ab) - 2) hi1++;
    if (j == N(tp2->ab) - 2) hi2++;
    //cout << "  q1= " << q1 << "\n";
    //cout << "  q2= " << q2 << "\n";
    //cout << "  dq= " << dq << "\n";
    //cout << "  sc= " << sc << "\n";
    if (dq >= 0) {
      double ww= as_double (V::disc() >> 1) / (as_double (dq) + 1.0e-100);
      int w= (int) min ((double) (hi2 - lo2), max (ww, 1.0));
      for (int ix2= lo2; ix2 < hi2; ix2 += w)
	accumulate_ver (lo1, hi1, ix2, min (ix2 + w, hi2), q1, sc);
    }
    else {
      double ww= as_double (V::disc() >> 1) / as_double (-dq);
      int w= (int) min ((double) (hi1 - lo1), max (ww, 1.0));
      for (int ix1= lo1; ix1 < hi1; ix1 += w)
	accumulate_hor (ix1, min (ix1 + w, hi1), lo2, hi2, q2, sc);
    }
  }

  inline Torn_polynomial
  multiply () {
    nat j_min= 0, j_max= 0;
    for (nat i=0; i<N(tp1->ab)-1; i++) {
      while (top_negligible (i, j_min))	j_min++;
      while (j_max < N(tp2->ab)-1 && !bot_negligible (i, j_max)) j_max++;
      for (nat j= j_min; j < j_max; j++) accumulate (i, j);
    }
    Polynomial p (pm, nn);
    return Torn_polynomial (p, tp1->val + tp2->val, ex);
  }

  inline Torn_polynomial
  square () {
    nat j_max= 1;
    for (nat i=0; i<N(tp1->ab)-1; i++) {
      while (j_max < N(tp2->ab)-1 && !bot_negligible (i, j_max)) j_max++;
      for (nat j= i+1; j < j_max; j++) accumulate (i, j);
    }
    W::Vec::add (pm, pm, nn);
    for (nat i=0; i<N(tp1->ab)-1; i++)
      accumulate (i, i);
    Polynomial p (pm, nn);
    return Torn_polynomial (p, tp1->val + tp2->val, ex);
  }
};

TMPL Torn_polynomial
operator * (const Torn_polynomial& tp1, const Torn_polynomial& tp2) {
  if (N (tp1->p) == 0 || N (tp2->p) == 0)
    return Torn_polynomial (Polynomial (0));
  if (N (tp1->p) == 1)
    return Torn_polynomial (tp1->p * tp2->p, tp1->val + tp2->val,
			    tp2->ex + tp1->ex[0]);
  if (N (tp2->p) == 1)
    return Torn_polynomial (tp1->p * tp2->p, tp1->val + tp2->val,
			    tp1->ex + tp2->ex[0]);
  torn_multiplier<C,V> tm (tp1, tp2);
  return tm.multiply ();
}

TMPL Torn_polynomial
square (const Torn_polynomial& tp) {
  if (N (tp->p) == 0)
    return Torn_polynomial (Polynomial (0));
  if (N (tp->p) == 1)
    return Torn_polynomial (square (tp->p), 2 * tp->val, 2 * tp->ex);
  torn_multiplier<C,V> tm (tp, tp);
  return tm.square ();
}

/******************************************************************************
* Graeffe and related operations
******************************************************************************/

TMPL Torn_polynomial
lshiftz (const Torn_polynomial& tp, const int& sh= 1) {
  int new_val= ((int) tp->val) + sh;
  if (new_val >= 0)
    return Torn_polynomial (tp->p, (nat) new_val, tp->ex);
  if (new_val + ((int) N(tp->p)) <= 0)
    return Torn_polynomial (Polynomial (0));
  return Torn_polynomial (range (tp->p , (nat) (-new_val), N (tp->p)), 0,
			  range (tp->ex, (nat) (-new_val), N (tp->p)));
}

TMPL Torn_polynomial
extract_mod (const Torn_polynomial& tp, nat i, nat p) {
  nat ii= (tp->val + i) % p;
  nat vv= (tp->val + i) / p;
  return Torn_polynomial (extract_mod (tp->p , ii, p), vv,
			  extract_mod (tp->ex, ii, p));
}

TMPL Torn_polynomial
graeffe (const Torn_polynomial& tp) {
  return
    square (extract_mod (tp, 0, 2)) -
    lshiftz (square (extract_mod (tp, 1, 2)));
}

#undef TMPL_DEF
#undef TMPL
#undef Torn
#undef Torn_vector
#undef Torn_polynomial
#undef Torn_polynomial_rep
#undef W
#undef Polynomial
#undef Exp
} // namespace mmx
#endif // __MMX_TORN_POLYNOMIAL_HPP
