
/******************************************************************************
* MODULE     : torn_tangent.hpp.hpp
* DESCRIPTION: Torn polynomials for fast numerically stable arithmetic
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_TORN_TANGENT_HPP
#define __MMX_TORN_TANGENT_HPP
#include <numerix/tangent.hpp>
#include <jorix/torn_polynomial.hpp>
namespace mmx {
template<typename C> struct standard_torn;
#define TMPL_DEF template<typename C, typename V= standard_torn<C> >
#define TMPL template<typename C, typename V>
#define Torn torn<C,V>
#define Torn_vector vector<Torn >
#define Torn_polynomial torn_polynomial<C,V>
#define Torn_polynomial_rep torn_polynomial_rep<C,V>
#define W V::PV
#define Polynomial typename V::PT
#define Exp typename V::ET
TMPL class torn_polynomial_rep;
TMPL class torn_polynomial;
TMPL Torn_vector compute_roots (const Torn_polynomial& tp, nat s, nat ss);

/******************************************************************************
* Variants for torn polynomials
******************************************************************************/

STMPL
struct standard_torn<tangent<complex<floating<> > > > {
  typedef tangent<complex<floating<> > > C;
  typedef C R;
  typedef polynomial_variant_helper<C>::PV PV;
  typedef polynomial<C,PV> PT;
  typedef rational ET;
  static inline ET get_exponent (const C& c) {
    return ET (exponent (c)); }
  static inline R get_power (const ET& e) {
    return C (exp2 (floating<> (e))); }
  static inline C shift (const C& c, const ET& e) {
    return c * get_power (e); }
  static inline ET get_magnitude (const C& c) {
    double mag= magnitude (c);
    return ET ((int) floor (66536 * mag)) / 65536; }
  static inline nat prec () { return mmx_bit_precision; }
  static inline nat disc () { return mmx_discrepancy; }
};

/******************************************************************************
* Root finding
******************************************************************************/

double ok_digits= 10000000;

TMPL Torn_vector
linear_roots (const Torn_polynomial& tp) {
  ASSERT (N(tp->p) == 2, "polynomial should have degree one");
  return vec<Torn > (Torn (-tp->p[0] / tp->p[1], tp->ex[0] - tp->ex[1]));
}

TMPL Torn_vector
quadratic_roots (const Torn_polynomial& tp) {
  ASSERT (N(tp->p) == 3, "polynomial should have degree two");
  ASSERT (N(tp->ab) == 2, "single slope expected");
  C a= tp->p[2], b= tp->p[1], c= tp->p[0];
  C d= sqrt (square (b) - 4*a*c);
  C x1= -b + d, x2= -b - d, x= (abs (x1) > abs (x2)? x1: x2);
  C z1= x / (2 * a), z2= (2 * c) / x;
  Torn v1 (z1, tp->ex[0] - tp->ex[1]);
  Torn v2 (z2, tp->ex[0] - tp->ex[1]);
  if (v1.e < v2.e || (v1.e == v2.e && abs (z1) < abs (z2)))
    return vec<Torn > (v1, v2);
  else
    return vec<Torn > (v2, v1);
}

TMPL Torn_vector
post_treat (const Torn_vector& tv, nat step) {
  typedef Scalar_type(C) K;
  integer n= integer (1) << step;
  Torn_vector r= fill<Torn > (N (tv));
  for (nat i=0; i<N(tv); i++) {
    C   z= tv[i].x;
    Exp e= tv[i].e;
    K   c= base (z);
    K   d= slope (z);
    K   x= K (n) * c / d;
    r[i]= Torn (C (x));
    Torn q= pow (tv[i] / pow (r[i], n), rational (1) / rational (n));
    double ok= -magnitude (base (untwist (q)) - 1);
    ok_digits= min (ok_digits, ok);
    cout << n << ", " << ok << ", "
	 << Re (base (untwist (tv[i] / pow (r[i], n)))) << "\n";
  }
  return r;
}

TMPL Torn_vector
subdivide_roots (const Torn_polynomial& tp, nat step, nat steps) {
  Torn_vector tv;
  nat start= 0;
  while (start+1 < N(tp->ab)) {
    nat end= start+1;
    while (end+1 < N(tp->ab)) {
      nat i1 = tp->ab[end-1];
      nat i2 = tp->ab[end];
      nat i3 = tp->ab[end+1];
      Exp sl1= (tp->ex[i2] - tp->ex[i1]) / (i2 - i1);
      Exp sl2= (tp->ex[i3] - tp->ex[i2]) / (i3 - i2);
      if (sl1 - sl2 >= V::prec () + V::disc ()) break;
      end++;
    }
    tv << compute_roots (range (tp, tp->ab[start], tp->ab[end] + 1),
			 step, steps);
    start= end;
  }
  return tv;
}

TMPL Torn_vector
graeffe_roots (const Torn_polynomial& tp, nat step, nat steps) {
  return subdivide_roots (graeffe (tp), step + 1, steps);
}

TMPL Torn_vector
compute_roots (const Torn_polynomial& tp, nat step, nat steps) {
  //cout << "Roots " << step << "\n";
  //cout << "Roots " << step << ", " << tp << "\n";
  if (N (tp->p) == 2)
    return post_treat (linear_roots (tp), step);
  if (N (tp->p) == 3 && N (tp->ab) == 2)
    return post_treat (quadratic_roots (tp), step);
  ASSERT (step <= steps, "multiple roots not yet treated");
  return graeffe_roots (tp, step, steps);
}

TMPL Torn_vector
compute_roots (const Torn_polynomial& tp) {
  double credit=
    (3.0 * log (1.0 + N(tp->p)) + 2.0 * log (V::prec ())) / log (2.0);
  Torn_vector tv= compute_roots (tp, 0, (nat) credit);
  return tv;
}

template<typename C> vector<C>
compute_roots (const polynomial<C>& p) {
  typedef tangent<C> TC;
  typedef typename Polynomial_variant(TC) TV;
  ok_digits= mmx_bit_precision;
  polynomial<C> dp= -derive (p);
  nat alloc_c= aligned_size<TC,TV> (N(p));
  TC* c= mmx_new<TC> (alloc_c);
  for (nat i=0; i<N(p); i++)
    c[i]= TC (p[i], i<N(p)-1? dp[i]: C(0));
  torn_polynomial<TC> tp (polynomial<TC> (c, N(p), alloc_c));
  const torn<TC> sh (TC (C (1.234e-3, 2.345e-3)));
  // TODO: rather choose w.r.t. one of the smallest slopes
  tp= shift (tp, sh);
  vector<torn<TC> > v= compute_roots (tp);
  v = v + sh;
  vector<C> r= fill<C> (N(v));
  for (nat i=0; i<N(v); i++)
    r[i]= base (untwist (v[i]));
  cout << "Correct digits= " << ok_digits << "\n";
  cout << "Lost    digits= " << mmx_bit_precision - ok_digits << "\n";
  return r;
}

#undef TMPL_DEF
#undef TMPL
#undef Torn
#undef Torn_vector
#undef Torn_polynomial
#undef Torn_polynomial_rep
#undef W
#undef Polynomial
#undef Exp
} // namespace mmx
#endif // __MMX_TORN_TANGENT_HPP
