
/******************************************************************************
* MODULE     : environment.hpp
* DESCRIPTION: Environments for the base evaluator
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __ENVIRONMENT_HPP
#define __ENVIRONMENT_HPP
#include <basix/string.hpp>
#include <basix/list.hpp>
#include <basix/table.hpp>
#include <basix/lisp_syntax.hpp>
#include <basix/routine.hpp>

//! \file

namespace mmx {

class environment_rep;
class environment {
INDIRECT_PROTO (environment, environment_rep)
  inline environment ();
  inline bool operator == (const environment& env) const;
  inline bool operator != (const environment& env) const;
  inline generic operator [] (const generic& x) const;
  inline generic& operator [] (const generic& x);
  friend inline list<string> strings_for_completion (const environment& env);
  friend inline void verify_if_unknown_types (const environment& env);
  friend inline void reset (const environment& env, const generic& x);
  friend inline bool is_nil (const environment& env);
  friend inline nat hash (const environment& env);
};

class environment_rep: public rep_struct {
public:
  table<generic,generic,eq_table> bindings;
  environment next;
  nat serial;       // time stamp for validity of converter cache
  nat next_serial;  // required time stamp for parent
  nat env_type;     // tag for special environments (modules and classes)
public:
  inline environment_rep ();
  inline environment_rep (const environment& e);
  inline ~environment_rep ();

  // environment access
  generic name () const;
  inline generic& set (const generic& var);
  inline void set (const generic& var, const generic& val);
  inline void reset (const generic& var);
  bool contains (const generic& var) const;
  generic get (const generic& var) const;
  bool get (const generic& var, generic& val) const;
  list<string> strings_for_completion () const;
  void verify_if_unknown_types () const;

  // converters
  void ensure_up_to_date () const;
  void set_converter (nat src, nat dest, const generic& val, nat trv, nat pen);
  generic get_converter (nat src, nat dest, nat& pen) const;

  friend class environment;
};
INDIRECT_NULL_IMPL (environment, environment_rep)

inline environment_rep::environment_rep ():
  serial (0), next_serial (0), env_type (0) {}
inline environment_rep::environment_rep (const environment& e):
  next (e), serial (0), next_serial (next->serial), env_type (0) {}
inline environment_rep::~environment_rep () {}
inline generic& environment_rep::set (const generic& var) {
  return bindings [var]; }
inline void environment_rep::set (const generic& var, const generic& val) {
  bindings [var]= val; }
inline void environment_rep::reset (const generic& var) {
  mmx::reset (bindings, var); }

inline environment::environment (): rep (NULL) {}
inline bool environment::operator == (const environment& env) const {
  return rep == env.rep; }
inline bool environment::operator != (const environment& env) const {
  return rep != env.rep; }
inline generic environment::operator [] (const generic& x) const {
  return rep->get (x); }
inline generic& environment::operator [] (const generic& x) {
  return rep->set (x); }
inline generic read (const environment& env, const generic& x) {
  return env[x]; }
inline void reset (const environment& env, const generic& x) {
  return env.rep->reset (x); }

//! Return the list of all the litterals which are entries of the bindings
inline list<string> strings_for_completion (const environment& env) {
  return env.rep->strings_for_completion (); }
  
//! Check if Unknown type occurs in signatures
inline void verify_if_unknown_types (const environment& env) {
  return env.rep->verify_if_unknown_types (); }
  
inline bool is_nil (const environment& env) {
  return env.rep == NULL; }
inline nat hash (const environment& env) {
  return as_hash (env.rep); }
inline syntactic flatten (const environment& env) {
  if (is_nil (env)) return as_syntactic (GEN_NIL);
  else return as_syntactic (env->name ()); }

PRINT_SUGAR(,environment)

inline environment global_environment () {
  return new environment_rep (); }
inline environment local_environment (const environment& next) {
  return new environment_rep (next); }

routine overloaded_routine (const generic& name, const environment& env);
generic equalize (const generic& fun, const vector<generic>& args);

} // namespace mmx
#endif // __ENVIRONMENT_HPP
