
/******************************************************************************
* MODULE     : block_polynomial.hpp
* DESCRIPTION: Multivariate polynomials using a block representation
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_BLOCK_POLYNOMIAL_HPP
#define __MMX_BLOCK_POLYNOMIAL_HPP
#include <algebramix/polynomial.hpp>
#include <basix/table.hpp>
#include <basix/list.hpp>
namespace mmx {
struct naive_block_polynomial: vector_naive {};
#define TMPL_DEF template<typename C, typename V= naive_block_polynomial>
#define TMPL template<typename C, typename V>
#define Polynomial block_polynomial<C,V>
#define Polynomial_rep block_polynomial_rep<C,V>
#define Monomial vector<int>
TMPL class block_polynomial;
TMPL inline nat N (const Polynomial& p);
TMPL inline nat dim (const Polynomial& p);
TMPL inline vector<nat> ord (const Polynomial& p);
TMPL inline C* seg (const Polynomial& p);

/******************************************************************************
* Multivariate block polynomials
******************************************************************************/

TMPL_DEF
class block_polynomial_rep REP_STRUCT {
protected:
  C*          coeffs;
  vector<nat> orders;

public:
  inline block_polynomial_rep (C* coeffs2, const vector<nat>& orders2):
    coeffs (coeffs2), orders (orders2) {}
  inline block_polynomial_rep (const vector<nat>& orders2):
    coeffs (mmx_new<C> (big_mul (orders2), 0)), orders (orders2) {}
  inline ~block_polynomial_rep () {
    mmx_delete<C> (coeffs, big_mul (orders)); }

  friend class Polynomial;
  friend nat N LESSGTR (const Polynomial& p);
  friend nat dim LESSGTR (const Polynomial& p);
  friend vector<nat> ord LESSGTR (const Polynomial& p);
  friend C* seg LESSGTR (const Polynomial& p);
};

TMPL_DEF
class block_polynomial {
INDIRECT_PROTO_2 (block_polynomial, block_polynomial_rep, C, V)
public:
  inline block_polynomial ():
    rep (new Polynomial_rep (vec<nat> ())) {}
  inline block_polynomial (const C& c):
    rep (new Polynomial_rep (vec<nat> ())) { rep->coeffs[0]= c; }
  inline block_polynomial (C* coeffs, const vector<nat>& orders):
    rep (new Polynomial_rep (coeffs, orders)) {}
  inline block_polynomial (const vector<nat>& orders):
    rep (new Polynomial_rep (orders)) {}
  C operator [] (const Monomial& m) const;
  C& operator [] (const Monomial& m);
  Polynomial operator [] (nat i) const;
  //friend nat deg (const Polynomial& s);
  //friend C* seg (const Polynomial& s);
};
INDIRECT_IMPL_2 (block_polynomial, block_polynomial_rep,
		 typename C, C, typename V, V)

/******************************************************************************
* Access
******************************************************************************/

TMPL inline nat N (const Polynomial& p) { return big_mul (p->orders); }
TMPL inline nat dim (const Polynomial& p) { return N (p->orders); }
TMPL inline vector<nat> ord (const Polynomial& p) { return p->orders; }
TMPL inline C* seg (const Polynomial& p) { return p->coeffs; }

inline nat
block_index (const Monomial& m, const vector<nat>& orders) {
  nat d= N(orders), s= 0, p= 1;
  ASSERT (N(m) == d, "dimensions don't match");
  for (nat k=0; k<d; k++) {
    s += m[k] * p;
    p *= orders[k];
  }
  return s;
}

TMPL C
Polynomial::operator [] (const Monomial& m) const {
  return rep->coeffs [block_index (m, rep->orders)];
}

TMPL C&
Polynomial::operator [] (const Monomial& m) {
  secure ();
  return rep->coeffs [block_index (m, rep->orders)];
}

TMPL Polynomial
Polynomial::operator [] (nat i) const {
  nat d= N (rep->orders);
  ASSERT (d>0, "dimension must be non-zero");
  vector<nat> orders= range (rep->orders, 0, d-1);
  nat p= big_mul (orders);
  C* coeffs= mmx_new<C> (p);
  V::copy (coeffs, rep->coeffs + i*p, p);
  return Polynomial (coeffs, orders);
}

/******************************************************************************
* User interface
******************************************************************************/

template<typename Op, typename C, typename V> nat
unary_hash (const Polynomial& p) {
  register nat i, h= 7122007, n= N(p);
  for (i=0; i<n; i++)
    h= (h<<1) ^ (h<<3) ^ (h>>29) ^ Op::op (seg(p)[i]);
  return h;
}

TMPL inline nat
hash (const Polynomial& p) {
  return unary_hash<hash_op> (p);
}

TMPL syntactic
flatten (const Polynomial& p) {
  nat d= dim (p);
  if (d == 0) return flatten (seg (p) [0]);
  else {
    syntactic sum= 0;
    for (nat i=0; i<ord(p)[d-1]; i++)
      sum= sum + flatten (p[i]) *
	         pow (access (syntactic ("x"), flatten (d)), flatten (i));
    return sum;
  }
}

PRINT_SUGAR(TMPL,Polynomial)

/******************************************************************************
* Arithmetic operations
******************************************************************************/

TMPL Polynomial
operator - (const Polynomial& p) {
  nat n= N(p);
  C* r= mmx_new<C> (n);
  V::neg (r, seg (p), n);
  return Polynomial (r, ord (p));
}

TMPL Polynomial
operator + (const Polynomial& p1, const Polynomial& p2) {
  ASSERT (ord (p1) == ord (p2), "sizes don't match");
  nat n= N(p1);
  C* r= mmx_new<C> (n);
  V::add (r, seg (p1), seg (p2), n);
  return Polynomial (r, ord (p1));
}

TMPL Polynomial
operator - (const Polynomial& p1, const Polynomial& p2) {
  ASSERT (ord (p1) == ord (p2), "sizes don't match");
  nat n= N(p1);
  C* r= mmx_new<C> (n);
  V::sub (r, seg (p1), seg (p2), n);
  return Polynomial (r, ord (p1));
}

template<typename C> void
accumulate (C* dest, const vector<nat>& ordr,
	    const C* src1, const vector<nat>& ord1,
	    const C* src2, const vector<nat>& ord2)
{
  if (N (ordr) == 0)
    dest[0] += src1[0] * src2[0];
  else {
    nat d= N(ordr);
    vector<nat> sordr= range (ordr, 0, d-1);
    vector<nat> sord1= range (ord1, 0, d-1);
    vector<nat> sord2= range (ord2, 0, d-1);
    nat pr= big_mul (sordr);
    nat p1= big_mul (sord1);
    nat p2= big_mul (sord2);
    for (nat i=0; i<ord1[d-1]; i++)
      for (nat j=0; j<ord2[d-1]; j++)
	accumulate (dest + (i+j)*pr, sordr,
		    src1 + i*p1    , sord1,
		    src2 + j*p2    , sord2);
  }
}

TMPL Polynomial
operator * (const Polynomial& p1, const Polynomial& p2) {
  ASSERT (dim (p1) == dim (p2), "dimensions don't match");
  nat d= dim (p1);
  if (N (p1) == 0 || N (p2) == 0) return Polynomial (vec<nat> (d));
  vector<nat> ord1= ord (p1);
  vector<nat> ord2= ord (p2);
  vector<nat> ordr= ord1 + ord2 - 1;
  nat n= big_mul (ordr);
  C* r= mmx_new<C> (n, C(0));
  accumulate (r, ordr, seg (p1), ord1, seg (p2), ord2);
  return Polynomial (r, ordr);
}

#undef TMPL_DEF
#undef TMPL
#undef Polynomial
#undef Polynomial_rep
#undef Monomial
} // namespace mmx
#endif // __MMX_BLOCK_POLYNOMIAL_HPP
