
/******************************************************************************
* MODULE     : block_polynomial_floating.hpp
* DESCRIPTION: Kronecker multiplication of multivariate floating polynomials
* COPYRIGHT  : (C) 2007  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_BLOCK_POLYNOMIAL_FLOATING_HPP
#define __MMX_BLOCK_POLYNOMIAL_FLOATING_HPP
#include <analyziz/polynomial_floating.hpp>
#include <multimix/block_polynomial_integer.hpp>
namespace mmx {
#define TMPL template<typename V>
#define Floating floating<V>
#define Rnd V::rnd()

TMPL vector<Floating >
newton_scales (const Floating* src, const vector<nat>& ords) {
  nat d= N(ords), n= big_mul (ords);
  Floating** aux= mmx_new<Floating*> (d);
  for (nat i=0; i<d; i++)
    aux[i]= mmx_new<Floating > (ords[i], Floating (0));
  for (nat i=0; i<n; i++) {
    nat p= 1;
    for (nat j=0; j<d; j++, p *= (j<d ? ords[j] : (nat) 1)) {
      nat k= (i/p) % ords[j];
      aux[j][k]= max (aux[j][k], abs(src[i]));
    }
  }
  Floating* r= mmx_new<Floating > (d);
  for (nat i=0; i<d; i++) {
    r[i]= newton_scale (aux[i], ords[i]);
    mmx_delete<Floating > (aux[i], ords[i]);
  }
  mmx_delete<Floating*> (aux, d);
  return vector<Floating > (r, d);
}

TMPL void
scale (Floating* dest, const Floating* src, const vector<nat>& ords,
       const Floating& c, const vector<Floating >& a)
{
  nat d= N(ords);
  nat p= big_mul (range (ords, 0, d-1));
  if (d == 1)
    scale (dest, src, ords[0], c, a[0]);
  else {
    nat r= ords[d-1];
    vector<nat>       sords= range (ords, 0, d-1);
    vector<Floating > sa   = range (a   , 0, d-1);
    if (r == 1) scale (dest, src, sords, c, sa);
    else {
      nat      i = 1;
      Floating ai= a[d-1];
      while ((i<<1) < r) {
	ai *= ai;
	i <<= 1;
      }
      scale (dest, src, append (sords, vec<nat> (i)), c, a);
      scale (dest+i*p, src+i*p, append (sords, vec<nat> (r-i)), c*ai, a);
    }
  }
}

TMPL void
kronecker_mul (Floating* dest,
	       const Floating* src1, const vector<nat>& ords1,
	       const Floating* src2, const vector<nat>& ords2)
{
  vector<nat>       ords= ords1 + ords2 - ((nat) 1);
  nat               n1  = big_mul (ords1);
  nat               n2  = big_mul (ords2);
  nat               n   = big_mul (ords);

  if (n1 <= 4 || n2 <= 4) {
    integer* mant1= mmx_new<integer> (n1);
    integer* mant2= mmx_new<integer> (n2);
    integer* dmant= mmx_new<integer> (n);
    mp_exp_t e1, e2;
    nat bits= V::prec() + V::disc();
    decompose (mant1, e1, src1, n1, bits);
    decompose (mant2, e2, src2, n2, bits);
    kronecker_mul (dmant, mant1, ords1, mant2, ords2);
    recompose (dest, dmant, n, e1+e2);
    mmx_delete<integer> (mant1, n1);
    mmx_delete<integer> (mant2, n2);
    mmx_delete<integer> (dmant, n);
  }
  else {
    Floating*         Src1= mmx_new<Floating > (n1);
    Floating*         Src2= mmx_new<Floating > (n2);
    Floating*         Dest= mmx_new<Floating > (n);
    vector<Floating > sc1 = newton_scales (src1, ords1);
    vector<Floating > sc2 = newton_scales (src2, ords2);
    vector<Floating > sc  = sup (sc1, sc2);
    scale (Src1, src1, ords1, Floating (1), Floating (1) / sc);
    scale (Src2, src2, ords2, Floating (1), Floating (1) / sc);

    integer* mant1= mmx_new<integer> (n1);
    integer* mant2= mmx_new<integer> (n2);
    integer* dmant= mmx_new<integer> (n);
    mp_exp_t e1, e2;
    nat bits= V::prec() + V::disc();
    decompose (mant1, e1, Src1, n1, bits);
    decompose (mant2, e2, Src2, n2, bits);
    kronecker_mul (dmant, mant1, ords1, mant2, ords2);
    recompose (Dest, dmant, n, e1+e2);
    mmx_delete<integer> (mant1, n1);
    mmx_delete<integer> (mant2, n2);
    mmx_delete<integer> (dmant, n);

    scale (dest, Dest, ords, Floating (1), sc);
    mmx_delete<Floating > (Src1, n1);
    mmx_delete<Floating > (Src2, n2);
    mmx_delete<Floating > (Dest, n);
  }
}

#undef TMPL
#undef Floating
#undef Rnd
} // namespace mmx
#endif // __MMX_BLOCK_POLYNOMIAL_FLOATING_HPP
