
/******************************************************************************
* MODULE     : expander.hpp
* DESCRIPTION: Simpler but less robust implementation of transseries
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__EXPANDER__HPP
#define __MMX__EXPANDER__HPP
#include <basix/routine.hpp>
#include <numerix/real.hpp>
#include <multimix/linear_combination.hpp>
#include <multimix/power_product.hpp>
namespace mmx {

class   expander;
struct  expander_power_product;
typedef real xcoefficient;
typedef power_product<xcoefficient,generic,expander_power_product> xmonomial;
typedef pair<xmonomial,xcoefficient> xterm;
typedef linear_combination<xcoefficient,xmonomial> expansion;

/******************************************************************************
* Generalized asymptotic scales
******************************************************************************/

xmonomial infinite_xmonomial (const generic& name);
xmonomial infinitesimal_xmonomial (const generic& name);
expander log (const xmonomial& m);
xmonomial exp_xmonomial (const expander& f);
xmonomial as_xmonomial (const generic& b);
generic main_coordinate (const expander& f);
generic as_coordinate (const expander& f);
syntactic flatten (const xmonomial& m);
syntactic flatten_term (const xterm& t);
int asymptotic_sign (const xmonomial& mon);
bool flatter (const xmonomial& m1, const xmonomial& m2);
bool flatter_eq (const xmonomial& m1, const xmonomial& m2);
bool steeper (const xmonomial& m1, const xmonomial& m2);
bool steeper_eq (const xmonomial& m1, const xmonomial& m2);
expander derive (const xmonomial& m, const generic& v);
expander substitute (const xmonomial& m, const generic& v, const expander& g);
xcoefficient eval (const xmonomial& m, const generic& v,
		   const xcoefficient& c);

struct expander_power_product {
  template<typename C, typename T>
  struct helper {
    typedef symbol<table<C,T>,equal_table> rep_type;
    typedef table<C,T> table_type;
    static inline rep_type construct (const table_type& t) {
      return rep_type (t); }
    static inline table_type access (const rep_type& r) {
      return *r; }
    static inline rep_type construct (const T& v, const C& c) {
      table_type t (0); t[v]= c; return construct (t); }
    template<typename M> static inline bool
    is_dominated (const M& m1, const M& m2) {
      return asymptotic_sign (m1 / m2) <= 0; }
    template<typename M> static inline bool
    is_negligible (const M& m1, const M& m2) {
      return asymptotic_sign (m1 / m2) < 0; }
  };
};

/******************************************************************************
* Expander
******************************************************************************/

class expander_rep REP_STRUCT {
protected:
  vector<expansion> v;
public:
  inline expander_rep () {}
  inline virtual ~expander_rep () {}
  inline expander me () const;
  virtual expansion access (nat k) = 0;
  friend class expander;
};

class expander {
INDIRECT_PROTO (expander, expander_rep)
public:
  expander (const int& c= 0);
  expander (const expansion& ex);
  expander (const xcoefficient& c, const xmonomial& m= 1);
  expansion operator [] (nat k) const;
};
INDIRECT_IMPL (expander, expander_rep)

inline expander expander_rep::me () const {
  return expander (this, true); }

/******************************************************************************
* Fundamental routines
******************************************************************************/

expansion expand (const expander& f, nat order= 3);
expander lengthen (const expander& f, int extra= 1);
expander shorten (const expander& f, int less= 1);
expander stabilize (const expander& f, nat delay= 3);
syntactic flatten (const expander& f, nat order= 3, nat delay= 3);
PRINT_SUGAR(,expander)

xterm        dominant_term (const expander& f);
xcoefficient dominant_coefficient (const expander& f);
xmonomial    dominant_monomial (const expander& f);
expander     dominant_part (const expander& f);
expander     remainder_part (const expander& f);
expander     infinitesimal_part (const expander& f);
expander     bounded_part (const expander& f);
expander     infinite_part (const expander& f);
expander     constant_part (const expander& f);
expander     non_constant_part (const expander& f);

/******************************************************************************
* Operations on expanders
******************************************************************************/

expander operator + (const expander& f, const expander& g);
expander operator - (const expander& f);
expander operator - (const expander& f, const expander& g);
expander operator * (const expander& f, const expander& g);
expander operator / (const expander& f, const expander& g);
expander exp (const expander& f);
expander log (const expander& f);
expander pow (const expander& f, const xcoefficient& c);
expander pow (const expander& f, const expander& g);
expander zeta (const expander& f);
expander derive (const expander& f, const generic& v);
expander integrate (const expander& f, const generic& v);
expander substitute (const expander& f, const generic& v, const expander& g);
expander compose (const expander& f, const expander& g);
expander sum (const expander& f, const generic& v);
expander product (const expander& f, const generic& v);
expander fixed_point (const routine& fun);
xcoefficient eval (const expander& f, const generic& v, const xcoefficient& c);

/******************************************************************************
* Equality testing and ordering
******************************************************************************/

inline expansion stable_expand (const expander& f) {
  return expand (stabilize (f, 3), 3); }
inline nat hash (const expander& f) {
  return hash (stable_expand (f)); }
inline bool operator == (const expander& f, const expander& g) {
  return stable_expand (f) == 0; }
inline bool operator != (const expander& f, const expander& g) {
  return stable_expand (f) != 0; }
inline int sign (const expander& f) {
  return sign (dominant_coefficient (f)); }
HARD_TO_SOFT_IDENTITY_SUGAR(,expander)
EQUAL_INT_SUGAR(,expander,expander)
COMPARE_SUGAR(,expander)
COMPARE_INT_SUGAR(,expander,expander)

} // namespace mmx
#endif // __MMX__EXPANDER__HPP
