
/******************************************************************************
* MODULE     : lex_series_linear.hpp
* DESCRIPTION: Vectors and matrices of multivariate power series
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_LEX_SERIES_LINEAR_HPP
#define __MMX_LEX_SERIES_LINEAR_HPP
#include <multimix/lex_series.hpp>
#include <basix/vector.hpp>
#include <algebramix/matrix.hpp>
namespace mmx {
#define TMPL_DEF template<typename C, typename V, typename W>
#define TMPL template<typename C, typename V, typename W>
#define Series lex_series<C,V>
#define Series_rep lex_series_rep<C,V>
#define Monomial vector<int>
#define Precision typename V::template truncated_helper<C>::precision_type

/******************************************************************************
* Vectors of series
******************************************************************************/

#define Series_vector lex_series<vector<C,W>,V>
#define Vector_series vector<lex_series<C,V>,W>
STYPE_TO_TYPE(TMPL,as_vector_type,Series_vector,Vector_series);
#undef Vector_series
#undef Series_vector

TMPL
class vector_access_lex_series_rep: public Series_rep {
  const lex_series<vector<C,W>,V> f;
  const nat i;
public:
  vector_access_lex_series_rep
    (const lex_series<vector<C,W>,V>& f2, nat i2):
      Series_rep (f2->d), f (f2), i (i2) {}
  syntactic expression () const {
    return access (flatten (f), flatten (i)); }
  void increase_precision (const Precision& p) {
    Series_rep::increase_precision (p);
    mmx::increase_precision (f, p); }
  C compute (const Monomial& m) {
    return f[m][i]; }
};

TMPL Series
access (const lex_series<vector<C,W>,V>& f, nat i) {
  return (Series_rep*)
    new vector_access_lex_series_rep<C,V,W> (f, i);
}

TMPL vector<Series,W>
as_vector (const lex_series<vector<C,W>,V>& f) {
  vector<C,W> first= f[V::monomial (f->d)];
  nat n= N(first);
  vector<Series,W> v (Series (0), n);
  for (nat i=0; i<n; i++)
    v[i]= access (f, i);
  return v;
}

template<typename C, typename V> C
first (const vector<C,V>& v) {
  if (is_a_scalar (v)) return v.scalar ();
  else if (N(v) == 0) return 0;
  else return v[0];
}

TMPL
class vector_lex_series_rep:
  public lex_series_rep<vector<C,W>,V>
{
  const vector<Series,W> v;
public:
  vector_lex_series_rep (const vector<Series,W>& v2):
    lex_series_rep<vector<C,W>,V> (first (v2) -> d),
    v (v2) {}
  syntactic expression () const {
    return flatten (v); }
  void increase_precision (const Precision& p) {
    lex_series_rep<vector<C,W>,V>::increase_precision (p);
    for (nat i=0; i<N(v); i++)
      mmx::increase_precision (v[i], p); }
  vector<C,W> compute (const Monomial& m) {
    nat i, n= N(v);
    vector<C,W> coeff (C (0), n);
    for (i=0; i<n; i++)
      coeff[i]= v[i][m];
    return coeff; }
};

TMPL lex_series<vector<C,W>,V>
as_lex_series (const vector<Series,W>& v) {
  return (lex_series_rep<vector<C,W>,V>*)
    new vector_lex_series_rep<C,V,W> (v);
}

TMPL lex_series<vector<C,W>,V>
from_vector (const vector<Series,W>& v) {
  return (lex_series_rep<vector<C,W>,V>*)
    new vector_lex_series_rep<C,V,W> (v);
}

/******************************************************************************
* Trigonometric functions and linear differential equations
******************************************************************************/

TMPL lex_series<vector<C,W>,V>
trig (const lex_series<vector<C,W>,V>& f) {
  return unary_implicit_lex_series<trig_op> (f);
}

template<typename C,typename V> vector<Series >
cos_sin (const Series& f) {
  return as_vector (trig (as_lex_series (vec (f))));
}

template<typename C,typename V> Series
cos (const Series& f) {
  return unary_implicit_lex_series<cos_op> (f);
}

template<typename C,typename V> Series
sin (const Series& f) {
  return unary_implicit_lex_series<sin_op> (f);
}

template<typename C,typename V> Series
tan (const Series& f) {
  return unary_implicit_lex_series<tan_op> (f);
}

/******************************************************************************
* Matrices of series
******************************************************************************/

#define Series_matrix lex_series<matrix<C,W>,V>
#define Matrix_series matrix<lex_series<C,V>,W>
STYPE_TO_TYPE(TMPL,as_matrix_type,Series_matrix,Matrix_series);
#undef Matrix_series
#undef Series_matrix

TMPL
class matrix_access_lex_series_rep: public Series_rep {
  lex_series<matrix<C,W>,V> f;
  nat i, j;
public:
  matrix_access_lex_series_rep
    (const lex_series<matrix<C,W>,V>& f2, nat i2, nat j2):
      Series_rep (f2->d), f (f2), i (i2), j (j2) {}
  syntactic expression () const {
    return access (flatten (f), flatten (i), flatten (j)); }
  void increase_precision (const Precision& p) {
    Series_rep::increase_precision (p);
    mmx::increase_precision (f, p); }
  C compute (const Monomial& mon) {
    return f[mon](i,j); }
};

TMPL Series
access (const lex_series<matrix<C,W>,V>& f, nat i, nat j) {
  return (Series_rep*)
    new matrix_access_lex_series_rep<C,V,W> (f, i, j);
}

TMPL matrix<Series,W>
as_matrix (const lex_series<matrix<C,W>,V>& f) {
  matrix<C,W> first= f[V::monomial (f->d)];
  nat nr= rows (first), nc= cols (first);
  matrix<Series,W> m (Series (0), nr, nc);
  for (nat i=0; i<nr; i++)
    for (nat j=0; j<nc; j++)
      m (i, j)= access (f, i, j);
  return m;
}

template<typename C, typename V> C
first (const matrix<C,V>& m) {
  if (is_a_scalar (m)) return m.scalar ();
  else if (rows(m)*cols(m) == 0) return 0;
  else return m(0,0);
}

TMPL
class matrix_lex_series_rep:
  public lex_series_rep<matrix<C,W>,V>
{
  const matrix<Series,W> m;
public:
  matrix_lex_series_rep (const matrix<Series,W>& m2):
    lex_series_rep<matrix<C,W>,V> (first (m2) -> d),
    m (m2) {}
  syntactic expression () const {
    return flatten (m); }
  void increase_precision (const Precision& p) {
    lex_series_rep<matrix<C,W>,V>::increase_precision (p);
    for (nat i=0; i<rows(m); i++)
      for (nat j=0; j<cols(m); j++)
	mmx::increase_precision (m (i, j), p); }
  matrix<C,W> compute (const Monomial& mon) {
    nat i, j, nr= rows (m), nc= cols (m);
    matrix<C,W> coeff (C (0), nr, nc);
    for (i=0; i<nr; i++)
      for (j=0; j<nc; j++)
	coeff (i, j)= m (i, j) [mon];
    return coeff; }
};

TMPL lex_series<matrix<C,W>,V>
as_lex_series (const matrix<Series,W>& m) {
  return (lex_series_rep<matrix<C,W>,V>*)
    new matrix_lex_series_rep<C,V,W> (m);
}

TMPL lex_series<matrix<C,W>,V>
from_matrix (const matrix<Series,W>& m) {
  return (lex_series_rep<matrix<C,W>,V>*)
    new matrix_lex_series_rep<C,V,W> (m);
}

/******************************************************************************
* Differential calculus
******************************************************************************/

template<typename C, typename V> vector<Series >
grad (const Series& f) {
  nat i, d= f->d;
  vector<Series > v= fill<Series > (d);
  for (i=0; i<d; i++)
    v[i]= derive (f, i);
  return v;
}

template<typename C, typename V> Series
div (const vector<Series >& v) {
  nat i, d= N(v);
  Series sum= 0;
  for (i=0; i<d; i++)
    sum= sum + derive (v[i], i);
  return sum;
}

template<typename C, typename V> vector<Series >
curl (const vector<Series >& v) {
  ASSERT (N(v) == 3, "vector of dimension three expected");
  return vec<Series > (derive (v[2], 1) - derive (v[1], 2),
		       derive (v[0], 2) - derive (v[2], 0),
		       derive (v[1], 0) - derive (v[0], 1));
}

template<typename C, typename V> Series
lapl (const Series& f) {
  nat i, d= f->d;
  Series sum= 0;
  for (i=0; i<d; i++)
    sum= sum + derive (derive (f, i), i);
  return sum;
}

#undef TMPL_DEF
#undef TMPL
#undef Series
#undef Series_rep
#undef Monomial
#undef Precision
} // namespace mmx
#endif // __MMX_LEX_SERIES_LINEAR_HPP
