
/******************************************************************************
* MODULE     : lex_series_relaxed.hpp
* DESCRIPTION: Relaxed multivariate power series
* COPYRIGHT  : (C) 2007  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_LEX_SERIES_RELAXED__HPP
#define __MMX_LEX_SERIES_RELAXED__HPP
#include <multimix/lex_series.hpp>
namespace mmx {
struct lex_series_naive;
#define Lex_series_variant(C) \
  lex_series_variant_helper<C>::SV
#define TMPL template<typename C, typename V>
#define Series lex_series<C,V>
#define Series_rep lex_series_rep<C,V>
#define Implicit_series_rep implicit_lex_series_rep<C,V>
#define Monomial vector<int>
#define Helper V::template truncated_helper<C>
#define Truncated typename Helper::truncated_type
#define Precision typename Helper::precision_type

/******************************************************************************
* Relaxed multiplication
******************************************************************************/

TMPL void
read (C* dest, const vector<nat>& sz, const Series& f, const Monomial& m) {
  Monomial lim (sz);
  for (iterator<Monomial > it= below (lim); busy (it); ++it)
    dest[block_index (*it, lim)]= f[V::multiply (m, *it)];
}

TMPL void
add (Truncated& dest, const Monomial& m, const C* src, const vector<nat>& sz) {
  Monomial lim (sz);
  for (iterator<Monomial > it= below (lim); busy (it); ++it)
    Helper::add (dest, V::multiply (m, *it), src[block_index (*it, lim)]);
}

TMPL
class relaxed_mul_lex_series_rep:
  public binary_lex_series_rep<mul_op,C,V> {
public:
  inline relaxed_mul_lex_series_rep
    (const Series& f, const Series& g):
      binary_lex_series_rep<mul_op,C,V> (f, g) {}
  void accumulate (const Monomial& m, const Monomial& m1, const Monomial& m2,
		   const vector<nat>& sz) {
    nat d= N(m);
    //cout << "Accumulate " << m1 << ", " << m2 << ", " << sz << lf << indent;
    if (d == 0) {
      nat ns= big_mul (sz);
      nat nd= big_mul (sz + sz - ((nat) 1));
      C* p1= mmx_new<C> (ns);
      C* p2= mmx_new<C> (ns);
      C* pr= mmx_new<C> (nd);
      if (before (m1, m2)) {
	read (p1, sz, this->f, m1);
	if (!V::equal (p1, C(0), ns)) {
	  read (p2, sz, this->g, m2);
	  kronecker_mul (pr, p1, sz, p2, sz);
	  add<C,V> (this->coeffs, V::multiply (m1, m2),
		    pr, sz + sz - ((nat) 1));
	}
	//else cout << "First zero" << lf;
      }
      else {
	read (p2, sz, this->g, m2);
	if (!V::equal (p2, C(0), ns)) {
	  read (p1, sz, this->f, m1);
	  kronecker_mul (pr, p1, sz, p2, sz);
	  add<C,V> (this->coeffs, V::multiply (m1, m2),
		    pr, sz + sz - ((nat) 1));
	}
	//else cout << "Second zero" << lf;
      }
      mmx_delete<C> (p1, ns);
      mmx_delete<C> (p2, ns);
      mmx_delete<C> (pr, nd);
    }
    else {
      nat k= m[d-1];
      Monomial    sm = range (m, 0, d-1);
      Monomial    sm1= cons (0, m1);
      Monomial    sm2= cons (0, m2);
      vector<nat> ssz= cons ((nat) 1, sz);
      if (k == 0) accumulate (sm, sm1, sm2, ssz);
      else {
	sm1[0]= 0; sm2[0]= k; ssz[0]= 1;
	accumulate (sm, sm1, sm2, ssz);
	sm1[0]= k; sm2[0]= 0; ssz[0]= 1;
	accumulate (sm, sm1, sm2, ssz);
	if ((k&1) == 0) {
	  nat p= 0;
	  k += 2;
	  while ((k&1) == 0 && k != 2) {
	    p++;
	    k= k >> 1;
	    sm1[0]= (1<<p)-1; sm2[0]= ((k-1)<<p)-1; ssz[0]= 1<<p;
	    accumulate (sm, sm1, sm2, ssz);
	    if (k == 2) continue;
	    sm1[0]= ((k-1)<<p)-1; sm2[0]= (1<<p)-1; ssz[0]= 1<<p;
	    accumulate (sm, sm1, sm2, ssz);
	  }
	}
      }
    }
    //cout << unindent;
  }

  C compute (const Monomial& m) {
    accumulate (m, vec<int> (), vec<int> (), vec<nat> ());
    Helper::commit (this->coeffs, m);
    return Helper::get (this->coeffs, m);
  }
};

TMPL Series
relaxed_mul (const Series& f, const Series& g) {
  return (Series_rep*) new relaxed_mul_lex_series_rep<C,V> (f, g);
}

/******************************************************************************
* Delayed relaxed multiplication
******************************************************************************/

TMPL
class delayed_relaxed_mul_lex_series_rep:
  public binary_lex_series_rep<mul_op,C,V> {
public:
  inline delayed_relaxed_mul_lex_series_rep
    (const Series& f, const Series& g):
      binary_lex_series_rep<mul_op,C,V> (f, g) {}
  void accumulate (const Monomial& m, const Monomial& m1, const Monomial& m2,
		   const vector<nat>& sz) {
    nat d= N(m);
    //cout << "Accumulate " << m1 << ", " << m2 << ", " << sz << lf << indent;
    if (d == 0) {
      nat ns= big_mul (sz);
      if (ns == 1) {
	if (before (m1, m2)) {
	  C c= this->f[m1];
	  if (c != 0)
	    Helper::add (this->coeffs, V::multiply (m1, m2), c * this->g[m2]);
	}
	else {
	  C c= this->g[m2];
	  if (c != 0)
	    Helper::add (this->coeffs, V::multiply (m1, m2), this->f[m1] * c);
	}
      }
      else {
	nat nd= big_mul (sz + sz - ((nat) 1));
	C* p1= mmx_new<C> (ns);
	C* p2= mmx_new<C> (ns);
	C* pr= mmx_new<C> (nd);
	if (before (m1, m2)) {
	  read (p1, sz, this->f, m1);
	  if (!V::equal (p1, C(0), ns)) {
	    read (p2, sz, this->g, m2);
	    kronecker_mul (pr, p1, sz, p2, sz);
	    add<C,V> (this->coeffs, V::multiply (m1, m2),
		      pr, sz + sz - ((nat) 1));
	  }
	  //else cout << "First zero" << lf;
	}
	else {
	  read (p2, sz, this->g, m2);
	  if (!V::equal (p2, C(0), ns)) {
	    read (p1, sz, this->f, m1);
	    kronecker_mul (pr, p1, sz, p2, sz);
	    add<C,V> (this->coeffs, V::multiply (m1, m2),
		      pr, sz + sz - ((nat) 1));
	  }
	  //else cout << "Second zero" << lf;
	}
	mmx_delete<C> (p1, ns);
	mmx_delete<C> (p2, ns);
	mmx_delete<C> (pr, nd);
      }
    }
    else {
      nat k= m[d-1];
      Monomial    sm = range (m, 0, d-1);
      Monomial    sm1= cons (0, m1);
      Monomial    sm2= cons (0, m2);
      vector<nat> ssz= cons ((nat) 1, sz);
      switch (k) {
      case 0:
	accumulate (sm, sm1, sm2, ssz);
	break;
      case 1:
	sm1[0]= 0; sm2[0]= 1; ssz[0]= 1;
	accumulate (sm, sm1, sm2, ssz);
	sm1[0]= 1; sm2[0]= 0; ssz[0]= 1;
	accumulate (sm, sm1, sm2, ssz);
	break;
      case 2:
	sm1[0]= 0; sm2[0]= 2; ssz[0]= 1;
	accumulate (sm, sm1, sm2, ssz);
	sm1[0]= 1; sm2[0]= 1; ssz[0]= 1;
	accumulate (sm, sm1, sm2, ssz);
	sm1[0]= 2; sm2[0]= 0; ssz[0]= 1;
	accumulate (sm, sm1, sm2, ssz);
	break;
      default:
	sm1[0]= 0; sm2[0]= k; ssz[0]= 1;
	accumulate (sm, sm1, sm2, ssz);
	sm1[0]= 1; sm2[0]= k-1; ssz[0]= 1;
	accumulate (sm, sm1, sm2, ssz);
	sm1[0]= k-1; sm2[0]= 1; ssz[0]= 1;
	accumulate (sm, sm1, sm2, ssz);
	sm1[0]= k; sm2[0]= 0; ssz[0]= 1;
	accumulate (sm, sm1, sm2, ssz);
	k -= 2;
	if ((k&1) == 0) {
	  nat p= 0;
	  k += 2;
	  while ((k&1) == 0 && k != 2) {
	    p++;
	    k= k >> 1;
	    sm1[0]= (1<<p); sm2[0]= ((k-1)<<p); ssz[0]= 1<<p;
	    accumulate (sm, sm1, sm2, ssz);
	    if (k == 2) continue;
	    sm1[0]= ((k-1)<<p); sm2[0]= (1<<p); ssz[0]= 1<<p;
	    accumulate (sm, sm1, sm2, ssz);
	  }
	}
      }
    }
    //cout << unindent;
  }

  C compute (const Monomial& m) {
    accumulate (m, vec<int> (), vec<int> (), vec<nat> ());
    Helper::commit (this->coeffs, m);
    return Helper::get (this->coeffs, m);
  }
};

TMPL Series
delayed_relaxed_mul (const Series& f, const Series& g) {
  return (Series_rep*)
    new delayed_relaxed_mul_lex_series_rep<C,V> (f, g);
}

/******************************************************************************
* Example: modified Euler equation from PhD. by Walter Pauls
******************************************************************************/

TMPL inline Series
xlapl (const Series& f, const C& pp, const C& pq, const C& qq) {
  return     pp * xderive (xderive (f, 0), 0) +
         2 * pq * xderive (xderive (f, 0), 1) +
             qq * xderive (xderive (f, 1), 1);
}

TMPL
class lininv_lex_series_rep: public Series_rep {
  Series f;
  C pp, pq, qq;
public:
  lininv_lex_series_rep
    (const Series& f2, const C& pp2, const C& pq2, const C& qq2):
      Series_rep (f2->d), f (f2), pp (pp2), pq (pq2), qq (qq2) {}
  syntactic expression () const {
    return apply ("lininv", flatten (f)); }
  void increase_precision (const Precision& p) {
    Series_rep::increase_precision (p);
    mmx::increase_precision (f, p); }
  C compute (const Monomial& m) {
    C k1= m[0];
    C k2= m[1];
    C mu= (1-k1-k2) * (pp*k1*k1 + 2*pq*k1*k2 + qq*k2*k2);
    ASSERT (mu != 0, "division by zero");
    return f[m] / mu; }
};

TMPL inline Series
lininv (const Series& f, const C& pp, const C& pq, const C& qq) {
  return (Series_rep*) new lininv_lex_series_rep<C,V> (f, pp, pq, qq);
}

TMPL
class euler_lex_series_rep: public Implicit_series_rep {
  C pp, pq, qq;
public:
  euler_lex_series_rep (const C& pp2, const C& pq2, const C& qq2):
    Implicit_series_rep (2), pp (pp2), pq (pq2), qq (qq2) {}
  syntactic expression () const {
    return "euler"; }
  Series initialize (Series& init, Monomial& lim) {
    init= Series (vec<int> (0, 1)) + Series (vec<int> (1, 0));
    lim = vec<int> (1, 1);
    Series me= this->me ();
    Series la= xlapl (me, pp, pq, qq);
    Series lh= delayed_relaxed_mul (xderive (me, 0), xderive (la, 1));
    Series rh= delayed_relaxed_mul (xderive (me, 1), xderive (la, 0));
    return lininv (lh - rh, pp, pq, qq);
  }
};

template<typename C> lex_series<C>
euler_series () {
  typedef typename Lex_series_variant(C) V;
  const vector<C> p= vec<C> (1, 0);
  const vector<C> q= vec<C> (0, 2);
  C pp= dot (p, p);
  C pq= dot (p, q);
  C qq= dot (q, q);
  Series_rep* rep= new euler_lex_series_rep<C,V> (pp, pq, qq);
  return implicit (Series (rep));
}

#undef TMPL
#undef Series
#undef Series_rep
#undef Implicit_series_rep
#undef Monomial
#undef Helper
#undef Truncated
#undef Precision
} // namespace mmx
#endif // __MMX_LEX_SERIES_RELAXED__HPP
