
/******************************************************************************
* MODULE     : lex_series_solve.hpp
* DESCRIPTION: Solving equations in multivariate power series
* COPYRIGHT  : (C) 2007  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_LEX_SERIES_SOLVE_HPP
#define __MMX_LEX_SERIES_SOLVE_HPP
#include <multimix/lex_series.hpp>
#include <algebramix/monomial.hpp>
namespace mmx {

/******************************************************************************
* Unknowns
******************************************************************************/

struct std_unknown {};
#define TMPL_DEF template<typename C, typename V= std_unknown>
#define TMPL template<typename C, typename V>
#define Unknown unknown<C,V>
#define Unknown_rep unknown_rep<C,V>
#define Monomial monomial<int>
#define Table table<C,Monomial >

TMPL class unknown;

TMPL_DEF
class unknown_rep REP_STRUCT {
public:
  static Monomial one;
  static table<Unknown,Monomial > eliminated;

private:
  Table t;

public:
  inline unknown_rep (): t (0) {}
  inline unknown_rep (const C& c): t (0) { t[one]= c; }
  inline unknown_rep (const C& c, const generic& g): t (0) {
    t[Monomial (g, 1)]= c; }
  inline unknown_rep (const Table& t2): t (t2) {}
  void simplify ();

  friend class Unknown;
  inline Table operator * () { return t; }
};

TMPL_DEF
class unknown {
INDIRECT_PROTO_2 (unknown, unknown_rep, C, V)
public:
  inline unknown ():
    rep (new Unknown_rep ()) {}
  inline unknown (int c):
    rep (new Unknown_rep (C (c))) {}
  inline unknown (const C& c):
    rep (new Unknown_rep (c)) {}
  inline unknown (const C& c, const generic& g):
    rep (new Unknown_rep (c, g)) {}
  inline unknown (const Table& t):
    rep (new Unknown_rep (t)) {}
  inline Table operator * () const {
    Unknown_rep* r= const_cast<Unknown_rep*> (rep);
    r->simplify ();
    return r->t; }
};
INDIRECT_IMPL_2 (unknown, unknown_rep, typename C, C, typename V, V)

TMPL Monomial Unknown_rep::one= Monomial ();
TMPL table<Unknown,Monomial > Unknown_rep::eliminated= Unknown ();

TMPL void
Unknown_rep::simplify () {
  Table u= t;
  bool again= true;
  while (again) {
    again= false;
    for (iterator<Monomial > it= entries (u); busy (it); ++it)
      if (eliminated->contains (*it)) {
	C c= u[*it];
	reset (u, *it);
	u += c * (*(eliminated[*it]));
	again= true;
	break;
      }
  }
  t= u;
}

TMPL void
constraint (const Unknown& z) {
  //cout << "Constraint: " << z << "\n";
  Table t= *z;
  simplify (t);
  if (N(t) <= 0) {
    ASSERT (N(t) == 0, "constraint cannot be met");
    return;
  }
  for (iterator<Monomial > it= entries (t); busy (it); ++it)
    if (*it != Unknown_rep::one) {
      C c= t[*it];
      reset (t, *it);
      Unknown_rep::eliminated[*it]= Unknown (-t / c);
      //cout << *it << " "; cout.flush ();
      //cout << "Eliminated: " << *it << " := " << Unknown (-t / c) << "\n";
      return;
    }
}

TMPL Unknown
operator + (const Unknown& x1, const Unknown& x2) {
  return Unknown ((*x1) + (*x2));
}

TMPL Unknown
operator - (const Unknown& x1) {
  return Unknown (-(*x1));
}

TMPL Unknown
operator - (const Unknown& x1, const Unknown& x2) {
  return Unknown ((*x1) - (*x2));
}

TMPL Unknown
operator * (const C& c, const Unknown& x1) {
  return Unknown (c * (*x1));
}

TMPL Unknown
operator * (const Unknown& x1, const C& c) {
  return Unknown ((*x1) * c);
}

TMPL Unknown
operator / (const Unknown& x1, const C& c) {
  return Unknown ((*x1) / c);
}

TMPL Unknown
operator * (const Unknown& x1, const Unknown& x2) {
  Table t1 (*x1), t2 (*x2), r (0);
  for (iterator<Monomial > it1= entries (t1); busy (it1); ++it1)
    for (iterator<Monomial > it2= entries (t2); busy (it2); ++it2)
      r[(*it1) * (*it2)] += t1[*it1] * t2[*it2];
  simplify (r);
  return Unknown (r);
}

TMPL syntactic
flatten (const Unknown& x) {
  syntactic sum= 0;
  Table t (*x);
  for (iterator<Monomial > it= entries (t); busy (it); ++it)
    sum= sum + flatten (t[*it]) * flatten (*it);
  return sum;
}

/*
TMPL
struct as_helper<C,Unknown > {
  static C
  cv (const Unknown& x) {
    Table t (*x);
    nat n= N(t);
    if (n == 0) return C(0);
    else if (n == 1 && (t->contains (Unknown_rep::one)))
      return t[Unknown_rep::one];
    else ERROR ("not a constant expression");
  }
};
*/

ARITH_INT_SUGAR(TMPL,Unknown,C);
PRINT_SUGAR(TMPL,Unknown)

#undef TMPL
#undef TMPL_DEF
#undef Unknown
#undef Unknown_rep
#undef Monomial
#undef Table

/******************************************************************************
* Unknown series
******************************************************************************/

#define TMPL template<typename C, typename V>
#define Unknown unknown<C>
#define Series lex_series<Unknown,V>
#define Series_rep lex_series_rep<Unknown,V>
#define Monomial vector<int>
#define Helper V::template truncated_helper<C>

TMPL
class unknown_lex_series_rep: public Series_rep {
  generic name;
public:
  unknown_lex_series_rep (const generic& name2, nat d):
    Series_rep (d), name (name2) {}
  syntactic expression () const { return as_syntactic (name); }
  Unknown compute (const Monomial& m) {
    nat d= this->d;
    ASSERT (N(m) == d, "dimension check failed");
    vector<generic> v= fill<generic> (d);
    for (nat i=0; i<d; i++) v[i]= as_generic (flatten (m[i]));
    return Unknown (1, access (name, v)); }
};

TMPL inline Series
unknown_lex_series (const generic& name, nat d) {
  return (Series_rep*)
    new unknown_lex_series_rep<C,V> (name, d);
}

TMPL void
constraint (const Series& f, nat deg= mmx_lex_series_precision) {
  increase_precision (f, Helper::total_degree (deg));
  if (f->d == 0) constraint (f[V::monomial ()]);
  else for (nat i=0; i<deg; i++)
    constraint (f[i], deg-i);
}

#undef TMPL
#undef Unknown
#undef Series
#undef Series_rep
#undef Monomial
#undef Helper

} // namespace mmx
#endif // __MMX_LEX_SERIES_SOLVE_HPP
