
/******************************************************************************
* MODULE     : linear_combination.hpp
* DESCRIPTION: Linear combinations
* COPYRIGHT  : (C) 2007  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__LINEAR_COMBINATION__HPP
#define __MMX__LINEAR_COMBINATION__HPP
#include <basix/table.hpp>
#include <basix/wrap.hpp>
#include <basix/vector_sort.hpp>
#include <basix/evaluator.hpp>
namespace mmx {
struct std_linear_combination;
#define TMPL_DEF \
  template<typename C, typename M, typename V= std_linear_combination>
#define TMPL template<typename C, typename M, typename V>
#define Sum linear_combination<C,M,V>
#define Term pair<M,C>
TMPL class linear_combination;

template<typename C, typename M> inline C
coeff (const Term& t) { return t.x2; }
template<typename C, typename M> inline M
mon (const Term& t) { return t.x1; }
template<typename C, typename M> inline syntactic
flatten_term (const Term& t) { return flatten (coeff(t)) * flatten (mon(t)); }
template<typename C, typename M> inline bool
operator >>= (const Term& t1, const Term& t2) { return mon (t1) >>= mon (t2); }

/******************************************************************************
* Sums
******************************************************************************/

struct std_linear_combination {
  template<typename C, typename M>
  struct helper {
    typedef table<C,M> rep_type;
    typedef table<C,M> table_type;
    static inline rep_type construct (const table_type& t) { return t; }
    static inline table_type access (const rep_type& r) { return r; }
  };
};

TMPL_DEF
class linear_combination {
public:
  typedef typename V::template helper<C,M> Helper;
  typedef typename Helper::rep_type rep_type;
  typedef typename Helper::table_type table_type;

private:
  rep_type rep;

public:
  inline linear_combination (const table_type& h):
    rep (Helper::construct (h)) {}
  inline table_type operator * () const {
    return Helper::access (rep); }
  inline linear_combination ():
    rep (Helper::construct (0)) {}
  inline linear_combination (const int& c):
    rep (Helper::construct (0)) {
      if (c != 0) rep[M (1)]= C (c); }
  inline linear_combination (const C& c):
    rep (Helper::construct (0)) {
      if (c != 0) rep[M (1)]= c; }
  inline linear_combination (const M& m):
    rep (Helper::construct (0)) {
      rep[m]= C (1); }
  template<typename K> inline linear_combination (const K& c, const M& m):
    rep (Helper::construct (0)) {
      if (c != 0) rep[m]= C (c); }
  inline C operator [] (const M& m) const {
    return Helper::access (rep) [m]; }

  inline friend nat hash (const Sum& s) { return hash (s.rep); }
  inline friend nat soft_hash (const Sum& s) { return soft_hash (s.rep); }
  inline friend nat hard_hash (const Sum& s) { return hard_hash (s.rep); }
  inline friend bool operator == (const Sum& s1, const Sum& s2) {
    return s1.rep == s2.rep; }
  inline friend bool operator != (const Sum& s1, const Sum& s2) {
    return s1.rep != s2.rep; }
  inline friend bool eq (const Sum& s1, const Sum& s2) {
    return eq (s1.rep, s2.rep); }
  inline friend bool neq (const Sum& s1, const Sum& s2) {
    return neq (s1.rep, s2.rep); }
  inline friend bool hard_eq (const Sum& s1, const Sum& s2) {
    return hard_eq (s1.rep, s2.rep); }
  inline friend bool hard_neq (const Sum& s1, const Sum& s2) {
    return hard_neq (s1.rep, s2.rep); }
};

EQUAL_INT_SUGAR(TMPL,Sum,Sum)

/******************************************************************************
* Basic routines
******************************************************************************/

TMPL inline nat
N (const Sum& s) {
  return N (*s); }

TMPL vector<Term >
sorted (const Sum& s) {
  vector<Term > v (iterate (*s));
  sort_leq<larger_equal_op> (v);
  return v;
}

TMPL syntactic
flatten (const Sum& s1, const Sum& s2) {
  vector<Term > v= sorted (s2);
  bool ok= true;
  syntactic sum= 0;
  for (nat i=0; i<N(v); i++) {
    if ((*s1)->contains (mon (v[i]))) {
      sum= sum + flatten_term (v[i]);
      ok= true;
    }
    else {
      if (ok) sum = sum + apply ("O", flatten (mon (v[i])));
      ok= false;
    }
  }
  return sum;
}

TMPL inline syntactic
flatten (const Sum& s) {
  return flatten (s, s);
}

PRINT_SUGAR(TMPL,Sum)

/******************************************************************************
* Support related functions
******************************************************************************/

TMPL Term
dominant_term (const Sum& s) {
  ASSERT (N(s) != 0, "zero does not admit a dominant term");
  iterator<Term > it= iterate (*s);
  Term dt= *it;
  for (++it; busy (it); ++it)
    if ((*it) >>= dt) dt= *it;
  return dt;
}

TMPL inline C
dominant_coefficient (const Sum& s) {
  if (N(s) == 0) return 0;
  return coeff (dominant_term (s)); }
TMPL inline M
dominant_monomial (const Sum& s) {
  return mon (dominant_term (s)); }
TMPL inline Sum
dominant_part (const Sum& s) {
  if (N (s) == 0) return s;
  Term t= dominant_term (s);
  return Sum (coeff (t), mon (t)); }
TMPL inline Sum
remainder_part (const Sum& s) {
  return s - dominant_part (s); }
TMPL inline Sum
infinitesimal_part (const Sum& s) {
  return Sum (unary_filter<is_infinitesimal_op> (*s)); }
TMPL inline Sum
bounded_part (const Sum& s) {
  return Sum (unary_filter<is_bounded_op> (*s)); }
TMPL inline Sum
infinite_part (const Sum& s) {
  return Sum (unary_filter<is_infinite_op> (*s)); }
TMPL inline Sum
constant_part (const Sum& s) {
  return Sum (unary_filter<is_constant_op> (*s)); }
TMPL inline Sum
non_constant_part (const Sum& s) {
  return Sum (unary_filter<is_non_constant_op> (*s)); }
TMPL inline Sum
common (const Sum& s1, const Sum& s2) {
  return Sum (common (*s1, *s2)); }

/******************************************************************************
* Basic arithmetic
******************************************************************************/

TMPL inline Sum
operator + (const Sum& s1, const Sum& s2) {
  return Sum ((*s1) + (*s2)); }
TMPL inline Sum
operator - (const Sum& s) {
  return Sum (-(*s)); }
TMPL inline Sum
operator - (const Sum& s1, const Sum& s2) {
  return Sum ((*s1) - (*s2)); }
TMPL inline Sum
operator * (const Sum& s1, const Sum& s2) {
  return Sum (binary_combine<mul_add_op> (*s1, *s2)); }

TMPL inline Sum
operator + (const Sum& s, const C& c) {
  return Sum ((*s) + c); }
TMPL inline Sum
operator + (const C& c, const Sum& s) {
  return Sum (c + (*s)); }
TMPL inline Sum
operator - (const Sum& s, const C& c) {
  return Sum ((*s) - c); }
TMPL inline Sum
operator - (const C& c, const Sum& s) {
  return Sum (c - (*s)); }
TMPL inline Sum
operator * (const Sum& s, const C& c) {
  return Sum ((*s) * c); }
TMPL inline Sum
operator * (const C& c, const Sum& s) {
  return Sum (c * (*s)); }

ARITH_INT_SUGAR(TMPL,Sum,Sum)

#undef TMPL_DEF
#undef TMPL
#undef Sum
#undef Term
} // namespace mmx
#endif // __MMX__LINEAR_COMBINATION__HPP
