
/******************************************************************************
* MODULE     : power_product.hpp
* DESCRIPTION: Power products
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_POWER_PRODUCT_HPP
#define __MMX_POWER_PRODUCT_HPP
#include <basix/table.hpp>
#include <basix/list.hpp>
#include <basix/wrap.hpp>
#include <basix/evaluator.hpp>
namespace mmx {
struct std_power_product;
#define TMPL_DEF \
  template<typename C, typename T= generic, typename V= std_power_product>
#define TMPL template<typename C, typename T, typename V>
#define Product power_product<C,T,V>
#define Stair list<Product >
TMPL class power_product;

/******************************************************************************
* Products
******************************************************************************/

struct std_power_product {
  template<typename C, typename T>
  struct helper {
    typedef table<C,T> rep_type;
    typedef table<C,T> table_type;
    static inline rep_type construct (const table_type& t) { return t; }
    static inline table_type access (const rep_type& r) { return r; }
    static inline rep_type construct (const T& v, const C& c) {
      table_type t (0); t[v]= c; return construct (t); }
    template<typename M> static inline bool
    is_dominated (const M& m1, const M& m2) {
      return (*m1) >= (*m2); }
    template<typename M> static inline bool
    is_negligible (const M& m1, const M& m2) {
      return (*m1) > (*m2); }
  };
};

TMPL_DEF
struct power_product {
public:
  typedef typename V::template helper<C,T> Helper;
  typedef typename Helper::rep_type rep_type;
  typedef typename Helper::table_type table_type;

private:
  rep_type rep;

public:
  inline power_product (const table_type& h):
    rep (Helper::construct (h)) {}
  inline table_type operator * () const {
    return Helper::access (rep); }
  inline power_product ():
    rep (Helper::construct (0)) {}
  inline power_product (const int& c):
    rep (Helper::construct (0)) { ASSERT (c == 1, "one expected"); }
  inline power_product (const T& v, const C& c= C(1)):
    rep (Helper::construct (v, c)) {}
  inline C operator [] (const T& v) const {
    return Helper::access (rep) [v]; }

  inline friend nat hash (const Product& p) { return hash (p.rep); }
  inline friend nat soft_hash (const Product& p) { return soft_hash (p.rep); }
  inline friend nat hard_hash (const Product& p) { return hard_hash (p.rep); }
  inline friend bool operator == (const Product& p1, const Product& p2) {
    return p1.rep == p2.rep; }
  inline friend bool operator != (const Product& p1, const Product& p2) {
    return p1.rep != p2.rep; }
  inline friend bool eq (const Product& p1, const Product& p2) {
    return eq (p1.rep, p2.rep); }
  inline friend bool neq (const Product& p1, const Product& p2) {
    return neq (p1.rep, p2.rep); }
  inline friend bool hard_eq (const Product& p1, const Product& p2) {
    return hard_eq (p1.rep, p2.rep); }
  inline friend bool hard_neq (const Product& p1, const Product& p2) {
    return hard_neq (p1.rep, p2.rep); }
};

/******************************************************************************
* Basic routines
******************************************************************************/

TMPL table<bool,T>
variables (const Product& p) {
  table<bool,T> vs;
  for (iterator<T> it= entries (*p); busy (it); ++it)
    vs[*it]= (p[*it] != 0);
  return vs;
}

TMPL T
as_variable (const Product& p) {
  T prod= 1;
  for (iterator<T> it= entries (*p); busy (it); ++it)
    prod= prod * pow (*it, construct (p[*it]));
  return prod;
}

TMPL syntactic
flatten (const Product& p) {
  syntactic prod= 1;
  for (iterator<T> it= entries (*p); busy (it); ++it)
    prod= prod * pow (as<generic> (*it), as<generic> (p[*it]));
  return flatten (prod);
}

WRAP_INDIRECT_IMPL(TMPL inline,Product)
PRINT_SUGAR(TMPL,Product)

/******************************************************************************
* Standard arithmetic
******************************************************************************/

TMPL inline Product operator * (const Product& p1, const Product& p2) {
  return Product ((*p1) + (*p2)); }
TMPL inline Product operator / (const Product& p1, const Product& p2) {
  return Product ((*p1) - (*p2)); }
TMPL inline Product operator / (int p1, const Product& p2) {
  return Product (p1) / p2; }
template<typename C, typename T, typename K, typename V> inline Product
pow (const Product& p, const K& c) {
  return Product (C(c) * (*p)); }
template<typename C, typename T, typename K, typename V> inline Product
nth_root (const Product& p, const K& c) {
  return Product ((*p) / C(c)); }

TMPL inline Product gcd (const Product& p1, const Product& p2) {
  return Product (inf (*p1, *p2)); }
TMPL inline Product lcm (const Product& p1, const Product& p2) {
  return Product (sup (*p1, *p2)); }

TMPL inline bool operator << (const Product& p1, const Product& p2) {
  return Product::Helper::is_negligible (p1, p2); }
TMPL inline bool operator <<= (const Product& p1, const Product& p2) {
  return Product::Helper::is_dominated (p1, p2); }
TMPL inline bool operator >> (const Product& p1, const Product& p2) {
  return Product::Helper::is_negligible (p2, p1); }
TMPL inline bool operator >>= (const Product& p1, const Product& p2) {
  return Product::Helper::is_dominated (p2, p1); }

TMPL inline bool operator == (const Product& p, const int& i) {
  return p == Product (i); }
TMPL inline bool operator != (const Product& p, const int& i) {
  return p != Product (i); }
TMPL inline bool operator << (const Product& p, const int& i) {
  return p << Product (i); }
TMPL inline bool operator <<= (const Product& p, const int& i) {
  return p <<= Product (i); }
TMPL inline bool operator >> (const Product& p, const int& i) {
  return p >> Product (i); }
TMPL inline bool operator >>= (const Product& p, const int& i) {
  return p >>= Product (i); }

/******************************************************************************
* Monomial substitutions
******************************************************************************/

TMPL table<Product,T>
identity_substitution (const Product& p) {
  // return identity map for variables occurring in p
  table<Product,T> r;
  for (iterator<T> it= entries (*p); busy (it); ++it)
    r[*it]= Product (*it, 1);
  return r;
}

template<typename C, typename T, typename K, typename U, typename V> Product
compose (const table<Product,U>& t, const power_product<K,U,V>& p) {
  Product r= 1;
  for (iterator<U> it= entries (*p); busy (it); ++it)
    r= r * pow (t[*it], C(p[*it]));
  return r;
}

template<typename C, typename T, typename K, typename U, typename X,
	 typename V>
table<power_product<C,T,V>,X>
compose (const table<power_product<C,T,V>,U>& t,
	 const table<power_product<K,U,V>,X>& u)
{
  table<power_product<C,T,V>,X> r;
  for (iterator<U> it= entries (*u); busy (it); ++it)
    r[*it]= compose (t, u[*it]);
  return r;
}

/******************************************************************************
* Widenings
******************************************************************************/

TMPL Product
ramify (const Product& p, const C& c) {
  Product r (1);
  for (iterator<T> it= entries (*p); busy (it); ++it)
    r= r * Product (pow (*it, 1/T(c)), c * p[*it]);
  return r;
}

template<typename C, typename T, typename K, typename U, typename V>
table<Product,T>
make_bounded (const table<power_product<K,U,V>,T>& t,
	      const Product& eps)
{
  // On input: Cartesian coordinates t and a Cartesian monomial eps
  // which represent a bounded monomial.
  // On output: a change of Cartesian coordinates in which eps is bounded
  if (eps <<= 1) return identity_substitution (eps);
  for (iterator<T> it= entries (*eps); busy (it); ++it)
    if (eps[*it] > 0) {
      Product p= eps / Product (*it, eps[*it]);
      if (compose (t, p) <<= 1) {
	table<Product,T> r= make_bounded (t, p);
	r[*it]= Product (*it, 1);
	return r;
      }
      else {
	C c= eps[*it];
	table<Product,T> r= make_bounded (t, 1/p);
	table<Product,T> s;
	for (iterator<T> it2= entries (r); busy (it2); ++it2)
	  s[*it2]= ramify (r[*it2], c);
	T z= pow (as_variable (eps), 1/T(c));
	if (z == 1) s[*it]= nth_root (compose (s, 1/p), c);
	else s[*it]= nth_root (Product (z, c) * compose (s, 1/p), c);
	return s;
      }
    }
  ERROR ("unexpected situation in make_bounded");
}

/******************************************************************************
* Stairs
******************************************************************************/

TMPL Stair
stair_cons (const Product& p, const Stair& l) {
  if (is_nil (l)) return cons (p, l);
  else if (p <<= car (l)) return l;
  else if (car (l) <<= p) return stair_cons (p, cdr (l));
  else return cons (car (l), stair_cons (p, cdr (l)));
}

TMPL Stair
stair_append (const Stair& l, const Stair& r) {
  if (is_nil (l)) return r;
  else return stair_cons (car (l), stair_append (cdr (l), r));
}

TMPL Stair
stair_mul (const Product& p, const Stair& l) {
  if (is_nil (l)) return l;
  else return cons (p * car (l), stair_mul (p, cdr (l)));
}

TMPL C
max_power_in (const Stair& l, const T& v) {
  if (is_nil (l)) return Minus_infinity(C);
  else return max (car (l) [v], max_power_in (cdr (l), v));
}

template<typename T, typename V> int
max_power_in (const list<power_product<int,T,V> >& l, const T& v) {
  if (is_nil (l)) return -((int) (((nat) (-1)) >> 1));
  else return max (car (l) [v], max_power_in (cdr (l), v));
}

TMPL Stair
project_up_to (const Stair& l, const T& v, int i) {
  if (is_nil (l)) return l;
  else {
    int j= car (l) [v];
    Stair r= project_up_to (cdr (l), v, i);
    if (j <= i) return stair_cons (car (l) / Product (v, j), r);
    else return r;
  }
}

#undef TMPL_DEF
#undef TMPL
#undef Product
#undef Stair
} // namespace mmx
#endif // __MMX_POWER_PRODUCT_HPP
