
/******************************************************************************
* MODULE     : ball.hpp
* DESCRIPTION: Balls
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_BALL_HPP
#define __MMX_BALL_HPP
#include <numerix/rounded.hpp>
namespace mmx {
#define TMPL_DEF template<typename C, \
  typename R=Default_radius_type(C),  \
  typename V=lipschitz_ball>
#define TMPL template<typename C,typename R,typename V>
#define Ball ball<C,R,V>
#define Up V::template rnd_helper<C>::UV
#define Down V::template rnd_helper<C>::DV
TMPL class ball;
TMPL inline C center (const Ball& z);
TMPL inline R radius (const Ball& z);
TMPL inline Ball operator - (const Ball& z1, const Ball& z2);

/******************************************************************************
* Ball variants
******************************************************************************/

struct lipschitz_ball {
  template<typename C>
  struct rnd_helper {
    typedef Round_up(C) UV;
    typedef Round_down(C) DV;
  };

  template<typename C,typename R> static inline void
  normalize (C& c, R& r) {
    typedef typename rnd_helper<C>::UV UV;
    VERIFY (r>=0, "negative radius");
    r= UV::template add_as<R> (r, rnd_error (c));
  }
};

struct truncated_lipschitz_ball: public lipschitz_ball {
  template<typename C,typename R> static void
  normalize (C& c, R& r) {
    typedef typename rnd_helper<C>::UV UV;
    VERIFY (r>=0, "negative radius");
    if (is_number (c) && is_number (r)) {
      if (step (c) < step (r)) {
	int p= precision (c) - mpfr_prec_round_up (step (r) - step (c));
	c= copy (c, (nat) max ((int) DEFAULT_LOW, p));
      }
    }
    else c= copy (c, DEFAULT_LOW);
    r= UV::template add_as<R> (r, rnd_error (c));
  }
};

/******************************************************************************
* The ball type
******************************************************************************/

TMPL_DEF
class ball {
MMX_ALLOCATORS
private:
  C c;
  R r;

public:
  inline void normalize () { V::normalize (c, r); }
  inline ball (): c (0), r (0) {}
  template<typename T> ball (const T& c2):
    c (as<C> (c2)), r (0) { normalize (); }
  template<typename CT,typename RT> inline ball (const ball<CT,RT,V>& z):
    c (as<C> (center (z))), r (as<R> (radius (z))) { normalize (); }
  inline ball (const C& x): c (copy (x)), r (0) { normalize (); }
  inline ball (const C& c2, const R& r2):
    c (c2), r (r2) { normalize (); }
  inline ball (const C& c2, const R& r2, bool flag):
    c (c2), r (r2) { if (flag) normalize (); }
  inline ball (const Real_type(Ball)& x, const Real_type(Ball)& y):
    c (gaussian (center (x), center (y))),
    r (Up::hypot (radius (x), radius (y))) { normalize (); }
  friend C center LESSGTR (const Ball& z);
  friend R radius LESSGTR (const Ball& z);

  inline Ball& operator <<= (int shift);
  inline Ball& operator >>= (int shift);
};

template<> struct symbolic_type_information<ball<floating<> > > {
  static const nat id= SYMBOLIC_BALL; };

STYPE_TO_TYPE(TMPL,center_type,Ball,C);
STYPE_TO_TYPE(TMPL,radius_type,Ball,R);
TMPL struct rounding_helper<Ball >: public rounding_helper<R> {};

TMPL
struct constant_helper<Ball > {
  static inline Ball nan () {
    return Ball (Nan (C), 0, false); }
  static inline Ball plus_infinity () {
    return Ball (Plus_infinity (C), 0, false); }
  static inline Ball minus_infinity () {
    return Ball (Minus_infinity (C), 0, false); }
  static inline Ball infinity () {
    return Ball (Infinity (C), 0, false); }
  static inline Ball fuzz () {
    return Ball (0, Plus_infinity (R), false); }
  static inline Ball log2 () { return Ball (Log2 (C), 0); }
  static inline Ball pi () { return Ball (Pi (C), 0); }
  static inline Ball euler () { return Ball (Euler (C), 0); }
  static inline Ball imaginary () {
    return Ball (Real_type(Ball) (0), Real_type(Ball) (1)); }
};

TMPL
struct make_ball_helper<Ball,C,R> {
  static inline Ball val (const C& c, const R& r) {
    return Ball (c, r); }
};

/******************************************************************************
* Basic routines
******************************************************************************/

TMPL inline C center (const Ball& z) {
  return z.c; }
TMPL inline R radius (const Ball& z) {
  return z.r; }
TMPL inline C lower (const Ball& z) {
  return Down::template sub_as<C> (center (z), radius (z)); }
TMPL inline C upper (const Ball& z) {
  return Up::template add_as<C> (center (z), radius (z)); }

TMPL inline Ball sharpen (const Ball& z) {
  return Ball (sharpen (center (z)), Radius_type(Ball) (0)); }
TMPL inline mp_exp_t exponent (const Ball& z) {
  return max (exponent (center (z)), exponent (radius (z))); }
TMPL inline double double_exponent (const Ball& z) {
  return max (double_exponent (center (z)), double_exponent (radius (z))); }
TMPL inline double magnitude (const Ball& z) {
  return max (magnitude (center (z)), magnitude (radius (z))); }

TMPL inline R abs_down (const Ball& z) {
  return Down::sub (Down::template abs_as<R> (center(z)), radius(z)); }
TMPL inline R abs_up (const Ball& z) {
  return Up::add (Up::template abs_as<R> (center(z)), radius(z)); }
TMPL inline R bnd_down (const Ball& z) {
  return Down::sub (Down::template as<R> (center(z)), radius(z)); }
TMPL inline R bnd_up (const Ball& z) {
  return Up::add (Up::template as<R> (center(z)), radius(z)); }

TMPL inline bool is_exact_zero (const Ball& z) {
  return is_exact_zero (center (z)) && is_exact_zero (radius (z)); }
TMPL inline bool is_number (const Ball& z) {
  return is_number (center (z)) && is_number (radius (z)); }
TMPL inline bool is_infinity (const Ball& z) {
  return is_infinity (center (z)) && is_number (radius (z)); }
TMPL inline bool is_fuzz (const Ball& z) {
  return !is_nan (center (z)) && is_infinity (radius (z)); }
TMPL inline bool is_nan (const Ball& z) {
  return is_nan (center (z)) || is_nan (radius (z)); }

template<typename Op, typename C, typename R, typename V> nat
unary_hash (const Ball& x) {
  nat h= Op::op (center (x));
  return (h<<3) ^ (h<<11) ^ (h>>29) ^ Op::op (radius (x));
}

template<typename Op, typename C, typename R, typename V> bool
binary_test (const Ball& x1, const Ball& x2) {
  return Op::op (center (x1), center (x2)) &&
         Op::op (radius (x1), radius (x2));
}

SOFT_IDENTITY_OP_SUGAR(TMPL,Ball)
HARD_IDENTITY_OP_SUGAR(TMPL,Ball)

TMPL syntactic
flatten (const Ball& z) {
  if (is_nan (z)) return Nan (syntactic);
  if (is_fuzz (z)) return Fuzz (syntactic);
  if (is_exact_zero (z)) return 0;
  return flatten_range (lower (z), upper (z));
  // return flatten ("ball", center (z), radius (z));
}

TMPL syntactic
flatten (const Ball& z, nat dd) {
  if (is_nan (z)) return Nan (syntactic);
  if (is_fuzz (z)) return Fuzz (syntactic);
  if (is_exact_zero (z)) return 0;
  return flatten_range (lower (z), upper (z), dd);
}

PRINT_SUGAR(TMPL,Ball)

/******************************************************************************
* Comparisons
******************************************************************************/

TMPL inline bool
is_zero (const Ball& z) {
  return !is_nan (center (z)) && !is_nan (radius (z)) &&
         abs (center (z)) <= radius (z);
}

TMPL inline bool
is_non_zero (const Ball& z) {
  return is_nan (center (z)) || is_nan (radius (z)) ||
         abs (center (z)) > radius (z);
}

TMPL inline int
var_sign (const Ball& z) {
  return is_zero (z)? 0: 1;
}

TMPL inline bool
is_negative_or_zero (const Ball& z) {
  return !is_nan (center (z)) && center (z) <= radius (z);
}

TMPL inline bool
is_negative (const Ball& z) {
  return !is_nan (center (z)) && center (z) < -radius (z);
}

TMPL inline bool
is_positive_or_zero (const Ball& z) {
  return !is_nan (center (z)) && center (z) >= -radius (z);
}

TMPL inline bool
is_positive (const Ball& z) {
  return !is_nan (center (z)) && center (z) > radius (z);
}

TMPL bool
operator == (const Ball& z1, const Ball& z2) {
  if (is_nan (z1) || is_nan (z2)) return is_nan (z1) && is_nan (z2);
  Ball d= z1 - z2;
  return abs (center (d)) <= radius (d);
}

TMPL bool
operator != (const Ball& z1, const Ball& z2) {
  if (is_nan (z1) || is_nan (z2)) return !(is_nan (z1) && is_nan (z2));
  Ball d= z1 - z2;
  return abs (center (d)) > radius (d);
}

TMPL inline bool
operator <= (const Ball& z1, const Ball& z2) {
  return is_negative_or_zero (z1 - z2);
}

TMPL inline bool
operator < (const Ball& z1, const Ball& z2) {
  return is_negative (z1 - z2);
}

TMPL inline bool
operator >= (const Ball& z1, const Ball& z2) {
  return is_positive_or_zero (z1 - z2);
}

TMPL inline bool
operator > (const Ball& z1, const Ball& z2) {
  return is_positive (z1 - z2);
}

EQUAL_INT_SUGAR(TMPL,Ball,Ball);
COMPARE_INT_SUGAR(TMPL,Ball,Ball);

TMPL nat
hash (const Ball& z) {
  (void) z;
  // return hash (center (z));
  return 123123123;
}

/******************************************************************************
* Other order related functions
******************************************************************************/

TMPL inline Ball
abs (const Ball& z) {
  return Ball (abs (center (z)), radius (z), false);
}

TMPL Ball
min (const Ball& z1, const Ball& z2) {
  if (z1 < z2) return z1;
  else if (z2 < z1) return z2;
  else return Ball (min (center (z1), center (z2)),
		    max (radius (z1), radius (z2)),
		    false);
}

TMPL Ball
max (const Ball& z1, const Ball& z2) {
  if (z1 > z2) return z1;
  else if (z2 > z1) return z2;
  else return Ball (max (center (z1), center (z2)),
		    max (radius (z1), radius (z2)),
		    false);
}

/******************************************************************************
* Basic arithmetic
******************************************************************************/

TMPL Ball
operator - (const Ball& z) {
  if (is_number (z) || is_infinity (z))
    return Ball (-center (z), radius (z), false);
  return z;
}

TMPL Ball
operator + (const Ball& z1, const Ball& z2) {
  if (is_number (z1) && is_number (z2))
    return Ball (center (z1) + center (z2),
		 Up::add (radius (z1), radius (z2)));
  if (is_nan (z1) || is_nan (z2)) return Nan (Ball);
  if (is_fuzz (z1) || is_fuzz (z2)) return Fuzz (Ball);
  if (is_infinity (z1) && is_infinity (z2) &&
      var_sign (center (z1)) == var_sign (-center (z2)))
    return Nan (Ball);
  return is_infinity (z1)? z1: z2;
}

TMPL Ball
operator - (const Ball& z1, const Ball& z2) {
  if (is_number (z1) && is_number (z2))
    return Ball (center (z1) - center (z2),
		 Up::add (radius (z1), radius (z2)));
  if (is_nan (z1) || is_nan (z2)) return Nan (Ball);
  if (is_fuzz (z1) || is_fuzz (z2)) return Fuzz (Ball);
  if (is_infinity (z1) && is_infinity (z2) &&
      var_sign (center (z1)) == var_sign (center (z2)))
    return Nan (Ball);
  return is_infinity (z1)? z1: -z2;
}

TMPL Ball
operator * (const Ball& z1, const Ball& z2) {
  if (is_number (z1) && is_number (z2)) {
    C c= center (z1) * center (z2);
    if (is_nan (c)) return Fuzz(Ball);
    return Ball (c, Up::add (Up::mul (abs_up (z1), radius (z2)),
			     Up::mul (radius (z1), abs_up (z2))));
  }
  if (is_nan (z1) || is_nan (z2)) return Nan (Ball);
  if (is_fuzz (z1) || is_fuzz (z2)) return Fuzz (Ball);
  if (is_infinity (z1)) {
    if (is_zero (z2)) return Fuzz (Ball);
    if (var_sign (z1) == var_sign (-z1)) return Infinity (Ball);
    return Ball (center (z1) * center (z2), 0, false);
  }
  else {
    if (is_zero (z1)) return Fuzz (Ball);
    if (var_sign (z2) == var_sign (-z2)) return Infinity (Ball);
    return Ball (center (z1) * center (z2), 0, false);
  }
}

TMPL inline Ball
square (const Ball& z) {
  if (is_number (z)) {
    C c= square (center (z));
    if (is_nan (c)) return Fuzz(Ball);
    return Ball (square (center (z)),
		 incexp2 (Up::mul (abs_up (z), radius (z)), 1));
  }
  if (is_infinity (z)) {
    if (var_sign (z) == var_sign (-z)) return Infinity (Ball);
    return Ball (square (center (z)), 0, false);
  }
  return z;
}

TMPL Ball
operator / (const Ball& z1, const Ball& z2) {
  if (is_zero (z2)) return Nan (Ball);
  if (is_number (z1) && is_number (z2))
    return z1 * Ball (1 / center (z2),
		      Up::div (radius (z2), Up::square (abs_down (z2))));
  if (is_nan (z1) || is_nan (z2)) return Nan (Ball);
  if (is_fuzz (z1) || is_fuzz (z2)) return Fuzz (Ball);
  if (is_infinity (z1)) {
    if (is_infinity (z2)) return Fuzz (Ball);
    if (var_sign (z1) == var_sign (-z1)) return Infinity (Ball);
    return Ball (center (z1) / center (z2), 0, false);
  }
  else return 0;
}

TMPL inline Ball
operator << (const Ball& z, int shift) {
  return Ball (incexp2 (center (z), shift), incexp2 (radius (z), shift), false);
}

TMPL inline Ball
operator >> (const Ball& z, int shift) {
  return Ball (decexp2 (center (z), shift), decexp2 (radius (z), shift), false);
}

TMPL inline Ball operator + (const C& z1, const Ball& z2) {
  return Ball (z1) + z2; }
TMPL inline Ball operator + (const Ball& z1, const C& z2) {
  return z1 + Ball (z2); }
TMPL inline Ball operator - (const C& z1, const Ball& z2) {
  return Ball (z1) - z2; }
TMPL inline Ball operator - (const Ball& z1, const C& z2) {
  return z1 - Ball (z2); }
TMPL inline Ball operator * (const C& z1, const Ball& z2) {
  return Ball (z1) * z2; }
TMPL inline Ball operator * (const Ball& z1, const C& z2) {
  return z1 * Ball (z2); }
TMPL inline Ball operator / (const C& z1, const Ball& z2) {
  return Ball (z1) / z2; }
TMPL inline Ball operator / (const Ball& z1, const C& z2) {
  return z1 / Ball (z2); }
TMPL inline Ball invert (const Ball& z) {
  return Ball (1) / z; }

ARITH_INT_SUGAR(TMPL,Ball,Ball);

/******************************************************************************
* In place arithmetic operations
******************************************************************************/

TMPL inline Ball&
Ball::operator <<= (int shift) {
  incexp2_assign (c, shift);
  incexp2_assign (r, shift);
  return *this;
}

TMPL inline Ball&
Ball::operator >>= (int shift) {
  decexp2_assign (c, shift);
  decexp2_assign (r, shift);
  return *this;
}

/******************************************************************************
* Elementary functions
******************************************************************************/

TMPL Ball
sqrt (const Ball& z) {
  if (is_negative_or_zero (z)) return Nan (Ball);
  return Ball (sqrt (center (z)),
	       Up::div (decexp2 (radius (z)), Down::sqrt (bnd_down (z))));
}

TMPL Ball
exp (const Ball& z) {
  return Ball (exp (center (z)),
	       Up::mul (Up::exp (bnd_up (z)), radius (z)));
}

TMPL Ball
log (const Ball& z) {
  if (is_negative_or_zero (z)) return Nan (Ball);
  return Ball (log (center (z)),
	       Up::div (radius (z), bnd_down (z)));
}

TMPL inline Ball
pow (const Ball& z, const Ball& y) {
  return exp (y * log (z));
}

TMPL Ball
cos (const Ball& z) {
  if (radius (z) >= R (3.14159))
    return Ball (cos (center (z)), radius (z));
  else {
    R u1= sin (bnd_down (z));
    R u2= sin (bnd_up (z));
    if (sign (u1) * sign (u2) != 1)
      return Ball (cos (center (z)), radius (z));
    return Ball (cos (center (z)),
		 Up::mul (max (abs (u1), abs (u2)), radius (z)));
  }
}

TMPL Ball
sin (const Ball& z) {
  if (radius (z) >= R (3.14159))
    return Ball (sin (center (z)), radius (z));
  else {
    R u1= cos (bnd_down (z));
    R u2= cos (bnd_up (z));
    if (sign (u1) * sign (u2) != 1)
      return Ball (sin (center (z)), radius (z));
    return Ball (sin (center (z)),
		 Up::mul (max (abs (u1), abs (u2)), radius (z)));
  }
}

TMPL Ball
tan (const Ball& z) {
  return sin (z) / cos (z);
}

TMPL inline Ball
cosh (const Ball& z) {
  return Ball (cosh (center (z)),
	       Up::mul (Up::sinh (bnd_up (abs (z))), radius (z)));
}

TMPL inline Ball
sinh (const Ball& z) {
  return Ball (sinh (center (z)),
	       Up::mul (Up::cosh (bnd_up (abs (z))), radius (z)));
}

TMPL Ball
tanh (const Ball& z) {
  return sinh (z) / cosh (z);
}

TMPL Ball
acos (const Ball& z) {
  Ball u= 1 - square (z);
  R v= bnd_down (u);
  if (v <= 0) return Nan (Ball);
  return Ball (acos (center (z)), Up::div (radius (z), Down::sqrt (v)));
}

TMPL Ball
asin (const Ball& z) {
  Ball u= 1 - square (z);
  R v= bnd_down (u);
  if (v <= 0) return Nan (Ball);
  return Ball (asin (center (z)), Up::div (radius (z), Down::sqrt (v)));
}

TMPL Ball
atan (const Ball& z) {
  R u= bnd_down (z);
  R v= Down::add (R (1), Down::square (u));
  if (u <= 0) return Ball (atan (center (z)), radius (z));
  else return Ball (atan (center (z)), Up::div (radius (z), v));
}

TMPL Ball
atan2 (const Ball& y, const Ball& x) {
  Ball z= square (x) + square (y);
  if (z <= 0) return Nan (Ball);
  R b = Up::div (R (1), bnd_down (z));
  R dx= Up::mul (bnd_up (abs (y)), b);
  R dy= Up::mul (bnd_up (abs (x)), b);
  return Ball (atan2 (center (x), center (y)),
	       Up::add (Up::mul (dx, radius (x)),
			Up::mul (dy, radius (y))));
}

HYPOT_SUGAR(TMPL,Ball)
INV_TRIGO_SUGAR(TMPL,Ball)
INV_HYPER_SUGAR(TMPL,Ball)
ARG_HYPER_SUGAR(TMPL,Ball)

#undef TMPL_DEF
#undef TMPL
#undef Ball
#undef Up
#undef Down
} // namespace mmx
#endif // __MMX_BALL_HPP
