
/******************************************************************************
* MODULE     : ball_complex.hpp
* DESCRIPTION: Complex balls
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_BALL_COMPLEX_HPP
#define __MMX_BALL_COMPLEX_HPP
#include <numerix/ball.hpp>
#include <numerix/complex.hpp>
namespace mmx {
#define TMPL_DEF template<typename C,typename R,typename V=lipschitz_ball>
#define TMPL template<typename C,typename R,typename V>
#define Real_ball ball<C,R,V>
#define Ball ball<complex<C>,R,V>
#define Up V::template rnd_helper<C>::UV
#define Down V::template rnd_helper<C>::DV

STYPE_TO_TYPE(TMPL,real_type,Ball,Real_ball);
STYPE_TO_TYPE(TMPL,abs_type,Ball,Real_ball);

TMPL inline int var_sign (const Ball& z) {
  if (z == 0) return 0; else return 1; }

/******************************************************************************
* Complex balls in Gaussian and polar representations
******************************************************************************/

TMPL Ball
gaussian (const Real_ball& re, const Real_ball& im= 0) {
  return Ball (gaussian (center (re), center (im)),
	       Up::hypot (radius (re), radius (im)));
}

TMPL Ball
polar (const Real_ball& r, const Real_ball& t) {
  return gaussian (r * cos (t), r * sin (t));
}

TMPL Real_ball
Re (const Ball& z) {
  return Real_ball (Re (center (z)), radius (z), false);
}

TMPL Real_ball
Im (const Ball& z) {
  return Real_ball (Im (center (z)), radius (z), false);
}

TMPL Real_ball
abs (const Ball& z) {
  if (is_infinity (z)) return Plus_infinity (Real_ball);
  return Real_ball (abs (center (z)), radius (z));
}

TMPL Real_ball
arg (const Ball& z) {
  return atan2 (Im (z), Re (z));
}

/******************************************************************************
* Further basic functions
******************************************************************************/

TMPL syntactic
flatten (const Ball& z) {
  if (is_nan (z)) return Nan (syntactic);
  if (is_fuzz (z)) return Fuzz (syntactic);
  if (is_infinity (z)) return Infinity (syntactic);
  if (is_exact_zero (z)) return 0;
  if (Re (z) != 0 && Im (z) == 0)
    return flatten (Re (z));
  if (Re (z) == 0 && Im (z) != 0)
    return flatten (Im (z)) * Imaginary (syntactic);
  return flatten (Re (z)) + flatten (Im (z)) * Imaginary (syntactic);
}

TMPL syntactic
flatten (const Ball& z, nat dd) {
  if (is_nan (z)) return Nan (syntactic);
  if (is_fuzz (z)) return Fuzz (syntactic);
  if (is_infinity (z)) return Infinity (syntactic);
  if (is_exact_zero (z)) return 0;
  if (Re (z) != 0 && Im (z) == 0)
    return flatten (Re (z), dd);
  if (Re (z) == 0 && Im (z) != 0)
    return flatten (Im (z), dd) * Imaginary (syntactic);
  return flatten (Re (z), dd) + flatten (Im (z), dd) * Imaginary (syntactic);
}

TMPL Ball
conj (const Ball& z) {
  return Ball (gaussian (Re (center (z)), -Im (center (z))),
	       radius (z), false);
}

TMPL Ball
times_i (const Ball& z) {
  return Ball (gaussian (-Im (center (z)), Re (center (z))),
	       radius (z), false);
}

TMPL Ball
over_i (const Ball& z) {
  return Ball (gaussian (Im (center (z)), -Re (center (z))),
	       radius (z), false);
}

/******************************************************************************
* Elementary functions
******************************************************************************/

TMPL Ball
sqrt (const Ball& z) {
  if (is_zero (z)) return Fuzz (Ball);
  return Ball (sqrt (center (z)),
	       Up::div (decexp2 (radius (z)), Down::sqrt (abs_down (z))));
}

TMPL Ball
exp (const Ball& z) {
  return Ball (exp (center (z)),
	       Up::mul (Up::exp (bnd_up (Re (z))), radius (z)));
}

TMPL Ball
log (const Ball& z) {
  if (is_zero (z)) return Nan (Ball);
  return Ball (log (center (z)),
	       Up::div (radius (z), abs_down (z)));
}

TMPL Ball
cos (const Ball& z) {
  return Ball (cos (center (z)),
	       Up::mul (Up::sinh (bnd_up (Im (z))), radius (z)));
}

TMPL Ball
sin (const Ball& z) {
  return Ball (sin (center (z)),
	       Up::mul (Up::cosh (bnd_up (Im (z))), radius (z)));
}

TMPL Ball
tan (const Ball& z) {
  return sin (z) / cos (z);
}

TMPL inline Ball
cosh (const Ball& z) {
  return Ball (cosh (center (z)),
	       Up::mul (Up::sinh (bnd_up (abs (Re (z)))), radius (z)));
}

TMPL inline Ball
sinh (const Ball& z) {
  return Ball (sinh (center (z)),
	       Up::mul (Up::cosh (bnd_up (abs (Re (z)))), radius (z)));
}

TMPL Ball
tanh (const Ball& z) {
  return sinh (z) / cosh (z);
}

TMPL Ball
acos (const Ball& z) {
  return over_i (acosh (z));
}

TMPL Ball
asin (const Ball& z) {
  return over_i (asinh (times_i (z)));
}

TMPL Ball
atan (const Ball& z) {
  return over_i (atanh (times_i (z)));
}

#undef TMPL_DEF
#undef TMPL
#undef Real_ball
#undef Ball
#undef Up
#undef Down
} // namespace mmx
#endif // __MMX_BALL_COMPLEX_HPP
