
/******************************************************************************
* MODULE     : complex.hpp
* DESCRIPTION: Complexified numbers
* COPYRIGHT  : (C) 2005  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_COMPLEX_HPP
#define __MMX_COMPLEX_HPP
#include <basix/string.hpp>
#include <numerix/mmx_gmp.hpp>
#include <basix/type_props.hpp>
#include <basix/operators.hpp>
namespace mmx {
#define TMPL_DEF template<typename C>
#define TMPL template<typename C>
#define Complex complex<C>
TMPL class complex;
TMPL inline C Re (const Complex& z);
TMPL inline C Im (const Complex& z);
TMPL inline C& Re (Complex& z);
TMPL inline C& Im (Complex& z);

/******************************************************************************
* The complex type
******************************************************************************/

TMPL_DEF
class complex {
MMX_ALLOCATORS;
private:
  C re;
  C im;
public:
  inline complex () {}
  template<typename T> inline complex (const T& x):
    re (as<C> (x)), im (0) {}
  template<typename T> inline complex (const complex<T>& z):
    re (Re (z)), im (Im (z)) {}
  template<typename XT,typename YT> inline
  complex (const XT& x, const YT& y):
    re (x), im (y) {}
  friend C Re LESSGTR (const Complex& z);
  friend C Im LESSGTR (const Complex& z);
  friend C& Re LESSGTR (Complex& z);
  friend C& Im LESSGTR (Complex& z);

  inline Complex& operator += (const Complex& z);
  inline Complex& operator -= (const Complex& z);
  inline Complex& operator *= (const C& x);
  inline Complex& operator /= (const C& x);
  inline Complex& operator *= (const Complex& z);
  inline Complex& operator /= (const Complex& z);
  inline Complex& operator <<= (int shift);
  inline Complex& operator >>= (int shift);
};

STYPE_TO_TYPE(TMPL,scalar_type,Complex,C);
STYPE_TO_TYPE(TMPL,real_type,Complex,C);
STYPE_TO_TYPE(TMPL,abs_type,Complex,C);
STYPE_TO_TYPE(TMPL,center_type,Complex,complex<Center_type(C) >);
STYPE_TO_TYPE(TMPL,radius_type,Complex,Radius_type(C));
STYPE_TO_TYPE(TMPL,default_radius_type,Complex,Default_radius_type(C));
STYPE_TO_TYPE(TMPL,user_radius_type,Complex,User_radius_type(C));
TMPL struct rounding_helper<Complex >: public rounding_helper<C> {};

TMPL
struct constant_helper<Complex > {
  static inline Complex nan () {
    return Complex (Nan (C), Nan (C)); }
  static inline Complex infinity () {
    return Complex (Infinity (C), Infinity (C)); }
  static inline Complex fuzz () {
    return Complex (Fuzz (C), Fuzz(C)); }
  static inline Complex log2 () { return Complex (Log2 (C), 0); }
  static inline Complex pi () { return Complex (Pi (C), 0); }
  static inline Complex euler () { return Complex (Euler (C), 0); }
  static inline Complex imaginary () { return Complex (C (0), C (1)); }
};

TMPL
struct fast_helper<Complex > {
  typedef complex<Fast_type(C) > fast_type;
  static inline fast_type dd (const Complex& z) {
    return complex<Fast_type (C) > (fast (Re (z)), fast (Im (z))); }
  static inline Complex uu (const fast_type& z) {
    return complex<C> (slow<C> (Re (z)), slow<C> (Im (z))); }
};

template<typename T,typename F>
struct as_helper<complex<T>,complex<F> > {
  static inline complex<T> cv (const complex<F>& z) {
    return complex<T> (as<T> (Re (z)), as<T> (Im (z))); }
};

/******************************************************************************
* Basic routines
******************************************************************************/

TMPL inline C Re (const Complex& z) { return z.re; }
TMPL inline C Im (const Complex& z) { return z.im; }
TMPL inline C& Re (Complex& z) { return z.re; }
TMPL inline C& Im (Complex& z) { return z.im; }

TMPL inline bool is_exact_zero (const Complex& z) {
  return is_exact_zero (Re (z)) && is_exact_zero (Im (z)); }
TMPL inline bool is_number (const Complex& z) {
  return is_number (Re (z)) && is_number (Im (z)); }
TMPL inline bool is_nan (const Complex& z) {
  return is_nan (Re (z)) || is_nan (Im (z)); }
TMPL inline bool is_infinity (const Complex& z) {
  return !is_nan (z) && (is_infinity (Re (z)) || is_infinity (Im (z))); }
TMPL inline bool is_fuzz (const Complex& z) {
  return !is_nan (z) && (is_fuzz (Re (z)) || is_fuzz (Im (z))); }

TMPL inline Complex copy (const Complex& z) {
  return Complex (copy (Re (z)), copy (Im (z))); }
TMPL inline Complex copy (const Complex& z, nat prec) {
  return Complex (copy (Re (z), prec), copy (Im (z), prec)); }
TMPL inline Complex sharpen (const Complex& z) {
  return Complex (sharpen (Re (z)), sharpen (Im (z))); }
TMPL inline nat precision (const Complex& z) {
  return max (precision (Re (z)), precision (Im (z))); }
TMPL inline mp_exp_t exponent (const Complex& z) {
  return max (exponent (Re (z)), exponent (Im (z))); }
TMPL inline mp_exp_t step (const Complex& z) {
  return max (step (Re (z)), step (Im (z))); }

template<typename Op, typename C> inline nat
unary_hash (const Complex& z) {
  nat h= Op::op (Re (z));
  return (h<<1) ^ (h<<5) ^ (h>>29) ^ Op::op (Im (z));
}

template<typename Op, typename C> bool
binary_test (const Complex& z1, const Complex& z2) {
  return Op::op (Re (z1), Re (z2)) && Op::op (Im (z1), Im (z2));
}

TRUE_IDENTITY_OP_SUGAR(TMPL,Complex)
SOFT_IDENTITY_OP_SUGAR(TMPL,Complex)
HARD_IDENTITY_OP_SUGAR(TMPL,Complex)

TMPL inline int
var_sign (const Complex& z) {
  if (z == 0) return 0; else return 1;
}

EQUAL_INT_SUGAR(TMPL,Complex,Complex);

TMPL syntactic
flatten (const Complex& z) {
  if (is_nan (z)) return Nan (syntactic);
  if (is_fuzz (z)) return Fuzz (syntactic);
  if (is_infinity (z)) return Infinity (syntactic);
  if (is_exact_zero (z)) return 0;
  if (Re (z) != 0 && Im (z) == 0)
    return flatten (Re (z));
  if (Re (z) == 0 && Im (z) != 0)
    return flatten (Im (z)) * Imaginary (syntactic);
  return flatten (Re (z)) + flatten (Im (z)) * Imaginary (syntactic);
}

TMPL syntactic
flatten (const Complex& z, nat dd) {
  if (is_nan (z)) return Nan (syntactic);
  if (is_fuzz (z)) return Fuzz (syntactic);
  if (is_infinity (z)) return Infinity (syntactic);
  if (is_exact_zero (z)) return 0;
  if (Re (z) != 0 && Im (z) == 0)
    return flatten (Re (z), dd);
  if (Re (z) == 0 && Im (z) != 0)
    return flatten (Im (z), dd) * Imaginary (syntactic);
  return flatten (Re (z), dd) + flatten (Im (z), dd) * Imaginary (syntactic);
}

PRINT_SUGAR(TMPL,Complex)

/******************************************************************************
* Polar coordinates and complex conjugation
******************************************************************************/

TMPL inline Complex
gaussian (const C& re, const C& im= 0) {
  return Complex (re, im);
}

TMPL inline Complex
polar (const C& r, const C& t) {
  return Complex (r * cos (t), r * sin (t));
}

TMPL inline C
abs (const Complex& z) {
  return hypot (Re (z), Im (z));
}

TMPL inline C
arg (const Complex& z) {
  return atan2 (Im (z), Re (z));
}

TMPL inline C
norm (const Complex& z) {
  return square (Re (z)) + square (Im (z));
}

TMPL inline Complex
conj (const Complex& z) {
  return Complex (Re (z), -Im (z));
}

TMPL inline Complex
times_i (const Complex& z) {
  return Complex (-Im (z), Re (z));
}

TMPL inline Complex
over_i (const Complex& z) {
  return Complex (Im (z), -Re (z));
}

/******************************************************************************
* Basic arithmetic
******************************************************************************/

TMPL inline Complex
operator + (const Complex& z1, const Complex& z2) {
  return Complex (Re (z1) + Re (z2), Im (z1) + Im (z2));
}

TMPL inline Complex
operator + (const Complex& z1, const C& z2) {
  return Complex (Re (z1) + z2, Im (z1));
}

TMPL inline Complex
operator + (const C& z1, const Complex& z2) {
  return Complex (z1 + Re (z2), Im (z2));
}

TMPL inline Complex
operator - (const Complex& z) {
  return Complex (-Re (z), -Im (z));
}

TMPL inline Complex
operator - (const Complex& z1, const Complex& z2) {
  return Complex (Re (z1) - Re (z2), Im (z1) - Im (z2));
}

TMPL inline Complex
operator - (const Complex& z1, const C& z2) {
  return Complex (Re (z1) - z2, Im (z1));
}

TMPL inline Complex
operator - (const C& z1, const Complex& z2) {
  return Complex (z1 - Re (z2), -Im (z2));
}

TMPL inline Complex
operator * (const C& x, const Complex& z) {
  return Complex (x * Re (z), x * Im (z));
}

TMPL inline Complex
operator * (const Complex& z, const C& x) {
  return Complex (Re (z) * x, Im (z) * x);
}

TMPL inline Complex
operator / (const Complex& z, const C& x) {
  return Complex (Re (z) / x, Im (z) / x);
}

TMPL inline Complex
operator / (const C& x, const Complex& z) {
  C a= x / norm (z);
  return Complex (a * Re (z), -a * Im (z));
}

TMPL inline Complex
operator * (const Complex& z1, const Complex& z2) {
  return Complex (Re (z1) * Re (z2) - Im (z1) * Im (z2),
		     Re (z1) * Im (z2) + Im (z1) * Re (z2));
}

TMPL inline Complex
operator / (const Complex& z1, const Complex& z2) {
  C a= norm (z2);
  return Complex ((Re (z1) * Re (z2) + Im (z1) * Im (z2)) / a,
		     (Im (z1) * Re (z2) - Re (z1) * Im (z2)) / a);
}

TMPL inline Complex
invert (const Complex& z) {
  return C(1) / z;
}

TMPL inline Complex
operator << (const Complex& z, int shift) {
  return Complex (incexp2 (Re (z), shift), incexp2 (Im (z), shift));
}

TMPL inline Complex
operator >> (const Complex& z, int shift) {
  return Complex (decexp2 (Re (z), shift), decexp2 (Im (z), shift));
}

ARITH_INT_SUGAR(TMPL,Complex,C);

/******************************************************************************
* In place arithmetic operations
******************************************************************************/

TMPL inline Complex&
Complex::operator += (const Complex& z) {
  re += z.re;
  im += z.im;
  return *this;
}

TMPL inline Complex&
Complex::operator -= (const Complex& z) {
  re -= z.re;
  im -= z.im;
  return *this;
}

TMPL inline Complex&
Complex::operator *= (const C& x) {
  re *= x;
  im *= x;
  return *this;
}

TMPL inline Complex&
Complex::operator /= (const C& x) {
  re /= x;
  im /= x;
  return *this;
}

TMPL inline Complex&
Complex::operator *= (const Complex& z) {
  C RE= re, zRE= z.re;
  re= re * z.re - im * z.im;
  im= RE * z.im + im * zRE;
  return *this;
}

TMPL inline Complex&
Complex::operator /= (const Complex& z) {
  C RE= re;
  C a = norm (z);
  re= (re * z.re + im * z.im) / a;
  im= (im * z.re - RE * z.im) / a;
  return *this;
}

TMPL inline Complex&
Complex::operator <<= (int shift) {
  re <<= shift;
  im <<= shift;
  return *this;
}

TMPL inline Complex&
Complex::operator >>= (int shift) {
  re >>= shift;
  im >>= shift;
  return *this;
}

TMPL inline void
neg (Complex& r) {
  neg (Re (r));
  neg (Im (r));
}

TMPL inline void
neg (Complex& r, const Complex& z1) {
  neg (Re (r), Re (z1));
  neg (Im (r), Im (z1));
}

TMPL inline void
add (Complex& r, const Complex& z1, const Complex& z2) {
  add (Re (r), Re (z1), Re (z2));
  add (Im (r), Im (z1), Im (z2));
}

TMPL inline void
sub (Complex& r, const Complex& z1, const Complex& z2) {
  sub (Re (r), Re (z1), Re (z2));
  sub (Im (r), Im (z1), Im (z2));
}

TMPL inline void
mul (Complex& r, const C& c, const Complex& z) {
  mul (Re (r), c, Re (z));
  mul (Im (r), c, Im (z));
}

TMPL inline void
mul (Complex& r, const Complex& z, const C& c) {
  mul (Re (r), Re (z), c);
  mul (Im (r), Im (z), c);
}

TMPL inline void
div (Complex& r, const Complex& z, const C& c) {
  div (Re (r), Re (z), c);
  div (Im (r), Im (z), c);
}

/******************************************************************************
* Exponentiation and logarithms
******************************************************************************/

TMPL Complex
sqrt (const Complex& z) {
  if (Re (z) > 0) {
    C r= sqrt ((hypot (Re (z), Im (z)) + Re (z)) / C(2));
    return Complex (r, (Im (z) / r) / C(2));
  }
  else {
    C a= hypot (Re (z), Im (z));
    if (a == 0) return 0;
    else {
      C i = sqrt ((a - Re (z)) / C(2));
      if (Im (z) < 0) i= -i;
      return Complex ((Im (z) / i) / C(2), i);
    }
  }
}

TMPL inline Complex
exp (const Complex& z) {
  return polar (exp (Re (z)), Im (z));
}

TMPL inline Complex
log (const Complex& z) {
  return Complex (log (abs (z)), arg (z));
}

TMPL inline Complex
pow (const Complex& z, const Complex& y) {
  C r= log (abs (z));
  C t= arg (z);
  return polar (exp (r * Re (y) - Im (y) * t),
		Im (y) * r + Re (y) * t);
}

TMPL inline Complex
pow (const Complex& z, const C& y) {
  return exp (y * log (z));
}

TMPL inline Complex
pow (const C& z, const Complex& y) {
  return exp (y * log (z));
}

TMPL inline Complex
cos (const Complex& z) {
  return Complex (cos (Re (z)) * cosh (Im (z)),
		     -sin (Re (z)) * sinh (Im (z)));
}

TMPL inline Complex
sin (const Complex& z) {
  return Complex (sin (Re (z)) * cosh (Im (z)),
		     cos (Re (z)) * sinh (Im (z)));
}

TMPL inline Complex
tan (const Complex& z) {
  return sin (z) / cos (z);
}

TMPL inline Complex
acos (const C& z) {
  return over_i (acosh (z));
}

TMPL inline Complex
asin (const C& z) {
  return over_i (asinh (times_i (z)));
}

TMPL inline Complex
atan (const C& z) {
  return over_i (atanh (times_i (z)));
}

TMPL inline Complex
acos (const Complex& z) {
  return over_i (acosh (z));
}

TMPL inline Complex
asin (const Complex& z) {
  return over_i (asinh (times_i (z)));
}

TMPL inline Complex
atan (const Complex& z) {
  return over_i (atanh (times_i (z)));
}

TMPL inline Complex
cosh (const Complex& z) {
  return Complex (cosh (Re (z)) * cos (Im (z)),
		     sinh (Re (z)) * sin (Im (z)));
}

TMPL inline Complex
sinh (const Complex& z) {
  return Complex (sinh (Re (z)) * cos (Im (z)),
		     cosh (Re (z)) * sin (Im (z)));
}

INV_TRIGO_SUGAR(TMPL,Complex)
INV_HYPER_SUGAR(TMPL,Complex)
ARG_HYPER_SUGAR(TMPL,Complex)

/******************************************************************************
* Miscellaneous
******************************************************************************/

template<typename C> inline C
rnd_error (const complex<C>& z) {
  typedef Round_up(C) Up;
  return Up::add (rnd_error (Re (z)), rnd_error (Im (z)));
}

#undef TMPL_DEF
#undef TMPL
#undef Complex
} // namespace mmx
#endif // __MMX_COMPLEX_HPP
