
/******************************************************************************
* MODULE     : int.hpp
* DESCRIPTION: additional operations for hardware integers
* COPYRIGHT  : (C) 2008  Gregoire Lecerf
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__INT__HPP
#define __MMX__INT__HPP

namespace mmx {

//! Hereafter C must a genuine C/C++ int type, signed or not.

// These two macro prevents the compiler from complaining
// when n = 8 * sizeof(C)
#define MMX_SAFE_LEFT_SHIFT_INT(C,a,n)				\
  (((n) >= 8*sizeof(C)) ? ((C) 0) : (((C) a) << ((n)/2)) << ((n)-(n)/2))

#define MMX_SAFE_RIGHT_SHIFT_INT(C,a,n)				\
  (((n) >= 8*sizeof(C)) ? ((C) 0) : (((C) a) >> ((n)/2)) >> ((n)-(n)/2))

//! Get int type of size at least s, void if impossible
template<nat s>
struct unsigned_int_with_size_at_least_helper {
private:
  template<bool b, typename T, typename E>
  struct if_helper { typedef T ret; };
  template<typename T, typename E>
  struct if_helper<false,T,E> { typedef E ret; };
public:
  typedef
    typename if_helper<sizeof(unsigned char)>=s, unsigned char,
    typename if_helper<sizeof(short unsigned int)>=s, short unsigned int,
    typename if_helper<sizeof(unsigned int)>=s, unsigned int,
    typename if_helper<sizeof(long unsigned int)>=s, long unsigned int,
    typename if_helper<sizeof(long long unsigned int)>=s,
                              long long unsigned int,
      void >::ret >::ret >::ret >::ret >::ret type ; };

//! Get int with double size
template<typename C>
struct unsigned_int_with_double_size_helper {
  typedef typename
  unsigned_int_with_size_at_least_helper<2*sizeof(C)>::type
  type; };

//! Get the corresponding signed integer type
template<typename C>
struct signed_of_helper {
  typedef void type; };

template<>
struct signed_of_helper<double> {
  typedef double type; };

template<>
struct signed_of_helper<char> {
  typedef signed char type; };

template<>
struct signed_of_helper<signed char> {
  typedef signed char type; };

template<>
struct signed_of_helper<short int> {
  typedef short int type; };

template<>
struct signed_of_helper<int> {
  typedef int type; };

template<>
struct signed_of_helper<long int> {
  typedef long int type; };

template<>
struct signed_of_helper<long long int> {
  typedef long long int type; };

template<>
struct signed_of_helper<unsigned char> {
  typedef signed char type; };

template<>
struct signed_of_helper<short unsigned int> {
  typedef short int type; };

template<>
struct signed_of_helper<unsigned int> {
  typedef int type; };

template<>
struct signed_of_helper<long unsigned int> {
  typedef long int type; };

template<>
struct signed_of_helper<long long unsigned int> {
  typedef long long int type; };

//! Get the unsigned corresponding integer type
template<typename C>
struct unsigned_of_helper {
  typedef void type; };

template<>
struct unsigned_of_helper<char> {
  typedef unsigned char type; };

template<>
struct unsigned_of_helper<signed char> {
  typedef unsigned char type; };

template<>
struct unsigned_of_helper<short int> {
  typedef unsigned short int type; };

template<>
struct unsigned_of_helper<int> {
  typedef unsigned int type; };

template<>
struct unsigned_of_helper<long int> {
  typedef long unsigned int type; };

template<>
struct unsigned_of_helper<long long int> {
  typedef long long unsigned int type; };

template<>
struct unsigned_of_helper<unsigned char> {
  typedef unsigned char type; };

template<>
struct unsigned_of_helper<short unsigned int> {
  typedef short unsigned int type; };

template<>
struct unsigned_of_helper<unsigned int> {
  typedef unsigned int type; };

template<>
struct unsigned_of_helper<long unsigned int> {
  typedef long unsigned int type; };

template<>
struct unsigned_of_helper<long long unsigned int> {
  typedef long long unsigned int type; };

//! Tell if the integer type is signed
template<typename C>
struct is_signed_helper {
  static const bool value = false; };

template<>
struct is_signed_helper<signed char> {
  static const bool value = true; };

template<>
struct is_signed_helper<char> {
  static const bool value = true; };

template<>
struct is_signed_helper<short int> {
  static const bool value = true; };

template<>
struct is_signed_helper<int> {
  static const bool value = true; };

template<>
struct is_signed_helper<long int> {
  static const bool value = true; };

template<>
struct is_signed_helper<long long int> {
  static const bool value = true; };

//! Store the bitsize of abs (p) in field value.
template<typename uC, nat s, uC x>
struct int_bitsize_helper_rec {
  static const nat s2  = s >> 1;
  static const uC mask = (uC) MMX_SAFE_LEFT_SHIFT_INT(uC,-1,s2);
  static const uC size = (x & mask) ?
    (s2 + int_bitsize_helper_rec< uC, s2, (x >> s2)>::size) :
    int_bitsize_helper_rec< uC, s2, x >::size;
};
  
template<typename uC, uC x>
struct int_bitsize_helper_rec <uC, 1, x> {
  static const uC size = (x == 0) ? 0 : 1; };

template<typename uC, uC x>
struct int_bitsize_helper_rec <uC, 0, x> {
  static const uC size = 0; };

template<typename C, C p>
struct int_bitsize_helper {
  typedef typename unsigned_of_helper<C>::type uC;
  static const uC up = (p < 0) ? -p : p;

  static const nat value =
    int_bitsize_helper_rec < uC, 8*sizeof(C), up >::size;
};

//! Return the bitsize of abs (p).
template<typename C> static inline nat
int_bitsize (const C& p) {
  typedef typename unsigned_of_helper<C>::type uC;
  if (p == 0) return 0;
  uC up = abs (p);
  nat s = 0;
  nat k = 4 * sizeof(C);
  uC mask = ((uC) -1) << k;
  while (k != 0) {
    if (up & mask) { up >>= k; s += k; }
    k >>= 1;
    mask >>= k;
  }
  return up == 0 ? s : s + 1;
}

template<typename C>
class long_int_mul_op {
  typedef typename unsigned_of_helper<C>::type uC;
  static const nat n2 = 4 * sizeof (uC);
  static const nat n  = 2 * n2;
  static const uC lo_mask = ((uC) -1) >> n2;
  static const uC carry = ((uC) 1) << n2;
  // Let B = (1 << n2),  R = B^2.

  static inline uC
  lo (const uC& s) { return s & lo_mask; }
    
  static inline uC
  hi (const uC& s) { return s >> n2; }

public:
  //! dest1 R + dest0 = a * b.
  static void
  op (uC& dest1, uC& dest0, const uC& a, const uC& b) {
    uC t0, t1;
    uC lo_a = lo (a);
    uC hi_a = hi (a);
    uC lo_b = lo (b);
    uC hi_b = hi (b);
    dest0 = lo_a * lo_b;
    dest1 = hi_a * hi_b;
    t0 = hi_a * lo_b;
    t1 = t0 + lo_a * hi_b;
    if (t1 < t0) dest1 += carry;
    t0 = lo (t1) << n2;
    dest0 += t0;
    dest1 += hi (t1);
    if (dest0 < t0) dest1++;
  }
};

template<typename C>
struct long_int_rshift_op {
  typedef typename unsigned_of_helper<C>::type uC;
  static const nat n2 = 4 * sizeof (uC);
  static const nat n  = 2 * n2;
  // Let B = (1 << n2),  R = B^2.

  //! dest1 R + dest0 >>= s
  static inline void
  op (uC& dest1, uC& dest0, nat s) {
    if (s == 0) return;
    if (s < n) {
      dest0 = (dest0 >> s) | (dest1 << (n - s));
      dest1 = dest1 >> s;
    }
    else {
      dest0 = dest1 >> (s - n);
      dest1 = 0;
    }
  }
    
  //! dest1 R + dest0 >>= s.
  //! Return true iff some nonzero bits have been pushed out.
  static inline bool
  op_b (uC& dest1, uC& dest0, nat s) {
    bool ans;
    if (s == 0) return false;
    if (s < n) {
      ans = (dest0 & (((uC) -1) >> (n - s))) != 0;
      dest0 = (dest0 >> s) | (dest1 << (n - s));
      dest1 = dest1 >> s;
    }
    else {
      ans = dest0 != 0 || (dest1 & (((uC) -1) >> (2*n - s))) != 0;
      dest0 = dest1 >> (s - n);
      dest1 = 0;
    }
    return ans;
  }
};

template<typename C>
struct long_int_lshift_op {
  typedef typename unsigned_of_helper<C>::type uC;
  static const nat n2 = 4 * sizeof (uC);
  static const nat n  = 2 * n2;
  // Let B = (1 << n2),  R = B^2.

  //! dest1 R + dest0 <<= s
  static inline void
  op (uC& dest1, uC& dest0, nat s) {
    if (s == 0) return;
    if (s < n) {
      dest1 = (dest1 << s) | (dest0 >> (n - s));
      dest0 = dest0 << s;
    }
    else {
      dest1 = dest0 << (s - n);
      dest0 = 0;
    }
  }
};

template<typename C>
struct long_int_sub_op {
  typedef typename unsigned_of_helper<C>::type uC;

  //! dest1 R + dest0 -= (s1 R + s0), with R = (1 << n) 
  static inline void
  op (uC& dest1, uC& dest0, const uC& s1, const uC& s0) {
    if (dest0 >= s0) {
      dest0 -= s0;
      dest1 -= s1;
    }
    else {
      dest0 -= s0;
      dest1 -= s1 + 1;
    }
  }
};

template<typename C>
struct long_int_ge_op {
  typedef typename unsigned_of_helper<C>::type uC;

  //! s1 R + s0 >= t1 R + t0.
  static inline bool
  op (const uC& s1, const uC& s0, const uC& t1, const uC& t0) {
    if (s1 > t1) return true;
    if (s1 < t1) return false;
    return s0 >= t0;
  }
};

}// namespace mmx
#endif //__MMX__INT__HPP
