
/******************************************************************************
* MODULE     : interval.hpp
* DESCRIPTION: Intervals
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_INTERVAL_HPP
#define __MMX_INTERVAL_HPP
#include <numerix/rounded.hpp>
namespace mmx {
#define TMPL_DEF template<typename C,typename V=std_interval>
#define TMPL template<typename C,typename V>
#define Interval interval<C,V>
#define Up V::template rnd_helper<C>::UV
#define Down V::template rnd_helper<C>::DV
TMPL class interval;
TMPL inline C lower (const Interval& z);
TMPL inline C upper (const Interval& z);

/******************************************************************************
* Interval variants
******************************************************************************/

struct std_interval {
  template<typename C>
  struct rnd_helper {
    typedef Round_up(C) UV;
    typedef Round_down(C) DV;
  };
};

/******************************************************************************
* The interval type
******************************************************************************/

TMPL_DEF
class interval {
MMX_ALLOCATORS
private:
  C l;
  C r;

public:      
  typedef C boundary_type;

  inline interval (): l (0), r (0) {}
  template<typename T> inline interval (const T& x):
    l (Down::template as<C> (x)),
    r (Up::template as<C> (x)) {}
  template<typename T> inline interval (const interval<T,V>& x):
    l (Down::template as<C> (lower (x))),
    r (Up::template as<C> (upper (x))) {}
  template<typename LT,typename UT> inline
  interval (const LT& l2, const UT& u2):
    l (Down::template as<C> (l2)),
    r (Up::template as<C> (u2)) {}
  inline interval (const C& x): l (x), r (x) {}
  inline interval (const C& l2, const C& r2): l (l2), r (r2) {}
  friend C lower LESSGTR (const Interval& x);
  friend C upper LESSGTR (const Interval& x);

  static inline Interval nan ();
  static inline Interval plus_infinity ();
  static inline Interval minus_infinity ();
  static inline Interval infinity ();
  static inline Interval fuzzy ();
  static inline Interval log2 ();
  static inline Interval pi ();
  static inline Interval euler ();

  inline Interval& operator += (const Interval& x);
  inline Interval& operator -= (const Interval& x);
  inline Interval& operator <<= (int shift);
  inline Interval& operator >>= (int shift);
};

STYPE_TO_TYPE(TMPL,center_type,Interval,C);
STYPE_TO_TYPE(TMPL,radius_type,Interval,C);
TMPL struct rounding_helper<Interval >: public rounding_helper<C> {};

TMPL
struct constant_helper<Interval > {
  static inline Interval nan () {
    return Interval (Nan (C), Nan (C)); }
  static inline Interval plus_infinity () {
    return Interval (Plus_infinity (C), Plus_infinity (C)); }
  static inline Interval minus_infinity () {
    return Interval (Minus_infinity (C), Minus_infinity (C)); }
  static inline Interval infinity () {
    return Interval (Infinity (C), Infinity (C)); }
  static inline Interval fuzz () {
    return Interval (Minus_infinity (C), Plus_infinity (C)); }
  static inline Interval log2 () {
    return Interval (Down::template log2<C>(), Up::template log2<C>()); }
  static inline Interval pi () {
    return Interval (Down::template pi<C>(), Up::template pi<C>()); }
  static inline Interval euler () {
    return Interval (Down::template euler<C>(), Up::template euler<C>()); }
  static inline Interval catalan () {
    return Interval (Down::template catalan<C>(), Up::template catalan<C>()); }
};

TMPL
struct make_interval_helper<Interval,C> {
  static inline Interval val (const C& l, const C& r) {
    return Interval (l, r); }
};

/******************************************************************************
* Basic routines
******************************************************************************/

TMPL inline C lower (const Interval& x) { return x.l; }
TMPL inline C upper (const Interval& x) { return x.r; }
TMPL inline C center (const Interval& x) {
  return (lower (x) + upper (x)) >> 1; }
TMPL inline C radius (const Interval& x) {
  if (is_infinity (x)) return 0;
  return Up::add (Up::sub (upper (x), lower (x)) >> 1, rnd_error (center (x))); }

TMPL inline Interval sharpen (const Interval& x) {
  C c= center (x); return Interval (c, c); }

TMPL inline bool intersect (Interval& r, const Interval &a, const Interval &b) {
    if ((upper(a) < lower(b)) || (lower(a) > upper(b))) return false;
    r = Interval(std::max(lower(a), lower(b)), std::min(upper(a), upper(b)));
    return true;
}

TMPL inline bool is_exact_zero (const Interval& x) {
  return is_exact_zero (lower (x)) && is_exact_zero (upper (x)); }
TMPL inline bool is_number (const Interval& x) {
  return is_number (lower (x)) && is_number (upper (x)); }
TMPL inline bool is_infinity (const Interval& x) {
  return is_infinity (lower (x)) && upper (x) == lower (x); }
TMPL inline bool is_fuzz (const Interval& x) {
  return is_infinity (lower (x)) && is_infinity (upper (x)) &&
    lower (x) != upper (x); }
TMPL inline bool is_nan (const Interval& x) {
  return is_nan (lower (x)) || is_nan (upper (x)); }

TMPL inline Interval copy (const Interval& x, nat prec) {
  return Interval (copy (lower (x), prec), copy (upper (x), prec)); }
TMPL inline nat precision (const Interval& x) {
  return max (precision (lower (x)), precision (upper (x))); }
TMPL inline mp_exp_t exponent (const Interval& x) {
  return max (exponent (lower (x)), exponent (upper (x))); }
TMPL inline mp_exp_t step (const Interval& x) {
  return step (upper (x) - lower (x)); }

template<typename Op, typename C, typename V> nat
unary_hash (const Interval& x) {
  nat h= Op::op (lower (x));
  return (h<<3) ^ (h<<15) ^ (h>>29) ^ Op::op (upper (x));
}

template<typename Op, typename C, typename V> bool
binary_test (const Interval& x1, const Interval& x2) {
  return Op::op (lower (x1), lower (x2)) && Op::op (upper (x1), upper (x2));
}

SOFT_IDENTITY_OP_SUGAR(TMPL,Interval)
HARD_IDENTITY_OP_SUGAR(TMPL,Interval)

TMPL syntactic
flatten (const Interval& x) {
  if (is_nan (x)) return Nan (syntactic);
  if (lower (x) == 0 && upper (x) == 0) return 0;
  return flatten_range (lower (x), upper (x));
  // return flatten ("interval", lower (x), upper (x));
}

TMPL syntactic
flatten (const Interval& x, nat dd) {
  if (is_nan (x)) return Nan (syntactic);
  if (lower (x) == 0 && upper (x) == 0) return 0;
  return flatten_range (lower (x), upper (x), dd);
}

PRINT_SUGAR(TMPL,Interval)

/******************************************************************************
* Comparisons
******************************************************************************/

TMPL inline bool
is_zero (const Interval& x) {
  return is_number (x) && sign (lower (x)) * sign (upper (x)) <= 0;
}

TMPL inline bool
is_non_zero (const Interval& x) {
  return !is_number (x) || sign (lower (x)) * sign (upper (x)) > 0;
}

TMPL inline bool
operator == (const Interval& x1, const Interval& x2) {
  if (is_nan (x1) || is_nan (x2)) return is_nan (x1) && is_nan (x2);
  return lower (x1) <= upper (x2) && lower (x2) <= upper (x1);
}

TMPL inline bool
operator != (const Interval& x1, const Interval& x2) {
  if (is_nan (x1) || is_nan (x2)) return !(is_nan (x1) && is_nan (x2));
  return lower (x1) > upper (x2) || lower (x2) > upper (x1);
}

TMPL inline bool
operator <= (const Interval& x1, const Interval& x2) {
  if (is_nan (x1) || is_nan (x2)) return false;
  return lower (x1) <= upper (x2);
}

TMPL inline bool
operator < (const Interval& x1, const Interval& x2) {
  if (is_nan (x1) || is_nan (x2)) return false;
  return lower (x2) > upper (x1);
}

TMPL inline bool
operator >= (const Interval& x1, const Interval& x2) {
  if (is_nan (x1) || is_nan (x2)) return false;
  return lower (x2) <= upper (x1);
}

TMPL inline bool
operator > (const Interval& x1, const Interval& x2) {
  if (is_nan (x1) || is_nan (x2)) return false;
  return lower (x1) > upper (x2);
}

EQUAL_INT_SUGAR(TMPL,Interval,Interval);
COMPARE_INT_SUGAR(TMPL,Interval,Interval);

/******************************************************************************
* Other order related functions
******************************************************************************/

TMPL inline Interval
abs (const Interval& x) {
  if (is_nan (x)) return x;
  C r= max (abs (lower (x)), abs (upper (x)));
  if (sign (lower (x)) * sign (upper (x)) <= 0) return Interval (C(0), r);
  return Interval (min (abs (lower (x)), abs (upper (x))), r);
}

TMPL Interval
min (const Interval& x1, const Interval& x2) {
  return Interval (min (lower (x1), lower (x2)), min (upper (x1), upper (x2)));
}

TMPL Interval
max (const Interval& x1, const Interval& x2) {
  return Interval (max (lower (x1), lower (x2)), max (upper (x1), upper (x2)));
}

TMPL inline C abs_down (const Interval& x) { return lower (abs (x)); }
TMPL inline C abs_up (const Interval& x) { return upper (abs (x)); }
TMPL inline C bnd_down (const Interval& x) { return lower (x); }
TMPL inline C bnd_up (const Interval& x) { return upper (x); }

/******************************************************************************
* Basic arithmetic
******************************************************************************/

TMPL inline Interval
operator - (const Interval& x) {
  return Interval (-upper (x), -lower (x));
}

TMPL inline Interval
operator + (const Interval& x1, const Interval& x2) {
  return Interval (Down::add (lower (x1), lower (x2)),
		   Up::add (upper (x1), upper (x2)));
}

TMPL inline Interval
operator - (const Interval& x1, const Interval& x2) {
  return Interval (Down::sub (lower (x1), upper (x2)),
		   Up::sub (upper (x1), lower (x2)));
}

TMPL Interval
operator * (const Interval& x1, const Interval& x2) {
  if (lower (x1) >= 0)
    return Interval (lower (x2) >= 0?
		     Down::mul (lower (x1), lower (x2)):
		     Down::mul (upper (x1), lower (x2)),
		     upper (x2) >= 0?
		     Up::mul (upper (x1), upper (x2)):
		     Up::mul (lower (x1), upper (x2)));
  else if (upper (x1) <= 0)
    return Interval (upper (x2) >= 0?
		     Down::mul (lower (x1), upper (x2)):
		     Down::mul (upper (x1), upper (x2)),
		     lower (x2) >= 0?
		     Up::mul (upper (x1), lower (x2)):
		     Up::mul (lower (x1), lower (x2)));
  else if (lower (x2) >= 0)
    return Interval (Down::mul (lower (x1), upper (x2)),
		     Up::mul (upper (x1), upper (x2)));
  else if (upper (x2) <= 0)
    return Interval (Down::mul (upper (x1), lower (x2)),
		     Up::mul (lower (x1), lower (x2)));
  else
    return Interval (min (Down::mul (lower (x1), upper (x2)),
			  Down::mul (upper (x1), lower (x2))),
		     max (Up::mul (lower (x1), lower (x2)),
			  Up::mul (upper (x1), upper (x2))));
}

TMPL inline Interval
square (const Interval& x) {
  if (lower (x) >= 0)
    return Interval (Down::mul (lower (x), lower (x)),
		     Up::mul (upper (x), upper (x)));
  else if (upper (x) <= 0)
    return Interval (Down::mul (upper (x), upper (x)),
		     Up::mul (lower (x), lower (x)));
  else return Interval (Down::mul (lower (x), upper (x)),
			max (Up::mul (lower (x), lower (x)),
			     Up::mul (upper (x), upper (x))));
}

TMPL Interval
operator / (const Interval& x1, const Interval& x2) {
  if (is_zero (x2)) return Nan (Interval);
  else if (lower (x2) > 0)
    return Interval (lower (x1) > 0?
		     Down::div (lower (x1), upper (x2)):
		     Down::div (lower (x1), lower (x2)),
		     upper (x1) > 0?
		     Up::div (upper (x1), lower (x2)):
		     Up::div (upper (x1), upper (x2)));
  else if (upper (x2) < 0)
    return Interval (upper (x1) > 0?
		     Down::div (upper (x1), upper (x2)):
		     Down::div (upper (x1), lower (x2)),
		     lower (x1) > 0?
		     Up::div (lower (x1), lower (x2)):
		     Up::div (lower (x1), upper (x2)));
  else return Nan (Interval);
}

TMPL inline Interval
operator << (const Interval& x, int shift) {
  return Interval (lower (x) << shift, upper (x) << shift);
}

TMPL inline Interval
operator >> (const Interval& x, int shift) {
  return Interval (lower (x) >> shift, upper (x) >> shift);
}

ARITH_INT_SUGAR(TMPL,Interval,Interval);

/******************************************************************************
* In place arithmetic operations
******************************************************************************/

TMPL inline Interval&
Interval::operator += (const Interval& x) {
  l= Down::add (l, lower (x));
  r= Up::add (r, upper (x));
  return *this;
}

TMPL inline Interval&
Interval::operator -= (const Interval& x) {
  l= Down::sub (l, upper (x));
  r= Up::sub (r, lower (x));
  return *this;
}

TMPL inline Interval&
Interval::operator <<= (int shift) {
  l <<= shift;
  r <<= shift;
  return *this;
}

TMPL inline Interval&
Interval::operator >>= (int shift) {
  l >>= shift;
  r >>= shift;
  return *this;
}

/******************************************************************************
* Elementary functions
******************************************************************************/

TMPL Interval
sqrt (const Interval& x) {
  return Interval (Down::sqrt (lower (x)), Up::sqrt (upper (x)));
}

TMPL Interval
hypot (const Interval& x, const Interval& y) {
  Interval z= square (x) + square (y);
  if (z > 0) return sqrt (z);
  else return Interval (0, Up::sqrt (upper (z)));
}

TMPL Interval
exp (const Interval& x) {
  return Interval (Down::exp (lower (x)), Up::exp (upper (x)));
}

TMPL Interval
log (const Interval& x) {
  return Interval (Down::log (lower (x)), Up::log (upper (x)));
}

TMPL inline Interval
pow (const Interval& x, const Interval& y) {
  return exp (y * log (x));
}

TMPL Interval
cos (const Interval& x) {
  if (Up::sub (upper (x), lower (x)) >= C (3.14159)) return Interval (-1, 1);
  else {
    int s1= -sign (sin (lower (x)));
    int s2= -sign (sin (upper (x)));
    if (s1 >= 0 && s2 >= 0)
      return Interval (Down::cos (lower (x)), Up::cos (upper (x)));
    else if (s1 <= 0 && s2 <= 0)
      return Interval (Down::cos (upper (x)), Down::cos (lower (x)));
    else if (s1 >= 0)
      return Interval (min (Down::cos (lower (x)),
			    Down::cos (upper (x))), C (1));
    else
      return Interval (C (-1), max (Down::cos (lower (x)),
				    Down::cos (upper (x))));
  }
}

TMPL Interval
sin (const Interval& x) {
  if (Up::sub (upper (x), lower (x)) >= C (3.14159)) return Interval (-1, 1);
  else {
    int s1= sign (cos (lower (x)));
    int s2= sign (cos (upper (x)));
    if (s1 >= 0 && s2 >= 0)
      return Interval (Down::sin (lower (x)), Up::sin (upper (x)));
    else if (s1 <= 0 && s2 <= 0)
      return Interval (Down::sin (upper (x)), Down::sin (lower (x)));
    else if (s1 >= 0)
      return Interval (min (Down::sin (lower (x)),
			    Down::sin (upper (x))), C (1));
    else
      return Interval (C(-1), max (Down::sin (lower (x)),
				   Down::sin (upper (x))));
  }
}

TMPL Interval
tan (const Interval& x) {
  return sin (x) / cos (x);
}

TMPL inline Interval
cosh (const Interval& x) {
  if (is_nan (x)) return Nan (Interval);
  else if (lower (x) >= 0)
    return Interval (Down::cosh (lower (x)), Up::cosh (upper (x)));
  else if (upper (x) <= 0)
    return Interval (Down::cosh (upper (x)), Up::cosh (lower (x)));
  else return Interval (1, max (Up::cosh (lower (x)), Up::cosh (upper (x))));
}

TMPL inline Interval
sinh (const Interval& x) {
  return Interval (Down::sinh (lower (x)), Up::sinh (upper (x)));
}

TMPL Interval
tanh (const Interval& x) {
  return sinh (x) / cosh (x);
}

TMPL Interval
acos (const Interval& x) {
  return Interval (Down::acos (upper (x)), Up::acos (lower (x)));
}

TMPL Interval
asin (const Interval& x) {
  return Interval (Down::asin (lower (x)), Up::asin (upper (x)));
}

TMPL Interval
atan (const Interval& x) {
  return Interval (Down::atan (lower (x)), Up::atan (upper (x)));
}

TMPL Interval
atan2 (const Interval& y, const Interval& x) {
  if (sign (lower (x)) > 0) return atan (y/x);
  if (sign (lower (y)) > 0) return atan (-x/y) + (Pi (Interval) >> 1);
  if (sign (upper (y)) < 0) return atan (-x/y) - (Pi (Interval) >> 1);
  if (sign (upper (x)) < 0) return atan (y/x) + Pi (Interval);
  C pi= Up::template pi<C> ();
  return Interval (-pi, pi);
}

INV_TRIGO_SUGAR(TMPL,Interval)
INV_HYPER_SUGAR(TMPL,Interval)
ARG_HYPER_SUGAR(TMPL,Interval)

#undef TMPL_DEF
#undef TMPL
#undef Interval
#undef Up
#undef Down
} // namespace mmx
#endif // __MMX_INTERVAL_HPP
