
/******************************************************************************
* MODULE     : modular.hpp
* DESCRIPTION: modular arithmetic
* COPYRIGHT  : (C) 2008  Gregoire Lecerf
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__MODULAR__HPP
#define __MMX__MODULAR__HPP
#include <basix/basix.hpp>
#include <basix/wrap.hpp>
#include <numerix/modulus.hpp>

namespace mmx {
#define Modular_variant(M) modular_variant_helper<M>::MV
#define TMPL_DEF \
  template<typename M, typename V= typename Modular_variant(M)>
#define TMPL template<typename M, typename V>
#define Modulus modulus<C,V>
#define Modular modular<M,V>

template<typename M>
struct modular_variant_helper;
TMPL class modular;

/******************************************************************************
* Variant
******************************************************************************/

// Local storages of the modulus
struct modular_local {
  template<typename M>
  class modulus_storage {
    static inline M& dyn_modulus () {
      static M modulus = M ();
      return modulus; }
  public:
    static inline void set_modulus (const M& p) { dyn_modulus () = p; }
    static inline M get_modulus () { return dyn_modulus (); }
  };
};

// Local storages of the modulus
template<typename C, C p>
struct modular_fixed {
  template<typename M>
  class modulus_storage {
    static inline M& dyn_modulus () {
      static fixed_value <C, p> _p;
      static M modulus = M (_p);
      return modulus; }
  public:
    static inline void set_modulus (const M& p) {
      ERROR ("modular_fixed: can not change the modulus");
    }
    static inline M get_modulus () { return dyn_modulus (); }
  };
};

/******************************************************************************
* Variant helper, default storage of the modulus is "local"
******************************************************************************/

template<typename M>
struct modular_variant_helper {
  typedef modular_local MV; };

/******************************************************************************
* General modulars
******************************************************************************/

TMPL_DEF
class modular {
MMX_ALLOCATORS
private:
  typedef typename M::base C;
  typedef typename V::template modulus_storage<M> S;
public:
  C rep;
  static inline M get_modulus () { return S::get_modulus (); }
  static inline void set_modulus (const M& p) { S::set_modulus (p); }

  inline C operator * () const {
    C dest;
    decode_mod (dest, rep, get_modulus ());
    return dest; }

  inline modular () { encode_mod (rep, C (), get_modulus ()); }

  template<typename O>
  inline modular (const O& s) {
    encode_mod (rep, ((C) s), get_modulus ()); }

  inline modular (const Modular& s) : rep (s.rep) {}

  inline Modular& operator = (const Modular& a) {
    rep = a.rep;
    return *this; }

  inline bool operator == (const Modular& a) const {
    return rep == a.rep; }

  inline bool operator != (const Modular& a) const {
    return rep != a.rep; }
};

template<typename TM, typename TV, typename SM, typename SV>
struct as_helper<modular<TM,TV>,modular<SM,SV> > {
  static inline modular<TM,TV>
  cv (const modular<SM,SV>& p) { return modular<TM,TV> (*p); }
};

template<typename C, typename V, typename W>
struct as_helper<C,modular<modulus<C,V>,W> > {
  static inline C
  cv (const modular<modulus<C,V>,W>& p) { return *p; }
};

WRAP_WRAPPED_IMPL(TMPL inline,Modular)

/******************************************************************************
* Arithmetic
******************************************************************************/

TMPL inline Modular&
operator += (Modular& dest, const Modular& s) {
  add_mod (dest.rep, s.rep, Modular::get_modulus ());
  return dest; }

TMPL inline Modular&
operator -= (Modular& dest, const Modular& s) {
  sub_mod (dest.rep, s.rep, Modular::get_modulus ());
  return dest; }

TMPL inline Modular&
operator *= (Modular& dest, const Modular& s) {
  mul_mod (dest.rep, s.rep, Modular::get_modulus ());
  return dest; }

TMPL inline Modular&
operator /= (Modular& dest, const Modular& s) {
  div_mod (dest.rep, s.rep, Modular::get_modulus ());
  return dest; }

TMPL inline Modular
operator + (const Modular& a, const Modular& b) {
  Modular c;
  add_mod (c.rep, a.rep, b.rep, Modular::get_modulus ());
  return c; }

TMPL inline Modular
operator - (const Modular& a) {
  Modular c;
  neg_mod (c.rep, a.rep, Modular::get_modulus ());
  return c; }

TMPL inline Modular
operator - (const Modular& a, const Modular& b) {
  Modular c;
  sub_mod (c.rep, a.rep, b.rep, Modular::get_modulus ());
  return c; }

TMPL inline Modular
operator * (const Modular& a, const Modular& b) {
  Modular c;
  mul_mod (c.rep, a.rep, b.rep, Modular::get_modulus ());
  return c; }

TMPL inline Modular
operator / (const Modular& a, const Modular& b) {
  Modular c;
  div_mod (c.rep, a.rep, b.rep, Modular::get_modulus ());
  return c; }

TMPL inline Modular
invert (const Modular& a) {
  Modular c;
  inv_mod (c.rep, a.rep, Modular::get_modulus ());
  return c; }

/******************************************************************************
* Output routine
******************************************************************************/

WRAP_PRINT_IMPL(TMPL inline,Modular)

template<typename V, typename W> inline ostream& 
operator << (ostream& out, const modular<modulus<char,W>, V>& c) {
  return out << ((int) *c); }

template<typename V, typename W> inline ostream& 
operator << (ostream& out, const modular<modulus<signed char,W>, V>& c) {
  return out << ((int) *c); }

template<typename V, typename W> inline ostream& 
operator << (ostream& out, const modular<modulus<unsigned char,W>, V>& c) {
  return out << ((int) *c); }

template<typename V, typename W> inline syntactic
flatten (const modular<modulus<unsigned char,W>, V>& c) {
  return as_syntactic (*c); }

/******************************************************************************
* Syntactic sugar
******************************************************************************/

ARITH_INT_SUGAR(TMPL,Modular,Modular)
POOR_MAN_SQRT_SUGAR(TMPL,Modular)
POOR_MAN_ELEMENTARY_SUGAR(TMPL,Modular)

#undef TMPL_DEF
#undef TMPL
#undef Modulus
#undef Modular
} // namespace mmx
#endif //__MMX__MODULAR__HPP
