
/******************************************************************************
* MODULE     : modulus_naive_int.hpp
* DESCRIPTION: modulus for modular arithmetic over hardware integers
* COPYRIGHT  : (C) 2008  Gregoire Lecerf
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__MODULUS_NAIVE_INT__HPP
#define __MMX__MODULUS_NAIVE_INT__HPP
#include <numerix/int.hpp>
#include <numerix/modulus_naive.hpp>

namespace mmx {

//! Hereafter C must a genuine C/C++ int type, signed or not.
//! If C has bit-size k and if m <= k is the maximum number of bits
//! allowed for the modulus then the modulus 0 actually stands for 2^m.

#define TMPL template<typename C, typename M>

/******************************************************************************
* Variant
******************************************************************************/
// The modulus 'p' is always taken to be positive
// The modulars are in the range [0, p-1]

template<nat size>
struct modulus_maximum_size_int {

  template<typename C, bool is_signed>
  struct _helper {
    static const nat value = (size < 8*sizeof(C) ? size : 8*sizeof(C)-1);
  };

  template<typename C>
  struct _helper<C, false> {
    static const nat value = (size <= 8*sizeof(C) ? size : 8*sizeof(C));
  };

  template<typename C>
  struct maximum_size_helper {
    static const nat value = _helper<C, is_signed_helper<C>::value>::value;
    static inline C dyn_value () { return value; }
  };

  template<typename C>
  struct maximum_value_helper {
    static const nat _size = maximum_size_helper<C>::value;
    typedef typename unsigned_of_helper<C>::type uC;
    static const C value = MMX_SAFE_LEFT_SHIFT_INT (uC, 1, _size) - 1;
    static inline C dyn_value () { return value; }
  };
};

template<typename V>
struct modulus_normalization_naive_int : public V {

  template<typename C> static inline bool
  normalize (C& p) {
    typedef typename unsigned_of_helper<C>::type uC;
    static const uC a = MMX_SAFE_LEFT_SHIFT_INT(uC, 1,
                         V::template maximum_size_helper<C>::value);
    p = abs (p);
    if ((uC) p == a || p == 0) { p = a; return true; }
    return p <= V::template maximum_value_helper<C>::dyn_value ();
  }
};

template<typename V>
struct modulus_reduction_naive_int : public V {
  template<typename C> static inline void 
  reduce_mod_core (C& dest, const C& p) {
    if (p != 0) dest %= p; }

  TMPL static inline void 
  reduce_mod (C& dest, const M& m) {
    reduce_mod_core (dest, (C) m.p); }

  TMPL static inline void 
  reduce_mod (C& dest, const C& s, const M& m) {
    dest = s;
    reduce_mod_core (dest, (C) m.p); }
};

template<typename V>
struct modulus_add_naive_int : public V {

  TMPL static inline void
  neg_mod (C& dest, const M& m) {
    if (dest != 0) dest = m.p - dest; }

  TMPL static inline void
  neg_mod (C& dest, const C& s, const M& m) {
    if (s != 0) dest = m.p - s; else dest = s; }

  template<typename C> static inline void
  add_mod_without_overflow (C& dest, const C& s, const C& p) {
    dest += s;
    if (dest >= p) dest -= p; }

  template<typename C> static inline void
  add_mod_with_overflow (C& dest, const C& s, const C& p) {
    dest += s;
    if (dest < s) { dest -= p; return; }
    if (dest >= p) dest -= p; }

  // Signed integer involve no overflow
  template<typename C, bool b> 
  struct add_mod_helper {
    static inline void
    op (C& dest, const C& s, const C& p) {
      typedef typename unsigned_of_helper<C>::type uC;
      uC t = dest;
      uC up = p;
      uC us = s;
      add_mod_without_overflow (t, us, up);
      dest = t; }
  };

  // Unsigned integer can cause overflow
  template<typename C> 
  struct add_mod_helper<C,false> {
    static inline void
    op (C& dest, const C& s, const C& p) {
      if (V::template maximum_size_helper<C>::value >= 8*sizeof(C))
	add_mod_with_overflow (dest, s, p);
      else
	add_mod_without_overflow (dest, s, p); }
  };

  template<typename C> static inline void
  add_mod_core (C& dest, const C& s, const C& p) {
    add_mod_helper<C,is_signed_helper<C>::value>::op (dest, s, p); }

  TMPL static inline void
  add_mod (C& dest, const C& s, const M& m) {
    add_mod_core (dest, s, m.p); }
  
  TMPL static inline void
  add_mod (C& dest, const C& s1, const C& s2, const M& m) {
    dest = s1;
    add_mod (dest, s2, m); }
  
  template<typename C> static inline void
  sub_mod_core (C& dest, const C& s, const C& p) {
    if (dest < s) dest += p;
    dest -= s; }

  TMPL static inline void
  sub_mod (C& dest, const C& s, const M& m) {
    sub_mod_core (dest, s, m.p); }

  TMPL static inline void
  sub_mod (C& dest, const C& s1, const C& s2, const M& m) {
    dest = s1;
    sub_mod (dest, s2, m); }
};
 
template<typename V>
struct modulus_mul_naive_int : public V {

  template <typename C, typename D>
  struct mul_mod_helper {
    static inline void op (C& dest, const C& s, const C& p) {
      D a (dest);	
      a *= s;
      if (p != 0) a %= p;
      dest = (C) a;
    }
  };

  template <typename C>
  struct mul_mod_helper<C, void> {
    static const nat half_size = 4 * sizeof (C);
    static const nat size = 2 * half_size;
    static const C lo_mask = ((C) -1) >> half_size;
    static const C hi_mask = ((C) -1) << half_size;
    // Let B = (1 << half_size)
    static inline C
    lo (const C& s) { return s & lo_mask; }
    
    static inline C
    hi (const C& s) { return (s & hi_mask) >> half_size; }

    static inline void
    half_lshift_mod(C& dest, const C& p, const C& lo_p,const C& hi_p) {
      // p = hi_p * B + lo_p, return (dest * B) % p
      // It is required that hi_p != 0
      C q = dest / hi_p;
      C r = dest - q * hi_p;
      q *= lo_p;
      q %= p;
      dest = r << half_size;
      V::sub_mod_core(dest, q, p);
    }

    static inline void op (C& dest, const C& s, const C& p) {
      C hi_p = hi (p);
      if (hi_p) {
	C lo_p = lo (p);
	C lo_dest = lo (dest);
	C lo_s = lo (s);
	C hi_dest = hi (dest);
	C hi_s = hi (s);
	dest = lo_dest * lo_s;
	reduce_mod_core (dest, p);
	C t = lo_dest * hi_s;
	half_lshift_mod (t, p, lo_p, hi_p);
	V::add_mod_core (dest, t, p);
	t = hi_dest * lo_s;
	half_lshift_mod (t, p, lo_p, hi_p);
	V::add_mod_core(dest, t, p);
	t = hi_dest * hi_s;
	half_lshift_mod (t, p, lo_p, hi_p);
	half_lshift_mod (t, p, lo_p, hi_p);
	V::add_mod_core(dest, t, p);
      }
      else {
	dest *= s;
	if (p != 0) dest %= p;
      }
    }
  };

  TMPL static inline void
  mul_mod (C& dest, const C& s, const M& m) {
    if (is_signed_helper<C>::value) {
      typedef typename unsigned_of_helper<C>::type uC;
      uC t = dest;
      mul_mod (t, (uC) s, m);
      dest = t;
    }
    else {
      typedef typename unsigned_int_with_double_size_helper<C>::type D;
      mul_mod_helper<C,D>::op (dest, s, m.p);
    }
  }

  TMPL static inline void
  mul_mod (C& dest, const C& s1, const C& s2, const M& m) {
    dest = s1;
    mul_mod (dest, s2, m); }

  TMPL static inline void
  mul_mod (C& dest, const C& s, const M& m, const C& x) {
    mul_mod (dest, s, m); }
};

template<typename V>
struct modulus_inv_naive_int : public V {

  template<typename C> static void
  inv_mod_signed (C& a, const C& p) {
    typedef typename unsigned_of_helper<C>::type uC;
    typedef typename signed_of_helper<C>::type sC;
    sC coa0=0, coa1=1, r0=p, r1=a, q, t;	
    if ((p == 0) && (r1 != 0)) {
      q = (((uC) r0) - ((uC) r1)) / ((uC) r1) + 1;
      t = r0 - q * r1;
      r0 = r1;
      r1 = t;
      t = coa1;
      coa1 = coa0 - q * coa1;
      coa0 = t;
    }
    while (r1 != 0) {
      q = r0 / r1;
      t = r0 - q * r1;
      r0 = r1;
      r1 = t;
      t = coa1;
      coa1 = coa0 - q * coa1;
      coa0 = t;
    }
    if (r0 != 1)
      ERROR ("inv_mod: argument is not invertible");
    a = coa0 < 0 ? p+coa0 : coa0;
  }

  template<typename C> static void
  inv_mod_unsigned (C& a, const C& p) {
    C coa0=0, coa1=1, r0=p, r1=a, q, t;	
    bool toggle = true;
    if ((p == 0) && (r1 != 0)) {
      q = (r0-r1) / r1 + 1;
      t = r0 - q * r1;
      r0 = r1;
      r1 = t;
      t = coa1;
      coa1 = coa0 - q * coa1;
      coa0 = t;
      toggle = !toggle;
    }
    while (r1 != 0) {
      q = r0 / r1;
      t = r0 - q * r1;
      r0 = r1;
      r1 = t;
      t = coa1;
      coa1 = coa0 - q * coa1;
      coa0 = t;
      toggle = !toggle;
    }
    if (r0 != 1)
      ERROR ("inv_mod: argument is not invertible");
    a = (toggle && (coa0 != 0)) ? p+coa0 : coa0;
  }

  TMPL static inline void
  inv_mod (C& a, const M& m) {
    if (V::template maximum_size_helper<C>::value >= 8*sizeof(C))
      inv_mod_unsigned(a, m.p);
    else
      inv_mod_signed(a, m.p); }

  TMPL static inline void
  inv_mod (C& dest, const C& s, const M& m) {
    dest = s;
    inv_mod (dest, m); }
};

template<typename V>
struct modulus_div_naive_int : public V {

  TMPL static inline void 
  div_mod (C& dest, const C& s, const M& m) {
    C t = s;
    V::inv_mod (t, m);
    V::mul_mod (dest, t, m); }

  TMPL static inline void 
  div_mod (C& dest, const C& s1, const C& s2, const M& m) {
    dest = s1;
    div_mod (dest, s2, m); }
};

template<typename V>
struct modulus_encoding_naive_int : public V {
  TMPL static inline void 
  encode_mod (C& dest, const C& s, const M& m) {
    typedef typename unsigned_of_helper<C>::type uC;
    if (sign (s) < 0) {
      uC tmp = - ((uC) s);
      V::reduce_mod (tmp, m);
      dest = (C) (((uC) m.p) - tmp);
    }
    else
      dest = s;
      V::reduce_mod (dest, m);
  }
  TMPL static inline void 
  decode_mod (C& dest, const C& s, const M& m) {
    (void) m;
    dest = s; }
};

template<int size>
struct modulus_naive_int:
  public modulus_encoding_naive_int<
         modulus_div_naive_int<
         modulus_inv_naive_int<
	 modulus_mul_naive_int<
	 modulus_add_naive_int<
         modulus_reduction_naive_int<
         modulus_normalization_naive_int<
	 modulus_maximum_size_int<size> > > > > > > > {};

/******************************************************************************
* Helpers specializations for hardware integers
******************************************************************************/

template<>
struct modulus_variant_helper<unsigned char> {
  typedef modulus_naive_int<8*sizeof(unsigned char)> MV; };

template<>
struct modulus_variant_helper<short unsigned int> {
  typedef modulus_naive_int<8*sizeof(short unsigned int)> MV; };

template<>
struct modulus_variant_helper<unsigned int> {
  typedef modulus_naive_int<8*sizeof(unsigned int)> MV; };

template<>
struct modulus_variant_helper<long unsigned int> {
  typedef modulus_naive_int<8*sizeof(long unsigned int)> MV; };

template<>
struct modulus_variant_helper<long long unsigned int> {
  typedef modulus_naive_int<8*sizeof(long long unsigned int)> MV; };

template<>
struct modulus_variant_helper<char> {
  typedef modulus_naive_int<8*sizeof(char)-1> MV; };

template<>
struct modulus_variant_helper<signed char> {
  typedef modulus_naive_int<8*sizeof(char)-1> MV; };

template<>
struct modulus_variant_helper<short int> {
  typedef modulus_naive_int<8*sizeof(short int)-1> MV; };

template<>
struct modulus_variant_helper<int> {
  typedef modulus_naive_int<8*sizeof(int)-1> MV; };

template<>
struct modulus_variant_helper<long int> {
  typedef modulus_naive_int<8*sizeof(long int)-1> MV; };

template<>
struct modulus_variant_helper<long long int> {
  typedef modulus_naive_int<8*sizeof(long long int)-1> MV; };

#undef TMPL
} // namespace mmx
#endif //__MMX__MODULUS_NAIVE_INT__HPP
