
/******************************************************************************
* MODULE     : modulus_preinverse_int.hpp
* DESCRIPTION: modulus with precomputed inverse for hardware integers
* COPYRIGHT  : (C) 2008  Gregoire Lecerf
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

// Implementation details are to be found in doc/texmacs/modular.tm

#ifndef __MMX__MODULUS_PREINVERSE_INT__HPP
#define __MMX__MODULUS_PREINVERSE_INT__HPP

#include <numerix/int.hpp>
#include <numerix/modulus_naive_int.hpp>

namespace mmx {

//! Hereafter C is a genuine C++ hardware int type, signed or not.
//! If C has bit-size n then the modulus 0 behaves as 2^n.

#define TMPL template<typename C>

/******************************************************************************
* Variant
******************************************************************************/

template<typename V>
struct modulus_mul_preinverse_int : public V {

/*
  Dynamic computation of the auxiliaries
*/
  template <typename C, typename D> 
  struct _dynamic_inverse_helper {
    typedef typename unsigned_of_helper<C>::type uC;
    typedef typename unsigned_of_helper<D>::type uD;
    static const nat m = V::template maximum_size_helper<C>::value;
    // p must be normalized
    static inline uC
    op (const uC& p, nat r, nat s, nat t) {
      (void) r;
      if (p == 1 && s+t == (nat) -1) return 0; // only case when s+t < 0
      uD loc_p = (p == 0) ? ((uD) 1) << m : (uD) ((uC) p);
      return (((uD) 1) << (s+t)) / loc_p; }
  };
  
  template <typename C>
  struct _dynamic_inverse_helper<C,void> {
    static const nat n2 = 4 * sizeof(C);
    static const nat n  = 8 * sizeof(C);
    static const nat m  = V::template maximum_size_helper<C>::value;
    typedef typename unsigned_of_helper<C>::type uC;
    typedef long_int_mul_op<C>    long_mul;
    typedef long_int_rshift_op<C> long_rshift;
    typedef long_int_sub_op<C>    long_sub;
    
    static uC
    _q0 (const uC& p, nat r, nat s, nat t, nat s0, nat s1) {
      // p != 1 or 2
      uC q0;
      if (r <= s0)
	q0 = (((uC) 1) << (s0+r-1)) / p;
      else {
	uC p1 = (p == 0) ? ((uC) 1) << s0 : p >> (r-s0);
	uC p0 = p & (((uC) -1) >> (n+s0-r));
	uC d = ((uC) 1) << s1;
	q0 = d / p1;
	uC b = (d % p1) << (r-s0);
	uC c = q0 * p0;
	if (c >= p) { q0--; c -= p; }
	if (b < c) q0--;
      }
      return q0;
    }
    
    static uC
    op (const uC& p, nat r, nat s, nat t) {
      // If p is 1 or 2 then s0 == 0.
      // p must be normalized.
      if (p == 1) {
	if (m+2 <= n) return 2;
	if (m+1 <= n) return 1;
	return 0;
      }
      if (p == 2) {
	if (m+2 <= n) return 4;
	if (m+1 <= n) return 2;
	return 1;
      }
      nat s0 = (s+t-(r-1)) >> 1;
      nat s1 = 2*s0 - 1;
      uC q0 = _q0 (p, r, s, t, s0, s1);
      uC a0 = 0, a1 = 0;
      long_mul::op (a1, a0, q0 * q0, p);
      if (long_rshift::op_b (a1, a0, r)) a0++;
      uC c = (q0 << s0) - a0;
      uC d;
      if (m+1 <= n || s1+r <= n) {
	if (s1+r-1 >= n) d = 0; else d = (((uC) 1) << (s1+r-1));
	d -= c*p;
	if (d >= p) { c++; d-=p; }
      }
      else {
	uC d1, e0, e1;
	long_mul::op (e1, e0, c, p);
	d = 0;
	d1 = ((uC) 1) << (s1+r-1-n);
	long_sub::op (d1, d, e1, e0);
	if ((d1 != 0) || (d >= p)) { c++; d-=p; }
      }
      uC d2 = d << 1;
      c <<= 1;
      if ((d2 >= p) || (d2 < d)) { c++; d = d2-p; } else d = d2;
      if (s1+1 == s+t-(r-1)) return c;
      d2 = d << 1;
      c <<= 1;
      if ((d2 >= p) || (d2 < d)) c++;
      return c;
    }
  };
  
  TMPL static inline nat
  dyn_r (const C& p) {
    typedef typename unsigned_of_helper<C>::type uC;
    static const nat m = V::template maximum_size_helper<C>::value;
    C abs_p = abs (p);
    nat r = int_bitsize (abs_p);
    return (r == 0) ? m : ( (uC) abs_p == (((uC) 1) << (r-1)) ? r-1 : r );
  }

  TMPL static inline nat
  dyn_s (const C& p, nat r) {
    static const nat m = V::template maximum_size_helper<C>::value;
    return m+2 <= 8 * sizeof(C) ? r-2 : r-1;
  }

  TMPL static inline nat
  dyn_t (const C& p, nat r) {
    static const nat n = 8 * sizeof(C);
    static const nat m = V::template maximum_size_helper<C>::value;
    return (m+2 <= n) ? r+3 : ((m+1 <= n) ? r+1 : r);
  }

  TMPL static inline C
  dyn_q (const C& p, nat r, nat s, nat t) {
    typedef typename unsigned_of_helper<C>::type uC;
    typedef typename unsigned_int_with_double_size_helper<C>::type D;
    return _dynamic_inverse_helper<C,D>::op ((uC) p, r, s, t);
  }

/*
  Static computation of the auxiliaries
*/
  //! Compute the auxiliary quantities attached to the modulus p.
  template<typename C, C p>
  struct auxiliaries_helper {
    static const nat m = V::template maximum_size_helper<C>::value;
    typedef typename unsigned_of_helper<C>::type uC;
    typedef typename unsigned_int_with_double_size_helper<uC>::type uD;

    static const nat n2= 4*sizeof(C);
    static const nat n = 8*sizeof(C);
    static const uC abs_p = (uC) ((p < 0) ? -p : p);
    static const uC a  = MMX_SAFE_LEFT_SHIFT_INT(uC,1,m);
    static const uC up = (abs_p == 0) ? a : abs_p;
    static const nat bitsize = int_bitsize_helper<uC, up>::value;
    static const nat _b = (bitsize == 0) ? n : bitsize-1;
    static const nat r  = (bitsize == 0) ? n :
      ( (up == MMX_SAFE_LEFT_SHIFT_INT(uC,1,_b)) ?
	bitsize-1 : bitsize );
    static const nat s = (m+2 <= n) ? r-2 : r-1;
    static const nat t = (m+2 <= n) ? r+3 : ((m+1 <= n) ? r+1 : r);

    template <typename D, typename Dummy> 
    struct _inverse_helper {
      static const uD _up = (up == 0) ? ((uD) 1) << m : (uD) up;
      static const nat _s = (s+t >= 2*n) ? 2*n : s+t;
      static const uD _t  = (uD) (MMX_SAFE_LEFT_SHIFT_INT(uD,1,_s));
      static const uC value = (up == 1 && m >= n) ?
                              (uC) 0 : (uC) (_t / (uD) _up);
      static inline C dyn_value () { return value; }
    };

    template <typename Dummy> 
    struct _inverse_helper <void, Dummy> {
      static inline C dyn_value () { return dyn_q ((C) up, r, s, t); }
    };

    static inline C q () {
      return _inverse_helper<uD, void>::dyn_value (); }
  };

/*
  Modular product when int of double size are available
*/

  template <typename C, typename D, nat m>
  struct mul_mod_helper {
    static const nat n = 8 * sizeof(C);
    // C and D are unsigned.
    static inline void
    op (C& dest, const C& src, const C& p, const C& q, nat r, nat s, nat t) {
      D a (dest);
      a *= src;
      if (m + 2 <= n) {
	C hi_a = a >> s;
	C b = (((D) hi_a) * q) >> t;
	dest = ((C) a) - b * p;
	if (dest >= p) dest -=p;
	return;
      }
      D hi_a = a >> s;
      C b = (hi_a * q) >> t;
      D c = a - ((D) b) * p;
      if (m + 1 <= n) {
	if (c >= p) {
	  dest = c - p;
	  if (dest >= p)
	    dest -= p;
	}
	else
	  dest = c;
	return;
      }
      if (c >= p) {
	c -= p;
	if (c >= p) {
	  c -= p;
	  if (c >= p) 
	    c -= p;
	}
      }
      dest = c;
    }
  };

/*
  Modular product when int of double size are not available
*/

  template <typename C, nat m>
  struct mul_mod_helper<C, void, m> {
    // C is unsigned.
    static const nat n2 = 4 * sizeof(C);
    static const nat n  = 8 * sizeof(C);
    typedef long_int_mul_op<C>    long_mul;
    typedef long_int_rshift_op<C> long_rshift;
    typedef long_int_sub_op<C>    long_sub;
    typedef long_int_ge_op<C>     long_ge;

    static void
    op (C& dest, const C& src, const C& p, const C& q,
	nat r, nat s, nat t) {

      C a1, a0, hi_a1, hi_a0;
      C tmp = dest;
      
      long_mul::op (a1, a0, tmp, src);
      hi_a1 = a1;
      hi_a0 = a0;
      long_rshift::op (hi_a1, hi_a0, s);
      C b1, b0;
      long_mul::op (b1, b0, hi_a0, q);
      if (m >= n) {if (hi_a1 != 0) b1 += q;}
      if (m+1 <= n) {
	if (t >= n)
	  if (n >= t) b0 = b1 << (n-t); else  b0 = b1 >> (t-n);
	else
	  b0 = (b0 >> t) | (b1 << (n-t));
      }
      else
	long_rshift::op (b1, b0, t);
      if (m+2 <= n) {
	dest = a0 - b0*p;
	dest = min (dest, dest - p);
	return;
      }
      C d1, d0;
      long_mul::op (d1, d0, b0, p);
      long_sub::op (a1, a0, d1, d0);
      if (m+1 <= n) {
	if (long_ge::op (a1, a0, 0, p)) {
	  dest = a0 - p;
	  if (dest >= p)
	    dest -=p; }
	else dest = a0;
	return;
      }
      if (long_ge::op (a1, a0, 0, p)) {
	long_sub::op (a1, a0, 0, p);
	if (long_ge::op (a1, a0, 0, p)) {
	  long_sub::op (a1, a0, 0, p);
	  if (long_ge::op (a1, a0, 0, p)) 
	    a0 -= p;
	}
      }
      dest = a0;
    }
  };
  
  template<typename C, typename M> static inline void
  mul_mod (C& dest, const C& src, const M& x) {
    static const nat m = V::template maximum_size_helper<C>::value;
    typedef typename unsigned_of_helper<C>::type uC;
    typedef typename unsigned_int_with_double_size_helper<uC>::type uD;
    uC tmp = dest;
    mul_mod_helper<uC,uD,m>::op (tmp, (uC) src, (uC) x.p, (uC) x.q, 
				 x.r, x.s, x.t);
    dest = tmp; }

  template<typename C, typename M> static inline void
  mul_mod (C& dest, const C& s1, const C& s2, const M& x) {
    dest = s1;
    mul_mod (dest, s2, x); }
};


template<nat m>
struct modulus_preinverse_int:
  public modulus_encoding_naive_int<
         modulus_div_naive_int<
         modulus_inv_naive_int<
	 modulus_mul_preinverse_int<
	 modulus_add_naive_int<
         modulus_reduction_naive_int<
	 modulus_normalization_naive_int<
	 modulus_maximum_size_int<m> > > > > > > > {};

/******************************************************************************
* Specialized modulus
******************************************************************************/

template<typename C, nat m>
class modulus <C, modulus_preinverse_int<m> > {
MMX_ALLOCATORS
public:
  C p;
  C q;
  nat r;
  nat s;
  nat t;
  typedef C base;
  typedef modulus_preinverse_int<m> variant;
  typedef modulus <C, modulus_preinverse_int<m> > self_type;

  inline C operator * () const { return p; }

  inline modulus () : p (0) {
    r = variant::dyn_r (p);
    s = variant::dyn_s (p, r);
    t = variant::dyn_t (p, r);
    q = variant::dyn_q (p, r, s, t);
  }

  template<typename X>
  inline modulus (const X& x) {
    p = (C) x;
    ASSERT (variant::normalize (p), "modulus: bad value");
    r = variant::dyn_r (p);
    s = variant::dyn_s (p, r);
    t = variant::dyn_t (p, r);
    q = variant::dyn_q (p, r, s, t);
  }

  template<typename X, X _p>
  inline modulus (const fixed_value<X, _p>& x) {
    typedef typename variant::template auxiliaries_helper<C, _p> A;
    C loc_p = _p;
    ASSERT (variant::normalize (loc_p), "modulus: bad value");
    p = A::up;
    q = A::q ();
    r = A::r;
    s = A::s;
    t = A::t; }
    
  inline modulus (const self_type& x) : p (x.p), q (x.q),
					r (x.r), s (x.s),
					t (x.t) {}

  inline self_type&
  operator = (const self_type& x) {
    p = x.p;
    q = x.q;
    r = x.r;
    s = x.s;
    t = x.t;
    return *this; }
};

#undef TMPL
} // namespace mmx
#endif //__MMX__MODULUS_PREINVERSE_INT__HPP
