
/******************************************************************************
* MODULE     : tangent.hpp
* DESCRIPTION: Efficient computations in first order jet spaces C[X]/(X^2)
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_TANGENT_HPP
#define __MMX_TANGENT_HPP
#include <numerix/rational.hpp>
#include <basix/type_props.hpp>
namespace mmx {
#define TMPL_DEF template<typename C, typename D=C>
#define TMPL template<typename C,typename D>
#define Tangent tangent<C,D>
#define Abs_tangent tangent<Abs_type(C),Abs_type(D) >
#define Real_tangent tangent<Real_type(C),Real_type(D) >
TMPL class tangent;
TMPL inline C base  (const Tangent& z);
TMPL inline D slope (const Tangent& z);
TMPL inline Tangent operator - (const Tangent& z1, const Tangent& z2);

/******************************************************************************
* The tangent type
******************************************************************************/

TMPL_DEF
class tangent {
MMX_ALLOCATORS
private:
  C c;
  D d;

public:
  inline tangent (): c (0), d (0) {}
  template<typename T> tangent (const T& c2): c (c2), d (0) {}
  template<typename CT,typename DT> inline tangent (const tangent<CT,DT>& z):
    c (base (z)), d (slope (z)) {}
  inline tangent (const C& z): c (copy (z)), d (0) {}
  inline tangent (const C& c2, const D& d2): c (c2), d (d2) {}
  friend C base LESSGTR (const Tangent& z);
  friend D slope LESSGTR (const Tangent& z);
};

TMPL struct rounding_helper<Tangent >: public rounding_helper<C> {};

TMPL
struct constant_helper<Tangent > {
  static inline Tangent nan () { return Tangent (Nan (C), 0); }
  static inline Tangent plus_infinity () {
    return Tangent (Plus_infinity (C), 0); }
  static inline Tangent minus_infinity () {
    return Tangent (Minus_infinity (C), 0); }
  static inline Tangent infinity () { return Tangent (Infinity (C), 0); }
  static inline Tangent fuzz () { return Tangent (Fuzz (C), 0); }
  static inline Tangent log2 () { return Tangent (Log2 (C), 0); }
  static inline Tangent pi () { return Tangent (Pi (C), 0); }
  static inline Tangent euler () { return Tangent (Euler (C), 0); }
  static inline Tangent imaginary () { return Tangent (Imaginary (C), 0); }
};

/******************************************************************************
* Basic routines
******************************************************************************/

TMPL inline C base (const Tangent& z) {
  return z.c; }
TMPL inline D slope (const Tangent& z) {
  return z.d; }

template<typename Op, typename C, typename D> nat
unary_hash (const Tangent& x) {
  nat h= Op::op (base (x));
  return (h<<3) ^ (h<<11) ^ (h>>29) ^ Op::op (slope (x));
}

template<typename Op, typename C, typename D> bool
binary_test (const Tangent& x1, const Tangent& x2) {
  return Op::op (base (x1), base (x2)) &&
         Op::op (slope (x1), slope (x2));
}

TRUE_IDENTITY_OP_SUGAR(TMPL,Tangent)
SOFT_IDENTITY_OP_SUGAR(TMPL,Tangent)
HARD_IDENTITY_OP_SUGAR(TMPL,Tangent)

TMPL syntactic
flatten (const Tangent& z) {
  return apply ("tangent", flatten (base (z)), flatten (slope (z)));
}

PRINT_SUGAR(TMPL,Tangent)

/******************************************************************************
* Comparisons
******************************************************************************/

TMPL inline bool
operator <= (const Tangent& z1, const Tangent& z2) {
  return base (z1) < base (z2) || z1 == z2;
}

TMPL inline bool
operator < (const Tangent& z1, const Tangent& z2) {
  return base (z1) < base (z2);
}

TMPL inline bool
operator >= (const Tangent& z1, const Tangent& z2) {
  return base (z1) > base (z2) || z1 == z2;
}

TMPL inline bool
operator > (const Tangent& z1, const Tangent& z2) {
  return base (z1) > base (z2);
}

EQUAL_INT_SUGAR(TMPL,Tangent,Tangent);
COMPARE_INT_SUGAR(TMPL,Tangent,Tangent);

/******************************************************************************
* Other order related functions
******************************************************************************/

TMPL inline Tangent
min (const Tangent& z1, const Tangent& z2) {
  if (z1 <= z2) return z1; else return z2;
}

TMPL Tangent
max (const Tangent& z1, const Tangent& z2) {
  if (z1 >= z2) return z1; else return z2;
}

/******************************************************************************
* Basic arithmetic
******************************************************************************/

TMPL inline Tangent
operator - (const Tangent& z) {
  return Tangent (-base (z), -slope (z));
}

TMPL inline Tangent
operator + (const Tangent& z1, const Tangent& z2) {
  return Tangent (base (z1) + base (z2), slope (z1) + slope (z2));
}

TMPL inline Tangent
operator - (const Tangent& z1, const Tangent& z2) {
  return Tangent (base (z1) - base (z2), slope (z1) - slope (z2));
}

TMPL inline Tangent
operator * (const Tangent& z1, const Tangent& z2) {
  return Tangent (base (z1) * base (z2),
		  base (z1) * slope (z2) + slope (z1) * base (z2));
}

TMPL inline Tangent
operator * (const C& z1, const Tangent& z2) {
  return Tangent (z1 * base (z2), z1 * slope (z2));
}

TMPL inline Tangent
operator * (const Tangent& z1, const C& z2) {
  return Tangent (base (z1) * z2, slope (z1) * z2);
}

TMPL inline Tangent
square (const Tangent& z) {
  return Tangent (square (base (z)), 2 * base (z) * slope (z));
}

TMPL inline Tangent
operator / (const Tangent& z1, const Tangent& z2) {
  C inv= 1 / base (z2);
  return Tangent (inv * base (z1),
		  inv * (slope (z1) - inv * base (z1) * slope (z2)));
}

TMPL inline Tangent
operator / (const C& z1, const Tangent& z2) {
  C inv= 1 / base (z2);
  return Tangent (inv * z1, - square (inv) * z1 * slope (z2));
}

TMPL inline Tangent
operator / (const Tangent& z1, const C& z2) {
  C inv= 1 / z2;
  return Tangent (inv * base (z1), inv * slope (z1));
}

TMPL inline Tangent
operator << (const Tangent& z, int shift) {
  return Tangent (base (z) << shift, slope (z) << shift);
}

TMPL inline Tangent
operator >> (const Tangent& z, int shift) {
  return Tangent (base (z) >> shift, slope (z) >> shift);
}

ARITH_INT_SUGAR(TMPL,Tangent,Tangent);

/******************************************************************************
* Elementary functions
******************************************************************************/

TMPL inline Tangent
sqrt (const Tangent& z) {
  C s= sqrt (base (z));
  return Tangent (s, slope (z) / (2 * s));
}

TMPL inline Tangent
exp (const Tangent& z) {
  C e= exp (base (z));
  return Tangent (e, e * slope (z));
}

TMPL inline Tangent
log (const Tangent& z) {
  return Tangent (log (base (z)), slope (z) / base (z));
}

template<typename C, typename D, typename K> inline Tangent
pow (const Tangent& z, const K& y) {
  return exp (Tangent (y) * log (z));
}

TMPL inline Tangent
cos (const Tangent& z) {
  return Tangent (cos (base (z)), -sin (base (z)) * slope (z));
}

TMPL inline Tangent
sin (const Tangent& z) {
  return Tangent (sin (base (z)), cos (base (z)) * slope (z));
}

TMPL inline Tangent
tan (const Tangent& z) {
  return sin (z) / cos (z);
}

TMPL inline Tangent
cosh (const Tangent& z) {
  return Tangent (cosh (base (z)), sinh (base (z)) * slope (z));
}

TMPL inline Tangent
sinh (const Tangent& z) {
  return Tangent (sinh (base (z)), cosh (base (z)) * slope (z));
}

TMPL Tangent
tanh (const Tangent& z) {
  return sinh (z) / cosh (z);
}

TMPL Tangent
acos (const Tangent& z) {
  C df= -1 / sqrt (1 - square (base (z)));
  return Tangent (acos (base (z)), df * slope (z));
}

TMPL Tangent
asin (const Tangent& z) {
  C df= 1 / sqrt (1 - square (base (z)));
  return Tangent (asin (base (z)), df * slope (z));
}

TMPL Tangent
atan (const Tangent& z) {
  C df= 1 / (1 + square (base (z)));
  return Tangent (asin (base (z)), df * slope (z));
}

TMPL Tangent
atan2 (const Tangent& y, const Tangent& x) {
  ERROR ("not yet implemented");
  return Tangent (atan2 (base (y), base (x)), 0);
}

HYPOT_SUGAR(TMPL,Tangent)
INV_TRIGO_SUGAR(TMPL,Tangent)
INV_HYPER_SUGAR(TMPL,Tangent)
ARG_HYPER_SUGAR(TMPL,Tangent)

/******************************************************************************
* Miscellaneous routines
******************************************************************************/

TMPL inline nat precision (const Tangent& z) {
  return precision (base (z)); }
TMPL inline double magnitude (const Tangent& z) {
  return magnitude (base (z)); }
TMPL inline mp_exp_t exponent (const Tangent& z) {
  return exponent (base (z)); }
TMPL inline double double_exponent (const Tangent& z) {
  return double_exponent (base (z)); }

STYPE_TO_TYPE(TMPL,scalar_type,Tangent,C);
STYPE_TO_TYPE(TMPL,real_type,Tangent,Real_tangent);
STYPE_TO_TYPE(TMPL,abs_type,Tangent,Abs_tangent);

TMPL inline Abs_tangent
abs (const Tangent& z) {
  // FIXME: this is not really nice
  return Abs_tangent (abs (base (z)), abs (slope (z)));
}

TMPL inline Real_tangent
Re (const Tangent& z) {
  return Real_tangent (Re (base (z)), Re (slope (z)));
}

TMPL inline Real_tangent
Im (const Tangent& z) {
  return Real_tangent (Im (base (z)), Im (slope (z)));
}

#undef TMPL_DEF
#undef TMPL
#undef Tangent
#undef Abs_tangent
#undef Real_tangent
} // namespace mmx
#endif // __MMX_TANGENT_HPP
