
/******************************************************************************
* MODULE     : twin.hpp
* DESCRIPTION: Twin numbers
* COPYRIGHT  : (C) 2007  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_TWIN_HPP
#define __MMX_TWIN_HPP
#include <basix/operators.hpp>
namespace mmx {
struct std_twin {};
#define TMPL_DEF template<typename C, typename D=generic, typename V=std_twin>
#define TMPL template<typename C, typename D, typename V>
#define Twin twin<C,D,V>
TMPL class twin;
TMPL inline C car (const Twin& x);
TMPL inline D cdr (const Twin& x);

/******************************************************************************
* The twin type
******************************************************************************/

TMPL_DEF
class twin {
MMX_ALLOCATORS;
private:
  C x1;
  D x2;
public:
  inline twin (): x1 (0), x2 (0) {}
  template<typename T> inline twin (const T& x):
    //x1 (x), x2 (x) {}
    x1 (as<C> (x)), x2 (as<D> (x)) {}
  template<typename T, typename U> inline twin (const T& y1, const U& y2):
    // x1 (y1), x2 (y2) {}
    x1 (as<C> (y1)), x2 (as<D> (y2)) {}
  inline twin (const int& x):
    x1 (x), x2 (x) {}
  template<typename T> inline twin (const T& y1, const int& y2):
    x1 (as<C> (y1)), x2 (y2) {}
  template<typename W> inline twin (const twin<C,V,W>& x):
    x1 (car (x)), x2 (cdr (x)) {}
  friend C car LESSGTR (const Twin& x);
  friend D cdr LESSGTR (const Twin& x);
};

TMPL
struct constant_helper<Twin > {
  static inline Twin nan () {
    return Twin (Nan (C), Nan (D)); }
  static inline Twin infinity () {
    return Twin (Infinity (C), Infinity (D)); }
  static inline Twin plus_infinity () {
    return Twin (Plus_infinity (C), Plus_infinity (D)); }
  static inline Twin minus_infinity () {
    return Twin (Minus_infinity (C), Minus_infinity (D)); }
  static inline Twin fuzz () {
    return Twin (Fuzz (C), Fuzz(D)); }
  static inline Twin log2 () { return Twin (Log2 (C), Log2 (D)); }
  static inline Twin pi () { return Twin (Pi (C), Pi (D)); }
  static inline Twin euler () { return Twin (Euler (C), Euler (D)); }
  static inline Twin imaginary () { return Twin (Imaginary(C), Imaginary(D)); }
};

template<typename C1, typename D1, typename V1,
	 typename C2, typename D2, typename V2>
struct as_helper<twin<C1,D1,V1>,twin<C2,D2,V2> > {
  static inline twin<C1,D1,V1> cv (const twin<C2,D2,V2>& x) {
    return twin<C1,D1,V1> (as<C1> (car (x)), as<D1> (cdr (x))); }
};

TMPL generic
construct (const Twin& x) {
  return construct (as<generic> (x));
}

/******************************************************************************
* Basic routines
******************************************************************************/

TMPL inline C car (const Twin& x) { return x.x1; }
TMPL inline D cdr (const Twin& x) { return x.x2; }

TMPL inline Twin copy (const Twin& x) {
  return Twin (copy (car (x)), copy (cdr (x))); }

template<typename Op, typename C, typename D, typename V> nat
unary_hash (const Twin& x) {
  return Op::op (car (x));
}

TMPL inline syntactic
flatten (const Twin& x) {
  return flatten (cdr (x));
}

PRINT_SUGAR(TMPL,Twin)

/******************************************************************************
* Equality and ordering relations
******************************************************************************/

TMPL bool operator == (const Twin& x, const Twin& y) {
  return car (x) == car (y); }
TMPL bool operator != (const Twin& x, const Twin& y) {
  return car (x) != car (y); }
TMPL bool operator < (const Twin& x, const Twin& y) {
  return car (x) < car (y); }
TMPL bool operator <= (const Twin& x, const Twin& y) {
  return car (x) <= car (y); }
TMPL bool operator > (const Twin& x, const Twin& y) {
  return car (x) > car (y); }
TMPL bool operator >= (const Twin& x, const Twin& y) {
  return car (x) >= car (y); }

EQUAL_INT_SUGAR(TMPL,Twin,Twin);
COMPARE_INT_SUGAR(TMPL,Twin,Twin);

TMPL nat hash (const Twin& x) { return hash (car (x)); }
TMPL nat soft_hash (const Twin& x) { return soft_hash (cdr (x)); }
TMPL nat hard_hash (const Twin& x) { return hard_hash (cdr (x)); }

TMPL bool eq (const Twin& x, const Twin& y) {
  return eq (cdr (x), cdr (y)); }
TMPL bool neq (const Twin& x, const Twin& y) {
  return neq (cdr (x), cdr (y)); }
TMPL bool hard_eq (const Twin& x, const Twin& y) {
  return hard_eq (cdr (x), cdr (y)); }
TMPL bool hard_neq (const Twin& x, const Twin& y) {
  return hard_neq (cdr (x), cdr (y)); }

/******************************************************************************
* Basic arithmetic
******************************************************************************/

TMPL inline Twin
operator + (const Twin& x1, const Twin& x2) {
  return Twin (car (x1) + car (x2), cdr (x1) + cdr (x2));
}

TMPL inline Twin
operator - (const Twin& x) {
  return Twin (-car (x), -cdr (x));
}

TMPL inline Twin
operator - (const Twin& x1, const Twin& x2) {
  return Twin (car (x1) - car (x2), cdr (x1) - cdr (x2));
}

TMPL inline Twin
operator * (const Twin& x1, const Twin& x2) {
  return Twin (car (x1) * car (x2), cdr (x1) * cdr (x2));
}

TMPL inline Twin
operator / (const Twin& x1, const Twin& x2) {
  return Twin (car (x1) / car (x2), cdr (x1) / cdr (x2));
}

TMPL inline Twin
operator << (const Twin& x, int shift) {
  return Twin (car (x) << shift, cdr (x) << shift);
}

TMPL inline Twin
operator >> (const Twin& x, int shift) {
  return Twin (car (x) >> shift, cdr (x) >> shift);
}

ARITH_INT_SUGAR(TMPL,Twin,Twin)

/******************************************************************************
* Elementary functions
******************************************************************************/

TMPL Twin sqrt (const Twin& x) {
  return Twin (sqrt (car (x)), sqrt (cdr (x))); }
TMPL Twin exp (const Twin& x) {
  return Twin (exp (car (x)), exp (cdr (x))); }
TMPL Twin log (const Twin& x) {
  return Twin (log (car (x)), log (cdr (x))); }
TMPL inline Twin pow (const Twin& x1, const Twin& x2) {
  return Twin (pow (car (x1), car (x2)), pow (cdr (x1), cdr (x2))); }
TMPL inline Twin pow (const Twin& x1, const int& x2) {
  return pow (x1, Twin (x2)); }
TMPL Twin cos (const Twin& x) {
  return Twin (cos (car (x)), cos (cdr (x))); }
TMPL Twin sin (const Twin& x) {
  return Twin (sin (car (x)), sin (cdr (x))); }
TMPL Twin tan (const Twin& x) {
  return Twin (tan (car (x)), tan (cdr (x))); }
TMPL Twin acos (const Twin& x) {
  return Twin (acos (car (x)), acos (cdr (x))); }
TMPL Twin asin (const Twin& x) {
  return Twin (asin (car (x)), asin (cdr (x))); }
TMPL Twin atan (const Twin& x) {
  return Twin (atan (car (x)), atan (cdr (x))); }
TMPL Twin cosh (const Twin& x) {
  return Twin (cosh (car (x)), cosh (cdr (x))); }
TMPL Twin sinh (const Twin& x) {
  return Twin (sinh (car (x)), sinh (cdr (x))); }
TMPL Twin tanh (const Twin& x) {
  return Twin (tanh (car (x)), tanh (cdr (x))); }
TMPL Twin acosh (const Twin& x) {
  return Twin (acosh (car (x)), acosh (cdr (x))); }
TMPL Twin asinh (const Twin& x) {
  return Twin (asinh (car (x)), asinh (cdr (x))); }
TMPL Twin atanh (const Twin& x) {
  return Twin (atanh (car (x)), atanh (cdr (x))); }
TMPL Twin atan2 (const Twin& x, const Twin& y) {
  return Twin (atan2 (car (x), car (y)), atan2 (cdr (x), cdr (y))); }
TMPL Twin hypot (const Twin& x, const Twin& y) {
  return Twin (hypot (car (x), car (y)), hypot (cdr (x), cdr (y))); }

/******************************************************************************
* Miscellaneous
******************************************************************************/

TMPL inline bool is_exact_zero (const Twin& x) {
  return is_exact_zero (car (x)); }
TMPL inline bool is_number (const Twin& x) {
  return is_number (car (x)); }
TMPL inline bool is_nan (const Twin& x) {
  return is_nan (car (x)); }
TMPL inline bool is_infinity (const Twin& x) {
  return is_infinity (car (x)); }
TMPL inline bool is_fuzz (const Twin& x) {
  return is_fuzz (car (x)); }

TMPL inline Twin sharpen (const Twin& x) {
  return Twin (sharpen (car (x)), cdr (x)); }
TMPL inline nat precision (const Twin& x) {
  return precision (car (x)); }
TMPL inline mp_exp_t exponent (const Twin& x) {
  return exponent (car (x)); }
TMPL inline mp_exp_t step (const Twin& x) {
  return step (car (x)); }
TMPL inline int var_sign (const Twin& x) {
  return var_sign (car (x)); }

#undef TMPL_DEF
#undef TMPL
#undef Twin
} // namespace mmx
#endif // __MMX_TWIN_HPP
