#ifndef SYNAPS_UPOL_EUCLIDEAN_H
#define SYNAPS_UPOL_EUCLIDEAN_H

#include <subdivix/binomial.hpp>
#include <subdivix/UPOLDSE_m.hpp>

//======================================================================
namespace mmx {
//======================================================================


/// Module for Euclidean ring.
namespace EUCLIDEAN
{
  using UPOLDSE::lcoeff;
  using UPOLDSE::degree;

  /** Quotient and remainder of \c a by \c b.
      The polynomial \c a is modified. At the end of the computation,
      it is equal to the remainder in the euclidean division. The type
      \c R should provide a \c degree function and a direct access
      operator. **/
  template<class R>
  void div_rem(R & q, R & a, const R & b)
  {
    
    typedef typename R::iterator iterator;
    R t;
    typename R::value_type lb=b[degree(b)];
    q=0;
    while (degree(a)>=degree(b)) {
      R m( a[degree(a)]/lb, degree(a)-degree(b) );
      t=b*m;
      t[degree(a)]=a[degree(a)];
      a-=t;
	q += m;
    }
  }
    

  /** Compute the gcd of \c a and \c b with the coefficienst of the Bezout identity  
      stored in u0, v0. **/
  template<class R>
  R extended_gcd(R& u0,  R& v0, const R& a, const R& b)
    {
      R r0(a), r1(b), q, p;
      u0= 1; v0=0;
      R u1(0), v1(1);
      while (r1 !=0) 
	{
	  div_rem(q, r0, r1);
	  p = u0-u1*q; u0=u1; u1=p;
	  p = v0-v1*q; v0=v1; v1=p;
	  p = r0; r0=r1; r1=p;
	}
      return r0;
    }
    

  /** Pseudo quotient and remainder of \c a by \c b.
	The polynomial \c a is modified. At the end of the computation,
	it is equal to the remainder in the euclidean division. The type
	\c R should provide a \c degree function and a direct access
	operator. **/
  template<class R>
	void pseudo_div_rem_greedy(R & q, R & a, const R & b)
	{
	  typedef typename R::iterator iterator;
	  R t;
	  typename R::value_type lb=b[degree(b)];
	  q=0;
	  while (degree(a)>=degree(b)) {
		R m(a[degree(a)],degree(a)-degree(b));
		t=b*m;
		a*=lb;
		t[degree(a)]=a[degree(a)];
		a-=t;
		q+=m;
	  }
	}


  /** Pseudo quotient and remainder of \c f by \c g.
	At the end of the computation, \c q is equal to pseudo quotient and
	\c r is equal to the remainder in the euclidean division.
	**/
  template<class Poly>
	void prem(const Poly& f, const Poly& g, Poly& q, Poly & r)
	{
	  r = f * pow(lcoeff(g), degree(f) - degree(g) + 1);
	  q = 0;
	  while (degree(r) >= degree(g)) {
		int delta =  degree(r) - degree(g);
		q = q + Poly(1,delta) * (lcoeff(r)/lcoeff(g));
		r = r - ( (g * Poly(1, delta)) * (lcoeff(r)/lcoeff(g)) );
	  }
	  return;
	}


  /** Pseudo quotient and remainder of \c f by \c g.
   *  Returns the pseudo remainder, while \c q holds the pseudo quotient.
   **/
  template<class Poly>
	Poly prem(const Poly& f, const Poly& g, Poly& q)
	{
	  Poly r(f * pow(lcoeff(g), degree(f) - degree(g) + 1) );
	  q = 0;
	  while (degree(r) >= degree(g)) {
		int delta =  degree(r) - degree(g);
		q = q + Poly(1,delta) * (lcoeff(r)/lcoeff(g));
		r = r - ( (g * Poly(1, delta)) * (lcoeff(r)/lcoeff(g)) );
	  }
	  return r;
	}


  /** Pseudo quotient and remainder of \c f by \c g.
   * Returns the pseudo remainder.
   ***/
  template<class Poly>
    Poly prem(const Poly& f, const Poly& g)
    {
      Poly r(f * pow(lcoeff(g), degree(f) - degree(g) + 1) );
      Poly q = 0;
      while (degree(r) >= degree(g)) {
	int delta =  degree(r) - degree(g);
	//		q = q + Poly(1,delta) * (lcoeff(r)/lcoeff(g));
	r = r - ( (g * Poly(1, delta)) * (lcoeff(r)/lcoeff(g)) );
      }
      return r;
    }
  
  /** Pseudo quotient and remainder of \c f by \c g.
   * Returns the pseudo quotient.
   ***/
  template<class Poly>
  Poly pquo(const Poly& f, const Poly& g)
  {
    Poly r(f * pow(lcoeff(g), degree(f) - degree(g) + 1) );
    Poly q = 0;
    while (degree(r) >= degree(g)) {
      int delta =  degree(r) - degree(g);
      q = q + Poly(1,delta) * (lcoeff(r)/lcoeff(g));
      r = r - ( (g * Poly(1, delta)) * (lcoeff(r)/lcoeff(g)) );
    }
    return q;
  }
  
  
  template<class R>
  void pseudo_div_rem(R & q, R & a, const R & b)
    {
      typedef typename R::iterator iterator;
      R t(degree(b)+1, AsSize());
      typename R::value_type lb=b[degree(b)];
      q=0;
      int c=0, da = degree(a), db = degree(b), dab =da-db+1;
      while (da>=db) {
	R m(a[da],da-db);
	t=b*m;
	c++;
	a*=lb;
	t[da]=a[da];
	a-=t;
	// a.resize(da);
	q+=m;
	da=degree(a);
      }
      a *= pow(lb, dab-c);
    }
  
}
//CLOSE


//======================================================================
} //  namespace mmx 
//======================================================================

#endif // SYNAPS_UPOL_EUCLIDEAN

