/*********************************************************************
*      This file is part of the source code of SYNAPS kernel.        *
*   Author(s): B. Mourrain, GALAAD, INRIA                            *
**********************************************************************/
#ifndef mmx_realroot_sturm_Zmod_HPP
#define mmx_realroot_sturm_Zmod_HPP
//--------------------------------------------------------------------
#include <iostream>
#include <cstdlib>
#include <subdivix/let.hpp>
//--------------------------------------------------------------------
namespace mmx {
//--------------------------------------------------------------------
/** \file                                                           */
/**
   Class used to represent modular numbers. The internal storage is
   of type @T@. The computations are performed modulo @p@.
   The number @p@ should be a prime number, otherwise
   inversion and exponentiation may not be valid.
 **/
template <unsigned int p, class T=int>
class Zmod
{
private:
  T rep_;

public:
  Zmod(): rep_(0) {}
  Zmod(int n): rep_(n<0 ? p-(-n)%p : n%p ){}
  Zmod(const char* s);
  Zmod(const Zmod<p,T> & z): rep_(z.rep_) {}

  T rep() const {return rep_;}
  Zmod<p,T>&  operator= (const Zmod<p,T>  z)
    {rep_=z.rep_; return *this;}
  Zmod<p,T>&  operator+=(const Zmod<p,T>  z)
    {rep_ += z.rep_; if(rep_>=(int)p) rep_-=p; return *this;}
  Zmod<p,T>&  operator-=(const Zmod<p,T>  z)
    {rep_ -= z.rep_; if(rep_<0) rep_+=p; return *this;}
  Zmod<p,T>&  operator*=(const Zmod<p,T>  z)
    {(rep_ *= z.rep_)%= p; if(rep_<0) rep_+=p; return *this;}
  Zmod<p,T>&  operator/=(const Zmod<p,T> &  z);
  bool   operator==(const Zmod<p,T> & z) const {return (rep_==z.rep_);}
  bool   operator!=(const Zmod<p,T> & z) const {return !(*this==z);}

  friend inline
  Zmod<p,T> operator-(const Zmod<p,T> & a) {return Zmod<p,T>(-a.rep_);}
  friend inline
  Zmod<p,T> operator+(const Zmod<p,T>  & a, const  Zmod<p,T> & b){return (Zmod<p,T>(a.rep_+b.rep_));}
  friend inline
  Zmod<p,T> operator-(const Zmod<p,T> & a, const  Zmod<p,T> & b){return (Zmod<p,T>(a.rep_-b.rep_));}
  friend inline
  Zmod<p,T> operator*(const Zmod<p,T> & a, const  Zmod<p,T> & b){return (Zmod<p,T>(a.rep_*b.rep_));}
  friend inline
  Zmod<p,T> operator/(const Zmod<p,T> & a, const  Zmod<p,T> & b){return (Zmod<p,T>(a)/=b);}

  friend
  std::ostream &  operator<<(std::ostream&  os, const Zmod<p,T> & z) {return (os << z.rep_);}
  friend
  std::istream & operator>>(std::istream&  is, Zmod<p,T> & z) {return (is >> z.rep_);}

};


//----------------------------------------------------------------------
/// Conversion of a string to the cooresponding modular number.
//----------------------------------------------------------------------
template <unsigned int p, class T>
Zmod<p,T>::Zmod(const char* s) :rep_(0)
{
  int i=0; while(s[i]!='\0') i++;
  unsigned b=1;
  while(i>0){
    rep_+=(s[i-1]-'0')*b; rep_%=p;
    b*=10; b%=p;
    i--;
  }
}
//----------------------------------------------------------------------
/// Division operator, based on Euclid's algorithm.
//----------------------------------------------------------------------
template <unsigned int p, class T>
Zmod<p,T> & Zmod<p,T>::operator/=(const Zmod<p,T> & z)
{
  Zmod<p,T> u(*this);
  T r=1, a=p, b=z.rep_,  u1=0, u2=1, q;
  if(z==0){ std::cerr<< "xxxx division by zero"<<std::endl;}
  assert(a>0);
  while(b != 0)
    {
      q = a/b;
      r = a-q*b; 
      a=b; 
      b=r;
      r=u2;
      u2=u1-q*u2; 
      u1=r; 
    }
  if(u1<0) u1+=p;
  rep_ *= u1 ; rep_ %= p;
  return *this;
}
//----------------------------------------------------------------------
/// Fast binary exponentiation.
//----------------------------------------------------------------------
template <unsigned int p, class T>
Zmod<p,T> operator^(const Zmod<p,T> a, const  int n)
{
  //  std::cout<<a<<"^"<<n;
  int m;
  Zmod<p,T> r=1, e;
  if(n <0){e=1/a;m=-n;}else {e=a;m=n;}
  while(m !=0) {
    if(m%2){
      m -=1; m/=2; r*=e; e*=e;
    }else{
      m/=2; e*=e;
    }
  }
  //  std::cout <<" = "<<r<<std::endl;
  return r;
}
//--------------------------------------------------------------------
template <unsigned int p, class T>
bool with_plus_sign(const Zmod<p,T>& z)
{ 
  return (z.rep()>0);
}
//--------------------------------------------------------------------
} //namespace mmx
//--------------------------------------------------------------------
#endif // mmx_realroot_sturm_Zmod_HPP

