#ifndef subdivix_euclidean_hpp
#define subdivix_euclidean_hpp
/********************************************************************/
#include <subdivix/Seq.hpp>

namespace mmx {

struct euclidean {


  /** Quotient and remainder of \c a by \c b.
      The polynomial \c a is modified. At the end of the computation,
      it is equal to the remainder in the euclidean division. The type
      \c R should provide a \c degree function and a direct access
      operator. **/
  template<class R> static
  void div_rem(R & q, R & a, const R & b)
  {
    
    typedef typename R::iterator iterator;
    R t;
    typename R::value_type lb=b[degree(b)];
    q=0;
    while (degree(a)>=degree(b)) {
      R m( a[degree(a)]/lb, degree(a)-degree(b) );
      t=b*m;
      t[degree(a)]=a[degree(a)];
      a-=t;
	q += m;
    }
  }

  /** Compute the gcd of \c a and \c b. **/
  template<class R> static
  R gcd(const R& a, const R& b)
  {
    R r0(a), r1(b), q, p;
    while (r1 !=0) 
      {
	div_rem(q, r0, r1);
	p = r0; r0=r1; r1=p;
      }
    return r0;
  }

  /** Compute the gcd of \c a and \c b with the coefficienst of the 
      Bezout identity stored in u0, v0. **/
  template<class R> static
  R extended_gcd(R& u0,  R& v0, const R& a, const R& b)
  {
    R r0(a), r1(b), q, p;
    u0= 1; v0=0;
    R u1(0), v1(1);
    while (r1 !=0) 
      {
	div_rem(q, r0, r1);
	p = u0-u1*q; u0=u1; u1=p;
	p = v0-v1*q; v0=v1; v1=p;
	p = r0; r0=r1; r1=p;
      }
    return r0;
  }

  template<class R> static
  void pseudo_div(R& q, R& a, const R& b)
  {
    typedef typename R::iterator iterator;
    R t(degree(b)+1, AsSize());
    typename R::value_type lb=b[degree(b)];
    q=0;
    int c=0, da = degree(a), db = degree(b), dab =da-db+1;
    while (da>=db) {
      R m(a[da],da-db);
      t=b*m;
      c++;
      a*=lb;
      t[da]=a[da];
      a-=t;
      q+=m;
      da=degree(a);
    }
    a *= pow(lb, dab-c);
  }

  template<class Polynomial> static inline 
  Polynomial prem(const Polynomial& a, const Polynomial& b)
  {
    Polynomial r=a,q;
    pseudo_div(q, r, b);
    return r;
  }

  template<class Polynomial> static inline 
  Polynomial pquo(const Polynomial& a, const Polynomial& b)
  {
    Polynomial r=a, q;
    pseudo_div(q, r, b);
    return q;
  }



};

}// namespace mmx
#endif //subdivix_subresultant_hpp
