/*********************************************************************
*      This file is part of the source code of SYNAPS kernel.        *
*   Author(s): B. Mourrain, GALAAD, INRIA                            *
**********************************************************************
History: 
**********************************************************************/
#ifndef synaps_linalg_matrdse_H
#define synaps_linalg_matrdse_H
//----------------------------------------------------------------------
#include <sstream>
#include <list>
#include <subdivix/let.hpp>
#include <subdivix/type.hpp>
#include <subdivix/texp.hpp>
//#include <subdivix/traits.hpp>
#include <subdivix/vectdse.hpp>
#include <subdivix/seq1d.hpp>
#include <subdivix/linalg_rep2d.hpp>
#include <subdivix/MATRDSE_m.hpp>

#include "matrdse_texp.hpp"

//--------------------------------------------------------------------
namespace mmx {
//====================================================================
/** \file matrdse.hpp
    \brief Dense matrices.
 **/
//--------------------------------------------------------------------
/**
Interface for dense matrices. 
The class \c R is the container corresponding to the internal 
representation. The data are owned by the interface.
The inplace arithmetic operations are members of this class.
The external arithmetic operators +,-,*,/ are defined through
template expression. The different operations of this class are 
implemented in the module \c MATRDSE.
The container \c R should provide the following definitions
or methods:
\code
  typename R::value_type;
  typename R::size_type;
  typename R::iterator;

  R(size_type i, size_type j, value_type* t);
  R(size_type i, size_type j, value_type* t, char c);
  R(size_type s, Alloc);
  R(size_type s, C *t);
  R(iterator b, iterator e);
  void R::reserve(size_type m, size_type m);
 
        value_type & operator() (size_type i, size_type j);
  const value_type & operator() (size_type i, size_type j) const;

  size_type nbrow() const;
  size_type nbcol() const;
\endcode

The \c value_type is the type of the coefficients. The \c size_type
is the type of the indices used to access the elements.  The
\c row_iterator is the type of the forward iterators on the elements of a
row. The \c column_iterator is the type of the forward iterators on the
elements of a column.The \c row_type is the type of the row views.  The
\c col_type is the type of the column views.  The internal representation
type \c R is \c rep_type.

**/

template<class C, class R=linalg::rep2d<C> > 
struct matrdse 
{
  typedef vectdse<C> vector_type;
  typedef vectdse<C> row_type;
  typedef vectdse<C> col_type;
  shared_object<R>  data;
  R &       rep()       {return (*data);}
  const R & rep() const {return (*data);}
  
  typedef typename R::size_type     size_type;
  typedef C                         value_type;
  typedef C                         coeff_t;
    // typedef typename R::row_iterator  row_iterator;
  typedef typename R::col_iterator  col_iterator;
  
  
  typedef R                         rep_type;
  typedef matrdse<C, R>               self_t;
  
  matrdse() {}
  matrdse(const R & r): data(r) {}
  matrdse(size_type m, size_type n) { rep().resize(m,n); }

  /// Construct a \c m by \c n matrix with entries all equal to \c c.
  matrdse(size_type m, size_type n, const value_type & c) { 
    using MATRDSE::init; rep().resize(m,n); init(rep(),c); 
  }

  matrdse(const value_type & c, size_type m, size_type n) { 
    using MATRDSE::init; rep().resize(m,n); init(rep(),c); 
  }

  matrdse(const value_type & c) { rep().resize(1,1); (*this)(0,0)=c;}

  /** 
      Construct a \c m by \c n matrix from a string containing the 
      coefficients of the matrix separated by spaces. 
      These coefficients are read row by row.
  **/
  matrdse( size_type m, size_type n, const char* str, ByRow ){ 
    using MATRDSE::init; std::istringstream ins(str); init(rep(),m,n,ins,ByRow()); 
  }
  /** 
      Same as above but the coefficients are read column by column.
  **/
  matrdse(size_type m, size_type n, const char * str, ByCol= ByCol()) { 
    using MATRDSE::init; std::istringstream ins(str); init(rep(),m,n,ins,ByCol()); 
  };

  /** initialize the matrix from the coefficient data pointed by \c src 
      the coefficients are assumed to be stored in row major mode
      the same function exists in column major mode (ByRow or ByCol)
  **/
  template<class T>
  matrdse(size_type m, size_type n, const value_type * src, const T & t = ByCol()) { 
    using MATRDSE::init; init( this->rep(), m, n, src, t ); 
  };
  // Copy constructors.
  matrdse(const self_t & M): data(M.data) {
  } 
  // Assignement operators
  self_t & operator=(const self_t & M) {
    data = M.data; return *this;
  } 
  
    
  template<class X>
  self_t & operator=(const X & M);
  
  /// return the number of rows 
  size_type nbrow() const {
    return rep().nbrow();
  } 
  
  /// return the number of columns
  size_type nbcol() const {
    return rep().nbcol();
  }
  
  /// return the (i,j) coefficient
  value_type&       operator() (size_type i, size_type j) {return rep()(i,j);}
  
  /// return the (i,j) coefficient
  const value_type& operator() (size_type i, size_type j) const {return rep()(i,j);} 
  
  //  typedef typename meta::binary_operator_prototype< linalg::_submatrix_, self_t, Range2d >::F submatrix_t;
  /// return the submatrix given by rows in range \c I, columns in range \c J
  //  submatrix_t  operator()(const Range & I, const Range & J) {  return submatrix(*this,Range2d(I,J)); };
  
  /// return the submatrix given by the range \c rg
  //  submatrix_t  operator()(const Range2d & rg ) {  return submatrix(*this,rg); };
  
  template<class X>
  self_t & operator+=(const X & x) { *this = *this + x; return *this; }
  
  template<class X>
  self_t & operator-=(const X & x) { *this = *this - x; return *this; };
  
  template<class X>
  self_t & operator*=(const X & x)  { *this =  *this * x; return *this; };
  
  template<class X>
  self_t & operator/=(const X & x)  { MATRDSE::div_ext( *this, x); return *this; };
  /// Equality test
  bool operator== (const self_t & v) const { 
    using MATRDSE::equal; return equal(*this,v); 
  };
  
  bool operator!= (const self_t & v) const { return !(*this==v); }
  
  /// Rows and columns operations.
  self_t & swaprow(size_type i, size_type j) {
    using MATRDSE::swap_row; swap_row(rep(),i,j); return *this;
  };
  
  self_t & swapcol(size_type i, size_type j) { 
    using MATRDSE::swap_col; swap_col(rep(),i,j); return *this;
  };
  
  self_t & add_row(size_type i,size_type j, const value_type & c = 1) {
    using MATRDSE::add_row; add_row(rep(),i,j,c); return *this;
  };
  
  self_t & add_col(size_type i,size_type j, const value_type & c = 1) {
    using MATRDSE::add_col; add_col(rep(),i,j,c); return *this;
  };
  
  /// multiply row \c i by the scalar \c c
  self_t & multrow(size_type i, const value_type & c = 1) {
    using MATRDSE::mult_row; mult_row(*this,i,c); return *this;
  };
  
  /// multiply column \c i by the scalar \c c
  self_t & mult_col(size_type i, const value_type & c = 1)
  {
    using MATRDSE::mult_col; mult_col(*this,i,c); return *this;
  };
  
  /// inplace transposition.
  self_t & transpose() {
    using MATRDSE::in_place_transpose; in_place_transpose(rep()); return *this; 
  };

  /// resize 
  self_t & resize(size_type i,size_type j) { 
    this->rep().resize(i,j); 
    return *this; 
  }
};
//====================================================================
template<class C, class R> inline 
unsigned  nbrow(const matrdse<C,R> & M) {return M.nbrow();}
template<class C, class R> inline 
unsigned  nbcol(const matrdse<C,R> & M) {return M.nbcol();}
//====================================================================
template<class C, class R>
matrdse<C,R>  transpose(const matrdse<C,R> & M)
{
  matrdse<C,R> TM(M); TM.transpose();
  return (TM);  
}

} //namespace mmx
//====================================================================
#include "matrdse_let.hpp"
#include "matrdse_operators.hpp"
#include "matrdse_meta.hpp"
#include "matrdse_io.hpp"
//====================================================================
namespace mmx {
//--------------------------------------------------------------------
template<class C, class R>
template<class X> inline
matrdse<C,R> & matrdse<C,R>::operator=(const X & M)
{ 
  using let::assign; assign(*this,M); return *this;  
} 

//--------------------------------------------------------------------
template<class C, class R> inline 
const C& read(const matrdse<C,R>& m, unsigned i, unsigned j) { return m(i,j); }


template<class C, class R>
bool eq (const matrdse<C,R>& m1, const matrdse<C,R>& m2) {return m1==m2;}
template<class C, class R>
bool neq(const matrdse<C,R>& m1, const matrdse<C,R>& m2) {return m1!=m2;}


template<class C, class R>
unsigned hash (const matrdse<C,R>& m) 
{
  register unsigned i, j, h= 214365, nr= m.nbrow(), nc= m.nbcol();
  for (i=0; i<nr; i++)
    for (j=0; j<nc; j++)
      h= (h<<1) ^ (h<<5) ^ (h>>27) ^ hash(m (i, j));
  return h;
}
template<class C, class R>
unsigned soft_hash (const matrdse<C,R>& m) {return hash(m);}


template<class C, class R>
matrdse<C,R> copy(const matrdse<C,R>& m) {return matrdse<C,R>(nbrow(m), nbcol(m),m.rep().begin(), ByCol());}


template<class C, class R>
bool is_square_matrix(const matrdse<C,R>& m) {return (nbrow(m)==nbcol(m));}
//--------------------------------------------------------------------
} //namespace mmx
#endif // synaps_linalg_matrdse_H


