
/******************************************************************************
* MODULE     : matrix_glue.hpp
* DESCRIPTION: Glue for matrices
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_MATRIX_GLUE_HPP
#define __MMX_MATRIX_GLUE_HPP
#include <subdivix/matrdse.hpp>
#include <subdivix/MATRDSE_m.hpp>
#include <subdivix/MATRDSE_pivoting.hpp>
#include <basix/category_meta.hpp>
#include <basix/category_lift_glue.hpp>
#include <basix/tuple.hpp>
#include <basix/row_tuple.hpp>
#include <numerix/integer.hpp>

namespace mmx {
#define TMPL template<typename C>
#define Matrix matrdse<C>

TMPL generic
flatten (const Matrix& m) {
  if (m.nbrow()*m.nbcol()==1) return flatten (m (0, 0));
  int i, j, nr= m.nbrow(), nc= m.nbcol();
  generic v= comma ();
  for (i=0; i<nr; i++) {
    generic h= comma ();
    for (j=0; j<nc; j++)
      h= comma (h, flatten (m (i, j)));
    v= comma (v, xrow (h));
  }
  return xsqtuple (v);
}


TMPL nat N(const Matrix& m) {return (m.nbrow()*m.nbcol());}
TMPL nat rows (const Matrix& m) {return m.nbrow();}
TMPL nat cols (const Matrix& m) {return m.nbcol();}

TMPL row_tuple<C>
make_row (const tuple<C>& t) {
  return row_tuple<C> (cons (GEN_ROW, cdr (*t)));
}

template<typename Cond, typename C> void
define_standard_row () {
  define_type<always,row_tuple<C> > (gen ("Row", type_name<C> ()));
  define<always> (GEN_ROW, make_row<C>);
  define_from_iterator<always,C> (GEN_ROW);
}

TMPL Matrix
make_matrdse_tuple_row (const tuple<row_tuple<C> >& t) {
  nat i, j, rows= N(t), cols= N(t[0]);
  //  C dummy;
  Matrix r (rows, cols);
  for (i=0; i<rows; i++) {
    ASSERT (N(t[i]) == cols, "unequal row lengths");
    for (j=0; j<cols; j++)
      r(i,j)= t[i][j];
  }
  return r;
}

TMPL Matrix
make_matrdse_tuple_list (const tuple<list<C> >& t) {
  nat i, j, rows= N(t), cols= N(t[0]);
  //  C dummy;
  Matrix r (rows, cols);
  for (i=0; i<rows; i++) {
    ASSERT (N(t[i]) == cols, "unequal row lengths");
    for (j=0; j<cols; j++)
      r(i,j)= t[i][j];
  }
  return r;
}

TMPL Matrix
make_matrdse_tuple (const tuple<C>& t) {
  nat j, rows= 1, cols= N(t);
  Matrix r ( rows, cols);
  for (j=0; j<cols; j++)
    r(0,j)= t[j];
  return r;
}

TMPL Matrix
make_matrdse_list (const integer& nbr, const integer& nbc, const list<C>& t) {
  nat l= N(t), m=as_int(nbr), n=as_int(nbc) ;
  Matrix r(m,n);
  for (nat i=0; i<m ; i++)
    for (nat j=0; j<n ; j++)
      if(i+m*j<l)
	r(i,j)= read(t,i+m*j);
  return r;
}



TMPL
class matrdse_element_rep: public alias_rep<C> {
  alias<Matrix > a;
  nat row, col;
public:
  inline matrdse_element_rep (const alias<Matrix >& a2, nat row2, nat col2):
    a (a2), row (row2), col (col2) {}
  C  get  () const { return read (get_alias (a), row, col); }
  C& open () const { return open_alias (a) (row, col); }
  void close () const { close_alias (a); }
};

TMPL alias<C>
matrdse_element (const alias<Matrix >& a, const integer& i, const integer& j) {
  return new matrdse_element_rep<C> (a, as_int (i), as_int (j));
}

TMPL inline integer matrdse_size (const Matrix& m) { return N(m); }
TMPL inline integer matrdse_rows (const Matrix& m) { return rows(m); }
TMPL inline integer matrdse_cols (const Matrix& m) { return cols(m); }



template<typename T, typename F, typename TR, typename FR>
struct cast_helper<matrdse<T,TR>,
                   matrdse<F,FR> >
{
  static inline matrdse<T,TR>
  cv (const matrdse<F,FR>& m) {
    matrdse<T,TR> r(m.nbrow(),m.nbcol());
    for (unsigned  i=0; i<m.nbrow(); i++) 
      for (unsigned  j=0; j<m.nbcol(); j++) 
	r(i,j)= as<T> (m(i,j));
    return r;
  }
};

//--------------------------------------------------------------------
TMPL bool operator< (const Matrix& v,const Matrix& w) { return true;}
TMPL bool operator> (const Matrix& v,const Matrix& w) { return true;}
TMPL bool operator>=(const Matrix& v,const Matrix& w) { return true;}
TMPL bool operator<=(const Matrix& v,const Matrix& w) { return true;}

TMPL Matrix operator+ (const C& c, const Matrix& m2) { return m2; }
TMPL Matrix operator+ (const Matrix& m1, const C& c)  { return m1;}
TMPL Matrix operator- (const C& c, const Matrix& m2) { return m2; }
TMPL Matrix operator- (const Matrix& m1, const C& c)  { return m1;}
//--------------------------------------------------------------------
template<class C, class R>
matrdse<C,R> operator-(const matrdse<C,R>& m) 
{
    matrdse<C,R> r(m); return r;
}


TMPL inline Matrix  matrdse_transpose (const Matrix& m)  { return MATRDSE::trans_copy (m); }
TMPL inline integer matrdse_rank (const Matrix& m)       { return MATRDSE::rank (m); }
TMPL inline Matrix  matrdse_triangulate (const Matrix& m){ return MATRDSE::triangulate (m); }
TMPL inline Matrix  matrdse_ker (const Matrix& m)        { return MATRDSE::kernel (m); }
TMPL inline Matrix  matrdse_im (const Matrix& m)         { return MATRDSE::image (m); }
TMPL inline Matrix  matrdse_coker (const Matrix& m)      { return MATRDSE::cokernel (m); }
TMPL inline Matrix  matrdse_coim (const Matrix& m)       { return MATRDSE::coimage (m); }
TMPL inline C       matrdse_det (const Matrix& m)        { return MATRDSE::det (m); }
TMPL inline Matrix  matrdse_invert (const Matrix& m)     { return MATRDSE::invert(m); }
TMPL inline Matrix  matrdse_hjoin (const Matrix& m1, const Matrix& m2) { return MATRDSE::horizontal_join (m1, m2); }
TMPL inline Matrix  matrdse_vjoin (const Matrix& m1, const Matrix& m2) { return MATRDSE::vertical_join (m1, m2); }
TMPL inline Matrix  matrdse_krylov (const Matrix& m1, const Matrix& m2){ return MATRDSE::krylov (m1, m2); }

template<typename Hyp, typename C> void
define_matrdse () {
  typedef both<has<C,ring_cat>,Hyp> Cond;
  typedef both<has<C,field_cat>,Hyp> Is_Field;
  //  typedef always Cond;
  
  define_standard_row<Cond,C> ();
  define_type<Cond,Matrix > (gen ("MatrDse", type_name<C> ()));
  define<Cond> ("matrdse", make_matrdse_tuple_row<C>);
  define<Cond> ("matrdse", make_matrdse_tuple_list<C>);
  define<Cond> ("matrdse", make_matrdse_tuple<C>);
  define<Cond> ("matrdse", make_matrdse_list<C>);

  //  define_upgrader<Cond,Matrix,C> ();
  define_from_iterator<Cond,C> ("matrdse");

  define<Cond> (GEN_SIZE, matrdse_size<C>);
  define<Cond> ("rows", matrdse_rows<C>);
  define<Cond> ("columns", matrdse_cols<C>);
  define<Cond> (GEN_ACCESS, matrdse_element<C>);

  define_ring<Cond,Matrix > ();
//  lift_scalar_ring_ops<Cond,Matrix,C> ();

  define<Cond> ("horizontal_join", matrdse_hjoin<C>);
  define<Cond> ("vertical_join"  , matrdse_vjoin<C>);

  define<Cond> ("transpose", matrdse_transpose<C>);
  define<Cond> ("rank", matrdse_rank<C>);
  define<Cond> ("triangulate", matrdse_triangulate<C>);
  define<Cond> ("ker", matrdse_ker<C>);
  define<Cond> ("img", matrdse_im<C>);
  define<Cond> ("coker", matrdse_coker<C>);
  define<Cond> ("coimg", matrdse_coim<C>);
  define<Cond> ("det", matrdse_det<C>);
  define<Is_Field> ("inverse", matrdse_invert<C>);
  define<Cond> ("krylov", matrdse_krylov<C>);

}

#undef TMPL
#undef Matrix
} // namespace mmx
#endif // __MMX_MATRIX_GLUE_HPP
