/*********************************************************************
*      This file is part of the source code of SYNAPS kernel.        *
*      Author(s): B. Mourrain, GALAAD, INRIA                         *
**********************************************************************
$Id: monomial.hpp,v 1.1 2005/07/11 11:17:56 mourrain Exp $
**********************************************************************/
#ifndef SYNAPS_MPOL_MONOM_H
#define SYNAPS_MPOL_MONOM_H
//--------------------------------------------------------------------
#include <assert.h>
#include <map>
#include <string>
#include <cstring>
#include <subdivix/shared_object.hpp>
#include <subdivix/parser.hpp>
#include <subdivix/texp.hpp>
#include <subdivix/VECTOR_m.hpp>
#include <subdivix/MONOMIAL_m.hpp>
#include <subdivix/dynamicexp.hpp>
#include <subdivix/Variables.hpp>

#define TMPL template<class R, class O>
#define Monomial monomial<R,O>
//======================================================================
namespace mmx {
//======================================================================

//--------------------------------------------------------------------
/** \file monomial.hpp                                                   */
//--------------------------------------------------------------------
/**
  Implementation of monomials. The monomial class has
  tow template parameters, \c C the arithmetic class of the coefficients and
  \c R a container of int to store the powers.
**/
template <class C, class R=dynamicexp<unsigned> >
struct monomial {

  //data
  C                 _coef;
#ifdef SHARED_MONOM
  shared_object<R>  _expt;
  R &       rep()       {return (*_expt);}
  const R & rep() const {return (*_expt);}
  typedef  shared_object<R> exp_t;
#else
  R  _expt;
  R &       rep()       {return (_expt);}
  const R & rep() const {return (_expt);}
  typedef  R exp_t;
#endif

  static Variables var;

  // Type definitions
  typedef int                    index_t;
  typedef R                      container_t;
  typedef typename R::exponent_t exponent_t; //typename R::coeff_t exponent_t;
  typedef typename R::size_type  size_type;
  typedef C                      coeff_t;
  typedef monomial<C,R>             self_t;

  // Constructors
  monomial(): _coef(),_expt()
    {
      rep().resize(0); //COUNT<self_t>(' ');
    }

  monomial(const C & c, const exp_t & r):_coef(c), _expt(r){}
  monomial(const C & c): _coef(c),_expt()
    {
      rep().resize(0);
    } 
  monomial(int i):_coef(i), _expt()
    {
      rep().resize(0);
    }
  /// monomial x[i]**d
  monomial(int i, int d);                 
  /// monomial x[i]**d
  monomial(const std::string & s, int d, Variables& vars = Variables::default_ );  
  monomial(const C & c, unsigned s, const AsSize & S):_coef(c),_expt(s){}

  /// monomial c*x[i]**d
  monomial(const C & c, int i, int d =1); 
  monomial(const C & c, int s, int *t):_coef(c), _expt(s){
    for(int i=0;i<s;i++) rep()[i]= t[i];
  }
  monomial(const char *, Variables& vars =  monomial::var);

  //Copy constructor
  monomial(const monomial & m):_coef(m._coef),_expt(m._expt) {}

  // Destructor
  // no dynamic allocation at this level

  // Assignement
  self_t & operator = (const self_t & m)
    {
      _coef = m.coeff(); 
      _expt = m._expt; 
      return *this;
    };

  // Accessors
  const C & coeff () const { return _coef; }
  C& coeff ()       { return _coef; }

  void set_coeff (const C & c) { _coef = c; }

  unsigned size()  const    { return rep().size(); }
  int      nvars() const    { return lvar(rep()); }
  int      l_variable() const { return lvar(rep()); }
  unsigned nbvar() const    { return lvar(rep())+1; }

  exponent_t operator[] (size_type i) const { return rep()[i]; }

  self_t & set_expt (size_type i, exponent_t d)
    {rep().set_expt(i,d);return *this; }

  // boolean operators and functions
  bool operator==(int n) const;
  bool operator!=(int n) const {return !(*this==n);}


  friend bool operator == (const self_t & m1, const self_t & m2) 
  {
    return m1._coef == m2._coef && m1.rep() == m2.rep();
  }

  friend bool  operator != (const self_t & m1, const self_t & m2) 
  {
    return !(m1 == m2);
  }

  friend bool IsComparable (const self_t & m1, const self_t & m2)
  {
    return m1.rep() == m2.rep();
  }

  // Arithmetic operators
  self_t   operator  - () const {return self_t(-_coef,rep());}
  self_t & operator += (const self_t &);
  self_t & operator -= (const self_t &);
  self_t & operator *= (const self_t &);
  self_t & operator *= (const C &);
  self_t & operator /= (const C &);

};
//====================================================================
template<class C, class R>
bool eq (const monomial<C,R>& v1, const monomial<C,R>& v2) {return v1==v2;}
template<class C, class R>
bool neq(const monomial<C,R>& v1, const monomial<C,R>& v2) {return v1!=v2;}

template<class C, class R>
unsigned hash (const monomial<C,R>& v) 
{
  register unsigned i, h= 214365^hash(v.coeff()), n= v.size();
  for (i=0; i<n; i++)
      h= (h<<1) ^ (h<<5) ^ (h>>27) ^ hash(v[i]);
  return h;
}
template<class C, class R>
unsigned soft_hash (const monomial<C,R>& m) {return hash(m);}
//======================================================================
template<class R,class O> Variables monomial<R,O>::var;
//----------------------------------------------------------------------//
// template<class R,class O> unsigned monomial<R,O>::_number_var=1;
// template<class R,class O> std::map<std::string,int> monomial<R,O>::_index_of_var;
// template<class R,class O> std::map<int,std::string> monomial<R,O>::_var_of_index;
//----------------------------------------------------------------------//
namespace let 
{
  template<class C, class R> inline
  void assign( C & r, const monomial<C,R>& m )
  {
    r = m.coeff();
  };
  template<class C, class R> inline
  const C&  convert( const monomial<C,R> & m, const meta::As<C>& ) { return m.coeff(); };
};
//----------------------------------------------------------------------//
#define MONOM  monomial<R,O>
  declare_binary_operator(TMPL,MONOM,MONOM,meta::_mul_,operator*);
  declare_binary_operator(TMPL,MONOM,MONOM,meta::_div_,operator/);
#undef MONOM
//----------------------------------------------------------------------//
template <class C,class R> inline
void mul(monomial<C,R> & a, const monomial<C,R>& m1, const monomial<C,R>& m2)
{
   a.coeff() = (m1.coeff()*m2.coeff());
   if (a.coeff() !=0)
     {
       a.rep().reserve(std::max(m1.rep().size(),m2.rep().size()));
       using namespace VECTOR; add(a.rep(),m1.rep(),m2.rep());
     }
   else
     {
       erase(a.rep());
     }
}
//----------------------------------------------------------------------//
template <class R,class O> inline
void div(monomial<R,O> & a, const monomial<R,O>& m1, const monomial<R,O>& m2)
{
  a.coeff() = (m1.coeff()/m2.coeff());
  if (a.coeff() !=0)
    {
      a.rep().reserve(std::max(m1.rep().size(),m2.rep().size()));
      VECTOR::sub(a.rep(),m1.rep(),m2.rep());
    }
  else
    {
      erase(a.rep());
    }
}
//----------------------------------------------------------------------
template <class C, class R> inline
typename R::degree_t degree(const monomial<C,R> & m)
{
  return degree(m.rep());
}
//----------------------------------------------------------------------//
template<class M>
M MGcd(const M & m1, const M & m2) {
  int d = Gcd(m1.coeff(),m2.coeff());
  M res(d);
  int v = Min(m1.nvars(),m2.nvars());
  for (int i = 0; i <= v; ++i)
    res.set_expt(i, Min(m1[i],m2[i]));
  return res;
}

template <class R,class O> inline
monomial<R,O> div(const monomial<R,O> & m1, const monomial<R,O> & m2) {
  monomial<R,O> res(1); res*=(m1.coeff()/m2.coeff());
  for (int i = 0; i <= m1.nvars(); ++i)
    res.set_expt(i,m1[i] - m2[i]);
  return res;
}

template <class R,class O> inline
monomial<R,O> div_quo(const monomial<R,O> & m1, const monomial<R,O> & m2) {
  monomial<R,O> res(1); res*=(m1.coeff()/m2.coeff());
  for (int i = 0; i <= m1.nvars(); ++i)
    res.set_expt(i,m1[i] - m2[i]);
  return res;
}



template <class R,class O> inline
monomial<R,O> pow(const monomial<R,O>& m1, int n)
{
  monomial<R,O> res(m1);
  for (int i = 0; i <= m1.nvars(); ++i)
    res.set_expt(i,m1[i]*n);
  return res;
}
//--------------------------------------------------------------------
/// Construction of the monomial $x_{i}^{d}$.
template <class C, class R>
monomial<C,R>::monomial(int i, int d):_coef(1)
{
  if(d == 0)
    rep().resize(0);
  else
    {
      rep().reserve(i+1);
      for(int j=0;j<i;j++) rep().set_expt(j,0);
      rep().set_expt(i,d);
    }
}
//----------------------------------------------------------------------
/// Construction of the monomial $x_{i}^{d}$.
template <class C, class R>
monomial<C,R>::monomial(const std::string & s, int d, 
			Variables& vars 		  ):_coef(1)
{
  if(d == 0)
    rep().resize(0);
 else
     {
       int i=vars[s];
       rep().reserve(i+1);
       for(int j=0;j<i;j++) rep().set_expt(j,0);
       rep().set_expt(i,d);
     }
 }

//----------------------------------------------------------------------
/// Construction of the monomial $c\,x_{i}^{d}$.
template <class C, class R>
monomial<C,R>::monomial(const C & c, int i, int d ):_coef(c)
{
  if(d == 0)
    rep().resize(0);
  else
    {
      rep().reserve(i+1);
      for(int j=0;j<i;j++) rep().set_expt(j,0);
      rep().set_expt(i,d);
    }
}
//----------------------------------------------------------------------
extern "C" char *synaps_inputptr;
template <class C, class R>
monomial<C,R>::monomial(const char * s, Variables&  vars )
{
  int n = strlen(s);
  if (s[n-1]=='\n') {n--;}
  if(s[n-1]==';')
    {
      synaps_input = new char[n+1];
      memcpy(synaps_input,s,n); synaps_input[n]='\0';
      synaps_inputlim = synaps_input + n;
    }
  else
    {
      synaps_input = new char[n+2];
      memcpy(synaps_input,s,n); 
      synaps_input[n]=';'; synaps_input[n+1]='\0';
      synaps_inputlim = synaps_input + n+1;
    }
  synaps_inputptr=synaps_input;
  *this = monomial<C,R>(1);
  int Ask;

  for (;;) {
    Ask = yylex();
    if (Ask == COEF) { /* monomiale constant */
      C c;
      let::assign(c,yylval);
      free(yylval);
      *this *= c;
    }
    else if (Ask == XID) {
      char* p = strchr(yylval,'^');
      if (p == NULL)
	{
	  *this *= monomial<C,R>(vars[std::string(yylval)],1);
	}
      else {
	p++;
	*this *= monomial<C,R>(vars[std::string(yylval,p-1)], atoi(p));
      }
      free(yylval);
    }
    else if (Ask == TERMINATE) {
      break;
    }
  }
  delete[] synaps_input;
}
//----------------------------------------------------------------------
template <class C, class R> inline
bool monomial<C,R>::operator== (int n) const
{
  return (coeff()==n);
}

//----------------------------------------------------------------------
// arithemtic operators
//----------------------------------------------------------------------
template <class C, class R> inline
monomial<C,R> & monomial<C,R>::operator += (const monomial<C,R> & m)
{
  coeff() += m.coeff();
  if (coeff()==0) erase(rep());
  return *this;
}

template <class C, class R> inline
monomial<C,R> & monomial<C,R>::operator -= (const monomial<C,R> & m)
{
  coeff() -= m.coeff();
  if (coeff()==0) erase(rep());
  return *this;
}

template <class C, class R> inline
monomial<C,R>  operator +  (const monomial<C,R> & m1, const monomial<C,R> & m2)
{
  monomial<C,R> res(m1);
  res.coeff() += m2.coeff();
  return (*res);
}

template <class C, class R> inline
monomial<C,R> operator - (const monomial<C,R> & m1, const monomial<C,R> & m2)
{
        monomial<C,R> res(m1);
	res.coeff() -= m2.coeff();
	return (res);
}

template <class C, class R> inline
monomial<C,R> & monomial<C,R>::operator *= (const monomial<C,R> & m)
{
  using namespace VECTOR;
  coeff() *= m.coeff();
  if (coeff()!=0)
    add(rep(),m.rep());
  else
    erase(rep());
  return *this;
}

template <class C, class R> inline
monomial<C,R> & monomial<C,R>::operator *= (const C & c)
{
  coeff() *= c;
  if (coeff()==0) erase(rep());
  return *this;
}

template <class C, class R> inline
monomial<C,R> & monomial<C,R>::operator /= (const C & c)
{
  assert(c !=0);
  coeff() /= c;
  if (coeff()==0) erase(rep());
  return *this;
}

template <class C, class R> inline
monomial<C,R> operator * (const monomial<C,R> & m1, const C & c2)
{
  monomial<C,R> res(m1);
  res.coeff() *= c2;
  if (res.coeff() ==0) erase(res.rep());
  return (res);
}
template <class C, class R> inline
monomial<C,R> operator * (const C & c2, const monomial<C,R> & m1)
{
  return(m1*c2);
}

template <class C, class R> inline
monomial<C,R> operator / (const monomial<C,R> & m1, const C & c2)
{
  monomial<C,R> res(m1);
  res.coeff() /= c2;
  if (res.coeff() ==0) erase(res.rep());
  return (res);
}


//----------------------------------------------------------------------
// operateurs d'entrees/sorties
//----------------------------------------------------------------------
/// Output operator.
template <class C, class R> inline
std::ostream & operator << (std::ostream & os, const monomial <C,R> & m)
{
  if (m.coeff() ==0 )    os<<"(0)";
  else {
    MONOMIAL::print(os,m);
  }
  return os;
}

//--------------------------------------------------------------------
// namespace MONOMIAL 
// {

template<class C, class R> inline
C monomial_convert(const monomial<C,R>& m, unsigned i, meta::As<C>)
{
  return m.coeff();
}

template<class c1, class R1> struct upoldse;
template<class C, class R1, class R2> inline
upoldse<C,R1>  monomial_convert(const monomial<C,R2>& m, unsigned i, meta::As< upoldse<C,R1> > mth)
{
  return upoldse<C,R1>(m.coeff(),m[i]);
}
    
template<class C1, class O1, class R1> struct mpoldst;
template<class C, class O1, class R1, class R2> inline
mpoldst<C,O1,R1>  monomial_convert(const monomial<C,R2>& m, unsigned i, meta::As<mpoldst<C,O1,R1> > mth)
{
  monomial<C,R2> m1(m);
  if (i<m.nbvar())
    m1.set_expt(i,0);
  return mpoldst<C,O1,R1>(m1);
}
//}

//======================================================================
} //  namespace mmx 
//======================================================================
#undef TMPL
#undef Monomial
#endif // SYNAPS_MONOM_H

