#ifndef SYNAPS_MPOLDSE_VIEW_H
#define SYNAPS_MPOLDSE_VIEW_H

#include <subdivix/shared_object.hpp>
#include <subdivix/mpoldst.hpp>
#include <subdivix/multivariate_dense_monomial.hpp>
#include <subdivix/multivariate_dense_bernstein.hpp>
#include <subdivix/xml.hpp>

namespace mmx {
/**
   Interface for multivariate dense polynomials. The type @R@ is the
   representation of the polynomial (eg. in monomial or bernstein basis). 
   The type @C@ is the coefficient type of the polynomial.
**/
template< class C, class R = multivariate_dense::bernstein<C> > 
struct mpoldse //: arithm::overloaded
{
  typedef C                               value_type;
  typedef C                               coeff_t;
  typedef typename R::size_type           size_type;
  typedef R                               rep_type;
  typedef typename R::iterator            iterator;
  typedef typename R::const_iterator      const_iterator;
  typedef mpoldse<C,R>                    self_t;
  shared_object<R> data;
  R &       rep()       { return (*data);}
  const R & rep() const { return (*data);}
  mpoldse( int nvr, const int * szs, const int * vrs );
  mpoldse( const C& x = 0 );
  template<class O, class E> 
  mpoldse( const mpoldst<C,O,E>& mpol );
  mpoldse( const char * s ) ;
  template<class X> 
  mpoldse( const X& x );
  
  iterator               begin()        {return rep().begin();}
  const_iterator         begin()  const {return rep().begin();}
  iterator               end()          {return rep().end();}
  const_iterator         end()    const {return rep().end();}
  
  bool operator==( const C& c ) const;
  bool operator==( const mpoldse& mp ) const;
  
  mpoldse & operator=( const char * name);
  template<class X>
  mpoldse & operator= ( const X & x );

  template<class O, class E>
  mpoldse& operator = ( const mpoldst<C,O,E>& mpol );
  template<class X>
  mpoldse& operator+= ( const X & x ) 
  { 
    *this = *this + x; return *this; 
  } ;
  
  mpoldse& operator+= ( const mpoldse& );
  mpoldse& operator+= ( const C& );
  template<class X>
  mpoldse& operator-= ( const X & x ) { *this = *this - x; return *this; };
  mpoldse& operator-= ( const mpoldse& );
  mpoldse& operator-= ( const C& );
  template<class X>
  mpoldse& operator*= ( const X & x ) { *this = *this * x; return *this; };
  mpoldse& operator*= ( const C& );  
  mpoldse& operator*= ( const mpoldse& );
  const int * vrs() const { return rep().env.vrs(); };
  const int * szs() const { return rep().env.szs(); };
  unsigned size() const { return rep().data.size(); }
  int nvars() const;
  int nbvar() const { return nvars(); }
  int nvr() const { return nvars(); };
  int varid( int i ) const;
  const C& operator[]( int i ) const;
  C& operator[]( int i ) ;
  //  std::string varname( int i ) const ;
  //  void lexsort();
};

template<class C, class R>
int nbvar( const mpoldse<C,R>& mp ) { return mp.nvr(); };


template<class C, class R>
int degree( const mpoldse<C,R>& mp, int v ) { return mp.rep().env.szs()[v]; };

template<class C, class R> 
std::ostream& operator<<( std::ostream& o, const mpoldse<C,R>& mp );



template<class C, class R>
mpoldse<C,R> operator^( const mpoldse<C,R>& a, unsigned n );


template<class C, class R> 
bool operator==( const C& c, const mpoldse<C,R>& mp );

template<class C, class R> 
bool operator !=( const mpoldse<C,R>& p, const mpoldse<C,R>& q ) {return !(p==q);}

/** 
    Multivariate Polynomial Differentiation.
    Compute the partial derivative of @polynom@ w.r.t @v@ the vth variable of @polynom@.
    The result is stored in @derivative@.
 **/
template<class Coeff, class Rep>
mpoldse<Coeff,Rep> diff(const mpoldse<Coeff,Rep>& polynom, int v );

/**
   Multivariate Polynomial Evaluation.
   The type {\tt Prm } must a container type, @parameters@ corresponds to the $0..nvr()$ coordinates.
   The computation is done in the {\tt Sup Type} of the {\tt Contained Type } of {\tt Prm} and {\tt Coeff}.
   The type {\tt Result } is the type of the result, it is assumed to have an operator= on the computation type.
**/
template<class Result, class Coeff, class Rep, class Parameters>
Result eval(const mpoldse<Coeff,Rep>& polynom, const Parameters& parameters ) ;

template<class It>
void mpoldseSystem( It b, It e )
{
  multivariate_dense::eenv tmp(b->rep().env);
  It i = b;
  i++;
  for ( ; i != e; i ++ )
    tmp = multivariate_dense::eenv::mrgvrs(tmp,i->rep().env);
  for ( i = b; i != e; i ++ )
    i->rep().env = multivariate_dense::eenv::rewrite(tmp,i->rep().env);
};

namespace xml {

  /// Read a univariate polynomial, enclosed in @polynomial@ tags.
  template<class ISTREAM, typename C,typename R>
  bool read(ISTREAM& is, mpoldse<C,R>& p)
  {
    if(skip(is,begin_polynomial))
      {
        std::string s; char c;
        while((c=is.get()) != EOF && c != '<')
          {
            if(c != ' ' && c != '\n') s+=c;
          }
        s+='\0';
        p = mpoldse<C,R>(s.data());
        return true;
      }
    else
      {
        p=mpoldse<C,R>(0); return false;
      }
  }


}; //namespace xml

} //namespace mmx
//====================================================================
#include <subdivix/mpoldse_c.hpp>
//====================================================================
#endif
