/*********************************************************************
*      This file is part of the source code of the SYNAPS library.
*      Author(s): B. Mourrain, GALAAD, INRIA                 
**********************************************************************
History:
    03/03/00 by bm, update for the new version
    02/03/99 by dbondy,
		  - modification of the operator *=. "p*=p;" fonctionne.
    24/02/99 by dbondy,
        - add a constructor for the mpoldst class form a char *.
        - add a typedef for the arithmetic coefficient.
    12/12/98 by bm, change p in rep, add size method.
    1/12/98  by dbondy modified the structure of the file for a better
    documentation.
$Id: mpoldst.H,v 1.1 2005/07/11 11:17:56 mourrain Exp $
**********************************************************************/
#ifndef SYNAPS_MPOL_MPOL_H
#define SYNAPS_MPOL_MPOL_H

#include <stdlib.h>
#include <string.h>
#include <string>
#include <list>

#include <subdivix/shared_object.hpp>
#include <subdivix/parser.hpp>
#include <subdivix/let.hpp>
#include <subdivix/Variables.hpp>
#include <subdivix/dynamicexp.hpp>
#include <subdivix/monomial.hpp>
#include <subdivix/DegLex.hpp>
#include <subdivix/MPOLDST_m.hpp>

namespace mmx {
//======================================================================
#define FORALL template <class C, class O, class R>
#define Polynomial mpoldst<C,O,R>

namespace linalg { template<class C> struct rep1d; }

//----------------------------------------------------------------------
/// Multivariate polynomials.
/**
   Interface for multivariate polynomials. The type @R@ is the
   representation of the polynomial (eg. a list of monomial). The type
   @O@ is the class specifying the order on the monomials.
**/
template <class C,
          class O=DegLex,
          class R=std::list<monomial<C,dynamicexp<unsigned> > > >
struct mpoldst {

  shared_object<R>  data;
  R &       rep()       {return (*data);}
  const R & rep() const {return (*data);}

  // Types
  typedef typename R::value_type             monom_t;
  typedef C                                  coeff_t;
  typedef O                                  order_t;
  typedef typename R::iterator               iterator;
  typedef typename R::const_iterator         const_iterator;
  typedef typename R::reverse_iterator       reverse_iterator;
  typedef typename R::const_reverse_iterator const_reverse_iterator;
  typedef mpoldst<C,O,R>                        self_t;

  // Constructors
  mpoldst():data() {rep().resize(0);}
  mpoldst(const monom_t & m): data()
    {
      rep().resize(0); rep().insert(rep().begin(),m);
    }

  mpoldst(const coeff_t & c):data()
    {
      if(c !=0) {
	rep().resize(0); rep().insert(rep().begin(),monom_t(c));
      }else
	rep().resize(0);
    }

  mpoldst(const int & i): data()
    {
      if(i !=0)
	{ rep().resize(0);rep().insert(rep().begin(),monom_t(i));}
      else
	{ rep().resize(0);}
    }

  mpoldst(int, const monom_t *);
  mpoldst(const self_t & P): data(P.data) {}
  mpoldst(const char *, Variables& vars =  monom_t::var);
  mpoldst(const char *, const Variables& vars);

//   template<class S>
//   mpoldst(const VAL<S>& M) {using let::assign; assign(*this,M.rep());}


  template<class Q>
  mpoldst(const mpoldst<C,Q,R>& p) : data()
  {
    for(const_iterator it=p.begin();it != p.end(); it++)
      (*this) += *it;
  }
  
  // Destructor
  // no dynamic allocation at this level.

  //  static Variable var;
  static int index(const std::string & s)     { return monom_t::var[s]; }
  static std::string var(int i)               { return monom_t::var[i]; }
  static void var(int i, const std::string& s){ monom_t::var.let(i,s); }
  static int  var(char* s)                    { monom_t::var=s; return monom_t::var[s]; }


  // Assignment operators
  self_t & operator =(const self_t & x)  {data=x.data; return *this;}
  self_t & operator =(const coeff_t & c) {*this=self_t(c); return *this;}
  self_t & operator =(int n)             {*this=self_t(n); return *this;}
  self_t & operator =(const monom_t & m) {*this=self_t(m); return *this;}

//   template<class S>
//   self_t & operator =(const VAL<S> & M)  {assign(*this,M.rep());return *this;}

  self_t & operator*=(const coeff_t &);
  self_t & operator/=(const coeff_t &);

  self_t & operator*=(const monom_t &);
  self_t & operator/=(const monom_t &);
  //  self_t & operator+=( const monom_t & m );
  
  self_t & operator+=(const self_t &);
  self_t & operator-=(const self_t &);
  self_t & operator*=(const self_t &);
  self_t & operator/=(const self_t & Q);

  // Boolean operators and functions
  bool operator==(int n) const;
  bool operator!=(int n) const {return !(*this==n);}

  // Iterator functions
  iterator                begin ()       {return rep().begin(); }
  const_iterator          begin () const {return rep().begin(); }
  reverse_iterator       rbegin ()       {return rep().rbegin();}
  const_reverse_iterator rbegin () const {return rep().rbegin();}

  iterator                  end ()       {return rep().end(); }
  const_iterator            end () const {return rep().end(); }
  reverse_iterator         rend ()       {return rep().rend();}
  const_reverse_iterator   rend () const {return rep().rend();}

  /// Return the number of terms in the polynomial.
  int size()  const {return rep().size();}
  /// Return the number of variables involved in the polynomial.
  int nbvar() const {return MPOLDST::nbvar(*this);}

  /// Evaluate the polynomial at the "point" p.
  template<class T>  coeff_t operator()(const T & p) const;

};

//--------------------------------------------------------------------
template<class C, class O, class R>
int nbvars(const mpoldst<C,O,R>& p) { return MPOLDST::nbvar(p); }
//======================================================================
// template<class C, class O, class R> Variable mpoldst<C,O,R>::var;
//======================================================================
// Input/Output operator
// template<class C, class O, class R>
// std::istream & operator >>(std::istream &, mpoldst<C,O,R> &);
template<class C, class O, class R>
std::ostream & operator <<(std::ostream & os, const mpoldst<C,O,R>& p)
{
  MPOLDST::print(os,p); return os;
}
// template<class C, class O, class R>
// bool operator == (const mpoldst<C,O,R> &, const mpoldst<C,O,R> &);
// template<class C, class O, class R>
// bool operator != (const mpoldst<C,O,R> &, const mpoldst<C,O,R> &);
// template<class C, class O, class R>
// bool operator < (const mpoldst<C,O,R> &, const mpoldst<C,O,R> &);
// template<class C, class O, class R>
// bool operator <= (const mpoldst<C,O,R> &, const mpoldst<C,O,R> &);
// template<class C, class O, class R>
// bool operator > (const mpoldst<C,O,R> &, const mpoldst<C,O,R> &);
// template<class C, class O, class R>
// bool operator >= (const mpoldst<C,O,R> &, const mpoldst<C,O,R> &);
//======================================================================
// ARITHMETIC OPERATIONS.
//======================================================================
#ifdef WITH_TEMPLATE_EXP
	#include <synaps/mpoldst_with_texp.hpp>
#else
//======================================================================
template <class C, class O, class R>
mpoldst<C,O,R> operator+(const mpoldst<C,O,R> & a, const mpoldst<C,O,R> & b);
//----------------------------------------------------------------------
template <class C, class O, class R>
mpoldst<C,O,R> operator+(const mpoldst<C,O,R> & a,
		      const typename mpoldst<C,O,R>::coeff_t & c);
//----------------------------------------------------------------------
template <class C, class O, class R>
mpoldst<C,O,R> operator+(const mpoldst<C,O,R> & a,
		      const typename mpoldst<C,O,R>::monom_t & c);
//----------------------------------------------------------------------
template <class C, class O, class R>
mpoldst<C,O,R> operator+(const typename mpoldst<C,O,R>::coeff_t & c,const mpoldst<C,O,R> & a);
//----------------------------------------------------------------------
template <class C, class O, class R>
mpoldst<C,O,R> operator-(const mpoldst<C,O,R> & a, const mpoldst<C,O,R> & b);
//----------------------------------------------------------------------
template <class C, class O, class R> inline
mpoldst<C,O,R> operator-(const mpoldst<C,O,R> & a,const typename mpoldst<C,O,R>::coeff_t & c);
//----------------------------------------------------------------------
template <class C, class O, class R> inline
mpoldst<C,O,R> operator-(const typename mpoldst<C,O,R>::coeff_t & c,const mpoldst<C,O,R> & a);
//----------------------------------------------------------------------
template <class C, class O, class R> inline
mpoldst<C,O,R> operator-(const mpoldst<C,O,R> & a);
//----------------------------------------------------------------------
template <class C, class O, class R>
mpoldst<C,O,R> operator*(const mpoldst<C,O,R> & a, const mpoldst<C,O,R> & b);
//----------------------------------------------------------------------
template <class C, class O, class R>
mpoldst<C,O,R> operator*(const mpoldst<C,O,R> & a, const C & c);
//----------------------------------------------------------------------
template <class C, class O, class R> inline
mpoldst<C,O,R> operator*(const typename mpoldst<C,O,R>::coeff_t & c,
		      const mpoldst<C,O,R> & a);
//----------------------------------------------------------------------
template <class C, class O, class R> inline
mpoldst<C,O,R> operator*(int c,const mpoldst<C,O,R> & a);
//----------------------------------------------------------------------
template <class C, class O, class R> inline
mpoldst<C,O,R> operator*(const mpoldst<C,O,R> & a,
		      const typename mpoldst<C,O,R>::monom_t & m);
//----------------------------------------------------------------------
template <class C, class O, class R> inline
mpoldst<C,O,R> operator*(const typename mpoldst<C,O,R>::monom_t & m,
		      const mpoldst<C,O,R> & a);
//----------------------------------------------------------------------
template <class C, class O, class R> inline
mpoldst<C,O,R> operator/(const mpoldst<C,O,R> & a, const mpoldst<C,O,R> & b) ;
//----------------------------------------------------------------------
template <class C, class O, class R> inline
mpoldst<C,O,R> operator%(const mpoldst<C,O,R> & a, const mpoldst<C,O,R> & b) ;
//----------------------------------------------------------------------
template <class C, class O, class R>
mpoldst<C,O,R> operator^(const mpoldst<C,O,R> & a, unsigned int n);
#endif
//----------------------------------------------------------------------
/// Return the number of variables of @p@.
template <class C, class O, class R>
int NbVar(const mpoldst<C,O,R> & p)
{
  return MPOLDST::nbvar(p);
}

/// Return the number of variables of @p@.
template <class C, class O, class R>
int LastVar(const mpoldst<C,O,R> & p)
{
  return MPOLDST::lvar(p);
}


//====================================================================
/// Total degree of the polynomial @p@.
template <class C, class O, class R> inline
int degree (const mpoldst<C,O,R> & p)
{
  return MPOLDST::degree(p.rep());
}
/// Degree of the polynomial @p@ in the variable of index @i@.
template <class C, class O, class R> inline
int degree (const mpoldst<C,O,R> & p, unsigned int i)
{
  return MPOLDST::degree(p.rep(),i);
}
//====================================================================
template <class C, class O, class R> inline
unsigned size(const mpoldst<C,O,R> & p) {return p.size();}
//====================================================================
/// Derivative of @p@ with respect to the variable of index @i@.
template <class C, class O, class R> inline
mpoldst<C,O,R> diff(const mpoldst<C,O,R> & p, unsigned i)
{
  mpoldst<C,O,R> r;
  MPOLDST::diff(r,p,i);
  return r;
}
//----------------------------------------------------------------------
/** Derivative of @p@ with respect to the variable named @x@. It uses
the map @monom_t::index_of_var@, to get the index of the variable.
**/
template <class C, class O, class R> inline
mpoldst<C,O,R> diff(const mpoldst<C,O,R> & p, char* x, 
		    Variables& vars = mpoldst<C,O,R>::monom_t::var)
{
  typedef typename mpoldst<C,O,R>::monom_t monom_t;
  int i = vars[x];
  if(i<0)
    return mpoldst<C,O,R>(0);
  else
    {
      mpoldst<C,O,R> r;
      MPOLDST::diff(r,p,i);
      return r;
    }
}
//====================================================================
template <class C, class O, class R> inline
mpoldst<C,O,R> square(const mpoldst<C,O,R> & p) {return p*p;}
//====================================================================
namespace UPOLDSE 
{
  template <class C, class O, class R>
  std::ostream & print_as_coeff(std::ostream & os, const mpoldst<C,O,R> & p)
  {
    return os<<"+("<<p<<")";
  }
}
//----------------------------------------------------------------------
// Constructors
//----------------------------------------------------------------------
template <class C,class O, class R>
inline mpoldst<C,O,R>::mpoldst (int s, const monom_t *t)
{
	for (int i = 0; i<s; i++) *this += t[i];
}

extern "C" char *synaps_inputptr;
template <class C,class O, class R>
inline mpoldst<C,O,R>::mpoldst (const char * s, Variables&  vars ): data() {
  rep().resize(0);
  int n = strlen(s);
  if (s[n-1]=='\n') {n--;}
  if(s[n-1]==';')
    {
      synaps_input = new char[n+1];
      memcpy(synaps_input,s,n); synaps_input[n]='\0';
      synaps_inputlim = synaps_input + n;
    }
  else
    {
      synaps_input = new char[n+2];
      memcpy(synaps_input,s,n); 
      synaps_input[n]=';'; synaps_input[n+1]='\0';
      synaps_inputlim = synaps_input + n+1;
    }
  synaps_inputptr=synaps_input;
  monom_t m(1);
  bool sign = true;
  bool first = true;
  int Ask;

  for (;;) {
    Ask = yylex();
    if (Ask == COEF) { /* monomiale constant */
      coeff_t c;
      let::assign(c,yylval);
      free(yylval);
      m *= c;
    }
    else if (Ask == XID) {
      char* p = strchr(yylval,'^');
      if (p == NULL)
	{
	  m *= monom_t(vars[std::string(yylval)],1);
	}
      else {
	p++;
	m *= monom_t(vars[std::string(yylval,p-1)], atoi(p));
      }
      free(yylval);
    }
    else if (Ask == MULT);
    else if (Ask == SUM) {
      if (sign) {
	if(m!=0) *this += m;
	m = monom_t(1);
      } else {
	if(m!=0) *this -= m;
	sign = true;
	m = monom_t(1);
      }
    }
    else if (Ask == MINUS) {
      if (first)
	sign = false;
      else if (sign) {
	if(m!=0) *this += m;
	sign = false;
	m = monom_t(1);
      }
      else {
	if(m!=0) *this -= m;
	m = monom_t(1);
      }
    }
    else if (Ask == TERMINATE) {
      if (sign){
	if(m!=0) *this += m;
      }else{
	if(m!=0) *this -= m;
      }
      break;
    }
    first = false;
  }
  delete[] synaps_input;
}
//----------------------------------------------------------------------
template <class C,class O, class R>
  inline mpoldst<C,O,R>::mpoldst (const char * s, const Variables&  vars )
{
  Variables V(vars);
  *this= mpoldst<C,O,R>(s,V);
}
//----------------------------------------------------------------------
// template <class C,class O, class R>
// inline mpoldst<C,O,R> & mpoldst<C,O,R>::operator += (const typename mpoldst<C,O,R>::monom_t & m){
//   if (m.coeff()!=0)
//     MPOLDST::insert<C,O,R>(rep(), rep().begin(), m);
//   return *this;
// }
//----------------------------------------------------------------------
template <class C,class O, class R>
inline mpoldst<C,O,R> & mpoldst<C,O,R>::operator+= (const mpoldst<C,O,R> & P)
{
  if (P==0) return *this;
  if (*this==0)
    {*this = P; return *this;}
  else
    MPOLDST::add<O,R>(rep(),P.rep());
  return *this;
}
//----------------------------------------------------------------------
template <class C,class O, class R>
inline mpoldst<C,O,R> & mpoldst<C,O,R>::operator-= (const mpoldst<C,O,R> & P){
  if (P==0) return *this;
  if (*this==0) {*this = -P; return *this;}
  *this += (-P);
  return *this;
}
//----------------------------------------------------------------------
template <class C,class O, class R>
inline mpoldst<C,O,R> & mpoldst<C,O,R>::operator*= (const typename mpoldst<C,O,R>::coeff_t & c){
  mpoldst<C,O,R> P;
  for (iterator i = begin(); i != end(); i++) {
    (*i)*= c;
  }
  return *this;
}
//----------------------------------------------------------------------
template <class C,class O, class R>
inline mpoldst<C,O,R> & mpoldst<C,O,R>::operator/= (const typename mpoldst<C,O,R>::coeff_t & c){
  mpoldst<C,O,R> P;
  for (iterator i = begin(); i != end(); i++) { (*i)/= c;
  }
  return *this;
}
//----------------------------------------------------------------------
template <class C,class O, class R>
inline mpoldst<C,O,R> & mpoldst<C,O,R>::operator/= (const typename mpoldst<C,O,R>::monom_t & m){

  if ( (*this!=0)  && (m.coeff() !=0) ) {
    mpoldst<C,O,R> P(*this);
    for (iterator i = P.begin(); i != P.end(); i++) {
      int n=m.nvars();
      if (n<=(*i).nvars())
	while (((*i)[n] >= m[n] ) && ( n >= 0)) n--;
      if (n<0)
	*i=div(*i,m);
      else return *this;
    }
    (rep()=P.rep());
  }
  return *this;
}
//----------------------------------------------------------------------
template <class C,class O, class R>
inline mpoldst<C,O,R> & mpoldst<C,O,R>::operator*= (const typename mpoldst<C,O,R>::monom_t & m){
 if(m !=0)
   {
     mpoldst<C,O,R> P;
     for (iterator i = begin(); i != end(); i++) {
       (*i)*= m;
     }
   }
 else
   *this=0;
 return *this;
}
//----------------------------------------------------------------------
template <class C,class O, class R>
inline  mpoldst<C,O,R> & mpoldst<C,O,R>::operator*= (const mpoldst<C,O,R> & b)
{
  assert(&rep() != &b.rep());
  if(*this !=0) {
    if(b!=0) {
      const self_t a(*this);
      *this=0;
      MPOLDST::mul(rep(),a.rep(),b.rep(),O());
    } else {
      (*this)= self_t(0);
    }
  }
  return *this;
}
//----------------------------------------------------------------------
template <class C,class O, class R>
inline  mpoldst<C,O,R> & mpoldst<C,O,R>::operator/= (const mpoldst<C,O,R> & Q){
  assert(Q != 0);
  mpoldst<C,O,R> P(*this), quo;
  MPOLDST::div_rem_x(quo,P,Q);
  *this=quo;
  return *this;
}
//----------------------------------------------------------------------
template <class C,class O, class R>
template<class T>
typename mpoldst<C,O,R>::coeff_t
mpoldst<C,O,R>::operator()(const T & p) const
{
  coeff_t r(0);
  for(const_iterator it=begin(); it!=end(); ++it)
    {
      coeff_t s(it->coeff());
      for(unsigned int i=0;i<(*it).rep().size();++i)
	s*=pow(p[i],(int)(*it)[i]);
      r+=s;
    }
  return r;
}
//----------------------------------------------------------------------
// Boolean operators and functions
//----------------------------------------------------------------------
template <class C,class O, class R> inline
bool mpoldst<C,O,R>::operator ==(int n) const
{
  if(n==0)
    if (rep().size()==0)
      return true;
    else
      return false;
  else
    if (rep().size() ==0)
      return false;
    else
      return(rep().size()==1 && (rep().begin()->coeff() ==n));
 ///
 /* if ( n==0 && rep().size() == 0 ) 
    return true;
  if (rep().size() == 0 )   return false;
  return(rep().size()==1 && (rep().begin()->coeff() ==n));
 */
}
//----------------------------------------------------------------------
template <class C,class O, class R>
bool operator== (const mpoldst<C,O,R> & P, const mpoldst<C,O,R> & Q)
{
	return P.rep() == Q.rep();
}
//----------------------------------------------------------------------
template <class C,class O, class R>
bool operator!= (const mpoldst<C,O,R> & P, const mpoldst<C,O,R> & Q)
{
	return !(P.rep() == Q.rep());
}
//----------------------------------------------------------------------
template <class C,class O, class R>
bool operator< (const mpoldst<C,O,R> & P, const mpoldst<C,O,R> & Q)
{
	typename mpoldst<C,O,R>::const_iterator iP = P.begin();
	typename mpoldst<C,O,R>::const_iterator iQ = Q.begin();
	while (iP != P.end() && iQ != Q.end() && IsComparable(*iP,*iQ)) {
		++iP;
		++iQ;
	}
	if (iQ == Q.end())
		return false;
	if (iP == P.end())
		return true;
	return O::less(*iP,*iQ);
}
//----------------------------------------------------------------------
template <class C,class O, class R>
bool operator> (const mpoldst<C,O,R> & P, const mpoldst<C,O,R> & Q){
	return Q < P;
}
//----------------------------------------------------------------------
template <class C,class O, class R>
bool operator>= (const mpoldst<C,O,R> & P, const mpoldst<C,O,R> & Q){
	return !(P < Q);
}
//----------------------------------------------------------------------
template <class C,class O, class R>
bool operator<= (const mpoldst<C,O,R> & P, const mpoldst<C,O,R> & Q){
	return !(Q < P);
}
//======================================================================
// Input/putput operators
//======================================================================
/// Input operator.
template <class C,class O, class R>
std::istream & operator>> (std::istream & is, mpoldst<C,O,R> & P)
{
	std::string s,ss = "";
	int p;
	for (;;) {
		is >> s;
		p = s.find(';');
		if (p== -1) s.find(':');
		if (p== -1) s.find(',');
		if (p== -1)
			ss += s;
		else {
			ss += s.substr(0,p+1);
			break;
		}
	}
	P = mpoldst<C,O,R>(ss.c_str());
	return is;
}
//----------------------------------------------------------------------
/// Output operator with variable environment.
template <class C, class O, class R>
std::ostream & print (std::ostream & os, const mpoldst<C,O,R> & P, 
		      const Variables& V= mpoldst<C,O,R>::monom_t::var)
{
  MPOLDST::print(os,P,V); return os;
}


//----------------------------------------------------------------------
/// Sum of polynomials
template <class C, class O, class R>
mpoldst<C,O,R> operator+(const mpoldst<C,O,R> & a, const mpoldst<C,O,R> & b)
{
  mpoldst<C,O,R> p;
  if(&p == &a && &p == &b)
    MPOLDST::add<O,R>(p.rep(),mpoldst<C,O,R>(b).rep());
  else if(&p == &a)
    MPOLDST::add<O,R>(p.rep(),b.rep());
  else if(&p == &b)
    MPOLDST::add<O,R>(p.rep(),a.rep());
  else
    MPOLDST::add<O,R>(p.rep(),a.rep(),b.rep());
  return p;
}
//----------------------------------------------------------------------
/// Sum of a polynomial and a scalar.
template <class C, class O, class R>
mpoldst<C,O,R> operator+(const mpoldst<C,O,R> & a,
		      const typename mpoldst<C,O,R>::coeff_t & c)
{
  return a + mpoldst<C,O,R>(c);
}
//--------------------------------------------------------------------
/// Sum of a polynomial and a monom.
template <class C, class O, class R>
mpoldst<C,O,R> operator+(const mpoldst<C,O,R> & a,
		      const typename mpoldst<C,O,R>::monom_t & m)
{
  return a + mpoldst<C,O,R>(m);
}
//----------------------------------------------------------------------
/// Sum of a scalar and a polynomial.
template <class C, class O, class R>
mpoldst<C,O,R> operator+(const typename mpoldst<C,O,R>::coeff_t & c,
		      const mpoldst<C,O,R> & a)
{
  return a + mpoldst<C,O,R>(c);
}
//----------------------------------------------------------------------
// operator -
template <class C, class O, class R>
mpoldst<C,O,R> operator-(const mpoldst<C,O,R> & a, const mpoldst<C,O,R> & b)
{
  return (a + (-b));
}
template <class C, class O, class R> inline
mpoldst<C,O,R> operator-(const mpoldst<C,O,R> & a,
		      const typename mpoldst<C,O,R>::coeff_t & c)
{
  return a - mpoldst<C,O,R>(c);
}
template <class C, class O, class R> inline
mpoldst<C,O,R> operator-(const typename mpoldst<C,O,R>::coeff_t & c,
		      const mpoldst<C,O,R> & a)
{
  return mpoldst<C,O,R>(c) - a;
}
//--------------------------------------------------------------------
template <class C, class O, class R> inline
mpoldst<C,O,R> operator-(const mpoldst<C,O,R> & a,
		      const typename mpoldst<C,O,R>::monom_t & m)
{
  return a - mpoldst<C,O,R>(m);
}
template <class C, class O, class R> inline
mpoldst<C,O,R> operator-(const typename mpoldst<C,O,R>::monom_t & m,
		      const mpoldst<C,O,R> & a)
{
  return mpoldst<C,O,R>(m) - a;
}
//----------------------------------------------------------------------
// Unary operator -
template <class C, class O, class R> inline
mpoldst<C,O,R> operator-(const mpoldst<C,O,R> & a)
{
  return a*((typename mpoldst<C,O,R>::coeff_t)-1);
}
//----------------------------------------------------------------------
// Product of polynomials
template <class C, class O, class R>
mpoldst<C,O,R> operator*(const mpoldst<C,O,R> & a, const mpoldst<C,O,R> & b)
{
  mpoldst<C,O,R> p;
  if(a !=0 && b !=0)
    MPOLDST::mul(p.rep(),a.rep(),b.rep(),O());
  else
    p= mpoldst<C,O,R>(0);
  return p;
}
//----------------------------------------------------------------------
// Product of a polynomial by a "scalar"
template <class C, class O, class R>
mpoldst<C,O,R> operator*(const mpoldst<C,O,R> & a, const C & c)
{
  mpoldst<C,O,R> p;
  if(a!=0 && c!=0)
    MPOLDST::mul_ext(p.rep(),a.rep(),c);
 else
    p= mpoldst<C,O,R>(0);
  return p;
}
//----------------------------------------------------------------------
template <class C, class O, class R> inline
mpoldst<C,O,R> operator*(const typename mpoldst<C,O,R>::coeff_t & c,
		     const mpoldst<C,O,R> & a)
{
  return a*c;
}
//----------------------------------------------------------------------
template <class C, class O, class R> inline
mpoldst<C,O,R> operator*(int c,const mpoldst<C,O,R> & a)
{
  typedef typename mpoldst<C,O,R>::coeff_t coeff_t;
  return a*coeff_t(c);
}
//----------------------------------------------------------------------
template <class C, class O, class R> inline
mpoldst<C,O,R> operator*(const mpoldst<C,O,R> & a,
		      const typename mpoldst<C,O,R>::monom_t & m)
{
  mpoldst<C,O,R> p;
  if(a!=0)
    MPOLDST::mul_ext(p.rep(),a.rep(),m);
  else
    p= mpoldst<C,O,R>(0);
  return p;
}
//----------------------------------------------------------------------
template <class C, class O, class R> inline
mpoldst<C,O,R> operator*(const typename mpoldst<C,O,R>::monom_t & m,
		     const mpoldst<C,O,R> & a)
{
  return a*m;
}
//----------------------------------------------------------------------
template <class C, class O, class R> inline
mpoldst<C,O,R> operator/(const mpoldst<C,O,R> & a, const mpoldst<C,O,R> & b)
{
  assert(b != 0);
  mpoldst<C,O,R> r(a), q;
  MPOLDST::div_rem_x(q,r,b);
  return q;
}
//----------------------------------------------------------------------
// Division of a polynomial by a "scalar"
template <class C, class O, class R>
mpoldst<C,O,R> operator/(const mpoldst<C,O,R> & a, const C & c)
{
  mpoldst<C,O,R> p;
  MPOLDST::mul_ext(p.rep(),a.rep(),C(1)/c);
  return p;
}
//----------------------------------------------------------------------
template <class C, class O, class R> inline
mpoldst<C,O,R> operator%(const mpoldst<C,O,R> & a, const mpoldst<C,O,R> & b)
{
  assert(b != 0);
  mpoldst<C,O,R> r(a), q;
  MPOLDST::div_rem_x(q,r,b);
  return r;
}
//----------------------------------------------------------------------
template <class C, class O, class R>
mpoldst<C,O,R> operator^(const mpoldst<C,O,R> & a, unsigned int n)
{
  mpoldst<C,O,R> p;
  if(n ==0) p = 1;
  else {
    p=1;
    unsigned N = n;
    while(N>0) {p*=a; N--;}
  }
  return p;
}


// //--------------------------------------------------------------------
// /** This function let the variable of index \c i appear as the string
//     \c s. This association is also used, when a polynomial is parsed.
//     In other words, if the variable \c s is parsed, then its index is
//     \c i.
// **/
// template<class T> void Variable(unsigned i, const std::string & s)
// {
//   T::monom_t::_var_of_index[i+1]=s;
//   T::monom_t::_index_of_var[s]=i+1;
// }
// //----------------------------------------------------------------------
// /**  Return the variable of index \c i **/
// template<class T> std::string Variable(unsigned i)
// {
//   return T::monom_t::var_of_index(i);
// }
// //----------------------------------------------------------------------
// /**  Return the index of a variable \c s **/
// template<class T> unsigned IndexOf(const std::string & s)
// {
//   return T::monom_t::index_of_var(s);
// }
//--------------------------------------------------------------------
template<class C, class O, class R, class X>
mpoldst<C,O,R> subs( int variable, const X & value, class mpoldst<C,O,R> & polynom )
{
  //  using namespace MPOLDST;
  return MPOLDST::subs( variable, value, polynom );
};
//--------------------------------------------------------------------
FORALL Polynomial numerator(const Polynomial & p)
{
  typedef typename mpoldst<C,O,R>::monom_t monom_t;
  C m(1);
  for(typename mpoldst<C,O,R>::const_iterator it=p.begin(); it !=p.end(); ++it) 
    {
      m=lcm(m,denominator(it->coeff()));
    }
  return p*m;
  //   mpoldst<U> r; U c;
  //   for(typename mpoldst<C>::const_iterator it=p.begin(); it !=p.end(); ++it) 
  //     {
  //       c = numerator(it->coeff())*(m/denominator(it->coeff()));
  //       monom_t mn(c,it->rep());
  //       r.rep().push_back(mn);
  //     }
  //   return r;
}
//======================================================================
#undef Polynomial
#undef FORALL

} // namespace mmx 
//======================================================================
#endif // SYNAPS_MPOL_MPOL_H

