/********************************************************************
 *   This file is part of the source code of the SYNAPS library.
 *   Author(s): J.P. Pavone, GALAAD, INRIA
 *   $Id: fpu.h,v 1.1 2005/07/11 09:15:23 mourrain Exp $
 ********************************************************************/
#ifndef SYNAPS_NUMERICS_FPU_HPP
#define SYNAPS_NUMERICS_FPU_HPP
//--------------------------------------------------------------------
#include <fenv.h>
//--------------------------------------------------------------------
namespace mmx {
//--------------------------------------------------------------------
namespace numerics
{
  inline int get_cw()  { return fegetround(); };
  inline int get_rnd() { return fegetround(); };
  inline void set_cw( int cw ) { fesetround(cw);};
  inline static int rnd_up() { return FE_UPWARD;};
  inline static int rnd_dw() { return FE_DOWNWARD;};
  inline static int rnd_nr() { return FE_TONEAREST;};
  inline static int rnd_z()  { return FE_TOWARDZERO;};
  template< typename T > struct LongVersion { typedef T result_t; };
  template<> struct LongVersion<float>  { typedef long double result_t; };
  template<> struct LongVersion<double> { typedef long double result_t; };
  template< class T > struct fpu_rounding { 
    static const int id = 0; 
    typedef int rnd_t;
    inline static rnd_t  getrnd() { return 0; };
    inline static void   setrnd( rnd_t r ) { };
    inline static int rnd_up() { return 0; };
    inline static int rnd_dw() { return 0; };
    inline static int rnd_nr() { return 0; };
    inline static int rnd_z()  { return 0; };
    fpu_rounding( int ){}; 
  };
  template<> struct fpu_rounding<long double> 
  { 
    typedef int rnd_t;
    inline static rnd_t  getrnd() { return fegetround(); };
    inline static void   setrnd( rnd_t r ) { fesetround(r); };
    inline static int rnd_up() { return FE_UPWARD;};
    inline static int rnd_dw() { return FE_DOWNWARD;};
    inline static int rnd_nr() { return FE_TONEAREST;};
    inline static int rnd_z()  { return FE_TOWARDZERO;};
    static const int id = 1;
    rnd_t m_prev;
    bool  m_chg;
    fpu_rounding( rnd_t rnd ) 
    {
      m_chg = false;
      m_prev = getrnd();
      if ( m_prev != rnd ) 
	{ 
	  m_chg = true;
	  setrnd(rnd); 
	};
    };
    ~fpu_rounding() { 
      if ( m_chg ) setrnd(m_prev);
    };
  };
 
  
  template< typename T >
  struct rounding : fpu_rounding< typename LongVersion< T >::result_t >  
  { rounding( typename fpu_rounding<typename LongVersion<T>::result_t >::rnd_t rnd ) : fpu_rounding< typename LongVersion< T >::result_t >( rnd ) {}; };
  
  template< typename T > inline T div_dw(const T& a, const T& b) { rounding<T> r(rounding<T>::rnd_dw()); return a/b; };
  template< typename T > inline T div_up(const T& a, const T& b) { rounding<T> r(rounding<T>::rnd_up()); return a/b; };
  template< typename T > inline T add_dw(const T& a, const T& b) { rounding<T> r(rounding<T>::rnd_dw()); return a+b; };
  template< typename T > inline T add_up(const T& a, const T& b) { rounding<T> r(rounding<T>::rnd_up()); return a+b; };
  template< typename T > inline T mul_dw(const T& a, const T& b) { rounding<T> r(rounding<T>::rnd_dw()); return a*b; };
  template< typename T > inline T mul_up(const T& a, const T& b) { rounding<T> r(rounding<T>::rnd_up()); return a*b; };
  template< typename T > inline T sub_dw(const T& a, const T& b) { rounding<T> r(rounding<T>::rnd_dw()); return a-b; };
  template< typename T > inline T sub_up(const T& a, const T& b) { rounding<T> r(rounding<T>::rnd_up()); return a-b; };

  namespace rdw
  {
    template< typename T > inline  
    T upadd( const T& a, const T& b ) { T tmp = -a; tmp -= b; return -tmp;  }; 
    template< typename T > inline  
    T dwadd( const T& a, const T& b ) { return    a+b;    };
    template< typename T > inline  
    T upsub( const T& a, const T& b ) { T tmp =  b; tmp -= a; return -tmp;  };
    template< typename T > inline  
    T dwsub( const T& a, const T& b ) { return    a-b;    };
    template< typename T > inline  
    T upmul( const T& a, const T& b ) { T tmp = -a; tmp *= b; return -tmp;  };
    template< typename T > inline  
    T dwmul( const T& a, const T& b ) { return    a*b;    };
    template< typename T > inline  
    T updiv( const T& a, const T& b ) { T tmp = -a; tmp /= b; return -tmp;  };
    template< typename T > inline  
    T dwdiv( const T& a, const T& b ) { return    a/b;    };
    template< typename T > inline  
    T updet( const T& a, const T& b, const T& c, const T& d ) 
    { return upsub(upmul(a,d),dwmul(b,c)); };
    template< typename T > inline
    T dwdet( const T& a, const T& b, const T& c, const T& d ) 
    { return dwsub(dwmul(a,d),upmul(b,c)); };  
  };
 
  namespace rup
  {
    template< typename T > inline  
    T upadd( const T& a, const T& b ) { return    a+b;    };
    template< typename T > inline  
    T dwadd( const T& a, const T& b ) { T tmp = -a; tmp -= b; return -tmp;  }; 
    template< typename T > inline  
    T upsub( const T& a, const T& b ) { return    a-b;    };
    template< typename T > inline  
    T dwsub( const T& a, const T& b ) { T tmp =  b; tmp -= a; return -tmp;  };
    template< typename T > inline  
    T upmul( const T& a, const T& b ) { return    a*b;    };
    template< typename T > inline  
    T dwmul( const T& a, const T& b ) { T tmp = -a; tmp *= b; return -tmp;  };
    template< typename T > inline  
    T updiv( const T& a, const T& b ) { return    a/b;    };
    template< typename T > inline  
    T dwdiv( const T& a, const T& b ) { T tmp = -a; tmp /= b; return -tmp;  };
    template< typename T > inline  
    T updet( const T& a, const T& b, const T& c, const T& d ) 
    { return upsub(upmul(a,d),dwmul(b,c)); };
    template< typename T > inline
    T dwdet( const T& a, const T& b, const T& c, const T& d ) 
    { return dwsub(dwmul(a,d),upmul(b,c)); };
  }
}; //namespace numerics
//======================================================================
} //namespace mmx
//======================================================================
#endif //
