/*******************************************************************
 *   This file is part of the source code of the SYNAPS kernel.
 *   Author(s): 
 *        E. Tsigaridas <et@di.uoa.gr>
 *           Department of Informatics and Telecommunications
 *           University of Athens (Greece).
 *   Modification:
 *        B. Mourrain, GALAAD, INRIA
 ********************************************************************/

#ifndef _subdivix_solve_contfrac_hpp_
#define _subdivix_solve_contfrac_hpp_

#include <subdivix/Interval.hpp>
#include <subdivix/upoldse.hpp>
#include <subdivix/upoldse_bound.hpp>
#include <subdivix/sign.hpp>
#include <subdivix/Seq.hpp>
#include <subdivix/solver.hpp>
#include <subdivix/contfrac_intervaldata.hpp>
#include <subdivix/contfrac_lowerbound.hpp>


namespace mmx {

  template<class K, class B> struct continued_fraction_isolate: public K
  {
    typedef typename K::integer           coefficient;
    typedef typename K::rational          rational;
    typedef upoldse<coefficient>          polynomial;
    typedef typename K::interval_rational root_t;
    typedef Seq<root_t>                   sol_t;
    
    typedef IntervalData< coefficient, polynomial>  data;  

    static inline root_t as_root(const data& ID)
    {
      return as_interval<rational>(ID);
    }
    static inline bool        is_empty(const data& ID)         {return ID.s==0;}
    static inline bool        is_good(const data& ID)          {return ID.s==1;}
    static inline coefficient lower_bound(const polynomial& p) {return B::lower_bound(p);}
  };


  template<class K, class B> struct continued_fraction_approximate: public K
  {
    typedef typename K::integer           coefficient;
    typedef typename K::rational          rational;
    typedef typename K::interval_rational interval_rational;
    typedef upoldse<coefficient>          polynomial;
    typedef interval_rational             root_t;
    typedef Seq<root_t>                   sol_t;
    typedef IntervalData< coefficient, polynomial>  data;  

    static unsigned prec;
    static inline void     set_precision(unsigned n){ prec = n;}
    static inline unsigned get_precision() { return prec; }

    static inline root_t as_root(const data& ID)
    {
      return as_interval<rational>(ID);
    }

    static inline bool is_empty(const data& ID) {return ID.s==0;}
    static inline bool is_good (const data& ID) 
    {
      //      std::cout << __FUNCTION__ <<" "<<ID.p<<" "<<ID.s<<std::endl; 
      if(ID.s !=1)
	return false;
      else
	{
	  integer N = ID.a*ID.d-ID.b*ID.c, D=ID.c*ID.d ;
	  if(D==0)
	    return false;
	  if(N<0) N = -N;
	  N<<=prec;
	  if(D<0) D = -D;
	  if(N<D)  
	    return true;
	  else
	    return false;
	}      
    }
    static inline coefficient lower_bound(const polynomial& p) {return B::lower_bound(p);}

  };

  template<class K, class B> unsigned continued_fraction_approximate<K,B>::prec = 
	      solver_default_precision;

  template < class K >
  struct continued_fraction_subdivision : public K
  {
    typedef typename K::integer       RT;
    typedef typename K::integer       integer;
    typedef typename K::rational      FT;
    typedef typename K::rational      rational;
    typedef typename K::floating      floating;
    typedef typename K::root_t        root_t;
    typedef typename K::polynomial    polynomial_integer;
    typedef upoldse<rational>         polynomial_rational;
    typedef upoldse<floating>         polynomial_floating;
    typedef typename K::data          data;

    static void set_precision(int prec) {K::set_precision(prec);}

    static void solve_positive(Seq<root_t>& sol, const polynomial_integer& f, bool posneg)
    {
      //      std::cout << __FUNCTION__ << " "<<f<<" "<<posneg<<std::endl; 

      typedef typename K::data IntervalData;  

            IntervalData ID(1, 0, 0, 1, f, 0);
            if (!posneg) 
            {      
                ID.a=-1;
                for (int i = 1; i <= degree(f); i+=2) ID.p[i] *= -1;
            }
            ID.s = sign_variation(ID.p);
	
            if ( K::is_empty(ID) ) { return; }
            if ( K::is_good(ID)  ) {
                //std::cout << "A) Sign variation: 1"<<std::endl;;
                FT B = bound_root( f, Cauchy<FT>());
                if ( posneg )
                     sol << root_t( FT(0), B);
                else sol << root_t( FT(0), FT(-B));
                return;
            }

            solve_homography(sol,ID);
    }
  
    template<class output>
    static void solve_homography(output& sol, const data& ID)
    {
      //            std::cout << __FUNCTION__ << " "<<ID.p<<std::endl; 

            polynomial_integer X(2, AsSize()); X[1] = RT(1); X[0]=RT(0);

            int iters  = 0;
            const RT One(1);
            FT Zero(0);
            //    std::cout << "Polynomial is " << ID.p << std::endl;

            typedef data IntervalData;  
            std::stack< IntervalData > S;
            S.push( ID );
            while ( !S.empty() ) {
                ++iters;
                IntervalData ID = S.top();
                S.pop();

		// std::cout << "Polynomial is " << ID.p << std::endl;
                // Steps 3 - 4: Choose the method for computing the lower bound 
                //--------------------------------------------------------------------
                RT a = K::lower_bound(ID.p);
                //    long k = Bd.lower_power_2( ID.p);
                //    std::cout<< "\t Lower bound is "<< a<< std::endl;
    
                if ( a > RT(1) ) {
                    scale(ID,a);
                    a = RT(1);
                }
                //--------------------------------------------------------------------
                // Step 5 //
                if ( a >= RT(1) ) {
                    shift(ID, a);
                    //      	std::cout<<"Shift by "<<a<<std::endl;

                    if ( ID.p[0] == RT(0)) 
                    {
                        div_x(ID.p,1);
                        sol << root_t( to_FT( ID.b, ID.d), to_FT( ID.b, ID.d ));
                    }
                    ID.s = sign_variation( ID.p);
                }
                //int s= Kernel::init_shift(ID);
                //--------------------------------------------------------------------
                if ( K::is_empty(ID) ) 
                    continue; 
                else if( K::is_good(ID) ) {
		  sol << K::as_root(ID);
                    continue;
                }
                //--------------------------------------------------------------------
                // Step 6
                IntervalData I1;
                shift_by_1( I1, ID);

                unsigned long r  = 0;
    
                if (I1.p[0] == RT(0)) {
                    //	std::cout << "Zero at end point"<<std::endl;;
		  sol << root_t( to_FT( I1.b, I1.d), to_FT(I1.b, I1.d) );
                    div_x(I1.p,1);
                    r = 1;
                }
                I1.s = sign_variation( I1.p);

                IntervalData I2;
                I2.s = ID.s - I1.s - r;
                reverse_shift_homography(I2,ID);
		
		// std::cout <<I1.s <<" "<<I2.s<<std::endl;
                // Step 7;
                if ( !K::is_empty(I2) && !K::is_good(I2) ) {

                    //p2 = p2( 1 / (x+1));
                    reverse( I2.p, ID.p);
                    shift_by_1(I2.p);

                    if ( I2.p[0] == 0) {
                        div_x(I2.p,1);
                    }
                    I2.s = sign_variation( I2.p);
                }

                // Step 8
                if ( I1.s < I2.s ) {std::swap( I1, I2); }
    
                // Step 9
                if ( K::is_good(I1) ) {
                    //	std::cout << "C) Sign variation: 1"<<std::endl;;
		  sol << K::as_root(I1);
                } else if ( !K::is_empty(I1) ) {
                    S.push(I1); //IntervalData( I1.a, I1.b, I1.c, I1.d, I1.p, I1.s));
                }
                // Step 10
                if ( K::is_good(I2) ) {
                    //	std::cout << "D) Sign variation: 1"<<std::endl;;
                    sol <<  K::as_root(I2);
                } else if ( !K::is_empty(I2) ) {
                    S.push(I2); //IntervalData( I2.a, I2.b, I2.c, I2.d, I2.p, I2.s));
                }
            } // while
            //    std::cout << "-------------------- iters: " << iters << std::endl; 
            return;  
    }

    template < class output> inline
    static void solve(output& sol,  const polynomial_integer& f, int  mult=1)
    {
      //      std::cout << __FUNCTION__ << " I "<<f<<std::endl; 
            Seq < root_t > isolating_intervals;
            // Check if 0 is a root
            int idx;
            for (idx = 0; idx <= degree( f); ++idx) {
                if ( f[idx] != 0 ) break;
            }
  
            polynomial_integer p;
            if ( idx == 0 ) { p = f; } 
            else {
                p.resize( degree(f) + 1 - idx);
                for (int j = idx; j <= degree( f); ++j) 
                    p[j-idx] = f[j];
            }
  
            // std::cout << "Solving: " << p << std::endl; 
            // Isolate the negative roots
            //    for (int i = 1; i <= degree(p); i+=2) p[i] *= -1;
            solve_positive( isolating_intervals, p, false);
            // Is 0 a root?
            if (idx != 0) isolating_intervals << root_t( 0, 0);
            //    for (int i = 1; i <= degree(p); i+=2) p[i] *= -1;
            solve_positive( isolating_intervals, p, true);

            //  sort( isolating_intervals.begin(), isolating_intervals.end(), CompareFIT());
            //  std::cout << "ok sorting" << std::endl; 

            for (unsigned i = 0 ; i < isolating_intervals.size(); ++i) {
                if ( singleton( isolating_intervals[i]) ) {
		  if ( lower( isolating_intervals[i]) == 0 ) continue;
      
                    polynomial_integer t( 2, AsSize());
                    t[1] = denominator( lower(isolating_intervals[i]));
                    t[0] = - numerator( lower(isolating_intervals[i]));
      
                    polynomial_integer dummy;
                    UPOLDSE::div_rem( dummy, p, t);
                    p = dummy;
                }
            }
  
            //  std::cout << "now p: " << p << ",  #nr: " << isolating_intervals.size() << std::endl; 

            //  std::cout << "Done...." << std::endl; 
            for (unsigned i = 0 ; i < isolating_intervals.size(); ++i) 
	      {sol << isolating_intervals[i];}
            //    return sol;
    }

    template < class output> inline
    static void solve(output& sol,  const polynomial_rational& f)
    {
      //      std::cout << __FUNCTION__ << " R "<<f<<std::endl; 
      solve(sol, UPOLDSE::numer<polynomial_integer> (f));
    }

    template < class output> inline
    static void solve(output& sol,  const polynomial_floating& f)
    {
      //      std::cout << __FUNCTION__ << " R "<<f<<std::endl; 
      polynomial_rational p(degree(f)+1,AsSize());

      for(int i=0;i<degree(f)+1;i++) 
	{
	  p[i] = as<rational>(f[i]);
	}
      solve(sol,p);
    }

  };

  //--------------------------------------------------------------------
  template<class K> 
  struct solver_continued_fraction_approximate: 
    public continued_fraction_subdivision<
              continued_fraction_approximate<K,
	         AkritasBound<typename K::integer> > > 
  {

    typedef continued_fraction_subdivision<
              continued_fraction_approximate<K,
	         AkritasBound<typename K::integer> > >
                                                 base_t;
    typedef typename K::interval_rational        interval_t;
    typedef Seq<interval_t>                      solutions_t;
    typedef typename base_t::polynomial          polynomial_integer;
    typedef typename base_t::polynomial_rational polynomial_rational;
    typedef typename base_t::polynomial_floating polynomial_floating;


    template < class output > inline
    static void solve(output& sol, const polynomial_integer& f)
    {
       base_t::solve(sol, f);
    }


    template < class output > inline
    static void solve(output& sol, const polynomial_rational& f)
    {
       base_t::solve(sol, f);
    }
    template < class output > inline
    static void solve(output& sol, const polynomial_floating& f)
    {
       base_t::solve(sol, f);
    }

    template < class output > inline
    static void solve(output& sol, const polynomial_integer& f, int prec)
    {
      base_t::set_precision(prec);
      solve(sol, f);
      base_t::set_precision(solver_default_precision);
    }

    template < class output> inline
    static void solve(output& sol, const polynomial_integer& f, const interval_t& I)
    {
      typename K::integer 
	a= numerator(upper(I)), 
	b= numerator(lower(I)), 
	c= denominator(upper(I)),  
	d= denominator(lower(I));
      IntervalData<integer, polynomial_integer> D = 
	as_interval_data(a,b,c,d,f);
      solve_homography(sol,D);
    }

    template < class output> inline
    static void solve(output& sol, const polynomial_rational& f, const interval_t& I)
    {
      solve(sol, UPOLDSE::numer<polynomial_integer>(f), I);
    }


  };

  template<class K> 
  struct solver_continued_fraction_isolate: 
    public continued_fraction_subdivision<
              continued_fraction_isolate<K,
	         AkritasBound<typename K::integer> > > {};


  //====================================================================
  template<class M, class K=Numerix> struct ContFrac{};

  //--------------------------------------------------------------------
  template<class K> struct ContFrac<Isolate,K>
  {
    typedef Seq<typename K::interval_rational> solutions_t;
    typedef solver_continued_fraction_isolate<K> solver_t;
  };
  //--------------------------------------------------------------------
  template<class K> struct ContFrac<Approximate,K>
  {
    typedef typename K::interval_rational        interval_t;
    typedef Seq<interval_t>                      solutions_t;
    typedef solver_continued_fraction_approximate<K> solver_t;

    static inline void set_precision(unsigned p)
    {
       continued_fraction_approximate<K,AkritasBound<typename K::integer> >::set_precision(p);
    }
    
  };

  //--------------------------------------------------------------------

//--------------------------------------------------------------------
} //namespace mmx 
#endif // _subdivix_solve_contfrac_hpp_
