/*******************************************************************
 *   This file is part of the source code of the SYNAPS kernel.
 *   Author(s): 
 *        E. Tsigaridas <et@di.uoa.gr>
 *           Department of Informatics and Telecommunications
 *           University of Athens (Greece).
 *   Modification:
 *        B. Mourrain, GALAAD, INRIA
 *        V. Sharma, GALAAD, INRIA
 ********************************************************************/

#ifndef _SYNAPS_SOLVE_CONTFRAC_H_
#define _SYNAPS_SOLVE_CONTFRAC_H_

#include <subdivix/Interval.hpp>
#include <subdivix/upoldse.hpp>
#include <subdivix/upoldse_bound.hpp>
#include <subdivix/sign.hpp>
#include <subdivix/Seq.hpp>

#include <subdivix/contfrac_intervaldata.hpp>
#include <subdivix/contfrac_lowerbound.hpp>


namespace mmx {

    template<class K, class L = AkritasBound<typename K::integer> > struct ContFrac_
    {
        typedef K                           Kernel;
        typedef typename K::integer         RT;
        typedef typename K::integer         integer;
        typedef typename K::rational        FT;
        typedef typename K::rational        rational;
        typedef Interval<FT>                FIT;
        typedef Interval<FT>                root_t;
        typedef upoldse<integer>            Poly;
        typedef L                           bound;
        typedef ContFrac_<K,L>               self_t;

        typedef IntervalData< RT, Poly>     data;  

    };

    template<class C, class V> struct Solver;
    template < class NT , class K, class LowerBound >
    struct Solver<NT, ContFrac_ <K, LowerBound> >
    {
        typedef NT                        RT;
        typedef typename K::integer       integer;
        typedef typename K::rational      FT;
        typedef typename K::rational      rational;
        typedef Interval<FT>              FIT;
        typedef Interval<rational>        root_t;
        typedef upoldse<integer>          Poly;
        typedef upoldse<rational>         PolyFT;
        typedef LowerBound                bound;
        typedef ContFrac_<K,LowerBound>    Kernel;


        void solve_positive(Seq<FIT>& sol, const Poly& f, bool posneg)
        {
            //    std::cout << __FUNCTION__ << " "<<f<<" "<<posneg<<std::endl; 

            typedef IntervalData< RT, Poly> IntervalData;  

            IntervalData ID(1, 0, 0, 1, f, 0);
            if (!posneg) 
            {      
                ID.a=-1;
                for (int i = 1; i <= degree(f); i+=2) ID.p[i] *= -1;
            }
            ID.s = sign_variation(ID.p);
	
            if ( ID.s == 0 ) { return; }
            if ( ID.s == 1 ) {
                //std::cout << "A) Sign variation: 1"<<std::endl;;
                FT B = bound_root( f, Cauchy<FT>());
                if ( posneg )
                {
                    //	  std::cout<<as_interval<rational>( FT(0), B)<<std::endl;;
                    sol << as_interval<rational>( FT(0), B);
                }
                else sol << as_interval<rational>( FT(0), FT(-B));
                return;
            }

            solve_homography(sol,ID);
        }
  
        void solve_homography(Seq<FIT>& sol, const IntervalData< RT, Poly>& ID)
        {
            //    std::cout << __FUNCTION__ << " "<<ID.p<<std::endl; 

            Poly X(2, AsSize()); X[1] = RT(1); X[0]=RT(0);

            int iters  = 0;
            const RT One(1);
            FT Zero(0);
            //    std::cout << "Polynomial is " << ID.p << std::endl;

            typedef IntervalData< RT, Poly> IntervalData;  
            std::stack< IntervalData > S;
            S.push( ID );
            while ( !S.empty() ) {
                ++iters;
                IntervalData ID = S.top();
                S.pop();

                // std::cout << "Polynomial is " << ID.p << std::endl;
                // Steps 3 - 4: Choose the method for computing the lower bound 
                //--------------------------------------------------------------------
                LowerBound Bd;

                RT a = Bd.lower_bound(ID.p);
                //    long k = Bd.lower_power_2( ID.p);
                //    std::cout<< "\t Lower bound is "<< a<< std::endl;
    
                if ( a > RT(1) ) {
                    scale(ID,a);
                    a = RT(1);
                }
                //--------------------------------------------------------------------
                // Step 5 //
                if ( a >= RT(1) ) {
                    shift(ID, a);
                    //      	std::cout<<"Shift by "<<a<<std::endl;

                    if ( ID.p[0] == RT(0)) 
                    {
                        div_x(ID.p,1);
                        sol << as_interval<rational>( to_FT( ID.b, ID.d), 
                                                      to_FT( ID.b, ID.d ));
                    }
                    ID.s = sign_variation( ID.p);
                }
                //int s= Kernel::init_shift(ID);
                //--------------------------------------------------------------------
                if ( ID.s == 0 ) 
                    continue; 
                else if(ID.s==1) {
                    sol << as_interval<rational>(ID);
                    continue;
                }
                //--------------------------------------------------------------------
                // Step 6
                IntervalData I1;
                shift_by_1( I1, ID);

                unsigned long r  = 0;
    
                if (I1.p[0] == RT(0)) {
                    //	std::cout << "Zero at end point"<<std::endl;;
                    sol << as_interval<rational>( to_FT( I1.b, I1.d), to_FT(I1.b, I1.d) );
                    div_x(I1.p,1);
                    r = 1;
                }
                I1.s = sign_variation( I1.p);

                IntervalData I2;
                I2.s = ID.s - I1.s - r;
                reverse_shift_homography(I2,ID);
    
                // Step 7;
                if ( I2.s > 1 ) {

                    //p2 = p2( 1 / (x+1));
                    reverse( I2.p, ID.p);
                    shift_by_1(I2.p);

                    if ( I2.p[0] == 0) {
                        div_x(I2.p,1);
                    }
                    I2.s = sign_variation( I2.p);
                }

                // Step 8
                if ( I1.s < I2.s ) {std::swap( I1, I2); }
    
                // Step 9
                if ( I1.s == 1 ) {
                    //	std::cout << "C) Sign variation: 1"<<std::endl;;
                    sol << as_interval<rational>(I1);
                } else if ( I1.s > 1) {
                    S.push(I1); //IntervalData( I1.a, I1.b, I1.c, I1.d, I1.p, I1.s));
                }

                // Step 10
                if ( I2.s == 1 ) {
                    //	std::cout << "D) Sign variation: 1"<<std::endl;;
                    sol << as_interval<rational>(I2);
                } else if ( I2.s > 1) {
                    S.push(I2); //IntervalData( I2.a, I2.b, I2.c, I2.d, I2.p, I2.s));
                }
            } // while
            //    std::cout << "-------------------- iters: " << iters << std::endl; 
            return;  
        }

        template < class output> inline
        void solve(output& sol,  const Poly& f, int  mult=1)
        {

            Seq < FIT > isolating_intervals;
            // Check if 0 is a root
            int idx;
            for (idx = 0; idx <= degree( f); ++idx) {
                if ( f[idx] != 0 ) break;
            }
  
            Poly p;
            if ( idx == 0 ) { p = f; } 
            else {
                p.resize( degree(f) + 1 - idx);
                for (int j = idx; j <= degree( f); ++j) 
                    p[j-idx] = f[j];
            }
  
            // std::cout << "Solving: " << p << std::endl; 
            // Isolate the negative roots
            //    for (int i = 1; i <= degree(p); i+=2) p[i] *= -1;
            solve_positive( isolating_intervals, p, false);
            // Is 0 a root?
            if (idx != 0) isolating_intervals << FIT( 0, 0);
            //    for (int i = 1; i <= degree(p); i+=2) p[i] *= -1;
            solve_positive( isolating_intervals, p, true);

            //  sort( isolating_intervals.begin(), isolating_intervals.end(), CompareFIT());
            //  std::cout << "ok sorting" << std::endl; 

            for (unsigned i = 0 ; i < isolating_intervals.size(); ++i) {
                if ( singleton( isolating_intervals[i]) ) {
                    if ( lower( isolating_intervals[i]) == 0 ) continue;
      
                    Poly t( 2, AsSize());
                    t[1] = denominator( isolating_intervals[i].lower());
                    t[0] = - numerator( isolating_intervals[i].lower());
      
                    Poly dummy;
                    UPOLDSE::div_rem( dummy, p, t);
                    p = dummy;
                }
            }
  
            //  std::cout << "now p: " << p << ",  #nr: " << isolating_intervals.size() << std::endl; 

            //  std::cout << "Done...." << std::endl; 
            for (unsigned i = 0 ; i < isolating_intervals.size(); ++i) {sol << isolating_intervals[i];}
            //    return sol;
        }

        template < class output> inline
        void solve(output& sol,  const PolyFT& f, int  mult=1)
        {
            Poly p(degree(f)+1,AsSize());
            integer d=1;
            for(int i=0;i<degree(f)+1;i++) d=lcm(d,numerator(f[i]));
            for(int i=0;i<degree(f)+1;i++) 
                p[i]=numerator(f[i])* (d/denominator(f[i]));
            solve(sol,p,mult);
        }
    };


} //namespace mmx 


#endif // _SYNAPS_SOLVE_CONTFRAC_H_
