/*******************************************************************
 *   This file is part of the source code of the SYNAPS kernel.
 *   Author(s): B. MOurrain, GALAAD, INRIA
 ********************************************************************/
#ifndef synaps_solver_uvp_snc_hpp
#define synaps_solver_uvp_snc_hpp
/********************************************************************/
#include <subdivix/solver_uvp_sleeve.hpp>
#include <subdivix/solver_uvp_contfrac.hpp>
//====================================================================
namespace mmx {

  //  struct SymbNum {};

  template<class K, class L = AkritasBound<typename K::integer> > struct SymbNum 
  {
    static Solver<double,Sleeve<K> > slv;

    typedef typename K::integer         integer;
    typedef typename K::rational        rational;
    typedef typename K::floating        floating;
    typedef typename K::ieee            ieee;

    typedef K                           Kernel;
    typedef typename K::integer         RT;
    typedef typename K::rational        FT;
    typedef Interval<FT>                FIT;
    typedef Interval<FT>                root_t;
    typedef upoldse<integer>            Poly;

   
  };
  template<class K, class L> Solver<double,Sleeve<K> > SymbNum<K,L>::slv;

  /// Class
  template < class NT , class K, class L >
  struct Solver<NT, SymbNum <K, L> > : public Solver< NT, ContFrac_<K,L> >

  {
    typedef NT                          RT;
    typedef typename K::integer         integer;
    typedef typename K::rational        FT;
    typedef typename K::rational        rational;
    typedef Interval<FT>                FIT;
    typedef Interval<rational>          root_t;
    typedef upoldse<integer>            Poly;
    typedef SymbNum<K,L>                Kernel;
    typedef IntervalData<integer, Poly> Data;  
    typedef Solver<NT, SymbNum <K, L> > self_t;

    root_t to_sol(const Data& I)
    {
      rational left, right;
      if ( I.c == RT(0)) 
	left = sign(I.a) * bound_root( I.p, Cauchy<rational>());
      else 
	left = rational(I.a)/I.c;
  
      if ( I.d == RT(0)) 
	right = sign(I.b) * bound_root( I.p, Cauchy<rational>()); 
      else 
	right = rational(I.b)/I.d;
      //      std::cout << I.a<<"/"<<I.c<<", "<<I.b<<"/"<<I.d<<std::endl;
      return root_t(left, right);
    }

    void solve_positive(Seq<FIT>& sol, const Poly& f, bool posneg)
    {
      //  std::cout << __FUNCTION__ << std::endl; 
      Data ID( 1, 0, 0, 1, f, 0);
      if (!posneg) 
	{      
	  ID.a=-1;
	  for (int i = 1; i <= degree(f); i+=2) ID.p[i] *= -1;
	}
      ID.s = sign_variation(ID.p);

      //      unsigned long s = sign_variation(f);

      if ( ID.s == 0 ) { return; }
      if ( ID.s == 1 ) {
	//      std::cout << "A) Sign variation: 1"<<std::endl;;
	FT B = bound_root( f, Cauchy<RT>());
	if ( posneg )
	  {
	    sol << as_interval<rational>( FT(0), B);
	  }
	else sol << as_interval<rational>( FT(0), FT(-B));
	return;
      }

      Seq<Interval<double> > rsol; 
      Kernel::slv.solve_positive(rsol,ID.p,Approx());
      //	std::cout<< "<polynomial>"<< ID.p<<"</polynomial>"<< std::endl;
      //	std::cout<< "\t Approximate roots "<< rsol<< std::endl;


      std::vector<double> up(degree(f)), dw(degree(f));
      Poly D= diff(f);
      rational mx = as<rational>(VECTOR::max_abs(D));
      assert( mx != 0 );
      
      upoldse<Interval<double> > dp (degree(f),AsSize());
      {
	numerics::rounding<double> rnd( numerics::rnd_up() );
	for (int i = 0; i < degree(f); i ++ )  
	  dp[i].upper()=as<double>(as<rational>(D[i]/mx));
	//dp[i].upper()=as<double>(D[i]);
      }
      {
	numerics::rounding<double> rnd( numerics::rnd_dw() );
	for (int i = 0; i < degree(f); i ++ )  
	  dp[i].lower()=as<double>(as<rational>(D[i]/mx));
	//dp[i].lower()=as<double>(D[i]);
      }
      //      std::cout <<dp<<std::endl;

      //      Poly X(2, AsSize()); X[1] = RT(1); X[0]=RT(0);

      for(unsigned i=0;i<rsol.size();i++)
	{
	  //	  std::cout <<rsol[i]<<" "<<dp(rsol[i])<<std::endl;
	  if((rsol[i].lower()!=-INFINITY) && (rsol[i].upper()!=INFINITY) 
	     && !contain_zero(dp(rsol[i])))
	    {
	      sol<<Interval<rational>(as<rational>(rsol[i].lower()),
				      as<rational>(rsol[i].upper()));
	      continue;
	    }
	  else
	    {
	      Data PH(ID);
 	      integer a = (int)floor(rsol[i].lower());
 	      shift(PH,a);
 	      reverse(PH);
 	      shift_by_1(PH);
 	      if (PH.p[0] == 0) {
 		  //		  std::cout << "Zero at end point"<<std::endl;;
 		sol << as_interval<rational>( to_FT( PH.b, PH.d), to_FT(PH.b, PH.d) );
 		div_x(PH.p,1);
 	      }
 	      if ( PH.s == 1 ) {
 		  //	std::cout << "C) Sign variation: 1"<<std::endl;;
 		  sol << as_interval<rational>(PH);
 		} 
	      this->solve_homography(sol,PH);
	    }
	}
    }
    //--------------------------------------------------------------------
    template < class output> inline
    void solve(output& sol,  const Poly& f, int  mult=1)
    {
      
      Seq < FIT > isolating_intervals;
      // Check if 0 is a root
      int idx;
      for (idx = 0; idx <= degree( f); ++idx) {
	if ( f[idx] != 0 ) break;
      }
  
      Poly p;
      if ( idx == 0 ) { p = f; } 
      else {
	p.resize( degree(f) + 1 - idx);
	for (int j = idx; j <= degree( f); ++j) 
	  p[j-idx] = f[j];
      }
      
      //std::cout << "Solving: " << p << std::endl; 
      // Isolate the negative roots
      solve_positive( isolating_intervals, p, false);
      // Is 0 a root?
      if (idx != 0) isolating_intervals << FIT( 0, 0);
      solve_positive( isolating_intervals, p, true);
      
      //  sort( isolating_intervals.begin(), isolating_intervals.end(), CompareFIT());
      //  std::cout << "ok sorting" << std::endl; 
      
      for (unsigned i = 0 ; i < isolating_intervals.size(); ++i) {
	if ( singleton( isolating_intervals[i]) ) {
	  if ( lower( isolating_intervals[i]) == 0 ) continue;
	  
	  Poly t( 2, AsSize());
	  t[1] = denominator( isolating_intervals[i].lower());
	  t[0] = - numerator( isolating_intervals[i].lower());
	  
	  Poly dummy;
	  UPOLDSE::div_rem( dummy, p, t);
	  p = dummy;
	}
      }
      
      //  std::cout << "now p: " << p << ",  #nr: " << isolating_intervals.size() << std::endl; 
      //  std::cout << "Done...." << std::endl; 
      for (unsigned i = 0 ; i < isolating_intervals.size(); ++i) {sol << isolating_intervals[i];}
      //    return sol;
    }
    //--------------------------------------------------------------------
  };
//====================================================================
} //namespace mmx
#endif  //synaps_solver1d_sleeve_hpp
