/********************************************************************
 *   This file is part of the source code of the SYNAPS library.
 *   Author(s): F.Carreras, B.Mourrain, O.Ruatta, C.Tanasescu
 *   $Id: SquareFree.hpp,v 1.1 2005/07/11 08:12:26 mourrain Exp $
 ********************************************************************/
#ifndef SYNAPS_UPOL_SQUAREFREE_H_
#define SYNAPS_UPOL_SQUAREFREE_H_
//--------------------------------------------------------------------
#include <list>
#include <vector>

#include <subdivix/Seq.hpp>
#include <subdivix/upoldse.hpp>
#include <subdivix/gcd.hpp>


namespace mmx {

/*************************************************************************/
/**
 *      Square-free part of an univariate polynomial using gcd.
 *      We assume that C is a field.
 **/
template<class C>
upoldse<C> square_free(const upoldse<C> & p)
{
  if(p==0)
    return upoldse<C>(0);
  else
    {
      typedef upoldse<C> Pol;
      Pol h = p, dp = diff(h);
      Pol g = gcd(h,dp);
      h/= g;
      // h/= h[degree(h)];
      return h;
    }
}
/*************************************************************************/
template<class C>
upoldse<upoldse<C> > square_free2(const upoldse<upoldse<C> > & p)
{
  typedef upoldse<upoldse<C> > PolPol;
  typedef upoldse<C> Pol;
  PolPol h = p;
  PolPol dp = diff(p);
  PolPol g = gcd2(h,dp);

  h/=g;
  Pol lc = h[degree(h)];
  h /= lc[degree(lc)];
  return h;
}

/*************************************************************************/
/**
 * \brief Computes a square-free factorisation based on Yun's algorithm.
 * \param $f$ The polynomial that is to be factorized
 * \return A sequence of polynomials. The first element in the sequence is always zero.
 * The position of the polynomial indicates its multiplicity as a factor of $f$
 */
template < typename _POL >
Seq< _POL >
square_free_factorisation( const _POL& f)
{


  typedef _POL    Poly;

  Seq< Poly > res;

  //  res.resize( UPOLDSE::degree(f) );

  Poly df = UPOLDSE::diff(f);

  Poly c = gcd(f, df);
  
  if ( UPOLDSE::degree( c) <= 0) {
    res<<f;
    return res;
  }
  
  //  int n = 0;
  Poly w = f / c;
  Poly y = df / c;
  Poly z = y - UPOLDSE::diff(w);
  Poly g;
  
  while ( true ) {
      g = gcd( w, z);
      if ( UPOLDSE::degree( g) > 0) {
	res<< g;
      }
      else
	res<< Poly(1);

      w /= g;
      y = z / g;
      z = y - UPOLDSE::diff( w);
      if ( UPOLDSE::degree(z) <= 0) break;
  }

  res<<w;
  return res;
}


template < typename RT >
Seq< upoldse<RT> >
square_free_factorization( const upoldse<RT>& f, SturmSeq<RT>& sth)
{

  typedef upoldse<RT>    Poly;
  Seq< Poly > res;

  Poly df = UPOLDSE::diff(f);

  Poly c;
  sth = SturmSeq<RT>( f / content(f), df, SUBRES());

  typename SturmSeq<RT>::reverse_iterator it = sth.rep().rbegin();
  
  for (unsigned i = 0; i < sth.size(); ++i) {
    if ( degree( sth[i]) >= 0 )  sth[i] /= content(sth[i]);
  }

  for(; (it!= sth.rep().rend()) && (*it == 0) ; ++it);
  RT temp = content( *it);
  c = (*it) / temp;
  
  if ( degree(c) <= 0) {
    res<<f;
    return res;
  }
  
  //  int n = 0;

  Poly w = f / c;
  Poly y = df / c;
  Poly z = y - diff(w);
  Poly g;
  
  while ( true ) {
      g = gcd( w, z);
      if ( degree(g) > 0) {
	res << g;
      }
      else
	res <<Poly(1);
      w /= g;
      y = z / g;
      z = y - diff( w);
      if ( degree(z) <= 0) break;
  }
  res<<w;
  return res;
}
} //namespace mmx
//--------------------------------------------------------------------
#include <subdivix/Zmod.hpp>
namespace mmx 
{
  template < typename RT >
  bool is_square_free_mod(const upoldse<RT>& f)
  {
    typedef  Zmod<31991>  RTmod;
    RT M(31991), r;
    upoldse< RTmod > p(degree(f)+1, AsSize());
    for(int i=0; i<= degree(f); i++)
      {
	r = f[i] % M;
	p[i]= RTmod(mpz_get_si(r.get_mpz_t()));
      }
    SturmSeq<RTmod> sth = SturmSeq<RTmod>( p, diff(p), HABICHT());
    //    std::cout << sth <<std::endl;
    return sth[sth.size()-1]!=0;
  }
}
//--------------------------------------------------------------------
#endif // SYNAPS_UPOL_SQUAREFREE_H_
