/*********************************************************************
*      This file is part of the source code of SYNAPS library.       *
*      Author(s): B. Mourrain, GALAAD, INRIA                         *
**********************************************************************
$Id: upoldse.hpp,v 1.2 2005/08/29 14:30:17 mourrain Exp $
**********************************************************************/
#ifndef SYNAPS_UPOL_UPOLDSE_H
#define SYNAPS_UPOL_UPOLDSE_H
#include <subdivix/shared_object.hpp>
#include <subdivix/parser.hpp>
#include <stdlib.h>
#include <string.h>
#include <subdivix/let.hpp>
#include <subdivix/UPOLDSE_m.hpp>
#include <subdivix/EUCLIDEAN_m.hpp>
#include <subdivix/linalg_rep1d.hpp>

#define TMPL template<typename C, typename R>
#define Polynomial upoldse<C,R>

namespace mmx {


//----------------------------------------------------------------------
template<class I,class V, class N> struct upol_iterator;
//======================================================================
namespace upol {
//======================================================================
/**
   Array polynomial representation, as a subclass of
   @linalg::rep1d<C>@.
**/
template <class C>
struct rep : public linalg::rep1d<C> {

  typedef C coeff_t;
  typedef typename linalg::rep1d<C>::size_type size_type;
  typedef typename linalg::rep1d<C>::iterator iterator;
  typedef typename linalg::rep1d<C>::const_iterator const_iterator;
  typedef typename linalg::rep1d<C>::const_reverse_iterator const_reverse_iterator;
  typedef typename linalg::rep1d<C>::reverse_iterator reverse_iterator;

  int degree_;

  rep(): linalg::rep1d<C>(), degree_(-1) {}
  rep(size_type s, AsSize): linalg::rep1d<C>(s), degree_(s-1) {}
  rep(size_type s, C *t): linalg::rep1d<C>(s,t), degree_(s-1)
    {checkdegree(*this);}
  rep(C* b, C *e): linalg::rep1d<C>(b,e)
    {degree_=this->size_-1;checkdegree(*this);}
  rep(const rep<C> & p): linalg::rep1d<C>(p), degree_(p.degree_)
    {}//{cout <<"copy rep"<<endl;}

  void resize(size_type n)
    {this->linalg::rep1d<C>::resize(n); degree_=n-1;}

  iterator          end()         { return iterator(this->tab_+degree_+1); }
  const_iterator    end()   const { return const_iterator(this->tab_+degree_+1); }

  reverse_iterator       rbegin()
    {return reverse_iterator(this->tab_+degree_+1); }
  const_reverse_iterator rbegin() const
    {return const_reverse_iterator(this->tab_+degree_+1); }

  rep<C> & operator=(const rep<C> & v)
    {this->linalg::rep1d<C>::operator=(v); degree_=v.degree_; return *this;}

};
//-------------------------------------------------------------------------
template <class C> inline
int checkdegree(rep<C> & a)
{
  int l = a.degree_;
  while(l >=0 && (a.tab_[l]==0)) {l--;}
  a.degree_=l;
  return a.degree_;
}
//-------------------------------------------------------------------------
template <class C>
void assign(rep<C> & a, const rep<C> & b)
{
  assign((linalg::rep1d<C> &)a,(const linalg::rep1d<C> &)b);
  a.degree_ = b.degree_;
}
//-------------------------------------------------------------------------
template<class I, class V, class N>
struct upol_iterator
{
public:
  typedef upol_iterator<I,V,N> self_t;

  upol_iterator<I,V,N>(): my_ptr(), my_idx(0)            {}
  upol_iterator(V* t): my_ptr(t), my_idx(0)        {}
  upol_iterator(V* t, int n): my_ptr(t), my_idx(n) {}
  upol_iterator(const upol_iterator<I,V,N> & t):
    my_ptr(t.my_ptr), my_idx(t.my_idx) {}

  self_t&
  operator++()  {my_ptr++; my_idx++; return *this;}

  self_t&
  operator--()  {my_ptr--; my_idx--; return *this;}

  V   operator*() const {return *my_ptr;}
  V & operator*()       {return *my_ptr;}

  int index()   {return my_idx;}

  bool operator!=( self_t const& rhs) { return my_ptr != rhs.my_ptr; }
  bool operator<=( self_t const& rhs) { return my_ptr <= rhs.myy_ptr; }

private:
  I  my_ptr;
  N  my_idx;
};

//======================================================================
}//CLOSE upol
//======================================================================




/// Univariate polynomials.
/**
Univariate polynomials, with a dense representation given by the container
@R@. Forward and backward iterators are provided, as well as the arithmetic
operations @+,-,+=,-=,*,*=,/, /=, %,%=@. The operators @%@ and @/@
correspond respectively to the quotient and remainder computation in the
Euclidean division.
The @operator()(const C &)@  corresponds to the evaluation of the polynomial.

The function @checkdegree@ is called at the end of an
arithmetic operation, to reajust the variable @degree_@, if it exists
in the container @R@.

The container \texttt{R} should provide the following signatures:
\begin{SIGNATURE}
  typename R::value_type coeff_t;
  typedef typename R::size_type;
  typedef typename R::iterator;

  R(size_type s, Alloc);
  R(size_type s, C *t);
  R(iterator b, iterator e);
  void R::reserve(size_type n);
  iterator  R::begin()       ;
  iterator  R::begin() const ;
  iterator  R::end()         ;
  iterator  R::end()   const ;
  iterator  R::rbegin()      ;
  iterator  R::rbegin() const;
  iterator  R::rend()        ;
  iterator  R::rend()   const;
  size_type R::size();
  int  degree(const R &);
\end{SIGNATURE}

\index{Univariate polynomial, dense} \help{upoldse}
**/
template<class C, class R=upol::rep<C> >
struct upoldse {

#ifdef NOREFCOUNT
  R data;
  R &       rep()       {return data;}
  const R & rep() const {return data;}
#else
  shared_object<R> data;
  R &       rep()       {return *data;}
  const R & rep() const {return *data;}
#endif
  typedef C                            value_type;
  typedef C                            coeff_t;
  typedef typename R::size_type        size_type;
  typedef R                            rep_type;
  typedef typename R::iterator         iterator;
  typedef typename R::const_iterator   const_iterator;
  typedef typename R::reverse_iterator reverse_iterator;
  typedef typename R::const_reverse_iterator const_reverse_iterator;
  typedef Polynomial                self_t;

  static char* variable;
  
  upoldse() : data() {}
  upoldse(const R & r) {data(r);}
  upoldse(const int n);
  //  upoldse( const coeff_t & c );
  upoldse(size_type n, AsSize A)
    {rep().resize(n);VECTOR::set_cst(rep(),(value_type)0);}
  upoldse(value_type c, size_type n = 0 );
  upoldse(size_type n, value_type * t);
  upoldse(iterator b, iterator e);

  upoldse(const char *);
  upoldse(const self_t & P) {data=P.data;} //COUNT<self_t>('c');}


  static void set_variable(const char *s) { 
    variable = new char[strlen(s)+1];
    strcpy(self_t::variable, s);
  }

  iterator               begin()        {return rep().begin();}
  const_iterator         begin()  const {return rep().begin();}
  iterator               end()          {return rep().end();}
  const_iterator         end()    const {return rep().end();}
  reverse_iterator       rbegin()       {return rep().rbegin();}
  const_reverse_iterator rbegin() const {return rep().rbegin();}
  reverse_iterator       rend()         {return rep().rend();}
  const_reverse_iterator rend()   const {return rep().rend();}

  self_t & operator=(const self_t & P)
    {
      data=P.data;
      //  COUNT<self_t>('=');
      return *this;
    }

//   template<class S>
//   self_t & operator =(const VAL<S> & M)
//   {let::assign(*this,M.rep());return *this;}

  bool operator==(int n) const;
  bool operator!=(int n) const {return !(*this==n);}

  bool operator==(const self_t & p) const;
  bool operator!=(const self_t & p) const {return !(*this==p);}

  self_t & operator+= (const self_t  & P);
  self_t & operator+= (const value_type & c);

  self_t & operator-= (const self_t  & P);
  self_t & operator-= (const value_type & c);

  self_t & operator*= (const self_t  & P);
  self_t & operator*= (const value_type & c);

  self_t & operator/= (const self_t  & Q);
  self_t & operator/= (const value_type & c);

  self_t & operator%= (const self_t  & Q);
  self_t & operator%= (const value_type & x );
  int get_degree() const 
  { 
    using UPOLDSE::degree; return degree(rep());
  }
  //  int degree() const { return get_degree(); };
  size_type   size() const { return (rep().size()); }
  void resize(size_type i) { rep().resize(i); }

  inline  value_type &  operator[] (size_type i)
  {
    assert(i<size()); return (rep())[i];
  }
  inline  value_type    operator[] (size_type i) const
  {
    if(i< std::max((size_type)(this->get_degree()+1),(size_type)0))
      return (rep())[i]; else return (value_type)0;
  }

  inline value_type coeff(size_type i) const { return (*this)[i];}

  template<class T> T operator()(const T & c) const;
  
  bool operator<( const value_type & c ) { return false; };
};
//====================================================================
TMPL unsigned hash(const Polynomial& p){ return 1;}
TMPL unsigned  soft_hash(const Polynomial& p){ return 1;}
TMPL bool eq(const Polynomial& p, const Polynomial& q) { return p==q;}

template<class C, class R> inline C read(const Polynomial& m, unsigned i) { return m[i]; }

template<class C, class R> inline bool adjust_denum( const Polynomial & ) { return false; };
//======================================================================
TMPL char* Polynomial::variable= (char *) "x";

// /** return a reference to the output characters of the variable. Allow to
//     change it:  @Variable<upoldse<C,R> >()=const_cast<"y">@.
// **/
// template<class P> char*& Variable(void) {return P::variable;}

//======================================================================
/// non equality test
TMPL  inline
bool neq(const Polynomial& p, const Polynomial& q)
{
  return (p!=q);
}
//====================================================================
TMPL inline unsigned size(const Polynomial& p) {return p.size();}
//--------------------------------------------------------------------
/** Output the degree, that is the index of the last non-zero
    coefficient.   
**/
TMPL inline int degree(const Polynomial& p)
{
  using UPOLDSE::degree;  return degree(p.rep());
}
//----------------------------------------------------------------------
/** Compute the derivative of a polynomial. **/
TMPL  inline
Polynomial const diff(const Polynomial& p)
{
  if(degree(p)>0)
    {
      Polynomial r(degree(p),AsSize());
      using UPOLDSE::diff; diff(r.rep(),p.rep());
      return r;
    }
  else
    return 0;
}
//----------------------------------------------------------------------
/** Compute the $k^{\rm th}$ derivative of a polynomial. **/
TMPL  inline
Polynomial const diff(const Polynomial& p, unsigned int k)
{
  if(degree(p)>(int)k-1)
    {
      Polynomial r(p);
      using UPOLDSE::diff;
      for(unsigned i=0; i<k;++i) diff(r.rep(),r.rep());
      return r;
    }
  else
    return 0;
}
//--------------------------------------------------------------------
TMPL Polynomial shift (const Polynomial& p, const C& c) 
{
  Polynomial r(p); UPOLDSE::shift (r,c); return r;
}
//--------------------------------------------------------------------
TMPL Polynomial scale (const Polynomial& p, const C& c) 
{
  Polynomial r(p.size(),AsSize()); UPOLDSE::scale(r,p,c); return r;
}
//--------------------------------------------------------------------
TMPL Polynomial reverse(const Polynomial& p) 
{
  Polynomial r(p);
  UPOLDSE::reverse(r,r.size());
  return r;
}
//--------------------------------------------------------------------
TMPL inline Polynomial square (const Polynomial& p) {return p*p;}
//====================================================================
// THE ARITHMETIC OPERATORS
//====================================================================
#ifndef UPOL_WITH_TEMP_EXP
//--------------------------------------------------------------------
/// Addition of polynomials.
TMPL inline Polynomial operator+(const Polynomial & a, const Polynomial & b)
{
  Polynomial result(std::max(degree(a),degree(b))+1, AsSize());
  using UPOLDSE::add; add(result.rep(),a.rep(),b.rep());
  using UPOLDSE::checkdegree; checkdegree(result.rep());
  return result;
}
//----------------------------------------------------------------------
/// Substraction of polynomials.
TMPL
inline Polynomial
  operator-(const Polynomial & a, const Polynomial & b)
{

  Polynomial result(std::max(degree(a),degree(b))+1,AsSize());
  using UPOLDSE::sub; sub(result.rep(),a.rep(),b.rep());
  using UPOLDSE::checkdegree; checkdegree(result.rep());
  return result;
}
//----------------------------------------------------------------------}
/// Unary substraction.
TMPL
inline Polynomial operator-(const Polynomial & a)
{
  Polynomial result(degree(a)+1,AsSize());
  using UPOLDSE::mul_ext; mul_ext(result.rep(),a.rep(),-1);
  return result;
}
//----------------------------------------------------------------------
/// Product of polynomials.
TMPL inline
Polynomial operator*(const Polynomial & a, const Polynomial & b)
{
  Polynomial result(std::max(degree(a)+degree(b)+1,0),AsSize());
  using UPOLDSE::mul;         mul(result.rep(),a.rep(),b.rep());
  using UPOLDSE::checkdegree; checkdegree(result.rep());
  return result;
}
//----------------------------------------------------------------------
/// Scalar addition.
TMPL inline
Polynomial operator+(const Polynomial & a,
		       const typename Polynomial::value_type & c)
{
  if(degree(a) >= 0)
    {
      Polynomial r(a);
      using UPOLDSE::add_cst; add_cst(r.rep(),c);
      return r;
    }
  else
    return Polynomial(c,1);
}
//----------------------------------------------------------------------
/// Scalar addition.
TMPL inline
Polynomial operator+(const typename Polynomial::value_type & c,
		       const Polynomial & a)
{
  return a+c;  
}
//----------------------------------------------------------------------
/// Scalar substraction
TMPL inline
Polynomial operator-(const Polynomial & a,
		       const typename Polynomial::value_type & c)
{
  typedef typename Polynomial::value_type coeff_t;
  if(degree(a) >= 0)
    {
      Polynomial r(a);
      using UPOLDSE::sub_cst; sub_cst(r.rep(),c);
      return r;
    }
  else
    {
    }
  return Polynomial(coeff_t(-c),1);
}
//----------------------------------------------------------------------
/// Scalar addition.
TMPL inline
Polynomial operator-(const typename Polynomial::value_type & c,
		        const Polynomial & a)
{
  return  (-a)+c;   
}
//----------------------------------------------------------------------
/// Scalar multiplication.
TMPL inline
Polynomial operator*(const Polynomial & a,
		       const typename Polynomial::value_type & c)
{
  Polynomial result(degree(a)+1,AsSize());
  using UPOLDSE::mul_ext;
  mul_ext(result.rep(),a.rep(),c);
  checkdegree(result.rep());
  return result;
}

//----------------------------------------------------------------------
/// Scalar multiplication.
TMPL inline
Polynomial operator*(const typename Polynomial::value_type & c,
		       const Polynomial & a)
{
  return a*c;
}


//----------------------------------------------------------------------
/// Quotient of polynomials.
TMPL  inline
Polynomial operator/(const Polynomial & a, const Polynomial & b)
{
  assert(b!=0);
  Polynomial result(0), tmp(a);
  UPOLDSE::div_rem(result,tmp,b);
  return result;
}
TMPL  inline
Polynomial quo(const Polynomial & a, const Polynomial & b)
{
  return a/b;
}
//----------------------------------------------------------------------
/// Quotient of a polynomial by a scalar.
TMPL  inline
Polynomial operator/(const Polynomial & a,
		       const typename Polynomial::value_type &  c)
{
  assert(c!=0);
  Polynomial result(degree(a)+1,AsSize());
  using namespace UPOLDSE;
  div_ext(result.rep(),a.rep(),c);
  checkdegree(result.rep());
  return result;

}

//----------------------------------------------------------------------
/// Remainder of polynomials.
TMPL  inline
Polynomial operator%(const Polynomial & a, const Polynomial & b)
{
  assert(b!=0);
  Polynomial result(a);
  result %= b;
  return result;
}

/// Remainder of polynomials.
TMPL  inline
Polynomial rem(const Polynomial & a, const Polynomial & b)
{
  return a%b;
}

TMPL  inline
Polynomial operator%(const Polynomial & a, const C & b)
{
  assert(b!=0);
  Polynomial result(a);
  result %= b;
  return result;
}
//----------------------------------------------------------------------
/// Pseudo-remainder of polynomials.
TMPL  inline
Polynomial prem(const Polynomial & a, const Polynomial & b)
{
  assert(b!=0);
  Polynomial r(a), q;
  EUCLIDEAN::pseudo_div_rem(q,r,b);
  return r;
}
//----------------------------------------------------------------------
/// Power of a polynomial.
TMPL  inline
Polynomial operator^(const Polynomial & a, const unsigned int & n)
{
  Polynomial result(1);
  //  ALGEBRA::power(result,a,n);
  for(unsigned int i=0; i< n; i++) result *= a;
  return result;
}
#else
//======================================================================
/// Addition of polynomials.
TMPL
inline VAL<OP<'+',Polynomial,Polynomial > >
  operator+(const Polynomial & a, const Polynomial & b)
{
  return VAL<OP<'+',Polynomial,Polynomial > >
    (OP<'+',Polynomial,Polynomial >(a,b));
}
//----------------------------------------------------------------------
/// Substraction of polynomials.
TMPL
inline VAL<OP<'-',Polynomial,Polynomial > >
  operator-(const Polynomial & a, const Polynomial & b)
{
  return VAL<OP<'-',Polynomial,Polynomial > >
    (OP<'-',Polynomial,Polynomial >(a,b));
}
//----------------------------------------------------------------------}
///Unary substraction.
TMPL
inline VAL<OP<'-',Polynomial,Polynomial > >
  operator-(const Polynomial & a)
{
  return VAL<OP<'-',Polynomial,Polynomial > >
    (OP<'-',Polynomial,Polynomial >(Polynomial(0),a));
}
//----------------------------------------------------------------------
/// Product of polynomials.
TMPL  inline
VAL<OP<'*',Polynomial,Polynomial > >
  operator*(const Polynomial & a, const Polynomial & b)
{
  return VAL<OP<'*',Polynomial,Polynomial > >
   (OP<'*',Polynomial,Polynomial >(a,b));
}
//----------------------------------------------------------------------
/// Scalar multiplication.
TMPL inline
VAL<OP<'.',Polynomial,typename Polynomial::value_type> >
  operator*(const Polynomial & a, const typename Polynomial::value_type & c)
{
  typedef typename Polynomial::value_type C;
  return VAL<OP<'.',Polynomial,C> >(OP<'.',Polynomial,C>(a,c));
}
//----------------------------------------------------------------------
/// Quotient of polynomials.
TMPL  inline
VAL<OP<'/',Polynomial,Polynomial > >
  operator/(const Polynomial & a, const Polynomial & b)
{
  return VAL<OP<'/',Polynomial,Polynomial > >
   (OP<'/',Polynomial,Polynomial >(a,b));
}
//----------------------------------------------------------------------
/// Remainder of polynomials.
TMPL  inline
VAL<OP<'%',Polynomial,Polynomial > >
  operator%(const Polynomial & a, const Polynomial & b)
{
  return VAL<OP<'%',Polynomial,Polynomial > >
   (OP<'%',Polynomial,Polynomial >(a,b));
}
//----------------------------------------------------------------------
/// Power of a polynomial.
TMPL  inline
VAL<OP<'^',Polynomial,unsigned int> >
  operator^(const Polynomial & a, const unsigned int & n)
{
  return VAL<OP<'^',Polynomial,unsigned int> >
   (OP<'^',Polynomial,unsigned int>(a,n));
}
//======================================================================
// ASSIGNEMENTS
//======================================================================
namespace let {
  template<class C, class R> inline
  void assign(Polynomial & p,
	      const OP<'+',Polynomial, Polynomial >& M)
  {
    assert(&(p.rep())!=&(M.op1.rep())); 
    assert(&(p.rep())!=&(M.op2.rep()));
    p.rep().resize(std::max(degree(M.op1),degree(M.op2))+1);
    using namespace UPOLDSE;
    init(p.rep(),0);
    add(p.rep(),M.op1.rep(),M.op2.rep());
    checkdegree(p.rep());
  }


  template<class C, class R,class X> inline
  void assign(Polynomial & p,
	      const OP<'+',Polynomial, VAL<X> >& M)
  {
    assert(&(p.rep())!=&(M.op1.rep())); 
    // assert(&(p.rep())!=&(M.op2.rep()));
    Polynomial tmp=M.op2;
    p.rep().resize(std::max(degree(M.op1),degree(tmp))+1);
    using namespace UPOLDSE;
    init(p.rep(),0);
    add(p.rep(),M.op1.rep(),tmp);
    checkdegree(p.rep());
  }

  template<class C, class R,class X> inline
  void assign(Polynomial & p,
	      const OP<'+',VAL<X>,Polynomial >& M)
  {
    assert(&(p.rep())!=&(M.op2.rep())); 
    Polynomial tmp=M.op2;
    p.rep().resize(std::max(degree(M.op2),degree(tmp))+1);
    using UPOLDSE::init;  init(p.rep(),0);
    using UPOLDSE::add;   add(p.rep(),tmp,M.op2.rep());
    using UPOLDSE::checkdegree; checkdegree(p.rep());
  }
  //----------------------------------------------------------------------
  template<class C, class R> inline
  void assign(Polynomial & p,
	      const OP<'-',Polynomial, Polynomial >& M)
  {
    assert(&(p.rep())!=&(M.op1.rep())); assert(&(p.rep())!=&(M.op2.rep()));
    p.rep().resize(std::max(degree(M.op1),degree(M.op2))+1);
    using namespace UPOLDSE;
    init(p.rep(),0);
    sub(p.rep(),M.op1.rep(),M.op2.rep());
    checkdegree(p.rep());
  }
  //----------------------------------------------------------------------
  template<class C, class R> inline
  void assign(Polynomial & p,
	      const OP<'*',Polynomial, Polynomial >& M)
  {
    using namespace UPOLDSE;
    if( &(p.rep())==&(M.op1.rep()) || &(p.rep())==&(M.op2.rep()) )
      {
	Polynomial tmp;
	tmp.rep().resize(std::max(degree(M.op1)+degree(M.op2)+1,0));
	init(tmp.rep(),0);
	mul(tmp.rep(),M.op1.rep(),M.op2.rep());
	p=tmp;
      }
    else
      {
	p.rep().resize(std::max(degree(M.op1)+degree(M.op2)+1,0));
	init(p.rep(),0);
	mul(p.rep(),M.op1.rep(),M.op2.rep());
      }
  }
  //----------------------------------------------------------------------
  template<class C, class R> inline
  void assign(Polynomial & p,
	      const OP<'.',Polynomial,typename R::value_type>& M)
  {
    p.rep().resize(degree(M.op1)+1);
    using namespace UPOLDSE;
    mul_scalar(p.rep(),M.op1.rep(),M.op2);
  }
  //----------------------------------------------------------------------
  template<class C, class R> inline
  void assign(Polynomial & p,
	      const OP<'/',Polynomial,Polynomial >& M)
  {
    assert(M.op2!=0);
    Polynomial top1(M.op1);
    UPOLDSE::div_rem(p,top1,M.op2);
  }
  //----------------------------------------------------------------------
  template<class C, class R> inline
  void assign(Polynomial & r,
	      const OP<'%',Polynomial,Polynomial >& M)
  {
    assert(M.op2!=0);
    r= M.op1;
    Polynomial tmp;
    UPOLDSE::div_rem(tmp,r,M.op2);
  }
  //--------------------------------------------------------------------
  template<class C, class R> inline
  void assign(Polynomial & r,
	      const OP<'^',Polynomial,unsigned int> M)
  {
    r=M.op1;
    for(unsigned int i=1;i<M.op2;i++) r*=M.op1;
  }
  
} // namespace let
//======================================================================
#endif //UPOL_WITH_TEMP_EXP
//======================================================================
namespace let 
{
  //--------------------------------------------------------------------
  /** Assign a univariate polynomial with coefficients of type @D@ and
      internal representation @S@ to a polynomial with coefficients of type
      @C@. It uses the function @let::assign@ on the coefficients.
  **/
  template<class C, class R, class D, class S>  inline
  void assign(Polynomial & u, const upoldse<D,S> & P)
  {
    //    using let::assign_x;
    u.resize(degree(P)+1);
    for(int k=0; k< u.size();++k) let::assign(u[k],P[k]);
  }
  //--------------------------------------------------------------------
  /** Similar to the previous @assign@ function, which is used for this 
      convertion. 
  **/
  template<class POL, class C, class S>
    POL convert(const upoldse<C,S> & P, meta::As<POL>)
  {
    using let::assign;
    POL u(degree(P)+1, AsSize());
    assign(u,P);
    return u;
  }
}
//======================================================================
// INPUT OUTPUT
//======================================================================
/// Output operator.
TMPL inline
std::ostream & operator<<(std::ostream & os, const Polynomial & p)
{
  using UPOLDSE::print;
  return print(os,p.rep(),Polynomial::variable);
}
//----------------------------------------------------------------------
/// Input operator. The variable is @x@ or @x0@. Ended by a @;@.
TMPL
std::istream & operator>>(std::istream & is, Polynomial & P);
//====================================================================
template <class C, class R, class X>
int SignAt(const Polynomial& p, const X& x)
{
  return sign(p(x));
}
//====================================================================
template<class C, class R>
Polynomial::upoldse(int n)
{
  if(n==0)
    resize(0);
  else
    {
      rep().resize(1);
      rep()[0]=(C)n;
      checkdegree(rep());
    }
}
//----------------------------------------------------------------------
/// Construct the monomial \f$ c x^{n} \f$.
template<class C, class R>
Polynomial::upoldse(value_type c, size_type n)
{
  using UPOLDSE::set_monomial;  set_monomial(rep(),c,n);
}
//----------------------------------------------------------------------
/** Construct the polynomial, which coefficients are @n@ values of the array
    @v@, in the internal representation basis.
**/
template<class C, class R>
Polynomial::upoldse(size_type n, value_type * v)
{
  rep().resize(n);
  iterator it=rep().begin(), iv=v;
  for(unsigned int i=0;i<n;i++) *it++=*iv++;
  using UPOLDSE::checkdegree;  checkdegree(rep());
}
//----------------------------------------------------------------------
/** Construct the polynomial, which coefficients are the values from iterator 
    position @b@ to @e@, in the internal representation basis.
**/
template<class C, class R>
Polynomial::upoldse(iterator b, iterator e)
{
  rep().resize(std::distance(b,e));
  iterator it=rep().begin(), iv=b;
  while(iv != e) *it++=*iv++;
  using UPOLDSE::checkdegree;  checkdegree(rep());
}
//----------------------------------------------------------------------
/** Constructor of a univariate polynomial from a string.
Example: @Polynomial("x0^2+x0^3-1")@.
**/
TMPL
Polynomial::upoldse(const char * s) {
  rep().resize(0);
  VECTOR::set_cst(rep(),0);
  int n = strlen(s);
  if(s[n-1]==';')
    {
      synaps_input = new char[n+1];
      memcpy(synaps_input,s,n); synaps_input[n]='\0';
      synaps_inputlim = synaps_input + n;
    }
  else
    {
      synaps_input = new char[n+2];
      memcpy(synaps_input,s,n); synaps_input[n]=';'; synaps_input[n+1]='\0';
      synaps_inputlim = synaps_input + n+1;
    }
  synaps_inputptr = synaps_input;

  self_t m(1);
  bool sign  = true;
  bool first = true;
  int Ask=0;
  for (;;) {
    Ask = yylex();
    if (Ask == COEF) { /* Monome constant */
      value_type c;//(yylval);
      let::assign(c,yylval);
      m *= c;

    }
    else if (Ask == XID) { /* Variable */
      char * p = strchr(yylval,'^');
      if (p == NULL){
	m = self_t(m[0],1);
      } else {
	p++;
	//	PRINT_DEBUG(" ^");
	m = self_t(m[0],atoi(p));
	//	PRINT_DEBUG(m);
      }
    }
    else if (Ask == MULT); /* Multiplication */
    else if (Ask == SUM) { /* Addition */
      if (sign) {
	// 	PRINT_DEBUG("+");
	// 	PRINT_DEBUG(m);
	// 	PRINT_DEBUG("r="<<*this);
	*this += m;
	//	PRINT_DEBUG("r="<<*this);
	m = self_t(1);
      }
      else {
	*this -= m;
	sign = true;
	m = self_t(1);
      }
    }
    else if (Ask == MINUS) { /* Soustraction */
      if (first)
	sign = false;
      else if (sign) {
	*this += m;
	sign = false;
	m = self_t(1);
      }
      else {
	*this -= m;
	m = self_t(1);
      }
//      PRINT_DEBUG("r="<<*this);
    }
    else if (Ask == TERMINATE) { /* ; */
      if (sign)
	*this += m;
      else
	*this -= m;
      break;
    }
    first = false;
  }
  delete[] synaps_input;
}
//----------------------------------------------------------------------
/// Equality test to 0,1,\dots
template<class C, class R> inline
bool Polynomial::operator==(int n) const
{
  if(n==0 && degree(*this)<0)
    return(true);
  else{
    return( (degree(*this)==0) && ((*this)[0]==n) );
  }
}
//----------------------------------------------------------------------
/// Equality test to 0,1,\dots
template<class C, class R> inline
bool Polynomial::operator==(const Polynomial & p) const
{
  return &(rep())== &p.rep();
}
//----------------------------------------------------------------------
template<class C, class R> inline
Polynomial & Polynomial::operator+= (const Polynomial & P)
{
  using UPOLDSE::add; add(rep(),P.rep());
  using UPOLDSE::checkdegree; checkdegree(rep());
  return *this;
}
//--------------------------------------------------------------------
template<class C, class R> inline
Polynomial& Polynomial::operator+=(const C& c)
{
  if(degree(*this) >= 0)
    {
      using UPOLDSE::add_cst; add_cst(rep(),c);
    }
  else
    *this=Polynomial(c,1);
  return *this;
}
//----------------------------------------------------------------------
template<class C, class R> inline
Polynomial & Polynomial::operator-= (const Polynomial & P)
{
  rep().resize(std::max(degree(*this),degree(P))+1);
  using UPOLDSE::sub; sub(rep(),P.rep());
  using UPOLDSE::checkdegree; checkdegree(rep());
  return *this;
}
//--------------------------------------------------------------------
template<class C, class R> inline
Polynomial& Polynomial::operator-=(const C& c)
{
  if(degree(*this) >= 0)
    {
      using UPOLDSE::add_cst; add_cst(rep(),-c);
    }
  else
    *this=Polynomial(c,1);
  return *this;
}
//----------------------------------------------------------------------
template<class C, class R> inline
Polynomial & Polynomial::operator*= (const Polynomial & P)
{
  //  using namespace UPOLDSE;  mul(rep(),P.rep());
  (*this)= (*this) * P;
  return (*this);
}
//----------------------------------------------------------------------
template<class C, class R> inline
Polynomial & Polynomial::operator*= (const typename Polynomial::value_type & c)
{
  if (c != 0)
    {
      using UPOLDSE::mul_ext; mul_ext(rep(),rep(),c);
    }
  else 
    {
      rep().resize(0);// rep()[0]=0;
    }
  return *this;
}
//----------------------------------------------------------------------
/// Inplace division by a scalar.
template<class C, class R> inline
Polynomial & Polynomial::operator/= (const typename Polynomial::value_type & c)
{
  using namespace VECTOR; div_ext(rep(),rep(),c);
  return *this;
}
//----------------------------------------------------------------------
/// Inplace quotient in the Euclidian division.
TMPL inline
Polynomial & Polynomial::operator/= (const Polynomial & Q) {
  using namespace UPOLDSE;
  assert(Q!=0);
  if (*this ==0) return *this;
  if (&rep()==&Q.rep()) {*this=1;return *this;}
  Polynomial quo(0);
  UPOLDSE::div_rem(quo,*this,Q);
  *this=quo;
  return *this;
}
//-----------------------------------------------------------------------------
/// Inplace remainder in the Euclidian division.
TMPL inline
Polynomial & Polynomial::operator%= (const Polynomial & Q)
{
  assert(Q!=0);
  if (*this ==0) return *this;
  if (&rep()==&Q.rep()) {*this=0;return *this;}
  Polynomial quo;
  UPOLDSE::div_rem(quo,*this,Q);
  return *this;
}

template<class C, class R> inline 
Polynomial & Polynomial::operator%=( const C & c )
{
  using namespace UPOLDSE;
  coeff_modulo(rep(),c);
  checkdegree(rep());
  return *this;
};

//-----------------------------------------------------------------------------
/** Evaluation of a polynomial, using Horner scheme. The output value is
    let by the type of @c@.    **/
TMPL template<class T> inline
T Polynomial::operator()(const T & c) const
{
  using UPOLDSE::eval;
  return eval(rep(),c);
}
//----------------------------------------------------------------------
// Input operator. The variable is @x@ or @x0@. Ended by a @;@.
TMPL inline
std::istream & operator>>(std::istream & is, Polynomial & P)
{
  std::string s, ss = "";
  int p;
  for (;;) {
    is >> s;
    p = s.find(';');
    if (p == -1)
      ss += s;
    else {
      ss += s.substr(0,p+1);
      break;
    }
  }
  P = Polynomial(ss.c_str());
  return is;
}

namespace UPOLDSE {
  TMPL std::ostream & print_as_coeff(std::ostream & os, const Polynomial & p)
  {
    return os<<"+("<<p<<")";
  }
}

TMPL inline bool with_plus_sign(const Polynomial & p) {return true;}

template<class C, class R>
void checkdegree( Polynomial & p )
{
  using namespace UPOLDSE;
  checkdegree(p);
};

//template<class C, class R> 
//Polynomial abs( const Polynomial & p ) { if ( p[degree(p)] < 0 ) return -p; };



template<class C> struct ispolynomial { typedef meta::false_t T; };
template<class C, class R > struct ispolynomial< Polynomial > { typedef meta::true_t T; };
//template<class C>  struct base_ring

//======================================================================
} //namespace mmx
//====================================================================
#undef TMPL
#undef Polynomial
#endif // SYNAPS_UPOL_UPOLDSE_H

