/*********************************************************************
*   This file is part of the source code of the SYNAPS environment.  *
*   Author(s): B. Mourrain, GALAAD, INRIA                            *
**********************************************************************/
#ifndef synaps_linalg_vectdse_H
#define synaps_linalg_vectdse_H
//----------------------------------------------------------------------
#include <sstream>
#include <vector>
#include <subdivix/shared_object.hpp>
#include <subdivix/texp.hpp>
//#include <subdivix/Range.hpp>
#include <subdivix/seq1d.hpp>
#include <subdivix/let.hpp>
#include <subdivix/linalg_rep1d.hpp>
#include <subdivix/VECTOR_m.hpp>
#include <subdivix/MATRDSE_m.hpp>
//----------------------------------------------------------------------
namespace mmx {
//----------------------------------------------------------------------
/** \file vectdse.hpp 
    \brief Dense vectors

    Dense vectors with the usual operations of a vector space.
**/
/**
The standard class of vectors.
The container type @R@ should provide the following signature:
\code
  typedef size_type;   // type of the index.
  typedef value_type;  // type of the coefficients.
  typedef iterator;    // type of the iterator.
  R(size_type n);
  R(size_type n, value_type* t);
  R(iterator b,iterator e);
  R& R::operator=(const R &)
  size_type    R::size()
  void R::resize(size_type);
  iterator       R::begin();
  const_iterator R::begin() const;
  iterator       R::end();
  const_iterator R::end() const;
  value_type & R::operator[] (size_type i);
  value_type   R::operator[] (size_type i) const;
\endcode

The constructor @R(size_type i)@ allocate the memory but nothing is
supposed to be known about the allocated values.
The constructor @R(iterator_t b, iterator_t e)@ allocates the memory
and fill it with the values from the iterator from @b@ to @e@.
The constructor @R(size_type n, const char* str)@ build the array
of size @n@ from the  string @str@.
The function @void R::resize(size_type s)@ reallocate the memory if
necessary and copies the elements of @*this@,
or use the current memory space if @s@ is less that the
actual space.
**/
template <class C, class R = linalg::rep1d<C> >
struct vectdse   {

  shared_object<R>  data;
  R &       rep()       {return (*data);}
  const R & rep() const {return (*data);}

  typedef typename R::size_type          size_type;
  typedef C                              value_type;
  typedef typename R::iterator           iterator;
  typedef typename R::const_iterator     const_iterator;
  typedef typename R::reverse_iterator   reverse_iterator;
  typedef typename R::const_reverse_iterator const_reverse_iterator;
  typedef R                              rep_type;
  typedef vectdse<C,R>                   self_t;
  
  operator meta::template_expression<self_t>(){ return meta::template_expression<self_t>(*this); };

  vectdse(): data() {}
  vectdse(const R & r):data(r){}
  vectdse(R* r):data(r){}
  vectdse(const shared_object<R> & r):data(r){}
  vectdse(int n) {rep().resize(1); (*this)[0]=C(n);}
  vectdse(size_type s) {rep().resize(s);}
  vectdse(size_type s, const AsSize & mth) {rep().resize(s);}
  vectdse(size_type s, value_type* t) {rep()=R(t,t+s);}
  vectdse(size_type s, const value_type& v)
	  {rep().resize(s); for(unsigned i=0;i<s;i++) (*this)[i]=v;}
  vectdse(iterator b,iterator e) {rep()=R(b,e);}
  vectdse(size_type m, const char * str);

  //  inline operator generic () { return generic(data.body); }

  // Copy constructor
  vectdse(const self_t & r): data(r.data) {}
  template<class X>   vectdse( const meta::template_expression<X> & x );
  template<class C1,class R1> vectdse( const vectdse<C1,R1> & v );
  template<class X>   self_t & operator=(const X & exp);
  
  iterator               begin()        {return rep().begin();}
  iterator               end()          {return rep().end();}
  const_iterator         begin()  const {return rep().begin();}
  const_iterator         end()    const {return rep().end();}
  reverse_iterator       rbegin()       {return rep().rbegin();}
  reverse_iterator       rend()         {return rep().rend();}
  const_reverse_iterator rbegin() const {return rep().rbegin();}
  const_reverse_iterator rend  () const {return rep().rbegin();}

  self_t & operator =(const self_t & V) { using let::assign; assign(data,V.data); return *this; }
  self_t & operator+=(const self_t & v)     { add(*this,v); return *this; };
  self_t & operator-=(const self_t & v)     { sub(*this,v); return *this; };
  self_t & operator*=(const value_type & c) { mul(*this,c); return *this; };
  self_t & operator/=(const value_type & c) { div(*this,c); return *this; };
  template<class X> 
  self_t & operator+=(const X & x) { *this = *this + x; return *this; };
  template<class X>
  self_t & operator-=(const X & x) { *this = *this - x; return *this; };
  template<class X>
  self_t & operator*=(const X & x) { *this = *this * x; return *this; };
  template<class X>
  self_t & operator/=(const X & x) { *this = *this / x; return *this; };
  
  value_type & operator[] (size_type i)       {return rep()[i];}
  value_type & set        (size_type i)       {return rep()[i];}
  const value_type & operator[] (size_type i) const {return rep()[i];}
  value_type   coeff      (size_type i) const {return rep()[i];}

  bool operator== (const self_t & v) const
  { 
    using VECTOR::equal;
    return equal(this->rep(),v.rep());
  };
  
  bool operator!= (const self_t & v) const { return !(*this==v); }

  //  vectdse<C,seq1d<value_type,iterator> > operator[](const Range & I);
  
  size_type size() const {return rep().size();}
  self_t resize(int n)  {rep().resize(n); return *this;}

};
//======================================================================
template<class C, class R> inline 
const C& read(const vectdse<C,R>& m, unsigned i) { return m[i]; }

template<class C, class R>
bool eq (const vectdse<C,R>& v1, const vectdse<C,R>& v2) {return v1==v2;}
template<class C, class R>
bool neq(const vectdse<C,R>& v1, const vectdse<C,R>& v2) {return v1!=v2;}

template<class C, class R>
unsigned hash (const vectdse<C,R>& v) 
{
  register unsigned i, h= 214365, n= v.size();
  for (i=0; i<n; i++)
      h= (h<<1) ^ (h<<5) ^ (h>>27) ^ hash(v[i]);
  return h;
}
template<class C, class R>
unsigned soft_hash (const vectdse<C,R>& m) {return hash(m);}

template<class C, class R>
typename vectdse<C,R>::const_iterator iterate(const vectdse<C,R>& m) {return m.begin();}

//======================================================================
/*{
 The @value_type@ is the type of the coefficients. The @size_type@ is the
 type of the indices used to access the elements of the vectors.  The
 @iterator@ is the type of the forward iterators on the vectors.  The
 internal representation type @R@ is @rep_type@.
}*/
namespace let
{

  template<class A>
  void assign( std::pair<A,A> & a, const vectdse<A> & b )
    {
      assign(a.first,b[0]);
      assign(a.second ,b[1]);
    };

  template<class Ca,class Ra,class Cb, class Rb> inline
  void assign( vectdse<Ca,Ra> & a, const vectdse<Cb,Rb> & b )
  {
    using VECTOR::assign;
    assign(a.rep(),b.rep());
  };
  //----------------------------------------------------------------------
  template<class Ca, class Ra,class Cb, class  Rb> inline
  void copy(vectdse<Ca,Ra> & v, vectdse<Cb,Rb>* t)
  {
    using VECTOR::assign;
    assign(v.rep(), (*t).rep());
  }
  
  template<class Ca, class Ra, class Cb, class Rb> inline
  void copy( vectdse<Ca,Ra> & a, const vectdse<Cb,Rb> & b )
  {
    using VECTOR::assign; assign(a.rep(),b.rep());
  };

  template<class Ca, class Ra, class TEXP> inline
  void copy( vectdse<Ca,Ra> & a, const TEXP & b )
  {
    using let::assign; assign(a,b);
  }
}
//======================================================================
#include "vectdse_operators.hpp"

//======================================================================
// INPUT OUTPUT
//======================================================================
/// Output operator.
template<class C, class R>
std::ostream & operator<<(std::ostream & os, const vectdse<C,R> & V)
{
  using namespace VECTOR;
  return print(os,V.rep());
}
//----------------------------------------------------------------------
/**
  Input operator for standard matrices. The input format is of the form
  @s c0 c1 ...@ where @s@ is the number of elements.
**/
template<class C, class R>  inline
std::istream & operator>>(std::istream & is, vectdse<C,R> & V)
{
  using namespace VECTOR;
  return read(is,V);
}

//--------------------------------------------------------------------
/**
   Constructor  for a size @m@ and a string containing the elements.
   Example: @vectdse(3,"1 2 3")@.
**/
template <class C, class R> 
vectdse<C,R>::vectdse(size_type m, const char * str)
{
  rep()= R(m);
  std::istringstream ins(str);
  for(size_type i=0; i< m && ins.good(); i++)	ins >> rep()[i];
}

//----------------------------------------------------------------------
// template <class C, class R> 
// vectdse<C,seq1d<typename vectdse<C,R>::value_type,typename vectdse<C,R>::iterator> > 
//    vectdse<C,R>::operator[](const Range & I) 
// {
//   using namespace VECTOR;
//   typedef typename vectdse<C,R>::value_type value_type;
//   typedef typename vectdse<C,R>::iterator   iterator;
//   typedef seq1d<value_type,iterator>     rep_type;
      
//   typename vectdse<C,R>::size_type n=I.i2-I.i1+1;
//   return vectdse<C,rep_type>(rep_type(n,create(iterator(),rep(),I.i1),
// 				      create(iterator(),rep(),I.i2+1)));
// }
//----------------------------------------------------------------------
template<class C, class R>
template<class X> inline
vectdse<C,R>::vectdse( const meta::template_expression<X> & x )
{
  using let::assign;  assign(*this,x);
};

template<class C, class R>
template<class X> inline
vectdse<C,R> & vectdse<C,R>::operator= ( const X & x )
{
  using let::assign;  assign(*this,x); return *this;
};

template<class C, class R> 
template<class C1,class R1>  vectdse<C,R>::vectdse( const vectdse<C1,R1> & v )
{
  let::assign(*this,v);
};


namespace meta
{
  template<class Y, class C> struct instanceof< vectdse<C>, Y  >  { typedef vectdse<Y> T; };
};
} //namespace mmx
//====================================================================
#endif //synaps_linalg_vectdse_H

