
/******************************************************************************
* MODULE     : simplify_regroup.hpp
* DESCRIPTION: Limited factorization and polynomial regrouping
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __SIMPLIFY_REGROUP__HPP
#define __SIMPLIFY_REGROUP__HPP
#include <symbolix/symbolic.hpp>
#include <symbolix/wrap_sum.hpp>
#include <symbolix/wrap_product.hpp>
#include <basix/vector_sort.hpp>
namespace mmx {

template<typename V>
struct simplify_regroup: public V {
  typedef typename V::S S;
  typedef typename V::W W;
  typedef typename V::sum sum;
  typedef typename V::product product;
  using V::ZERO;
  using V::ONE;
  using V::expression;
  using V::is_scalar;
  using V::term;
  using V::power;
  using V::div_expand_left;
  using V::size;

  //----- Counting occurrences ------------------------------------------------

  static table<nat, S>
  count_power (const sum& s) {
    table<nat,S> count (0);
    for (nat i=0; i<N(s); i++) {
      S sc= co (s, i);
      S sm= mo (s, i);
      if (is<product> (sm)) {
	product p= as<product> (sm);
	for (nat j=0; j<N(p); j++) {
	  S pc= co (p, j);
	  S pm= mo (p, j);
	  if (is<sum> (pc)) {
	    sum e= as<sum> (pc);
	    for (nat k=0; k<N(e); k++) {
	      S et= term (mo (e, k), co (e, k));
	      count [power (pm, et)] ++;
	    }
	  }
	  else count [power (pm, pc)] ++;
	}
      }
      else if (sm != ONE) count [sm] ++;
      if (sc != ONE) count [sc] ++;
    }
    return count;
  }

  //----- Regrouping ----------------------------------------------------------

  static pair<S,S>
  common_quasi_factor (const sum& x) {
    //cout << "x= " << *x << "\n";
    table<nat, S> count= count_power (x);
    //cout << "c= " << count << "\n";
    S c= ONE;
    vector<S> v;
    for (iterator<S> it= entries (count); busy (it); ++it)
      if (3*count[*it] > 2*N(*x)) {
	if (is_scalar (*it)) c= c * (*it);
	else v << (*it);
      }
    //cout << "p= " << pair<S,S> (c, big_mul (v)) << "\n";
    return pair<S,S> (c, big_mul (v));
  }

  static S
  fewnomial_base (const S& x) {
    if (is<product> (x)) {
      product p= as<product> (x);
      ASSERT (N(p) == 1, "internal error: power expected");
      S b= mo (p, 0);
      S e= co (p, 0);
      if (is_scalar (e)) return b;
      else if (is<sum> (e)) {
	sum s= as<sum> (e);
	ASSERT (N(s) == 1, "internal error: term expected");
	return power (b, mo (s, 0));
      }
      else return x;
    }
    else return x;
  }

  static S
  fewnomial_exponent (const S& x) {
    if (is<product> (x)) {
      product p= as<product> (x);
      ASSERT (N(p) == 1, "internal error: power expected");
      S b= mo (p, 0);
      S e= co (p, 0);
      if (is_scalar (e)) return e;
      else if (is<sum> (e)) {
	sum s= as<sum> (e);
	ASSERT (N(s) == 1, "internal error: term expected");
	return co (s, 0);
      }
      else return ONE;
    }
    else if (x == ONE) return ZERO;
    else return ONE;
  }

  static S
  fewnomial_extract (const S& x, const S& var) {
    if (is<product> (x)) {
      product p= as<product> (x);
      for (nat i=0; i<N(p); i++) {
	S b= mo (p, i);
	S e= co (p, i);
	if (is<sum> (e)) {
	  sum s= as<sum> (e);
	  for (nat j=0; j<N(s); j++) {
	    S m= mo (s, j);
	    S c= co (s, j);
	    if (power (b, m) == var)
	      return power (b, term (m, c));
	  }
	}
	else if (b == var) return power (b, e);
	else if (p[i] == var) return var;
      }
      return ONE;
    }
    else if (x == var) return x;
    else return ONE;
  }

  static S
  best_fewnomial_variable (const sum& x) {
    table<nat, S> count= count_power (x);
    table<nat, S> score (0);
    //cout << "sum  = " << *x << "\n";
    for (iterator<S> it= entries (count); busy (it); ++it) {
      S mon= *it;
      if (is_scalar (mon)) continue;
      S var= fewnomial_base (mon);
      //cout << "  var= " << dump (var) << ", " << size (var) << "\n";
      score[var] += (2*count[mon] - 1);
    }
    //cout << "score= " << score << "\n";
    S best= ONE;
    nat best_score= 0;
    for (iterator<S> it= entries (score); busy (it); ++it)
      if (score[*it] > 2 && size (*it) * score[*it] > best_score) {
	best= *it;
	best_score= size (*it) * score[*it];
      }
    //cout << "best = " << best << "\n";
    return best;
  }

  //----- Flattening sums -----------------------------------------------------

  static syntactic
  expression_fewnomial (const sum& s) {
    //cout << "s= " << dump (*s) << "\n";
    nat n= N(s);
    if (n <= 1) return V::expression (s);
    S var= best_fewnomial_variable (s);
    if (var == ONE) return V::expression (s);
    //cout << "var= " << var << "\n";
    table<vector<S>,S> accu;
    for (nat i=0; i<N(s); i++) {
      //cout << "  handle " << s[i] << " in " << var << "\n";
      S f= fewnomial_extract (mo (s, i), var);
      //cout << "    f= " << f << "\n";
      S e= fewnomial_exponent (f);
      //cout << "    e= " << e << "\n";
      S c= s[i] / f;
      //cout << "    c= " << c << "\n";
      if (!contains (accu, e)) accu[e]= vector<S> ();
      accu[e] << c;
    }
    vector<S> expos (entries (accu));
    ASSERT (N (expos) >= 2, "internal error: at least two terms expected");
    //cout << "expos= " << expos << "\n";
    for (nat i=0; i<N(expos); i++)
      ASSERT (is<integer> (expos[i]) || is<rational> (expos[i]) ||
	      is<floating<> > (expos[i]) || is<ball<floating<> > > (expos[i]),
	      "internal error: exponents cannot be ordered");
    sort_leq<gtr_op> (expos);
    //cout << "expos= " << expos << "\n";
    syntactic r= 0;
    for (nat i=0; i<N(expos); i++) {
      syntactic sc= flatten (big_add (accu[expos[i]]));
      syntactic sm= flatten (pow (var, expos[i]));
      bool old= set_frac_flag (true);
      r= r + sc * sm;
      (void) set_frac_flag (old);
    }
    return r;
  }

  static syntactic
  expression (const sum& x) {
    nat n= N(x);
    if (n <= 1) return V::expression (x);
    pair<S,S> f= common_quasi_factor (x);
    if (f.x1 == ONE && f.x2 == ONE) return expression_fewnomial (x);
    S q= div_expand_left (S (x), (f.x1 * f.x2));
    ASSERT (is<sum> (q), "internal error: sum expected");
    syntactic c= flatten (f.x1);
    syntactic s= expression_fewnomial (as<sum> (q));
    syntactic m= flatten (f.x2);
    bool old= set_frac_flag (true);
    syntactic r= c * s * m;
    (void) set_frac_flag (old);
    return r;
  }
};

} // namespace mmx
#endif // __SIMPLIFY_REGROUP__HPP
