
/******************************************************************************
* MODULE     : symbolic_apply.hpp
* DESCRIPTION: Application of operators
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __SYMBOLIC_APPLY__HPP
#define __SYMBOLIC_APPLY__HPP
#include <symbolix/symbolic.hpp>
#include <basix/symbol.hpp>
#include <basix/memoize.hpp>

namespace mmx {

template<typename V>
struct symbolic_apply: public V {
  typedef typename V::S S;
  typedef typename V::W W;
  typedef typename V::compound compound;
  typedef typename V::sum sum;
  typedef typename V::product product;
  typedef typename V::derivative derivative;
  typedef typename V::function function;
  using V::apply;
  using V::PLUS;
  using V::MINUS;
  using V::TIMES;
  using V::OVER;
  using V::SQUARE;
  using V::SQRT;
  using V::POWER;
  using V::EXP;
  using V::LOG;
  using V::COS;
  using V::SIN;
  using V::TAN;
  using V::ARCCOS;
  using V::ARCSIN;
  using V::ARCTAN;
  using V::DERIVATIVE;
  using V::LAMBDA;
  using V::expression;

  //----- Standard functions --------------------------------------------------

  static S apply_plus (const vector<S>& v) {
    return W::add (cdr (v)); }
  static S apply_minus (const vector<S>& v) {
    ASSERT (N(v) == 3, "two arguments expected");
    return W::sub (v[1], v[2]); }
  static S apply_times (const vector<S>& v) {
    return W::mul (cdr (v)); }
  static S apply_over (const vector<S>& v) {
    ASSERT (N(v) == 3, "two arguments expected");
    return W::div (v[1], v[2]); }
  static S apply_square (const vector<S>& v) {
    ASSERT (N(v) == 2, "one argument expected");
    return W::square (v[1]); }
  static S apply_sqrt (const vector<S>& v) {
    ASSERT (N(v) == 2, "one argument expected");
    return W::sqrt (v[1]); }
  static S apply_power (const vector<S>& v) {
    ASSERT (N(v) == 3, "two arguments expected");
    return W::pow (v[1], v[2]); }
  static S apply_exp (const vector<S>& v) {
    ASSERT (N(v) == 2, "one argument expected");
    return W::exp (v[1]); }
  static S apply_log (const vector<S>& v) {
    ASSERT (N(v) == 2, "one argument expected");
    return W::log (v[1]); }
  static S apply_cos (const vector<S>& v) {
    ASSERT (N(v) == 2, "one argument expected");
    return W::cos (v[1]); }
  static S apply_sin (const vector<S>& v) {
    ASSERT (N(v) == 2, "one argument expected");
    return W::sin (v[1]); }
  static S apply_tan (const vector<S>& v) {
    ASSERT (N(v) == 2, "one argument expected");
    return W::tan (v[1]); }
  static S apply_arccos (const vector<S>& v) {
    ASSERT (N(v) == 2, "one argument expected");
    return W::acos (v[1]); }
  static S apply_arcsin (const vector<S>& v) {
    ASSERT (N(v) == 2, "one argument expected");
    return W::asin (v[1]); }
  static S apply_arctan (const vector<S>& v) {
    ASSERT (N(v) == 2, "one argument expected");
    return W::atan (v[1]); }
  static S apply_derivative (const vector<S>& v) {
    ASSERT (N(v) >= 2, "at least one argument expected");
    vector<nat> o= fill<nat> (N(v) - 2);
    for (nat i=0; i<N(v)-2; i++) {
      ASSERT (is<integer> (v[i+2]), "positive integer expected");
      o[i]= as_int (as<integer> (v[i+2]));
    }
    return derivative (v[1], o); }
  static S apply_lambda (const vector<S>& v) {
    ASSERT (N(v) >= 2, "at least one argument expected");
    return function (range (v, 1, N(v) - 1), v [N(v) - 1]); }
  
  typedef S (*nary_function) (const vector<S>&);
  typedef table<nary_function,S> nary_function_table;
  static nary_function_table
  compute_standard_apply () {
    nary_function_table t ((nary_function) NULL);
    t[PLUS]      = apply_plus;
    t[MINUS]     = apply_minus;
    t[TIMES]     = apply_times;
    t[OVER]      = apply_over;
    t[SQUARE]    = apply_square;
    t[SQRT]      = apply_sqrt;
    t[POWER]     = apply_power;
    t[EXP]       = apply_exp;
    t[LOG]       = apply_log;
    t[COS]       = apply_cos;
    t[SIN]       = apply_sin;
    t[TAN]       = apply_tan;
    t[ARCCOS]    = apply_arccos;
    t[ARCSIN]    = apply_arcsin;
    t[ARCTAN]    = apply_arctan;
    // t[DERIVATIVE]= apply_derivative;
    // FIXME: uncommenting yields infinite loop on 'derive (f (x, x^2), x^^2)'
    t[LAMBDA]    = apply_lambda;
    return t;
  }

  //----- Functions f for which f (g) (x) = f (g (x)) ------------------------

  static table<bool,S>
  compute_associative_apply_set () {
    table<bool,S> s (false);
    s[EXP]    = true;
    s[LOG]    = true;
    s[COS]    = true;
    s[SIN]    = true;
    s[TAN]    = true;
    s[ARCCOS] = true;
    s[ARCSIN] = true;
    s[ARCTAN] = true;
    return s;
  }

  //----- Dispatch and arithmetic on functions --------------------------------

  static S
  apply (const vector<S>& v) {
    if (N(v) == 0) return V::apply (v);
    switch (symbolic_type (v[0])) {
    case SYMBOLIC_INTEGER:
    case SYMBOLIC_RATIONAL:
    case SYMBOLIC_FLOATING:
    case SYMBOLIC_BALL:
    case SYMBOLIC_REAL:
      return v[0];
    case SYMBOLIC_LITERAL: {
      static nary_function_table standard_apply=
	compute_standard_apply ();
      if (contains (standard_apply, v[0]))
	return standard_apply[v[0]] (v);
      else return V::apply (v);
    }
    case SYMBOLIC_COMPOUND: {
      static table<bool,S> associative_apply_set=
	compute_associative_apply_set ();
      if (contains (associative_apply_set, v[0][0])) {
	vector<S> largs= cdr (*as<compound> (v[0]));
	vector<S> rargs= cdr (v);
	vector<S> args = fill<S> (N (largs));
	for (nat i=0; i<N(args); i++)
	  args[i]= apply (largs[i], rargs);
	return apply (v[0][0], args);
      }
      return V::apply (v);
    }
    case SYMBOLIC_SUM: {
      vector<S> args= cdr (v);
      sum s= as<sum> (v[0]);
      nat n= N(s);
      vector<S> r= fill<S> (n);
      for (nat i=0; i<n; i++)
	r[i]= co (s, i) * W::apply (cons (mo (s, i), args));
      return big_add (r);
    }
    case SYMBOLIC_PRODUCT: {
      vector<S> args= cdr (v);
      product p= as<product> (v[0]);
      nat n= N(p);
      vector<S> r= fill<S> (n);
      for (nat i=0; i<n; i++)
	r[i]= W::pow (W::apply (cons (mo (p, i), args)), co (p, i));
      return big_mul (r);
    }
    case SYMBOLIC_DERIVATIVE:
      return V::apply (v);
    case SYMBOLIC_FUNCTION: {
      vector<S> args= cdr (v);
      function f= as<function> (v[0]);
      return W::replace (body (f), arguments (f), args);
    }
    default:
      return V::apply (v);
    }
  }

  static inline S apply (const S& f) {
    return apply (vec (f)); }
  static inline S apply (const S& f, const S& a) {
    return apply (vec (f, a)); }
  static inline S apply (const S& f, const S& a, const S& b) {
    return apply (vec (f, a, b)); }
  static inline S apply (const S& f, const S& a, const S& b, const S& c) {
    return apply (vec (f, a, b, c)); }
  static inline S apply (const S& f, const vector<S>& a) {
    return apply (cons (f, a)); }

  //----- Unapply and flattening functions ------------------------------------

  static inline S unapply (const S& f, const vector<S>& vars) {
    return function (vars, f); }

  static syntactic
  expression (const function& f) {
    vector<S> a= arguments (f);
    syntactic b= flatten (body (f));
    nat n= N(a);
    if (n == 1) return mmx::apply (":->", flatten (a[0]), b);
    else {
      vector<syntactic> v= fill<syntactic> (n);
      for (nat i=0; i<n; i++) v[i]= flatten (a[i]);
      return mmx::apply (":->", mmx::apply ("()", v), b);
    }
  }
};

} // namespace mmx
#endif // __SYMBOLIC_APPLY__HPP
