
/******************************************************************************
* MODULE     : symbolic_default.hpp
* DESCRIPTION: Default syntactic operations on symbolic expressions
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __SYMBOLIC_DEFAULT__HPP
#define __SYMBOLIC_DEFAULT__HPP
#include <symbolix/symbolic.hpp>
namespace mmx {

template<typename V>
struct symbolic_default: public V {
  typedef typename V::S S;
  using V::apply;

  static S ZERO;
  static S ONE;
  static S MINUS_ONE;
  static S TWO;
  static S ONE_HALF;

  static S MINUS;
  static S PLUS;
  static S TIMES;
  static S OVER;
  static S SQUARE;
  static S SQRT;
  static S POWER;
  static S MATHE;
  static S EXP;
  static S LOG;
  static S COS;
  static S SIN;
  static S TAN;
  static S ARCCOS;
  static S ARCSIN;
  static S ARCTAN;

  static S DERIVATIVE;
  static S LAMBDA;
  static S ACCESS;
  static S MATHD;
  static S DERIVE;
  static S EQUAL;
  static S FAIL;
  static S UNINIT;

  static inline S neg (const S& s) {
    return apply (MINUS, s); }
  static inline S square (const S& s) {
    return apply (TIMES, s, s); }
  static inline S add (const S& s1, const S& s2) {
    return apply (PLUS, s1, s2); }
  static inline S sub (const S& s1, const S& s2) {
    return apply (MINUS, s1, s2); }
  static inline S mul (const S& s1, const S& s2) {
    return apply (TIMES, s1, s2); }
  static inline S div (const S& s1, const S& s2) {
    return apply (OVER, s1, s2); }

  static inline S sqrt (const S& s) {
    return apply (SQRT, s); }
  static inline S exp (const S& s) {
    return apply (EXP, s); }
  static inline S log (const S& s) {
    return apply (LOG, s); }
  static inline S pow (const S& s1, const S& s2) {
    return apply (POWER, s1, s2); }
  static inline S cos (const S& s) {
    return apply (COS, s); }
  static inline S sin (const S& s) {
    return apply (SIN, s); }
  static inline S tan (const S& s) {
    return apply (TAN, s); }
  static inline S acos (const S& s) {
    return apply (ARCCOS, s); }
  static inline S asin (const S& s) {
    return apply (ARCSIN, s); }
  static inline S atan (const S& s) {
    return apply (ARCTAN, s); }

  // NOTE: we might consider using 'equal' instead of 'eq'
  // in order to get floating and ball comparisons right
  static inline bool equal (const S& s1, const S& s2) {
    return eq_op::op (*s1, *s2); }
  static inline bool unequal (const S& s1, const S& s2) {
    return neq_op::op (*s1, *s2); }
  static inline bool lesseq (const S& s1, const S& s2) {
    return eq_op::op (*s1, *s2); }
  static inline bool less (const S& s1, const S& s2) {
    return neq_op::op (*s1, *s2); }
  static inline bool gtreq (const S& s1, const S& s2) {
    return eq_op::op (*s1, *s2); }
  static inline bool gtr (const S& s1, const S& s2) {
    return neq_op::op (*s1, *s2); }

  static inline S derive (const S& s1, const S& s2) {
    return apply (DERIVE, s1, s2); }
};

template<typename V> typename V::S
symbolic_default<V>::ZERO= typename V::S (0);
template<typename V> typename V::S
symbolic_default<V>::ONE= typename V::S (1);
template<typename V> typename V::S
symbolic_default<V>::MINUS_ONE= typename V::S (-1);
template<typename V> typename V::S
symbolic_default<V>::TWO= typename V::S (2);
template<typename V> typename V::S
symbolic_default<V>::ONE_HALF=
  symbolic_default<V>::ONE / symbolic_default<V>::TWO;

template<typename V> typename V::S
symbolic_default<V>::MINUS= typename V::S ("-");
template<typename V> typename V::S
symbolic_default<V>::PLUS= typename V::S ("+");
template<typename V> typename V::S
symbolic_default<V>::TIMES= typename V::S ("*");
template<typename V> typename V::S
symbolic_default<V>::OVER= typename V::S ("/");
template<typename V> typename V::S
symbolic_default<V>::SQRT= typename V::S ("sqrt");
template<typename V> typename V::S
symbolic_default<V>::SQUARE= typename V::S ("square");
template<typename V> typename V::S
symbolic_default<V>::POWER= typename V::S ("^");
template<typename V> typename V::S
symbolic_default<V>::MATHE= typename V::S ("mathe");
template<typename V> typename V::S
symbolic_default<V>::EXP= typename V::S ("exp");
template<typename V> typename V::S
symbolic_default<V>::LOG= typename V::S ("log");
template<typename V> typename V::S
symbolic_default<V>::COS= typename V::S ("cos");
template<typename V> typename V::S
symbolic_default<V>::SIN= typename V::S ("sin");
template<typename V> typename V::S
symbolic_default<V>::TAN= typename V::S ("tan");
template<typename V> typename V::S
symbolic_default<V>::ARCCOS= typename V::S ("arccos");
template<typename V> typename V::S
symbolic_default<V>::ARCSIN= typename V::S ("arcsin");
template<typename V> typename V::S
symbolic_default<V>::ARCTAN= typename V::S ("arctan");

template<typename V> typename V::S
symbolic_default<V>::DERIVATIVE= typename V::S ("derivative");
template<typename V> typename V::S
symbolic_default<V>::LAMBDA= typename V::S ("lambda");
template<typename V> typename V::S
symbolic_default<V>::ACCESS= typename V::S (".[]");
template<typename V> typename V::S
symbolic_default<V>::MATHD= typename V::S ("mathd");
template<typename V> typename V::S
symbolic_default<V>::DERIVE= typename V::S ("derive");
template<typename V> typename V::S
symbolic_default<V>::EQUAL= typename V::S ("=");
template<typename V> typename V::S
symbolic_default<V>::FAIL= typename V::S ("fail");
template<typename V> typename V::S
symbolic_default<V>::UNINIT= typename V::S ("uninit");

} // namespace mmx
#endif // __SYMBOLIC_DEFAULT__HPP
