
/******************************************************************************
* MODULE     : symbolic_derive.hpp
* DESCRIPTION: Derivation of symbolic expressions
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __SYMBOLIC_DERIVE__HPP
#define __SYMBOLIC_DERIVE__HPP
#include <symbolix/symbolic.hpp>
#include <symbolix/wrap_sum.hpp>
#include <symbolix/wrap_product.hpp>
#include <symbolix/wrap_derivative.hpp>
#include <symbolix/wrap_function.hpp>
#include <symbolix/wrap_partial.hpp>
#include <basix/memoize.hpp>
namespace mmx {

template<typename S, typename Var>
struct derivative_helper {
  typedef S Ret;
  static inline bool is_canonical (const Var& v) {
    (void) v; return false; }
  static inline bool equal (const S& x, const Var& v) {
    return x == v; }
  static inline Ret derive (const S& x, const Var& v) {
    if (x == v) return 1; else return 0; }
};

template<typename W>
struct derivative_helper<symbolic<W>,wrap_partial<W> > {
  typedef symbolic<W> Ret;
  static inline bool is_canonical (const wrap_partial<W>& v) {
    return is<integer> (variable (v)); }
  static inline bool equal (const symbolic<W>& x, const wrap_partial<W>& v) {
    return x == variable (v); }
  static Ret derive (const symbolic<W>& x, const wrap_partial<W>& v) {
    if (x == variable (v)) return multiplier (v);
    else if (is<wrap_derivative<W> > (x)) {
      wrap_derivative<W> df= as<wrap_derivative<W> > (x);
      if (is<integer> (variable (v))) return mmx::derive (df, number (v));
      return replace (x, root (df), derive (root (df), v));
    }
    else if (is<integer> (variable (v)))
      return mmx::derive (wrap_derivative<W> (x), number (v));
    else return 0;
  }
};

template<typename W, typename Var>
typename derivative_helper<symbolic<W>,Var>::Ret
derive_cached (const symbolic<W>& x, const Var& v) {
  typedef symbolic<W> S;
  typedef typename derivative_helper<symbolic<W>,Var>::Ret R;
  MEMOIZE_BINARY (std_memoizer,R,S,Var,derive_cached,x,v,derive(x,v));
}

#define R typename derivative_helper<S,Var>::Ret

template<typename V>
struct symbolic_derive: public V {
  typedef typename V::S S;
  typedef typename V::W W;
  typedef typename V::compound compound;
  typedef typename V::sum sum;
  typedef typename V::product product;
  typedef typename V::derivative derivative;
  typedef typename V::function function;
  using V::apply;
  using V::ZERO;
  using V::ONE;
  using V::MINUS_ONE;
  using V::EXP;
  using V::LOG;
  using V::COS;
  using V::SIN;
  using V::TAN;
  using V::ARCCOS;
  using V::ARCSIN;
  using V::ARCTAN;
  using V::ACCESS;
  using V::expression;

  //----- derivatives of standard functions -----------------------------------

  static S dlog (const S& x) { return ONE / x; }
  static S dexp (const S& x) { return exp (x); }
  static S dcos (const S& x) { return -sin (x); }
  static S dsin (const S& x) { return cos (x); }
  static S dtan (const S& x) { return ONE / square (cos (x)); }
  static S dacos (const S& x) { return MINUS_ONE / sqrt (ONE - square (x)); }
  static S dasin (const S& x) { return ONE / sqrt (ONE - square (x)); }
  static S datan (const S& x) { return ONE / (ONE + square (x)); }

  typedef S (*unary_function) (const S&);
  typedef table<unary_function,S> unary_function_table;
  static unary_function_table
  compute_unary_derivative_table () {
    unary_function_table t ((unary_function) NULL);
    t [EXP]= dexp;
    t [LOG]= dlog;
    t [COS]= dcos;
    t [SIN]= dsin;
    t [TAN]= dtan;
    t [ARCCOS]= dacos;
    t [ARCSIN]= dasin;
    t [ARCTAN]= datan;
    return t;
  }

  //----- derivatives of compound expressions ---------------------------------

  template<typename Var> static R
  derive (const compound& x, const Var& y) {
    vector<S> v= *x;
    switch (N(v)) {
    case 0:
    case 1:
      return ZERO;
    case 2:
      {
	static unary_function_table unary_derivative_table=
	  compute_unary_derivative_table ();
	unary_function df= unary_derivative_table [v[0]];
	if (df == NULL) break;
	return mul_expand_left (derive (v[1], y), df (v[1]));
      }
    }
    vector<S> args   = cdr (v);
    vector<R> partial= fill<R> (N(args) + 1);
    for (nat i=0; i<N(args); i++) {
      R dfi= derive (v[0], wrap_partial<W> (S (integer (i))));
      R dai= derive (args[i], y);
      partial[i]= mul_expand (dai, apply (dfi, args));
    }
    partial[N(args)]= apply (derive (v[0], y), args);
    return big_add (partial);
  }

  //----- derivatives of sums, products, derivatives and functions ------------

  template<typename Var> static R
  derive (const sum& s, const Var& y) {
    nat n= N(s);
    vector<R> v= fill<R> (n);
    for (nat i=0; i<n; i++)
      v[i]= co (s, i) * derive_cached (mo (s, i), y);
    return big_add (v);
  }
  
  template<typename Var> static R
  derive (const product& p, const Var& y) {
    nat n= N (p);
    vector<R> v= fill<R> (n);
    for (nat i=0; i<n; i++)
      v[i]= log (mo (p, i)) * derive (co (p, i), y) +
	    (co (p, i) / mo (p, i)) * derive_cached (mo (p, i), y);
    return mul_expand_left (big_add (v), p);
  }

  template<typename Var> static R
  derive (const function& f, const Var& y) {
    vector<S> a= arguments (f);
    if (derivative_helper<S,Var>::is_canonical (y))
      return unapply (derive_cached (body (f), a[number (y)]), a);
    else {
      for (nat i=0; i<N(a); i++)
	if (derivative_helper<S,Var>::equal (a[i], y)) return ZERO;
      return unapply (derive_cached (body (f), y), a);
    }
  }

  //----- dispatch ------------------------------------------------------------
  
  template<typename Var> static R
  derive (const S& x, const Var& v) {
    //cout << "Derive " << x << ", " << v << "\n";
    switch (symbolic_type (x)) {
    case SYMBOLIC_INTEGER:
    case SYMBOLIC_RATIONAL:
    case SYMBOLIC_FLOATING:
    case SYMBOLIC_BALL:
    case SYMBOLIC_REAL:
      return ZERO;
    case SYMBOLIC_LITERAL:
      return derivative_helper<S,Var>::derive (x, v);
    case SYMBOLIC_COMPOUND:
      if (is_func (x, ACCESS)) // FIXME: a bit too brutal
	return derivative_helper<S,Var>::derive (x, v);
      return derive (as<compound> (x), v);
    case SYMBOLIC_SUM:
      return derive (as<sum> (x), v);
    case SYMBOLIC_PRODUCT:
      return derive (as<product> (x), v);
    case SYMBOLIC_DERIVATIVE:
      return derivative_helper<S,Var>::derive (x, v);
    case SYMBOLIC_FUNCTION:
      return derive (as<function> (x), v);
    default:
      return ZERO;
    }
  }

  static S
  derive (const S& x, const S& v) {
    //cout << "Derive " << x << ", " << v << "\n";
    if (is_func (v, S ("variate"), 2))
      return derive (x, wrap_partial<W> (v[1], v[2]));
    else 
      return derive (x, wrap_partial<W> (v));
  }

  static S
  derive (const S& x, const vector<S>& v) {
    S r= x;
    for (nat i=0; i<N(v); i++)
      r= derive (r, v[i]);
    return r;
  } 

  //----- Flattening derivatives of functions ---------------------------------

  static syntactic
  expression (const derivative& df) {
    vector<nat> o= orders (df);
    if (N (o) == 1) {
      syntactic r= flatten (root (df));
      for (nat i=0; i<o[0]; i++)
	r= mmx::apply (".'", r);
      return r;
    }
    else {
      vector<syntactic> v= vec (flatten (root (df)));
      for (nat i=0; i<N(o); i++)
	for (nat j=0; j<o[i]; j++)
	  v << flatten (integer (i));
      return mmx::apply ("derivative", v);
    }
  }
};

#undef R

} // namespace mmx
#endif // __SYMBOLIC_DERIVE__HPP
