
/******************************************************************************
* MODULE     : symbolic_internal.hpp
* DESCRIPTION: Internal syntactic operations on symbolic expressions
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __SYMBOLIC_INTERNAL__HPP
#define __SYMBOLIC_INTERNAL__HPP
#include <symbolix/symbolic.hpp>
#include <basix/symbol.hpp>
#include <numerix/ball.hpp>
#include <numerix/real.hpp>
#include <symbolix/wrap_sum.hpp>
#include <symbolix/wrap_product.hpp>
#include <symbolix/wrap_derivative.hpp>
#include <symbolix/wrap_function.hpp>
namespace mmx {
template<typename W> struct wrap_sum;
template<typename W> struct wrap_product;

/******************************************************************************
* For debugging
******************************************************************************/

template<typename C> inline syntactic
dump (const C& x) {
  return flatten (x);
}

template<> inline syntactic
dump (const generic& x) {
  return apply ("generic", flatten (x));
}

template<typename X1, typename X2> syntactic
dump (const pair<X1,X2>& x) {
  return apply ("[]", dump (x.x1), dump (x.x2));
}

template<typename C, typename V> syntactic
dump (const symbol<C,V>& x) {
  return apply ("symbol", flatten (inside (x)), dump (*x));
}

template<typename C, typename V> syntactic
dump (const vector<C,V>& v) {
  vector<syntactic> a= fill<syntactic> (N(v));
  for (nat i=0; i<N(v); i++) a[i]= dump (v[i]);
  return apply ("[]", a);
}

template<typename C, typename T, typename V> syntactic
dump (const sparse_vector<C,T,V>& v) {
  vector<syntactic> a= fill<syntactic> (N(v));
  for (nat i=0; i<N(v); i++) a[i]= dump (v[i]);
  return apply ("[]", a);
}

template<typename W> syntactic
dump (const symbolic<W>& s) {
  return apply ("symbolic", W::dump_expression (s));
}

/******************************************************************************
* Basic routines
******************************************************************************/

template<typename RR, typename WW>
struct symbolic_internal {
  typedef RR R;
  typedef WW W;
  typedef symbolic<W> S;
  typedef symbol<string> literal;
  typedef symbol<vector<S> > compound;
  typedef wrap_sum<W> sum;
  typedef wrap_product<W> product;
  typedef wrap_derivative<W> derivative;
  typedef wrap_function<W> function;

  /****************************************************************************
  * Macro-like routines
  ****************************************************************************/

  static inline bool is_scalar (const S& s) {
    register nat tp= symbolic_type (*s);
    return tp >= SYMBOLIC_INTEGER && tp <= SYMBOLIC_REAL; }    
  static inline bool is_atomic (const S& s) {
    register nat tp= symbolic_type (*s);
    return tp <= SYMBOLIC_LITERAL; }    
  static inline bool is_composite (const S& s) {
    register nat tp= symbolic_type (*s);
    return tp >= SYMBOLIC_COMPOUND; }    
  
  /****************************************************************************
  * Constructors
  ****************************************************************************/

  static inline R construct () {
    return as<R> (integer (0)); }
  template<typename C> static inline R construct (const C& c) {
    return as<R> (c); }
  static inline R construct (const string& s) {
    return as<R> (literal (s)); }

  static R construct (const generic& g) {
    if (is<mmx::literal> (g)) return construct (as<mmx::literal> (g));
    else if (is<mmx::compound> (g)) return construct (as<mmx::compound> (g)); 
    else if (is<S> (g)) return *as<S> (g);
    else return g;
  }
  static inline R construct (const mmx::literal& s) {
    return as<R> (literal (as_string (s))); }
  static R construct (const mmx::compound& c) {
    vector<generic> v= as_vector (c);
    nat n= N(v);
    vector<S> r= fill<S> (n);
    for (nat i=0; i<n; i++) r[i]= S (construct (v[i]));
    return *W::apply (r);
  }

  static inline R construct (const rational& c) {
    if (denominator (c) == 1) return as<R> (numerator (c));
    else return as<R> (c); }
  static inline R construct (const floating<>& c) {
    if (c == 0) return as<R> (integer (0));
    else return as<R> (c); }
  static inline R construct (const ball<floating<> >& c) {
    if (c == 0) return as<R> (integer (0));
    else return as<R> (c); }
  static inline R construct (const real& c) {
    if (c == 0) return as<R> (integer (0));
    else return as<R> (c); }

  static inline R construct (const char* s) {
    return construct (literal (string (s))); }
  static inline R construct (const int& c) {
    return construct (integer (c)); }
  static inline R construct (const nat& c) {
    return construct (integer (c)); }
  static inline R construct (const double& c) {
    return construct (floating<> (c)); }

  /****************************************************************************
  * Composite expressions
  ****************************************************************************/

  static inline S apply (const vector<S>& a) {
    return compound (a); }
  static inline S apply (const S& f) {
    return apply (vec (f)); }
  static inline S apply (const S& f, const S& a) {
    return apply (vec (f, a)); }
  static inline S apply (const S& f, const S& a, const S& b) {
    return apply (vec (f, a, b)); }
  static inline S apply (const S& f, const S& a, const S& b, const S& c) {
    return apply (vec (f, a, b, c)); }
  static inline S apply (const S& f, const vector<S>& a) {
    return apply (cons (f, a)); }

  static nat arity (const S& x) {
    switch (symbolic_type (x)) {
    case SYMBOLIC_COMPOUND:
      return N (*as<compound> (x));
    case SYMBOLIC_SUM:
      return max (N (as<sum> (x)), (nat) 2) + 1;
    case SYMBOLIC_PRODUCT:
      return max (N (as<product> (x)), (nat) 2) + 1;
    case SYMBOLIC_DERIVATIVE:
      return N (orders (as<derivative> (x))) + 2;
    case SYMBOLIC_FUNCTION:
      return N (arguments (as<function> (x))) + 2;
    default:
      ERROR ("composite symbolic expression expected");
      return 0;
    }
  }

  static S access (const S& x, nat i) {
    switch (symbolic_type (x)) {
    case SYMBOLIC_COMPOUND: {
      vector<S> v= *as<compound> (x);
      ASSERT (i < N(v), "index out of range");
      return v[i];
    }
    case SYMBOLIC_SUM: {
      sum s= as<sum> (x);
      ASSERT (i <= max (N(s), (nat) 2), "index out of range");
      if (N(s) == 1) {
	if (i == 0) return "*";
	if (i == 1) return co (s, 0);
	if (i == 2) return mo (s, 0);
      }
      else {
	if (i == 0) return "+";
	else return s[i-1];
      }
    }
    case SYMBOLIC_PRODUCT: {
      product p= as<product> (x);
      ASSERT (i <= max (N(p), (nat) 2), "index out of range");
      if (N(p) == 1) {
	if (i == 0) return "^";
	if (i == 1) return mo (p, 0);
	if (i == 2) return co (p, 0);
      }
      else {
	if (i == 0) return "*";
	else return p[i-1];
      }
    }
    case SYMBOLIC_DERIVATIVE: {
      derivative df= as<derivative> (x);
      ASSERT (i < N (orders (df)) + 2, "index out of range");
      if (i == 0) return "derivative";
      if (i == 1) return root (df);
      return integer (orders (df) [i+2]);
    }
    case SYMBOLIC_FUNCTION: {
      function f= as<function> (x);
      nat n= N (arguments (f));
      ASSERT (i < n + 2, "index out of range");
      if (i == 0) return "lambda";
      if (i == n + 1) return body (f);
      return arguments (f) [i+1];
    }
    default:
      ERROR ("composite symbolic expression expected");
      return 0;
    }
  }

  static vector<S> components (const S& x) {
    switch (symbolic_type (x)) {
    case SYMBOLIC_COMPOUND:
      return *as<compound> (x);
    case SYMBOLIC_SUM: {
      sum s= as<sum> (x);
      if (N(s) == 1) return vec (S ("*"), co (s, 0), mo (s, 0));
      vector<S> r= fill<S> (N(s) + 1);
      r[0]= S ("+");
      for (nat i=0; i<N(s); i++) r[i+1]= s[i];
      return r;
    }
    case SYMBOLIC_PRODUCT: {
      product p= as<product> (x);
      if (N(p) == 1) return vec (S ("^"), mo (p, 0), co (p, 0));
      vector<S> r= fill<S> (N(p) + 1);
      r[0]= S ("*");
      for (nat i=0; i<N(p); i++) r[i+1]= p[i];
      return r;
    }
    case SYMBOLIC_DERIVATIVE: {
      derivative df= as<derivative> (x);
      vector<nat> o= orders (df);
      nat n= N (o);
      vector<S> r= fill<S> (n + 2);
      r[0]= S ("derivative");
      r[1]= root (df);
      for (nat i=0; i<n; i++) r[i+2]= integer (o[i]);
      return r;
    }
    case SYMBOLIC_FUNCTION: {
      function f= as<function> (x);
      vector<S> a= arguments (f);
      nat n= N (a);
      vector<S> r= fill<S> (n + 2);
      r[0]= S ("lambda");
      r[n+1]= body (f);
      for (nat i=0; i<n; i++) r[i+1]= a[i];
      return r;
    }
    default:
      ERROR ("composite symbolic expression expected");
      return vector<S> ();
    }
  }

  /****************************************************************************
  * Flattening
  ****************************************************************************/

  static syntactic expression (const S& x) {
    if (is<literal> (x)) return flatten (mmx::literal (*as<literal> (x)));
    else if (is<compound> (x)) {
      vector<S> v= *as<compound> (x);
      nat n= N(v);
      vector<generic> r= fill<generic> (n);
      for (nat i=0; i<n; i++) r[i]= as_generic (expression (v[i]));
      return as_syntactic (as<generic> (mmx::compound (r)));
    }
    else return flatten (*x);
  }

  static syntactic dump_expression (const S& x) {
    switch (symbolic_type (x)) {
    case SYMBOLIC_INTEGER:
      return dump (as<integer> (x));
    case SYMBOLIC_RATIONAL:
      return dump (as<rational> (x));
    case SYMBOLIC_FLOATING:
      return dump (as<floating<> > (x));
    case SYMBOLIC_BALL:
      return dump (as<ball<floating<> > > (x));
    case SYMBOLIC_REAL:
      return dump (as<real> (x));
    case SYMBOLIC_LITERAL:
      return dump (as<literal> (x));
    case SYMBOLIC_COMPOUND:
      return dump (as<compound> (x));
    case SYMBOLIC_SUM:
      return dump (as<sum> (x));
    case SYMBOLIC_PRODUCT:
      return dump (as<product> (x));
    case SYMBOLIC_DERIVATIVE:
      return dump (as<derivative> (x));
    case SYMBOLIC_FUNCTION:
      return dump (as<function> (x));
    default:
      return dump (*x);
    }
  }
};

} // namespace mmx
#endif // __SYMBOLIC_INTERNAL__HPP
