
/******************************************************************************
* MODULE     : symbolic_product.hpp
* DESCRIPTION: Symbolic power products with scalar exponents
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __SYMBOLIC_PRODUCT__HPP
#define __SYMBOLIC_PRODUCT__HPP
#include <symbolix/symbolic.hpp>
#include <symbolix/wrap_sum.hpp>
#include <symbolix/wrap_product.hpp>
namespace mmx {

template<typename V>
struct symbolic_product: public V {
  typedef typename V::S S;
  typedef typename V::sum sum;
  typedef typename V::product product;
  using V::expression;
  using V::is_scalar;
  using V::ZERO;
  using V::ONE;
  using V::MINUS_ONE;
  using V::TWO;
  using V::ONE_HALF;

  //----- Powers --------------------------------------------------------------

  static inline bool is_power (const S& x) {
    return is<product> (x) && N (as<product> (x)) == 1; }
  static inline S power (const S& m, const S& c) {
    if (m == ONE) return ONE;
    if (c == ONE) return m;
    return product (m, c); }
  static inline S get_base (const S& x) {
    if (is<product> (x)) return co (as<product> (x), 0);
    else return ONE; }
  static inline S get_exponent (const S& x) {
    if (is<product> (x)) {
      product p= as<product> (x);
      if (N(p) == 1) return mo (p, 0);
      else return product_to_symbolic (sqrtn (p, co (p, 0)));
    }
    else return x; }

  //----- Conversion between products and symbolic expressions ----------------

  static product
  symbolic_to_product (const S& x) {
    if (is<product> (x)) return as<product> (x);
    else if (is<sum> (x)) {
      S lc= co (as<sum> (x), 0);
      if (lc == ONE) return product (x, ONE);
      else return product (lc, ONE) *
	          symbolic_to_product (div_scalar_right (x, lc));
    }
    else if (x == 1) return product ();
    else return product (x, ONE);
  }

  static S
  product_to_symbolic (const product& p) {
    typedef pair<symbol<S>,S> Term;
    typedef sparse_vector<S,symbol<S>,hard_eq_sparse_vector> Terms;
    //cout << "Convert " << dump (p) << "\n";
    nat l= N(p), n= 0;
    Term* r= mmx_new<Term> (l);
    S sc= ONE;
    for (nat k=0; k<N(p); k++) {
      S x= mo (p, k);
      S e= co (p, k);
      if (e == ZERO);
      else if (is_scalar (x)) {
	if (e == ONE) sc= sc * x;
	else if (e == MINUS_ONE) sc= sc / x;
	else if (is<integer> (e) ||
		 (is<rational> (e) && denominator (as<rational> (e)) == 1))
	  {
	    integer i= is<integer> (e)?
	      as<integer> (e): numerator (as<rational> (e));
	    if (abs (i) <= 10) {
	      if (i >= 0) sc= sc * binpow (x, (nat) as_int (i));
	      else sc= sc / binpow (x, (nat) as_int (-i));
	    }
	    else r[n++]= (*p)[k];
	  }
	else r[n++]= (*p)[k];
      }
      else r[n++]= (*p)[k];
    }

    if (n == 0) {
      mmx_delete<Term> (r, l);
      return sc;
    }
    if (n == 1 && r[0].x2 == ONE) {
      S x= *(r[0].x1);
      mmx_delete<Term> (r, l);
      if (sc == ONE) return x;
      return sum (x, sc);
    }
    S prod= product (Terms (r, n, l));
    if (sc == ONE) return prod;
    return sum (prod, sc);
  }

  //----- Expanded multiplication and division --------------------------------

  static S
  mul_scalar_right (const S& x1, const S& sc) {
    if (is_scalar (x1)) return V::mul (x1, sc);
    return sum_to_symbolic (symbolic_to_sum (x1) * sc);
  }

  static S
  div_scalar_right (const S& x1, const S& sc) {
    if (is_scalar (x1)) return V::mul (x1, sc);
    return sum_to_symbolic (symbolic_to_sum (x1) / sc);
  }

  static S
  mul_expand_left (const S& x, const S& y) {
    sum s= symbolic_to_sum (x);
    nat n= N(s);
    vector<S> v= fill<S> (n);
    for (nat i=0; i<n; i++)
      v[i]= co (s, i) * (mo (s, i) * y);
    return big_add (v);
  }

  static S
  div_expand_left (const S& x, const S& y) {
    sum s= symbolic_to_sum (x);
    nat n= N(s);
    vector<S> v= fill<S> (n);
    for (nat i=0; i<n; i++)
      v[i]= co (s, i) * (mo (s, i) / y);
    return big_add (v);
  }

  static S
  mul_expand (const S& x, const S& y) {
    sum s1= symbolic_to_sum (x);
    sum s2= symbolic_to_sum (y);
    nat n1= N(s1), n2= N(s2);
    vector<S> v= fill<S> (n1 * n2);
    for (nat i1=0; i1<n1; i1++)
      for (nat i2=0; i2<n2; i2++)
	v[i1*n2+i2]=
	  term ((mo (s1, i1) * mo (s2, i2)), (co (s1, i1) * co (s2, i2)));
    return big_add (v);
  }

  //----- Default multiplication and division ---------------------------------

  static bool
  is_sum_free (const S& x) {
    if (is<sum> (x)) return false;
    if (!is<product> (x)) return true;
    product p= as<product> (x);
    for (nat i=0; i<N(p); i++)
      if (is<sum> (mo (p, i)))
	return false;
    return true;
  }

  static S
  mul (const S& x1, const S& x2) {
    if (is_scalar (x1)) {
      if (is_scalar (x2)) return V::mul (x1, x2);
      else if (x1 == ZERO) return ZERO;
      else if (x1 == ONE) return x2;
      return sum_to_symbolic (symbolic_to_sum (x2) * x1);
    }
    if (is_scalar (x2)) {
      if (x2 == ZERO) return ZERO;
      else if (x2 == ONE) return x1;
      return sum_to_symbolic (symbolic_to_sum (x1) * x2);
    }
    if (is<sum> (x1) && is_sum_free (x2))
      return mul_expand_left (x1, x2);
    if (is<sum> (x2) && is_sum_free (x1))
      return mul_expand_left (x2, x1);
    return product_to_symbolic (symbolic_to_product (x1) *
				symbolic_to_product (x2));
  }

  static S
  div (const S& x1, const S& x2) {
    if (is_scalar (x1)) {
      if (is_scalar (x2)) return V::div (x1, x2);
      else if (x1 == ZERO) return ZERO;
    }
    if (is_scalar (x2)) {
      if (x2 == ZERO) { ERROR ("division by zero"); }
      else if (x2 == ONE) return x1;
      return sum_to_symbolic (symbolic_to_sum (x1) / x2);
    }
    if (is<sum> (x1) && is_sum_free (x2))
      return div_expand_left (x1, x2);
    return product_to_symbolic (symbolic_to_product (x1) /
				symbolic_to_product (x2));
  }

  static S
  mul (const vector<S>& v) {
    switch (N(v)) {
    case 0: return ONE;
    case 1: return v[0];
    case 2: return mul (v[0], v[1]);
    case 3: return mul (v[0], mul (v[1], v[2]));
    case 4: return mul (mul (v[0], v[1]), mul (v[2], v[3]));
    default: return mul (mul (range (v, 0, N(v)>>1)),
			 mul (range (v, N(v)>>1, N(v))));
    }
  }

  //----- Taking powers -------------------------------------------------------

  static S square (const S& x1) {
    if (is_scalar (x1)) return V::square (x1);
    return product_to_symbolic (mmx::pow (symbolic_to_product (x1), TWO));
  }

  static S sqrt (const S& x1) {
    if (is_scalar (x1)) return V::sqrt (x1);
    return product_to_symbolic (mmx::pow (symbolic_to_product (x1), ONE_HALF));
  }

  static S pow (const S& x1, const S& x2) {
    if (is_scalar (x1)) {
      //if (is_scalar (x2)) return V::pow (x1, x2);
      if (x1 == ZERO) return ZERO;
      if (x1 == ONE) return ONE;
    }
    if (is_scalar (x2)) {
      if (x2 == ZERO) return ONE;
      if (x2 == ONE) return x1;
      if (x2 == MINUS_ONE) return ONE / x1;
    }
    return product_to_symbolic (mmx::pow (symbolic_to_product (x1), x2));
  }

  //----- Flattening products -------------------------------------------------

  static syntactic
  expression (const product& p) {
    nat n= N(p);
    vector<syntactic> v= fill<syntactic> (n);
    for (nat i=0; i<n; i++) {
      syntactic c= flatten (co (p, i));
      syntactic m= flatten (mo (p, i));
      bool old= set_frac_flag (true);
      v[i]= mmx::pow (m, c);
      (void) set_frac_flag (old);
    }
    return ordered_product (v);
  }
};

} // namespace mmx
#endif // __SYMBOLIC_PRODUCT__HPP
