
/******************************************************************************
* MODULE     : symbolic_scalar.hpp
* DESCRIPTION: Scalaring of symbolic operations for simple types
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __SYMBOLIC_SCALAR__HPP
#define __SYMBOLIC_SCALAR__HPP
#include <symbolix/symbolic.hpp>
#include <numerix/ball.hpp>
namespace mmx {
#define Floating floating<>
#define Ball ball<floating<> >

struct var_div_op: public div_op {
  using div_op::op;
  static inline rational op (const integer& i, const integer& j) {
    return rational (i) / rational (j); }
};

template<typename V>
struct symbolic_scalar: public V {
  typedef typename V::S S;

  /****************************************************************************
  * Unary arithmetic operations
  ****************************************************************************/

  template<typename Op, S (*fall_back) (const S&)> static S
  arithmetic_unary (const S& x) {
    switch (symbolic_type (x)) {
    case SYMBOLIC_INTEGER:
      return S (Op::op (as<integer> (x)));
    case SYMBOLIC_RATIONAL:
      return S (Op::op (as<rational> (x)));
    case SYMBOLIC_FLOATING:
      return S (Op::op (as<Floating > (x)));
    case SYMBOLIC_BALL:
      return S (Op::op (as<Ball > (x)));
    case SYMBOLIC_REAL:
      return S (Op::op (as<real> (x)));
    default:
      return fall_back (x);
    }
  }

  static inline S neg (const S& x) {
    return arithmetic_unary<neg_op,V::neg> (x); }
  static inline S square (const S& x) {
    return arithmetic_unary<square_op,V::square> (x); }

  /****************************************************************************
  * Binary arithmetic operations
  ****************************************************************************/

  template<typename Op, S (*fall_back) (const S&, const S&)> static S
  arithmetic_binary (const S& x1, const S& x2) {
    switch (symbolic_type (x1)) {
    case SYMBOLIC_INTEGER:
      switch (symbolic_type (x2)) {
      case SYMBOLIC_INTEGER:
	return S (Op::op (as<integer> (x1), as<integer> (x2)));
      case SYMBOLIC_RATIONAL:
	return S (Op::op (rational (as<integer> (x1)), as<rational> (x2)));
      case SYMBOLIC_FLOATING:
	return S (Op::op (Floating (as<integer> (x1)), as<Floating > (x2)));
      case SYMBOLIC_BALL:
	return S (Op::op (Ball (as<integer> (x1)), as<Ball > (x2)));
      case SYMBOLIC_REAL:
	return S (Op::op (real (as<integer> (x1)), as<real> (x2)));
      default:
	return fall_back (x1, x2);
      }
    case SYMBOLIC_RATIONAL:
      switch (symbolic_type (x2)) {
      case SYMBOLIC_INTEGER:
	return S (Op::op (as<rational> (x1), rational (as<integer> (x2))));
      case SYMBOLIC_RATIONAL:
	return S (Op::op (as<rational> (x1), as<rational> (x2)));
      case SYMBOLIC_FLOATING:
	return S (Op::op (Floating (as<rational> (x1)), as<Floating > (x2)));
      case SYMBOLIC_BALL:
	return S (Op::op (Ball (as<rational> (x1)), as<Ball > (x2)));
      case SYMBOLIC_REAL:
	return S (Op::op (real (as<rational> (x1)), as<real> (x2)));
      default:
	return fall_back (x1, x2);
      }
    case SYMBOLIC_FLOATING:
      switch (symbolic_type (x2)) {
      case SYMBOLIC_INTEGER:
	return S (Op::op (as<Floating > (x1), Floating (as<integer> (x2))));
      case SYMBOLIC_RATIONAL:
	return S (Op::op (as<Floating > (x1), Floating (as<rational> (x2))));
      case SYMBOLIC_FLOATING:
	return S (Op::op (as<Floating > (x1), as<Floating > (x2)));
      case SYMBOLIC_BALL:
	return S (Op::op (Ball (as<Floating > (x1)), as<Ball > (x2)));
      case SYMBOLIC_REAL:
	return S (Op::op (real (as<Floating > (x1)), as<real> (x2)));
      default:
	return fall_back (x1, x2);
      }
    case SYMBOLIC_BALL:
      switch (symbolic_type (x2)) {
      case SYMBOLIC_INTEGER:
	return S (Op::op (as<Ball > (x1), Ball (as<integer> (x2))));
      case SYMBOLIC_RATIONAL:
	return S (Op::op (as<Ball > (x1), Ball (as<rational> (x2))));
      case SYMBOLIC_FLOATING:
	return S (Op::op (as<Ball > (x1), Ball (as<Floating > (x2))));
      case SYMBOLIC_BALL:
	return S (Op::op (as<Ball > (x1), as<Ball > (x2)));
      case SYMBOLIC_REAL:
	return S (Op::op (real (as<Ball > (x1)), as<real> (x2)));
      default:
	return fall_back (x1, x2);
      }
    case SYMBOLIC_REAL:
      switch (symbolic_type (x2)) {
      case SYMBOLIC_INTEGER:
	return S (Op::op (as<real> (x1), real (as<integer> (x2))));
      case SYMBOLIC_RATIONAL:
	return S (Op::op (as<real> (x1), real (as<rational> (x2))));
      case SYMBOLIC_FLOATING:
	return S (Op::op (as<real> (x1), real (as<Floating > (x2))));
      case SYMBOLIC_BALL:
	return S (Op::op (as<real> (x1), real (as<Ball > (x2))));
      case SYMBOLIC_REAL:
	return S (Op::op (as<real> (x1), as<real> (x2)));
      default:
	return fall_back (x1, x2);
      }
    default:
      return fall_back (x1, x2);
    }
  }

  static inline S add (const S& x1, const S& x2) {
    return arithmetic_binary<add_op,V::add> (x1, x2); }
  static inline S sub (const S& x1, const S& x2) {
    return arithmetic_binary<sub_op,V::sub> (x1, x2); }
  static inline S mul (const S& x1, const S& x2) {
    return arithmetic_binary<mul_op,V::mul> (x1, x2); }
  static inline S div (const S& x1, const S& x2) {
    return arithmetic_binary<var_div_op,V::div> (x1, x2); }

  /****************************************************************************
  * Unary transcendental operations
  ****************************************************************************/

  template<typename Op, S (*fall_back) (const S&)> static S
  transcendental_unary (const S& x) {
    switch (symbolic_type (x)) {
    case SYMBOLIC_FLOATING:
      return S (Op::op (as<Floating > (x)));
    case SYMBOLIC_BALL:
      return S (Op::op (as<Ball > (x)));
    case SYMBOLIC_REAL:
      return S (Op::op (as<real> (x)));
    default:
      return fall_back (x);
    }
  }

  static inline S sqrt (const S& x) {
    return transcendental_unary<sqrt_op,V::sqrt> (x); }
  static inline S exp (const S& x) {
    return transcendental_unary<exp_op,V::exp> (x); }
  static inline S log (const S& x) {
    return transcendental_unary<log_op,V::log> (x); }
  static inline S cos (const S& x) {
    return transcendental_unary<cos_op,V::cos> (x); }
  static inline S sin (const S& x) {
    return transcendental_unary<sin_op,V::sin> (x); }
  static inline S tan (const S& x) {
    return transcendental_unary<tan_op,V::tan> (x); }
  static inline S acos (const S& x) {
    return transcendental_unary<acos_op,V::acos> (x); }
  static inline S asin (const S& x) {
    return transcendental_unary<asin_op,V::asin> (x); }
  static inline S atan (const S& x) {
    return transcendental_unary<atan_op,V::atan> (x); }

  /****************************************************************************
  * Binary arithmetic tests
  ****************************************************************************/

  template<typename Op, bool (*fall_back) (const S&, const S&)> static bool
  arithmetic_test (const S& x1, const S& x2) {
    switch (symbolic_type (x1)) {
    case SYMBOLIC_INTEGER:
      switch (symbolic_type (x2)) {
      case SYMBOLIC_INTEGER:
	return Op::op (as<integer> (x1), as<integer> (x2));
      case SYMBOLIC_RATIONAL:
	return Op::op (rational (as<integer> (x1)), as<rational> (x2));
      case SYMBOLIC_FLOATING:
	return Op::op (Floating (as<integer> (x1)), as<Floating > (x2));
      case SYMBOLIC_BALL:
	return Op::op (Ball (as<integer> (x1)), as<Ball > (x2));
      case SYMBOLIC_REAL:
	return Op::op (real (as<integer> (x1)), as<real> (x2));
      default:
	return fall_back (x1, x2);
      }
    case SYMBOLIC_RATIONAL:
      switch (symbolic_type (x2)) {
      case SYMBOLIC_INTEGER:
	return Op::op (as<rational> (x1), rational (as<integer> (x2)));
      case SYMBOLIC_RATIONAL:
	return Op::op (as<rational> (x1), as<rational> (x2));
      case SYMBOLIC_FLOATING:
	return Op::op (Floating (as<rational> (x1)), as<Floating > (x2));
      case SYMBOLIC_BALL:
	return Op::op (Ball (as<rational> (x1)), as<Ball > (x2));
      case SYMBOLIC_REAL:
	return Op::op (real (as<rational> (x1)), as<real> (x2));
      default:
	return fall_back (x1, x2);
      }
    case SYMBOLIC_FLOATING:
      switch (symbolic_type (x2)) {
      case SYMBOLIC_INTEGER:
	return Op::op (as<Floating > (x1), Floating (as<integer> (x2)));
      case SYMBOLIC_RATIONAL:
	return Op::op (as<Floating > (x1), Floating (as<rational> (x2)));
      case SYMBOLIC_FLOATING:
	return Op::op (as<Floating > (x1), as<Floating > (x2));
      case SYMBOLIC_BALL:
	return Op::op (Ball (as<Floating > (x1)), as<Ball > (x2));
      case SYMBOLIC_REAL:
	return Op::op (real (as<Floating > (x1)), as<real> (x2));
      default:
	return fall_back (x1, x2);
      }
    case SYMBOLIC_BALL:
      switch (symbolic_type (x2)) {
      case SYMBOLIC_INTEGER:
	return Op::op (as<Ball > (x1), Ball (as<integer> (x2)));
      case SYMBOLIC_RATIONAL:
	return Op::op (as<Ball > (x1), Ball (as<rational> (x2)));
      case SYMBOLIC_FLOATING:
	return Op::op (as<Ball > (x1), Ball (as<Floating > (x2)));
      case SYMBOLIC_BALL:
	return Op::op (as<Ball > (x1), as<Ball > (x2));
      case SYMBOLIC_REAL:
	return Op::op (real (as<Ball > (x1)), as<real> (x2));
      default:
	return fall_back (x1, x2);
      }
    case SYMBOLIC_REAL:
      switch (symbolic_type (x2)) {
      case SYMBOLIC_INTEGER:
	return Op::op (as<real> (x1), real (as<integer> (x2)));
      case SYMBOLIC_RATIONAL:
	return Op::op (as<real> (x1), real (as<rational> (x2)));
      case SYMBOLIC_FLOATING:
	return Op::op (as<real> (x1), real (as<Floating > (x2)));
      case SYMBOLIC_BALL:
	return Op::op (as<real> (x1), real (as<Ball > (x2)));
      case SYMBOLIC_REAL:
	return Op::op (as<real> (x1), as<real> (x2));
      default:
	return fall_back (x1, x2);
      }
    default:
      return fall_back (x1, x2);
    }
  }

  static inline bool equal (const S& x1, const S& x2) {
    return arithmetic_test<equal_op,V::equal> (x1, x2); }
  static inline bool unequal (const S& x1, const S& x2) {
    return arithmetic_test<unequal_op,V::unequal> (x1, x2); }
  static inline bool lesseq (const S& x1, const S& x2) {
    return arithmetic_test<lesseq_op,V::lesseq> (x1, x2); }
  static inline bool less (const S& x1, const S& x2) {
    return arithmetic_test<less_op,V::less> (x1, x2); }
  static inline bool gtreq (const S& x1, const S& x2) {
    return arithmetic_test<gtreq_op,V::gtreq> (x1, x2); }
  static inline bool gtr (const S& x1, const S& x2) {
    return arithmetic_test<gtr_op,V::gtr> (x1, x2); }
};

#undef Floating
#undef Ball
} // namespace mmx
#endif // __SYMBOLIC_SCALAR__HPP
