
/******************************************************************************
* MODULE     : symbolic_solve.hpp
* DESCRIPTION: Very rudimentary equation solver
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __SYMBOLIC_SOLVE__HPP
#define __SYMBOLIC_SOLVE__HPP
#include <symbolix/symbolic.hpp>
#include <basix/symbol.hpp>
#include <basix/memoize.hpp>
#include <basix/vector_sort.hpp>
namespace mmx {

struct size_leq {
  template<typename C> static inline bool
  op (const C& x, const C& y) {
    return size (x) <= size (y);
  }
};

template<typename V>
struct symbolic_solve: public V {
  typedef typename V::S S;
  typedef typename V::sum sum;
  typedef typename V::product product;
  using V::replace;
  using V::ZERO;
  using V::ONE;
  using V::EXP;
  using V::LOG;
  using V::COS;
  using V::SIN;
  using V::TAN;
  using V::ARCCOS;
  using V::ARCSIN;
  using V::ARCTAN;
  using V::EQUAL;
  using V::FAIL;
  using V::UNINIT;

  //----- Solving a single equation -------------------------------------------

  template<typename SP> static nat
  find_unique (const SP& sp, const S& x, S& s, S& r) {
    nat counter= 0;
    for (nat i=0; i<N(sp); i++) {
      if (depends (co (sp, i), x) || depends (mo (sp, i), x)) {
	s= co (sp, i);
	r= mo (sp, i);
	if (depends (s, x)) counter++;
	if (depends (r, x)) counter++;
      }
    }
    return counter;
  }

  static S
  solve (const S& l, const S& r, const S& x) {
    //cout << "Solve " << l << " = " << r << " in " << x << "\n";
    if (l == r) return x;
    if (is<sum> (l) || (is<sum> (r) && depends (r, x))) {
      S y= r - l;
      if (!is<sum> (y)) return solve (y, ZERO, x);
      S sc, rad;
      if (find_unique (as<sum> (y), x, sc, rad) != 1) return FAIL;
      S z= sc * rad - y;
      if (depends (sc, x)) return solve (sc, z / rad, x);
      return solve (rad, z / sc, x);
    }
    if (is<product> (l) || (is<product> (r) && depends (r, x))) {
      S y= r / l;
      if (!is<product> (y)) return solve (y, ONE, x);
      S p, rad;
      if (find_unique (as<product> (y), x, p, rad) != 1) return FAIL;
      S z= pow (rad, p) / y;
      if (depends (p, x)) return solve (p, log (z) / log (rad), x);
      return solve (rad, pow (z, ONE/p), x);
    }

    nat nl= occurrences (l, x), nr= occurrences (r, x);
    if (nl + nr != 1) return FAIL;
    if (nl == 0) return solve (r, l, x);
    if (l == x) return r;
    
    if (is_func (l, EXP, 1)) return solve (l[1], log (r), x);
    if (is_func (l, LOG, 1)) return solve (l[1], exp (r), x);
    if (is_func (l, COS, 1)) return solve (l[1], acos (r), x);
    if (is_func (l, SIN, 1)) return solve (l[1], asin (r), x);
    if (is_func (l, TAN, 1)) return solve (l[1], atan (r), x);
    if (is_func (l, ARCCOS, 1)) return solve (l[1], cos (r), x);
    if (is_func (l, ARCSIN, 1)) return solve (l[1], sin (r), x);
    if (is_func (l, ARCTAN, 1)) return solve (l[1], tan (r), x);

    return FAIL;
  }

  static S
  solve (const S& z, const S& x) {
    //cout << "Solving " << z << " in " << x << "\n";
    if (is_func (z, EQUAL, 2)) return solve (z[1], z[2], x);
    return solve (z, ZERO, x);
  }

  //----- Solving a system of equations ---------------------------------------

  static vector<S>
  replace (const vector<S>& v, const table<S,S>& t) {
    nat n= N(v);
    vector<S> r= fill<S > (n);
    for (nat i=0; i<n; i++)
      r[i]= replace (v[i], t);
    return r;
  }

  static table<S,S>
  replace (const table<S,S>& a, const table<S,S>& t) {
    table<S,S> r (ZERO);
    for (iterator<S > it= entries (a); busy (it); ++it)
      r[*it]= replace (a[*it], t);
    return r;
  }

  static void
  find_focus (const vector<S>& eqs, const vector<S>& vars,
	      nat& eq_i, nat& var_i)
  {
    for (eq_i=0; eq_i<N(eqs); eq_i++)
      for (var_i=0; var_i<N(vars); var_i++)
	if (occurrences (eqs[eq_i], vars[var_i]) == 1)
	  return;
    for (eq_i=0; eq_i<N(eqs); eq_i++)
      for (var_i=0; var_i<N(vars); var_i++)
	if (occurrences (eqs[eq_i], vars[var_i]) >= 1)
	  return;
    eq_i= 0;
    var_i= N(vars);
  }

  static vector<S>
  solve (const vector<S>& l, const vector<S>& r) {
    //cout << "Solve " << l << " in " << r << "\n";
    if (N(r) == 0) return r;
    table<S,S> t (ZERO);
    vector<S> eqs = l;
    vector<S> vars= r;
    sort_leq<size_leq> (eqs);
    while (N(eqs) != 0) {
      //cout << "Solving " << eqs << ", " << vars << "\n";
      nat eq_i, var_i;
      find_focus (eqs, vars, eq_i, var_i);
      //cout << "Solve " << eq_i << ", " << var_i << "\n";
      S eq = eqs [eq_i ];
      S var= (var_i < N(vars)? vars[var_i]: UNINIT);
      S sol= solve (eq, var);
      //cout << "Solved " << var << " := " << sol << "\n";
      if (sol == FAIL) return fill<S> (FAIL, N(r));
      eqs= append (range (eqs, 0, eq_i), range (eqs, eq_i+1, N(eqs)));
      if (var_i < N(vars)) {
	t [var]= sol;
	eqs= replace (eqs, t);
	t  = replace (t  , t);
	vars= append (range (vars, 0, var_i), range (vars, var_i+1, N(vars)));
      }
    }

    vector<S> ret;
    for (nat i=0; i<N(r); i++)
      ret << (contains (t, r[i])? t[r[i]]: r[i]);
    //cout << "Solutions " << ret << "\n";
    return ret;
  }
};

} // namespace mmx
#endif // __SYMBOLIC_SOLVE__HPP
