
/******************************************************************************
* MODULE     : symbolic_substitute.hpp
* DESCRIPTION: Syntactic routines on symbolic expressions
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __SYMBOLIC_SUBSTITUTE__HPP
#define __SYMBOLIC_SUBSTITUTE__HPP
#include <symbolix/symbolic.hpp>
#include <symbolix/wrap_sum.hpp>
#include <symbolix/wrap_product.hpp>
#include <symbolix/wrap_derivative.hpp>
#include <symbolix/wrap_function.hpp>
#include <symbolix/wrap_partial.hpp>
#include <basix/symbol.hpp>
#include <basix/memoize.hpp>
namespace mmx {

template<typename W> nat
size_cached (const symbolic<W>& x) {
  typedef symbolic<W> S;
  MEMOIZE_UNARY (std_memoizer,nat,S,size_cached,x,size(x));
}

template<typename W> nat
occurrences_cached (const symbolic<W>& x,
		    const table<bool,symbolic<W> >& t) {
  typedef symbolic<W> S;
  typedef table<bool,S> Set;
  MEMOIZE_BINARY (std_memoizer,nat,S,Set,occurrences_cached,x,t,
		  occurrences(x,t));
}

template<typename W> symbolic<W>
replace_cached (const symbolic<W>& x,
		const table<symbolic<W>,symbolic<W> >& t) {
  typedef symbolic<W> S;
  typedef table<S,S> Table;
  MEMOIZE_BINARY (std_memoizer,S,S,Table,replace_cached,x,t,replace(x,t));
}

template<typename V>
struct symbolic_substitute: public V {
  typedef typename V::S S;
  typedef typename V::W W;
  typedef typename V::compound compound;
  typedef typename V::sum sum;
  typedef typename V::product product;
  typedef typename V::derivative derivative;
  typedef typename V::function function;
  using V::ONE;

  //----- expression size -----------------------------------------------------

  static nat
  size (const S& x) {
    switch (symbolic_type (x)) {
    case SYMBOLIC_INTEGER:
    case SYMBOLIC_RATIONAL:
    case SYMBOLIC_FLOATING:
    case SYMBOLIC_BALL:
    case SYMBOLIC_REAL:
    case SYMBOLIC_LITERAL:
      return 1;
    case SYMBOLIC_COMPOUND: {
      nat sz= 0;
      vector<S> v= *as<compound> (x);
      for (nat i=0; i<N(v); i++)
	sz += size (v[i]);
      return sz;
    }
    case SYMBOLIC_SUM: {
      nat sz= 1;
      sum s= as<sum> (x);
      for (nat i=0; i<N(s); i++) {
	if (co (s, i) != ONE) sz += size (co (s, i));
	if (mo (s, i) != ONE) sz += size_cached (mo (s, i));
      }
      return sz;
    }
    case SYMBOLIC_PRODUCT: {
      nat sz= 1;
      product p= as<product> (x);
      for (nat i=0; i<N(p); i++) {
	if (co (p, i) != ONE) sz += size (co (p, i));
	sz += size_cached (mo (p, i));
      }
      return sz;
    }
    case SYMBOLIC_DERIVATIVE: {
      derivative df= as<derivative> (x);
      return size_cached (root (df)) + N (orders (df)) + 1;
    }
    case SYMBOLIC_FUNCTION: {
      function f= as<function> (x);
      nat sz= size_cached (body (f)) + 1;
      vector<S> a= arguments (f);
      for (nat i=0; i<N(a); i++)
	sz += size_cached (a[i]);
      return sz;
    }
    default:
      return 0;
    }
  }

  //----- dependency counting -------------------------------------------------

  static nat
  occurrences (const S& x, const table<bool,S>& t) {
    nat nr= 0;
    if (contains (t, x)) nr= 1;
    switch (symbolic_type (x)) {
    case SYMBOLIC_INTEGER:
    case SYMBOLIC_RATIONAL:
    case SYMBOLIC_FLOATING:
    case SYMBOLIC_BALL:
    case SYMBOLIC_REAL:
    case SYMBOLIC_LITERAL:
      return nr;
    case SYMBOLIC_COMPOUND: {
      vector<S> v= *as<compound> (x);
      for (nat i=0; i<N(v); i++)
	nr += occurrences (v[i], t);
      return nr;
    }
    case SYMBOLIC_SUM: {
      sum s= as<sum> (x);
      for (nat i=0; i<N(s); i++) {
	nr += occurrences (co (s, i), t);
	nr += occurrences_cached (mo (s, i), t);
      }
      return nr;
    }
    case SYMBOLIC_PRODUCT: {
      product p= as<product> (x);
      for (nat i=0; i<N(p); i++) {
	nr += occurrences (co (p, i), t);
	nr += occurrences_cached (mo (p, i), t);
      }
      return nr;
    }
    case SYMBOLIC_DERIVATIVE: {
      derivative df= as<derivative> (x);
      return occurrences_cached (root (df), t);
    }
    case SYMBOLIC_FUNCTION: {
      function f= as<function> (x);
      vector<S> a= arguments (f);
      table<bool,S> u= copy (t);
      for (nat i=0; i<N(a); i++) reset (u, a[i]);
      return occurrences_cached (body (f), u);
    }
    default:
      return nr;
    }
  }

  static inline nat
  occurrences (const S& x, const S& y) {
    table<bool,S> t (false);
    t[y]= true;
    return occurrences (x, t);
  }

  static inline nat
  occurrences (const S& x, const vector<S>& v) {
    table<bool,S> t (false);
    for (nat i=0; i<N(v); i++)
      t[v[i]]= true;
    return occurrences (x, t);
  }

  static inline bool
  depends (const S& x, const table<bool,S>& s) {
    return occurrences (x, s) > 0;
  }

  static inline bool
  depends (const S& x, const S& y) {
    return occurrences (x, y) > 0;
  }

  static inline bool
  depends (const S& x, const vector<S>& v) {
    return occurrences (x, v) > 0;
  }

  //----- substitution --------------------------------------------------------
  
  static S
  replace (const S& x, const table<S,S>& t) {
    //cout << "Replace " << x << ", " << v << "\n";
    switch (symbolic_type (x)) {
    case SYMBOLIC_INTEGER:
    case SYMBOLIC_RATIONAL:
    case SYMBOLIC_FLOATING:
    case SYMBOLIC_BALL:
    case SYMBOLIC_REAL:
      return x;
    case SYMBOLIC_LITERAL:
      if (contains (t, x)) return t[x];
      return x;
    case SYMBOLIC_COMPOUND: {
      if (contains (t, x)) return t[x];
      vector<S> v= *as<compound> (x);
      nat n= N(v);
      vector<S> r= fill<S> (n);
      for (nat i=0; i<n; i++)
	r[i]= replace (v[i], t);
      return W::apply (r);
    }
    case SYMBOLIC_SUM: {
      if (contains (t, x)) return t[x];
      sum s= as<sum> (x);
      nat n= N(s);
      vector<S> v= fill<S> (n);
      for (nat i=0; i<n; i++)
	v[i]= co (s, i) * replace_cached (mo (s, i), t);
      return big_add (v);
    }
    case SYMBOLIC_PRODUCT: {
      if (contains (t, x)) return t[x];
      product p= as<product> (x);
      nat n= N(p);
      vector<S> v= fill<S> (n);
      for (nat i=0; i<n; i++)
	v[i]= pow (replace_cached (mo (p, i), t), replace (co (p, i), t));
      return big_mul (v);
    }
    case SYMBOLIC_DERIVATIVE: {
      derivative df= as<derivative> (x);
      S fun= replace_cached (root (df), t);
      if (fun == root (df)) return x;
      vector<nat> o= orders (df);
      for (nat i=0; i<N(o); i++)
	for (nat j=0; j<o[i]; j++)
	  fun= derive (fun, wrap_partial<W> (S (integer (i))));
      return fun;
    }
    case SYMBOLIC_FUNCTION: {
      function f= as<function> (x);
      vector<S> a= arguments (f);
      table<S,S> u= copy (t);
      for (nat i=0; i<N(a); i++) reset (u, a[i]);
      return function (a, replace_cached (body (f), u));
    }
    default:
      return x;
    }
  }

  static S
  replace (const S& x, const S& w, const S& b) {
    table<S,S> t;
    t[w]= b;
    return replace (x, t);
  }

  static S
  replace (const S& x, const vector<S>& w, const vector<S>& b) {
    ASSERT (N(w) == N(b), "vectors of unequal lengths");
    table<S,S> t;
    for (nat i=0; i<N(w); i++)
      t[w[i]]= b[i];
    return replace (x, t);
  }
};

} // namespace mmx
#endif // __SYMBOLIC_SUBSTITUTE__HPP
