
/******************************************************************************
* MODULE     : symbolic_sum.hpp
* DESCRIPTION: Linear combinations of symbolic expressions
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __SYMBOLIC_SUM__HPP
#define __SYMBOLIC_SUM__HPP
#include <symbolix/symbolic.hpp>
#include <symbolix/wrap_sum.hpp>
#include <symbolix/wrap_product.hpp>
namespace mmx {

template<typename V>
struct symbolic_sum: public V {
  typedef typename V::S S;
  typedef typename V::W W;
  typedef typename V::sum sum;
  typedef typename V::product product;
  using V::expression;
  using V::is_scalar;
  using V::ZERO;
  using V::ONE;

  //----- Terms ---------------------------------------------------------------

  static inline bool is_term (const S& x) {
    return is<sum> (x) && N (as<sum> (x)) == 1; }
  static inline S term (const S& m, const S& c) {
    if (m == ONE) return c;
    if (c == ONE) return m;
    return sum (m, c); }
  static inline S get_coefficient (const S& x) {
    if (is<sum> (x)) return co (as<sum> (x), 0);
    else return ONE; }
  static inline S get_monomial (const S& x) {
    if (is<sum> (x)) {
      sum s= as<sum> (x);
      if (N(s) == 1) return mo (s, 0);
      else return sum_to_symbolic ((*s) / co (s, 0));
    }
    else return x; }

  //----- Conversion between sums and symbolic expressions --------------------
  
  static sum
  symbolic_to_sum (const S& x) {
    if (is<sum> (x)) return as<sum> (x);
    if (is_scalar (x)) {
      if (x == ZERO) return sum ();
      return sum (ONE, x);
    }
    return sum (x, ONE);
  }

  static S
  sum_to_symbolic (const sum& s) {
    if (N (s) == 0) return ZERO;
    if (N (s) == 1) {
      if (mo (s, 0) == ONE) return co (s, 0);
      if (co (s, 0) == ONE) return mo (s, 0);
    }
    return s;
  }

  //----- Addition and subtraction --------------------------------------------

  static S neg (const S& x1) {
    if (is_scalar (x1)) return V::neg (x1);
    return sum_to_symbolic (-symbolic_to_sum (x1)); }
  static S add (const S& x1, const S& x2) {
    if (is_scalar (x1) && is_scalar (x2)) return V::add (x1, x2);
    return sum_to_symbolic (symbolic_to_sum (x1) + symbolic_to_sum (x2)); }
  static S sub (const S& x1, const S& x2) {
    if (is_scalar (x1) && is_scalar (x2)) return V::sub (x1, x2);
    return sum_to_symbolic (symbolic_to_sum (x1) - symbolic_to_sum (x2)); }

  static S
  add (const vector<S>& v) {
    switch (N(v)) {
    case 0: return ZERO;
    case 1: return v[0];
    case 2: return add (v[0], v[1]);
    case 3: return add (v[0], add (v[1], v[2]));
    case 4: return add (add (v[0], v[1]), add (v[2], v[3]));
    default: return add (add (range (v, 0, N(v)>>1)),
			 add (range (v, N(v)>>1, N(v))));
    }
  }

  //----- Flattening sums -----------------------------------------------------

  static syntactic
  expression (const sum& s) {
    nat n= N(s);
    vector<syntactic> v= fill<syntactic> (n);
    for (nat i=0; i<n; i++) {
      syntactic c= flatten (co (s, i));
      syntactic m= flatten (mo (s, i));
      bool old= set_frac_flag (true);
      v[i]= c * m;
      (void) set_frac_flag (old);
    }
    return ordered_sum (v);
  }
};

} // namespace mmx
#endif // __SYMBOLIC_SUM__HPP
